// ------------------------------------------------------------------
//  $Id: application_base.h 1348 2013-08-16 00:45:07Z secanell $
//
//  Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat.
//  Copyright (C) 2012 by Valentin N. Zingan, University of Alberta.
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
// ------------------------------------------------------------------

#ifndef _DEALII_APPFRAME_APPLICATION_BASE_H_
#define _DEALII_APPFRAME_APPLICATION_BASE_H_

#include <base/parameter_handler.h>

#include <boost/shared_ptr.hpp>

#include <appframe/fe_vectors.h>
#include <appframe/application_data.h>
#include <appframe/event.h>

using namespace dealii;

namespace AppFrame
{

/**
 * The vector class used by applications.
 */
typedef BlockVector<double> FEVector;

/**
 * Base class for applications. \ref get_application()
 *
 * <h3>Use of applications</h3>
 *
 * <h3>Application chains</h3>
 *
 * The idea of these application classes is the possibility to build a
 * chain out of these in order to have several predefined nested
 * solvers. For this, we distinguish applications roughly in two
 * classes, both derived from ApplicationBase:
 *
 * <ol>
 * <li> Terminal applications, which implement the real finite element
 * code like computing residuals on mesh cells, assembling matrices
 * and solving linear systems.
 *
 * <li>Non-terminal applications derived from ApplicationCopy; these
 * usually implement a new solve() function as an iterative solver
 * around another application. They implement all functions of the
 * ApplicationBase interface, either forwarding them to the next inner
 * application by ApplicationCopy or by providing their own
 * implementation.
 * </ol>
 *
 * Usually an class in a chain communicates values with its next outer
 * class through function arguments. Nevertheless, at least the
 * terminal application will require values from even outer
 * applications in the chain in order to compute residuals and
 * matrices correctly. For these, the mechanism of named data
 * provided by ApplicationData was introduced. Each class can store
 * auxiliary data under a unique name there for use in inner
 * iterations.
 *
 * @note This class declares a lot of virtual functions, which will
 * issue a warning to deallog if not overloaded. It does not provide
 * implementations.
 *
 * @author Guido Kanschat, 2006
 */

// ------------------------------

class ApplicationBase : public Subscriptor
{
public:

  /**
   * Constructor for an
   * application. One task of the
   * constructor is declaring
   * parameters for an
   * application. Since this is
   * done consecutively for derived
   * classes, all parameters may be
   * declared in a modular
   * way. Derived classes should
   * create unique subsections in
   * the ParameterHandler.
   *
   * @note Actually, this
   * constructor does nothing,
   * since the base class declares
   * no parameters.
   */
  ApplicationBase(boost::shared_ptr<ApplicationData> ex_data = boost::shared_ptr<ApplicationData>());

  /**
   * Copy constructor.
   * Generate a new
   * application
   * sharing the same
   * ApplicationData with
   * another one.
   */
  ApplicationBase(const ApplicationBase& other);

  /**
   * Virtual destructor.
   */
  virtual ~ApplicationBase() { }

  /**
   * Declare parameters for a
   * parameter file.
   *
   * Pure virtual function
   * MUST be redeclared in derived
   * classes.
   */
  virtual void declare_parameters(ParameterHandler& param) = 0; // No implementation here

  /**
   * Print default parameters for the application to a file.
   * Note that this member function MUST be called after declare_parameters.
   * For example:
   * \code
   * ApplicationBase app;
   * ParamterHandler param;
   * app.declare_parameters(param);
   * app.print_parameters_to_file(param, "default_parameters.txt", ParameterHandler::XML);
   * \endcode
   */
  void print_parameters_to_file(ParameterHandler& param,
                                const std::string& file_name,
                                const ParameterHandler::OutputStyle& style);

  /**
   * Pure virtual initialization
   * function. Here, the parameters
   * declared in the constructor
   * are actually read from
   * ParameterHandler.
   *
   * This function must
   * be overloaded in any derived
   * class using its own values
   * from the parameter file.
   *
   * Initialization functions of
   * derived classes must make sure
   * to call all functions
   * _initialize() of the base classes.
   */
  virtual void initialize(ParameterHandler& param) = 0; // No implementation here

  /**
   * Generates the next mesh
   * depending on the mesh
   * generation parameters.
   */
  virtual void remesh();

  /**
   * Initialize vector to problem
   * size.
   */
  virtual void init_vector(FEVector& dst) const;

  /**
   * Initialize vector to problem
   * size.
   */
  virtual void start_vector(FEVector& dst,
                            std::string caller) const;

  /**
   * Compute residual of
   * <tt>src</tt> and store it
   * into <tt>dst</tt>. The bool variable is used
   * to specify if boundary conditions should be applied
   * using FilteredMatrix.
   */
  virtual double residual(FEVector& dst,
                          const FEVectors& src,
                          bool apply_boundaries = true);

  /**
   * Solve the system assembled
   * with right hand side
   * <tt>rhs</tt> and return the
   * result in <tt>start</tt>.
   */
  virtual void solve(FEVector& start,
                     const FEVectors& rhs);

  /**
   * Solve the dual system
   * assembled with right hand
   * side <tt>rhs</tt> and return
   * the result in <tt>start</tt>.
   */
  virtual void Tsolve(FEVector& start,
                      const FEVectors& rhs);

  /**
   * Error estimation.
   */
  virtual double estimate(const FEVectors& src);

  /**
   * Evaluate whatever the
   * simulation was made for.
   */
  virtual double evaluate(const FEVectors& src);

  /**
   * Write the mesh in the format
   * specified by the ParameterHandler.
   */
  virtual void grid_out(const std::string& filename) const;

  /**
   * Write data in the format
   * specified by the ParameterHandler.
   */
  virtual void data_out(const std::string& filename,
                        const FEVectors& src);
  
  virtual void data_out(const std::string& filename,
                        const FEVectors& src,
								const std::vector<std::string> );
  
  /**
   * Get access to the protected variable #data
   */
  boost::shared_ptr<ApplicationData> get_data()             { return data; }

  /**
   * Get read-only access to the protected variable #data
   */
  const boost::shared_ptr<ApplicationData> get_data() const { return data; }

  /**
   * Return a unique
   * identification string for
   * this application. In particular, it returns
   * the name of the class in which the object you requested
   * id() from belongs to. For example:
   * cBase* a = new cBase
   * cBase* b = new cDerived
   * then,
   * a->id() will return cBase and
   * b->id() will return cDerived
   */
  virtual std::string id() const           { return std::string( typeid(*this).name() ); }

  /**
   * Add a reason for
   * assembling. The new reason
   * is binary or'd with the
   * previous ones.
   */
  virtual void notify(const Event& reason) { notifications += reason; }

  /**
   * Reset the application class
   * such that a call to
   * initialize() is possible
   * again and will produce the
   * same result as if the object
   * was fresh.
   */
  virtual void clear()                     { notifications.all(); }

  /**
   * Clear all #notifications.
   */
  void clear_events()                      { notifications.clear(); }

protected:

  /**
   * Print caller name.
   */
  void print_caller_name(const std::string& caller_name) const;

  /**
   * Object for auxiliary
   * data. This object is
   * always valid, since it is
   * either handed to the
   * constructor or created
   * there.
   */
  boost::shared_ptr<ApplicationData> data;

  /**
   * Accumulate reasons for
   * reassembling here. If any of
   * those is set, the function
   * solve() of a terminal
   * application must take care
   * of reassembling the matrix.
   */
  Event notifications;
};

// ------------------------------

inline
ApplicationBase::ApplicationBase(boost::shared_ptr<ApplicationData> ex_data)
:
Subscriptor(),
data(ex_data)
{
  deallog << "  Application";
  notifications.all();

  if(!data) //if(data == NULL)
    data = boost::shared_ptr<ApplicationData>(new ApplicationData);
}

inline
ApplicationBase::ApplicationBase(const ApplicationBase& other)
:
Subscriptor(),
data(other.data)
{
  deallog << "Base->";
  notifications.all();
}

inline
void
ApplicationBase::print_parameters_to_file(ParameterHandler& param,
                                          const std::string& file_name,
                                          const ParameterHandler::OutputStyle& style)
{
  std::ofstream file_out;
  file_out.open(file_name.c_str());
  deallog << "Printing parameter handler" << std::endl;
  param.print_parameters(file_out, style);
  file_out.close();
}

inline
void
ApplicationBase::remesh()
{
  print_caller_name(__FUNCTION__);
}

inline
void
ApplicationBase::init_vector(FEVector&) const
{
  print_caller_name(__FUNCTION__);
}

inline
void
ApplicationBase::start_vector(FEVector& dst,
                              std::string) const
{
  print_caller_name(__FUNCTION__);
  init_vector(dst);
}

inline
double
ApplicationBase::residual(FEVector&,
                          const FEVectors&,
                          bool)
{
  print_caller_name(__FUNCTION__);
  return 0.;
}

inline
void
ApplicationBase::solve(FEVector&,
                       const FEVectors&)
{
  print_caller_name(__FUNCTION__);
}

inline
void
ApplicationBase::Tsolve(FEVector&,
                        const FEVectors&)
{
  print_caller_name(__FUNCTION__);
}

inline
double
ApplicationBase::estimate(const FEVectors&)
{
  print_caller_name(__FUNCTION__);
  return 0.;
}

inline
double
ApplicationBase::evaluate(const FEVectors&)
{
  print_caller_name(__FUNCTION__);
  return 0.;
}

inline
void
ApplicationBase::grid_out(const std::string&) const
{
  print_caller_name(__FUNCTION__);
}

inline
void
ApplicationBase::data_out(const std::string&,
                          const FEVectors&)
{
  print_caller_name(__FUNCTION__);
}

inline
void
ApplicationBase::data_out(const std::string&,
                          const FEVectors&,
								  const std::vector<std::string> )
{
  print_caller_name(__FUNCTION__);
}

inline
void
ApplicationBase::print_caller_name(const std::string& caller_name) const
{
  const std::type_info& info = typeid(*this);
  deallog << "Pure function " << caller_name << " called in Class " << info.name() << std::endl;
}

}

#endif