// ------------------------------------------------------------------
//  $Id: application_data.h 1348 2013-08-16 00:45:07Z secanell $
//
//  Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat.
//  Copyright (C) 2012 by Valentin N. Zingan, University of Alberta.
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
// ------------------------------------------------------------------

#ifndef _DEALII_APPFRAME_APPLICATION_DATA_H_
#define _DEALII_APPFRAME_APPLICATION_DATA_H_

#include <lac/vector_memory.h>
#include <lac/block_vector.h>

#include <algorithm>
#include <fstream>

using namespace dealii;

namespace AppFrame
{

/**
 * The vector class used by applications.
 */
typedef BlockVector<double> FEVector;

/**
 * Here we handle general data of applications. It is the idea of this
 * class that all applications inheriting from ApplicationBase share
 * the <b>same</b> ApplicationData object, thus accessing the same
 * vector pools and using the object as a means for communicating
 * global information.
 *
 * First purpose is providing vector memory objects for Vector and
 * BlockVector objects for applications. These objects, #vector_pool
 * and #block_vector_pool can be accessed directly.
 *
 * Furthermore, it provides a map which allows applications to
 * exchange data in named registers. To this end, the functions
 * enter(std::string,const double&) and enter(std::string,const FEVector&)
 * can be used to enter scalar and
 * vector data, respectively. These data are entered as references, so
 * the controlling application can change them at any time. All other
 * applications sharing the same ApplicationData can access them read
 * only through the functions scalar() and vector().
 */

class ApplicationData
{
public:

  /**
   * Destructor, releasing all data.
   */
 ~ApplicationData();

  /**
   * Register a named scalar.
   */
  void enter(std::string name,
             const double& s);

  /**
   * Register a named vector.
   */
  void enter(std::string name,
             const FEVector& v);

  /**
   * Delete a previously registered
   * scalar.
   */
  void erase_scalar(std::string name);

  /**
   * Delete a previously registered
   * vector.
   */
  void erase_vector(std::string name);

  /**
   * Get read-only access to a
   * registered scalar.
   * It only offers read access and
   * returns a null pointer, if
   * the name has not been
   * registered.
   */
  const double* scalar(std::string name) const;

  /**
   * Get read-only access to a
   * registered vector.
   * It only offers read access and
   * returns a null pointer, if
   * the name has not been
   * registered.
   */
  const FEVector* vector(std::string name) const;

  /**
   * List all stored objects to deallog.
   */
  void log() const;

  /**
   * VectorMemory object for
   * simple vectors. All
   * applications should allocate
   * their vectors here, so they
   * can be reused and operating
   * system memory management can
   * be avoided.
   */
  GrowingVectorMemory< Vector<double> > vector_pool;

  /**
   * VectorMemory object for
   * block vectors. All
   * applications should allocate
   * their vectors here, so they
   * can be reused and operating
   * system memory management can
   * be avoided.
   */
  GrowingVectorMemory<FEVector> block_vector_pool;

  /**
   * The typedef for the map of
   * scalars.
   */
  typedef std::map< std::string, const double* > scalar_map;

  /**
   * The typedef for the map of
   * vectors.
   */
  typedef std::map< std::string, const FEVector* > vector_map;

  /**
   * Exception thrown when a
   * named scalar or vector was
   * searched but not found.
   */
  DeclException2(ExcNotFound,
                 char*,
                 std::string,
                 << "A " << arg1 << " with name " << arg2 << " was not stored in this data object");

private:

  /**
   * A map linking names of data
   * to actual scalars.
   */
  scalar_map named_scalars;

  /**
   * A map linking names of data
   * to actual vectors.
   */
  vector_map named_vectors;
};

inline
ApplicationData::~ApplicationData()
{ }

inline
void
ApplicationData::enter(std::string name,
                       const double& s)
{
  named_scalars.insert(std::make_pair(name, &s));
}

inline
void
ApplicationData::enter(std::string name,
                       const FEVector& v)
{
  named_vectors.insert(std::make_pair(name, &v));
}

inline
void
ApplicationData::erase_scalar(std::string name)
{
  scalar_map::iterator iter = named_scalars.find(name);
  Assert(iter != named_scalars.end(), ExcNotInitialized());
  named_scalars.erase(iter);
}

inline
void
ApplicationData::erase_vector(std::string name)
{
  vector_map::iterator iter = named_vectors.find(name);
  Assert(iter != named_vectors.end(), ExcNotInitialized());
  named_vectors.erase(iter);
}

inline
const double*
ApplicationData::scalar(std::string name) const
{
  scalar_map::const_iterator iter = named_scalars.find(name);
  AssertThrow(iter != named_scalars.end(), ExcNotFound("scalar", name));
  return iter->second;
}

inline
const FEVector*
ApplicationData::vector(std::string name) const
{
  vector_map::const_iterator iter = named_vectors.find(name);
  AssertThrow(iter != named_vectors.end(), ExcNotFound("vector", name));
  return iter->second;
}

inline
void
ApplicationData::log() const
{
  deallog.push("scalar");
  for(scalar_map::const_iterator iter = named_scalars.begin(); iter != named_scalars.end(); ++iter)
    deallog << iter->first << '\t' << iter->second << std::endl;
  deallog.pop();

  deallog.push("vector");
  for(vector_map::const_iterator iter = named_vectors.begin(); iter != named_vectors.end(); ++iter)
    deallog << iter->first << '\t' << iter->second->size() << std::endl;
  deallog.pop();
}

}

#endif