// ------------------------------------------------------------------
//  $Id: base.h 1348 2013-08-16 00:45:07Z secanell $
//
//  Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat.
//  Copyright (C) 2012 by Valentin N. Zingan, University of Alberta.
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
// ------------------------------------------------------------------

#ifndef _DEALII_APPFRAME_BASE_H_
#define _DEALII_APPFRAME_BASE_H_

#include <appframe/fe_vectors.h>
#include <appframe/application_data.h>
#include <appframe/event.h>
#include <appframe/application_base.h>
#include <appframe/application_copy.h>

using namespace dealii;

/**
 * \brief Functions and classes of the application framework.
 *
 * Application classes in this framework are derived from
 * ApplicationBase, where the interface a driver program may use is
 * defined.
 *
 * The only driver program available right now is the function
 * mesh_loop(), which repeats discretizing and solving an application
 * on a sequence of meshes.
 *
 * ???
 *
 * Therefore, application programs may be very short. Take for example
 * <tt>test/newton.cc</tt>:
 *
 * \dontinclude newton.cc
 * \skipline include
 * \until }
 *
 * First, we include the files containing the definition of the Newton
 * application and of the <tt>SquareApplication</tt> used by Newton.
 *
 * Then, we implement the required function get_application(), first
 * creating an instance of <tt>SquareApplication</tt> and initializing
 * Newton's method with it. It returns a pointer to the last method in
 * the chain, which is Newton.
 *
 * @author Guido Kanschat, 2006
 */

namespace AppFrame
{

/**
 * Open a logfile. The name of the
 * logfile can be given as an
 * optional <tt>char*</tt>; a
 * suffix of this name will be
 * removed and the suffix
 * <tt>.dlg</tt> will be added. If
 * no name is given, the name of
 * the file will be the host name.
 *
 * After opening the file,
 * <tt>deallog</tt> will
 * automatically log into the file.
 */
  extern void open_logfile(const std::string& filename = std::string(""));

/**
 * Declare some additional global
 * parameters and try to read
 * parameter files.
 */
  extern void declare_parameter_files(ParameterHandler& param);

/**
 * Read some additional global
 * parameters and try to read
 * parameter files.
 *
 * This function attempts to read
 * <tt>common.prm</tt>,
 * <tt>output.prm</tt> and
 * <tt>filename.prm</tt> from the
 * path for parameter files set in
 * PathSearch.
 */
  extern void read_parameter_files(ParameterHandler& param,
                                   const std::string& filename = std::string(""));

/**
 * General initialization of the
 * AppFrame in order to simplify
 * main functions. Call this
 * function first of all in your
 * program.
 */
  extern void initialize(int argc,
                         char** argv);

/**
 * Copy of the number of parameters
 * for main, set by initialize().
 */
  extern int argc;

/**
 * Copy of the parameters
 * for main, set by initialize().
 */
  extern char** argv;

}

#endif