//---------------------------------------------------------------------------
//    $Id: block_matrix_application.h 1415 2013-08-30 16:14:41Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//    Copyright (C) 2007-13 by Marc Secanell
//
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__appframe__block_matrix_application_h
#define __deal2__appframe__block_matrix_application_h

#include <base/quadrature.h>
#include <lac/full_matrix.h>
#include <lac/block_sparse_matrix.h>
#include <lac/block_sparsity_pattern.h>
#include <lac/filtered_matrix.h>
#include <lac/vector.h>
#include <grid/tria.h>
#include <grid/tria_iterator.h>
#include <grid/tria_accessor.h>
#include <dofs/dof_tools.h>
#include <numerics/matrix_tools.h>

#include <appframe/dof_application.h>
#include <appframe/matrix_base.h>
#include <boost/shared_ptr.hpp>


namespace AppFrame
{
    /**
     * Application handling matrices and assembling linear systems of
     * equations.
     *
     * This class inherits the information on Triangulation and DoFHandler
     * from its base class DoFApplication. It adds information on the
     * linear system, in particular the #sparsities.
     *
     * Additionally, it holds a BlockSparseMatrix as the system matrix of
     * the finite element problem. Generic loops over mesh cells and faces
     * are implemented in order to assemble() the system matrices and
     * compute the residual().  These functions use the virtual local
     * integration functions declared here and implemented by the derived
     * actual application class.
     *
     * @note You probably miss functions for building the right hand side
     * of a linear system. They are missing deliberately, since the linear
     * case is very special and can be handled in the framework of
     * nonlinear problems. So, if you have a linear problem, code the
     * local matrices for residual() so they ignore the BlockVector
     * <tt>src</tt>.
     *
     * <h3>Deriving your own application</h3>
     *
     * 
     * @author Guido Kanschat, 2006
     */
    template <int dim>
    class BlockMatrixApplication :
    public DoFApplication<dim>
    {
    public:
        ///@name Constructors and Initialization
        //@{
        /**
         * Constructor for an object,
         * owning its own mesh and dof handler.
         *
         * If <tt>data</tt> is null,
         * a new handler for
         * application data is
         * generated.
         *
         * see constructor of DoFApplication.
         */
        BlockMatrixApplication(boost::shared_ptr<ApplicationData> data =
        boost::shared_ptr<ApplicationData>());
        
        /**
         * Constructor for an object,
         * borrowing mesh and dof
         * handler from another
         * object. If
         * <tt>triangulation_only</tt>
         * is true, only the
         * triangulation is borrowed.
         *
         * see constructor of DoFApplication.
         */
        BlockMatrixApplication(DoFApplication<dim>&,
                               bool triangulation_only);
        
        /**
         * Initialize data of this
         * class.
         * @note remesh_dofs() has to be called before this member function
         * since remesh_dofs() initializes data that is used in this routine. 
         * In particular, dof and block_indices.
         */
        void _initialize(ParameterHandler& param);
        
        /**
         * Declare parameters related
         * to matrices.
         *
         * @note The function
         * declare_parameters of
         * derived applications should
         * always call the one of its
         * base class.
         */
        virtual void declare_parameters(ParameterHandler& param);
        /**
         * Initialize application
         */
        virtual void initialize (ParameterHandler& param);  
        //@}
        ///@name Data management functions
        //@{
        /**
         * Add a named vector from
         * #data to be handed to the
         * cell_matrix() function via
         * CellInfo::values and
         * CellInfo::derivatives. 
         */
        
        void add_vector_for_cell_matrix(std::string name,
                                        unsigned int block,
                                        unsigned int components,
                                        bool values,
                                        bool derivatives);
        
        /**
         * Add a named vector from
         * #data to be handed to the
         * bdry_matrix() and
         * face_matrix() functions via
         * FaceInfo::values and
         * FaceInfo::derivatives.
         */
        void add_vector_for_flux_matrix(std::string name,
                                        unsigned int block,
                                        unsigned int components,
                                        bool values,
                                        bool derivatives);
        //@}
        /**
         * Multithreaded assemble
         * function called by
         * assemble().
         */
        void do_assemble(
            const FEVectors&,
            typename DoFHandler<dim>::active_cell_iterator begin,
            typename DoFHandler<dim>::active_cell_iterator end,
            Threads::Mutex& mutex);
        
        /**
         * A call back function for
         * assemble(), called after all
         * cell matrices have been
         * entered, but before the face
         * matrices are computed.
         */
        virtual void post_cell_assemble();
        
        ///@name ApplicationBase interface
        //@{			                                             
        /**
         * Initialize sparsity patterns
         * and matrices for the new
         * mesh.
         */
        void remesh_matrices();
        /**
         * Refine grid accordingly to the Refinement options under "Grid Generation"
         * in the parameter file.
         */
        virtual void remesh();
        
        /**
         * Loop over all cells and
         * assemble the system
         * #matrices by using the local
         * matrix integration functions
         * cell_matrix(), bdry_matrix()
         * and face_matrix() provided
         * by the derived class.
         */
        void assemble(const FEVectors&);
        
        
        /**
         * Compute the Jacobian of the system of equations, 
         * \f[
         * J(i,j) = \frac{\partial R_i}{\partial u_j}
         * \f]
         * by using forward differences. The solution is perturbed at each degree of freedom 
         * and the residual function is used to compute the residual for each perturbed solution.
         * Forward differences are used to compute the Jacobian.
         * 
         * The Jacobian is stored in this->matrix and a version with the boundary conditions already
         * applied is stored in this->filtered_matrix
         * 
         * @param[in] src A vector of BlockVectors containting the solution at the previous
         *  Newton step and the residual.
         * @param[in] delta The step size used for numerical differenciation.
         * 
         * \note This routine is used in combination with a Netwon nonlinear solver since it
         * requires src to contain a field named "Newton iterate" with the solution at the
         * previous Newton step.
         * 
         * \author M. Secanell, 2013
         */
        void assemble_numerically(const FEVectors& src,
                                  const double delta = 1e-6);
        /**
         * Redefinition of residual_constraints() in
         * DoFHandler. In this member function, I adjust the residual
         * in order to account for hanging nodes and I apply 
         * Dirichlet boundary conditions given by
         * dirichlet_bc().
         */
        
        void residual_constraints(FEVector& dst) const;
        
        //@}
        /// @name Local integrators
        //@{
        /**
         * Integration of local
         * bilinear form. Here we loop over the quadrature
         * points and over degrees of freedom in order to compute
         * the matrix for the cell.
         */
        virtual void cell_matrix(MatrixVector& cell_matrices,
                                 const typename DoFApplication<dim>::CellInfo& cell);
        
        /**
         * Integration of local
         * bilinear form.
         */
        virtual void bdry_matrix(MatrixVector& face_matrices,
                                 const typename DoFApplication<dim>::FaceInfo& face);
        
        /**
         * Integration of local
         * bilinear form.
         */
        virtual void face_matrix(MatrixVector& matrices11,
                                 MatrixVector& matrices12,
                                 MatrixVector& matrices21,
                                 MatrixVector& matrices22,
                                 const typename DoFApplication<dim>::FaceInfo& face1,
                                 const typename DoFApplication<dim>::FaceInfo& face2);
        
        /**
         * Member function used to set dirichlet boundary conditions.
         * This function is application specific and it only computes the boundary_value
         * values that are used to constraint the linear system of equations that is being
         * solved
         */
        //@}
        /// @name Dirichlet boundary conditions
        //@{
        virtual void dirichlet_bc(std::map<unsigned int, double>& boundary_values) const;
        //@}
    
    protected:
        /**
         * Couplings through cell bilinear forms.
         * cell_coupling allows to specify which variables
         * couple with which equations. This is used 
         * by DoFTools to generate a sparsity pattern
         * that does not contain elements unless specified in the
         * cell_coupings, therefore, reducing the amount of memory
         * needed.
         * 
         * @note cell_coupings is a two dimensional table (i.e. a matrix) of 
         * DoFTools::Coupling objects. DoFTools::Coulings takes the values: 
         * always (components couple), zero (no coupling) and nonzero
         *
         * Example: For the 2D Stokes equations
         * \f{eqnarray*}
         * \frac{d^2u}{dx^2} +\frac{d^2u}{dy^2} + \frac{dp}{dx} = 0 \\
         * \frac{d^2v}{dx^2} +\frac{d^2v}{dy^2} + \frac{dp}{dy} = 0 \\
         * \frac{du}{dx} +\frac{dv}{dy} = 0
         *					\f}
         * we need to initialize cell_coupling to:
         * @verbatim
         * cell_coupling.reinit(3,3);
         * cell_coupling(0,0) = DoFTools::always;
         * cell_coupling(0,1) = DoFTools::zero;
         * cell_couplings(0,2) = DoFTools::always;
         * cell_couplings(1,0) = DoFTools::zero; 
         * ...
         * @endverbatim
         */
        Table<2, DoFTools::Coupling> cell_couplings;
        
        /**
         * Couplings through flux
         * bilinear forms
         */
        Table<2, DoFTools::Coupling> flux_couplings;
        
        /**
         * Quadrature rule for matrix
         * assembling on cells.
         */
        boost::shared_ptr<Quadrature<dim> > quadrature_assemble_cell;
        
        /**
         * Quadrature rule for matrix
         * assembling on faces.
         */
        boost::shared_ptr<Quadrature<dim-1> > quadrature_assemble_face;

    private:
        /**
             * Sparsity patterns.
             */
            BlockSparsityPattern sparsities;      
            
    protected:
        /**
         * Storage for the actual
         * matrix.
         */
        BlockSparseMatrix<double> matrix;
        /**
         * Wrapper for the actual matrix that contains all the 
         * boundary conditions.
         * This is the matrix that should be used to solve
         * the linear system of equations and NOT matrix.
         */
        FilteredMatrix<BlockVector<double> > filtered_matrix;
                
    private:
        /**
         * The set of all vector data to be used in cell_matrix().
         */
        MeshWorker::VectorSelector cell_matrix_values;
        /**
         * The set of all vector data
         * to be used in
         * bdry_matrix() and
         * face_matrix().
         */
        MeshWorker::VectorSelector flux_matrix_values;    
   
    };
}
                    
#endif
