//---------------------------------------------------------------------------
//    $Id: dof_application.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__appframe__dof_application_h
#define __deal2__appframe__dof_application_h

#include <appframe/base.h>
#include <appframe/residual.h>
#include <appframe/mesh_worker.h>
#include <boost/shared_ptr.hpp>
#include <lac/block_indices.h>
#include <grid/grid_out.h>
#include <grid/grid_tools.h>
#include <lac/constraint_matrix.h>
#include <fe/fe_values.h>
#include <numerics/data_out.h>
#include <numerics/solution_transfer.h>

namespace dealii
{
    template <int, int> class Triangulation;
    template <int, int> class Mapping;
    template <int, int> class FiniteElement;
    template <int, int> class DoFHandler;
}


namespace AppFrame
{
    template<int dim> class DoFApplication;
    
    /**
     * The structure indicating which blocks of which stored vectors
     * should be handed over to the local integration functions for
     * residuals and matrices.
     *
     * @author Guido Kanschat, 2007
     */
    struct VectorSelector
    {
        /**
         * The name of the vector
         * as stored in
         * ApplicationData
         */
        std::string name;
        /**
         * The block number for
         * which values should be
         * computed
         */
        unsigned int block;
        /**
         * The number of components
         * the result should
         * contain. In order for
         * this to work properly,
         * this should be divisible
         * by the block size of the
         * element.
         */
        unsigned int components;
        
        /**
         * Which values should be
         * computed. May be a
         * combination of
         * <tt>update_values</tt>
         * and <tt>update_gradients</tt>
         */
        dealii::UpdateFlags flags;
    };
    
    /**
     * The class using CellInfo and FaceInfo to compute local contributions to a 1form.
     *
     * Objects of this class are handed to the function
     * DoFApplication::integrate_1form(). Derived classes for instance
     * compute local contributions to the residual.
     *
     * In addition to the local integration functions cell(), bdry() and
     * face(), objects of this class provide control information to the
     * DoFApplication loop:
     *
     * <ul>
     *
     * <li> #cell_quadrature, #bdry_quadrature and #face_quadrature
     * determine the quadrature rules used on cells, boundary and interior
     * faces, respectively. A rule with zero points means that the loop
     * over these entities should be skipped.
     *
     * </ul>
     *
     * @author Guido Kanschat, 2007
     */
    template <int dim>
    class LocalIntegrator
    {
    public:
        typedef typename MeshWorker::InfoObjects::IntegrationInfo<dim, FEValuesBase<dim, dim> > CellInfo;
        typedef typename MeshWorker::InfoObjects::IntegrationInfo<dim, FEFaceValuesBase<dim, dim> > FaceInfo;
        
        /** Virtual constructor */
        LocalIntegrator();
        /** Virtual destructor */
        virtual ~LocalIntegrator();
        
        /** Integration of local cell form. */
        virtual void cell(FEVector& cell_vector, const CellInfo& cell) = 0;
        
        /// Integration of local boundary form.
        virtual void bdry(FEVector& face_vector, const FaceInfo& face) = 0;
        
        /// Integration of local interior face form.
        virtual void face(FEVector& face_vector1, FEVector& face_vector2,
                          const FaceInfo& face1, const FaceInfo& face2) = 0;
                          
        /// An optional function called after the integration loop
        virtual void post_loop(FEVector& dst, const FEVectors& srcs);
                          
        /// The quadrature rule used on cells.
        Quadrature<dim> cell_quadrature;
                          
        /// The quadrature rule used on boundary faces.
        Quadrature<dim-1> bdry_quadrature;
        
        /// The quadrature rule used on interior faces.
        Quadrature<dim-1> face_quadrature;
        
        /** Assign n-point Gauss
         * quadratures to each of the
         * quadrature rules. Here, a
         * size of zero points means
         * that no loop over these grid
         * entities should be
         * performed.
         */
        void set_gauss_quadrature(unsigned int n_cell_points,
                                  unsigned int n_bdry_points,
                                  unsigned int n_face_points);
    };
    
    /**
     * @deprecated This class is a shortcut, but should not be needed in
     * future applications.
     *
     * The Integrator computing the residual.
     *
     * This class implements the residual, using the functions
     * DoFApplication::cell_residual(), DoFApplication::bdry_residual()
     * and DoFApplication::face_residual().
     *
     * @author Guido Kanschat, 2007
     */
    template <int dim>
    class LocalResidual : public MeshWorker::Assembler::ResidualLocalBlocksToGlobalBlocks<double>,
    public MeshWorker::WorkerObjects::IntegrationWorker<dim>
    {
    private:
        SmartPointer<DoFApplication<dim> > app;
    public:
        /// Constructor
        void initialize(DoFApplication<dim>* a,
                        const MeshWorker::BlockInfo& bi,
                        FEVectors& m)
        {
            app = a;
            MeshWorker::Assembler::ResidualLocalBlocksToGlobalBlocks<double>::initialize(bi, m);
        }
        
        /**
         * Do the work on a cell.
         */
        void cell(const typename DoFApplication<dim>::CellInfo& cell);
        
        /**
         * Do the work on a boundary face.
         */
        void bdry(const typename DoFApplication<dim>::FaceInfo& face);
        
        /**
         * Do the work on an interior face.
         */
        void face(const typename DoFApplication<dim>::FaceInfo& face1,
                  const typename DoFApplication<dim>::FaceInfo& face2);
    };
    
    /**
     * @deprecated This class is a shortcut, but should not be needed in
     * future applications.
     *
     * The LocalIntegrator computing an error estimate based on jumps of
     * the normal derivative.
     *
     * This class implements the estimate, using the functions
     * DoFApplication::cell_estimate(), DoFApplication::bdry_estimate()
     * and DoFApplication::face_estimate().
     *
     * @author Guido Kanschat, 2007
     */
    template <int dim>
    class LocalEstimate : public MeshWorker::Assembler::CellsAndFaces<double>,
    public MeshWorker::WorkerObjects::IntegrationWorker<dim>
    {
    private:
        SmartPointer<DoFApplication<dim> > app;
    public:
        /// Constructor
        void initialize(DoFApplication<dim>* a,
                        const MeshWorker::BlockInfo&,
                        FEVectors& m)
        {
            app = a;
            MeshWorker::Assembler::CellsAndFaces<double>::initialize(m, true);
        }
        
        /**
         * Do the work on a cell.
         */
        void cell(const typename DoFApplication<dim>::CellInfo& cell);
        
        /**
         * Do the work on a boundary face.
         */
        void bdry(const typename DoFApplication<dim>::FaceInfo& face);
        
        /**
         * Do the work on an interior face.
         */
        void face(const typename DoFApplication<dim>::FaceInfo& face1,
                  const typename DoFApplication<dim>::FaceInfo& face2);
    };
    
//     /**
//      * Integrate the residual using
//      * DoFApplication::integrate_1form(). Optionally, this class can use
//      * its own LocalIntegrator or use the standard residual of the
//      * DoFApplication.
//      *
//      * @author Guido Kanschat, 2008
//      */
//     template <int dim>
//     class DoFResidual : public Residual
//     {
//     public:
//         /**
//          * Constructor initializing
//          * this object to compute the
//          * residual using
//          * <tt>application</tt> and the
//          * provided <tt>local_residual</tt>.
//          */
//         DoFResidual(const DoFApplication<dim>& application,
//                     const LocalIntegrator<dim>& local_residual);
//         
//         virtual double operator() (FEVector& dst, const FEVectors& src) const;
//         
//     private:
//         /**
//          * The DoFApplication class
//          * providing the function
//          * DoFApplication::integrate_1form().
//          */
//         SmartPointer<const DoFApplication<dim> > app;
//         
//         /**
//          * The object integrating the
//          * local residuals.
//          */
//         LocalIntegrator<dim> local_residual;
//     };
//     
  
    /**
     * Base class for all terminal applications, i.e. all applications requiring a Triangulation and a
     * handler for degrees of freedom.
     *
     * The mesh as well as the dof handler may be created by this class,
     * which is the default, or they may be provided by another object, in
     * which case they must be specified in the constructor.
     *
     * Note that in this class the received or created dof handler is associated
     * to the finite element given by the argument "Element" on the parameter
     * file. Therefore, this class in not responsible to generate the system of equations to
     * be solved, only to initialize the dof handler
     * "Element" can either be a single element of a FESystem. In the latter case, the
     * nomenclature used in the paramter file is:
     s et Element = FESys*tem[element1_type(element1_degree)^number_of_elements1-...-elementN_type(elementN_degree)^number_of_elementsN]
     * Example: set Element = FESystem[FE_DGQ(0)-FE_Q(1)^2]
     *
     * @author Guido Kanschat, 2006
     */
    template <int dim>
    class DoFApplication
    :
    public ApplicationBase
    {
    public:
        /** This is a redefinition of CellInfo in order to call MeshWorker objects instead of the previous implementation
         *		 in AppFrame
         */
        typedef typename MeshWorker::InfoObjects::IntegrationInfo<dim, FEValuesBase<dim> > CellInfo;
        /** This is a redefinition of CellInfo in order to call MeshWorker objects instead of the previous implementation
         *		 in AppFrame
         */
        typedef typename MeshWorker::InfoObjects::IntegrationInfo<dim, FEFaceValuesBase<dim> > FaceInfo;
        /**
         * Constructor for an object,
         * owning its own mesh and dof handler.
         *
         * If <tt>data</tt> is null, a new handler for
         * application data is generated.
         */
        DoFApplication(boost::shared_ptr<ApplicationData> data = boost::shared_ptr<ApplicationData>(),
                       bool multigrid = false);
        
        /**
         * Constructor for an object owning its own mesh and dof
         * handler and creating new ApplicationData.
         */
        DoFApplication(bool multigrid);
        
        /**
         * Constructor for an object, borrowing mesh and dof
         * handler from another object. If <tt>triangulation_only</tt>
         * is true, only the triangulation is borrowed.
         *
         * @note A pointer to <tt>DH2</tt> must be
         * convertible to a pointer to <tt>DH</tt>.
         */
        DoFApplication(DoFApplication<dim>&,
                       bool triangulation_only,
                       bool multigrid = false);
        
        /**
         * Destructor which deletes
         * owned objects.
         */
        ~DoFApplication();
        
        /**
         * Declare parameters related to mesh generation and
         * finite elements. Most importantly, here we declare "Element".
         * "Element" is declared under the subsection "Discretization":
         * "Element" is the element that will be associated
         * to the dof handler (by default FE_Q(1)). This can either be a single element of a FESystem.
         * In the latter case, the
         * nomenclature used in the paramter file is:
         * set Element = FESystem[element1_type(element1_degree)^number_of_elements1-
         * ...-elementN_type(elementN_degree)^number_of_elementsN]
         * Example: set Element = FESystem[FE_DGQ(0)-FE_Q(1)^2]
         *
         * @note The function
         * declare_parameters of
         * derived applications should
         * always call the one of its
         * base class.
         */
        virtual void declare_parameters(ParameterHandler& param);
        
        /**
         * Initialize from parameter
         * values.
         */
        void _initialize (ParameterHandler& param);
        
        /**
         * Add the vector to be transfered from one
         * mesh to the next.
         */
        void add_vector_for_transfer(FEVector*);
        /**
         * Delete the vector to be
         * transfered from one
         * mesh to the next.
         */
        void delete_vector_for_transfer();
        /**
         * Initialize dof handler, count the dofs in each block
         * and renumber the dofs. Different numbering schemes can be easily
         * implemented by reimplementing this function
         * in a derived class and do the sorting after having
         * called this function. Remember though to
         * always sort by component in the end.
         */
        virtual void remesh_dofs();
        
        virtual void initialize (ParameterHandler& param);
        /**
         * Reinitialize the BlockVector dst such that it contains
         * block_size.size() blocks. Each block is
         * reinitialized to dimensions block_sizes[i]. Note that
         * block_sizes contains the number of degrees of freedom
         * per block. So, init_vector could be used to
         * reinitialize the solution vector.  
         * 
         * @note: init_vector can only be called after
         * remesh_dofs has been called otherwise block_sizes will
         * not have been initialized.
         */
        virtual void init_vector (FEVector& dst) const;
        
        /**
         * Refine grid accordingly to the Refinement options under "Grid Generation"
         * in the parameter file
         */
        virtual void remesh();
        /**
         * Evaluate a functional during postprocessing such
         * as drag or lift on an aerodynamics problem
         */
        virtual double evaluate(const FEVectors&);
        /**
         * Estimate the error. This is the loop over all cells and
         * faces, using the virtual local functions
         * - cell_residual(),
         * - face_residual() and
         * - bdry_residual().  
         * 
         * While these are usually implemented for a specific
         * problem, a default implementation measuring
         * discontinuity of the normal derivative (and, if
         * #interior_fluxes or #boundary_fluxes are true)
         * the function itself.
         */
        virtual double estimate(const FEVectors& src);
        
        /**
         * Loop over all cells to compute the residual. Uses
         * the virtual local functions
         * - cell_residual(),
         * - face_residual() and
         * - bdry_residual() 
         * to be implemented by derived classes.
         *
         * This function uses residual_adjust_boundary()
         * right before computing the norm of the residual as
         * return value.
         */
        virtual double residual(FEVector& dst, const FEVectors& src, bool apply_boundaries = true);
        
        /**
         * Apply boundary conditions and hanging node constraints
         * to a residual vector after it has been computed by
         * residual(). This function is used in residual() and its
         * default implementation is void.
         *
         * @note Reimplement this function if you want to
         * constrain the residual. A typical implementation would be
         * <pre>
         * hanging_node_constraints.condense(dst);
         * constrain_boundary(v, true);
         * </pre>
         */
        virtual void residual_constraints(FEVector& dst) const;
        
        ///@name Data output
        //@{
        
        /**
         * Output the grid used to solve the problem.
         */
        virtual void grid_out(const std::string& basename);
        
        /**
         * \deprecated This function is deprecated.
         */
        virtual void data_out(const std::string& basename,
                              const FEVectors&   src);
        
        /**
         * \deprecated This function is deprecated.
         */
        virtual void data_out(const std::string&              basename,
                              const FEVectors&                src,
                              const std::vector<std::string>& solution_names);
        
        /**
         * This function outputs
         * the results of a computation.
         *
         * Parameters:
         *
         * - \p basename is a base of an output file name,
         *
         * - \p solution       is the global solution computed in DoFs,
         * - \p solution_names is the component names of the \p solution,
         *
         * - \p postprocessing       is what you compute in DoFs (or in cells in the case of piece-wise constant functionals) after the \p solution has been computed,
         * - \p postprocessing_names is the component names of the \p postprocessing.
         */
        virtual void data_out(const std::string&              basename,
                              
                              const FEVector&                 solution,
                              const std::vector<std::string>& solution_names,
                              
                              const FEVector&                 postprocessing       = FEVector(),
                              const std::vector<std::string>& postprocessing_names = std::vector<std::string>());
        
        /**
         * \p solution_interpretations identifies whether some \p solution_names are scalars or parts of a vector.
         * Note that if \p solution_interpretations is empty, all \p solution_names are treated as scalars.
         */
        std::vector< DataComponentInterpretation::DataComponentInterpretation > solution_interpretations;
        
        /**
         * \p postprocessing_interpretations identifies whether some \p postprocessing_names are scalars or parts of a vector.
         * Note that if \p postprocessing_interpretations is empty, all \p postprocessing_names are treated as scalars.
         */
        std::vector< DataComponentInterpretation::DataComponentInterpretation > postprocessing_interpretations;
        
        /**
         * \p output_materials_and_levels if \p true then visualized, otherwise suppressed.
         * Initialized as \p true in the constructors.
         */
        bool output_materials_and_levels;
        
        /**
         * Vector that will be used by \p data_out to store the
         * material ids.
         */
        Vector<double> output_materials;
        
        /**
         * Vector that will be used by \p data_out to store the
         * refinement levels at each cell.
         */
        Vector<double> output_levels;
        
        //@}
               
        /**
         * Compute the amount of memory
         * needed by this object.
         */
        unsigned int memory_consumption() const;
        
        /**
         * Integration of estimate
         * inside a cell.
         *
         * The implementation in this
         *class simply returns zero.
         */
        virtual double cell_estimate(const CellInfo&);
        
        /**
         * Integration of estimate on a boundary face.
         *
         * The implementation in this class simply returns zero.
         */
        virtual double bdry_estimate(const FaceInfo&);
        
        /**
         * Integration of estimate on an interior face
         *
         * The implementation in this class simply returns the
         * jump of the normal derivative. If #interior_fluxes is set,
         * then the jump of the function is added with
         * appropriate weight.
         */
        virtual double face_estimate(const FaceInfo&,
                                     const FaceInfo&);
        
        /**
         * Integration of local linear form.
         */
        virtual void cell_residual(FEVector& cell_vector,
                                   const CellInfo& cell);
        
        /**
         * Integration of local linear form.
         */
        virtual void bdry_residual(FEVector& face_vector,
                                   const FaceInfo& face);
        
        /**
         * Integration of local linear form.
         */
        virtual void face_residual(FEVector& face_vector1,
                                   FEVector& face_vector2,
                                   const FaceInfo& face1,
                                   const FaceInfo& face2);
        /**
         * The function integrating 1-forms as for instance used
         * by residual(). Since the concept of a 1-form is
         * simply taking a set of source vectors and local integrators to produce a set
         * of output vectors, we abstract this from the actual residual function
         * here.
         *
         * The function takes an argument of type LocalIntegrator, which is usually a derived
         * class. LocalResidual itself will produce the residual by applying the virtual functions
         * - cell_residual(),
         * - bdry_residual() and
         * - face_residual().
         */
        template<class BOX, class WORKER>
        void integrate_1form(BOX& box, WORKER& local_forms, FEVector& dst, const FEVectors& src);
        
        /**
         * The function integrating functionals using local
         * integrators. The result consists of two vectors
         * containing the local contributions to the
         * functionals of the cells and faces, respectively.
         *
         * The destination argument is expected to have two
         * FEVector objects at its front, named "cells" and         
         * "faces" for cell and face contributions, respectively.
         *
         * It is possible to compute several functionals at a
         * time in which case these two FEVector objects should be
         * block vectors with one block for each functional. The         
         * destination vector for the LocalIntegrator will be
         * adjusted to the correct size.
         *
         * The function takes an argument of type
         * LocalIntegrator, which is usually a derived class. 
         * LocalEstimate itself will produce the residual by applying the
         * virtual functions
         * - cell_estimate(),
         * - bdry_estimate() and
         * - face_estimate().
         *
         * @note The function LocalIntegrator::face() receives the same result
         * vector twice. Therefore, only one of them should be filled.
         */
        void integrate_functional(LocalEstimate<dim>& local_forms,
                                  FEVectors& dst,
                                  const FEVectors& src);
        
        /**
         * Function to copy a triangulation
         * object for use after refinement
         */
        void store_triangulation(Triangulation<dim>& new_tr)
        {
            new_tr.copy_triangulation(*this->tr);
        }
        
        /**
         * Function to perform the transfer of a solution on a refined grid to the initial coarse grid
         */
        void transfer_solution_to_coarse_mesh(Triangulation<dim>& tr_coarse,
                                              AppFrame::FEVector& coarse_solution,
                                              AppFrame::FEVector& refined_solution);
        
        /**
         * Function to transfer a solution on a refined grid to the initial coarse grid
         */
        void read_init_solution(AppFrame::FEVector& dst, bool &good_solution) const;
        
        /**
         * Controls verbosity of certain functions. Zero is
         * quiet and default is one.
         */
        unsigned int verbosity;
        
    protected:
        /**
         * Apply either homogeneous or inhomogeneous
         * #boundary_constraints.
         */
        void constrain_boundary(FEVector& v, bool homogeneous) const;
        
        /**
         * List of all nodes constrained by a strong
         * boundary condition, together with a value to be assigned.
         *
         * This map must be filled by the application and is used
         * in constrain_boundary().
         */
        std::map<unsigned int, double> boundary_constraints;
        
        /**
         * Constraint Matrix object. This object contains a list of the
         * constraints from the hanging nodes.
         * It needs to be used to remove hanging nodes
         * from the system matrix and rhs before the
         * system of equations is solved using
         * @verbatim
         * hanging_node_constraints.condense(system_matrix);
         * hanging_node_constraints.condense(system_rhs);
         * @endverbatim
         * and it needs to be used to add the hanging nodes to the solution
         * once the system is solved using
         * @verbatim
         * hanging_node_constraints.distribute(solution);
         * @endverbatim
         */
        ConstraintMatrix hanging_node_constraints; 
        
        /**
         * The mapping used for the
         * transformation of mesh
         * cells.
         *
         * @note Should be set by
         * initialize() in a derived
         * class before _initialize()
         * is called.
         */
        boost::shared_ptr<Mapping<dim> > mapping;
        
        /**
         * Degree used for polynomial mapping of arbitrary order.
         * 
         * By mapping we mean the transformation between the unit cell (i.e. the unit line, square, or cube) to the cells in real space.
         * Before we have implicitly used linear or d-linear mappings and have not noticed this at all, since this is what happens if we do not do anything special.
         * However, if the domain has curved boundaries, there are cases where the piecewise linear approximation of the boundary (i.e. by straight line segments) is not sufficient,
         * and we may want that the computational domain is an approximation to the real domain using curved boundaries as well.
         *
         * If the boundary approximation uses piecewise quadratic parabolas to approximate the true boundary,
         * then we say that this is a quadratic or \f$ Q_2 \f$ approximation.
         * If we use piecewise graphs of cubic polynomials, then this is a \f$ Q_3 \f$ approximation, and so on.
         * 
         * For some differential equations, it is known that piecewise linear approximations of the boundary, i.e. \f$ Q_1 \f$ mappings, are not sufficient
         * if the boundary of the domain is curved.
         * Examples are the biharmonic equation using \f$ C_1 \f$ elements,
         * or the Euler equation on domains with curved reflective boundaries.
         * In these cases, it is necessary to compute the integrals using a higher order mapping.
         * The reason, of course, is that if we do not use a higher order mapping,
         * the order of approximation of the boundary dominates the order of convergence of the entire numerical scheme,
         * irrespective of the order of convergence of the discretization in the interior of the domain.
         */      
        unsigned int mapping_degree;    
        
        /**
         * Pointer to the Triangulation
         * object.
         */
        boost::shared_ptr<Triangulation<dim> > tr;
        
        /**
         * The finite element used in
         * #dof. This can be a single
         * element or a FESystem
         */
        boost::shared_ptr<FiniteElement<dim> > element;
        
        /**
         * Pointer to the DoFHandler
         * object.
         */
        boost::shared_ptr<DoFHandler<dim> > dof;
        
        /**
         * Pointer to the MGDoFHandler
         * object.
         */
        boost::shared_ptr<MGDoFHandler<dim> > mg_dof;
        
        MeshWorker::BlockInfo block_info;
        
        /**
         * The result of error estimation by cell. This
         * variable will be used for adaptive mesh refinement.
         */
        //      Vector<double> cell_errors;
        Vector<float> cell_errors;
        /**
         * The result of error estimation by face.
         */
        Vector<float> face_errors;
        
        /**
         * The object for writing grids.
         */
        GridOut g_out;
        
        /**
         * The object for writing data.
         */
        DataOut<dim, DoFHandler<dim> > d_out;
        
        /**
         * Vector for adjusting output to vector data.
         */
        std::vector<DataComponentInterpretation::DataComponentInterpretation>
        data_interpretation;
        
        /**
         * Refinement parameter from parameter file.
         */
        std::string refinement;
        
        /**
         * Initial refinement from parameter file
         */
        unsigned int initial_refinement;
        
        /**
         * Refinement threshold for adaptive method from parameter
         * file.
         */
        double refinement_threshold;
        
        /**
         * Coarsening threshold for
         * adaptive method from parameter
         * file.
         */
        double coarsening_threshold;
        
        /**
         * Flag for sorting with Cuthill McKee algorithm.
         *
         * @note Read from parameter file
         */
        bool sort_cuthill;
        
        /**
         * Direction for downstream sorting. No downstream
         * sorting if this vector is zero.
         *
         * @note Read from parameter file
         */
        Point<dim> sort_direction;
        
        /**
         * List of vector names to be
         * transfered from one grid to
         * the next.
         */
        std::vector<FEVector*> transfer_vectors;
        /**
         * Quadrature rule for residual
         * computation on cells.
         */
        Quadrature<dim> quadrature_residual_cell;
        
        /**
         * Quadrature rule for residual
         * computation on boundary faces.
         */
        Quadrature<dim-1> quadrature_residual_bdry;
        
        /**
         * Quadrature rule for residual
         * computation on faces.
         */
        Quadrature<dim-1> quadrature_residual_face;
        
        /**
         * Extend the integration loops in assemble() and residual()
         * also to boundary faces.
         */
        bool boundary_fluxes;
        
        /**
         * Extend the integration loops in assemble() and residual()
         * also to interior faces.
         */
        bool interior_fluxes;
        
    public:
        
        /**
         * Fill local data vector with values of the stored finite
         * element function vectors.
         */
        template <class INFO, typename TYPE>
        void fill_local_data(const INFO& info,
                             const std::vector<VectorSelector>& data_vector,
                             std::vector<std::vector<std::vector<TYPE> > >& data) const;
                             
    private:
        /**
         * Sort the degrees of freedom. In a derived class,
         * sorting order and schemes can be changed by
         * overloading this function.
         */
        virtual void sort_dofs();
        
        /**
         * After computing the error contributions on faces and
         * cells, this function adds half of the contribution
         * computed by face_estimate() and the contribution
         * computed by bdry_estimate() to the adjacent cell. It
         * deletes #face_errors in the end.
         *
         * This function can be overridden by derived
         * classes to handle both estimates separately.
         */
        virtual void distribute_face_to_cell_errors();
        
        /**
         * Compute a global estimation value from #cell_errors. In
         * order for this to be useful, #face_errors should be
         * transfered to cells first, for instance by
         * distribute_face_to_cell_errors().
         *
         * This function computes the sum of the #cell_errors and
         * takes the square root in the end.  Reimplementation in
         * derived classes allows for different norms.
         *
         * The function is called by estimate() to compute its
         * return value.
         */
        virtual double global_from_local_errors() const;
        friend class LocalResidual<dim>;
        friend class LocalEstimate<dim>;
    };   
}

#endif