//---------------------------------------------------------------------------
//    $Id: matrix_application.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__appframe__matrix_application_h
#define __deal2__appframe__matrix_application_h

#include <base/quadrature.h>
#include <base/thread_management.h>
#include <base/multithread_info.h>
#include <lac/full_matrix.h>
#include <lac/sparse_matrix.h>
#include <lac/block_sparsity_pattern.h>
#include <grid/tria.h>
#include <grid/tria_iterator.h>
#include <grid/tria_accessor.h>
#include <dofs/dof_tools.h>
#include <base/mg_level_object.h>
#include <appframe/dof_application.h>
#include <appframe/matrix_base.h>
#include <boost/shared_ptr.hpp>

namespace AppFrame
{
/**
 * Application handling matrices and assembling linear systems of
 * equations.
 *
 * This class inherits the information on Triangulation and DoFHandler
 * from its base class DoFApplication. It adds information on the
 * linear system, in particular the #sparsity.
 *
 * Furthermore, it contains a vector of matrix blocks. This vector
 * should be filled with void matrices and suitable block coordinates
 * by the initialization routine of a derived class. After that,
 * remesh_matrices() will reinitialize these matrices with appropriate
 * sparsity patterns after each remesh().
 *
 * Finally, generic loops over mesh cells and faces are implemented in
 * order to assemble() the system matrices and compute the residual().
 * These functions use the virtual local integration functions
 * declared here and implemented by the derived actual application
 * class.
 *
 * @note You probably miss functions for building the right hand side
 * of a linear system. They are missing deliberately, since the linear
 * case is very special and can be handled in the framework of
 * nonlinear problems. So, if you have a linear problem, code the
 * local matrices for residual() so they ignore the BlockVector
 * <tt>src</tt>.
 *
 * <h3>Deriving your own application</h3>
 *
 * Writing your own application derived from MatrixApplication
 * requires setting up some data either in the constructor of your
 * application or in its _initialize() function. Furthermore, some
 * virtual functions should be implemented. Here's the list of things
 * to do:
 *
 * <h4>Providing information to MatrixApplication</h4>
 *
 * MatrixApplication automizes the process of computing residuals and
 * matrices. In order to do this efficiently, it requires some data
 * structures to be set up in the beginning.
 *
 * @note Those of the following data not depending on the parameter
 * file should be set up in the constructor, while your function
 * _initialize() can be used to set up those depending on these
 * parameters.
 *
 * <ol><li>All applications must fill the fields #cell_couplings,
 * #flux_couplings if face integrals enter the formulation. These two
 * objects are essentiel for saving memory and computation time.
 *
 * <li>If multilevel support is required, then also #mg_cell_couplings
 * and #mg_flux_couplings must be filled as well.
 *
 * <li>If the formulation does not include integrals over interior or
 * boundary faces, the variables #boundary_fluxes and #interior_fluxes
 * should be set to <tt>false</tt> to save computation time.
 *
 * <li>You must add matrices that will be filled by assemble() using
 * add_matrix(). If using multigrid, do the same with add_mg_matrix().
 *
 * <li>Finally, for problems depending on a previous finite element
 * solution (e.g. in Newton's method or a timestepping scheme), you
 * must tell MatrixApplication the name of the vector to be filled
 * into the CellInfo and FaceInfo objects. This is done using the
 * functions ???.
 *
 * </ol>
 * 
 * @author Guido Kanschat, 2006, 2007
 */
  template <int dim>
  class MatrixApplication :
      public DoFApplication<dim>
  {
    public:
				       /**
					* @brief The type of matrices stored.
					*/
      typedef MatrixBlock<SparseMatrix<double> > MatrixType;
				       /**
					* @brief The type of multigrid
					* matrices stored.
					*/
      typedef MatrixBlock<MGLevelObject<SparseMatrix<double> > > MGMatrixType;

				       /**
					* The assembler object.
					*/
      typedef MeshWorker::Assembler::MatrixLocalBlocksToGlobalBlocks<SparseMatrix<double> >
      Assembler;
      
				       /**
					* The multigrid assembler object.
					*/
      typedef MeshWorker::Assembler::MGMatrixLocalBlocksToGlobalBlocks<SparseMatrix<double> >
      MGAssembler;
      
///@name Constructors and Initialization
//@{
				       /**
					* Constructor for an object,
					* owning its own mesh and dof handler.
					*
					* If <tt>data</tt> is null,
					* a new handler for
					* application data is
					* generated.
					*
					* see constructor of DoFApplication.
					*/
      MatrixApplication(boost::shared_ptr<ApplicationData> data =
			boost::shared_ptr<ApplicationData>(),
			bool multigrid = false);

				       /**
					* Constructor for an object,
					* owning its own mesh and dof
					* handler and having its own data.
					*/
      MatrixApplication(bool multigrid);
      
				       /**
					* Constructor for an object,
					* borrowing mesh and dof
					* handler from another
					* object. If
					* <tt>triangulation_only</tt>
					* is true, only the
					* triangulation is borrowed.
					*
					* see constructor of DoFApplication.
					*/
      MatrixApplication(DoFApplication<dim>&,
			bool triangulation_only,
			bool multigrid = false);

				       /**
					* Initialize data of this
					* class.
					*
					* @note remesh_dofs() has to
					* be called before this member
					* function since remesh_dofs()
					* initializes data that is
					* used in this routine.  In
					* particular, dof and
					* block_indices.
					*/
      void _initialize(ParameterHandler& param);
      
				       /**
					* Declare parameters related
					* to matrices.
					*
					* @note The function
					* declare_parameters of
					* derived applications should
					* always call the one of its
					* base class.
					*/
      virtual void declare_parameters(ParameterHandler& param);

				       /**
					* The virtual initialize
					* function to be replaced by
					* derived applications
					*/
      virtual void initialize (ParameterHandler& param);
//@}
///@name Data management functions
//@{
      
				       /**
					* Add a new matrix to the list
					* of #matrices. In order to
					* initialize the matrix
					* correctly, MatrixApplication
					* needs to know the row and
					* column this matrix is in.
					*
					* Consecutive calls of
					* add_matrix() generate a
					* sequence of matrices in the
					* same order/
					*/
      void add_matrix(unsigned int row, unsigned int col);

				       /**
					* Add a new multigrid matrix
					* to #mg_matrices. See
					* add_matrix() for details.
					*/
      void add_mg_matrix(unsigned int row, unsigned int col);
      
				       /**
					* Add a named vector from
					* #data to be handed to the
					* cell_matrix() function via
					* CellInfo::values and
					* CellInfo::derivatives. 
					*/
      void add_vector_for_cell_matrix(std::string name,
				      unsigned int block,
				      unsigned int components,
				      bool values,
				      bool derivatives);
				       /**
					* Add a named vector from
					* #data to be handed to the
					* bdry_matrix() and
					* face_matrix() functions via
					* FaceInfo::values and
					* FaceInfo::derivatives.
					*/
      void add_vector_for_flux_matrix(std::string name,
				      unsigned int block,
				      unsigned int components,
				      bool values,
				      bool derivatives);
//@}
      
      				       /**
					* A call back function for
					* assemble(), called after all
					* cell matrices have been
					* entered, but before the face
					* matrices are computed.
					*/
      virtual void post_cell_assemble();

///@name ApplicationBase interface
//@{
      
				       /**
					* Initialize sparsity patterns
					* and matrices for the new
					* mesh.
					*/
      void remesh_matrices();

				       /**
					* Refine grid accordingly to
					* the Refinement options under
					* "Grid Generation" in the
					* parameter file.
					*/
      virtual void remesh();
 
				       /**
					* Loop over all cells and
					* assemble the system
					* #matrices by using the local
					* matrix integration functions
					* cell_matrix(), bdry_matrix()
					* and face_matrix() provided
					* by the derived class.
					*/
      void assemble(const FEVectors&);
//@}
/// @name Local integrators
//@{
      
				       /**
					* Integration of local
					* bilinear form. Here we loop over the quadrature
					* points and over degrees of freedom in order to compute
					* the matrix for the cell.
					*/
      virtual void cell_matrix(MatrixVector& cell_matrices,
			       const typename DoFApplication<dim>::CellInfo& cell);
      
				       /**
					* Integration of local
					* bilinear form.
					*/
      virtual void bdry_matrix(MatrixVector& face_matrices,
			       const  typename DoFApplication<dim>::FaceInfo& face);
      
				       /**
					* Integration of local
					* bilinear form.
					*/
      virtual void face_matrix(MatrixVector& matrices11,
			       MatrixVector& matrices12,
			       MatrixVector& matrices21,
			       MatrixVector& matrices22,
			       const  typename DoFApplication<dim>::FaceInfo& face1,
			       const  typename DoFApplication<dim>::FaceInfo& face2);

				       /**
					* Integration of local
					* bilinear form. Here we loop over the quadrature
					* points and over degrees of freedom in order to compute
					* the matrix for the cell.
					*
					* @note Default implementation calls cell_matrix.
					*/
      virtual void mg_cell_matrix(MatrixVector& cell_matrices,
				  const typename DoFApplication<dim>::CellInfo& cell);
      
				       /**
					* Integration of local
					* bilinear form.
					*
					* @note Default implementation calls bdry_matrix.
					*/
      virtual void mg_bdry_matrix(MatrixVector& face_matrices,
				  const typename DoFApplication<dim>::FaceInfo& face);
      
				       /**
					* Integration of local
					* bilinear form.
					*
					* @note Default implementation calls face_matrix.
					*/
      virtual void mg_face_matrix(MatrixVector& matrices11,
				  MatrixVector& matrices12,
				  MatrixVector& matrices21,
				  MatrixVector& matrices22,
				  const typename DoFApplication<dim>::FaceInfo& face1,
				  const typename DoFApplication<dim>::FaceInfo& face2);
				       //@}
      template <class ASSEMBLER>
      class LocalMatrixIntegrator : public ASSEMBLER,
				    public MeshWorker::WorkerObjects::IntegrationWorker<dim>
      {
	  SmartPointer<MatrixApplication<dim> > app;
	  bool multigrid;
	public:
	  LocalMatrixIntegrator(bool mg = false)
	    : multigrid(mg)
	    {}
	  
	  void initialize(MatrixApplication<dim>* a,
			  const MeshWorker::BlockInfo& bi, std::vector<typename ASSEMBLER::MatrixPtr>& m)
	    {
	      app = a;
	      ASSEMBLER::initialize(bi, m);
	    }
	  
					   /**
					    * Do the work on a cell.
					    */
	  void cell(const typename DoFApplication<dim>::CellInfo& cell)
	    {
	      if (multigrid)
		app->mg_cell_matrix(this->M11, cell);
	      else
		app->cell_matrix(this->M11, cell);
	      assemble(cell);
	    }
	  
	  
					   /**
					    * Do the work on a boundary face.
					    */
	  void bdry(const typename DoFApplication<dim>::FaceInfo& face)
	    {
	      if (multigrid)
		app->mg_bdry_matrix(this->M11, face);
	      else
		app->bdry_matrix(this->M11, face);
	      assemble(face);
	    }
	  
					   /**
					    * Do the work on an interior face.
					    */
	  void face(const typename DoFApplication<dim>::FaceInfo& face1,
		    const typename DoFApplication<dim>::FaceInfo& face2)
	    {
	      if (multigrid)
		app->mg_face_matrix(this->M11, this->M12, this->M21, this->M22, face1, face2);
	      else
		app->face_matrix(this->M11, this->M12, this->M21, this->M22, face1, face2);
	      assemble(face1, face2);
	    }
      };
      
    protected:
				       /**
					* Couplings through cell bilinear forms.
					* cell_coupling allows to specify which variables
					* couple with which equations. This is used 
					* by DoFTools to generate a sparsity pattern
					* that does not contain elements unless specified in the
					* cell_coupings, therefore, reducing the amount of memory
					* needed.
					* 
					* @note cell_coupings is a two dimensional table (i.e. a matrix) of 
					* DoFTools::Coupling objects. DoFTools::Coulings takes the values: 
					* always (components couple), zero (no coupling) and nonzero
					*
					* Example: For the 2D Stokes equations
					* \f{eqnarray*}
					* \frac{d^2u}{dx^2} +\frac{d^2u}{dy^2} + \frac{dp}{dx} = 0 \\
					* \frac{d^2v}{dx^2} +\frac{d^2v}{dy^2} + \frac{dp}{dy} = 0 \\
					* \frac{du}{dx} +\frac{dv}{dy} = 0
					\f}
					* we need to initialize cell_coupling to:
					* @verbatim
					* cell_coupling.reinit(3,3);
					* cell_coupling(0,0) = DoFTools::always;
					* cell_couplings(0,2) = DoFTools::always;
					* cell_coupling(1,1) = DoFTools::always;
					* ...
					* @endverbatim
					*/
      Table<2, DoFTools::Coupling> cell_couplings;
      
				       /**
					* Couplings through flux
					* bilinear forms
					*/
      Table<2, DoFTools::Coupling> flux_couplings;

				       /**
					* Same as #cell_couplings, but
					* for multilevel matrices.
					*/
      Table<2, DoFTools::Coupling> mg_cell_couplings;

				       /**
					* Same as #flux_couplings, but
					* for multilevel matrices.
					*/
      Table<2, DoFTools::Coupling> mg_flux_couplings;

				       /**
					* Quadrature rule for matrix
					* assembling on cells.
					*/
      boost::shared_ptr<Quadrature<dim> > quadrature_assemble_cell;

				       /**
					* Quadrature rule for matrix
					* assembling on faces.
					*/
      boost::shared_ptr<Quadrature<dim-1> > quadrature_assemble_face;
      
  private:
				       /**
					* Sparsity patterns.
					*/
      BlockSparsityPattern sparsity;
				       /**
					* Multi-grid sparsity patterns
					*/
      MGLevelObject<BlockSparsityPattern> mg_sparsity;
      
				       /**
					* Sparsity patterns for
					* #mg_interface_matrices in multigrid.
					*/
      MGLevelObject<BlockSparsityPattern> mg_sparsity_interface;
      
    protected:
				       /**
					* Storage for the actual
					* matrices. This vector will
					* be initialized by the
					* concrete application,
					* knowing how many matrices in
					* each component are required.
					*
					* It is suggested to use
					* add_matrix() instead of
					* initializing this array by
					* hand.
					*/
      std::vector<boost::shared_ptr<MatrixType> > matrices;
      
				       /**
					* Storage for multilevel matrices
					*/
      std::vector<boost::shared_ptr<MGMatrixType> > mg_matrices;

				       /**
					* The matrices at the
					* interface between two
					* refinement levels, coupling
					* coarse to fine.
					*/
      std::vector<boost::shared_ptr<MGMatrixType> > mg_interface_matrices_up;
      
				       /**
					* The matrices at the
					* interface between two
					* refinement levels, coupling
					* fine to coarse.
					*/
      std::vector<boost::shared_ptr<MGMatrixType> > mg_interface_matrices_down;
      
    private:
				       /**
					* The set of all vector data to be used in cell_matrix().
					*/
      MeshWorker::VectorSelector cell_matrix_values;
      
				       /**
					* The set of all vector data
					* to be used in
					* bdry_matrix() and
					* face_matrix().
					*/
      MeshWorker::VectorSelector flux_matrix_values;
  };
  
}

#endif
