//---------------------------------------------------------------------------
//    $Id: matrix_shop_cell.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __appshop__matrix_shop_cell_h
#define __appshop__matrix_shop_cell_h

#include <appframe/app_shop.h>
#include <appframe/matrix_base.h>

/**
 * \file
 * \brief Matrix integration routines on cells for standard bilinear forms.
 */
using namespace AppFrame;

namespace AppShop
{
  namespace Matrix
  {
    namespace Cell
    {
				       /**
					* The mass matrix.
					*
					* \f[
					* (a u,v)
					* \f]
					*/
      template <int dim>
      void mass (FullMatrix<double>& M,
		 const FEValuesBase<dim>& fe,
		 const double a = 1.);
				       /**
					* The mass matrix between
					* different elements.
					*
					* \f[
					* (a u,v)
					* \f]
					*
					* If <tt>component</tt> has
					* a nonnegative value and the
					* element is vector-valued,
					* then only the component
					* corresponding to this number
					* is assembled.
					*/
      template <int dim>
      void mass (FullMatrix<double>& M,
		 const FEValuesBase<dim>& fe,
		 const FEValuesBase<dim>& fetest,
		 const unsigned int component,
		 const double a = 1.);
				       /**
					* Mass matrix with varying coefficient.
					*
					* \f[
					* (a u,v)
					* \f]
					*
					* If <tt>divide</tt> is
					* <tt>true</tt>, then divide
					* by <i>a</i> instead of
					* multiplying.
					*/
      template <int dim>
      void mass_scaled (FullMatrix<double>& M,
			const FEValuesBase<dim>& fe,
			const std::vector<double>& a,
			bool divide);

				       /**
					* Mass matrix with varying coefficient
					* for different elements.
					*
					* \f[
					* (a u,v)
					* \f]
					*
					* If <tt>divide</tt> is
					* <tt>true</tt>, then divide
					* by <i>a</i> instead of
					* multiplying.
					*/
      template <int dim>
      void mass_scaled (FullMatrix<double>& M,
			const FEValuesBase<dim>& fe,
			const FEValuesBase<dim>& fetest,
			const unsigned int select,
			const std::vector<double>& factor,
		        bool divide = false);
      
				       /**
					* Linear advection in divergence
					* form. The derivative is on the
					* test function.
					*
					* \f[
					* -(u \mathbf w, \nabla v)
					* \f]
					*/
      template <int dim>
      void advection(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const std::vector<Point<dim> >& velocity);

				       /**
					* Linear advection in divergence
					* form. The derivative is on the
					* test function.
					*
					* \f[
					* -(u \mathbf w, \nabla v)
					* \f]
					*/
      template <int dim>
      void advection(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const VectorSlice<const std::vector<std::vector<double> > >& velocity,
	double factor = 1.);

/**
 * Cell matrix for divergence. The derivative is on the trial function.
 *
 * \f[
 * (\nabla \cdot u , v)
 * \f]
 *
 * @arg M Result matrix vector
 *
 * @arg indices: the indices in the vector where we find the matrices.
 *
 * @arg fe: finite element space for trial functions.
 * 
 * @arg fetest: finite element space for trial functions.
 */
      template <int dim>
      void div_primal(
	std::vector<MatrixBlock<FullMatrix<double> > >& M,
	const std::vector<unsigned int>& indices,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);

/**
 * Cell matrix for divergence. The derivative is on the test function.
 *
 * \f[
 * -(u , \nabla v)
 * \f]
 *
 * @arg M Result matrix vector
 *
 * @arg indices: the indices in the vector where we find the matrices.
 *
 * @arg fe: finite element space for trial functions.
 * 
 * @arg fetest: finite element space for trial functions.
 */
      template <int dim>
      void div(
	std::vector<MatrixBlock<FullMatrix<double> > >& M,
	const std::vector<unsigned int>& indices,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);

/**
 * Same as div_cell(), but with a possibly varying coefficient.
 *
 * \f[
 * -( \alpha u, \nabla v)
 * \f]
 */
      template <int dim>
      void div_scaled (
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest, 
	const std::vector<double>& factor);

/**
 * Directional derivative in direction <tt>i</tt>.
 *
 * \f[
 * -(u, \partial_i v)
 * \f]
 */
      template <int dim>
      void di(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	unsigned int i);

/**
 * Same as di_cell(), but with varying coefficient.
 *
 * \f[
 * -(\alpha u, \partial_i v)
 * \f]
 */
      template <int dim>
      void di_scaled (
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	unsigned int d,
	std::vector<double>& factor);

/**
 * Discretization for the transpose of the symmetric gradient.
 *
 * \f[
 * -(u, \varepsilon(v))
 * \f]
 */
      template <int dim>
      void eps_div(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest);

/**
 * Laplacian in weak form
 *
 * \f[
 * (\nu \nabla u, \nabla v)
 * \f]
 */
      template <int dim>
      void laplacian(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	double nu = 1.);
/**
 * Laplacian in weak form
 * with a Tensor coefficient to allow
 * anisotropic media
 * \f[
 * (K \nabla u, \nabla v)
 * \f]
 */
      template <int dim>
      void laplacian(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	Tensor<2,dim> nu);

/**
 * Laplacian with varying coefficient. If <tt>divide</tt> is
 * <tt>true</tt>, divide gradient u times gradient v by factor.
 *
 * \f[
 * (\nu \nabla u, \nabla v)
 * \f]
 */
      template <int dim>
      void laplacian_scaled(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const std::vector<double>& nu,
	bool divide = false);

/**
 * Laplacian with constant coefficient. If <tt>divide</tt> is
 * <tt>true</tt>, divide gradient u times gradient v by factor.
 *
 * \f[
 * (\nu \nabla u, \nabla v)
 * \f]
 */
      template <int dim>
      void laplacian_scaled(
	FullMatrix<double>& M,
	const FEValuesBase<dim>& fe,
	const double nu,
	bool divide = false);      
      
/**
 * Second order divergence form for vector valued elements.
 *
 * \f[
 * (\nabla \cdot u, \nabla \cdot v)
 * \f]
 */
      template <int dim>
	void div_div (FullMatrix<double>& M,
		      const FEValuesBase<dim>& fe,
		      const double factor = 1.);

    }
  }
}

#endif
