//---------------------------------------------------------------------------
//    $Id: mesh_loop.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__appframe__mesh_loop_h
#define __deal2__appframe__mesh_loop_h

#include <lac/solver_control.h>
#include <appframe/base.h>
#include <numerics/vector_tools.h>

#include <set>

namespace AppFrame
{
  template <int dim> class DoFApplication;
  
/**
 * Loop over a sequence of meshes and solve the inner application.
 *
 * The function solve of this application solves the innner
 * problem. All other virtual functions except initialize() are handed
 * to the inner application immediately.
 *
 * The iteration is controlled by an object of type SolverControl,
 * which is usually filled from the parameter file.
 *
 * @author Guido Kanschat, 2006
 */
  class MeshLoop :
      public ApplicationCopy
  {
    public:
				       /**
					* The Event raised by MeshLoop
					* if the mesh has changed.
					*/
      static const Event new_mesh;
      
				       /**
					* Constructor, receiving the
					* application solved by
					* solve().
					*/
      MeshLoop(ApplicationBase& app);
				       /**
					* Declare parameters of the
					* application pointed to and
					* ReductionControl parameters
					* for this object..
					*/
      virtual void declare_parameters(ParameterHandler& param);
      virtual void initialize (ParameterHandler& param);      
      virtual void solve(FEVector& start, const FEVectors& rhs);
      virtual void data_out(const std::string& filename, const FEVectors& src);
				       /**
					* Change #base_name, the trunc
					* of the output file names.
					*/
      void set_name(std::string new_name);      
    private:
				       /**
					* A function called right
					* before the loop starts. It
					* allows a derived class to
					* initialize its data at that
					* point.
					*
					* @arg @p v is the iteration
					* vector in solve(), which
					* contains the solution after
					* each step.
					*/
      virtual void setup(FEVector& v);

				       /**
					* The function called to
					* perform the remeshing at the
					* beginning of all but the
					* first iterations.
					*
					* By default, just calls
					* Application::remesh() and
					* Application::start_vector(),
					* which should be replicated
					* by overloaded functions in
					* derived classes.
					*
					* @arg @p v is the solution
					* vector, which is filled by
					* Application::start_vector().
					*/
      virtual void do_remesh(FEVector& v);

				       /**
					* The function called in the
					* iteration right after
					* solving.
					*
					* By default, calls
					* Application::evaluate().
					*/
      virtual void do_evaluate(const FEVectors&);
      
				       /**
					* The object controlling the
					* mesh iteration.
					*/
      ReductionControl control;
				       /**
					* Base name of the file for
					* data output.
					*/
      std::string base_name;
  };

/**
 * A derivate form MeshLoop providing additional functionality which
 * needs information from a DoFApplication.
 *
 * This MeshLoop can compute the difference between two successive
 * solutions automatically.
 *
 * @author Guido Kanschat, 2009
 */
  template <int dim>
  class DoFMeshLoop :
      public MeshLoop
  {
    public:
      
				       /**
					* Constructor, receiving the
					* application solved by
					* solve().
					*/
      DoFMeshLoop(ApplicationBase& app, DoFApplication<dim>& dof_app);
      
      ~DoFMeshLoop();
				       /**
					* Declare parameters of the
					* application pointed to and
					* ReductionControl parameters
					* for this object..
					*/
      virtual void declare_parameters(ParameterHandler& param);      
      virtual void initialize (ParameterHandler& param);
      virtual void data_out(const std::string& filename, const FEVectors& src);
      
    private:
      virtual void setup(FEVector&);
      virtual void do_remesh(FEVector&);
      virtual void do_evaluate(const FEVectors&);
				       /**
					* Pointer to the DoFHandler.
					*/
      SmartPointer<DoFApplication<dim> > dof_app;
      
				       /**
					* The flag controlling,
					* whether the previous
					* solution should be
					* transfered to the new mesh.
					*
					* This function will not only
					* add the solution vector to
					* DoFApplication::transfer_vectors,
					* it also eliminates the call
					* to Application::start_vector
					* after remeshing.
					*/
      bool transfer_solution;

				       /**
					* A flag determining whether
					* the difference of two
					* successive meshes should be
					* provided to
					* Application::evaluate()
					* instead of the last
					* solution.
					*
					* If this is @p true,
					* Application::evaluate()
					* should compute errors with
					* respect to ZeroFunction
					* instead of the true solution.
					*/
      bool intrinsic_error;

				       /**
					* Additionally, output the
					* difference of two conseutive
					* steps.
					*/
      bool output_difference;
      
				       /**
					* Pointer needed to store the
					* old solution.
					*/
      FEVector* old_solution;      
  };
}

#endif
