//---------------------------------------------------------------------------
//    $Id: mesh_worker_assembler.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__mesh_worker_assembler_h
#define __deal2__mesh_worker_assembler_h

namespace MeshWorker
{

/**
 * The namespace containing objects that can be used to assemble data
 * computed on cells and faces into global objects. This can reach
 * from collecting the total error estimate from cell and face
 * contributions to assembling matrices and multilevel matrices.
 *
 * <h3>Data models</h3>
 *
 * The class chosen from this namespace determines which data model is
 * used. For the local as well as the global objects, we have the
 * choice between two models:
 *
 * <h4>The comprehensive data model</h4>
 *
 * This is the structure set up by the FESystem class. Globally, this
 * means, data is assembled into one residual vector and into one
 * matrix. These objects may be block vectors and block matrices, but
 * the process of assembling ignores this fact.
 *
 * Similarly, there is only a single cell vector and cell matrix,
 * respectively, which is indexed by all degrees of freedom of the
 * FESystem. When building the cell matrix, it is necessary to
 * distinguish between the different components of the system and
 * select the right operator for each pair.
 *
 * <h4>The blocked data model</h4>
 *
 * Here, all the blocks are treated separately (in spite of using
 * FESystem for its convenience in other places). For instance, no
 * block matrix is assembled, but a list of blocks, which can be
 * combined later by BlockMatrixArray. Locally, this means, that each
 * matrix block of a system is generated separately and assembled into
 * the corresponding global block.
 *
 * This approach is advantageous, if the number of matrices for each
 * block position in the global system is different. For instance,
 * block preconditioners for the Oseen problem require 3 pressure
 * matrices, but only one divergence and one advection-diffusion
 * operator for velocities.
 *
 * Additionally, this approach enables the construction of a system of
 * equations from building blocks for each equation and coupling
 * operator.
 *
 * Nevertheless, since a separate FEValues object must be created for
 * each base element, it is not quite clear a priori, which data model
 * is more efficient.
 *
 * @author Guido Kanschat, 2009
 */
  namespace Assembler
  {
/**
 * The class assembling local contributions to a functional into the
 * global functionals.
 *
 * 
 *
 * @author Guido Kanschat, 2009
 */
    template <typename number = double>
    class Functional
    {
      public:
	void initialize(unsigned int n);
					 /**
					  * Assemble the values #J1
					  * into the global vectors
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info);
	
					 /**
					  * Assemble both local values
					  * into the global vectors
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info1,
		      const InfoObjects::DoFInfo<dim>& info2);

					 /**
					  * The value oth the ith #result.
					  */
	number operator() (unsigned int i) const;
	
	std::vector<number> J1;
      private:
					 /**
					  * The values into which the
					  * results are added.
					  */
	std::vector<double> results;
    };

/**
 * Compute cell and face contributions of a functional, typically for
 * error estimates.
 *
 * @author Guido Kanschat, 2009
 */
    template <typename number = double>
    class CellsAndFaces
    {
      public:
	void initialize(FEVectors& residuals, bool separate_faces = true);
					 /**
					  * Assemble the values #J1
					  * into the global vectors
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info);
	
					 /**
					  * Assemble both local values
					  * into the global vectors
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info1,
		      const InfoObjects::DoFInfo<dim>& info2);

					 /**
					  * The value oth the ith #result.
					  */
	number operator() (unsigned int i) const;
	
	std::vector<number> J1;
	std::vector<number> J2;
      private:
	FEVectors results;
	bool separate_faces;
    };
    
    
/**
 * Assemble local residuals into global residuals.
 *
 * The global residuals are expected as an FEVectors object.
 * The local residuals are block vectors.
 *
 * Depending on whether the BlockInfo object was initialize with
 * BlockInfo::initialize_local(), the comprehensive or block data
 * model is used locally.
 *
 * In the block model, each of the blocks of the local vectors
 * corresponds to the restriction of a single block of the system to
 * this cell (@ref GlossBlock). Thus, the size of this local block is
 * the number of degrees of freedom of the corresponding base element
 * of the FESystem.
 *
 * @todo Comprehensive model currently not implemented.
 *
 * @author Guido Kanschat, 2009
 */
    template <typename number = double>
    class ResidualLocalBlocksToGlobalBlocks
    {
      public:
					 /**
					  * Copy the BlockInfo and the
					  * matrix pointers into local
					  * variables and initialize
					  * cell vectors.
					  *
					  * If the BlockInfo object
					  * has been initialized with
					  * BlockInfo::initialize_local(),
					  * the vectors #R1 and #R2
					  * will have as many blocks as
					  * the global system. If not,
					  * there will be only a
					  * single block.
					  */
	void initialize(const BlockInfo& block_info,
			FEVectors& residuals);

					 /**
					  * Assemble the residuals #R1
					  * into the global residuals
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info);
	
					 /**
					  * Assemble both local residuals
					  * into the global residual
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info1,
		      const InfoObjects::DoFInfo<dim>& info2);
	
					 /**
					  * The local residuals on the
					  * cell or on the first cell
					  * on a face.
					  */
	std::vector<BlockVector<number> > R1;
	
					 /**
					  * The local residuals on the
					  * second cell on a face.
					  */
	std::vector<BlockVector<number> > R2;
      private:
					 /**
					  * Assemble a single local
					  * residual into the global.
					  */
	void assemble( FEVector& global,
		       const BlockVector<number>& local,
		       const std::vector<unsigned int>& dof);
	
					 /**
					  * Information on block sizes
					  * and block start indices.
					  */
	BlockInfo block_info;
					 /**
					  * The global matrices,
					  * stored as a vector of
					  * pointers.
					  */
	FEVectors residuals;
    };

/**
 * The local matrices filled by the worker and then assembled into the
 * global system by the derived classes.
 *
 * @author Guido Kanschat, 2009
 */
    template <typename number>
    class LocalMatrixBlocks
    {
					 /**
					  * Initialize a single local
					  * matrix block. A helper
					  * function for initialize()
					  */
	void initialize_local(
	  MatrixBlock<FullMatrix<number> >& M,
	  unsigned int row, unsigned int col,
	  unsigned int n_rows, unsigned int n_cols);
      public:
					 /**
					  * Copy the BlockInfo and the
					  * matrix pointers into local
					  * variables and initialize
					  * cell matrix vectors.
					  */
	template <class MatrixPtr>
	void initialize(const BlockInfo& block_info,
			std::vector<MatrixPtr>& matrices);

					 /**
					  * The local matrices
					  * coupling degrees of
					  * freedom in the cell
					  * itself aor within the
					  * first cell on a face.
					  */
	std::vector<MatrixBlock<FullMatrix<number> > > M11;
	
					 /**
					  * The local matrices
					  * coupling degrees of
					  * freedom in the second
					  * cell on a face.
					  */
	std::vector<MatrixBlock<FullMatrix<number> > > M22;
	
					 /**
					  * The local matrices
					  * coupling test function
					  * degrees of freedom in the
					  * first cell and trial
					  * function degrees of
					  * freedom in the second cell
					  * of a face, respectively.
					  */
	std::vector<MatrixBlock<FullMatrix<number> > > M12;
	
					 /**
					  * The local matrices
					  * coupling test function
					  * degrees of freedom in the
					  * second cell and trial
					  * function degrees of
					  * freedom in the first cell
					  * of a face, respectively.
					  */
	std::vector<MatrixBlock<FullMatrix<number> > > M21;
	
					 /**
					  * Information on block sizes
					  * and block start indices.
					  */
	BlockInfo block_info;
    };
    
    
    
/**
 * A helper class assembling local matrices into global matrices.
 *
 * The global matrices are expected as a vector of MatrixBlock
 * objects, each containing a matrix object with a function
 * corresponding to SparseMatrix::add() and information on the block
 * row and column this matrix represents in a block system.
 *
 * The local matrices are expected as a similar vector of MatrixBlock
 * objects, but containing a FullMatrix.
 *
 * Like with ResidualLocalBlocksToGlobalBlocks, the initialization of
 * the BlockInfo object decides whether the comprehensive data model
 * or the block model is used.
 *
 * In the comprehensive model, each of the LocalMatrixBlocks has
 * coordinates (0,0) and dimensions equal to the number of degrees of
 * freedom of the FESystem.
 *
 * In the comprehensive model, each block has its own block
 * coordinates and the size depends on the associated
 * FESystem::base_element(). These blocks can be generated separately
 * and will be assembled into the correct matrix block by this object.
 *
 * @author Guido Kanschat, 2009
 */
    template <class MATRIX, typename number = double>
    class MatrixLocalBlocksToGlobalBlocks
      : public LocalMatrixBlocks<number>
	
    {
      public:
					 /// The object that is stored
	typedef boost::shared_ptr<MatrixBlock<MATRIX> > MatrixPtr;

					 /**
					  * Constructor, initializing
					  * the #threshold, which
					  * limits how small numbers
					  * may be to be entered into
					  * the matrix.
					  */
	MatrixLocalBlocksToGlobalBlocks(double threshold = 1.e-12);
	
					 /**
					  * Copy the BlockInfo and the
					  * matrix pointers into local
					  * variables and initialize
					  * cell matrix vectors.
					  */
	void initialize(const BlockInfo& block_info,
			std::vector<MatrixPtr>& matrices);

					 /**
					  * Assemble the matrices #M11
					  * into the global matrices
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info);
	
					 /**
					  * Assemble all local matrices
					  * into the global matrices
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info1,
		      const InfoObjects::DoFInfo<dim>& info2);
	
      private:
					 /**
					  * Assemble a single local
					  * matrix into a global one.
					  */
	void assemble(
	  MATRIX& global,
	  const FullMatrix<number>& local,
	  unsigned int block_row,
	  unsigned int block_col,
	  const std::vector<unsigned int>& dof1,
	  const std::vector<unsigned int>& dof2);
	
					 /**
					  * The global matrices,
					  * stored as a vector of
					  * pointers.
					  */
	std::vector<MatrixPtr> matrices;

					 /**
					  * The smallest positive
					  * number that will be
					  * entered into the global
					  * matrix. All smaller
					  * absolute values will be
					  * treated as zero and will
					  * not be assembled.
					  */
	const double threshold;
	
    };

/**
 * A helper class assembling local matrices into global multilevel
 * matrices. This class is the multilevel equivalent of
 * MatrixLocalBlocksToGlobalBlocks and documentation of that class
 * applies here to a large extend.
 *
 * The global matrices are expected as a vector of pointers to MatrixBlock
 * objects, each containing a MGLevelObject with matrices with a function
 * corresponding to SparseMatrix::add() and information on the block
 * row and column this matrix represents in a block system.
 *
 * The local matrices are a similar vector of MatrixBlock objects, but
 * containing a FullMatrix.
 *
 * If local refinement occurs, the Multigrid method needs more
 * matrices, two for continuous elements and another two if numerical
 * fluxes are computed on interfaces. The second set can be added
 * using initialize_edge_flux(). Once added, the contributions in all
 * participating matrices will be assembled from the cell and face
 * matrices automatically.
 *
 * @author Guido Kanschat, 2009
 */
    template <class MATRIX, typename number = double>
    class MGMatrixLocalBlocksToGlobalBlocks
      : public LocalMatrixBlocks<number>
    {
      public:
					 /// The object that is stored
	typedef boost::shared_ptr<MatrixBlock<MGLevelObject<MATRIX> > > MatrixPtr;

					 /**
					  * Constructor, initializing
					  * the #threshold, which
					  * limits how small numbers
					  * may be to be entered into
					  * the matrix.
					  */
	MGMatrixLocalBlocksToGlobalBlocks(double threshold = 1.e-12);
	
					 /**
					  * Copy the BlockInfo and the
					  * matrix pointers into local
					  * variables and initialize
					  * cell matrix vectors.
					  */
	void initialize(const BlockInfo& block_info,
			std::vector<MatrixPtr>& matrices);

					 /**
					  * Multigrid methods on
					  * locally refined meshes
					  * need additional
					  * matrices. For
					  * discontinuous Galerkin
					  * methods, these are two
					  * flux matrices across the
					  * refinement edge, which are
					  * set by this method.
					  */
	void initialize_edge_flux(std::vector<MatrixPtr>& up, std::vector<MatrixPtr>& down);
	
					 /**
					  * Assemble the matrices #M11
					  * into the global matrices
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info);
	
					 /**
					  * Assemble all local matrices
					  * into the global matrices
					  * and reset them to zero.
					  */
	template<int dim>
	void assemble(const InfoObjects::DoFInfo<dim>& info1,
		      const InfoObjects::DoFInfo<dim>& info2);
	
      private:
					 /**
					  * Assemble a single local
					  * matrix into a global one.
					  */
	void assemble(
	  MATRIX& global,
	  const FullMatrix<number>& local,
	  unsigned int block_row,
	  unsigned int block_col,
	  const std::vector<unsigned int>& dof1,
	  const std::vector<unsigned int>& dof2,
	  unsigned int level1,
	  unsigned int level2,
	  bool transpose = false);
	
					 /**
					  * The level matrices,
					  * stored as a vector of
					  * pointers.
					  */
	std::vector<MatrixPtr> matrices;
	
					 /**
					  * The flux matrix between
					  * the fine and the coarse
					  * level at refinement edges.
					  */
	std::vector<MatrixPtr> flux_down;
	
					 /**
					  * The flux matrix between
					  * the coarse and the fine
					  * level at refinement edges.
					  */
	std::vector<MatrixPtr> flux_up;
	
					 /**
					  * The smallest positive
					  * number that will be
					  * entered into the global
					  * matrix. All smaller
					  * absolute values will be
					  * treated as zero and will
					  * not be assembled.
					  */
	const double threshold;
	
    };

//----------------------------------------------------------------------//

    template <typename number>
    inline void
    Functional<number>::initialize(unsigned int n)
    {
      J1.resize(n);
      results.resize(n);
    }
    
    
    template <typename number>
    template<int dim>
    inline void
    Functional<number>::assemble(const InfoObjects::DoFInfo<dim>& info)
    {
      for (unsigned int i=0;i<results.size();++i)
	{
	  results[i] += J1[i];
	  J1[i] = 0.;
	}
    }
    

    template <typename number>
    template<int dim>
    inline void
    Functional<number>::assemble(const InfoObjects::DoFInfo<dim>& info1,
				 const InfoObjects::DoFInfo<dim>& info2)
    {
      for (unsigned int i=0;i<results.size();++i)
	{
	  results[i] += J1[i];
	  J1[i] = 0.;
	}
    }

    
    template <typename number>
    inline number
    Functional<number>::operator() (unsigned int i) const
    {
      AssertIndexRange(i, results.size());
      return results[i];
    }
    
//----------------------------------------------------------------------//

    template <typename number>
    inline void
    CellsAndFaces<number>::initialize(FEVectors& r, bool sep)
    {
      Assert(r.vector_name(0) == "cells", FEVectors::ExcNameMismatch(0, "cells"));
      Assert(r.vector_name(1) == "faces", FEVectors::ExcNameMismatch(1, "faces"));
      AssertDimension(r.vector(0).n_blocks(), r.vector(1).n_blocks());
      
      results = r;
      separate_faces = sep;

      J1.resize(results.vector(0).n_blocks());
      J2.resize(results.vector(0).n_blocks());
    }
    
    
    template <typename number>
    template<int dim>
    inline void
    CellsAndFaces<number>::assemble(const InfoObjects::DoFInfo<dim>& info)
    {
      for (unsigned int i=0;i<J1.size();++i)
	{
	  if (separate_faces &&
	      info.face_number != deal_II_numbers::invalid_unsigned_int)
	    results.vector(1).block(i)(info.face->user_index()) += J1[i];
	
	  else
	    results.vector(0).block(i)(info.cell->user_index()) += J1[i];
	  J1[i] = 0.;
	}
    }
    

    template <typename number>
    template<int dim>
    inline void
    CellsAndFaces<number>::assemble(const InfoObjects::DoFInfo<dim>& info1,
				    const InfoObjects::DoFInfo<dim>& info2)
    {
      for (unsigned int i=0;i<J1.size();++i)
	{
	  if (separate_faces)
	    {
	      results.vector(1).block(i)(info1.face->user_index()) += J1[i];
	      if (info2.face != info1.face)
		results.vector(1).block(i)(info2.face->user_index()) += J1[i];
	      J1[i] = 0.;
	    }
	  else
	    {
	    results.vector(0).block(i)(info1.cell->user_index()) += J1[i];
	    results.vector(0).block(i)(info2.cell->user_index()) += J2[i];
	    J1[i] = 0.;
	    J2[i] = 0.;
	    }
	}
    }
    

//----------------------------------------------------------------------//

    template <typename number>
    inline void
    ResidualLocalBlocksToGlobalBlocks<number>::initialize(
      const BlockInfo& bi,
      FEVectors& m)
    {
      block_info = bi;
      residuals = m;

      Assert(block_info.local.size() != 0, ExcNotInitialized());
      
      R1.resize(residuals.n_vectors());
      R2.resize(residuals.n_vectors());
      for (unsigned int i=0;i<R1.size();++i)
	{
	  R1[i].reinit(block_info.local);
	  R2[i].reinit(block_info.local);
	}
    }


    template <typename number>
    inline void
    ResidualLocalBlocksToGlobalBlocks<number>::assemble(
      FEVector& global,
      const BlockVector<number>& local,
      const std::vector<unsigned int>& dof)
    {
      for (unsigned int b=0;b<local.n_blocks();++b)
	for (unsigned int j=0;j<local.block(b).size();++j)
	  {
					     // The coordinates of
					     // the current entry in
					     // DoFHandler
					     // numbering, which
					     // differs from the
					     // block-wise local
					     // numbering we use in
					     // our local vectors
	    const unsigned int jcell = this->block_info.local.local_to_global(b, j);
	    global(dof[jcell]) += local.block(b)(j);
	  }
    }

    
    template <typename number>
    template <int dim>
    inline void
    ResidualLocalBlocksToGlobalBlocks<number>::assemble(
      const InfoObjects::DoFInfo<dim>& info)
    {
      for (unsigned int i=0;i<residuals.n_vectors();++i)
	{
	  assemble(residuals.vector(i), R1[i], info.indices);
	  R1[i] = 0.;
	}
    }

    
    template <typename number>
    template <int dim>
    inline void
    ResidualLocalBlocksToGlobalBlocks<number>::assemble(
      const InfoObjects::DoFInfo<dim>& info1,
      const InfoObjects::DoFInfo<dim>& info2)
    {
      for (unsigned int i=0;i<residuals.n_vectors();++i)
	{
	  assemble(residuals.vector(i), R1[i], info1.indices);
	  assemble(residuals.vector(i), R2[i], info2.indices);
	  R1[i] = 0.;
	  R2[i] = 0.;
	}
    }


//----------------------------------------------------------------------//


    template <typename number>
    inline void
    LocalMatrixBlocks<number>::initialize_local(
      MatrixBlock<FullMatrix<number> >& M,
      unsigned int row, unsigned int col,
      unsigned int n_rows, unsigned int n_cols)
    {
      M.row = row;
      M.column = col;
      M.matrix.reinit(n_rows, n_cols);
    }

    template <typename number>
    template <class MatrixPtr>
    inline void
    LocalMatrixBlocks<number>::initialize(
      const BlockInfo& bi,
      std::vector<MatrixPtr>& matrices)
    {
      block_info = bi;

      M11.resize(matrices.size());
      M12.resize(matrices.size());
      M21.resize(matrices.size());
      M22.resize(matrices.size());

      for (unsigned int i=0;i<matrices.size();++i)
	{
	  const unsigned int row = matrices[i]->row;
	  const unsigned int col = matrices[i]->column;
	  const unsigned int n_rows = block_info.local.block_size(row);
	  const unsigned int n_cols = block_info.local.block_size(col);

	  initialize_local(M11[i], row, col, n_rows, n_cols);
	  initialize_local(M12[i], row, col, n_rows, n_cols);
	  initialize_local(M21[i], row, col, n_rows, n_cols);
	  initialize_local(M22[i], row, col, n_rows, n_cols);
	}
    }


// ----------------------------------------------------------------------//
    
    template <class MATRIX, typename number>
    inline
    MatrixLocalBlocksToGlobalBlocks<MATRIX, number>::MatrixLocalBlocksToGlobalBlocks(
      double threshold)
		    :
		    threshold(threshold)
    {}
    
    
    template <class MATRIX, typename number>
    inline void
    MatrixLocalBlocksToGlobalBlocks<MATRIX, number>::initialize(
      const BlockInfo& bi,
      std::vector<MatrixPtr>& m)
    {
      LocalMatrixBlocks<number>::initialize(bi, m);
      matrices = m;
    }
    

    
    template <class MATRIX, typename number>
    inline void
    MatrixLocalBlocksToGlobalBlocks<MATRIX, number>::assemble(
      MATRIX& global,
      const FullMatrix<number>& local,
      unsigned int block_row,
      unsigned int block_col,
      const std::vector<unsigned int>& dof1,
      const std::vector<unsigned int>& dof2)
    {
      for (unsigned int j=0;j<local.n_rows();++j)
	for (unsigned int k=0;k<local.n_cols();++k)
	  if (std::fabs(local(j,k)) >= threshold)
	    {
					       // The coordinates of
					       // the current entry in
					       // DoFHandler
					       // numbering, which
					       // differs from the
					       // block-wise local
					       // numbering we use in
					       // our local matrices
	      const unsigned int jcell = this->block_info.local.local_to_global(block_row, j);
	      const unsigned int kcell = this->block_info.local.local_to_global(block_col, k);

					       // The global dof
					       // indices to assemble
					       // in. Since we may
					       // have face matrices
					       // coupling two
					       // different cells, we
					       // provide two sets of
					       // dof indices.
	      const unsigned int jglobal = this->block_info.global.global_to_local(dof1[jcell]).second;
	      const unsigned int kglobal = this->block_info.global.global_to_local(dof2[kcell]).second;
	      
	      global.add(jglobal, kglobal, local(j,k));
	    }
    }

    
    template <class MATRIX, typename number>
    template <int dim>
    inline void
    MatrixLocalBlocksToGlobalBlocks<MATRIX, number>::assemble(
      const InfoObjects::DoFInfo<dim>& info)
    {
      for (unsigned int i=0;i<matrices.size();++i)
	{
					   // Row and column index of
					   // the block we are dealing with
	  const unsigned int row = matrices[i]->row;
	  const unsigned int col = matrices[i]->column;

	  assemble(matrices[i]->matrix, this->M11[i].matrix, row, col, info.indices, info.indices);
	  this->M11[i].matrix = 0.;
	}
    }

    
    template <class MATRIX, typename number>
    template <int dim>
    inline void
    MatrixLocalBlocksToGlobalBlocks<MATRIX, number>::assemble(
      const InfoObjects::DoFInfo<dim>& info1,
      const InfoObjects::DoFInfo<dim>& info2)
    {
      for (unsigned int i=0;i<matrices.size();++i)
	{
					   // Row and column index of
					   // the block we are dealing with
	  const unsigned int row = matrices[i]->row;
	  const unsigned int col = matrices[i]->column;

	  assemble(matrices[i]->matrix, this->M11[i].matrix, row, col, info1.indices, info1.indices);
	  assemble(matrices[i]->matrix, this->M12[i].matrix, row, col, info1.indices, info2.indices);
	  assemble(matrices[i]->matrix, this->M21[i].matrix, row, col, info2.indices, info1.indices);
	  assemble(matrices[i]->matrix, this->M22[i].matrix, row, col, info2.indices, info2.indices);

	  this->M11[i].matrix = 0.;
	  this->M12[i].matrix = 0.;
	  this->M21[i].matrix = 0.;
	  this->M22[i].matrix = 0.;
	}
    }


// ----------------------------------------------------------------------//
    
    template <class MATRIX, typename number>
    inline
    MGMatrixLocalBlocksToGlobalBlocks<MATRIX, number>::MGMatrixLocalBlocksToGlobalBlocks(
      double threshold)
		    :
		    threshold(threshold)
    {}
    
    
    template <class MATRIX, typename number>
    inline void
    MGMatrixLocalBlocksToGlobalBlocks<MATRIX, number>::initialize(
      const BlockInfo& bi,
      std::vector<MatrixPtr>& m)
    {
      LocalMatrixBlocks<number>::initialize(bi, m);
      matrices = m;
    }
    

    template <class MATRIX, typename number>
    inline void
    MGMatrixLocalBlocksToGlobalBlocks<MATRIX, number>::initialize_edge_flux(
      std::vector<MatrixPtr>& up,
      std::vector<MatrixPtr>& down)
    {
      flux_up = up;
      flux_down = down;
    }
    
    
    template <class MATRIX, typename number>
    inline void
    MGMatrixLocalBlocksToGlobalBlocks<MATRIX, number>::assemble(
      MATRIX& global,
      const FullMatrix<number>& local,
      unsigned int block_row,
      unsigned int block_col,
      const std::vector<unsigned int>& dof1,
      const std::vector<unsigned int>& dof2,
      unsigned int level1,
      unsigned int level2,
      bool transpose)
    {
      for (unsigned int j=0;j<local.n_rows();++j)
	for (unsigned int k=0;k<local.n_cols();++k)
	  if (std::fabs(local(j,k)) >= threshold)
	    {
					       // The coordinates of
					       // the current entry in
					       // DoFHandler
					       // numbering, which
					       // differs from the
					       // block-wise local
					       // numbering we use in
					       // our local matrices
	      const unsigned int jcell = this->block_info.local.local_to_global(block_row, j);
	      const unsigned int kcell = this->block_info.local.local_to_global(block_col, k);

					       // The global dof
					       // indices to assemble
					       // in. Since we may
					       // have face matrices
					       // coupling two
					       // different cells, we
					       // provide two sets of
					       // dof indices.
	      const unsigned int jglobal = this->block_info.levels[level1].global_to_local(dof1[jcell]).second;
	      const unsigned int kglobal = this->block_info.levels[level2].global_to_local(dof2[kcell]).second;

	      if (transpose)
		global.add(kglobal, jglobal, local(j,k));
	      else
		global.add(jglobal, kglobal, local(j,k));
	    }
    }

    
    template <class MATRIX, typename number>
    template <int dim>
    inline void
    MGMatrixLocalBlocksToGlobalBlocks<MATRIX, number>::assemble(
      const InfoObjects::DoFInfo<dim>& info)
    {
      const unsigned int level = info.cell->level();
      
      for (unsigned int i=0;i<matrices.size();++i)
	{
					   // Row and column index of
					   // the block we are dealing with
	  const unsigned int row = matrices[i]->row;
	  const unsigned int col = matrices[i]->column;

	  assemble(matrices[i]->matrix[level], this->M11[i].matrix, row, col,
		   info.indices, info.indices, level, level);
	  
	  this->M11[i].matrix = 0.;
	}
    }

    
    template <class MATRIX, typename number>
    template <int dim>
    inline void
    MGMatrixLocalBlocksToGlobalBlocks<MATRIX, number>::assemble(
      const InfoObjects::DoFInfo<dim>& info1,
      const InfoObjects::DoFInfo<dim>& info2)
    {
      const unsigned int level1 = info1.cell->level();
      const unsigned int level2 = info2.cell->level();
      
      for (unsigned int i=0;i<matrices.size();++i)
	{
					   // Row and column index of
					   // the block we are dealing with
	  const unsigned int row = matrices[i]->row;
	  const unsigned int col = matrices[i]->column;

	  if (level1 == level2)
	    {
	      assemble(matrices[i]->matrix[level1], this->M11[i].matrix, row, col, info1.indices, info1.indices, level1, level1);
	      assemble(matrices[i]->matrix[level1], this->M12[i].matrix, row, col, info1.indices, info2.indices, level1, level2);
	      assemble(matrices[i]->matrix[level1], this->M21[i].matrix, row, col, info2.indices, info1.indices, level2, level1);
	      assemble(matrices[i]->matrix[level1], this->M22[i].matrix, row, col, info2.indices, info2.indices, level2, level2);
	    }
	  else
	    {
	      Assert(level1 > level2, ExcNotImplemented());
	      if (flux_up.size() != 0)
		{
						   // Do not add M22,
						   // which is done by
						   // the coarser cell
		  assemble(matrices[i]->matrix[level1], this->M11[i].matrix, row, col,
			   info1.indices, info1.indices, level1, level1);	      
		  assemble(flux_up[i]->matrix[level1], this->M12[i].matrix, row, col,
			   info1.indices, info2.indices, level1, level2, true);
		  assemble(flux_down[i]->matrix[level1], this->M21[i].matrix, row, col,
			   info2.indices, info1.indices, level2, level1);
		}
	    }
	  
	  this->M11[i].matrix = 0.;
	  this->M12[i].matrix = 0.;
	  this->M21[i].matrix = 0.;
	  this->M22[i].matrix = 0.;
	}
    }
  }
}


#endif
