// ------------------------------------------------------------------
//
// $Id: mesh_worker_info.h 1348 2013-08-16 00:45:07Z secanell $
//
// Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat.
// Copyright (C) 2012 by Valentin N. Zingan, University of Alberta.
//
// This file is subject to QPL and may not be  distributed
// without copyright and license information. Please refer
// to the file deal.II/doc/license.html for the  text  and
// further information on this license.
//
// ------------------------------------------------------------------

#ifndef _DEALII_APPFRAME_MESH_WORKER_INFO_H_
#define _DEALII_APPFRAME_MESH_WORKER_INFO_H_

#include <base/geometry_info.h>
#include <base/quadrature_lib.h>
#include <base/mg_level_object.h>

#include <lac/block_indices.h>

#include <fe/fe.h>
#include <fe/fe_tools.h>

#include <multigrid/mg_tools.h>

#include <appframe/matrix_base.h>

#include <boost/shared_ptr.hpp>

namespace dealii
{
  template<int,int> class DoFHandler;
  template<int,int> class MGDoFHandler;
}

using namespace dealii;
using namespace AppFrame;

// ################################################################################################

/**
 * A collection of functions and classes for the mesh loops that are
 * an ubiquitous part of each finite element program.
 *
 * The workhorse of this namespace is the loop() function, which implements a
 * completely generic loop over all mesh entities (cells and faces).
 *
 * The loop() depends on certain objects handed to it as
 * arguments. These objects can be structured into categories
 * corresponding to the namespaces where they are defined.
 *
 * @author Guido Kanschat, 2009
 */

// ##### CATEGORY-1 ################################################# INFO_OBJECTS ################
//
// ##### MeshWorker::BlockInfo                       #####
// ##### MeshWorker::VectorSelector                  #####
// ##### MeshWorker::InfoObjects::DoFInfo            #####
// ##### MeshWorker::InfoObjects::IntegrationInfo    #####
// ##### MeshWorker::InfoObjects::IntegrationInfoBox #####

/**
 * InfoObjects provide the data used by other MeshWorker
 * objects.
 *
 * They can be considered as an extension of mesh @ref Iterators
 * in the sense that they are updated for each mesh cell or face,
 * and provide additional data on these mesh entities.
 *
 * @author Guido Kanschat, 2009
 */

namespace MeshWorker
{

/**
 * @brief A small structure collecting the different BlockIndices of FEVector vectors (for instance, solution) involved in the computations.
 *
 * @author Guido Kanschat, 2009
 */

struct BlockInfo : public Subscriptor
{
  /**
   * @brief The block structure of the global FEVector solution.
   */
  BlockIndices global;

  /**
   * @brief The block structure of a local (cell) FEVector solution.
   */
  BlockIndices local;

  /**
   * @brief The multilevel block structure of the global FEVector solution.
   */
  std::vector<BlockIndices> levels;

  /**
   * A vector of base elements.
   *
   * @note index -> value = # block -> # fe in an FESystem object
   */
  std::vector<unsigned int> base_element;

  /**
   * A vector containing the internal
   * renumbering of degrees of freedom from the
   * standard ordering on a cell to a
   * block wise ordering on the same cell.
   *
   * @note index -> value = standard # dof -> block wise # dof
   */
  std::vector<unsigned int> local_renumbering;

  /**
   * @brief Fill the structure globally with values describing the DoFHandler.
   */
  template<int dim, int spacedim>
  void initialize(const DoFHandler<dim, spacedim>& dof_handler)
  {
    const FiniteElement<dim, spacedim>& fe = dof_handler.get_fe();
    std::vector<unsigned int> sizes(fe.n_blocks());
    DoFTools::count_dofs_per_block(dof_handler, sizes);
    global.reinit(sizes);
  }

  /**
   * @brief Fill the structure locally with values describing the DoFHandler.
   */
  template<int dim, int spacedim>
  void initialize_local(const DoFHandler<dim, spacedim>& dof_handler)
  {
    const FiniteElement<dim, spacedim>& fe = dof_handler.get_fe();
    std::vector<unsigned int> sizes(fe.n_blocks());

    base_element.resize(fe.n_blocks());

    for(unsigned int i = 0; i < base_element.size(); ++i)
      base_element[i] = fe.block_to_base_index(i).first;

    local_renumbering.resize(fe.n_dofs_per_cell());
    FETools::compute_block_renumbering(fe,
                                       local_renumbering,
                                       sizes,
                                       false);
    local.reinit(sizes);
  }

  /**
   * @brief Fill the structure globally with values describing the MGDoFHandler.
   */
  template<int dim, int spacedim>
  void initialize(const MGDoFHandler<dim, spacedim>& mg_dof_handler)
  {
    initialize((DoFHandler<dim, spacedim>&) mg_dof_handler);

    std::vector< std::vector<unsigned int> > sizes(mg_dof_handler.get_tria().n_levels(),
                                                   std::vector<unsigned int>(mg_dof_handler.get_fe().n_blocks())
                                                  );

    MGTools::count_dofs_per_block(mg_dof_handler, sizes);

    levels.resize(sizes.size());
    for(unsigned int i = 0; i < sizes.size(); ++i)
      levels[i].reinit(sizes[i]);
  }
};

/**
 * A class that selects the named FEVector vectors in
 * FEVectors and assigns the types of information extracted.
 * These types are values, gradients, and hessians.
 *
 * @author Guido Kanschat, 2009
 */

class VectorSelector
{
public:

  /**
   * Default constructor for an
   * empty object.
   */
  VectorSelector() { }

  /**
   * Copy constructor, copying
   * everything except the
   * cache_values, cache_gradients, and
   * cache_hessians.
   */
  VectorSelector(const VectorSelector& other);

  /**
   * Add a FEVector name as it is
   * stored in FEVectors
   * to the list of selections.
   * The arguments are
   * the name and indicators
   * showing the types of
   * information extracted for this name.
   */
  void add_vector(const std::string& name,
                  bool               values    = true,
                  bool               gradients = false,
                  bool               hessians  = false);

  /**
   * Based on the list of selections,
   * cache the numbers of FEVector vectors
   * in FEVectors appropriately.
   * Available containers are
   * cache_values, cache_gradients, and
   * cache_hessians.
   */
  void cache(const FEVectors& src);

  /**
   * Check whether any FEVector in FEVectors is selected.
   */
  bool empty() const
  {
    return (selection.size() == 0);
  }

  /**
   * Returns true if values are selected for any FEVector in FEVectors.
   */
  bool has_values() const
  {
    return selected_values;
  }

  /**
   * Returns true if gradients are selected for any FEVector in FEVectors.
   */
  bool has_gradients() const
  {
    return selected_gradients;
  }

  /**
   * Returns true if hessians are selected for any FEVector in FEVectors.
   */
  bool has_hessians() const
  {
    return selected_hessians;
  }

  /**
   * The number of FEVector vectors in FEVectors selected for getting values.
   */
  unsigned int n_values() const
  {
    return cache_values.size();
  }

  /**
   * The number of FEVector vectors in FEVectors selected for getting gradients.
   */
  unsigned int n_gradients() const
  {
    return cache_gradients.size();
  }

  /**
   * The number of FEVector vectors in FEVectors selected for getting hessians.
   */
  unsigned int n_hessians() const
  {
    return cache_hessians.size();
  }

  /**
   * The FEVector index in FEVectors for the ith value.
   */
  unsigned int value_index(unsigned int i) const
  {
    AssertIndexRange(i, cache_values.size());
    return cache_values[i];
  }

  /**
   * The FEVector index in FEVectors for the ith gradient.
   */
  unsigned int gradient_index(unsigned int i) const
  {
    AssertIndexRange(i, cache_gradients.size());
    return cache_gradients[i];
  }

  /**
   * The FEVector index in FEVectors for the ith hessian.
   */
  unsigned int hessian_index(unsigned int i) const
  {
    AssertIndexRange(i, cache_hessians.size());
    return cache_hessians[i];
  }

  /**
   * Print the info on the
   * selected FEVector vectors
   * in FEVectors to the stream.
   */
  template<typename STREAM>
  void print(STREAM&          s,
             const FEVectors& v) const;

private:

  /**
   * The structure which stores
   * the name of the selected FEVector
   * as it is stored in FEVectors
   * and the types of information
   * extracted.
   *
   * The structure is then stored in the
   * list of selections.
   */
  struct ListEntry
  {
    /// The name of the selected FEVector as it is stored in FEVectors.
    std::string name;

    /// Update values for this name.
    bool values;

    /// Update gradients for this name.
    bool gradients;

    /// Update hessians for this name.
    bool hessians;
  };

  /**
   * The list of selections.
   */
  std::vector<ListEntry> selection;

  /**
   * Set to @p true if values are selected for any FEVector in FEVectors.
   */
  bool selected_values;

  /**
   * Set to @p true if gradients are selected for any FEVector in FEVectors.
   */
  bool selected_gradients;

  /**
   * Set to @p true if hessians are selected for any FEVector in FEVectors.
   */
  bool selected_hessians;

  /**
   * Cache indices of FEVector vectors in FEVectors for which values have been selected.
   */
  std::vector<unsigned int> cache_values;

  /**
   * Cache indices of FEVector vectors in FEVectors for which gradients have been selected.
   */
  std::vector<unsigned int> cache_gradients;

  /**
   * Cache indices of FEVector vectors in FEVectors for which hessians have been selected.
   */
  std::vector<unsigned int> cache_hessians;
};

// --- implementation ---

inline
VectorSelector::VectorSelector(const VectorSelector& other)
:
selection(other.selection),
selected_values(other.selected_values),
selected_gradients(other.selected_gradients),
selected_hessians(other.selected_hessians)
{ }

inline
void
VectorSelector::add_vector(const std::string& name,
                           bool               values,
                           bool               gradients,
                           bool               hessians)
{
  ListEntry e;
  e.name      = name;
  e.values    = values;
  e.gradients = gradients;
  e.hessians  = hessians;

  selected_values    |= values;
  selected_gradients |= gradients;
  selected_hessians  |= hessians;

  selection.push_back(e);
}

inline
void
VectorSelector::cache(const FEVectors& src)
{
  cache_values.resize(0);
  cache_gradients.resize(0);
  cache_hessians.resize(0);

  for(std::vector<ListEntry>::const_iterator iter  = selection.begin();
                                             iter != selection.end();
                                           ++iter)
  {
    const unsigned int k = src.find_vector(iter->name);

    Assert( k != deal_II_numbers::invalid_unsigned_int,
            ApplicationData::ExcNotFound("vector", iter->name) );

    if(iter->values)    cache_values.push_back(k);
    if(iter->gradients) cache_gradients.push_back(k);
    if(iter->hessians)  cache_hessians.push_back(k);
  }
}

template<typename STREAM>
inline
void
VectorSelector::print(STREAM&          s,
                      const FEVectors& v) const
{
  s << "values:   ";
  for(unsigned int i = 0; i < cache_values.size(); ++i)
    s << " '" << v.vector_name(cache_values[i]) << '\'';
  s << std::endl << "gradients:   ";
  for(unsigned int i = 0; i < cache_gradients.size(); ++i)
    s << " '" << v.vector_name(cache_gradients[i]) << '\'';
  s << std::endl << "hessians:   ";
  for(unsigned int i = 0; i < cache_hessians.size(); ++i)
    s << " '" << v.vector_name(cache_hessians[i]) << '\'';
  s << std::endl;
}

/**
 * Additionally to the VectorSelector class we collect some helper
 * functions that choose the right access
 * functions in the VectorSelector class itself
 * depending on the data type being
 * used.
 *
 * Note that the last argument of
 * these helper functions is never
 * used. It is there just for the
 * type identification.
 */

inline
unsigned int
selector_size(const VectorSelector& s,
              const std::vector<double>&)
{
  return s.n_values();
}

template<int dim>
inline
unsigned int
selector_size(const VectorSelector& s,
              const std::vector< Tensor<1, dim> >&)
{
  return s.n_gradients();
}

template<int dim>
inline
unsigned int
selector_size(const VectorSelector& s,
              const std::vector< Tensor<2, dim> >&)
{
  return s.n_hessians();
}

inline
unsigned int
selector_index(const VectorSelector& s,
               unsigned int          i,
               const std::vector<double>&)
{
  return s.value_index(i);
}

template<int dim>
inline
unsigned int
selector_index(const VectorSelector& s,
               unsigned int          i,
               const std::vector< Tensor<1, dim> >&)
{
  return s.gradient_index(i);
}

template<int dim>
inline
unsigned int
selector_index(const VectorSelector& s,
               unsigned int          i,
               const std::vector< Tensor<2, dim> >&)
{
  return s.hessian_index(i);
}

namespace InfoObjects
{

/**
 * Helper functions computing the
 * desired data in each quadrature
 * point of a mesh entity by calling
 * FEValuesBase::get_function_values(),
 * FEValuesBase::get_function_grads(),
 * and
 * FEValuesBase::get_function_hessians().
 *
 * @param fe_values: The FEValues
 * object.
 *
 * @param fe_vector: The global finite
 * element function in the form of a global
 * node FEVector.
 *
 * @param local_dof_indices: The local DoF
 * indices associated with the current mesh entity.
 *
 * @param first_index: The first
 * index in @p local_dof_indices to be used.
 *
 * @param n_indices: The number of
 * indices in @p local_dof_indices to be
 * used.
 *
 * @param result: The result.
 * The access to ith component in
 * qth quadrature point is result[i][q].
 */

template<int dim, typename TYPE>
inline
void
fill_data(const FEValuesBase<dim>&          fe_values,
          const FEVector&                   fe_vector,
          const std::vector<unsigned int>&  local_dof_indices,
          unsigned int                      first_index,
          unsigned int                      n_indices,
          std::vector< std::vector<TYPE> >& result)
{
  Assert(false, ExcNotImplemented());
}

template<int dim>
inline
void
fill_data(const FEValuesBase<dim>&            fe_values,
          const FEVector&                     fe_vector,
          const std::vector<unsigned int>&    local_dof_indices,
          unsigned int                        first_index,
          unsigned int                        n_indices,
          std::vector< std::vector<double> >& result)
{
  fe_values.get_function_values(fe_vector,
                                make_slice(local_dof_indices, first_index, n_indices),
                                result,
                                true);
}

template<int dim>
inline
void
fill_data(const FEValuesBase<dim>&                     fe_values,
          const FEVector&                              fe_vector,
          const std::vector<unsigned int>&             local_dof_indices,
          unsigned int                                 first_index,
          unsigned int                                 n_indices,
          std::vector< std::vector< Tensor<1,dim> > >& result)
{
  fe_values.get_function_grads(fe_vector,
                               make_slice(local_dof_indices, first_index, n_indices),
                               result,
                               true);
}

template<int dim>
inline
void
fill_data(const FEValuesBase<dim>&                     fe_values,
          const FEVector&                              fe_vector,
          const std::vector<unsigned int>&             local_dof_indices,
          unsigned int                                 first_index,
          unsigned int                                 n_indices,
          std::vector< std::vector< Tensor<2,dim> > >& result)
{
  fe_values.get_function_hessians(fe_vector,
                                  make_slice(local_dof_indices, first_index, n_indices),
                                  result,
                                  true);
}

/**
 * Very basic info class only containing information on geometry and
 * degrees of freedom of a mesh entity.
 *
 * The information in this class is usually used by one of the
 * Assembler classes. This information is also needed in mesh
 * based matrices (often referred to as matrix free methods).
 *
 * This class operates in two different modes, corresponding to the
 * data models discussed in the Assembler namespace documentation.
 *
 * The choice of the local data model is triggered by the vector
 * #BlockInfo::local_renumbering, which in turn is usually filled by
 * #BlockInfo::initialize_local(). If this function has been called, or
 * if this vector has been changed from zero-length, then local dof
 * indices stored in this class will automatically be renumbered to
 * reflect the local block structure.
 *
 * The BlockInfo object is stored here as a pointer. Therefore, if the
 * local block structure changes, for instance because of the mesh refinement,
 * the DoFInfo class will automatically use this new structure.
 *
 * @author Guido Kanschat, 2009
 * @author Valentin N. Zingan, 2012
 */

template<int dim, int spacedim = dim>
class DoFInfo
{
public:

  /**
   * Constructor.
   */
  DoFInfo(const BlockInfo& block_info);

  /**
   * @deprecated Constructor with ignored first argument.
   */
  DoFInfo(const FEVectors&,
          const BlockInfo& block_info);

  /**
   * Set the current cell and
   * fill #indices.
   */
  template<typename DHCellIterator>
  void reinit(const DHCellIterator& c)
  {
    get_indices(c);

    dof_cell        = c;
    dof_active_cell = c;
    cell            = static_cast<typename Triangulation<dim>::cell_iterator> (c);

    face_number = deal_II_numbers::invalid_unsigned_int;
    sub_number  = deal_II_numbers::invalid_unsigned_int;
  }

  /**
   * Set the current cell and face and
   * fill #indices if the #cell
   * changed.
   */
  template<typename DHCellIterator, typename DHFaceIterator>
  void reinit(const DHCellIterator& c,
              const DHFaceIterator& f,
              const unsigned int    n)
  {
    if(   cell.state() != IteratorState::valid || cell != static_cast<typename Triangulation<dim>::cell_iterator> (c)   )
    {
      get_indices(c);
    }

    dof_cell        = c;
    dof_active_cell = c;
    cell            = static_cast<typename Triangulation<dim>::cell_iterator> (c);

    dof_face = f;
    face     = static_cast<typename Triangulation<dim>::face_iterator> (f);

    face_number = n;
    sub_number  = deal_II_numbers::invalid_unsigned_int;
  }

  /**
   * Set the current cell, face, and subface and
   * fill #indices if the #cell
   * changed.
   */
  template<typename DHCellIterator, typename DHFaceIterator>
  void reinit(const DHCellIterator& c,
              const DHFaceIterator& f,
              const unsigned int    n,
              const unsigned int    s)
  {
    if(   cell.state() != IteratorState::valid || cell != static_cast<typename Triangulation<dim>::cell_iterator> (c)   )
    {
      get_indices(c);
    }

    dof_cell        = c;
    dof_active_cell = c;
    cell            = static_cast<typename Triangulation<dim>::cell_iterator> (c);

    dof_face = f;
    face     = static_cast<typename Triangulation<dim>::face_iterator> (f);

    face_number = n;
    sub_number  = s;
  }

  /**
   * The current DoFHandler<dim> cell.
   */
  typename DoFHandler<dim>::cell_iterator dof_cell;

  /**
   * The current DoFHandler<dim> active cell.
   */
  typename DoFHandler<dim>::active_cell_iterator dof_active_cell;

  /**
   * The current DoFHandler<dim> face.
   */
  typename DoFHandler<dim>::face_iterator dof_face;

  /**
   * The current Triangulation<dim> cell.
   */
  typename Triangulation<dim>::cell_iterator cell;

  /**
   * The current Triangulation<dim> face.
   */
  typename Triangulation<dim>::face_iterator face;

  /**
   * The number of the current
   * face on the current cell.
   *
   * This number is
   * deal_II_numbers::invalid_unsigned_int
   * if the DoFInfo object is
   * initialized with a cell only.
   */
  unsigned int face_number;

  /**
   * The number of the current
   * subface on the current face
   * of the current cell.
   *
   * This number is
   * deal_II_numbers::invalid_unsigned_int
   * if the DoFInfo object is
   * initialized with a cell and face only.
   */
  unsigned int sub_number;

  /**
   * The local DoF indices associated
   * with the current mesh entity.
   */
  std::vector<unsigned int> indices;

  /**
   * The block structure of the problem at hand.
   */
  SmartPointer<const BlockInfo> block_info;

private:

  /// Fill index vector.
  void get_indices(const typename DoFHandler<dim, spacedim>::cell_iterator c);

  /// Fill index vector with this level indices.
  void get_indices(const typename MGDoFHandler<dim, spacedim>::cell_iterator c);

  /**
   * Auxiliary vector.
   */
  std::vector<unsigned int> indices_org;
};

// --- implementation ---

template<int dim, int spacedim>
inline
DoFInfo<dim, spacedim>::DoFInfo(const BlockInfo& block_info)
:
block_info(&block_info,
           typeid(*this).name())
{ }

template<int dim, int spacedim>
inline
DoFInfo<dim, spacedim>::DoFInfo(const FEVectors&,
                                const BlockInfo& block_info)
:
block_info(&block_info,
           typeid(*this).name())
{ }

template<int dim, int spacedim>
inline
void
DoFInfo<dim, spacedim>::get_indices(const typename DoFHandler<dim, spacedim>::cell_iterator c)
{
  indices.resize(c->get_fe().dofs_per_cell);

  if(block_info->local_renumbering.size() == 0)
  {
    c->get_dof_indices(indices);
  }
  else
  {
    indices_org.resize(c->get_fe().dofs_per_cell);
    c->get_dof_indices(indices_org);

    for(unsigned int i = 0; i < indices.size(); ++i)
      indices[this->block_info->local_renumbering[i]] = indices_org[i];
  }
}

template<int dim, int spacedim>
inline
void
DoFInfo<dim, spacedim>::get_indices(const typename MGDoFHandler<dim, spacedim>::cell_iterator c)
{
  indices.resize(c->get_fe().dofs_per_cell);

  if(block_info->local_renumbering.size() == 0)
  {
    c->get_mg_dof_indices(indices);
  }
  else
  {
    indices_org.resize(c->get_fe().dofs_per_cell);
    c->get_mg_dof_indices(indices_org);

    for(unsigned int i = 0; i < indices.size(); ++i)
      indices[this->block_info->local_renumbering[i]] = indices_org[i];
  }
}

/**
 * This class is created for the objects handed to the local integration functions.
 *
 * The objects of this class contain one or more objects of either type FEValues (cell)
 * or FEFaceValues (face) or FESubfaceValues (subface) to be used in the local
 * integration functions. They are stored in an array of pointers either to the base
 * class FEValuesBase (cell) or to the base class FEFaceValuesBase (face, subface).
 *
 * Therefore, the actual type @p FEVALUES must be either FEValues (cell)
 * or FEFaceValues (face) or FESubfaceValues (subface).
 * The template parameter @p FEVALUESBASE must be either
 * FEValuesBase (cell) or FEFaceValuesBase (face, subface).
 *
 * Additionally, this class includes containers to store the values (#values),
 * gradients (#gradients), and hessians (#hessians) of finite element functions,
 * stored in #global_data, in the quadrature points of mesh entities.
 * You can use initialize_selector() function to select the FEVector vectors
 * in #global_data by names.
 *
 * <h3>Integration models</h3>
 *
 * This class supports two local integration models.
 * One is the
 * standard model suggested by the use of FESystem. Namely, there is
 * one FEValuseBase object in this class, containing all shape
 * functions of the whole system, and having as many components as the
 * system has. Using this model involves loops over all system shape
 * functions. It requires to identify the system components
 * for each shape function and to select the correct bilinear form,
 * usually in an @p if or @p switch statement.
 *
 * The second local integration model builds one FEValuesBase object per
 * base element of the system. The degrees of freedom on each cell are
 * renumbered by block, such that they represent the same block
 * structure as the global system. Objects performing the integration
 * can then process each block separately, which improves reusability
 * of code considerably.
 *
 * @note As described in DoFInfo, the use of the local block model is
 * triggered by calling BlockInfo::initialize_local() before
 * using initialize() in this class.
 *
 * @author Guido Kanschat, 2009
 * @author Valentin N. Zingan, 2012
 */

template<int dim, typename FEVALUESBASE>
class IntegrationInfo : public DoFInfo<dim>
{
public:

  /**
   * Constructor.
   */
  IntegrationInfo(const BlockInfo& block_info);

  /**
   * @deprecated Constructor.
   */
  IntegrationInfo(const FEVectors& data,
                  const BlockInfo& block_info);

  /**
   * Build internal structures like #fevalv
   * and allocate memory for
   * #values,
   * #gradients,
   * #hessians.
   *
   * @param FE_VALUES is an object
   * of the actual type FEVALUES
   * which is
   * either
   * FEValues (cell)
   * or
   * FEFaceValues (face)
   * or
   * FESubfaceValues (subface).
   * @note #FE_VALUES is really not
   * needed here and only serves to
   * define the actual type FEVALUES.
   *
   * @param fe is the finite
   * element (for instance, FESystem object)
   * stored in the DoFHandler.
   * It is used in the constructor
   * of #FE_VALUES.
   *
   * @param mapping is the Mapping
   * object used to map the
   * mesh cells.
   * It is used in the constructor
   * of #FE_VALUES.
   *
   * @param quadrature is a
   * quadrature formula.
   * It is used in the constructor
   * of #FE_VALUES.
   *
   * @param flags are the
   * UpdateFlags.
   * They are used in the constructor
   * of #FE_VALUES.
   */
  template<typename FEVALUES>
  void initialize(const FEVALUES*                                 FE_VALUES,
                  const FiniteElement<dim>&                       fe,
                  const Mapping<dim>&                             mapping,
                  const Quadrature<FEVALUES::integral_dimension>& quadrature,
                  const UpdateFlags                               flags);

  /**
   * Initialize the
   * #selector.
   * Cache the
   * #selector.
   */
  void initialize_selector(const VectorSelector& s)
  {
    selector = s;

    if(global_data)
      selector.cache(*global_data);
  }

  /**
   * Initialize the
   * #global_data.
   * Cache the
   * #selector.
   */
  void initialize_data(const FEVectors& data)
  {
    global_data = &data;
    selector.cache(*global_data);
  }

  /**
   * Resize #fevalv to 0.
   * In other words, clear #fevalv.
   */
  void clear()
  {
    fevalv.resize(0);
  }

  /// Access to a single actual FEVALUES object.
  const FEVALUESBASE& fe() const
  {
    AssertDimension(fevalv.size(), 1);
    return *fevalv[0];
  }

  /// Access to a group of actual FEVALUES objects.
  const FEVALUESBASE& fe(unsigned int i) const
  {
    Assert( i < fevalv.size(), ExcIndexRange(i, 0, fevalv.size()) );
    return *fevalv[i];
  }

  /// (ZVN) Access to \p fe_val_unsplit.
  const FEVALUESBASE& get_fe_val_unsplit() const
  {
    return *fe_val_unsplit;
  }

  /**
   * Reinitialize internal data
   * for use on a cell.
   */
  template<typename DHCellIterator>
  void reinit(const DHCellIterator& c);

  /**
   * Reinitialize internal data
   * for use on a face.
   */
  template<typename DHCellIterator, typename DHFaceIterator>
  void reinit(const DHCellIterator& c,
              const DHFaceIterator& f,
              const unsigned int    fn);

  /**
   * Reinitialize internal data
   * for use on a subface.
   */
  template<typename DHCellIterator, typename DHFaceIterator>
  void reinit(const DHCellIterator& c,
              const DHFaceIterator& f,
              const unsigned int    fn,
              const unsigned int    sn);

  /**
   * @deprecated This is the
   * old version not using
   * the VectorSelector.
   *
   * Use the finite element
   * functions in #global_data
   * and fill the vectors
   * #values,
   * #gradients,
   * #hessians.
   */
  template<typename TYPE>
  void fill_local_data(std::vector< std::vector< std::vector<TYPE> > >& data,
                       bool                                             split_fevalues) const;

  /// This is true if we assemble for multigrid.
  bool multigrid;

  /**
   * The smart pointer
   * to the FEVectors object
   * called #global_data.
   * It contains
   * the named scalars
   * and named vectors
   * of an application.
   */
  SmartPointer<const FEVectors> global_data;

  /**
   * This object selects
   * the named FEVector vectors
   * in #global_data and
   * assigns the types
   * of information extracted.
   * These types are values, gradients, and hessians.
   */
  VectorSelector selector;

  /**
   * The vector containing the
   * values of finite element
   * functions in the quadrature
   * points.
   *
   * Example: nth solution, mth component, value at the qth quadrature point = values[n][m][q]
   */
  std::vector< std::vector< std::vector<double> > > values;

  /**
   * The vector containing the
   * gradients of finite element
   * functions in the quadrature
   * points.
   *
   * Example: nth solution, mth component, gradient at the qth quadrature point = gradients[n][m][q]
   */
  std::vector< std::vector< std::vector< Tensor<1,dim> > > > gradients;

  /// @deprecated Reference to gradients for compatibility.
  std::vector< std::vector< std::vector< Tensor<1,dim> > > >& derivatives;

  /**
   * The vector containing the
   * hessians of finite element
   * functions in the quadrature
   * points.
   *
   * Example: nth solution, mth component, hessian at the qth quadrature point = hessians[n][m][q]
   */
  std::vector< std::vector< std::vector< Tensor<2,dim> > > > hessians;

private:

  /**
   * Use the finite element
   * functions in #global_data
   * and fill the vectors
   * #values,
   * #gradients,
   * #hessians.
   */
  template<typename TYPE>
  void fill_local_data(std::vector< std::vector< std::vector<TYPE> > >& data,
                       VectorSelector&                                  selector,
                       bool                                             split_fevalues) const;

  /// The vector of smart pointers to FEVALUESBASE objects.
  std::vector< boost::shared_ptr<FEVALUESBASE> > fevalv;

  /**
   * (ZVN) In the mode # 1 (or in the main mode), AppFrame splits
   * an FEVALUESBASE object in several objects.
   * Sometimes there is a need to keep the unsplit
   * FEVALUESBASE object as well, even while in mode # 1.
   */
  boost::shared_ptr<FEVALUESBASE> fe_val_unsplit;
};

// --- implementation ---

template<int dim, typename FEVALUESBASE>
inline
IntegrationInfo<dim, FEVALUESBASE>::IntegrationInfo(const BlockInfo& block_info)
:
DoFInfo<dim>(block_info),
multigrid(false),
global_data(0, typeid(*this).name()),
derivatives(gradients),
fevalv(0)
{ }

template<int dim, typename FEVALUESBASE>
inline
IntegrationInfo<dim, FEVALUESBASE>::IntegrationInfo(const FEVectors& data,
                                                    const BlockInfo& block_info)
:
DoFInfo<dim>(block_info),
multigrid(false),
global_data(&data),
derivatives(gradients),
fevalv(0)
{ }

template<int dim, typename FEVALUESBASE>
template<typename FEVALUES>
inline
void
IntegrationInfo<dim, FEVALUESBASE>::initialize(const FEVALUES*,
                                               const FiniteElement<dim>&                       fe,
                                               const Mapping<dim>&                             mapping,
                                               const Quadrature<FEVALUES::integral_dimension>& quadrature,
                                               const UpdateFlags                               flags)
{
  if( this->block_info->local_renumbering.size() != 0 ) // mode # 1 - the main mode
  {
    fevalv.resize(fe.n_base_elements());
    for(unsigned int i = 0; i < fevalv.size(); ++i)
    {
      fevalv[i] = boost::shared_ptr<FEVALUESBASE>( new FEVALUES(mapping,
                                                                fe.base_element(i),
                                                                quadrature,
                                                                flags)
                                                 );
    }

    // --- (ZVN) fe_val_unsplit initialization ---

    fe_val_unsplit = boost::shared_ptr<FEVALUESBASE>( new FEVALUES(mapping,
                                                                   fe,
                                                                   quadrature,
                                                                   flags)
                                                    );
  }
  else // mode # 2
  {
    fevalv.resize(1);
    fevalv[0] = boost::shared_ptr<FEVALUESBASE>( new FEVALUES(mapping,
                                                              fe,
                                                              quadrature,
                                                              flags)
                                               );
  }

  values.resize(global_data->n_vectors(),
                std::vector< std::vector<double> >(fe.n_components(),
                                                   std::vector<double>(quadrature.size())
                                                  )
               );
  gradients.resize(global_data->n_vectors(),
                   std::vector< std::vector< Tensor<1,dim> > >(fe.n_components(),
                                                               std::vector< Tensor<1,dim> >(quadrature.size())
                                                              )
                  );
  hessians.resize(global_data->n_vectors(),
                  std::vector< std::vector< Tensor<2,dim> > >(fe.n_components(),
                                                              std::vector< Tensor<2,dim> >(quadrature.size())
                                                             )
                 );
}

template<int dim, typename FEVALUESBASE>
template<typename DHCellIterator>
inline
void
IntegrationInfo<dim, FEVALUESBASE>::reinit(const DHCellIterator& c)
{
  DoFInfo<dim>::reinit(c);

  for(unsigned int i = 0; i < fevalv.size(); ++i)
  {
    FEVALUESBASE&  fe_values_base = *fevalv[i];
    FEValues<dim>& fe_values      = dynamic_cast<FEValues<dim>&>(fe_values_base);
    fe_values.reinit(this->cell);
  }

  // --- (ZVN) ---

  FEVALUESBASE&  fe_values_base_unsplit = *fe_val_unsplit;
  FEValues<dim>& fe_values_unsplit      = dynamic_cast<FEValues<dim>&>(fe_values_base_unsplit);
  fe_values_unsplit.reinit(this->dof_active_cell);

  // --- (ZVN) ---

  const bool split_fevalues = this->block_info->local_renumbering.size() != 0;
  fill_local_data(values,
                  selector,
                  split_fevalues);
  fill_local_data(gradients,
                  selector,
                  split_fevalues);
  fill_local_data(hessians,
                  selector,
                  split_fevalues);
}

template<int dim, typename FEVALUESBASE>
template<typename DHCellIterator, typename DHFaceIterator>
inline
void
IntegrationInfo<dim, FEVALUESBASE>::reinit(const DHCellIterator& c,
                                           const DHFaceIterator& f,
                                           const unsigned int    fn)
{
  DoFInfo<dim>::reinit(c,
                       f,
                       fn);

  for(unsigned int i = 0; i < fevalv.size(); ++i)
  {
    FEVALUESBASE&      fe_values_base = *fevalv[i];
    FEFaceValues<dim>& fe_face_values = dynamic_cast<FEFaceValues<dim>&>(fe_values_base);
    fe_face_values.reinit(this->cell,
                          fn);
  }

  // --- (ZVN) ---

  FEVALUESBASE&      fe_values_base_unsplit = *fe_val_unsplit;
  FEFaceValues<dim>& fe_face_values_unsplit = dynamic_cast<FEFaceValues<dim>&>(fe_values_base_unsplit);
  fe_face_values_unsplit.reinit(this->dof_active_cell,
                                fn);

  // --- (ZVN) ---

  const bool split_fevalues = this->block_info->local_renumbering.size() != 0;
  fill_local_data(values,
                  selector,
                  split_fevalues);
  fill_local_data(gradients,
                  selector,
                  split_fevalues);
  fill_local_data(hessians,
                  selector,
                  split_fevalues);
}

template<int dim, typename FEVALUESBASE>
template<typename DHCellIterator, typename DHFaceIterator>
inline
void
IntegrationInfo<dim, FEVALUESBASE>::reinit(const DHCellIterator& c,
                                           const DHFaceIterator& f,
                                           const unsigned int    fn,
                                           const unsigned int    sn)
{
  DoFInfo<dim>::reinit(c,
                       f,
                       fn,
                       sn);

  for(unsigned int i = 0; i < fevalv.size(); ++i)
  {
    FEVALUESBASE&         fe_values_base    = *fevalv[i];
    FESubfaceValues<dim>& fe_subface_values = dynamic_cast<FESubfaceValues<dim>&>(fe_values_base);
    fe_subface_values.reinit(this->cell,
                             fn,
                             sn);
  }

  // --- (ZVN) ---

  FEVALUESBASE&         fe_values_base_unsplit    = *fe_val_unsplit;
  FESubfaceValues<dim>& fe_subface_values_unsplit = dynamic_cast<FESubfaceValues<dim>&>(fe_values_base_unsplit);
  fe_subface_values_unsplit.reinit(this->dof_active_cell,
                                   fn,
                                   sn);

  // --- (ZVN) ---

  const bool split_fevalues = this->block_info->local_renumbering.size() != 0;
  fill_local_data(values,
                  selector,
                  split_fevalues);
  fill_local_data(gradients,
                  selector,
                  split_fevalues);
  fill_local_data(hessians,
                  selector,
                  split_fevalues);
}

template<int dim, typename FEVALUESBASE>
template<typename TYPE>
inline
void
IntegrationInfo<dim, FEVALUESBASE>::fill_local_data(std::vector< std::vector< std::vector<TYPE> > >& data,
                                                    bool                                             split_fevalues) const
{
  if(split_fevalues)
  {
    std::vector< std::vector<TYPE> > local_data;

    unsigned int comp = 0;
    for(unsigned int b = 0; b < this->block_info->base_element.size(); ++b)
    {
      const unsigned int       no_fe          = this->block_info->base_element[b];
      const FEValuesBase<dim>& fe_values_base = this->fe(no_fe);

      const unsigned int n_comp = fe_values_base.get_fe().n_components();
      local_data.resize(n_comp,
                        std::vector<TYPE>(fe_values_base.n_quadrature_points));

      for(unsigned int i = 0; i < this->global_data->n_vectors(); ++i)
      {
        const FEVector& src = this->global_data->vector(i);
        MeshWorker::InfoObjects::fill_data(fe_values_base,
                                           src,
                                           this->indices,
                                           this->block_info->local.block_start(b),
                                           fe_values_base.dofs_per_cell,
                                           local_data);

        for(unsigned int c = 0; c < local_data.size(); ++c)
          for(unsigned int k = 0; k < local_data[c].size(); ++k)
            data[i][comp+c][k] = local_data[c][k];
      }

      comp += n_comp;
    }
  }
  else
  {
    for(unsigned int i = 0; i < this->global_data->n_vectors(); ++i)
    {
      const FEVector&          src            = this->global_data->vector(i);
      const FEValuesBase<dim>& fe_values_base = this->fe();

      MeshWorker::InfoObjects::fill_data(fe_values_base,
                                         src,
                                         this->indices,
                                         0,
                                         fe_values_base.get_fe().dofs_per_cell,
                                         data[i]);
    }
  }
}

template<int dim, typename FEVALUESBASE>
template<typename TYPE>
inline
void
IntegrationInfo<dim, FEVALUESBASE>::fill_local_data(std::vector< std::vector< std::vector<TYPE> > >& data,
                                                    VectorSelector&                                  selector,
                                                    bool                                             split_fevalues) const
{
  if(split_fevalues)
  {
    std::vector< std::vector<TYPE> > local_data;

    unsigned int comp = 0;
    for(unsigned int b = 0; b < this->block_info->base_element.size(); ++b)
    {
      const unsigned int       no_fe          = this->block_info->base_element[b];
      const FEValuesBase<dim>& fe_values_base = this->fe(no_fe);

      const unsigned int n_comp = fe_values_base.get_fe().n_components();
      local_data.resize(n_comp,
                        std::vector<TYPE>(fe_values_base.n_quadrature_points));

      const unsigned int n = selector_size(selector,
                                           local_data[0]);

      for(unsigned int i = 0; i < n; ++i)
      {
        const FEVector& src = this->global_data->vector( selector_index(selector, i, local_data[0]) );
        MeshWorker::InfoObjects::fill_data(fe_values_base,
                                           src,
                                           this->indices,
                                           this->block_info->local.block_start(b),
                                           fe_values_base.dofs_per_cell,
                                           local_data);

        for(unsigned int c = 0; c < local_data.size(); ++c)
          for(unsigned int k = 0; k < local_data[c].size(); ++k)
            data[i][comp+c][k] = local_data[c][k];
      }

      comp += n_comp;
    }
  }
  else
  {
    for(unsigned int i = 0; i < this->global_data->n_vectors(); ++i)
    {
      const FEVector&          src            = this->global_data->vector(i);
      const FEValuesBase<dim>& fe_values_base = this->fe();

      MeshWorker::InfoObjects::fill_data(fe_values_base,
                                         src,
                                         this->indices,
                                         0,
                                         fe_values_base.get_fe().dofs_per_cell,
                                         data[i]);
    }
  }
}

/**
 * A simple container collecting five info objects required by the
 * integration loops.
 *
 * @author Guido Kanschat, 2009
 */

template<int dim>
class IntegrationInfoBox
{
public:

  /**
   * Constructor.
   */
  IntegrationInfoBox(const BlockInfo& block_info);

  /**
   * Function that initializes
   * five info objects required by the
   * integration loops.
   */
  template<typename WORKER>
  void initialize(const WORKER&             integrator,
                  const FiniteElement<dim>& fe,
                  const Mapping<dim>&       mapping,
                  const FEVectors&          global_data);

  /**
   * Five info objects required by the
   * integration loops.
   */
  IntegrationInfo<dim, FEValuesBase<dim>     > cell_info;
  IntegrationInfo<dim, FEFaceValuesBase<dim> > bdry_info;
  IntegrationInfo<dim, FEFaceValuesBase<dim> > face_info;
  IntegrationInfo<dim, FEFaceValuesBase<dim> > subface_info;
  IntegrationInfo<dim, FEFaceValuesBase<dim> > neighbor_info;
};

// --- implementation ---

template<int dim>
inline
IntegrationInfoBox<dim>::IntegrationInfoBox(const BlockInfo& block_info)
:
cell_info(block_info),
bdry_info(block_info),
face_info(block_info),
subface_info(block_info),
neighbor_info(block_info)
{ }

template<int dim>
template<typename WORKER>
inline
void
IntegrationInfoBox<dim>::initialize(const WORKER&             integrator,
                                    const FiniteElement<dim>& fe,
                                    const Mapping<dim>&       mapping,
                                    const FEVectors&          global_data)
{
  FEValues<dim>*        fe_values         = 0;
  FEFaceValues<dim>*    fe_face_values    = 0;
  FESubfaceValues<dim>* fe_subface_values = 0;

  cell_info.initialize_selector(integrator.cell_selector);
  bdry_info.initialize_selector(integrator.bdry_selector);
  face_info.initialize_selector(integrator.face_selector);
  subface_info.initialize_selector(integrator.face_selector);
  neighbor_info.initialize_selector(integrator.face_selector);

  cell_info.initialize_data(global_data);
  bdry_info.initialize_data(global_data);
  face_info.initialize_data(global_data);
  subface_info.initialize_data(global_data);
  neighbor_info.initialize_data(global_data);

  cell_info.initialize(fe_values,
                       fe,
                       mapping,
                       integrator.cell_quadrature,
                       integrator.cell_flags);

  bdry_info.initialize(fe_face_values,
                       fe,
                       mapping,
                       integrator.bdry_quadrature,
                       integrator.face_flags);

  face_info.initialize(fe_face_values,
                       fe,
                       mapping,
                       integrator.face_quadrature,
                       integrator.face_flags);

  subface_info.initialize(fe_subface_values,
                          fe,
                          mapping,
                          integrator.face_quadrature,
                          integrator.face_flags);

  neighbor_info.initialize(fe_face_values,
                           fe,
                           mapping,
                           integrator.face_quadrature,
                           integrator.ngbr_flags);
}

} // namespace InfoObjects

} // namespace MeshWorker

#endif