// ------------------------------------------------------------------
//
// $Id: mesh_worker_workers.h 1348 2013-08-16 00:45:07Z secanell $
//
// Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat.
// Copyright (C) 2012 by Valentin N. Zingan, University of Alberta.
//
// This file is subject to QPL and may not be  distributed
// without copyright and license information. Please refer
// to the file deal.II/doc/license.html for the  text  and
// further information on this license.
//
// ------------------------------------------------------------------

#ifndef _DEALII_APPFRAME_MESH_WORKER_WORKERS_H_
#define _DEALII_APPFRAME_MESH_WORKER_WORKERS_H_

#include <appframe/mesh_worker_info.h>

// ################################################################################################

/**
 * A collection of functions and classes for the mesh loops that are
 * an ubiquitous part of each finite element program.
 *
 * The workhorse of this namespace is the loop() function, which implements a
 * completely generic loop over all mesh entities (cells and faces).
 *
 * The loop() depends on certain objects handed to it as
 * arguments. These objects can be structured into categories
 * corresponding to the namespaces where they are defined.
 *
 * @author Guido Kanschat, 2009
 */

// ##### CATEGORY-2 ################################################# WORKER_OBJECTS ##############
//
// ##### MeshWorker::WorkerObjects::LocalWorker       #####
// ##### MeshWorker::WorkerObjects::IntegrationWorker #####

namespace MeshWorker // # 2
{

namespace WorkerObjects
{

// --- 1 ---

/**
 * Template for a class for those classes which do the actual work on cells
 * and faces.
 *
 * This class can serve as a base class for the actual worker class,
 * since, while we do not use virtual functions, we provide the
 * necessary interface for the mesh loops and the InfoObjects::DoFInfo
 * class here. Thus, they do not have to be reprogrammed.
 *
 * In particular, the mesh loops will require data elements
 * #boundary_fluxes and #interior_fluxes to determine whether the
 * loop over faces will be started at all.
 *
 * @author Guido Kanschat, 2009
 */

template<int dim>
class LocalWorker
{
public:

  /**
   * Constructor, setting
   * #boundary_fluxes
   * and
   * #interior_fluxes
   * to
   * @p true.
   */
  LocalWorker()
  {
    boundary_fluxes = true;
    interior_fluxes = true;
  }

  /**
   * Do the work on a cell.
   */
  void cell(const InfoObjects::DoFInfo<dim>& cell);

  /**
   * Do the work on a boundary face.
   */
  void bdry(const InfoObjects::DoFInfo<dim>& face);

  /**
   * Do the work on an interior face.
   */
  void face(const InfoObjects::DoFInfo<dim>& face1,
            const InfoObjects::DoFInfo<dim>& face2);

  /**
   * Computations on boundary
   * faces are on/off.
   */
  bool boundary_fluxes;

  /**
   * Computations on interior
   * faces are on/off.
   */
  bool interior_fluxes;
};

// --- 2 ---

/**
 * Basic worker class for integration of residuals.
 *
 * @author Guido Kanschat, 2009
 */

template<int dim>
class IntegrationWorker : public LocalWorker<dim>
{
public:

  /**
   * Constructor, which sets
   * the UpdateFlags flags
   * to their default values.
   */
  IntegrationWorker();

  /**
   * Initialize the
   * VectorSelector objects
   * #cell_selector,
   * #bdry_selector,
   * and
   * #face_selector.
   * If no selectors
   * are used, then values for
   * all named FEVector vectors in
   * FEVectors
   * will be computed in all quadrature
   * points.
   *
   * This function will also
   * add UpdateFlags to the
   * flags stored in this class.
   */
  void initialize_selectors(const VectorSelector& cs,
                            const VectorSelector& bs,
                            const VectorSelector& fs);

  /**
   * Add a FEVector name as it is
   * stored in FEVectors to some or
   * all selectors.
   */
  void add_selector(const std::string& name,
                    bool               values,
                    bool               gradients,
                    bool               hessians,
                    bool               cell,
                    bool               bdry,
                    bool               face);

  /**
   * Add additional UpdateFlags flags.
   */
  void add_update_flags(const UpdateFlags flags,
                        bool              cell = true,
                        bool              bdry = true,
                        bool              face = true,
                        bool              ngbr = true)
  {
    if(cell) cell_flags |= flags;
    if(bdry) bdry_flags |= flags;
    if(face) face_flags |= flags;
    if(ngbr) ngbr_flags |= flags;
  }

  /**
   * Assign n-point Gauss
   * quadratures to each of the
   * quadrature rules. Here, a
   * size of zero points means
   * that no loop over these grid
   * entities should be
   * performed.
   */
  void initialize_gauss_quadrature(unsigned int n_cell_points,
                                   unsigned int n_bdry_points,
                                   unsigned int n_face_points)
  {
    cell_quadrature = QGauss<dim>(n_cell_points);
    bdry_quadrature = QGauss<dim-1>(n_bdry_points);
    face_quadrature = QGauss<dim-1>(n_face_points);
  }

  /**
   * Select the FEVector vectors from
   * FEVectors
   * that should be computed in
   * the quadrature points on cells.
   */
  MeshWorker::VectorSelector cell_selector;

  /**
   * Select the FEVector vectors from
   * FEVectors
   * that should be computed in
   * the quadrature points on boundary faces.
   */
  MeshWorker::VectorSelector bdry_selector;

  /**
   * Select the FEVector vectors from
   * FEVectors
   * that should be computed in
   * the quadrature points on interior faces.
   */
  MeshWorker::VectorSelector face_selector;

  /**
   * The quadrature rule used
   * on cells.
   */
  Quadrature<dim> cell_quadrature;

  /**
   * The quadrature rule used
   * on boundary faces.
   */
  Quadrature<dim-1> bdry_quadrature;

  /**
   * The quadrature rule used
   * on interior faces.
   */
  Quadrature<dim-1> face_quadrature;

  /**
   * The set of update flags
   * for cell integration.
   *
   * Defaults to
   * #update_JxW_values.
   */
  UpdateFlags cell_flags;

  /**
   * The set of update flags
   * for boundary face integration.
   *
   * Defaults to
   * #update_normal_vectors
   * and
   * #update_JxW_values.
   */
  UpdateFlags bdry_flags;

  /**
   * The set of update flags
   * for interior face integration.
   *
   * Defaults to
   * #update_normal_vectors
   * and
   * #update_JxW_values.
   */
  UpdateFlags face_flags;

  /**
   * The set of update flags
   * for interior face integration.
   *
   * Defaults to
   * #update_default, since
   * quadrature weights are
   * taken from the interior
   * of the neighboring cell.
   */
  UpdateFlags ngbr_flags;
};

// --- implementation ---

template<int dim>
inline
IntegrationWorker<dim>::IntegrationWorker()
:
LocalWorker<dim>()
{
  cell_flags = update_JxW_values;
  bdry_flags = UpdateFlags(update_normal_vectors | update_JxW_values);
  face_flags = bdry_flags;
  ngbr_flags = update_default;
}

template<int dim>
inline
void
IntegrationWorker<dim>::initialize_selectors(const VectorSelector& cs,
                                             const VectorSelector& bs,
                                             const VectorSelector& fs)
{
  cell_selector = cs;
  bdry_selector = bs;
  face_selector = fs;

  if(cell_selector.has_values()    != 0) cell_flags |= update_values;
  if(cell_selector.has_gradients() != 0) cell_flags |= update_gradients;
  if(cell_selector.has_hessians()  != 0) cell_flags |= update_hessians;

  if(bdry_selector.has_values()    != 0) bdry_flags |= update_values;
  if(bdry_selector.has_gradients() != 0) bdry_flags |= update_gradients;
  if(bdry_selector.has_hessians()  != 0) bdry_flags |= update_hessians;

  if(face_selector.has_values()    != 0) face_flags |= update_values;
  if(face_selector.has_gradients() != 0) face_flags |= update_gradients;
  if(face_selector.has_hessians()  != 0) face_flags |= update_hessians;

  if(face_selector.has_values()    != 0) ngbr_flags |= update_values;
  if(face_selector.has_gradients() != 0) ngbr_flags |= update_gradients;
  if(face_selector.has_hessians()  != 0) ngbr_flags |= update_hessians;
}

template<int dim>
inline
void
IntegrationWorker<dim>::add_selector(const std::string& name,
                                     bool               values,
                                     bool               gradients,
                                     bool               hessians,
                                     bool               cell,
                                     bool               bdry,
                                     bool               face)
{
  if(cell) cell_selector.add_vector(name, values, gradients, hessians);
  if(bdry) bdry_selector.add_vector(name, values, gradients, hessians);
  if(face) face_selector.add_vector(name, values, gradients, hessians);

  if(cell_selector.has_values()    != 0) cell_flags |= update_values;
  if(cell_selector.has_gradients() != 0) cell_flags |= update_gradients;
  if(cell_selector.has_hessians()  != 0) cell_flags |= update_hessians;

  if(bdry_selector.has_values()    != 0) bdry_flags |= update_values;
  if(bdry_selector.has_gradients() != 0) bdry_flags |= update_gradients;
  if(bdry_selector.has_hessians()  != 0) bdry_flags |= update_hessians;

  if(face_selector.has_values()    != 0) face_flags |= update_values;
  if(face_selector.has_gradients() != 0) face_flags |= update_gradients;
  if(face_selector.has_hessians()  != 0) face_flags |= update_hessians;

  if(face_selector.has_values()    != 0) ngbr_flags |= update_values;
  if(face_selector.has_gradients() != 0) ngbr_flags |= update_gradients;
  if(face_selector.has_hessians()  != 0) ngbr_flags |= update_hessians;
}

} // namespace WorkerObjects

} // namespace MeshWorker # 2

#endif