//---------------------------------------------------------------------------
//    $Id: newton.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__appframe__newton_h
#define __deal2__appframe__newton_h

#include <appframe/base.h>
#include <appframe/residual.h>
#include <lac/solver_control.h>


namespace dealii
{
  class ParameterHandler;
  template <typename T> class BlockVector;
}

namespace AppFrame
{
  template <int dim> class DoFApplication;
  template <int dim> class LocalIntegrator;
/**
 * Application class performing Newton's iteration.
 *
 * @author Guido Kanschat, 2006
 */
  class Newton : public ApplicationCopy
  {
    public:
				       /**
					* The Event set by Newton if
					* convergence is becoming bad
					* and a new matrix should be
					* assembled.
					*/
      static const Event bad_derivative;
      
				       /**
					* Constructor, receiving the
					* application computing the
					* residual and solving the
					* linear problem.
					*/
      Newton (ApplicationBase& app);

				       /**
					* Constructor receiving the
					* application for solving the
					* linear problem and the
					* residual for computing
					* residuals.
					*/
      Newton (ApplicationBase& app, const Residual& res);
      
      virtual double residual(FEVector& start,
			      const FEVectors& rhs);
	
				       /**
					* The actual Newton solver.
					*/
      virtual void solve(FEVector& start,
			 const FEVectors& rhs);
	
      virtual void declare_parameters (ParameterHandler& param);
	
				       /**
					* Read the parameters.
					*/
      void _initialize (ParameterHandler& param);
	
      virtual void initialize (ParameterHandler& param);
	
				       /**
					* Set the maximal residual
					* reduction allowed without
					* triggering assembling in the
					* next step. Return the
					* previous value.
					*/
      double threshold(double new_value);

				       /**
					* Control object for the
					* Newton iteration.
					*/
      ReductionControl control;
    private:
				       /**
					* Default residual object if
					* the function
					* ApplicationBase::residual()
					* shoud be used.
					*/
      ApplicationResidual default_residual;
      
				       /**
					* The Residual object.
					*/
      SmartPointer<const Residual> compute_residual;
      
				       /**
					* This flag is set by the
					* function assemble(),
					* indicating that the matrix
					* must be assembled anew upon
					* start.
					*/
      bool assemble_now;

				       /**
					* A flag used to decide how
					* many stepsize iteration
					* should be made. Default is
					* the original value of 21.
					*
					* Enter zero here to turn of
					* stepsize control.
					*
					* @note Controlled by
					* <tt>Stepsize iterations</tt> in
					* parameter file
					*/
      unsigned int n_stepsize_iterations;
      
				       /**
					* Threshold for re-assembling matrix.
					*
					* If the quotient of two
					* consecutive residuals is
					* smaller than this threshold,
					* the system matrix is not
					* assembled in this step.
					*
					* @note This parameter should be
					* adjusted to the residual gain
					* of the inner solver.
					*/
      double assemble_threshold;
				       /**
					* Print updated solution after
					* each step into file
					* <tt>Newton_uNNN</tt>?
					*/
      bool debug_solution;
				       /**
					* Print Newton update after
					* each step into file
					* <tt>Newton_dNNN</tt>?
					*/
      bool debug_update;
				       /**
					* Print Newton residual after
					* each step into file
					* <tt>Newton_rNNN</tt>?
					*/
      bool debug_residual;
				       /**
					* Write debug output to
					* #deallog; the higher the
					* number, the more output.
					*/
      unsigned int debug;	
  };

}




#endif
