//---------------------------------------------------------------------------
//    $Id: residual_shop_bdry.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __appshop__residual_shop_bdry_h
#define __appshop__residual_shop_bdry_h

#include <appframe/app_shop.h>

/**
 * \file
 * \brief Residual integration routines on cells for standard bilinear forms.
 */

namespace AppShop
{
  namespace Residual
  {
    namespace Boundary
    {
				       /**
					* Upwind flux at boundary faces.
					*/
      template <int dim>
      void advection (
	Vector<double>& result,
	const std::vector<double>& input,
	const std::vector<Vector<double> >& data,
	unsigned int data_index,
	const FEFaceValuesBase<dim>& fe,
	const std::vector<Point<dim> >& velocity,
	double factor = 1.);
      
				       /**
					* Upwind flux at boundary faces.
					*/
      template <int dim>
      void advection (
	Vector<double>& result,
	const std::vector<double>& input,
	const std::vector<double>& data,
	const FEFaceValuesBase<dim>& fe,
	const VectorSlice<const std::vector<std::vector<double> > >& velocity,
	double factor = 1.);
      
				       /**
					* Vector valued upwind flux at
					* boundary faces.
					*
					* \f[
					* \left< u^\uparrow \mathbf w,
					* v \mathbf n \right>
					* \f]
					*/
      template <int dim>
      void advection (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const VectorSlice<const std::vector<std::vector<double> > >& data,
	const FEFaceValuesBase<dim>& fe,
	const VectorSlice<const std::vector<std::vector<double> > >& velocity,
	double factor = 1.);
				       /**
					* Product of solution and normal vector.
					*
					* \f[
					* \left< u\mathbf n, \mathbf v\right>
					* \f]
					*/
      template <int dim>
      void u_times_n (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEFaceValuesBase<dim>& fe,
	const FEFaceValuesBase<dim>& fetest,
	double factor = 1.);
      
				       /**
					* Product of solution and
					* normal vector, vector
					* valued.
					*
					* \f[
					* \left< u_i\mathbf n, \mathbf v\right>
					* \quad i=1\ldots d
					* \f]
					*/
      template <int dim>
      void u_times_n (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEFaceValuesBase<dim>& fe,
	const FEFaceValuesBase<dim>& fetest,
	double factor = 1.);
      
				       /**
					* Product of solution and normal vector.
					*
					* \f[
					* \left<a u\mathbf n, \mathbf v\right>
					* \f]
					*/
      template <int dim>
      void u_times_n_scaled (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEFaceValuesBase<dim>& fe,
	const FEFaceValuesBase<dim>& fetest,
	const std::vector<double>& a,
	bool divide = false);

				       /**
					* Dot product of solution and normal vector.
					*
					* \f[
					* \left<\mathbf u\cdot \mathbf n, v\right>
					* \f]
					*/
      template <int dim>
      void u_dot_n (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEFaceValuesBase<dim>& fe,
	const FEFaceValuesBase<dim>& fetest,
	double factor = 1.);

				       /**
					* Dot product of solution and normal vector.
					*
					* \f[
					* \left<a \mathbf u\cdot \mathbf n, v\right>
					* \f]
					*/
      template <int dim>
      void u_dot_n_scaled (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEFaceValuesBase<dim>& fe,
	const FEFaceValuesBase<dim>& fetest,
	const std::vector<double>& a,
	bool divide = false);

				       /**
					* Elasticity.
					*/
      template <int dim>
      void eps (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEFaceValuesBase<dim>& fe,
	const FEFaceValuesBase<dim>& fetest);
				       /**
					* Elasticity.
					*/
      template <int dim>
      void eps_div (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEFaceValuesBase<dim>& fe,
	const FEFaceValuesBase<dim>& fetest);
      
				       /**
					* Boundary mass matrix.
					*
					* \f[
					* \left<u, v\right>
					* \f]					
					*/
      template <int dim>
      void jump (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEFaceValuesBase<dim>& fe,
	const double factor = 1.);
      
      
				       /**
					* Boundary mass.
					*
					* \f[
					* \left<a u, v\right>
					* \f]					
					*/
      template <int dim>
      void jump_scaled (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEFaceValuesBase<dim>& fe,
	const std::vector<double>& a,
	const double factor = 1.,
	bool divide = false);
				       /**
					* Boundary jump, vector valued.
					*
					* \f[
					* \left<a (u-u^D), v\right>
					* \f]
					*/
      template <int dim>
      void jump_scaled (
	Vector<double>& result,
	const std::vector<double>& input,
	const std::vector<Vector<double> >& data,
	const unsigned int data_index,
	const FEFaceValuesBase<dim>& fe,
	const std::vector<double>& factor,
	const double factor2 = 1.,
	bool divide = false);
      
				       /**
					* Weak boundary condition for
					* the Laplacian by Nitsche.
					*/
      template <int dim>
      void nitsche (
	Vector<double>& result,
	const std::vector<double>& u,
	const std::vector<Tensor<1,dim> >& Du,
	const std::vector<double>& data,
	const FEFaceValuesBase<dim>& fe,
	double penalty,
	double factor = 1.);      
      
				       /**
					* Weak boundary condition for
					* the Laplacian by Nitsche for
					* vector valued elements.
					*/
      template <int dim>
      void nitsche (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& u,
	const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& Du,
	const VectorSlice<const std::vector<std::vector<double> > >& data,
	const FEFaceValuesBase<dim>& fe,
	double penalty,
	double factor = 1.);      
      
				       /**
					* Weak boundary condition for
					* the Laplacian by Nitsche.
					*/
      template <int dim>
      void nitsche (
	Vector<double>& result,
	const std::vector<double>& u,
	const std::vector<Tensor<1,dim> >& Du,
	const std::vector<Vector<double> >& data,
	const unsigned int data_index,
	const FEFaceValuesBase<dim>& fe,
	double penalty,
	double factor = 1.);      
      
				       /**
					* Weak boundary condition for
					* the Laplacian by Nitsche.
					*/
      template <int dim>
      void nitsche_scaled (
	Vector<double>& result,
	const std::vector<double>& u,
	const std::vector<Tensor<1,dim> >& Du,
	const std::vector<Vector<double> >& data,
	const unsigned int data_index,
	const FEFaceValuesBase<dim>& fe,
	const std::vector<double>& factor,
	double penalty);
      
				       /**
					* Weak boundary condition for
					* the Laplacian by Nitsche.
					*/
      template <int dim>
      void nitsche_scaled (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& u,
	const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& Du,
	const std::vector<Vector<double> >& data,
	const unsigned int data_index,
	const FEFaceValuesBase<dim>& fe,
	const std::vector<double>& factor,
	double penalty);      
    }
  }
}

#endif
