//---------------------------------------------------------------------------
//    $Id: residual_shop_cell.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __appshop__residual_shop_cell_h
#define __appshop__residual_shop_cell_h

#include <appframe/app_shop.h>

/**
 * \file
 * \brief Residual integration routines on cells for standard bilinear forms.
 */

namespace AppShop
{
  namespace Residual
  {
    namespace Cell
    {
				       /**
					* Linear advection in divergence
					* form. The derivative is on the
					* test function.
					*
					* \f[
					* -(u \mathbf w, \nabla v)
					* \f]
					*/
      template <int dim>
      void advection(
	Vector<double>& result,
	const std::vector<double>& input,
	const FEValuesBase<dim>& fe,
	const VectorSlice<const std::vector<std::vector<double> > >& velocity,
	double factor = 1.);

				       /**
					* Linear advection in
					* divergence form for vector
					* valued elements. The
					* derivative is on the test
					* function.
					*
					* \f[
					* -(u \mathbf w, \nabla v)
					* \f]
					*/
      template <int dim>
      void advection(
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEValuesBase<dim>& fe,
	const VectorSlice<const std::vector<std::vector<double> > >& velocity,
	double factor = 1.);

				       /**
					* Linear advection in divergence
					* form. The derivative is on the
					* test function.
					*
					* \f[
					* -(u \mathbf w, \nabla v)
					* \f]
					*/
      template <int dim>
      void advection(
	Vector<double>& result,
	const std::vector<double>& input,
	const FEValuesBase<dim>& fe,
	const std::vector<Point<dim> >& velocity,
	double factor = 1.);
				       /**
					* Directional derivative in strong form
					*
					* \f[
					* (\partial_i u, v)
					* \f]
					*/
      template <int dim>
      void di (
	Vector<double>& result,
	const std::vector<Tensor<1,dim> >& input,
	unsigned int direction,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);
				       /**
					* Directional derivative in weak form
					*
					* \f[
					* (-u,\partial_i v)
					* \f]
					*/
      template <int dim>
      void di (
	Vector<double>& result,
	const std::vector<double>& input,
	unsigned int direction,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);
				       /**
					* Divergence in strong form
					*
					* \f[
					* (\nabla \cdot \mathbf u, v)
					* \f]
					*/
      template <int dim>
      void div (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);
				       /**
					* Divergence in weak form
					*
					* \f[
					* -(\mathbf u, \nabla v)
					* \f]
					*/
      template <int dim>
      void div (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);
      				       /**
					* Scaled divergence in weak form
					*
					* \f[
					* -(a \mathbf u, \nabla v)
					* \f]
					*/
      template <int dim>
      void div_scaled (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	const std::vector<double>& a); 
				       /**
					* Gradient in strong form. The
					* element <tt>fetest</tt> must
					* be vector valued with
					* <tt>dim</tt> components.
					*
					* \f[
					* (\nabla u, \mathbf v)
					* \f]
					*/
      template <int dim>
      void grad (
	Vector<double>& result,
	const std::vector<Tensor<1,dim> >& input,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);
    
				       /**
					* Gradient in weak form. The
					* element <tt>fetest</tt> must
					* be vector valued with
					* <tt>dim</tt> components.
					*
					* \f[
					* -(u, \nabla \cdot \mathbf v)
					* \f]
					*/
      template <int dim>
      void grad (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest,
	double factor = 1.);
    
      				       /**
					* Elasticity.
					*/
      template <int dim>
      void eps_div (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest);
      				       /**
					* Elasticity.
					*/
      template <int dim>
      void eps (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEValuesBase<dim>& fe,
	const FEValuesBase<dim>& fetest);

				       /**
					* Laplacian in weak form
					*
					* \f[
					* -(\nu \nabla u, \nabla v)
					* \f]
					*/
      template <int dim>
      void laplacian_scaled (
	Vector<double>& result,
	const std::vector<Tensor<1,dim> >& input,
	const FEValuesBase<dim>& fe,
	const std::vector<double>& nu,
	bool divide = false);

				       /**
					* Laplacian in weak form, vector valued
					*
					* \f[
					* -(\nu \nabla u_i, \nabla v)
					* \quad i=1\ldots d
					* \f]
					*/
      template <int dim>
      void laplacian_scaled (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
	const FEValuesBase<dim>& fe,
	const std::vector<double>& nu,
	bool divide = false);
      
      
				       /**
					* Laplacian in weak form
					*
					* \f[
					* -(\nabla u, \nabla v)
					* \f]
					*/
      template <int dim>
      void laplacian  (
	Vector<double>& result,
	const std::vector<Tensor<1,dim> >& input,
	const FEValuesBase<dim>& fe,
	double factor = 1.);

                                       /**
					* Laplacian in weak form
					* with a factor that allows for anisotropy
					* \f[
					* -(\nabla u, K \nabla v)
					* \f]
					*/
      template <int dim>
      void laplacian  (
	Vector<double>& result,
	const std::vector<Tensor<1,dim> >& input,
	const FEValuesBase<dim>& fe,
	const Tensor<2,dim> factor);
				       /**
					* Vector-valued Laplacian in weak form
					*
					* \f[
					* -(\nabla u, \nabla v)
					* \f]
					*/
      template <int dim>
      void laplacian  (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
	const FEValuesBase<dim>& fe,
	double factor = 1.);
				       /**
					* Mass residual
					*
					* \f[
					* (f,v)
					* \f]
					*/
      template <int dim>
      void mass (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEValuesBase<dim>& fe,
	double factor = 1.);

				       /**
					* Mass residual for a slice of
					* a vector valued right hand
					* side.
					*
					* \f[
					* (f,v)
					* \f]
					*/
      template <int dim>
      void mass (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEValuesBase<dim>& fe,
	double factor = 1.);

				       /**
					* Mass residual for a vector
					* valued right hand side in
					* the format produced by
					* Function::vector_value_list.
					*
					* \f[
					* (f,v)
					* \f]
					*/
      template <int dim>
      void mass (
	Vector<double>& result,
	const std::vector<Vector<double> >& input,
	const unsigned int start_component,
	const FEValuesBase<dim>& fe,
	double factor = 1.);

				       /**
					* Residual for a nonlinear
					* power of u
					*
					* \f[
					* (u^p,v)
					* \f]
					*/
      template <int dim>
      void power (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEValuesBase<dim>& fe,
	const double power,
	double factor = 1.);

				       /**
					* Residual of an implicit Euler timestep
					*/
      template <int dim>
      void time_step (
	Vector<double>& result,
	const std::vector<double>& input,
	const std::vector<double>& last_time,
	const FEValuesBase<dim>& fe,
	const double time_step);

				       /**
					* Residual of an implicit Euler timestep
					*/
      template <int dim>
      void time_step (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const VectorSlice<const std::vector<std::vector<double> > >& last_time,
	const FEValuesBase<dim>& fe,
	const double time_step);

				       /**
					* Scaled mass residual
					*/
      template <int dim>
      void mass_scaled (
	Vector<double>& result,
	const std::vector<double>& input,
	const FEValuesBase<dim>& fe,
	const std::vector<double>& factor,
	bool divide = false);

				       /**
					* Scaled mass residual
					*/
      template <int dim>
      void mass_scaled (
	Vector<double>& result,
	const VectorSlice<const std::vector<std::vector<double> > >& input,
	const FEValuesBase<dim>& fe,
	const std::vector<double>& factor,
	bool divide=false);


    }
  }
}

#endif
