//---------------------------------------------------------------------------
//    $Id: block_matrix_application.cc 1415 2013-08-30 16:14:41Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//    Copyright (C) 2007-13 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------


#include <appframe/block_matrix_application.h>
#include <base/parameter_handler.h>
#include <base/quadrature_lib.h>
#include <dofs/dof_handler.h>
#include <fe/fe.h>
#include <fe/fe_values.h>

using namespace AppFrame;

//------------------------------
template <int dim>
BlockMatrixApplication<dim>::BlockMatrixApplication(
    boost::shared_ptr<ApplicationData> data)
:
DoFApplication<dim>(data)
{
    deallog << "->BlockMatrix";
}

//---------------------------------------------------------------------------
template <int dim>
BlockMatrixApplication<dim>::BlockMatrixApplication(
    DoFApplication<dim>& other,
    bool triangulation_only)
:
DoFApplication<dim>(other, triangulation_only)
{
    deallog << "->BlockMatrix";
}

//---------------------------------------------------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::declare_parameters(ParameterHandler& param)
{
    DoFApplication<dim>::declare_parameters(param);
}

//---------------------------------------------------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::_initialize(ParameterHandler& param)
{
    
    // Initialize default values for flux_couplings (Note that this can be redefined in your application)
    this->flux_couplings.reinit(this->element->n_blocks(), this->element->n_blocks());
    for (unsigned int i=0; i<this->element->n_blocks(); ++i)
        for (unsigned int j=0; j<this->element->n_blocks(); ++j)
            this->flux_couplings(i,j) = DoFTools::none;
        
        // Get relevant parameter values
        // and initialize objects
        // accordingly.
        param.enter_subsection("Discretization");
    param.enter_subsection("Matrix");
    if (true)
    {
        int n = param.get_integer("Quadrature cell");
        if (n<=0) n = this->element->degree - n;
        boost::shared_ptr<Quadrature<dim> > newq(new QGauss<dim>(n));
        quadrature_assemble_cell = newq;
    }
    if (true)
    {
        int n = param.get_integer("Quadrature face");
        if (n<=0) n = this->element->degree - n;
        boost::shared_ptr<Quadrature<dim-1> > newq(new QGauss<dim-1>(n));
        quadrature_assemble_face = newq;
    }
    param.leave_subsection();
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::initialize(ParameterHandler& param)
{
    DoFApplication<dim>::initialize(param);
    _initialize(param);
}

//---------------------------------------------------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::add_vector_for_cell_matrix(
    std::string name,
    unsigned int , //block,
    unsigned int , //components,
    bool values,
    bool derivatives)
{
    cell_matrix_values.add_vector(name, values, derivatives, false);
    /*VectorSelector S;
     *  S.name = name;
     *  S.block = block;
     *  S.components = components;
     *  if (values)
     *    cell_matrix_values.push_back(S);
     *  if (derivatives)
     *    cell_matrix_derivatives.push_back(S);  */
}

//---------------------------------------------------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::add_vector_for_flux_matrix(std::string name,
                                                        unsigned int , //block,
                                                        unsigned int , //components,
                                                        bool values,
                                                        bool derivatives)
{
    flux_matrix_values.add_vector(name, values, derivatives, false);
    /* VectorSelector S;
     *  S.name = name;
     *  S.block = block;
     *  S.components = components;
     *  if (values)
     *    flux_matrix_values.push_back(S);
     *  if (derivatives)
     *    flux_matrix_derivatives.push_back(S);  */
}

//---------------------------------------------------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::remesh_matrices()
{
    this->block_info.initialize_local(*this->dof);
    
    // clear matrices
    matrix.clear();
    // Make the list of constraints associated with hanging nodes
    this->hanging_node_constraints.clear();
    DoFTools::make_hanging_node_constraints (*this->dof,
                                             this->hanging_node_constraints);
    this->hanging_node_constraints.close();
    
    const unsigned int n_blocks = this->element->n_blocks();
    
    BlockCompressedSparsityPattern c_sparsity(n_blocks, n_blocks);
    for (unsigned int i=0;i<n_blocks;++i)
        for (unsigned int j=0;j<n_blocks;++j)
            c_sparsity.block(i,j).reinit(this->block_info.global.block_size(i),
                                         this->block_info.global.block_size(j));
            c_sparsity.collect_sizes();
        
        if (this->interior_fluxes)
            DoFTools::make_flux_sparsity_pattern(*this->dof, c_sparsity,
                                                 cell_couplings, flux_couplings);
            else
                DoFTools::make_sparsity_pattern(*this->dof, cell_couplings,
                                                c_sparsity);
                
                // Condense sparsity pattern to account for hanging nodes
                this->hanging_node_constraints.condense(c_sparsity);
            
            sparsities.copy_from(c_sparsity);
        sparsities.compress();
        
        matrix.reinit(sparsities);
        
        //////////////////////////////////////////////////////////////////////
        // Apply Dirichlet boundary conditions
        //////////////////////////////////////////////////////////////////////
        filtered_matrix.initialize(this->matrix);
        filtered_matrix.clear_constraints();
        std::map<unsigned int, double> boundary_values;
        this->dirichlet_bc(boundary_values);
        filtered_matrix.add_constraints(boundary_values);
}

//----------------------------
template <int dim>
void
BlockMatrixApplication<dim>::remesh()
{
    DoFApplication<dim>::remesh();
    remesh_matrices();
}

//----------------------------
template <int dim>
void
BlockMatrixApplication<dim>::assemble(const FEVectors& src)
{
    // Clear old matrices
    matrix = 0;
    
    typedef typename DoFHandler<dim>::active_cell_iterator active_cell_iterator;
    
    active_cell_iterator begin = this->dof->begin_active();
    active_cell_iterator end = this->dof->end();  
    Threads::Mutex mutex;
    do_assemble(src, begin, end, mutex);
    
    /* if (mg_matrices.size() != 0)
     *    {
     *      typedef typename MGDoFHandler<dim>::cell_iterator cell_iterator;
     *      const std::vector<std::pair<cell_iterator,cell_iterator> >
     *	thread_ranges = Threads::split_range<cell_iterator> (this->mg_dof->begin(),
     *							     this->mg_dof->end(), n_threads);
     *      
     *      for (unsigned int thread=0; thread<n_threads; ++thread)
     *	threads += Threads::spawn (*this, &MatrixApplication<dim>::mg_assemble)
     *		   (src, thread_ranges[thread].first, thread_ranges[thread].second, mutex);
     *      threads.join_all();
     }*/
    
    //////////////////////////////////////////////////////////////////////
    // Condense matrix
    //////////////////////////////////////////////////////////////////////
    this->hanging_node_constraints.condense(this->matrix);
    
    //////////////////////////////////////////////////////////////////////
    // Apply Dirichlet boundary conditions
    //////////////////////////////////////////////////////////////////////
    filtered_matrix.clear_constraints();
    std::map<unsigned int, double> boundary_values;
    this->dirichlet_bc(boundary_values);
    filtered_matrix.initialize(this->matrix);
    filtered_matrix.add_constraints(boundary_values);
    
}
//------------------------------
//------------------------------

// Define these so the multigrid and single grid codes become more
// similar
#define MATRIX this->matrix.block(block_row, block_col)
#define INDEX(i,l) this->block_info.global.global_to_local(i).second


template <int dim>
void
BlockMatrixApplication<dim>::do_assemble(const FEVectors& src,
                                         typename DoFHandler<dim>::active_cell_iterator begin,
                                         typename DoFHandler<dim>::active_cell_iterator end,
                                         Threads::Mutex& mutex)
{
    deallog.push("ass");
    Assert (quadrature_assemble_cell, ExcNotInitialized());
    Assert (quadrature_assemble_face, ExcNotInitialized());
    const Quadrature<dim>& cell_quadrature = * quadrature_assemble_cell;
    const Quadrature<dim-1>& face_quadrature = * quadrature_assemble_face;
    
    // Objects for cell and face data
    // handed down to the local
    // routines.
    FEValues<dim>* fevalues = 0;
    FEFaceValues<dim>* fefacevalues = 0;
    FESubfaceValues<dim>* fesubfacevalues = 0;
    typename DoFApplication<dim>::CellInfo cell_info(src, this->block_info);
    typename DoFApplication<dim>::FaceInfo bdry_info(src, this->block_info);
    typename DoFApplication<dim>::FaceInfo face_info(src, this->block_info);
    typename DoFApplication<dim>::FaceInfo subface_info(src, this->block_info);
    typename DoFApplication<dim>::FaceInfo neighbor_info(src, this->block_info);
    
    cell_info.initialize(fevalues, *this->element, *this->mapping, cell_quadrature,
                         UpdateFlags(update_q_points
                         | update_values
                         | update_gradients
                         | update_JxW_values));
    bdry_info.initialize(fefacevalues, *this->element, *this->mapping, face_quadrature,
                         UpdateFlags(update_q_points
                         | update_values
                         | update_gradients
                         | update_normal_vectors
                         | update_JxW_values));
    face_info.initialize(fefacevalues, *this->element, *this->mapping, face_quadrature,
                         UpdateFlags(update_q_points
                         | update_values
                         | update_gradients
                         | update_normal_vectors
                         | update_JxW_values));
    subface_info.initialize(fesubfacevalues, *this->element, *this->mapping, face_quadrature,
                            UpdateFlags(update_q_points
                            | update_values
                            | update_gradients
                            | update_normal_vectors
                            | update_JxW_values));
    neighbor_info.initialize(fefacevalues, *this->element, *this->mapping, face_quadrature,
                             UpdateFlags(update_values
                             | update_gradients
                             | update_normal_vectors
                             | update_JxW_values));
    
    // Initialize local data
    MatrixVector intint;
    MatrixVector intext;
    MatrixVector extint;
    MatrixVector extext;
    
    for (unsigned int i=0;i<matrix.n_block_rows();++i)
        for (unsigned int j=0;j<matrix.n_block_rows();++j)
        {
            if (cell_couplings(i,j) == DoFTools::none
                && flux_couplings(i,j) == DoFTools::none)
                continue;
            
            const unsigned int rows = this->block_info.local.block_size(i);
            const unsigned int cols = this->block_info.local.block_size(j);
            MatrixBlock<FullMatrix<double> > block(i, j);
            intint.push_back(block);
            intext.push_back(block);
            extint.push_back(block);
            extext.push_back(block);
            intint[intint.size()-1].matrix.reinit (rows, cols);
            intext[intint.size()-1].matrix.reinit (rows, cols);
            extint[intint.size()-1].matrix.reinit (rows, cols);
            extext[intint.size()-1].matrix.reinit (rows, cols);
        }
        
        this->tr->clear_user_flags ();
        
        typename DoFHandler<dim>::active_cell_iterator c;
        for (c = begin ; c != end ; ++c)
        {
            for (unsigned int i=0;i<intint.size();++i)
                intint[i].matrix = 0.;
            // Initialize local structures
            cell_info.reinit(c);
            
            // Fill local data vectors
            cell_info.fill_local_data(cell_info.values, true);
            cell_info.fill_local_data(cell_info.derivatives, true);
            cell_matrix(intint, cell_info);
            
            mutex.acquire();
            for (unsigned int i=0;i<intint.size();++i)
            {
                for (unsigned int j=0;j<intint[i].matrix.n_rows();++j)
                    for (unsigned int k=0;k<intint[i].matrix.n_cols();++k)
                        if (fabs(intint[i].matrix(j,k)) > 1.e-15)
                        {
                            const unsigned int block_row = intint[i].row;
                            const unsigned int block_col = intint[i].column;
                            const unsigned int
                            jcell = this->block_info.local.local_to_global(block_row, j);
                            const unsigned int
                            kcell = this->block_info.local.local_to_global(block_col, k);
                            
                            MATRIX.add(INDEX(cell_info.indices[jcell], level),
                                       INDEX(cell_info.indices[kcell], level),
                                       intint[i].matrix(j,k));
                        }
            }
            mutex.release();
        }
        
        this->post_cell_assemble();
        
        #if deal_II_dimension > 1
        
        // //////////////////////////////////////////////////////////////////////
        // // Integration of fluxes
        // //////////////////////////////////////////////////////////////////////
        if (this->interior_fluxes || this->boundary_fluxes)
            
            for (c = begin; c != end; ++c)
            {
                for (unsigned int face_nr=0;
                     face_nr < GeometryInfo<dim>::faces_per_cell;
                ++ face_nr)
                     {
                         typename DoFHandler<dim>::face_iterator face = c->face(face_nr);
                         
                         // Avoid computing this face
                         // a second time.
                         if (face->user_flag_set ())
                             continue;
                         
                         if (c->at_boundary(face_nr))
                         {
                             for (unsigned int i=0;i<intint.size();++i)
                                 intint[i].matrix = 0.;
                             
                             bdry_info.reinit(c, face, face_nr);
                             // Boundary values
                             // Fill local data vectors
                             bdry_info.fill_local_data(bdry_info.values, true);
                             bdry_info.fill_local_data(bdry_info.derivatives, true);
                             this->bdry_matrix(intint, bdry_info);
                             
                             for (unsigned int i=0;i<intint.size();++i)
                             {
                                 for (unsigned int j=0;j<intint[i].matrix.n_rows();++j)
                                     for (unsigned int k=0;k<intint[i].matrix.n_cols();++k)
                                         if (fabs(intint[i].matrix(j,k)) > 1.e-12)
                                         {
                                             const unsigned int block_row = intint[i].row;
                                             const unsigned int block_col = intint[i].column;
                                             const unsigned int
                                             jcell = this->block_info.local.local_to_global(block_row, j);
                                             const unsigned int
                                             kcell = this->block_info.local.local_to_global(block_col, k);
                                             
                                             MATRIX.add(INDEX(bdry_info.indices[jcell], level),
                                                        INDEX(bdry_info.indices[kcell], level),
                                                        intint[i].matrix(j,k));
                                         }
                             }
                         }
                         else if (this->interior_fluxes)
                         {
                             // Interior face
                             typename DoFHandler<dim>::cell_iterator
                             neighbor = c->neighbor(face_nr);
                             // Do refinement face
                             // from the coarse side
                             if (neighbor->level() < c->level())
                                 continue;
                             
                             unsigned int neighbor_face_nr = c->neighbor_of_neighbor(face_nr);
                             // Make sure this
                             // face is not worked
                             // on from the other
                             // side
                             mutex.acquire();
                             if (face->user_flag_set ())
                             {
                                 mutex.release();
                                 continue;
                             }
                             face->set_user_flag ();
                             if (!neighbor->has_children())
                                 neighbor->face(neighbor_face_nr)->set_user_flag();
                             mutex.release();
                             
                             if (neighbor->has_children())
                             {
                                 // Refinement face
                                 // needs additional
                                 // treatment
                                 for (unsigned int sub_nr = 0; sub_nr != face->number_of_children(); ++sub_nr)
                                 {
                                     typename DoFHandler<dim>::cell_iterator sub_neighbor
                                     = c->neighbor_child_on_subface(face_nr, sub_nr);
                                     
                                     for (unsigned int i=0;i<intint.size();++i)
                                     {
                                         intint[i].matrix = 0.;
                                         intext[i].matrix = 0.;
                                         extint[i].matrix = 0.;
                                         extext[i].matrix = 0.;
                                     }
                                     
                                     subface_info.reinit(c, face, face_nr, sub_nr);
                                     neighbor_info.reinit(sub_neighbor,
                                                          sub_neighbor->face(neighbor_face_nr),
                                                          neighbor_face_nr);
                                     
                                     // Fill local data vectors
                                     subface_info.fill_local_data(subface_info.values, true);
                                     subface_info.fill_local_data(subface_info.derivatives, true);
                                     neighbor_info.fill_local_data(neighbor_info.values, true);
                                     neighbor_info.fill_local_data(neighbor_info.derivatives, true);
                                     this->face_matrix(intint, intext, extint, extext, subface_info, neighbor_info);
                                     
                                     mutex.acquire();
                                     sub_neighbor->face(neighbor_face_nr)->set_user_flag ();
                                     for (unsigned int i=0;i<intint.size();++i)
                                     {
                                         for (unsigned int j=0;j<intint[i].matrix.n_rows();++j)
                                             for (unsigned int k=0;k<intint[i].matrix.n_cols();++k)
                                             {
                                                 const unsigned int block_row = intint[i].row;
                                                 const unsigned int block_col = intint[i].column;
                                                 const unsigned int
                                                 jcell = this->block_info.local.local_to_global(block_row, j);
                                                 const unsigned int
                                                 kcell = this->block_info.local.local_to_global(block_col, k);
                                                 
                                                 if (fabs(intint[i].matrix(j,k)) > 1.e-12)
                                                     MATRIX.add(INDEX(subface_info.indices[jcell], level),
                                                                INDEX(subface_info.indices[kcell], level),
                                                                intint[i].matrix(j,k));
                                                     if (fabs(intext[i].matrix(j,k)) > 1.e-12)
                                                         MATRIX.add(INDEX(subface_info.indices[jcell], level),
                                                                    INDEX(neighbor_info.indices[kcell], level),
                                                                    intext[i].matrix(j,k));
                                                         if (fabs(extint[i].matrix(j,k)) > 1.e-12)
                                                             MATRIX.add(INDEX(neighbor_info.indices[jcell], level),
                                                                        INDEX(subface_info.indices[kcell], level),
                                                                        extint[i].matrix(j,k));
                                                             if (fabs(extext[i].matrix(j,k)) > 1.e-12)
                                                                 MATRIX.add(INDEX(neighbor_info.indices[jcell], level),
                                                                            INDEX(neighbor_info.indices[kcell], level),
                                                                            extext[i].matrix(j,k));
                                             }
                                     }
                                     mutex.release();
                                 }
                             } else {
                                 // Regular interior face
                                 face_info.reinit(c, face, face_nr);
                                 neighbor_info.reinit(neighbor,
                                                      neighbor->face(neighbor_face_nr),
                                                      neighbor_face_nr);
                                 
                                 for (unsigned int i=0;i<intint.size();++i)
                                 {
                                     intint[i].matrix = 0.;
                                     intext[i].matrix = 0.;
                                     extint[i].matrix = 0.;
                                     extext[i].matrix = 0.;
                                 }
                                 
                                 // Fill local data vectors
                                 face_info.fill_local_data(face_info.values, true);
                                 face_info.fill_local_data(face_info.derivatives, true);
                                 neighbor_info.fill_local_data(neighbor_info.values, true);
                                 neighbor_info.fill_local_data(neighbor_info.derivatives, true);
                                 this->face_matrix(intint, intext, extint, extext, face_info, neighbor_info);
                                 
                                 mutex.acquire();
                                 for (unsigned int i=0;i<intint.size();++i)
                                 {
                                     for (unsigned int j=0;j<intint[i].matrix.n_rows();++j)
                                         for (unsigned int k=0;k<intint[i].matrix.n_cols();++k)
                                         {
                                             const unsigned int block_row = intint[i].row;
                                             const unsigned int block_col = intint[i].column;
                                             const unsigned int
                                             jcell = this->block_info.local.local_to_global(block_row, j);
                                             const unsigned int
                                             kcell = this->block_info.local.local_to_global(block_col, k);
                                             
                                             if (fabs(intint[i].matrix(j,k)) > 1.e-12)
                                                 MATRIX.add(INDEX(face_info.indices[jcell], level),
                                                            INDEX(face_info.indices[kcell], level),
                                                            intint[i].matrix(j,k));
                                                 if (fabs(intext[i].matrix(j,k)) > 1.e-12)
                                                     MATRIX.add(INDEX(face_info.indices[jcell], level),
                                                                INDEX(neighbor_info.indices[kcell], level),
                                                                intext[i].matrix(j,k));
                                                     if (fabs(extint[i].matrix(j,k)) > 1.e-12)
                                                         MATRIX.add(INDEX(neighbor_info.indices[jcell], level),
                                                                    INDEX(face_info.indices[kcell], level),
                                                                    extint[i].matrix(j,k));
                                                         if (fabs(extext[i].matrix(j,k)) > 1.e-12)
                                                             MATRIX.add(INDEX(neighbor_info.indices[jcell], level),
                                                                        INDEX(neighbor_info.indices[kcell], level),
                                                                        extext[i].matrix(j,k));
                                         }
                                 }
                                 mutex.release();
                             }
                         }
                         
                     } // faces
            } // cells
            
            #endif
            
            deallog.pop();
}

#undef MATRIX
#undef INDEX

//------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::assemble_numerically(const FEVectors& src,
                                                  const double delta)
{
    this->matrix=0.0;
    
    double l2_norm_residual;
    
    // Find the index where the solution vector is stored:
    unsigned int ind = src.find_vector("Newton iterate");
    
    // Vector where the pertrubed solution is stored:
    AppFrame::FEVector solution_delta;
    solution_delta = src.vector(ind);
    
    // FEVectors object that is passed to residual and that includes the perturbed solution:
    AppFrame::FEVectors solution_copy;
    solution_copy.add_vector(solution_delta, "Newton iterate" );
    
    // Create a vector where the residual is stored and initialize
    // it with the residual:
    AppFrame::FEVector residual;
    residual.reinit(src.vector(ind));       
    l2_norm_residual = this->residual(residual, solution_copy, false);
    
    // Create and initialize a vector where the residual obtained
    // using the perturbed solution will be stored:
    AppFrame::FEVector residual_copy;        
    residual_copy.reinit(src.vector(ind));
    
    // Loop over DOFs (columns)
    for (unsigned int j=0; j<residual.size(); j++)
    {
        // Perturb solution:
        solution_delta[j] += delta;
        
        // Compute residual
        l2_norm_residual = this->residual(residual_copy, solution_copy, false);
        
        // Loop over DOFs (rows) and copy entries in the Jacobian as required:
        for (unsigned int i=0; i<residual.size(); i++)
        {
            this->matrix.set(i,j,(residual_copy(i) - residual(i))/delta);
        }
        // Remove perturbation:
        solution_delta[j] -= delta;
    }
    
    //////////////////////////////////////////////////////////////////////
    // Condense matrix, i.e. correct for hanging nodes
    //////////////////////////////////////////////////////////////////////
    this->hanging_node_constraints.condense(this->matrix);
    
    //////////////////////////////////////////////////////////////////////
    // Apply Dirichlet boundary conditions
    //////////////////////////////////////////////////////////////////////
    filtered_matrix.clear_constraints();
    std::map<unsigned int, double> boundary_values;
    this->dirichlet_bc(boundary_values);
    filtered_matrix.initialize(this->matrix);
    filtered_matrix.add_constraints(boundary_values);
    
    // Note: In solve, if we want to apply BC to matrix, rhs and sol use:       
    MatrixTools::apply_boundary_values(boundary_values,
                                       this->matrix,
                                       solution_delta,
                                       solution_delta);
    
    /*
    // ========================================
    // USE FOR DEBUGGING ONLY:
    // ========================================
    std::cout<<"== PRINTING NUMERICAL MATRIX =="<<std::endl; 
    std::ofstream file;
    file.open("Numerical_matrix.dat");
    this->matrix.print_formatted(file, 5, true, 0, " 0 ");
    file.close();
    std::cout<<"============================"<<std::endl;
    // ========================================
    */
    
    
    // ========================================
    // USE FOR DEBUGGING ONLY:
    // ========================================
    /*
    //Analytical sensitivities:
    this->matrix=0.0;
    AppFrame::FEVectors sol;
    unsigned int ind = src.find_vector("Newton iterate");
    sol.add_vector(src.vector(ind),"Newton iterate");
    this->assemble(sol); //Note the second component of
           
    AppFrame::FEVector solution_delta;
    solution_delta = src.vector(ind);
             
    std::map<unsigned int, double> boundary_values;
    this->dirichlet_bc(boundary_values);
            
    MatrixTools::apply_boundary_values(boundary_values,
                                       this->matrix,
                                       solution_delta,
                                       solution_delta);
             
    std::cout<<"== PRINTING ANALYTICAL MATRIX =="<<std::endl; 
    std::ofstream file;
    file.open("Analytical_matrix.dat");
    this->matrix.print_formatted(file, 5, true, 0, " 0 ");
    file.close();
    std::cout<<"============================"<<std::endl;
    */
    // ========================================   
    
}


//------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::residual_constraints(FEVector& dst) const
{
    
    //////////////////////////////////////////////////////////////////////
    // Condense rhs
    //////////////////////////////////////////////////////////////////////
    this->hanging_node_constraints.condense(dst);
    
    //////////////////////////////////////////////////////////////////////
    // Apply Dirichlet boundary conditions
    //////////////////////////////////////////////////////////////////////
    filtered_matrix.apply_constraints(dst,
                                      false);  
}

//------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::post_cell_assemble()
{}

//------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::cell_matrix(
    MatrixVector& matrices,
    const typename DoFApplication<dim>::CellInfo&)
{
    deallog << "Void function " << __FUNCTION__ << " called" << std::endl;
    for (unsigned int i=0;i<matrices.size();++i)
        deallog << "Matrix[" << i << "]("
        << matrices[i].row << ','
        << matrices[i].column << ")\t"
        << matrices[i].matrix.n_rows() << 'x'
    << matrices[i].matrix.n_cols() << std::endl;
}

//------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::bdry_matrix(
    MatrixVector& matrices,
    const typename DoFApplication<dim>::FaceInfo&)
{
    deallog << "Void function " << __FUNCTION__ << " called" << std::endl;
    for (unsigned int i=0;i<matrices.size();++i)
        deallog << "Matrix[" << i << "]("
        << matrices[i].row << ','
        << matrices[i].column << ")\t"
        << matrices[i].matrix.n_rows() << 'x'
    << matrices[i].matrix.n_cols() << std::endl;
}

//------------------------------
template <int dim>
void
BlockMatrixApplication<dim>::face_matrix(
    MatrixVector& matrices11,
    MatrixVector& matrices12,
    MatrixVector& matrices21,
    MatrixVector& matrices22,
    const typename DoFApplication<dim>::FaceInfo&,
    const typename DoFApplication<dim>::FaceInfo&)
{
    deallog << "Void function " << __FUNCTION__ << " called" << std::endl;
    Assert (matrices11.size() == matrices12.size(),
            ExcDimensionMismatch(matrices11.size(), matrices12.size()));
    Assert (matrices11.size() == matrices21.size(),
            ExcDimensionMismatch(matrices11.size(), matrices21.size()));
    Assert (matrices11.size() == matrices22.size(),
            ExcDimensionMismatch(matrices11.size(), matrices22.size()));
    
    for (unsigned int i=0;i<matrices11.size();++i)
    {
        Assert (matrices11[i].matrix.n_rows() == matrices12[i].matrix.n_rows(),
                ExcDimensionMismatch(matrices11[i].matrix.n_rows(), matrices12[i].matrix.n_rows()));
        Assert (matrices11[i].matrix.n_cols() == matrices12[i].matrix.n_cols(),
                ExcDimensionMismatch(matrices11[i].matrix.n_cols(), matrices12[i].matrix.n_cols()));
        Assert (matrices11[i].matrix.n_rows() == matrices21[i].matrix.n_rows(),
                ExcDimensionMismatch(matrices11[i].matrix.n_rows(), matrices21[i].matrix.n_rows()));
        Assert (matrices11[i].matrix.n_cols() == matrices21[i].matrix.n_cols(),
                ExcDimensionMismatch(matrices11[i].matrix.n_cols(), matrices21[i].matrix.n_cols()));
        Assert (matrices11[i].matrix.n_rows() == matrices22[i].matrix.n_rows(),
                ExcDimensionMismatch(matrices11[i].matrix.n_rows(), matrices22[i].matrix.n_rows()));
        Assert (matrices11[i].matrix.n_cols() == matrices22[i].matrix.n_cols(),
                ExcDimensionMismatch(matrices11[i].matrix.n_cols(), matrices22[i].matrix.n_cols()));
        
        deallog << "Matrix[" << i << "]("
                << matrices11[i].row << ','
                << matrices11[i].column << ")\t"
                << matrices11[i].matrix.n_rows() << 'x'
                << matrices11[i].matrix.n_cols() << std::endl;
    }
}

//--------------------
template <int dim>
void
BlockMatrixApplication<dim>::dirichlet_bc(std::map<unsigned int, double>& /*boundary_values*/) const
{
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name() << std::endl;
}

//------------------------------
//------------------------------
//------------------------------
template class BlockMatrixApplication<deal_II_dimension>;
     