//---------------------------------------------------------------------------
//    $Id: dof_application.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

// The following classes are implemented:
// - LocalIntegrator
// - LocalResidual
// - LocalEstimate
// -
// - 
//
//

#include <appframe/dof_application.h>
#include <appframe/mesh_worker_loop.h>
#include <base/parameter_handler.h>
#include <base/data_out_base.h>
#include <base/quadrature_lib.h>
#include <lac/block_vector.h>
#include <grid/tria.h>
#include <grid/tria_iterator.h>
#include <grid/tria_accessor.h>
#include <grid/grid_out.h>
#include <grid/grid_refinement.h>
#include <dofs/dof_handler.h>
#include <dofs/dof_renumbering.h>
#include <dofs/dof_tools.h>
#include <multigrid/mg_tools.h>
#include <fe/fe.h>
#include <fe/mapping_q1.h>
#include <fe/mapping_q.h>
#include <fe/fe_tools.h>

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <typeinfo>

using namespace AppFrame;

//----------------------------------------------------------------------//
//----------------------------------------------------------------------//
//--------              LocalIntegrator                  ---------------//
//----------------------------------------------------------------------//
//----------------------------------------------------------------------//

template <int dim>
LocalIntegrator<dim>::LocalIntegrator()
{}

//----------------------------------------------------------------------//
template <int dim>
LocalIntegrator<dim>::~LocalIntegrator()
{}

//----------------------------------------------------------------------//
template<int dim>
void
LocalIntegrator<dim>::set_gauss_quadrature(
    const unsigned int cp,
    const unsigned int bp,
    const unsigned int fp)
{
    cell_quadrature = QGauss<dim>(cp);
    bdry_quadrature = QGauss<dim-1>(bp);
    face_quadrature = QGauss<dim-1>(fp);
}

//----------------------------------------------------------------------//
template<int dim>
void
LocalIntegrator<dim>::post_loop(FEVector&, const FEVectors&)
{}

//----------------------------------------------------------------------//
//----------------------------------------------------------------------//
//--------              LocalResidual                    ---------------//
//----------------------------------------------------------------------//
//----------------------------------------------------------------------//

template<int dim>
void
LocalResidual<dim>::cell(const typename DoFApplication<dim>::CellInfo& cell)
{
    app->cell_residual(R1[0], cell);
    assemble(cell);
}


template<int dim>
void
LocalResidual<dim>::bdry(const typename DoFApplication<dim>::FaceInfo& face)
{
    app->bdry_residual(R1[0], face);
    assemble(face);
}



template<int dim>
void
LocalResidual<dim>::face(const typename DoFApplication<dim>::FaceInfo& face1,
                         const typename DoFApplication<dim>::FaceInfo& face2)
{
    app->face_residual(R1[0], R2[0], face1, face2);
    assemble(face1, face2);
}


//----------------------------------------------------------------------//
//----------------------------------------------------------------------//
//--------              LocalEstimate                  ---------------//
//----------------------------------------------------------------------//
//----------------------------------------------------------------------//

template<int dim>
void
LocalEstimate<dim>::cell(const typename DoFApplication<dim>::CellInfo& cell)
{
    J1[0] = app->cell_estimate(cell);
    assemble(cell);
}



template<int dim>
void
LocalEstimate<dim>::bdry(const typename DoFApplication<dim>::FaceInfo& face)
{
    J1[0] = app->bdry_estimate(face);
    assemble(face);
}



template<int dim>
void
LocalEstimate<dim>::face(const typename DoFApplication<dim>::FaceInfo& face1,
                         const typename DoFApplication<dim>::FaceInfo& face2)
{
    J1[0] = app->face_estimate(face1, face2);
    assemble(face1, face2);
}



//----------------------------------------------------------------------//
//----------------------------------------------------------------------//
//--------              DoFResidual                  ---------------//
//----------------------------------------------------------------------//
//----------------------------------------------------------------------//
/*
 * template <int dim>
 * DoFResidual<dim>::DoFResidual(const DoFApplication<dim>& application,
 *                              const LocalIntegrator<dim>& local_residual)
 * :
 * app(&application),
 * local_residual(local_residual)
 * {}
 * 
 * //----------------------------------------------------------------------//
 * template <int dim>
 * double
 * DoFResidual<dim>::operator() (FEVector& dst, const FEVectors& src) const
 * {
 *    app->integrate_1form(dst, src);
 *    return dst.l2_norm();
 }
 */
//----------------------------------------------------------------------//
//----------------------------------------------------------------------//
//--------              DoFApplication                   ---------------//
//----------------------------------------------------------------------//
//----------------------------------------------------------------------//

template <int dim>
DoFApplication<dim>::DoFApplication(boost::shared_ptr<ApplicationData> data, 
                                    bool multigrid)
:
ApplicationBase(data),
verbosity(0),
tr(new Triangulation<dim>),
boundary_fluxes(true),
interior_fluxes(true),
output_materials_and_levels(true)
{
    deallog << "->DoF";
    if (multigrid)
    {
        boost::shared_ptr<MGDoFHandler<dim> >
        newdof(new MGDoFHandler<dim>(*tr));
        mg_dof = newdof;
        dof = mg_dof;
    }
    else
    {
        boost::shared_ptr<DoFHandler<dim> >
        newdof(new DoFHandler<dim>(*tr));
        dof = newdof;
    }
}


//----------------------------------------------------------------------//

template <int dim>
DoFApplication<dim>::DoFApplication(bool multigrid)
:
ApplicationBase(boost::shared_ptr<ApplicationData>()),
tr(new Triangulation<dim>),
boundary_fluxes(true),
interior_fluxes(true),
output_materials_and_levels(true)
{
    deallog << "->DoF";
    if (multigrid)
    {
        boost::shared_ptr<MGDoFHandler<dim> >
        newdof(new MGDoFHandler<dim>(*tr));
        mg_dof = newdof;
        dof = mg_dof;
    }
    else
    {
        boost::shared_ptr<DoFHandler<dim> >
        newdof(new DoFHandler<dim>(*tr));
        dof = newdof;
    }
}

//----------------------------------------------------------------------//


template <int dim>
inline
DoFApplication<dim>::DoFApplication(DoFApplication<dim>& other,
                                    bool triangulation_only,
                                    bool multigrid)
: ApplicationBase(other),
tr(other.tr)
{
    tr = other.tr;
    if (triangulation_only)
    {
        if (multigrid)
        {
            boost::shared_ptr<MGDoFHandler<dim> >
            newdof(new MGDoFHandler<dim>(*tr));
            mg_dof = newdof;
            dof = mg_dof;
        }
        else
        {
            boost::shared_ptr<DoFHandler<dim> >
            newdof(new DoFHandler<dim>(*tr));
            dof = newdof;
        }
    }
    else
    {
        dof = other.dof;
        mg_dof = other.mg_dof;
    }
}

//----------------------------------------------------------------------//

template <int dim>
DoFApplication<dim>::~DoFApplication()
{}

//----------------------------------------------------------------------//

template <int dim>
void
DoFApplication<dim>::declare_parameters(ParameterHandler& param)
{
    param.enter_subsection("Discretization");
    param.declare_entry("Element", "FE_Q(1)", Patterns::Anything(),
    "The finite element used for discretization. "
    "Any input for FETools::get_element_by_name "
    "is possible.");
    
    param.declare_entry("Mapping degree",
    "1",
    Patterns::Integer(1),
    "Degree used for polynomial mapping of arbitrary order");
    
    param.declare_entry("Boundary fluxes",
    "false",
    Patterns::Bool(),
    "");
    param.declare_entry("Interior fluxes",
    "false",
    Patterns::Bool(),
    "");
    param.enter_subsection("Matrix");
    param.declare_entry("Quadrature cell", "-1", Patterns::Integer(),
    "Gauss formula used for integrating the "
    "bilinear form on a cell. A negative value means "
    "adding its absolute value to polynomial degree of "
    "the shape functions.");
    param.declare_entry("Quadrature face", "-1", Patterns::Integer(),
    "Gauss formula used for integrating the bilinear "
    "form on faces. See above for more details.");
    param.leave_subsection();
    param.enter_subsection("Residual");
    param.declare_entry("Quadrature cell", "-1", Patterns::Integer(),
    "Gauss formula for integrating the residual on cells."
    " See above for more details.");
    param.declare_entry("Quadrature bdry", "-1", Patterns::Integer(),
    "Gauss formula used for integrating the residual "
    "on boundary faces. See above for more details.");
    param.declare_entry("Quadrature face", "-1", Patterns::Integer(),
    "Gauss formula for integrating the residual on faces."
    " See above for more details.");
    param.leave_subsection();
    param.leave_subsection();
    
    param.enter_subsection("Grid generation");
    param.declare_entry("Sort by component", "true", Patterns::Bool());
    param.declare_entry("Sort Cuthill-McKee", "false", Patterns::Bool());
    param.declare_entry("Initial refinement", "1", Patterns::Integer());
    param.declare_entry("Refinement", "global",
    Patterns::Selection("global|adaptive|single|steps|circle"));
    param.declare_entry("Refinement threshold", "0.3", Patterns::Double());
    param.declare_entry("Coarsening threshold", "0", Patterns::Double());
    param.declare_entry("Distort", "0", Patterns::Double());
    param.leave_subsection();
    
    param.enter_subsection ("Output");
    param.enter_subsection ("Grid");
    GridOut::declare_parameters(param);
    param.leave_subsection();
    param.enter_subsection("Data");
    DataOutInterface<dim>::declare_parameters(param);
    param.leave_subsection();
    param.leave_subsection();
}

template <int dim>
void
DoFApplication<dim>::_initialize (ParameterHandler& param)
{
    param.enter_subsection ("Grid generation");
    refinement = param.get("Refinement");
    initial_refinement = param.get_integer("Initial refinement");
    refinement_threshold = param.get_double("Refinement threshold");
    coarsening_threshold = param.get_double("Coarsening threshold");
    
    sort_cuthill = param.get_bool("Sort Cuthill-McKee");
    param.leave_subsection();
    
    
    param.enter_subsection("Discretization");
    boost::shared_ptr<FiniteElement<dim> >
    newelement(FETools::get_fe_from_name<dim>(param.get("Element")));
    element = newelement;
    deallog << element->get_name() << std::endl;
    
    mapping_degree = param.get_integer("Mapping degree");  
    
    if(mapping == 0)
    {
        if(mapping_degree == 1)
        {
            boost::shared_ptr<Mapping<dim> > newmapping(new MappingQ1<dim>());
            mapping = newmapping;
        }
        else
        {
            boost::shared_ptr<Mapping<dim> > newmapping(new MappingQ<dim>(mapping_degree));
            mapping = newmapping;
        }
    } 
    
    boundary_fluxes = param.get_bool("Boundary fluxes");
    interior_fluxes = param.get_bool("Interior fluxes");
    
    param.enter_subsection("Residual");
    if (true)
    {
        int n = param.get_integer("Quadrature cell");
        if (n<=0) n = this->element->degree - n;
        this->quadrature_residual_cell = QGauss<dim>(n);
    }
    if (true)
    {
        int n = param.get_integer("Quadrature bdry");
        if (n<=0) n = this->element->degree - n;
        this->quadrature_residual_bdry = QGauss<dim-1>(n);
    }
    if (true)
    {
        int n = param.get_integer("Quadrature face");
        if (n<=0) n = this->element->degree - n;
        this->quadrature_residual_face = QGauss<dim-1>(n);
    }
    param.leave_subsection();
    param.leave_subsection();
    
    // Parse output parameters
    param.enter_subsection ("Output");
    param.enter_subsection ("Grid");
    g_out.parse_parameters(param);
    param.leave_subsection();
    param.enter_subsection ("Data");
    d_out.parse_parameters(param);
    param.leave_subsection();
    param.leave_subsection();
    
}

template <int dim>
void
DoFApplication<dim>::add_vector_for_transfer(FEVector* v)
{
    Assert(transfer_vectors.size() == 0, ExcMessage("Only one vector can be transferred."
    " You need to delete the previous vector"));
    transfer_vectors.push_back(v);
}

template <int dim>
void
DoFApplication<dim>::delete_vector_for_transfer()
{
    transfer_vectors.pop_back();
}

template <int dim>
void
DoFApplication<dim>::initialize (ParameterHandler& param)
{
    _initialize(param);
}


template <int dim>
void
DoFApplication<dim>::init_vector (FEVector& dst) const
{
    dst.reinit(block_info.global);
}

template <int dim>
void
DoFApplication<dim>::remesh()
{
    if (refinement == "global")
    {
        for (typename Triangulation<dim>::active_cell_iterator
            cell = tr->begin_active();
        cell != tr->end(); ++cell)
            cell->set_refine_flag();
    }
    else if (refinement == "single")
    {
        tr->last_active()->set_refine_flag();
    }
    else if (refinement == "steps")
    {
        bool cell_refined = false;
        for (typename Triangulation<dim>::active_cell_iterator
            cell = tr->begin_active();
        cell != tr->end(); ++cell)
        {
            const Point<dim> p = cell->center();
            bool positive = p(0) > 0;
            if (dim>1 && p(1) <= 0)
                positive = false;
            if (dim>2 && p(2) <= 0)
                positive = false;
            if (positive)
            {
                cell->set_refine_flag();
                cell_refined = true;
            }
        }
        // If nothing was refined, do a
        // global refinement.
        if (!cell_refined)
            for (typename Triangulation<dim>::active_cell_iterator
                cell = tr->begin_active();
            cell != tr->end(); ++cell)
                cell->set_refine_flag();
    }
    else if (refinement == "adaptive")
    {
        GridRefinement::refine_and_coarsen_fixed_number(
            *tr, cell_errors, refinement_threshold, coarsening_threshold);
    }
    else
    {
        Assert (false, ExcNotImplemented());
    }
    
    if (transfer_vectors.size() > 0)
    {
        Assert (transfer_vectors.size() == 1, ExcNotImplemented());
        SolutionTransfer<dim, Vector<double> > transfer(*dof);
        Vector<double> aux(dof->n_dofs());
        for (unsigned int i=0;i<aux.size();++i)
            aux(i) = transfer_vectors[0]->operator()(i);
        
        transfer.prepare_for_coarsening_and_refinement(aux);
        tr->execute_coarsening_and_refinement();
        remesh_dofs();
        Vector<double> aux2(dof->n_dofs());
        transfer.interpolate(aux, aux2);
        init_vector(*transfer_vectors[0]);
        for (unsigned int i=0;i<aux2.size();++i)
            transfer_vectors[0]->operator()(i) = aux2(i);
        
    }
    else
    {
        tr->execute_coarsening_and_refinement();
        remesh_dofs();
    }
}


template <int dim>
void
DoFApplication<dim>::grid_out(const std::string& name)
{
    
    const std::string suffix = g_out.default_suffix();
    const std::string filename = name + suffix;
    if (suffix == std::string(""))
    {
        deallog << "Not writing " << filename << std::endl;
        return;
    }
    
    deallog << "Gridfile:" << filename << std::endl;
    std::ofstream out(filename.c_str());
    
    g_out.write(*tr, out);
}

template <int dim>
void
DoFApplication<dim>::data_out(    				//## Print out VTK file and make it so that we print out a vtk file at evey single point, essentially we don't over write it.
const std::string& basename,					// set up so that name depends on variable that the parametric study is based on.
const FEVectors& src)
{
    d_out.attach_dof_handler(*this->dof);
    const std::string suffix = d_out.default_suffix();
    if (suffix == "") return;
    
    output_levels.reinit(this->tr->n_active_cells());
    output_materials.reinit(this->tr->n_active_cells());
    
    unsigned int i=0;
    for (typename Triangulation<dim>::active_cell_iterator cell = this->tr->begin_active();
         cell != this->tr->end(); ++cell, ++i)
         {
             output_levels(i) = cell->level();
             output_materials(i) = cell->material_id();
         }
         
         d_out.add_data_vector(output_levels, "level", DataOut_DoFData<DoFHandler<dim>, dim>::type_cell_data);
    d_out.add_data_vector(output_materials, "material", DataOut_DoFData<DoFHandler<dim>, dim>::type_cell_data);
    
    for (unsigned int i=0;i<src.n_vectors();++i)
        d_out.add_data_vector(src.vector(i),
                              src.vector_name(i),
                              DataOut_DoFData<DoFHandler<dim>, dim>::type_dof_data,
                              data_interpretation);
        
        const std::string filename = basename + suffix;
    deallog << "Datafile:" << filename << std::endl;
    
    std::ofstream of (filename.c_str ());
    d_out.build_patches();
    
    this->d_out.write (of);
    this->d_out.clear();
}

template <int dim>
void
DoFApplication<dim>::data_out(
    const std::string& basename,
    const FEVectors& src,
    const std::vector<std::string>& solution_names)
{
    d_out.attach_dof_handler(*this->dof);
    
    output_levels.reinit(this->tr->n_active_cells());
    output_materials.reinit(this->tr->n_active_cells());
    
    unsigned int i=0;
    for (typename Triangulation<dim>::active_cell_iterator cell = this->tr->begin_active();
         cell != this->tr->end(); ++cell, ++i)
         {
             output_levels(i) = cell->level();
             output_materials(i) = cell->material_id();
         }
         
         d_out.add_data_vector(output_levels, "level", DataOut_DoFData<DoFHandler<dim>, dim>::type_cell_data);
    d_out.add_data_vector(output_materials, "material", DataOut_DoFData<DoFHandler<dim>, dim>::type_cell_data);
    
    for (unsigned int i=0;i<src.n_vectors();++i)
        d_out.add_data_vector(src.vector(i),
                              solution_names,
                              DataOut_DoFData<DoFHandler<dim>, dim>::type_dof_data,
                              data_interpretation);
}







// ---          ---
// --- data_out ---
// ---          ---

template<int dim>
void
DoFApplication<dim>::data_out(const std::string&              basename,
                              
                              const FEVector&                 solution,
                              const std::vector<std::string>& solution_names,
                              
                              const FEVector&                 postprocessing,
                              const std::vector<std::string>& postprocessing_names)
{
    d_out.attach_dof_handler(*this->dof);
    d_out.add_data_vector(solution,
                          solution_names,
                          DataOut<dim>::type_dof_data,
                          solution_interpretations);
    
    if( postprocessing.size() != 0 )
    {
        d_out.add_data_vector(postprocessing,
                              postprocessing_names,
                              DataOut<dim>::type_dof_data,
                              postprocessing_interpretations);
    }
    
    output_materials.reinit(this->tr->n_active_cells());
    output_levels.reinit(this->tr->n_active_cells());
    
    unsigned int i = 0;
    for( typename Triangulation<dim>::active_cell_iterator cell =  this->tr->begin_active();
        cell != this->tr->end();
    ++cell,
    ++i )
        {
            output_materials(i) = cell->material_id();
            output_levels(i)    = cell->level();
        }
        
        if( output_materials_and_levels )
        {
            d_out.add_data_vector(output_materials,
                                  "cell_material_ids",
                                  DataOut<dim>::type_cell_data);
            d_out.add_data_vector(output_levels,
                                  "cell_levels",
                                  DataOut<dim>::type_cell_data);
        }
        
        d_out.build_patches();
        std::ofstream output( (basename + d_out.default_suffix()).c_str() );
        d_out.write(output);
        d_out.clear();
}







template <int dim>
void
DoFApplication<dim>::remesh_dofs()
{
    deallog << "Levels:" << dof->get_tria().n_levels() << std::endl
    << "Cells:" << dof->get_tria().n_active_cells() << std::endl;
    if (mg_dof)
    {
        mg_dof->clear();
        mg_dof->distribute_dofs (*element);
    }
    else
    {
        dof->clear();
        dof->distribute_dofs (*element);
    }
    
    if (mg_dof)
        block_info.initialize(*mg_dof);
    else
        block_info.initialize(*dof);
    
    deallog << "Dofs:" << dof->n_dofs();
    for (unsigned int i=0;i<block_info.global.size();++i)
        deallog << ((i==0) ? '(' : ',')
        << block_info.global.block_size(i);
    deallog << ')';
        
        if (mg_dof)
        {
            for (unsigned int l=0;l<tr->n_levels();++l)
            {
                deallog << "MGDofs[" << l << "]:" << mg_dof->n_dofs(l);
                for (unsigned int i=0;i<block_info.levels[l].size();++i)
                    deallog << ((i==0) ? '(' : ',')
                    << block_info.levels[l].block_size(i);
                deallog << ')';
            }
        }
        
        sort_dofs();
        // Make the list of constraints associated with hanging nodes
        hanging_node_constraints.clear();
        DoFTools::make_hanging_node_constraints (*this->dof,
                                                 hanging_node_constraints);
        hanging_node_constraints.close();
        
        // Provide indices for cells and
        // faces to be used in error
        // estimation
        tr->clear_user_data();
        unsigned int k=0;
        for (typename Triangulation<dim>::active_cell_iterator i = tr->begin_active();
             i != tr->end(); ++i)
             {
                 i->set_user_index(k++);
             }
             
             k=0;
        for (typename Triangulation<dim>::face_iterator i = tr->begin_face();
             i != tr->end_face(); ++i)
             {
                 i->set_user_index(k++);
             }
}


template <int dim>
void
DoFApplication<dim>::sort_dofs()
{
    if (sort_direction.norm_square() != 0.)
    {
        deallog << " sorting downstream";
        DoFRenumbering::downstream(*dof, sort_direction);
        if (mg_dof)
            for (unsigned int l=0;l<tr->n_levels();++l)
                DoFRenumbering::downstream(*mg_dof, l, sort_direction);
    }
    
    if (sort_cuthill)
    {
        deallog << " sorting Cuthill-McKee";
        DoFRenumbering::Cuthill_McKee(*dof);
        if (mg_dof)
            for (unsigned int l=0;l<tr->n_levels();++l)
                DoFRenumbering::Cuthill_McKee(*mg_dof, l);
    }
    
    deallog << " sorting by component";
    DoFRenumbering::component_wise(*dof);
    if (mg_dof)
        DoFRenumbering::component_wise(*mg_dof);
    deallog << std::endl;
}

//---------------------------------------------------------------------------
// P. Dobson and M. Secanell
//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::transfer_solution_to_coarse_mesh(Triangulation<dim>& tr_coarse,
                                                      AppFrame::FEVector& coarse_solution,
                                                      AppFrame::FEVector& refined_solution)
{
    // Initialize a dof handler object with the triangulation
    DoFHandler<dim> dof_handler_coarse;
    const FiniteElement<dim>& fe_coarse(*this->element);
    dof_handler_coarse.initialize(tr_coarse,fe_coarse);
    dof_handler_coarse.distribute_dofs(fe_coarse);
    DoFRenumbering::component_wise(dof_handler_coarse);
    
    // Obtain the common cells between the two triangulations
    typedef std::list<std::pair<typename DoFHandler<dim>::cell_iterator,
    typename DoFHandler<dim>::cell_iterator> > CellList;
    
    CellList finest_common_cells = GridTools::get_finest_common_cells (dof_handler_coarse, *this->dof);
    
    Assert (fe_coarse.dofs_per_cell == this->element->dofs_per_cell, ExcInternalError() );
    
    // Assign a vector for the dof values on the cell to transfer between solutions
    Vector<double> local_dof_values(fe_coarse.dofs_per_cell);
    // Loop over the common cells and transfer the solution values
    for (typename CellList::const_iterator 	cell_pair = finest_common_cells.begin();
         cell_pair != finest_common_cells.end();
    ++cell_pair)
         {
             cell_pair->second->get_interpolated_dof_values(refined_solution, local_dof_values);
             cell_pair->first->set_dof_values_by_interpolation(local_dof_values, coarse_solution);
         }
         
         // 	// TODO: Read from parameter file??
         // 	std::string gridfile_format = "ucd";
         // 	std::cout << GridOut::get_output_format_names() << std::endl;
         // 	GridOut::OutputFormat grid_format = GridOut::parse_output_format(gridfile_format);
         // 	std::string grid_filename = "coarse_grid" + GridOut::default_suffix(grid_format);
         // 	std::ofstream grid_file(grid_filename);
         // 	GridOut().write (tr_coarse, grid_filename, grid_format);
         
         std::ofstream grid_file("._mesh_to_be_transfered.msh");
         GridOut grid_out;
         GridOutFlags::Msh msh_flags;
         
         // Save the boundary indicators on the grid
         msh_flags.write_faces = true;
         grid_out.set_flags(msh_flags);
         grid_out.write_msh(tr_coarse,grid_file);
         
         std::ofstream vector_out(".transfer_solution.FEVector");
         coarse_solution.block_write(vector_out);
         vector_out.close();
         
         deallog << "Coarse mesh and solution written to file" << std::endl;
}

//---------------------------------------------------------------------------
// P. Dobson and M. Secanell
//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::read_init_solution(AppFrame::FEVector& dst, bool &good_solution) const
{
    std::ifstream vector_in(".transfer_solution.FEVector");
    if(!vector_in.fail())
    {
        deallog << "Initializing solution with available data" << std::endl;
        dst.block_read(vector_in);
        vector_in.close();
        good_solution = true;
    }
    else
    {
        deallog << "Cannot read file containing initial solution" << std::endl;
        good_solution = false;
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::constrain_boundary(FEVector& v, bool homogeneous) const
{
    for (typename std::map<unsigned int, double>::const_iterator
        i = boundary_constraints.begin();
    i != boundary_constraints.end();
    ++i)
        v(i->first) = (homogeneous) ? 0. : (i->second);
}

//---------------------------------------------------------------------------
template <int dim>
double
DoFApplication<dim>::evaluate(const FEVectors&)
{
    deallog.push("evaluate");
    unsigned int n = block_info.global.size();
    unsigned int mem_tr = tr->memory_consumption();
    unsigned int mem_dof = dof->memory_consumption();
    deallog << "DoFApplication "
    << memory_consumption() << " bytes" << std::endl
    << "Triangulation  " << mem_tr
    << " bytes" << std::endl
    << "DoFHandler     " << mem_dof
    << " bytes" << std::endl
    << "Dofs  \t" << dof->n_dofs() << std::endl
    << "Blocks\t" << block_info.global.size();
    for (unsigned int i=0;i<n;++i)
        deallog << '\t' << block_info.global.block_size(i);
    deallog << std::endl
    << "Start \t";
    for (unsigned int i=0;i<n;++i)
        deallog << '\t' << block_info.global.local_to_global(i,0);
    deallog << std::endl;
    deallog.pop();
    return 0.;
}

//---------------------------------------------------------------------------
template <int dim>
unsigned int
DoFApplication<dim>::memory_consumption() const
{
    unsigned int result = sizeof (*this)
    + tr->memory_consumption()
    + dof->memory_consumption()
    + block_info.global.memory_consumption()
    + block_info.local.memory_consumption()
    + g_out.memory_consumption()
    + d_out.memory_consumption()
    - sizeof(block_info)
    - sizeof (d_out)
    - sizeof (g_out);
    return result;
}

//---------------------------------------------------------------------------
template <int dim>
double
DoFApplication<dim>::estimate(const FEVectors& src)
{
    LocalEstimate<dim> integrator;
    integrator.boundary_fluxes = true;
    integrator.interior_fluxes = true;
    
    FEVectors dst;
    BlockVector<double>* cells = this->data->block_vector_pool.alloc();
    BlockVector<double>* faces = this->data->block_vector_pool.alloc();
    std::vector<unsigned int> block_sizes(1, this->tr->n_active_cells());
    cells->reinit(block_sizes);
    block_sizes[0] = this->tr->n_faces();
    faces->reinit(block_sizes);
    dst.add_vector(*cells, "cells");
    dst.add_vector(*faces, "faces");
    
    integrator.initialize(this, block_info, dst);
    integrator.cell_quadrature = quadrature_residual_cell;
    integrator.bdry_quadrature = quadrature_residual_bdry;
    integrator.face_quadrature = quadrature_residual_face;
    
    //TODO: Replace this!!!
    MeshWorker::VectorSelector cs;
    MeshWorker::VectorSelector fs;
    for (unsigned int i=0;i<src.n_vectors();++i)
    {
        cs.add_vector(src.vector_name(i), true, true, true);
        fs.add_vector(src.vector_name(i), true, true, false);
    }
    integrator.initialize_selectors(cs, fs, fs);
    
    MeshWorker::InfoObjects::IntegrationInfoBox<dim> box(this->block_info);
    box.initialize(integrator, *this->element, *this->mapping, src);
    
    MeshWorker::integration_loop(dof->begin_active(), dof->end(), box, integrator);
    
    cell_errors.reinit(this->tr->n_active_cells());
    face_errors.reinit(this->tr->n_faces());
    cell_errors = cells->block(0);
    face_errors = faces->block(0);
    
    this->data->block_vector_pool.free(cells);
    this->data->block_vector_pool.free(faces);
    
    // If face estimates are to be
    // added to cell estimates, do this
    // here.
    distribute_face_to_cell_errors();
    
    // Compute the global estimate from
    // local contributions
    return global_from_local_errors();
}

//---------------------------------------------------------------------------
template <int dim>
double
DoFApplication<dim>::residual(FEVector& dst, const FEVectors& src, bool apply_boundaries)
{
    dst.reinit(this->block_info.global);
    
    LocalResidual<dim> integrator;
    integrator.boundary_fluxes = boundary_fluxes;
    integrator.interior_fluxes = interior_fluxes;
    
    FEVectors aux;
    aux.add_vector(dst, "Residual");
    integrator.initialize(this, block_info, aux);
    integrator.add_update_flags(update_values | update_gradients | update_q_points);
    MeshWorker::VectorSelector tmp;
    
    //TODO: Replace this!!!
    for (unsigned int i=0;i<src.n_vectors();++i)
        tmp.add_vector(src.vector_name(i), true, true, false);
    integrator.initialize_selectors(tmp, tmp, tmp);
    integrator.cell_quadrature = quadrature_residual_cell;
    integrator.bdry_quadrature = quadrature_residual_bdry;
    integrator.face_quadrature = quadrature_residual_face;
    
    MeshWorker::InfoObjects::IntegrationInfoBox<dim> box(this->block_info);
    box.initialize(integrator, *this->element, *this->mapping, src);
    
    MeshWorker::integration_loop(dof->begin_active(), dof->end(), box, integrator);
    
    if (apply_boundaries == true)
        residual_constraints(dst);
    
    return dst.l2_norm();
}

//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::residual_constraints(FEVector&) const
{}

//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::cell_residual(
    FEVector& vectors,
    const CellInfo&)
{
    deallog << "Void function " << __FUNCTION__ << " called" << std::endl;
    for (unsigned int i=0;i<vectors.n_blocks();++i)
        deallog << "Vector[" << i << "]\t"
        << vectors.block(i).size() << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::bdry_residual(
    FEVector& vectors,
    const FaceInfo&)
{
    deallog << "Void function " << __FUNCTION__ << " called" << std::endl;
    for (unsigned int i=0;i<vectors.n_blocks();++i)
        deallog << "Vector[" << i << "]\t"
        << vectors.block(i).size() << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::face_residual(
    FEVector& face_vector1,
    FEVector& face_vector2,
    const FaceInfo&,
    const FaceInfo&)
{
    Assert (face_vector1.n_blocks() == face_vector2.n_blocks(),
            ExcDimensionMismatch(face_vector1.n_blocks(), face_vector2.n_blocks()));
    
    deallog << "Void function " << __FUNCTION__ << " called" << std::endl;
    for (unsigned int i=0;i<face_vector1.n_blocks();++i)
    {
        Assert (face_vector1.block(i).size() == face_vector2.block(i).size(),
                ExcDimensionMismatch(face_vector1.block(i).size(), face_vector2.block(i).size()));
        deallog << "Vector[" << i << "]\t"
        << face_vector1.block(i).size() << std::endl;
    }
}

//---------------------------------------------------------------------------
template <int dim>
double
DoFApplication<dim>::cell_estimate(
    const CellInfo&)
{
    return 0.;
}

//---------------------------------------------------------------------------
template <int dim>
double
DoFApplication<dim>::bdry_estimate(const FaceInfo&)
{
    return 0.;
}

//---------------------------------------------------------------------------
template <int dim>
double
DoFApplication<dim>::face_estimate(
    const FaceInfo& info1,
    const FaceInfo& info2)
{
    const FEFaceValuesBase<dim>& fe1 = info1.fe(0);
    //  const FEFaceValuesBase<dim>& fe2 = info2.fe(0);
    const std::vector<std::vector<Tensor<1,dim> > >& Du1 = info1.derivatives[0];
    const std::vector<std::vector<Tensor<1,dim> > >& Du2 = info2.derivatives[0];
    
    double result = 0.;
    
    
    double h = info1.face->diameter();
    for (unsigned k=0;k<fe1.n_quadrature_points;++k)
    {
        const double dx = fe1.JxW(k);
        const Point<dim>& n = fe1.normal_vector(k);
        
        // Loop over all components
        for (unsigned int d=0;d<Du1.size();++d)
        {
            const double diff = (n * Du1[d][k]) - (n * Du2[d][k]);
            result += dx * h * diff * diff;
        }
    }
    return result;
}

//---------------------------------------------------------------------------
template <int dim>
void
DoFApplication<dim>::distribute_face_to_cell_errors()
{
    typename DoFHandler<dim>::active_cell_iterator c;
    for (c = this->dof->begin_active();
         c != this->dof->end();
    ++c)
         {
             const unsigned int cell_index = c->user_index();
             for (unsigned int f = 0; f< GeometryInfo<dim>::faces_per_cell;++f)
             {
                 const unsigned int face_index = c->face(f)->user_index();
                 if (c->at_boundary(f))
                     cell_errors(cell_index) += face_errors(face_index);
                 else
                     cell_errors(cell_index) += 0.5 * face_errors(face_index);
             }
         }
}

//---------------------------------------------------------------------------
template<int dim>
double
DoFApplication<dim>::global_from_local_errors() const
{
    double sum = 0.;
    for (unsigned int i=0;i<cell_errors.size();++i)
        sum += cell_errors(i);
    return std::sqrt(sum);
}

//---------------------------------------------------------------------------
template <int dim>
template <class INFO, typename TYPE>
void
DoFApplication<dim>::fill_local_data(const INFO& info,
                                     const std::vector<VectorSelector>& data_vector,
                                     std::vector<std::vector<std::vector<TYPE> > >& data) const
{
    const std::vector<unsigned int>& index = info.index;
    for (unsigned int i=0;i<data_vector.size();++i)
    {
        const FEVector* src = this->get_data()->vector(data_vector[i].name);
        const unsigned int block = data_vector[i].block;
        const unsigned int fe_no = this->element->block_to_base_index(data_vector[i].block).first;
        const FEValuesBase<dim>& fe = info.fe(fe_no);
        const unsigned int size = data[i].size()/fe.get_fe().n_components()*fe.dofs_per_cell;
        
        MeshWorker::InfoObjects::fill_data(fe, *src, index, this->block_info.local.block_start(block), size, data[i]);
    }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template class LocalIntegrator<deal_II_dimension>;
template class LocalResidual<deal_II_dimension>;
template class DoFApplication<deal_II_dimension>;
