//---------------------------------------------------------------------------
//    $Id: dof_mesh_loop.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <base/parameter_handler.h>
#include <base/logstream.h>
#include <base/function.h>
#include <lac/block_vector.h>
#include <appframe/mesh_loop.h>
#include <appframe/dof_application.h>

#include <boost/shared_ptr.hpp>

#include <fstream>
#include <iomanip>

using namespace AppFrame;
using namespace std;


template <int dim>
DoFMeshLoop<dim>::DoFMeshLoop(ApplicationBase& app, DoFApplication<dim>& dof_app)
		: MeshLoop(app), dof_app(&dof_app),
		  transfer_solution(false),
		  intrinsic_error(false),
		  output_difference(true)
{
  deallog << "->DoFMeshLoop";
}


template <int dim>
DoFMeshLoop<dim>::~DoFMeshLoop()
{
  if (old_solution)
    get_data()->block_vector_pool.free(old_solution);    
}


template <int dim>
void DoFMeshLoop<dim>::declare_parameters(ParameterHandler& param)
{
  MeshLoop::declare_parameters(param);
    
  param.enter_subsection("Mesh loop");
  param.declare_entry("Intrinsic error", "false", Patterns::Bool());  
  param.declare_entry("Output difference", "true", Patterns::Bool());  
  param.declare_entry("Transfer solution", "true", Patterns::Bool());  
  param.leave_subsection();
}


template <int dim>
void DoFMeshLoop<dim>::initialize(ParameterHandler& param)
{
  MeshLoop::initialize(param);
  
  param.enter_subsection("Mesh loop");
  intrinsic_error = param.get_bool("Intrinsic error");
  output_difference = param.get_bool("Output difference");
  transfer_solution = param.get_bool("Transfer solution");
  if (!intrinsic_error)
    output_difference = false;
  
  param.leave_subsection();
}


template <int dim>
void
DoFMeshLoop<dim>::data_out(const std::string& fname, const FEVectors& src)
{
  FEVectors out;
  if (output_difference)
    out.add_vector(*old_solution, "diff");
  out.merge(src);
  app->data_out(fname, out);
}


template <int dim>
void DoFMeshLoop<dim>::setup(FEVector& start)
{
  if (dof_app)
    {
      if (intrinsic_error || transfer_solution)
	dof_app->add_vector_for_transfer(&start);
      
      if (intrinsic_error)
	{
	  old_solution = get_data()->block_vector_pool.alloc();
	  old_solution->reinit(start);
	}
    }
}


template <int dim>
void DoFMeshLoop<dim>::do_remesh(FEVector& start)
{
  app->remesh();
  if (intrinsic_error)
    *old_solution = start;
  if (!transfer_solution)
    app->start_vector(start, "MeshLoop");
  else
    app->start_vector(start, "Constraints");
}


template <int dim>
void DoFMeshLoop<dim>::do_evaluate(const FEVectors& src)
{
  FEVectors vecs;
  
  app->evaluate(src);
  if (intrinsic_error)
    {
      old_solution->add(-1., src.vector(0));
      vecs.add_vector(*old_solution, "DoFMeshLoop::difference");
    }
  
  vecs.merge(src);
  app->evaluate(vecs);
}


template class DoFMeshLoop<deal_II_dimension>;
