// ------------------------------------------------------------------
//  $Id: event.cc 1348 2013-08-16 00:45:07Z secanell $
//
//  Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat.
//  Copyright (C) 2012 by Valentin N. Zingan, University of Alberta.
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
// ------------------------------------------------------------------

#include <appframe/event.h>

using namespace AppFrame;
std::vector<std::string> Event::names;

Event::Event()
:
all_true(false),
flags(names.size(), false)
{ }

Event
Event::assign(const char* name)
{
  unsigned int index = names.size();
  names.push_back(name);

  Event result;
  // The constructor generated an
  // object with all flags equal
  // zero. Now we set the new one.
  result.flags[index] = true;

  return result;
}

void
Event::all()
{
  all_true = true;
  std::fill(flags.begin(), flags.end(), true);
}

void
Event::clear()
{
  all_true = false;
  std::fill(flags.begin(), flags.end(), false);
}

bool
Event::test(const Event& other) const
{
  // First, test all_true in this object
  if(all_true)
    return true;

  const unsigned int n = flags.size();
  const unsigned int m = other.flags.size();
  const unsigned int n_min = std::min(n, m);

  // Now, if all_true set in the
  // other object, then all must be true
  // in this object
  if(other.all_true)
  {
    // Non existing flags are
    // always assumed false
    if(m > n)
      return false;

    // Test all flags separately
    // and return false if one is
    // not set
    for(std::vector<bool>::const_iterator iter = flags.begin(); iter != flags.end(); ++iter)
      if(!*iter)
        return false;
    // All flags are set
    return true;
  }

  // Finally, compare each flag
  // separately
  for(unsigned int i = 0; i < n_min; ++i)
    if(other.flags[i] && !flags[i])
      return false;
  for(unsigned int i = n_min; i < m; ++i)
    if(other.flags[i])
      return false;
  return true;
}

bool
Event::any() const
{
  if(all_true)
    return true;
  for(std::vector<bool>::const_iterator iter = flags.begin(); iter != flags.end(); ++iter)
    if(*iter)
      return true;
  return false;
}

Event&
Event::operator += (const Event& other)
{
  all_true |= other.all_true;
  if(all_true)
    return *this;

  if(flags.size() < other.flags.size())
    flags.resize(other.flags.size());
  for(unsigned int i = 0; i < flags.size(); ++i)
    flags[i] = flags[i] || other.flags[i];

  return *this;
}

Event&
Event::operator -= (const Event& other)
{
  if(!other.any())
    return *this;

  all_true = false;
  if(other.all_true)
  {
    for(std::vector<bool>::iterator iter = flags.begin(); iter != flags.end(); ++iter)
      *iter = false;
    return *this;
  }

  if(flags.size() < other.flags.size())
    flags.resize(other.flags.size());
  for(unsigned int i = 0; i < flags.size(); ++i)
    if(other.flags[i])
      flags[i] = false;

  return *this;
}

template<typename OS>
void
Event::print(OS& os) const
{
  if(all_true)
    os << " ALL";

  for(unsigned int i = 0; i < flags.size(); ++i)
    if(flags[i])
      os << ' ' << names[i];
}

template<typename OS>
void
Event::print_assigned(OS& os)
{
  for(unsigned int i = 0; i < names.size(); ++i)
    os << i << '\t' << names[i] << std::endl;
}

/**
 * Output shift operator for
 * events. Calls Event::print().
 *
 * @relates Event
 */
template<typename OS>
OS& operator << (OS& os,
                 const Event& event)
{
  event.print(os);
  return os;
}