// ------------------------------------------------------------------
//  $Id: file_utils.cc 1348 2013-08-16 00:45:07Z secanell $
//
//  Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat.
//  Copyright (C) 2012 by Valentin N. Zingan, University of Alberta.
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
// ------------------------------------------------------------------

#include <base/logstream.h>
#include <base/path_search.h>
#include <base/parameter_handler.h>

#include <appframe/base.h>

#include <boost/shared_ptr.hpp>
#include <fstream>

static boost::shared_ptr<std::ofstream> log_file;

int AppFrame::argc;
char** AppFrame::argv;

void
AppFrame::initialize(int argc, char** argv)
{
  AppFrame::argc = argc;
  AppFrame::argv = argv;

  PathSearch search("PARAMETERS");
  search.add_path("prm/");
}

void
AppFrame::open_logfile (const std::string& extname)
{
				   // If a parameter file was
				   // specified, use its base name for
				   // the logfile.
  std::string name = extname;
				   // If we did not get a filename,
				   // try to get it from the command
				   // line
  if (name == "" && AppFrame::argc > 1)
    {
      for (int i=1;i<argc;++i)
	{
	  std::string tmp = argv[i];
	  std::string::size_type pos = tmp.find_last_of(".log");
	  if (pos == tmp.size() - 1)
	    {
	      name = tmp;
	      break;
	    }
	}
    }
				   // Still no name, use hostname,
				   // allowing parallel programs to
				   // write into different logfiles.
  if (name == "")
    {
      char log_name[40];
      gethostname(log_name, 19);
      name = log_name;
    }

  std::string::size_type pos = name.rfind(".log");
  if (pos < name.size())
    {
      name.erase(pos);
    }
  name += ".log";

  log_file = boost::shared_ptr<std::ofstream> (new std::ofstream(name.c_str()));
				   // Log into file
  deallog.attach(*log_file);
}

void
AppFrame::declare_parameter_files (
  ParameterHandler& param)
  {
     // First, declare additional parameters related to logfiles.
  param.enter_subsection("Log");
  param.declare_entry("Logfile name", "logfile.log", Patterns::Anything());
  param.declare_entry("FileDepth", "10000", Patterns::Integer());
  param.declare_entry("ConsoleDepth", "10000", Patterns::Integer());
  param.declare_entry("Time", "false", Patterns::Bool());
  param.leave_subsection();

  }

void
AppFrame::read_parameter_files(ParameterHandler&  param,
                               const std::string& extname)
{
  std::string name = extname;

  if(!name.empty())
  {
         std::size_t pos = name.find(".");
         std::string ext = name.substr(pos);

         if( ext.compare(".prm") == 0 )
         {
                deallog << "Parameters: " << name << std::endl;
                param.read_input(name,
                                 true);
         }
         else if( ext.compare(".xml") == 0 )
         {
                std::filebuf fb;
                fb.open(name,
                        std::ios::in);
                std::istream is(&fb);
                deallog << "Parameters: " << name << std::endl;
                param.read_input_from_xml(is);
         }
         else
         {
                deallog << "Extension " << ext << " is not handled" << std::endl;
                AssertThrow( false , ExcNotImplemented() );
         }
  }

  if (argc == 1)
    {
      name = argv[0];
      name += ".prm";
      deallog << "Parameters: " << name << std::endl;
      param.read_input(name, false, true);
    }
  else
    {
      for (int i=1;i<argc;++i)
	{
	  name = argv[i];
	  std::string::size_type pos = name.find_last_of(".log");
	  if (pos == name.size() - 1) continue;
	  pos = name.find_last_of(".prm");
	  if (pos != name.size() - 1) continue;
	  deallog << "Parameters: " << name << std::endl;
	  param.read_input(name, true);
	}
    }

				   // Set parameters for execution log
  param.enter_subsection("Log");
  AppFrame::open_logfile(param.get("Logfile name"));
  deallog.depth_file(param.get_integer("FileDepth"));
  deallog.depth_console(param.get_integer("ConsoleDepth"));
  deallog.log_execution_time(param.get_bool("Time"));
  param.leave_subsection();
}