//---------------------------------------------------------------------------
//    $Id: matrix_application.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------


#include <appframe/matrix_application.h>
#include <appframe/mesh_worker_loop.h>
#include <base/parameter_handler.h>
#include <base/quadrature_lib.h>
#include <base/tensor.h>
#include <dofs/dof_handler.h>
#include <multigrid/mg_tools.h>
#include <fe/fe.h>
#include <fe/fe_values.h>

using namespace AppFrame;

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

template <int dim>
MatrixApplication<dim>::MatrixApplication(
    boost::shared_ptr<ApplicationData> data, bool multigrid)
:
DoFApplication<dim>(data, multigrid)
{
    deallog << "->Matrix";
}

//---------------------------------------------------------------------------

template <int dim>
MatrixApplication<dim>::MatrixApplication(bool multigrid)
:
DoFApplication<dim>(multigrid)
{
    deallog << "->Matrix";
}

//---------------------------------------------------------------------------

template <int dim>
MatrixApplication<dim>::MatrixApplication(
    DoFApplication<dim>& other,
    bool triangulation_only,
    bool multigrid)
:
DoFApplication<dim>(other, triangulation_only, multigrid)
{}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::declare_parameters(ParameterHandler& param)
{
    DoFApplication<dim>::declare_parameters(param);
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::_initialize(ParameterHandler& param)
{
    // Get relevant parameter values
    // and initialize objects
    // accordingly.
    param.enter_subsection("Discretization");
    param.enter_subsection("Matrix");
    if (true)
    {
        int n = param.get_integer("Quadrature cell");
        if (n<=0) n = this->element->degree - n;
        boost::shared_ptr<Quadrature<dim> > newq(new QGauss<dim>(n));
        quadrature_assemble_cell = newq;
    }
    if (true)
    {
        int n = param.get_integer("Quadrature face");
        if (n<=0) n = this->element->degree - n;
        boost::shared_ptr<Quadrature<dim-1> > newq(new QGauss<dim-1>(n));
        quadrature_assemble_face = newq;
    }
    param.leave_subsection();
    param.leave_subsection();
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::initialize(ParameterHandler& param)
{
    DoFApplication<dim>::initialize(param);
    _initialize(param);
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::add_matrix(unsigned int row, unsigned int col)
{
    boost::shared_ptr<MatrixType> M(new MatrixType(row, col));
    matrices.push_back(M);
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::add_mg_matrix(unsigned int row, unsigned int col)
{
    Assert (this->mg_dof != 0, ExcNotImplemented());
    boost::shared_ptr<MGMatrixType> M1(new MGMatrixType(row, col));
    boost::shared_ptr<MGMatrixType> M2(new MGMatrixType(row, col));
    boost::shared_ptr<MGMatrixType> M3(new MGMatrixType(row, col));
    mg_matrices.push_back(M1);
    mg_interface_matrices_up.push_back(M2);
    mg_interface_matrices_down.push_back(M3);
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::add_vector_for_cell_matrix(
    std::string name,
    unsigned int,
    unsigned int,
    bool values,
    bool derivatives)
{
    cell_matrix_values.add_vector(name, values, derivatives, false);
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::add_vector_for_flux_matrix(
    std::string name,
    unsigned int,
    unsigned int,
    bool values,
    bool derivatives)
{
    flux_matrix_values.add_vector(name, values, derivatives, false);
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::remesh_matrices()
{
    this->block_info.initialize_local(*this->dof);
    
    if (this->mg_dof)
    {
        // Reinit matrices first to
        // release sparsity pattern
        for (unsigned int k=0;k<mg_matrices.size();++k)
        {
            mg_matrices[k]->matrix.resize(0, this->tr->n_levels()-1);
            mg_matrices[k]->matrix.clear();
            mg_interface_matrices_up[k]->matrix.resize(0, this->tr->n_levels()-1);
            mg_interface_matrices_up[k]->matrix.clear();
            mg_interface_matrices_down[k]->matrix.resize(0, this->tr->n_levels()-1);
            mg_interface_matrices_down[k]->matrix.clear();
        }
        // Then reinit sparsity pattern
        mg_sparsity.resize(0, this->tr->n_levels()-1);
        mg_sparsity_interface.resize(0, this->tr->n_levels()-1);
    }
    
    // clear matrices
    for (unsigned int k=0;k<matrices.size();++k)
        matrices[k]->matrix.clear();
    
    // Make sparsity pattern
    sparsity.reinit(this->block_info.global.size(), this->block_info.global.size());
    BlockCompressedSparsityPattern c_sparsity(this->block_info.global, this->block_info.global);
    if (this->interior_fluxes)
        DoFTools::make_flux_sparsity_pattern(*this->dof, c_sparsity,
                                             cell_couplings, flux_couplings);
        else
            DoFTools::make_sparsity_pattern(*this->dof, cell_couplings,
                                            c_sparsity);
            
            // Condense sparsity pattern to account for hanging nodes
            // this->hanging_node_constraints.condense(c_sparsity);
            sparsity.copy_from(c_sparsity);
        
        for (unsigned int k=0;k<matrices.size();++k)
            matrices[k]->matrix.reinit(sparsity.block(matrices[k]->row,
                                                      matrices[k]->column));
            
            for (unsigned int k=0;k<matrices.size();++k)
                deallog << "Matrix: " << k << '\t'
                << matrices[k]->row << ','
                << matrices[k]->column << '\t'
                << matrices[k]->matrix.m() << 'x'
                << matrices[k]->matrix.n() << std::endl;
            
            if (this->mg_dof)
                for (unsigned int level=mg_sparsity.min_level();
                     level<=mg_sparsity.max_level();++level)
                     
                     {
                         mg_sparsity[level].reinit(this->block_info.levels[level].size(),
                                                   this->block_info.levels[level].size());
                         BlockCompressedSparsityPattern c_sparsity(this->block_info.levels[level],
                                                                   this->block_info.levels[level]);
                         BlockCompressedSparsityPattern ci_sparsity;
                         if (level>0)
                         {
                             mg_sparsity_interface[level].reinit(
                                 this->block_info.levels[level-1].size(),
                                                                 this->block_info.levels[level].size());
                             ci_sparsity.reinit(this->block_info.levels[level-1],
                                                this->block_info.levels[level]);
                         }
                         
                         if (this->interior_fluxes)
                         {
                             MGTools::make_flux_sparsity_pattern(
                                 *this->mg_dof, c_sparsity, level, mg_cell_couplings, mg_flux_couplings);
                             if (level>0)
                                 MGTools::make_flux_sparsity_pattern_edge(
                                     *this->mg_dof, ci_sparsity, level, mg_flux_couplings);
                         }
                         else
                             MGTools::make_sparsity_pattern(*this->mg_dof, c_sparsity, level);
                         mg_sparsity[level].copy_from(c_sparsity);
                         for (unsigned int k=0;k<mg_matrices.size();++k)
                             mg_matrices[k]->matrix[level].reinit(
                                 mg_sparsity[level].block(mg_matrices[k]->row, mg_matrices[k]->column));
                             if (level>0)
                             {
                                 mg_sparsity_interface[level].copy_from(ci_sparsity);
                                 for (unsigned int k=0;k<mg_interface_matrices_up.size();++k)
                                 {
                                     mg_interface_matrices_up[k]->matrix[level].reinit(
                                         mg_sparsity_interface[level].block(mg_interface_matrices_up[k]->row,
                                                                            mg_interface_matrices_up[k]->column));
                                     mg_interface_matrices_down[k]->matrix[level].reinit(
                                         mg_sparsity_interface[level].block(mg_interface_matrices_down[k]->row,
                                                                            mg_interface_matrices_down[k]->column));
                                 }
                             }
                             deallog << "Sparsity Level " << level << " complete" << std::endl;
                     }
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::remesh()
{
    DoFApplication<dim>::remesh();
    remesh_matrices();
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::assemble(const FEVectors& src)
{
    // Clear old matrices
    for (unsigned int i=0;i<matrices.size();++i)
    {
        matrices[i]->matrix = 0;
    }
    
    for (unsigned int k=0;k<mg_matrices.size();++k)
        for (unsigned int level=mg_matrices[k]->matrix.min_level();
             level<=mg_matrices[k]->matrix.max_level();++level)
             mg_matrices[k]->matrix[level] = 0;
        
        // Assemble global system
        deallog.push("ass");
    // Set up integrator class
    LocalMatrixIntegrator<Assembler> integrator;
    integrator.initialize(this, this->block_info, matrices);
    integrator.add_update_flags(update_values | update_gradients | update_q_points);
    integrator.initialize_selectors(cell_matrix_values, flux_matrix_values, flux_matrix_values);
    integrator.initialize_gauss_quadrature(quadrature_assemble_cell->size(),
                                           quadrature_assemble_face->size(),
                                           quadrature_assemble_face->size());
    
    if (this->verbosity)
    {
        src.list(deallog, this->verbosity);
        integrator.cell_selector.cache(src);
        integrator.cell_selector.print(deallog, src);
    }
    
    MeshWorker::InfoObjects::IntegrationInfoBox<dim> box(this->block_info);
    box.initialize(integrator, *this->element, *this->mapping, src);
    
    MeshWorker::integration_loop(this->dof->begin_active(), this->dof->end(), box, integrator);
    
    deallog.pop();
    
    if (mg_matrices.size() != 0)
    {
        deallog.push("mgass");
        LocalMatrixIntegrator<MGAssembler> mg_integrator;
        mg_integrator.initialize(this, this->block_info, mg_matrices);
        mg_integrator.initialize_edge_flux(mg_interface_matrices_up, mg_interface_matrices_down);
        mg_integrator.add_update_flags(update_values | update_gradients | update_q_points);
        mg_integrator.initialize_gauss_quadrature(quadrature_assemble_cell->size(),
                                                  quadrature_assemble_face->size(),
                                                  quadrature_assemble_face->size());
        
        MeshWorker::InfoObjects::IntegrationInfoBox<dim> mg_box(this->block_info);
        mg_box.initialize(mg_integrator, *this->element, *this->mapping, src);
        
        MeshWorker::integration_loop(this->mg_dof->begin(), this->mg_dof->end(), mg_box, mg_integrator);
        deallog.pop();
    }
}


//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::post_cell_assemble()
{}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::cell_matrix(
    MatrixVector& matrices,
    const typename DoFApplication<dim>::CellInfo& info)
{
    deallog << "Void function " << __FUNCTION__ << " called on level "
    << info.cell->level() << std::endl << "Indices ";
    for (unsigned int i=0;i<info.indices.size();++i)
        deallog << ' ' << info.indices[i];
    deallog << std::endl;
    for (unsigned int i=0;i<matrices.size();++i)
        deallog << "Matrix[" << i << "]("
        << matrices[i].row << ','
        << matrices[i].column << ")\t"
        << matrices[i].matrix.n_rows() << 'x'
        << matrices[i].matrix.n_cols() << std::endl;
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::bdry_matrix(
    MatrixVector& matrices,
    const typename DoFApplication<dim>::FaceInfo& info)
{
    deallog << "Void function " << __FUNCTION__ << " called on level "
    << info.cell->level() << std::endl << "Indices ";
    for (unsigned int i=0;i<info.indices.size();++i)
        deallog << ' ' << info.indices[i];
    deallog << std::endl;
    for (unsigned int i=0;i<matrices.size();++i)
        deallog << "Matrix[" << i << "]("
        << matrices[i].row << ','
        << matrices[i].column << ")\t"
        << matrices[i].matrix.n_rows() << 'x'
        << matrices[i].matrix.n_cols() << std::endl;
}

//---------------------------------------------------------------------------

template <int dim>
void
MatrixApplication<dim>::face_matrix(
    MatrixVector& matrices11,
    MatrixVector& matrices12,
    MatrixVector& matrices21,
    MatrixVector& matrices22,
    const typename DoFApplication<dim>::FaceInfo& info1,
    const typename DoFApplication<dim>::FaceInfo& info2)
{
    deallog << "Void function " << __FUNCTION__ << " called on levels "
    << info1.cell->level() << ','
        << info2.cell->level() << std::endl << "Indices1 ";
        for (unsigned int i=0;i<info1.indices.size();++i)
            deallog << ' ' << info1.indices[i];
        deallog << std::endl << "Indices2 ";
        for (unsigned int i=0;i<info2.indices.size();++i)
            deallog << ' ' << info2.indices[i];
        deallog << std::endl;
        
        Assert (matrices11.size() == matrices12.size(),
                ExcDimensionMismatch(matrices11.size(), matrices12.size()));
        Assert (matrices11.size() == matrices21.size(),
                ExcDimensionMismatch(matrices11.size(), matrices21.size()));
        Assert (matrices11.size() == matrices22.size(),
                ExcDimensionMismatch(matrices11.size(), matrices22.size()));
        
        for (unsigned int i=0;i<matrices11.size();++i)
        {
            Assert (matrices11[i].matrix.n_rows() == matrices12[i].matrix.n_rows(),
                    ExcDimensionMismatch(matrices11[i].matrix.n_rows(), matrices12[i].matrix.n_rows()));
            Assert (matrices11[i].matrix.n_cols() == matrices12[i].matrix.n_cols(),
                    ExcDimensionMismatch(matrices11[i].matrix.n_cols(), matrices12[i].matrix.n_cols()));
            Assert (matrices11[i].matrix.n_rows() == matrices21[i].matrix.n_rows(),
                    ExcDimensionMismatch(matrices11[i].matrix.n_rows(), matrices21[i].matrix.n_rows()));
            Assert (matrices11[i].matrix.n_cols() == matrices21[i].matrix.n_cols(),
                    ExcDimensionMismatch(matrices11[i].matrix.n_cols(), matrices21[i].matrix.n_cols()));
            Assert (matrices11[i].matrix.n_rows() == matrices22[i].matrix.n_rows(),
                    ExcDimensionMismatch(matrices11[i].matrix.n_rows(), matrices22[i].matrix.n_rows()));
            Assert (matrices11[i].matrix.n_cols() == matrices22[i].matrix.n_cols(),
                    ExcDimensionMismatch(matrices11[i].matrix.n_cols(), matrices22[i].matrix.n_cols()));
            
            deallog << "Matrix[" << i << "]("
            << matrices11[i].row << ','
            << matrices11[i].column << ")\t"
            << matrices11[i].matrix.n_rows() << 'x'
            << matrices11[i].matrix.n_cols() << std::endl;
        }
}


template <int dim>
void
MatrixApplication<dim>::mg_cell_matrix(
    MatrixVector& matrices,
    const typename DoFApplication<dim>::CellInfo& info)
{
    cell_matrix(matrices, info);
}


template <int dim>
void
MatrixApplication<dim>::mg_bdry_matrix(
    MatrixVector& matrices,
    const typename DoFApplication<dim>::FaceInfo& info)
{
    bdry_matrix(matrices, info);
}


template <int dim>
void
MatrixApplication<dim>::mg_face_matrix(
    MatrixVector& matrices11,
    MatrixVector& matrices12,
    MatrixVector& matrices21,
    MatrixVector& matrices22,
    const typename DoFApplication<dim>::FaceInfo& info1,
    const typename DoFApplication<dim>::FaceInfo& info2)
{
    face_matrix(matrices11, matrices12, matrices21, matrices22, info1, info2);
}



template class MatrixApplication<deal_II_dimension>;

