// $Id: matrix_shop_bdry.cc 1348 2013-08-16 00:45:07Z secanell $

#include <appframe/matrix_shop_bdry.h>

#include <base/point.h>
#include <base/vector_slice.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#include <iostream>

using namespace std; 

#define NAMEPREFIX AppShop::Matrix::Boundary

template<int dim>
void
NAMEPREFIX::nitsche (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& int_fe,
  const double penalty,
  const double factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int n_comp = int_fe.get_fe().n_components();
  
  Assert (M.m() == n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  Assert (M.n() == n_dofs, ExcDimensionMismatch(M.n(), n_dofs));

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k) * factor;
      const Point<dim>& n = int_fe.normal_vector(k);
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  for (unsigned int d=0;d<n_comp;++d)
	    M(i,j) += dx *
		      (int_fe.shape_value_component(i,k,d) * penalty * int_fe.shape_value_component(j,k,d)
		       - (n * int_fe.shape_grad_component(i,k,d)) * int_fe.shape_value_component(j,k,d)
		       - (n * int_fe.shape_grad_component(j,k,d)) * int_fe.shape_value_component(i,k,d));
    }
}


template<int dim>
void
NAMEPREFIX::nitsche_scaled (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& int_fe,
  const std::vector<double>& diffusion,
  const double penalty)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;

  Assert (M.m() == n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  Assert (M.n() == n_dofs, ExcDimensionMismatch(M.n(), n_dofs));
  
  if (int_fe.get_fe().is_primitive())
    {
      for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
	{
	  const double dx = int_fe.JxW(k);
	  const Point<dim>& n = int_fe.normal_vector(k);
	  const double nu = diffusion[k];
	  for (unsigned i=0;i<n_dofs;++i)
	    {
	      const double vi = int_fe.shape_value(i,k);
	      const double dnvi = n * int_fe.shape_grad(i,k);
	      for (unsigned j=0;j<n_dofs;++j)
		{
		  const double ui = int_fe.shape_value(j,k);
		  const double dnui = n * int_fe.shape_grad(j,k);
		  M(i,j) += dx*nu*(vi * penalty * ui - dnvi * ui - dnui * vi);
		}
	    }
	}
    }
  else
    {
      for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
	{
	  const double dx = int_fe.JxW(k);
	  const Point<dim>& n = int_fe.normal_vector(k);
	  const double nu = diffusion[k];
 	  for (unsigned int d=0;d<int_fe.get_fe().n_components();++d)
	    for (unsigned i=0;i<n_dofs;++i)
	      {
		const double vi = int_fe.shape_value_component(i,k,d);
		const double dnvi = n * int_fe.shape_grad_component(i,k,d);
		for (unsigned j=0;j<n_dofs;++j)
		  {
		    const double ui = int_fe.shape_value_component(j,k,d);
		    const double dnui = n * int_fe.shape_grad_component(j,k,d);
		    M(i,j) += dx*nu*(vi * penalty * ui - dnvi * ui - dnui * vi);
		  }
	      }
	}
    }
}


template<int dim>
void
NAMEPREFIX::nitsche_div (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& int_fe,
  const double penalty)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const Point<dim>& n = int_fe.normal_vector(k);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dvi = int_fe.shape_grad(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      const Tensor<1,dim>& Dui = int_fe.shape_grad(j,k);
	      for (unsigned int d1=0;d1<dim;++d1)
		for (unsigned int d2=0;d2<dim;++d2)
		  M(i+d1*n_dofs,j+d2*n_dofs) +=
		    dx * (-Dui[d2]*n(d1)*vi - Dvi[d1]*n(d2)*ui
			  +penalty*n(d1)*n(d2)*ui*vi);
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::jump_scaled (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& int_fe,
  const std::vector<double>& factor,
  const double factor2,
  bool divide)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  Assert (M.m() == n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  Assert (M.n() == n_dofs, ExcDimensionMismatch(M.n(), n_dofs));
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2 * ((divide)
				   ? (1./factor[k])
				   : (factor[k]));
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      M(i,j) += dx*nu*ui*vi;
	    }
	}
    }
}



template<int dim>
void
NAMEPREFIX::jump (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& int_fe,
  const double factor2)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  Assert (M.m() == n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  Assert (M.n() == n_dofs, ExcDimensionMismatch(M.n(), n_dofs));
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2;
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      M(i,j) += dx*nu*ui*vi;
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& fe,
  const std::vector<Point<dim> >& velocity,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
				   // Multiply the quadrature point
				   // index below with this factor to
				   // have simpler data for constant
				   // velocities.
  unsigned int constant_velocity = (velocity.size() == 1) ? 0 : 1;  
  
  Assert (M.m() == n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  Assert (M.n() == n_dofs, ExcDimensionMismatch(M.n(), n_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      const double nbeta = fe.normal_vector(k) * velocity[k * constant_velocity];

      if (nbeta > 0)
	{
	  for (unsigned i=0;i<n_dofs;++i)
	    {
	      for (unsigned j=0;j<n_dofs;++j)
		{
		  if (fe.get_fe().is_primitive())
		    M(i,j) += dx * nbeta
			      * fe.shape_value(i,k)
			      * fe.shape_value(j,k);
		  else
		    for (unsigned int d=0;d<dim;++d)
		      M(i,j) += dx * nbeta
				* fe.shape_value_component(i,k,d)
				* fe.shape_value_component(j,k,d);
		}
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell; 
  unsigned int n_components = fe.get_fe().n_components();
  AssertDimension(velocity.size(), dim);
  Assert (M.m() == n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  Assert (M.n() == n_dofs, ExcDimensionMismatch(M.n(), n_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);

      double nv = 0.;
      for (unsigned int d=0;d<dim;++d)
	nv += fe.normal_vector(k)[d] * velocity[d][k];
      
      if (nv > 0)
	{
	  for (unsigned i=0;i<n_dofs;++i)
	    for (unsigned j=0;j<n_dofs;++j)
	      {
		if (n_components == 1)
		  M(i,j) += dx * nv * fe.shape_value(i,k) * fe.shape_value(j,k);
		else
		  for (unsigned int c=0;c<n_components;++c)
		      M(i,j) += dx * nv * fe.shape_value_component(i,k,c)
				* fe.shape_value_component(j,k,c);
	      }
	}
    }
}



template<int dim>
void
NAMEPREFIX::u_dot_n (
  std::vector<MatrixBlock<FullMatrix<double> > >& M,
  const std::vector<unsigned int>& indices,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest,
  double factor)
{
  unsigned int fecomp = fe.get_fe().n_components();
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  Assert (indices.size() == dim/fecomp, ExcDimensionMismatch(indices.size(), dim/fecomp));  
  for (unsigned int i=0;i<indices.size();++i)
    {
      Assert (M[indices[i]].matrix.m()==t_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.m(), t_dofs));
      Assert (M[indices[i]].matrix.n()==n_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.n(), n_dofs));
    }
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const Tensor<1,dim> ndx = factor * fe.JxW(k) * fe.normal_vector(k);
      for (unsigned i=0;i<t_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  if (fe.get_fe().n_components() == 1)
	    for (unsigned int d=0;d<dim;++d)
	      M[indices[d]].matrix(i,j) += ndx[d]
					   * fe.shape_value(j,k)
					   * fetest.shape_value(i,k);
	  else
	    for (unsigned int d=0;d<dim;++d)
	      M[0].matrix(i,j) += ndx[d]
				  * fe.shape_value_component(j,k,d)
				  * fetest.shape_value(i,k);
    }
}


template<int dim>
void
NAMEPREFIX::u_dot_n_scaled (
  std::vector<MatrixBlock<FullMatrix<double> > >& M,
  const std::vector<unsigned int>& indices,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest,
  const std::vector<double>& factor,
  bool divide)
{
  unsigned int fecomp = fe.get_fe().n_components();
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  Assert (indices.size() == dim/fecomp, ExcDimensionMismatch(indices.size(), dim/fecomp));  
  for (unsigned int i=0;i<indices.size();++i)
    {
      Assert (M[indices[i]].matrix.m()==t_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.m(), t_dofs));
      Assert (M[indices[i]].matrix.n()==n_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.n(), n_dofs));
    }
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const Tensor<1,dim> ndx = ((divide) ? (1./factor[k]) : (factor[k]))
				* fe.JxW(k) * fe.normal_vector(k);
      for (unsigned i=0;i<t_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  if (fe.get_fe().n_components() == 1)
	    for (unsigned int d=0;d<dim;++d)
	      M[indices[d]].matrix(i,j) += ndx[d]
					   * fe.shape_value(j,k)
					   * fetest.shape_value(i,k);
	  else
	    for (unsigned int d=0;d<dim;++d)
	      M[0].matrix(i,j) += ndx[d]
				  * fe.shape_value_component(j,k,d)
				  * fetest.shape_value(i,k);
    }
}


template<int dim>
void
NAMEPREFIX::u_ni (
  FullMatrix<double>& M,
  const FEFaceValuesBase<dim>& fetest,
  const FEFaceValuesBase<dim>& fe,
  unsigned int d)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  Assert (M.m()==t_dofs, ExcDimensionMismatch(M.m(), t_dofs));
  Assert (M.n()==n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      const Point<dim>& n = fe.normal_vector(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vv = fetest.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double uu = fe.shape_value(j,k);
	      M(i,j) += dx * vv * uu * n(d);
	    }
	}
    }
}

#define dim deal_II_dimension

template void NAMEPREFIX::advection<dim> (
  FullMatrix<double>&, const FEFaceValuesBase<dim>&,
  const std::vector<Point<dim> >&, double);
template void NAMEPREFIX::advection<dim> (
  FullMatrix<double>&, const FEFaceValuesBase<dim>&,
  const VectorSlice<const std::vector<std::vector<double> > >&, double);
template void NAMEPREFIX::nitsche<dim> (
  FullMatrix<double>& M, const FEFaceValuesBase<dim>& int_fe,
  const double penalty, const double factor);
template void NAMEPREFIX::nitsche_scaled<dim> (
  FullMatrix<double>& M, const FEFaceValuesBase<dim>& int_fe,
  const std::vector<double>& diffusion, const double penalty);
template void NAMEPREFIX::nitsche_div<dim> (
  FullMatrix<double>&, const FEFaceValuesBase<dim>&, double);
template void NAMEPREFIX::u_dot_n<dim> (
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  const std::vector<unsigned int>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&, double);
template void NAMEPREFIX::u_dot_n_scaled<dim> (
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  const std::vector<unsigned int>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const std::vector<double>&, bool);

