//---------------------------------------------------------------------------
//    $Id: matrix_shop_cell.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <appframe/matrix_shop_cell.h>
#include <appframe/matrix_application.h> //This is so that we can use MatrixBlock

#include <base/point.h>
#include <base/vector_slice.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#include <iostream>


#define NAMEPREFIX AppShop::Matrix::Cell

using namespace AppFrame;

template<int dim>
void
NAMEPREFIX::mass (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) * factor;

      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned j=0;j<n_dofs;++j)
	    for (unsigned int d=0;d<n_components;++d)
	      M(i,j) += dx
			* fe.shape_value_component(j,k,d)
			  * fe.shape_value_component(i,k,d);
	}
    }
}


template<int dim>
void
NAMEPREFIX::mass (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  const unsigned int select,
  const double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
  const unsigned int t_components = fetest.get_fe().n_components();

  AssertDimension(M.m(), t_dofs);
  AssertDimension(M.n(), n_dofs);  
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) * factor;

      if (n_components == t_components)
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    for (unsigned j=0;j<n_dofs;++j)
	      for (unsigned int d=0;d<n_components;++d)
		M(i,j) += dx
			  * fe.shape_value_component(j,k,d)
			  * fetest.shape_value_component(i,k,d);
	  }
      else if (n_components == 1 && t_components != 1)
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    for (unsigned j=0;j<n_dofs;++j)
	      M(i,j) += dx
			* fe.shape_value(j,k)
			* fetest.shape_value_component(i,k,select);
	  }
      else if (t_components == 1 && n_components != 1)
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    for (unsigned j=0;j<n_dofs;++j)
	      M(i,j) += dx
			* fe.shape_value_component(j,k, select)
			* fetest.shape_value(i,k);
	  }
      else
	{
	  Assert (false, ExcNotImplemented());
	}
    }
}


template<int dim>
void
NAMEPREFIX::mass_scaled (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      const double nu = (divide)
			? (1./factor[k])
			: (factor[k]);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vv = fe.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double uu = fe.shape_value(j,k);
	      M(i,j) += dx * nu * uu * vv;
	    }
	}
    }
}

template <int dim>
void 
NAMEPREFIX::mass_scaled (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  const unsigned int select,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
  const unsigned int t_components = fetest.get_fe().n_components();

  AssertDimension(M.m(), t_dofs);
  AssertDimension(M.n(), n_dofs);  
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double nu = (divide)
			? (1./factor[k])
			: (factor[k]);
      const double dx = fe.JxW(k) * nu;

      if (n_components == t_components)
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    for (unsigned j=0;j<n_dofs;++j)
	    {
	      for (unsigned int d=0;d<n_components;++d)
	      {
		M(i,j) += dx
			  * fe.shape_value_component(j,k,d)
			  * fetest.shape_value_component(i,k,d);
		//deallog << "M(" << i << "," << j << ") " << M(i,j) << std::endl;
	      }
	    }
	  }
      else if (n_components == 1 && t_components != 1)
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    for (unsigned j=0;j<n_dofs;++j)
	      M(i,j) += dx
			* fe.shape_value(j,k)
			* fetest.shape_value_component(i,k,select);
	  }
      else if (t_components == 1 && n_components != 1)
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    for (unsigned j=0;j<n_dofs;++j)
	      M(i,j) += dx
			* fe.shape_value_component(j,k, select)
			* fetest.shape_value(i,k);
	  }
      else
	{
	  Assert (false, ExcNotImplemented());
	}
    }  
}
      
      
template<int dim>
void
NAMEPREFIX::laplacian (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned j=0;j<n_dofs;++j)
	    for (unsigned int d=0;d<n_components;++d)
	      M(i,j) += dx * factor *
		(fe.shape_grad_component(j,k,d) * fe.shape_grad_component(i,k,d));
	}
    }
}

template<int dim>
void
NAMEPREFIX::laplacian (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const Tensor<2,dim> factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();

  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned j=0;j<n_dofs;++j)
	    for (unsigned int d=0;d<n_components;++d)
	      M(i,j) += dx * 
		(fe.shape_grad_component(j,k,d) * factor * fe.shape_grad_component(i,k,d));
	}
    }
}

template<int dim>
void
NAMEPREFIX::laplacian_scaled (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
  {
    const double dx = fe.JxW(k);
    const double nu = (divide)
      ? (1./factor[k])
      : (factor[k]);
      for (unsigned i=0;i<n_dofs;++i)
      {
	for (unsigned j=0;j<n_dofs;++j)
	{
	  if (fe.get_fe().is_primitive())
	    M(i,j) += dx * nu * 
	      (fe.shape_grad(j,k) * fe.shape_grad(i,k));
	  else
	  {
	    for (unsigned int d=0;d<fe.get_fe().n_components();++d)
	    {
	      const Tensor<1,dim>& Du = fe.shape_grad_component(j,k,d);
	      const Tensor<1,dim>& Dv = fe.shape_grad_component(i,k,d);
	      M(i,j) += dx * nu * (Du * Dv);
	    }
	  }
	}
      }
    }
}

template<int dim>
void
NAMEPREFIX::laplacian_scaled (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const double factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
  {
    const double dx = fe.JxW(k);
    const double nu = (divide)
      ? (1./factor)
      : (factor);
      for (unsigned i=0;i<n_dofs;++i)
      {
	for (unsigned j=0;j<n_dofs;++j)
	{
	  if (fe.get_fe().is_primitive())
	    M(i,j) += dx * nu * 
	      (fe.shape_grad(j,k) * fe.shape_grad(i,k));
	  else
	  {
	    for (unsigned int d=0;d<fe.get_fe().n_components();++d)
	    {
	      const Tensor<1,dim>& Du = fe.shape_grad_component(j,k,d);
	      const Tensor<1,dim>& Dv = fe.shape_grad_component(i,k,d);
	      M(i,j) += dx * nu * (Du * Dv);
	    }
	  }
	}
      }
    }
}

template<int dim>
void
NAMEPREFIX::div_div (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  AssertDimension(fe.get_fe().n_components(), dim);
  AssertDimension(M.m(), n_dofs);
  AssertDimension(M.n(), n_dofs);  
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  for (unsigned int d1=0;d1<dim;++d1)
	    for (unsigned int d2=0;d2<dim;++d2)
	      M(i,j) += dx * fe.shape_grad_component(i,k,d1)[d1] * fe.shape_grad_component(j,k,d2)[d2];
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const std::vector<Point<dim> >& velocity)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
				   // Multiply the quadrature point
				   // index below with this factor to
				   // have simpler data for constant
				   // velocities.
  unsigned int constant_velocity = (velocity.size() == 1) ? 0 : 1;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      const Point<dim>& ww = velocity[k * constant_velocity];
      
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  for (unsigned int d=0;d<n_components;++d)
	    M(i,j) -= dx
	      * (fe.shape_grad_component(i,k,d) * ww)
	      * fe.shape_value_component(j,k,d);
    }
}



template<int dim>
void
NAMEPREFIX::advection (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  AssertDimension(velocity.size(), dim);
  
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
  AssertDimension(M.n(), n_dofs);
  AssertDimension(M.m(), n_dofs);
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  for (unsigned int c=0;c<n_components;++c)
	    for (unsigned int d=0;d<dim;++d)
	      M(i,j) -= dx * fe.shape_grad_component(i,k,c)[d] * velocity[d][k]
		* fe.shape_value_component(j,k,c);
    }
}



template<int dim>
void
NAMEPREFIX::div_primal (
  std::vector<MatrixBlock<FullMatrix<double> > >& M,
  const std::vector<unsigned int>& indices,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  double factor)
{
  unsigned int fecomp = fe.get_fe().n_components();
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;

  Assert (indices.size() == dim/fecomp, ExcDimensionMismatch(indices.size(), dim/fecomp));  
  for (unsigned int i=0;i<indices.size();++i)
    {
      Assert (M[indices[i]].matrix.m()==t_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.m(), t_dofs));
      Assert (M[indices[i]].matrix.n()==n_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.n(), n_dofs));
    }
  
  if (fe.get_fe().n_components() == 1)
    {
      for (unsigned k=0;k<fe.n_quadrature_points;++k)
	{
	  const double dx = factor * fe.JxW(k);
	  for (unsigned i=0;i<t_dofs;++i)
	    {
	      const Tensor<1,dim>& Dv = fetest.shape_grad(i,k);
	      for (unsigned j=0;j<n_dofs;++j)
		{
		  const double uu = fe.shape_value(j,k);
		  for (unsigned int d=0;d<dim;++d)
		    M[indices[d]].matrix(i,j) -= dx * Dv[d] * uu;
		}
	    }
	}
    }
  else
    {      
      for (unsigned k=0;k<fe.n_quadrature_points;++k)
	{
	  const double dx = fe.JxW(k);
	  for (unsigned i=0;i<t_dofs;++i)
	    {
	      const double vv = fetest.shape_value(i,k);
	      for (unsigned int d=0;d<dim;++d)
		for (unsigned j=0;j<n_dofs;++j)
		  {
		    const double du = fe.shape_grad_component(j,k,d)[d];
		    M[indices[0]].matrix(i,j) += dx * du * vv;
		  }
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::div (
  std::vector<MatrixBlock<FullMatrix<double> > >& M,
  const std::vector<unsigned int>& indices,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  double factor)
{
  unsigned int fecomp = fe.get_fe().n_components();
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;

  Assert (indices.size() == dim/fecomp, ExcDimensionMismatch(indices.size(), dim/fecomp));  
  for (unsigned int i=0;i<indices.size();++i)
    {
      Assert (M[indices[i]].matrix.m()==t_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.m(), t_dofs));
      Assert (M[indices[i]].matrix.n()==n_dofs,
	      ExcDimensionMismatch(M[indices[i]].matrix.n(), n_dofs));
    }
  
  if (fe.get_fe().n_components() == 1)
    {
      for (unsigned k=0;k<fe.n_quadrature_points;++k)
	{
	  const double dx = factor * fe.JxW(k);
	  for (unsigned i=0;i<t_dofs;++i)
	    {
	      const Tensor<1,dim>& Dv = fetest.shape_grad(i,k);
	      for (unsigned j=0;j<n_dofs;++j)
		{
		  const double uu = fe.shape_value(j,k);
		  for (unsigned int d=0;d<dim;++d)
		    M[indices[d]].matrix(i,j) -= dx * Dv[d] * uu;
		}
	    }
	}
    }
  else
    {      
      for (unsigned k=0;k<fe.n_quadrature_points;++k)
	{
	  const double dx = fe.JxW(k);
	  for (unsigned i=0;i<t_dofs;++i)
	    {
	      const Tensor<1,dim>& Dv = fetest.shape_grad(i,k);
	      for (unsigned int d=0;d<dim;++d)
		for (unsigned j=0;j<n_dofs;++j)
		  {
		    const double uu = fe.shape_value_component(j,k,d);
		    M[indices[0]].matrix(i,j) -= dx * uu * Dv[d];
		  }
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::div_scaled (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest, 
  const std::vector<double>& factor)
{
				   // Still transposed?
  Assert(false, ExcNotImplemented());
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;

  Assert (M.m()==t_dofs, ExcDimensionMismatch(M.m(), t_dofs));
  Assert (M.n()==dim*n_dofs, ExcDimensionMismatch(M.m(), dim*n_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vv = fe.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const Tensor<1,dim>& Du = fetest.shape_grad(j,k);
	      for (unsigned int d=0;d<dim;++d)
		M(j,i+d*n_dofs) -= factor[k]* dx * Du[d] * vv;
	    }
	}
    }
}

template<int dim>
void
NAMEPREFIX::di (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  unsigned int d)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;

  Assert (M.m()==t_dofs, ExcDimensionMismatch(M.m(), t_dofs));
  Assert (M.n()==n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vv = fetest.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const Tensor<1,dim>& Du = fe.shape_grad(j,k);
	      M(i,j) -= dx * Du[d] * vv;
	    }
	}
    }
}

template<int dim>
void
NAMEPREFIX::di_scaled (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  unsigned int d,
  std::vector<double>& factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;

  Assert (M.m()==t_dofs, ExcDimensionMismatch(M.m(), t_dofs));
  Assert (M.n()==n_dofs, ExcDimensionMismatch(M.m(), n_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vv = fe.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const Tensor<1,dim>& Du = fetest.shape_grad(j,k);
	      M(j,i) -= dx * Du[d] * vv * factor[k];
	    }
	}
    }
}

template<int dim>
void
NAMEPREFIX::eps_div (
  FullMatrix<double>& M,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  Assert (M.m()==dim*t_dofs, ExcDimensionMismatch(M.m(), dim*t_dofs));
  Assert (M.n()==symm_dim(dim)*n_dofs,
	  ExcDimensionMismatch(M.m(), symm_dim(dim)*n_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vv = fetest.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const Tensor<1,dim>& Du = fe.shape_grad(j,k);
	      for (unsigned int d1=0;d1<dim;++d1)
		{
		  M(i+d1*t_dofs,j+d1*n_dofs) -= dx * Du[d1] * vv;
		  for (unsigned int d2=d1+1;d2<dim;++d2)
		    {
		      Assert (dim==2, ExcNotImplemented());
		      const unsigned int d3 = dim + d2 - 1;
		      M(i+d1*t_dofs,j+d3*n_dofs) -= .5 * dx * Du[d1] * vv;
		      M(i+d1*t_dofs,j+d3*n_dofs) -= .5 * dx * Du[d2] * vv;
		    }
		}
	    }
	}
    }
}


#define dim deal_II_dimension

template void NAMEPREFIX::mass<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, double);

template void NAMEPREFIX::mass<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, const FEValuesBase<dim>&,
  unsigned int, double);

template void NAMEPREFIX::mass_scaled<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&,
  const std::vector<double>&, bool);

template void NAMEPREFIX::mass_scaled<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, const FEValuesBase<dim>&,
  unsigned int, const std::vector<double>&, bool);

template void NAMEPREFIX::laplacian<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, const double);

template void NAMEPREFIX::laplacian<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, const Tensor<2,dim>);

template void NAMEPREFIX::laplacian_scaled<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&,
  const std::vector<double>&, bool);

template void NAMEPREFIX::laplacian_scaled<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&,
  const double, bool);

template void NAMEPREFIX::div_div<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, double);

template void NAMEPREFIX::advection<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&,
  const std::vector<Point<dim> >&);

template void NAMEPREFIX::advection<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&,
  const VectorSlice<const std::vector<std::vector<double> > >&, double);

template void NAMEPREFIX::div_primal<dim> (
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  const std::vector<unsigned int>&,
  const FEValuesBase<dim>&, const FEValuesBase<dim>&, double);
template void NAMEPREFIX::div<dim> (
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  const std::vector<unsigned int>&,
  const FEValuesBase<dim>&, const FEValuesBase<dim>&, double);

template void NAMEPREFIX::div_scaled<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&,
  const FEValuesBase<dim>&, const std::vector<double>&);

template void NAMEPREFIX::di<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&,
  const FEValuesBase<dim>&, unsigned int);

template void NAMEPREFIX::di_scaled<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, const FEValuesBase<dim>&,
  unsigned int, std::vector<double>&);

template void NAMEPREFIX::eps_div<dim> (
  FullMatrix<double>&, const FEValuesBase<dim>&, const FEValuesBase<dim>&);
