// $Id: matrix_shop_face.cc 1348 2013-08-16 00:45:07Z secanell $

#include <appframe/matrix_shop_face.h>

#include <base/point.h>
#include <base/vector_slice.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#include <iostream>

#define NAMEPREFIX AppShop::Matrix::Face

template<int dim>
void
NAMEPREFIX::ip (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const double pen,
  const double int_factor,
  const double ext_factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  Assert(intint.n() == n_dofs, ExcDimensionMismatch(intint.n(), n_dofs));
  Assert(intint.m() == n_dofs, ExcDimensionMismatch(intint.m(), n_dofs));
  Assert(intext.n() == n_dofs, ExcDimensionMismatch(intext.n(), n_dofs));
  Assert(intext.m() == n_dofs, ExcDimensionMismatch(intext.m(), n_dofs));
  Assert(extint.n() == n_dofs, ExcDimensionMismatch(extint.n(), n_dofs));
  Assert(extint.m() == n_dofs, ExcDimensionMismatch(extint.m(), n_dofs));
  Assert(extext.n() == n_dofs, ExcDimensionMismatch(extext.n(), n_dofs));
  Assert(extext.m() == n_dofs, ExcDimensionMismatch(extext.m(), n_dofs));

  const double nui = int_factor;
  const double nue = (ext_factor < 0) ? int_factor : ext_factor;
  const double penalty = .5 * pen * (nui + nue);
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const Point<dim>& n = int_fe.normal_vector(k);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      if (int_fe.get_fe().n_components() == 1)
		{
		  const double vi = int_fe.shape_value(i,k);
		  const double dnvi = n * int_fe.shape_grad(i,k);
		  const double ve = ext_fe.shape_value(i,k);
		  const double dnve = n * ext_fe.shape_grad(i,k);
		  const double ui = int_fe.shape_value(j,k);
		  const double dnui = n * int_fe.shape_grad(j,k);
		  const double ue = ext_fe.shape_value(j,k);
		  const double dnue = n * ext_fe.shape_grad(j,k);
		  
		  intint(i,j) += dx*(-.5*nui*dnvi*ui-.5*nui*dnui*vi+penalty*ui*vi);
		  intext(i,j) += dx*( .5*nui*dnvi*ue-.5*nue*dnue*vi-penalty*vi*ue);
		  extint(i,j) += dx*(-.5*nue*dnve*ui+.5*nui*dnui*ve-penalty*ui*ve);
		  extext(i,j) += dx*( .5*nue*dnve*ue+.5*nue*dnue*ve+penalty*ue*ve);
		}
	      else
		for (unsigned int d=0;d<dim;++d)
		  {
		    const double vi = int_fe.shape_value_component(i,k,d);
		    const double dnvi = n * int_fe.shape_grad_component(i,k,d);
		    const double ve = ext_fe.shape_value_component(i,k,d);
		    const double dnve = n * ext_fe.shape_grad_component(i,k,d);
		    const double ui = int_fe.shape_value_component(j,k,d);
		    const double dnui = n * int_fe.shape_grad_component(j,k,d);
		    const double ue = ext_fe.shape_value_component(j,k,d);
		    const double dnue = n * ext_fe.shape_grad_component(j,k,d);
		    
		    intint(i,j) += dx*(-.5*nui*dnvi*ui-.5*nui*dnui*vi+penalty*ui*vi);
		    intext(i,j) += dx*( .5*nui*dnvi*ue-.5*nue*dnue*vi-penalty*vi*ue);
		    extint(i,j) += dx*(-.5*nue*dnve*ui+.5*nui*dnui*ve-penalty*ui*ve);
		    extext(i,j) += dx*( .5*nue*dnve*ue+.5*nue*dnue*ve+penalty*ue*ve);
		  }
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::ip_scaled (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<double>& diffusion,
  const double penalty)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  Assert(intint.n() == n_dofs, ExcDimensionMismatch(intint.n(), n_dofs));
  Assert(intint.m() == n_dofs, ExcDimensionMismatch(intint.m(), n_dofs));
  Assert(intext.n() == n_dofs, ExcDimensionMismatch(intext.n(), n_dofs));
  Assert(intext.m() == n_dofs, ExcDimensionMismatch(intext.m(), n_dofs));
  Assert(extint.n() == n_dofs, ExcDimensionMismatch(extint.n(), n_dofs));
  Assert(extint.m() == n_dofs, ExcDimensionMismatch(extint.m(), n_dofs));
  Assert(extext.n() == n_dofs, ExcDimensionMismatch(extext.n(), n_dofs));
  Assert(extext.m() == n_dofs, ExcDimensionMismatch(extext.m(), n_dofs));
  
  if (int_fe.get_fe().n_components() == 1)
    {
      for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
	{
	  const double dx = int_fe.JxW(k);
	  const Point<dim>& n = int_fe.normal_vector(k);
	  const double nu = diffusion[k];
	  for (unsigned i=0;i<n_dofs;++i)
	    {
	      const double vi = int_fe.shape_value(i,k);
	      const double dnvi = n * int_fe.shape_grad(i,k);
	      const double ve = ext_fe.shape_value(i,k);
	      const double dnve = n * ext_fe.shape_grad(i,k);
	      for (unsigned j=0;j<n_dofs;++j)
		{
		  const double ui = int_fe.shape_value(j,k);
		  const double dnui = n * int_fe.shape_grad(j,k);
		  const double ue = ext_fe.shape_value(j,k);
		  const double dnue = n * ext_fe.shape_grad(j,k);
		  
		  intint(i,j) += dx*nu*(-.5*dnvi*ui-.5*dnui*vi+penalty*ui*vi);
		  intext(i,j) += dx*nu*( .5*dnvi*ue-.5*dnue*vi-penalty*vi*ue);
		  extint(i,j) += dx*nu*(-.5*dnve*ui+.5*dnui*ve-penalty*ui*ve);
		  extext(i,j) += dx*nu*( .5*dnve*ue+.5*dnue*ve+penalty*ue*ve);
		}
	    }
	}
    }
  else
    {
      for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
	{
	  const double dx = int_fe.JxW(k);
	  const Point<dim>& n = int_fe.normal_vector(k);
	  const double nu = diffusion[k];
 	  for (unsigned int d=0;d<int_fe.get_fe().n_components();++d)
	    for (unsigned i=0;i<n_dofs;++i)
	      {
		const double vi = int_fe.shape_value_component(i,k,d);
		const double dnvi = n * int_fe.shape_grad_component(i,k,d);
		const double ve = ext_fe.shape_value_component(i,k,d);
		const double dnve = n * ext_fe.shape_grad_component(i,k,d);
		for (unsigned j=0;j<n_dofs;++j)
		  {
		    const double ui = int_fe.shape_value_component(j,k,d);
		    const double dnui = n * int_fe.shape_grad_component(j,k,d);
		    const double ue = ext_fe.shape_value_component(j,k,d);
		    const double dnue = n * ext_fe.shape_grad_component(j,k,d);
		    
		    intint(i,j) += dx*nu*(-.5*dnvi*ui-.5*dnui*vi+penalty*ui*vi);
		    intext(i,j) += dx*nu*( .5*dnvi*ue-.5*dnue*vi-penalty*vi*ue);
		    extint(i,j) += dx*nu*(-.5*dnve*ui+.5*dnui*ve-penalty*ui*ve);
		    extext(i,j) += dx*nu*( .5*dnve*ue+.5*dnue*ve+penalty*ue*ve);
		  }
	      }
	}
    }
}


template<int dim>
void
NAMEPREFIX::jump_scaled (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<double>& factor,
  const double factor2,
  bool divide)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  Assert(intint.n() == n_dofs, ExcDimensionMismatch(intint.n(), n_dofs));
  Assert(intint.m() == n_dofs, ExcDimensionMismatch(intint.m(), n_dofs));
  Assert(intext.n() == n_dofs, ExcDimensionMismatch(intext.n(), n_dofs));
  Assert(intext.m() == n_dofs, ExcDimensionMismatch(intext.m(), n_dofs));
  Assert(extint.n() == n_dofs, ExcDimensionMismatch(extint.n(), n_dofs));
  Assert(extint.m() == n_dofs, ExcDimensionMismatch(extint.m(), n_dofs));
  Assert(extext.n() == n_dofs, ExcDimensionMismatch(extext.n(), n_dofs));
  Assert(extext.m() == n_dofs, ExcDimensionMismatch(extext.m(), n_dofs));
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2 * ((divide)
				   ? (1./factor[k])
				   : (factor[k]));
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const double ve = ext_fe.shape_value(i,k);  
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      const double ue = ext_fe.shape_value(j,k);
	      intint(i,j) += dx*nu*ui*vi;
	      intext(i,j) -= dx*nu*ue*vi;
	      extint(i,j) -= dx*nu*ui*ve;
	      extext(i,j) += dx*nu*ue*ve;
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::jump (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const double factor2)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  Assert(intint.n() == n_dofs, ExcDimensionMismatch(intint.n(), n_dofs));
  Assert(intint.m() == n_dofs, ExcDimensionMismatch(intint.m(), n_dofs));
  Assert(intext.n() == n_dofs, ExcDimensionMismatch(intext.n(), n_dofs));
  Assert(intext.m() == n_dofs, ExcDimensionMismatch(intext.m(), n_dofs));
  Assert(extint.n() == n_dofs, ExcDimensionMismatch(extint.n(), n_dofs));
  Assert(extint.m() == n_dofs, ExcDimensionMismatch(extint.m(), n_dofs));
  Assert(extext.n() == n_dofs, ExcDimensionMismatch(extext.n(), n_dofs));
  Assert(extext.m() == n_dofs, ExcDimensionMismatch(extext.m(), n_dofs));

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2;
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const double ve = ext_fe.shape_value(i,k);  
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      const double ue = ext_fe.shape_value(j,k);
	      intint(i,j) += dx*nu*ui*vi;
	      intext(i,j) -= dx*nu*ue*vi;
	      extint(i,j) -= dx*nu*ui*ve;
	      extext(i,j) += dx*nu*ue*ve;
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<Point<dim> >& velocity)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
				   // Multiply the quadrature point
				   // index below with this factor to
				   // have simpler data for constant
				   // velocities.
  unsigned int constant_velocity = (velocity.size() == 1) ? 0 : 1;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nbeta =  int_fe.normal_vector(k)
	* velocity[k * constant_velocity];
      
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  if (int_fe.get_fe().is_primitive())
	    {
	      if (nbeta > 0)
		{
		  intint(i,j) += dx*nbeta
				 * int_fe.shape_value(j,k)
				 * int_fe.shape_value(i,k);
		  extint(i,j) -= dx*nbeta
				 * int_fe.shape_value(j,k)
				 * ext_fe.shape_value(i,k);
		}
	      else
		{
		  extext(i,j) -= dx*nbeta
				 * ext_fe.shape_value(j,k)
				 * ext_fe.shape_value(i,k);
		  intext(i,j) += dx*nbeta
				 * ext_fe.shape_value(j,k)
				 * int_fe.shape_value(i,k);
		}
	    }
	  else
	    {
	      for (unsigned int d=0;d<int_fe.get_fe().n_components();++d)
		if (nbeta > 0)
		  {
		    intint(i,j) += dx*nbeta
				   * int_fe.shape_value_component(j,k,d)
				   * int_fe.shape_value_component(i,k,d);
		    extint(i,j) -= dx*nbeta
				   * int_fe.shape_value_component(j,k,d)
				   * ext_fe.shape_value_component(i,k,d);
		  }
		else
		  {
		    extext(i,j) -= dx*nbeta
				   * ext_fe.shape_value_component(j,k,d)
				   * ext_fe.shape_value_component(i,k,d);
		    intext(i,j) += dx*nbeta
				   * ext_fe.shape_value_component(j,k,d)
				   * int_fe.shape_value_component(i,k,d);
		  }	      
	    }
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int n_components = int_fe.get_fe().n_components();
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = factor * int_fe.JxW(k);
      double nv = 0.;
      for (unsigned int d=0;d<dim;++d)
	nv += int_fe.normal_vector(k)(d) * velocity[d][k];
      
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned j=0;j<n_dofs;++j)
	  for (unsigned int d=0;d<n_components;++d)
	    if (nv > 0)
	      {
		intint(i,j) += dx*nv
			       * int_fe.shape_value_component(j,k,d)
			       * int_fe.shape_value_component(i,k,d);
		extint(i,j) -= dx*nv
			       * int_fe.shape_value_component(j,k,d)
			       * ext_fe.shape_value_component(i,k,d);
	      }
	    else
	      {
		extext(i,j) -= dx*nv
			       * ext_fe.shape_value_component(j,k,d)
			       * ext_fe.shape_value_component(i,k,d);
		intext(i,j) += dx*nv
			       * ext_fe.shape_value_component(j,k,d)
			       * int_fe.shape_value_component(i,k,d);
	      }	      
    }
}



template<int dim>
void
NAMEPREFIX::u_times_n (
  std::vector<MatrixBlock<FullMatrix<double> > >& intint,
  std::vector<MatrixBlock<FullMatrix<double> > >& intext,
  std::vector<MatrixBlock<FullMatrix<double> > >& extint,
  std::vector<MatrixBlock<FullMatrix<double> > >& extext,
  const std::vector<unsigned int>& indices,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  double factor)
{
  unsigned int fecomp = int_fe.get_fe().n_components();
  const unsigned int ni_dofs = int_fe.dofs_per_cell;
  const unsigned int ti_dofs = int_fetest.dofs_per_cell;
  const unsigned int ne_dofs = ext_fe.dofs_per_cell;
  const unsigned int te_dofs = ext_fetest.dofs_per_cell;

  Assert (indices.size() == dim/fecomp, ExcDimensionMismatch(indices.size(), dim/fecomp));  
  for (unsigned int i=0;i<indices.size();++i)
    {
      Assert (intint[indices[i]].matrix.m()==ti_dofs,
	      ExcDimensionMismatch(intint[indices[i]].matrix.m(), ti_dofs));
      Assert (intint[indices[i]].matrix.n()==ni_dofs,
	      ExcDimensionMismatch(intint[indices[i]].matrix.n(), ni_dofs));
      Assert (extext[indices[i]].matrix.m()==te_dofs,
	      ExcDimensionMismatch(extext[indices[i]].matrix.m(), te_dofs));
      Assert (extext[indices[i]].matrix.n()==ne_dofs,
	      ExcDimensionMismatch(extext[indices[i]].matrix.n(), ne_dofs));
    }

  const Point<dim>& n = int_fe.normal_vector(0);

  if (int_fe.get_fe().n_components() == 1)
    {
      for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
	{
	  const double dx = .5 * factor * int_fe.JxW(k);
	  
	  for (unsigned i=0;i<ti_dofs;++i)
	    {
	      const double vi = int_fetest.shape_value(i,k);
	      const double ve = ext_fetest.shape_value(i,k);
	      for (unsigned j=0;j<ni_dofs;++j)
		{
		  const double ui = int_fe.shape_value(j,k);
		  const double ue = ext_fe.shape_value(j,k);
		  for (unsigned int d=0;d<dim;++d)
		    {
		      intint[indices[d]].matrix(i,j) += dx*n(d)*ui*vi;
		      extint[indices[d]].matrix(i,j) -= dx*n(d)*ui*ve;
		      intext[indices[d]].matrix(i,j) += dx*n(d)*ue*vi;
		      extext[indices[d]].matrix(i,j) -= dx*n(d)*ue*ve;
		      
		    }	      
		}
	    }
	}
    }
  else
    {
      for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
	{
	  const double dx = .5 * factor * int_fe.JxW(k);
	  
	  for (unsigned i=0;i<ti_dofs;++i)
	    {
	      const double vi = int_fetest.shape_value(i,k);
	      const double ve = ext_fetest.shape_value(i,k);
	      for (unsigned j=0;j<ni_dofs;++j)
		{
		  for (unsigned int d=0;d<dim;++d)
		    {
		      const double ui = int_fe.shape_value_component(j,k,d);
		      const double ue = ext_fe.shape_value_component(j,k,d);
		      intint[indices[0]].matrix(i,j) += dx*n(d)*ui*vi;
		      extint[indices[0]].matrix(i,j) -= dx*n(d)*ui*ve;
		      intext[indices[0]].matrix(i,j) += dx*n(d)*ue*vi;
		      extext[indices[0]].matrix(i,j) -= dx*n(d)*ue*ve;
		      
		    }	      
		}
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::u_times_n_scaled (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<double>& int_factor,
  const std::vector<double>& ext_factor)
{
				   // Still transposed?
  Assert(false, ExcNotImplemented());
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      const double ue = ext_fe.shape_value(j,k);
	      for (unsigned int d=0;d<dim;++d)
		{
		  intint(i,j+d*n_dofs) += dx*.5*n(d)*ui*vi*int_factor[k];
		  extint(i,j+d*n_dofs) -= dx*.5*n(d)*ui*ve*int_factor[k];
		  intext(i,j+d*n_dofs) += dx*.5*n(d)*ue*vi*ext_factor[k];
		  extext(i,j+d*n_dofs) -= dx*.5*n(d)*ue*ve*ext_factor[k];
		}	      
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::div_super (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& ext_fe)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  double C12n = 0.;
  for (unsigned int d=0;d<dim;++d)
    C12n += n(d);
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      const double ue = ext_fe.shape_value(j,k);
	      for (unsigned int d=0;d<dim;++d)
		{
		  if (C12n<0)
		    {
		      extint(i,j+d*n_dofs) += dx*n(d)*ui*ve;
		      extext(i,j+d*n_dofs) -= dx*n(d)*ue*ve;
		    } else {		      			      
		      intint(i,j+d*n_dofs) += dx*n(d)*ui*vi;
		      intext(i,j+d*n_dofs) -= dx*n(d)*ue*vi;
		    }
		}      
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::u_ni (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  unsigned int d)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  Assert (intint.m()==t_dofs, ExcDimensionMismatch(intint.m(), t_dofs));
  Assert (intint.n()==n_dofs, ExcDimensionMismatch(intint.m(), n_dofs));
  Assert (intext.m()==t_dofs, ExcDimensionMismatch(intext.m(), t_dofs));
  Assert (intext.n()==n_dofs, ExcDimensionMismatch(intext.m(), n_dofs));
  Assert (extint.m()==t_dofs, ExcDimensionMismatch(extint.m(), t_dofs));
  Assert (extint.n()==n_dofs, ExcDimensionMismatch(extint.m(), n_dofs));
  Assert (extext.m()==t_dofs, ExcDimensionMismatch(extext.m(), t_dofs));
  Assert (extext.n()==n_dofs, ExcDimensionMismatch(extext.m(), n_dofs));
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = (int_fetest.get_fe().is_primitive())
			    ? int_fetest.shape_value(i,k)
			    : int_fetest.shape_value_component(i,k,d);
	  const double ve = (ext_fetest.get_fe().is_primitive())
			    ? ext_fetest.shape_value(i,k)
			    : ext_fetest.shape_value_component(i,k,d);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = (int_fe.get_fe().is_primitive())
				? int_fe.shape_value(j,k)
				: int_fe.shape_value_component(j,k,d);
	      const double ue = (ext_fe.get_fe().is_primitive())
				? ext_fe.shape_value(j,k)
				: ext_fe.shape_value_component(j,k,d);
	      intint(i,j) += dx*.5*n(d)*ui*vi;
	      extint(i,j) += dx*.5*n(d)*ui*ve;
	      intext(i,j) -= dx*.5*n(d)*ue*vi;
	      extext(i,j) -= dx*.5*n(d)*ue*ve;
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::u_ni_scaled (
  FullMatrix<double>& intint,
  FullMatrix<double>& intext,
  FullMatrix<double>& extint,
  FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  unsigned int d,
  const std::vector<double>& int_factor,
  const std::vector<double>& ext_factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  Assert (intint.m()==t_dofs, ExcDimensionMismatch(intint.m(), t_dofs));
  Assert (intint.n()==n_dofs, ExcDimensionMismatch(intint.n(), n_dofs));
  Assert (intext.m()==t_dofs, ExcDimensionMismatch(intext.m(), t_dofs));
  Assert (intext.n()==n_dofs, ExcDimensionMismatch(intext.n(), n_dofs));
  Assert (extint.m()==t_dofs, ExcDimensionMismatch(extint.m(), t_dofs));
  Assert (extint.n()==n_dofs, ExcDimensionMismatch(extint.n(), n_dofs));
  Assert (extext.m()==t_dofs, ExcDimensionMismatch(extext.m(), t_dofs));
  Assert (extext.n()==n_dofs, ExcDimensionMismatch(extext.n(), n_dofs));
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);
	  for (unsigned j=0;j<n_dofs;++j)
	    {
	      const double ui = int_fe.shape_value(j,k);
	      const double ue = ext_fe.shape_value(j,k);
	      intint(i,j) += dx*.5*n(d)*ui*vi*int_factor[k];
	      extint(i,j) += dx*.5*n(d)*ui*ve*int_factor[k];
	      intext(i,j) -= dx*.5*n(d)*ue*vi*ext_factor[k];
	      extext(i,j) -= dx*.5*n(d)*ue*ve*ext_factor[k];
	    }
	}
    }
}



#define dim deal_II_dimension


template void NAMEPREFIX::advection<dim> (
  FullMatrix<double>& intint, FullMatrix<double>& intext,
  FullMatrix<double>& extint, FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe, const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<Point<dim> >& velocity);
template void NAMEPREFIX::advection<dim> (
  FullMatrix<double>&, FullMatrix<double>&, FullMatrix<double>&, FullMatrix<double>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const VectorSlice<const std::vector<std::vector<double> > >&, double);
template void NAMEPREFIX::ip<dim> (
  FullMatrix<double>& intint, FullMatrix<double>& intext,
  FullMatrix<double>& extint, FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe, const FEFaceValuesBase<dim>& ext_fe,
  double, double, double);
template void NAMEPREFIX::ip_scaled<dim> (
  FullMatrix<double>& intint, FullMatrix<double>& intext,
  FullMatrix<double>& extint, FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe, const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<double>& diffusion, const double penalty);
template void NAMEPREFIX::jump<dim> (
  FullMatrix<double>& intint, FullMatrix<double>& intext,
  FullMatrix<double>& extint, FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>& int_fe, const FEFaceValuesBase<dim>& ext_fe,
  const double factor2);
template void NAMEPREFIX::jump_scaled<dim> (
  FullMatrix<double>&, FullMatrix<double>&,
  FullMatrix<double>&, FullMatrix<double>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const std::vector<double>&, double, bool);

template void NAMEPREFIX::u_times_n<dim> (
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  std::vector<MatrixBlock<FullMatrix<double> > >&,
  const std::vector<unsigned int>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&, double);
template void NAMEPREFIX::u_times_n_scaled<dim> (
  FullMatrix<double>& intint, FullMatrix<double>& intext,
  FullMatrix<double>& extint, FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const std::vector<double>&, const std::vector<double>&);

template void NAMEPREFIX::div_super<dim> (
  FullMatrix<double>& intint, FullMatrix<double>& intext,
  FullMatrix<double>& extint, FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&);
template void NAMEPREFIX::u_ni<dim> (
  FullMatrix<double>& intint,FullMatrix<double>& intext,
  FullMatrix<double>& extint,FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  unsigned int d);
template void NAMEPREFIX::u_ni_scaled<dim> (
  FullMatrix<double>& intint, FullMatrix<double>& intext,
  FullMatrix<double>& extint, FullMatrix<double>& extext,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  unsigned int, const std::vector<double>&, const std::vector<double>&);
