//---------------------------------------------------------------------------
//    $Id: newton.cc 1348 2013-08-16 00:45:07Z secanell $
//    Version: $Name:  $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <appframe/newton.h>
#include <base/parameter_handler.h>
#include <base/data_out_base.h>
#include <lac/block_vector.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <sstream>

using namespace AppFrame;

const Event Newton::bad_derivative = Event::assign("Newton");

Newton::Newton(ApplicationBase& app)
    : ApplicationCopy(app),
      default_residual(app),
      compute_residual(&default_residual),
      n_stepsize_iterations(21),
      assemble_threshold(0.),
      debug_solution(false),
      debug_update(false),
      debug_residual(false),
      debug(0)
{
  deallog << "->Newton";
}

Newton::Newton(ApplicationBase& app, const Residual& res)
    : ApplicationCopy(app),
      default_residual(app),
      compute_residual(&res),
      n_stepsize_iterations(21),
      assemble_threshold(0.),
      debug_solution(false),
      debug_update(false),
      debug_residual(false),
      debug(0)
{
  deallog << "->Newton";
}

double
Newton::threshold (double value)
{
    const double t = assemble_threshold;
    assemble_threshold = value;
    return t;
}

void
Newton::declare_parameters(ParameterHandler& param)
{
    param.enter_subsection("Newton");
    ReductionControl::declare_parameters (param);
    param.declare_entry("Assemble threshold", "0.", Patterns::Double());
    param.declare_entry("Stepsize iterations", "21", Patterns::Integer());
    param.declare_entry("Debug level", "0", Patterns::Integer());
    param.declare_entry("Debug solution", "false", Patterns::Bool());
    param.declare_entry("Debug update", "false", Patterns::Bool());
    param.declare_entry("Debug residual", "false", Patterns::Bool());
    param.leave_subsection();
    
    ApplicationCopy::declare_parameters(param);
}

void
Newton::_initialize (ParameterHandler& param)
{
    param.enter_subsection("Newton");
    control.parse_parameters (param);
    assemble_threshold = param.get_double("Assemble threshold");
    n_stepsize_iterations = param.get_double("Stepsize iterations");
    debug_solution = param.get_bool("Debug solution");
    debug_update   = param.get_bool("Debug update");
    debug_residual = param.get_bool("Debug residual");
    param.leave_subsection ();
}

void
Newton::initialize (ParameterHandler& param)
{
    ApplicationCopy::initialize(param);
    _initialize(param);
}


double
Newton::residual (FEVector& dst, const FEVectors& src)
{
  const double result = (*compute_residual)(dst, src);
  return result;
}

void
Newton::solve (FEVector& u, const FEVectors& in_vectors)
{
  deallog.push ("Newton");
  
  if (debug>2)
    deallog << "u: " << u.l2_norm() << std::endl;
  
  FEVector* Du = get_data()->block_vector_pool.alloc();
  FEVector* res = get_data()->block_vector_pool.alloc();
  
  res->reinit(u);
  FEVectors src1;
  FEVectors src2;
  src1.add_vector(u, "Newton iterate");
  src1.merge(in_vectors);
  src2.add_vector(*res, "Newton residual");
  src2.merge(src1);
  
  unsigned int step = 0;
				   // fill res with (f(u), v)
  double residual = (*compute_residual)(*res, src1);
  double old_residual = residual / assemble_threshold + 1;
  
  if (debug_residual)
    {
      std::ostringstream streamOut;
      streamOut << "Newton_" << std::setw(2) << std::setfill('0') << step;
      std::string name = streamOut.str();
      FEVectors out;
      out.add_vector(u, "solution");
      out.add_vector(u, "xxx");
      out.add_vector(*res, "residual");
      app->data_out(name, out);
    }
  
  while (control.check(step++, residual) == SolverControl::iterate)
    {
				       // assemble (Df(u), v)
	if (residual/old_residual >= assemble_threshold)
	  app->notify (bad_derivative);
	
	Du->reinit(u);
	try 
	  {
	    app->solve (*Du, src2);
	  }
	catch (SolverControl::NoConvergence& e)
	  {
	    deallog << "Inner iteration failed after "
		    << e.last_step << " steps with residual "
		    << e.last_residual << std::endl;
	  }
	
	u.add(-1., *Du);
	old_residual = residual;
	residual = (*compute_residual)(*res, src1);
	
	// Step size control
	unsigned int step_size = 0;
	while (residual >= old_residual)
	{
	    ++step_size;
	    if (step_size > n_stepsize_iterations)
	    {
	      deallog << "No smaller stepsize allowed!";
	      break;
	    }
	    if (control.log_history())
	      deallog << "Trying step size: 1/" << (1<<step_size)
		      << " since residual was " << residual << std::endl;
	    u.add(1./(1<<step_size), *Du);
	    residual = (*compute_residual)(*res, src1);
	}
	
	if (debug_solution || debug_update || debug_residual)
	  {
	    std::ostringstream streamOut;
	    streamOut << "Newton_" << std::setw(2) << std::setfill('0') << step;
	    std::string name = streamOut.str();
	    FEVectors out;
	    out.add_vector(u, "solution");
	    out.add_vector(*Du, "update");
	    out.add_vector(*res, "residual");	    
	    app->data_out(name, out);
	  }
    }
  
  get_data()->block_vector_pool.free(Du);
  get_data()->block_vector_pool.free(res);
  deallog.pop();

				   // in case of failure: throw
				   // exception
  if (control.last_check() != SolverControl::success)
    throw SolverControl::NoConvergence (control.last_step(),
					control.last_value());
				   // otherwise exit as normal

}
