//---------------------------------------------------------------------------
//    $Id: random.cc 1348 2013-08-16 00:45:07Z secanell $
//    Version: $Name:  $
//
//    Copyright (C) 2006, 2007, 2008 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <appframe/random.h>
#include <base/parameter_handler.h>
#include <lac/block_vector.h>
#include <lac/solver_control.h>

using namespace AppFrame;

const Event Random::new_value = Event::assign("Random");

double Random::amplitude () const
{
  return scaled_amplitude * RAND_MAX/2;
}


Random::Random(ApplicationBase& app,
	       bool store_vectors)
		: ApplicationCopy(app),
		  store_vectors(store_vectors)
{
  deallog << "->Random";
}


Random::~Random ()
{
  for (unsigned int i=0;i<results.size();++i)
    {
      FEVector* p = results[i];
      results[i] = 0;
      if (p != 0)
	app->get_data()->block_vector_pool.free(p);
    }
}


void
Random::declare_parameters(ParameterHandler& param)
{
    param.enter_subsection("Random");
    param.declare_entry("Number of samples", "10", Patterns::Integer());
    param.declare_entry("Amplitude", "1.", Patterns::Double());
    param.leave_subsection();
    
    ApplicationCopy::declare_parameters(param);
}

void
Random::_initialize (ParameterHandler& param)
{
    param.enter_subsection("Random");
    const unsigned int n = param.get_integer("Number of samples");
    results.resize(n);
    scaled_amplitude = param.get_double("Amplitude") * 2. / RAND_MAX;
    param.leave_subsection ();
}

void
Random::initialize (ParameterHandler& param)
{
    ApplicationCopy::initialize(param);
    _initialize(param);
}


void
Random::solve (FEVector& mean, const FEVectors& in_vectors)
{
  deallog.push ("Random");
  FEVector* f = get_data()->block_vector_pool.alloc();
  
  unsigned int i=0;
  while (i<results.size())
    {
				       // Only use more vectors if we
				       // store them. Else, always use
				       // first.
      const unsigned int vec_num = (store_vectors) ? i : 0;
      
      if (results[vec_num] == 0)
	results[vec_num] = app->get_data()->block_vector_pool.alloc();
      
      FEVector& v = *results[vec_num];
      v.reinit(mean);
      v = mean;
      
      for (unsigned int j=0;j<v.size();++j)
	v(j) += scaled_amplitude * (rand() - RAND_MAX/2);

      FEVectors src;
      src.add_vector(v, "Random");
      src.merge(in_vectors);
      
      f->reinit(mean);
      app->residual(*f, src);
      app->notify(new_value);
      try 
	{
	  app->solve(v, src);
	  if (!store_vectors)
	    app->evaluate(src);
	  ++i;
	}
      catch (SolverControl::NoConvergence& e)
	{
	  deallog << "Inner iteration failed after "
		  << e.last_step << " steps with residual "
		  << e.last_residual << std::endl;
	}
    }

  get_data()->block_vector_pool.free(f);
  
  deallog.pop ();
}

