//---------------------------------------------------------------------------
//    $Id: residual_shop_bdry.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <appframe/residual_shop_bdry.h>

#include <base/point.h>
#include <base/vector_slice.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#define NAMEPREFIX AppShop::Residual::Boundary

template<int dim>
void
NAMEPREFIX::nitsche(
  Vector<double>& result,
  const std::vector<double>& input_value,
  const std::vector<Tensor<1,dim> >& input_grad,
  const std::vector<double>& data,
  const FEFaceValuesBase<dim>& int_fe,
  double penalty,
  double factor)
{
  Assert(int_fe.get_fe().n_components() == 1,
	 ExcDimensionMismatch(int_fe.get_fe().n_components(), 1));
  
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k) * factor;
      const Point<dim>& n = int_fe.normal_vector(k);
      
      const double ui = input_value[k];
      const Tensor<1,dim>& Dui = input_grad[k];
      const double dnui = Dui * n;
      const double ue = data[k];
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dvi = int_fe.shape_grad(i,k);
	  const double dnvi = Dvi * n;
	  
	  result(i) += dx * (penalty*vi*(ui - ue)
			     - (ui-ue)*dnvi - dnui*vi);
	}
    }
}


template<int dim>
void
NAMEPREFIX::nitsche(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input_value,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input_grad,
  const VectorSlice<const std::vector<std::vector<double> > >& data,
  const FEFaceValuesBase<dim>& int_fe,
  double penalty,
  double factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int n_comp = int_fe.get_fe().n_components();
  
  Assert(input_value.size() == n_comp,
	 ExcDimensionMismatch(input_value.size(), n_comp));
  Assert(input_grad.size() == n_comp,
	 ExcDimensionMismatch(input_grad.size(), n_comp));
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k) * factor;
      const Point<dim>& n = int_fe.normal_vector(k);

      for (unsigned int d=0;d<n_comp;++d)
	{
	  const double ui = input_value[d][k];
	  const Tensor<1,dim>& Dui = input_grad[d][k];
	  const double dnui = Dui * n;
	  const double ue = data[d][k];
	  
	  for (unsigned i=0;i<n_dofs;++i)
	    {
	      const double vi = int_fe.shape_value_component(i,k,d);
	      const Tensor<1,dim>& Dvi = int_fe.shape_grad_component(i,k,d);
	      const double dnvi = Dvi * n;
	      
	      result(i) += dx * (penalty*vi*(ui - ue)
				 - (ui-ue)*dnvi - dnui*vi);
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::nitsche(
  Vector<double>& result,
  const std::vector<double>& input_value,
  const std::vector<Tensor<1,dim> >& input_grad,
  const std::vector<Vector<double> >& data,
  const unsigned int data_index,
  const FEFaceValuesBase<dim>& int_fe,
  double penalty,
  double factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k) * factor;
      const Point<dim>& n = int_fe.normal_vector(k);
      
      const double ui = input_value[k];
      const Tensor<1,dim>& Dui = input_grad[k];
      const double dnui = Dui * n;
      const double ue = data[k](data_index);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dvi = int_fe.shape_grad(i,k);
	  const double dnvi = Dvi * n;
	  
	  result(i) += dx * (penalty*vi*(ui - ue)
			     - (ui-ue)*dnvi - dnui*vi);
	}
    }
}


template<int dim>
void
NAMEPREFIX::nitsche_scaled(
      Vector<double>& result,
      const std::vector<double>& input_value,
      const std::vector<Tensor<1,dim> >& input_grad,
      const std::vector<Vector<double> >& data,
      const unsigned int data_index,
      const FEFaceValuesBase<dim>& int_fe,
      const std::vector<double>& factor,
      double penalty)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const Point<dim>& n = int_fe.normal_vector(k);
      const double nu = factor[k];
      
      const double ui = input_value[k];
      const Tensor<1,dim>& Dui = input_grad[k];
      const double dnui = Dui * n;
      const double ue = data[k](data_index);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dvi = int_fe.shape_grad(i,k);
	  const double dnvi = Dvi * n;
	  
	  result(i) += dx * nu * (penalty*vi*(ui - ue)
				  - (ui-ue)*dnvi
				  - dnui*vi);
	}
    }
}


template<int dim>
void
NAMEPREFIX::nitsche_scaled(
      Vector<double>& result,
      const VectorSlice<const std::vector<std::vector<double> > >& input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input_grad,
      const std::vector<Vector<double> >& data,
      const unsigned int data_index,
      const FEFaceValuesBase<dim>& int_fe,
      const std::vector<double>& factor,
      double penalty)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int fe_components = int_fe.get_fe().n_components();
  const unsigned int n_components = input_value.size();
  Assert (result.size() == n_components*n_dofs/fe_components,
	  ExcDimensionMismatch(result.size(), n_components*n_dofs/fe_components));

  if (int_fe.get_fe().is_primitive())
    for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
      {
	const double dx = int_fe.JxW(k);
	const Point<dim>& n = int_fe.normal_vector(k);
	const double nu = factor[k];
	
	for (unsigned int d=0;d<n_components;++d)
	  {
	    const double ui = input_value[d][k];
	    const Tensor<1,dim>& Dui = input_grad[d][k];
	    const double dnui = Dui * n;
	    const double ue = data[k](data_index+d);
	    
	    for (unsigned i=0;i<n_dofs;++i)
	      {
		const double vi = int_fe.shape_value(i,k);
		const Tensor<1,dim>& Dvi = int_fe.shape_grad(i,k);
		const double dnvi = Dvi * n;
		
		result(i+d*n_dofs) += dx * nu * (penalty*vi*(ui - ue)
						 - (ui-ue)*dnvi
						 - dnui*vi);
	      }
	  }
      }
  else
    for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
      {
	const double dx = int_fe.JxW(k);
	const Point<dim>& n = int_fe.normal_vector(k);
	const double nu = factor[k];
	
	for (unsigned int d=0;d<n_components;++d)
	  {
	    const double ui = input_value[d][k];
	    const Tensor<1,dim>& Dui = input_grad[d][k];
	    const double dnui = Dui * n;
	    const double ue = data[k](data_index+d);
	    
	    for (unsigned i=0;i<n_dofs;++i)
	      {
		const double vi = int_fe.shape_value_component(i,k,d);
		const Tensor<1,dim>& Dvi = int_fe.shape_grad_component(i,k,d);
		const double dnvi = Dvi * n;
		
		result(i) += dx * nu * (penalty*vi*(ui - ue)
					- (ui-ue)*dnvi
					- dnui*vi);
	      }
	  }
      }
}


template<int dim>
void
NAMEPREFIX::jump_scaled (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEFaceValuesBase<dim>& int_fe,
  const std::vector<double>& factor,
  const double factor2,
  bool divide)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2 * ((divide)
				   ? (1./factor[k])
				   : (factor[k]));
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const double ui = input[k];
	  
	  result(i) += dx*nu*ui*vi;
	}
    }
}


template<int dim>
void
NAMEPREFIX::jump_scaled (
  Vector<double>& result,
  const std::vector<double>& input,
  const std::vector<Vector<double> >& data,
  const unsigned int data_index,
  const FEFaceValuesBase<dim>& int_fe,
  const std::vector<double>& factor,
  const double factor2,
  bool divide)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2 * ((divide)
				   ? (1./factor[k])
				   : (factor[k]));
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const double ui = input[k];
	  const double ue = data[k](data_index);
	  
	  result(i) += dx*nu*(ui-ue)*vi;
	}
    }
}


template<int dim>
void
NAMEPREFIX::jump (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEFaceValuesBase<dim>& int_fe,
  const double factor2)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2;
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const double ui = input[k];
	  result(i) += dx*nu*ui*vi;
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& result,
  const std::vector<double>& input,
  const std::vector<Vector<double> >& data,
  unsigned int data_index,
  const FEFaceValuesBase<dim>& fe,
  const std::vector<Point<dim> >& velocity,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      const double nbeta = fe.normal_vector(k) * velocity[k];
      if (nbeta < 0)
	{
	  const double ue = data[k](data_index);
	  for (unsigned i=0;i<n_dofs;++i)
	    {
	      const double vv = fe.shape_value(i,k);
	      const double uu = input[k];
	      result(i) += dx * vv * nbeta * (ue - uu);
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const VectorSlice<const std::vector<std::vector<double> > >& data,
  const FEFaceValuesBase<dim>& fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
  AssertDimension (velocity.size(), dim);
  AssertDimension (input.size(), n_components);
  AssertDimension (data.size(), n_components);  
  AssertDimension(result.size(), n_dofs);
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      const Point<dim>& n = fe.normal_vector(k);
      
      double nv = 0.;
      for (unsigned int d=0;d<dim;++d)
	nv += n(d) * velocity[d][k];
      
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned int c=0;c<n_components;++c)
	  {
	    if (nv < 0)
	      result(i) += dx * nv * fe.shape_value_component(i,k,c) * data[c][k];
	    else
	      result(i) += dx * nv * fe.shape_value_component(i,k,c) * input[c][k];
	  }
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& result,
  const std::vector<double>& input,
  const std::vector<double>& data,
  const FEFaceValuesBase<dim>& fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  AssertDimension (velocity.size(), dim);
  AssertDimension(result.size(), n_dofs);
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      const Point<dim>& n = fe.normal_vector(k);
      
      double nv = 0.;
      for (unsigned int d=0;d<dim;++d)
	nv += n(d) * velocity[d][k];
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  if (nv < 0)
	    result(i) += dx * nv * fe.shape_value(i,k) * data[k];
	  else
	    result(i) += dx * nv * fe.shape_value(i,k) * input[k];
	}
    }
}


template<int dim>
void
NAMEPREFIX::u_times_n (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest,
  double factor)
{
  Assert(result.size() == dim*fetest.dofs_per_cell/fetest.get_fe().n_components(),
	 ExcDimensionMismatch(result.size(),dim*fetest.dofs_per_cell/fetest.get_fe().n_components()));
  Assert(input.size() == fe.n_quadrature_points,
	 ExcDimensionMismatch(input.size(), fe.n_quadrature_points));
  
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const Tensor<1,dim> ndx = factor * fe.normal_vector(k) * fe.JxW(k);
      for (unsigned i=0;i<t_dofs;++i)
	if (fetest.get_fe().n_components() == 1)
	  for (unsigned int d=0;d<dim;++d)
	    result(i+d*t_dofs) += ndx[d] * fetest.shape_value(i,k)
				  * input[k];
	else
	  for (unsigned int d=0;d<dim;++d)
	    result(i) += ndx[d] * fetest.shape_value_component(i,k,d)
			 * input[k];
    }
}


template<int dim>
void
NAMEPREFIX::u_times_n_scaled (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const Tensor<1,dim> ndx = ((divide) ? (1./factor[k]) : (factor[k]))
				* fe.normal_vector(k) * fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	if (fetest.get_fe().n_components() == 1)
	  for (unsigned int d=0;d<dim;++d)
	    result(i+d*t_dofs) += ndx[d] * fetest.shape_value(i,k)
				  * input[k];
	else
	  for (unsigned int d=0;d<dim;++d)
	    result(i) += ndx[d] * fetest.shape_value_component(i,k,d)
			 * input[k];
    }
}


template<int dim>
void
NAMEPREFIX::u_dot_n (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest,
  double factor)
{
  Assert (input.size() == dim, ExcDimensionMismatch(input.size(), dim));
  Assert (fetest.get_fe().n_components() == 1,
	  ExcDimensionMismatch(fetest.get_fe().n_components(), 1));
  
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const Tensor<1,dim> ndx = factor * fe.normal_vector(k) * fe.JxW(k);      

      for (unsigned i=0;i<t_dofs;++i)
	for (unsigned int d=0;d<dim;++d)
	  result(i) += ndx[d] * fetest.shape_value(i,k) * input[d][k];
    }
}


template<int dim>
void
NAMEPREFIX::u_dot_n_scaled (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const Tensor<1,dim> ndx = ((divide) ? (1./factor[k]) : (factor[k]))
				* fe.normal_vector(k) * fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	  for (unsigned int d=0;d<dim;++d)
	    result(i) += ndx[d] * fetest.shape_value(i,k) * input[d][k];
    }
}


template<int dim>
void
NAMEPREFIX::eps (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  Assert (input.size() == dim, ExcDimensionMismatch(input.size(), dim));
  Assert (result.size()==symm_dim(dim)*t_dofs,
	  ExcDimensionMismatch(result.size(), symm_dim(dim)*t_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      const Point<dim>& n = fe.normal_vector(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = fetest.shape_value(i,k);
	  for (unsigned int d=0;d<dim;++d)
	    {
	      const double ui = input[d][k];
	      result(i+d*t_dofs) -= dx * vi * ui * n(d);
	      
//	      for (unsigned int d2=d+1;d2<dim;++d2)
//		{
//		  const double ui2 = input[d2][k];
		  Assert (dim==2, ExcNotImplemented());
		  const unsigned int d3 = 2;
		  result(i+d3*t_dofs) -= dx * vi * ui * n(1-d);
//		}
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::eps_div (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEFaceValuesBase<dim>& fe,
  const FEFaceValuesBase<dim>& fetest)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  Assert (input.size() == symm_dim(dim),
	  ExcDimensionMismatch(input.size(), symm_dim(dim)));
  Assert (result.size()==dim*t_dofs,
	  ExcDimensionMismatch(result.size(), dim*t_dofs));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      const Point<dim>& n = fe.normal_vector(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = fetest.shape_value(i,k);
	  for (unsigned int d1=0;d1<dim;++d1)
	    {
	      const double ui1 = input[d1][k];
	      result(i+d1*t_dofs) -= dx * vi * ui1 * n(d1);
	      
	      for (unsigned int d2=d1+1;d2<dim;++d2)
		{
		  Assert (dim==2, ExcNotImplemented());
		  const unsigned int d3 = dim+d2-1;
		  const double ui3 = input[d3][k];
		  
		  result(i+d1*t_dofs) -= dx * vi * ui3 * n(d2);
		  result(i+d2*t_dofs) -= dx * vi * ui3 * n(d1);
		}
	    }
	}
    }
}

#define dim deal_II_dimension

template void NAMEPREFIX::nitsche<dim> (
  Vector<double>&, const std::vector<double>&, const std::vector<Tensor<1,dim> >&,
  const std::vector<double>&, const FEFaceValuesBase<dim>&, double, double);
template void NAMEPREFIX::nitsche<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >&,
  const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEFaceValuesBase<dim>&, double, double);
template void NAMEPREFIX::nitsche<dim> (
  Vector<double>&, const std::vector<double>&, const std::vector<Tensor<1,dim> >&,
  const std::vector<Vector<double> >&, unsigned int,
  const FEFaceValuesBase<dim>&, double, double);
template void NAMEPREFIX::nitsche_scaled<dim> (
  Vector<double>&, const std::vector<double>&, const std::vector<Tensor<1,dim> >&,
  const std::vector<Vector<double> >&, unsigned int, const FEFaceValuesBase<dim>&,
  const std::vector<double>&, double);
template void NAMEPREFIX::nitsche_scaled<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >&,
  const std::vector<Vector<double> >&, unsigned int,
  const FEFaceValuesBase<dim>&, const std::vector<double>&, double);
template void NAMEPREFIX::jump_scaled<dim> (
  Vector<double>&, const std::vector<double>&, const FEFaceValuesBase<dim>&,
  const std::vector<double>&, double, bool);
template void NAMEPREFIX::jump_scaled<dim> (
  Vector<double>&, const std::vector<double>&, const std::vector<Vector<double> >&,
  unsigned int, const FEFaceValuesBase<dim>&, const std::vector<double>&, double, bool);
template void NAMEPREFIX::jump<dim> (
  Vector<double>&, const std::vector<double>&, const FEFaceValuesBase<dim>&, double);
template void NAMEPREFIX::advection<dim> (
  Vector<double>&, const std::vector<double>&, const std::vector<Vector<double> >&,
  unsigned int, const FEFaceValuesBase<dim>& fe, const std::vector<Point<dim> >&, double);
template void NAMEPREFIX::advection<dim> (
  Vector<double>&, const std::vector<double>&, const std::vector<double>&,
  const FEFaceValuesBase<dim>&,
  const VectorSlice<const std::vector<std::vector<double> > >&, double factor);
template void NAMEPREFIX::advection<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const VectorSlice<const std::vector<std::vector<double> > >&, const FEFaceValuesBase<dim>&,
  const VectorSlice<const std::vector<std::vector<double> > >&, double factor);
template void NAMEPREFIX::u_times_n<dim> (
  Vector<double>&, const std::vector<double>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, double);
template void NAMEPREFIX::u_dot_n<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&, double);
template void NAMEPREFIX::u_times_n_scaled<dim> (
  Vector<double>&, const std::vector<double>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const std::vector<double>&, bool);
template void NAMEPREFIX::u_dot_n_scaled<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const std::vector<double>&, bool);
template void NAMEPREFIX::eps<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&);
template void NAMEPREFIX::eps_div<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&);
