//---------------------------------------------------------------------------
//    $Id: residual_shop_cell.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <appframe/residual_shop_cell.h>

#include <base/point.h>
#include <base/vector_slice.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#define NAMEPREFIX AppShop::Residual::Cell

template<int dim>
void
NAMEPREFIX::mass(
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) * factor;
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vv = fe.shape_value(i,k);

	  result(i) += dx * input[k] * vv;
	}
    }
}


template<int dim>
void
NAMEPREFIX::mass(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int fe_components = fe.get_fe().n_components();
  const unsigned int n_components = input.size();
  
  Assert (result.size() == n_components*n_dofs/fe_components,
	  ExcDimensionMismatch(result.size(), n_components*n_dofs/fe_components));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) * factor;
      
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned int d=0;d<n_components;++d)
	  result(i) += dx * fe.shape_value_component(i,k,d) * input[d][k];
    }
}


template<int dim>
void
NAMEPREFIX::mass (
  Vector<double>& result,
  const std::vector<Vector<double> >& input,
  const unsigned int start_component,
  const FEValuesBase<dim>& fe,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = fe.get_fe().n_components();
  
  Assert (result.size() == n_dofs,
	  ExcDimensionMismatch(result.size(), n_dofs));
  Assert (input.size() == fe.n_quadrature_points,
	  ExcDimensionMismatch(input.size(), fe.n_quadrature_points));
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    Assert (input[k].size() >= start_component+n_components,
	    ExcDimensionMismatch(input[k].size(),
				 start_component+n_components));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) * factor;
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned int d=0;d<n_components;++d)
	  result(i) += dx * fe.shape_value_component(i,k,d)
	    * input[k](d+start_component);
    }
}


template<int dim>
void
NAMEPREFIX::power(
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  double p,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) * factor;
      for (unsigned i=0;i<n_dofs;++i)
	result(i) += dx * std::pow(input[k], p) * fe.shape_value(i,k);
    }
}


template<int dim>
void
NAMEPREFIX::time_step(
  Vector<double>& result,
  const std::vector<double>& input,
  const std::vector<double>& previous,
  const FEValuesBase<dim>& fe,
  double time_step)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) / time_step;
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vv = fe.shape_value(i,k);

	  result(i) += dx * (input[k] - previous[k]) * vv;
	}
    }
}


template<int dim>
void
NAMEPREFIX::time_step(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const VectorSlice<const std::vector<std::vector<double> > >& previous,
  const FEValuesBase<dim>& fe,
  double time_step)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int fe_components = fe.get_fe().n_components();
  const unsigned int n_components = input.size();
  
  Assert (result.size() == n_components*n_dofs/fe_components,
	  ExcDimensionMismatch(result.size(), n_components*n_dofs/fe_components));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) / time_step;
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned int d=0;d<n_components;++d)
	    if (fe.get_fe().is_primitive())
	      result(i+d*n_dofs) += dx * fe.shape_value(i,k)
				    * (input[d][k] - previous[d][k]);
	    else
	      result(i) += dx * fe.shape_value_component(i,k,d)
			   * (input[d][k] - previous[d][k]);
	}
    }
}


template<int dim>
void
NAMEPREFIX::mass_scaled(
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      const double nu = (divide)
			? (1./factor[k])
			: (factor[k]);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vv = fe.shape_value(i,k);

	  result(i) += dx * nu * input[k] * vv;
	}
    }
}


template<int dim>
void
NAMEPREFIX::mass_scaled(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int fe_components = fe.get_fe().n_components();
  const unsigned int n_components = input.size();
  
  Assert (result.size() == n_components*n_dofs/fe_components,
	  ExcDimensionMismatch(result.size(), n_components*n_dofs/fe_components));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k) ;
      const double nu = (divide)
	                ? (1./factor[k])
			: (factor[k]);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned int d=0;d<n_components;++d)
	    if (fe.get_fe().is_primitive())
	      result(i+d*n_dofs) += dx * nu * fe.shape_value(i,k) * input[d][k];
	    else
	      result(i) += dx * nu * fe.shape_value_component(i,k,d) * input[d][k];
	}
    }
}



template<int dim>
void
NAMEPREFIX::laplacian_scaled (
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      const double nu = (divide)
			? (1./factor[k])
			: (factor[k]);
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const Tensor<1,dim>& Dv = fe.shape_grad(i,k);	  
	  result(i) += dx * nu * (input[k] * Dv);
	}
    }
}

template<int dim>
void
NAMEPREFIX::laplacian_scaled (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int fe_components = fe.get_fe().n_components();
  const unsigned int n_components = input.size();
  
  Assert (result.size() == n_components*n_dofs/fe_components,
	  ExcDimensionMismatch(result.size(), n_components*n_dofs/fe_components));
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {	  
      const double dx = fe.JxW(k);
      const double nu = (divide)
			? (1./factor[k])
			: (factor[k]);
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned int d=0;d<n_components;++d)
	  if (fe.get_fe().is_primitive())
	    result(i+d*n_dofs) += dx * nu * (input[d][k] * fe.shape_grad(i,k));
	    
	  else
	    result(i) += dx * nu * (input[d][k] * fe.shape_grad_component(i,k,d));
    }
}




template<int dim>
void
NAMEPREFIX::laplacian (
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe,
  double factor)
{
  const unsigned int nq = fe.n_quadrature_points;
  const unsigned int n_dofs = fe.dofs_per_cell;
  Assert(input.size() == nq, ExcDimensionMismatch(input.size(), nq));
  Assert(result.size() == n_dofs, ExcDimensionMismatch(result.size(), n_dofs));
  
  for (unsigned k=0;k<nq;++k)
    {
      const double dx = factor * fe.JxW(k);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const Tensor<1,dim>& Dv = fe.shape_grad(i,k);	  
	  result(i) += dx * (input[k] * Dv);
	}
    }
}

template<int dim>
void
NAMEPREFIX::laplacian (
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe,
  const Tensor<2,dim> factor)
{
  const unsigned int nq = fe.n_quadrature_points;
  const unsigned int n_dofs = fe.dofs_per_cell;
  Assert(input.size() == nq, ExcDimensionMismatch(input.size(), nq));
  Assert(result.size() == n_dofs, ExcDimensionMismatch(result.size(), n_dofs));
  
  for (unsigned k=0;k<nq;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const Tensor<1,dim>& Dv = fe.shape_grad(i,k);	  
	  result(i) += dx * (input[k] * factor * Dv);
	}
    }
}


template<int dim>
void
NAMEPREFIX::laplacian (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
  const FEValuesBase<dim>& fe,
  double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int fe_components = fe.get_fe().n_components();
  const unsigned int n_components = input.size();
  
  Assert (result.size() == n_components*n_dofs/fe_components,
	  ExcDimensionMismatch(result.size(), n_components*n_dofs/fe_components));
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {	  
      const double dx = factor * fe.JxW(k);
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned int d=0;d<n_components;++d)
	  result(i) += dx * (input[d][k] * fe.shape_grad_component(i,k,d));
    }
}




template<int dim>
void
NAMEPREFIX::div(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
  const FEValuesBase<dim>&,
  const FEValuesBase<dim>& fetest,
  const double factor)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  Assert (result.size() == t_dofs, ExcDimensionMismatch(result.size(), t_dofs));
  Assert(input.size() == dim, ExcDimensionMismatch(input.size(), dim));
  for (unsigned int d=0;d<dim;++d)
    Assert(input[d].size() == fetest.n_quadrature_points,
	   ExcDimensionMismatch(input[d].size(), fetest.n_quadrature_points));
  
  for (unsigned k=0;k<fetest.n_quadrature_points;++k)
    {
      const double dx = factor * fetest.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	for (unsigned int d=0;d<dim;++d)
	  result(i) += dx * input[d][k][d] * fetest.shape_value(i,k);
    }
}


template<int dim>
void
NAMEPREFIX::div(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>&,
  const FEValuesBase<dim>& fetest,
  const double factor)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  const unsigned int n_components = input.size() / dim;

  Assert(input.size() == n_components*dim, ExcNotMultiple(input.size(), dim));

  if (fetest.get_fe().n_components() == 1)
    for (unsigned k=0;k<fetest.n_quadrature_points;++k)
      {
	const double dx = factor * fetest.JxW(k);
	
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    const Tensor<1,dim>& vv = fetest.shape_grad(i,k);
	    for (unsigned int d1=0;d1<n_components;++d1)
	      for (unsigned int d2=0;d2<dim;++d2)
		result(i+d1*t_dofs) -= dx * input[d2+dim*d1][k] * vv[d2];
	  }
      }
  else
    for (unsigned k=0;k<fetest.n_quadrature_points;++k)
      {
	const double dx = factor * fetest.JxW(k);
	
	for (unsigned i=0;i<t_dofs;++i)
	  {
	    for (unsigned int d1=0;d1<dim;++d1)
	      for (unsigned int d2=0;d2<dim;++d2)
		result(i+d1*t_dofs) -= dx *fetest.shape_grad_component(i,k,d2)[d2]
				       * input[d2+dim*d1][k];
	  }
      }
}


template<int dim>
void
NAMEPREFIX::div_scaled(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  const std::vector<double>& factor)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const Tensor<1,dim>& Dv = fetest.shape_grad(i,k);
	  for (unsigned int d=0;d<dim;++d)
	    result(i) -= dx * factor[k]*input[d][k] * Dv[d];
	}
    }
}



template<int dim>
void
NAMEPREFIX::grad(
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  double factor)
{
  Assert (fetest.get_fe().n_components() == dim,
	  ExcDimensionMismatch(fetest.get_fe().n_components(), dim));
  Assert(result.size() == fetest.dofs_per_cell,
	 ExcDimensionMismatch(result.size(),fetest.dofs_per_cell));
  Assert(input.size() == fe.n_quadrature_points,
	 ExcDimensionMismatch(input.size(), fe.n_quadrature_points));
  
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	for (unsigned int d=0;d<dim;++d)
	  result(i) += dx * input[k][d] * fetest.shape_value_component(i,k,d);
    }
}


template<int dim>
void
NAMEPREFIX::di(
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  unsigned int direction,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  const double factor)
{
  Assert(direction<dim,ExcIndexRange(direction,0,dim));
  Assert(result.size() == fetest.dofs_per_cell,
	 ExcDimensionMismatch(result.size(),fetest.dofs_per_cell));
  Assert(input.size() == fe.n_quadrature_points,
	 ExcDimensionMismatch(input.size(), fe.n_quadrature_points));
  
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    for (unsigned i=0;i<t_dofs;++i)
      result(i) += fe.JxW(k) * factor * input[k][direction] * fetest.shape_value(i,k);
}


template<int dim>
void
NAMEPREFIX::di(
  Vector<double>& result,
  const std::vector<double>& input,
  unsigned int direction,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  const double factor)
{
  Assert(direction<dim,ExcIndexRange(direction,0,dim));
  Assert(result.size() == fetest.dofs_per_cell,
	 ExcDimensionMismatch(result.size(),fetest.dofs_per_cell));
  Assert(input.size() == fe.n_quadrature_points,
	 ExcDimensionMismatch(input.size(), fe.n_quadrature_points));
  
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    for (unsigned i=0;i<t_dofs;++i)
      result(i) -= fe.JxW(k) * factor * input[k] * fetest.shape_grad(i,k)[direction];
}


template<int dim>
void
NAMEPREFIX::grad(
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  double factor)
{
  Assert (fetest.get_fe().n_components() == dim,
	  ExcDimensionMismatch(fetest.get_fe().n_components(), dim));
  
  const unsigned int t_dofs = fetest.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	for (unsigned int d=0;d<dim;++d)
	  result(i) -= dx * input[k] * fetest.shape_grad_component(i,k,d)[d];
    }
}


template<int dim>
void
NAMEPREFIX::eps_div(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  Assert (input.size() == symm_dim(dim),
	  ExcDimensionMismatch(input.size(), symm_dim(dim)));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const Tensor<1,dim>& Dv = fetest.shape_grad(i,k);
	  for (unsigned int d1=0;d1<dim;++d1)
	    {
	      const double u1 = input[d1][k];
	      result(i+d1*t_dofs) += dx * Dv[d1] * u1;
	      for (unsigned int d2=d1+1;d2<dim;++d2)
		{
		  Assert (dim==2, ExcNotImplemented());
		  const unsigned int d3 = dim+d2-1;
		  const double u3 = input[d3][k];
		  result(i+d1*t_dofs) += dx * Dv[d2] * u3;
		  result(i+d2*t_dofs) += dx * Dv[d1] * u3;
		}
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::eps(
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest)
{
  const unsigned int t_dofs = fetest.dofs_per_cell;
  Assert (input.size() == dim, ExcDimensionMismatch(input.size(), dim));
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const Tensor<1,dim>& Dv = fetest.shape_grad(i,k);
	  for (unsigned int d1=0;d1<dim;++d1)
	    {
	      const double u1 = input[d1][k];
	      result(i+d1*t_dofs) += dx * Dv[d1] * u1;
//	      for (unsigned int d2=d1+1;d2<dim;++d2)
//		{
//		  const double u2 = input[d2][k];
	      Assert (dim==2, ExcNotImplemented());
	      const unsigned int d3 = 2;//dim+d2-1;
	      result(i+d3*t_dofs) += dx * u1 * Dv[1-d1];
//		}
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  const std::vector<Point<dim> >& velocity,
  const double factor)
{
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor * fe.JxW(k);
      const Point<dim>& ww = velocity[k];
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const Tensor<1,dim>& Dv = fe.shape_grad(i,k);
	  result(i) -= dx * input[k] * (ww * Dv);
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  AssertDimension (input.size(), fe.n_quadrature_points);
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor*fe.JxW(k);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned int d=0;d<dim;++d)
	    result(i) -= dx * input[k]
			 * (fe.shape_grad(i,k)[d] * velocity[d][k]);
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  AssertDimension (input.size(), fe.get_fe().n_components());
  const unsigned int n_dofs = fe.dofs_per_cell;
  const unsigned int n_components = input.size();
  
  for (unsigned k=0;k<fe.n_quadrature_points;++k)
    {
      const double dx = factor*fe.JxW(k);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned int c=0;c<n_components;++c)
	    for (unsigned int d=0;d<dim;++d)
	      result(i) -= dx * input[c][k]
			   * (fe.shape_grad_component(i,k,c)[d] * velocity[d][k]);
	}
    }
}

#define dim deal_II_dimension

template void NAMEPREFIX::mass<dim> (
  Vector<double>&, const std::vector<double>&,
  const FEValuesBase<dim>&, double);

template void NAMEPREFIX::mass<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >& ,
  const FEValuesBase<dim>&, double);

template void NAMEPREFIX::mass<dim> (
  Vector<double>&, const std::vector<Vector<double> >&,
  const unsigned int, const FEValuesBase<dim>&, double);

template void NAMEPREFIX::power<dim> (
  Vector<double>&, const std::vector<double>&,
  const FEValuesBase<dim>&, double, double);

template void NAMEPREFIX::time_step<dim> (
  Vector<double>& result,
  const std::vector<double>& input,
  const std::vector<double>& previous,
  const FEValuesBase<dim>& fe,
  double time_step);

template void NAMEPREFIX::time_step<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const VectorSlice<const std::vector<std::vector<double> > >& previous,
  const FEValuesBase<dim>& fe,
  double time_step);

template void NAMEPREFIX::mass_scaled<dim> (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide);

template void NAMEPREFIX::mass_scaled<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide);

template void NAMEPREFIX::laplacian_scaled<dim> (
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide);

template void NAMEPREFIX::laplacian_scaled<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
  const FEValuesBase<dim>& fe,
  const std::vector<double>& factor,
  bool divide);

template void NAMEPREFIX::laplacian<dim> (
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe, double);

template void NAMEPREFIX::laplacian<dim> (
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe, 
  const Tensor<2,dim>);

template void NAMEPREFIX::laplacian<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& input,
  const FEValuesBase<dim>& fe, double);

template void NAMEPREFIX::di(
  Vector<double>&, const std::vector<Tensor<1,dim> >&, unsigned int,
  const FEValuesBase<dim>&, const FEValuesBase<dim>&, double);

template void NAMEPREFIX::di(
  Vector<double>&, const std::vector<double>&, unsigned int,
  const FEValuesBase<dim>&, const FEValuesBase<dim>&, double);
template void NAMEPREFIX::div<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >&,
  const FEValuesBase<dim>&, const FEValuesBase<dim>&, double);

template void NAMEPREFIX::div<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEValuesBase<dim>&, const FEValuesBase<dim>&, double);

template void NAMEPREFIX::div_scaled<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest,
  const std::vector<double>& factor);

template void NAMEPREFIX::grad<dim> (
  Vector<double>& result,
  const std::vector<Tensor<1,dim> >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest, double);

template void NAMEPREFIX::grad<dim> (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest, double);

template void NAMEPREFIX::eps_div<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest);

template void NAMEPREFIX::eps<dim> (
  Vector<double>& result,
  const VectorSlice<const std::vector<std::vector<double> > >& input,
  const FEValuesBase<dim>& fe,
  const FEValuesBase<dim>& fetest);

template void NAMEPREFIX::advection<dim> (
  Vector<double>& result,
  const std::vector<double>& input,
  const FEValuesBase<dim>& fe,
  const std::vector<Point<dim> >& velocity, double);

template void NAMEPREFIX::advection<dim> (
  Vector<double>&, const std::vector<double>&,
  const FEValuesBase<dim>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  double);

template void NAMEPREFIX::advection<dim> (
  Vector<double>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEValuesBase<dim>&, const VectorSlice<const std::vector<std::vector<double> > >&,
  double);

