//---------------------------------------------------------------------------
//    $Id: residual_shop_face.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------
 
#include <appframe/residual_shop_face.h>

#include <base/point.h>
#include <base/vector_slice.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#define NAMEPREFIX AppShop::Residual::Face

template<int dim>
void
NAMEPREFIX::ip(
      Vector<double>& int_result,
      Vector<double>& ext_result,
      const std::vector<double>& int_input_value,
      const std::vector<Tensor<1,dim> >& int_input_grad,
      const std::vector<double>& ext_input_value,
      const std::vector<Tensor<1,dim> >& ext_input_grad,
      const FEFaceValuesBase<dim>& int_fe,
      const FEFaceValuesBase<dim>& ext_fe,
      double pen,
      double int_factor,
      double ext_factor)
{
  Assert(int_fe.get_fe().n_components() == 1,
	 ExcDimensionMismatch(int_fe.get_fe().n_components(), 1));
  Assert(ext_fe.get_fe().n_components() == 1,
	 ExcDimensionMismatch(ext_fe.get_fe().n_components(), 1));

  const double nui = int_factor;
  const double nue = (ext_factor < 0) ? int_factor : ext_factor;
  const double penalty = .5 * pen * (nui + nue);
  
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const Point<dim>& n = int_fe.normal_vector(k);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dvi = int_fe.shape_grad(i,k);
	  const double dnvi = Dvi * n;
	  const double ve = ext_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dve = ext_fe.shape_grad(i,k);
	  const double dnve = Dve * n;

	  const double ui = int_input_value[k];
	  const Tensor<1,dim>& Dui = int_input_grad[k];
	  const double dnui = Dui * n;
	  const double ue = ext_input_value[k];
	  const Tensor<1,dim>& Due = ext_input_grad[k];
	  const double dnue = Due * n;

	  int_result(i) += dx*(-.5*nui*dnvi*ui-.5*nui*dnui*vi+penalty*ui*vi);
	  int_result(i) += dx*( .5*nui*dnvi*ue-.5*nue*dnue*vi-penalty*vi*ue);
	  ext_result(i) += dx*(-.5*nue*dnve*ui+.5*nui*dnui*ve-penalty*ui*ve);
	  ext_result(i) += dx*( .5*nue*dnve*ue+.5*nue*dnue*ve+penalty*ue*ve);
	}
    }
}


template<int dim>
void
NAMEPREFIX::ip_scaled(
      Vector<double>& int_result,
      Vector<double>& ext_result,
      const std::vector<double>& int_input_value,
      const std::vector<Tensor<1,dim> >& int_input_grad,
      const std::vector<double>& ext_input_value,
      const std::vector<Tensor<1,dim> >& ext_input_grad,
      const FEFaceValuesBase<dim>& int_fe,
      const FEFaceValuesBase<dim>& ext_fe,
      const std::vector<double>& factor,
      double penalty)
{
  Assert(int_fe.get_fe().n_components() == 1,
	 ExcDimensionMismatch(int_fe.get_fe().n_components(), 1));
  Assert(ext_fe.get_fe().n_components() == 1,
	 ExcDimensionMismatch(ext_fe.get_fe().n_components(), 1));
  
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const Point<dim>& n = int_fe.normal_vector(k);
      const double nu = factor[k];
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dvi = int_fe.shape_grad(i,k);
	  const double dnvi = Dvi * n;
	  const double ve = ext_fe.shape_value(i,k);
	  const Tensor<1,dim>& Dve = ext_fe.shape_grad(i,k);
	  const double dnve = Dve * n;

	  const double ui = int_input_value[k];
	  const Tensor<1,dim>& Dui = int_input_grad[k];
	  const double dnui = Dui * n;
	  const double ue = ext_input_value[k];
	  const Tensor<1,dim>& Due = ext_input_grad[k];
	  const double dnue = Due * n;

	  int_result(i) += dx*nu*(-.5*dnvi*ui-.5*dnui*vi+penalty*ui*vi);
	  int_result(i) += dx*nu*( .5*dnvi*ue-.5*dnue*vi-penalty*vi*ue);
	  ext_result(i) += dx*nu*(-.5*dnve*ui+.5*dnui*ve-penalty*ui*ve);
	  ext_result(i) += dx*nu*( .5*dnve*ue+.5*dnue*ve+penalty*ue*ve);
	}
    }
}


template<int dim>
void
NAMEPREFIX::ip(
      Vector<double>& int_result,
      Vector<double>& ext_result,
      const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& int_input_grad,
      const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& ext_input_grad,
      const FEFaceValuesBase<dim>& int_fe,
      const FEFaceValuesBase<dim>& ext_fe,
      double pen,
      double int_factor,
      double ext_factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int fe_comp = int_fe.get_fe().n_components();
  const unsigned int mult = int_input_value.size() / fe_comp;
  
  const double nui = int_factor;
  const double nue = (ext_factor < 0) ? int_factor : ext_factor;
  const double penalty = .5 * pen * (nui + nue);
  
  Assert(int_input_value.size() == fe_comp,
	 ExcDimensionMismatch(int_input_value.size(), fe_comp));
  Assert(int_input_grad.size() == fe_comp,
	 ExcDimensionMismatch(int_input_grad.size(), fe_comp));
  Assert(ext_input_value.size() == fe_comp,
	 ExcDimensionMismatch(ext_input_value.size(), fe_comp));
  Assert(ext_input_grad.size() == fe_comp,
	 ExcDimensionMismatch(ext_input_grad.size(), fe_comp));
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const Point<dim>& n = int_fe.normal_vector(k);
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned int m=0;m<mult;++m)
	    for (unsigned int d=0;d<fe_comp;++d)
	      {
		const double vi = int_fe.shape_value_component(i,k,d);
		const Tensor<1,dim>& Dvi = int_fe.shape_grad_component(i,k,d);
		const double dnvi = Dvi * n;
		const double ve = ext_fe.shape_value_component(i,k,d);
		const Tensor<1,dim>& Dve = ext_fe.shape_grad_component(i,k,d);
		const double dnve = Dve * n;
		
		const double ui = int_input_value[d+fe_comp*m][k];
		const Tensor<1,dim>& Dui = int_input_grad[d+fe_comp*m][k];
		const double dnui = Dui * n;
		const double ue = ext_input_value[d+fe_comp*m][k];
		const Tensor<1,dim>& Due = ext_input_grad[d+fe_comp*m][k];
		const double dnue = Due * n;
		
		int_result(i+m*n_dofs) += dx*(-.5*nui*dnvi*ui-.5*nui*dnui*vi+penalty*ui*vi);
		int_result(i+m*n_dofs) += dx*( .5*nui*dnvi*ue-.5*nue*dnue*vi-penalty*vi*ue);
		ext_result(i+m*n_dofs) += dx*(-.5*nue*dnve*ui+.5*nui*dnui*ve-penalty*ui*ve);
		ext_result(i+m*n_dofs) += dx*( .5*nue*dnve*ue+.5*nue*dnue*ve+penalty*ue*ve);
	      }
	}
    }
}


template<int dim>
void
NAMEPREFIX::ip_scaled(
      Vector<double>& int_result,
      Vector<double>& ext_result,
      const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& int_input_grad,
      const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& ext_input_grad,
      const FEFaceValuesBase<dim>& int_fe,
      const FEFaceValuesBase<dim>& ext_fe,
      const std::vector<double>& factor,
      double penalty)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int fe_comp = int_fe.get_fe().n_components();
  const unsigned int mult = int_input_value.size() / fe_comp;
  
  Assert(int_input_value.size() == fe_comp,
	 ExcDimensionMismatch(int_input_value.size(), fe_comp));
  Assert(int_input_grad.size() == fe_comp,
	 ExcDimensionMismatch(int_input_grad.size(), fe_comp));
  Assert(ext_input_value.size() == fe_comp,
	 ExcDimensionMismatch(ext_input_value.size(), fe_comp));
  Assert(ext_input_grad.size() == fe_comp,
	 ExcDimensionMismatch(ext_input_grad.size(), fe_comp));
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const Point<dim>& n = int_fe.normal_vector(k);
      const double nu = factor[k];
      
      for (unsigned i=0;i<n_dofs;++i)
	{
	  for (unsigned int m=0;m<mult;++m)
	    for (unsigned int d=0;d<fe_comp;++d)
	      {
		const double vi = int_fe.shape_value_component(i,k,d);
		const Tensor<1,dim>& Dvi = int_fe.shape_grad_component(i,k,d);
		const double dnvi = Dvi * n;
		const double ve = ext_fe.shape_value_component(i,k,d);
		const Tensor<1,dim>& Dve = ext_fe.shape_grad_component(i,k,d);
		const double dnve = Dve * n;
		
		const double ui = int_input_value[d+fe_comp*m][k];
		const Tensor<1,dim>& Dui = int_input_grad[d+fe_comp*m][k];
		const double dnui = Dui * n;
		const double ue = ext_input_value[d+fe_comp*m][k];
		const Tensor<1,dim>& Due = ext_input_grad[d+fe_comp*m][k];
		const double dnue = Due * n;
		
		int_result(i+m*n_dofs) += dx*nu*(-.5*dnvi*ui-.5*dnui*vi+penalty*ui*vi);
		int_result(i+m*n_dofs) += dx*nu*( .5*dnvi*ue-.5*dnue*vi-penalty*vi*ue);
		ext_result(i+m*n_dofs) += dx*nu*(-.5*dnve*ui+.5*dnui*ve-penalty*ui*ve);
		ext_result(i+m*n_dofs) += dx*nu*( .5*dnve*ue+.5*dnue*ve+penalty*ue*ve);
	      }
	}
    }
}



template<int dim>
void
NAMEPREFIX::jump_scaled (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input_value,
  const std::vector<double>& ext_input_value,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<double>& factor,
  const double factor2,
  bool divide)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2 * ((divide)
				   ? (1./factor[k])
				   : (factor[k]));
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const double ve = ext_fe.shape_value(i,k);  
	  
	  const double ui = int_input_value[k];
	  const double ue = ext_input_value[k];
	  int_result(i) += dx*nu*ui*vi;
	  int_result(i) -= dx*nu*ue*vi;
	  ext_result(i) -= dx*nu*ui*ve;
	  ext_result(i) += dx*nu*ue*ve;
	}
    }
}



template<int dim>
void
NAMEPREFIX::jump (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input_value,
  const std::vector<double>& ext_input_value,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const double factor2)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      const double nu = factor2;
      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vi = int_fe.shape_value(i,k);
	  const double ve = ext_fe.shape_value(i,k);  
	  
	  const double ui = int_input_value[k];
	  const double ue = ext_input_value[k];
	  int_result(i) += dx*nu*ui*vi;
	  int_result(i) -= dx*nu*ue*vi;
	  ext_result(i) -= dx*nu*ui*ve;
	  ext_result(i) += dx*nu*ue*ve;
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input,
  const std::vector<double>& ext_input,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<Point<dim> >& velocity,
  double factor)
{
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const Tensor<1,dim>& n = int_fe.normal_vector(0);
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = factor * int_fe.JxW(k);
      const double nbeta =  n * velocity[k];

      for (unsigned i=0;i<n_dofs;++i)
	{
	  const double vv = (nbeta<0)
			    ? int_fe.shape_value(i,k)
			    : ext_fe.shape_value(i,k);
	  const double ui = int_input[k];
	  const double ue = ext_input[k];
	  if (nbeta < 0)
	    {
	      int_result(i) -= dx*nbeta*ui*vv;
	      int_result(i) += dx*nbeta*ue*vv;
	    }
	  else
	    {
	      ext_result(i) += dx*nbeta*ue*vv;
	      ext_result(i) -= dx*nbeta*ui*vv;
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input,
  const std::vector<double>& ext_input,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  AssertDimension(velocity.size(), dim);
  AssertDimension(int_input.size(), ext_input.size());
  
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = factor * int_fe.JxW(k);
      
      double nv = 0.;
      for (unsigned int d=0;d<dim;++d)
	nv += n(d) * velocity[d][k];
      
      for (unsigned i=0;i<n_dofs;++i)
	if (nv<0)
	  {
	    int_result(i) += dx * nv * ext_input[k] * int_fe.shape_value(i,k);
	    ext_result(i) -= dx * nv * ext_input[k] * ext_fe.shape_value(i,k);
	  }
	else
	  {
	    int_result(i) += dx * nv * int_input[k] * int_fe.shape_value(i,k);
	    ext_result(i) -= dx * nv * int_input[k] * ext_fe.shape_value(i,k);
	  }
    }
}



template<int dim>
void
NAMEPREFIX::advection (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double factor)
{
  AssertDimension(velocity.size(), dim);
  AssertDimension(int_input.size(), ext_input.size());
  AssertDimension(int_input.size(), int_fe.get_fe().n_components());
  
  const unsigned int n_dofs = int_fe.dofs_per_cell;
  const unsigned int n_components = int_input.size();

  const Point<dim>& n = int_fe.normal_vector(0);
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = factor * int_fe.JxW(k);
      
      double nv = 0.;
      for (unsigned int d=0;d<dim;++d)
	nv += n(d) * velocity[d][k];
      
      for (unsigned i=0;i<n_dofs;++i)
	for (unsigned int c=0;c<n_components;++c)
	  {
	    if (nv<0)
	      {
		int_result(i) += dx * nv * ext_input[c][k] * int_fe.shape_value_component(i,k,c);
		ext_result(i) -= dx * nv * ext_input[c][k] * ext_fe.shape_value_component(i,k,c);
	      }
	    else
	      {
		int_result(i) += dx * nv * int_input[c][k] * int_fe.shape_value_component(i,k,c);
		ext_result(i) -= dx * nv * int_input[c][k] * ext_fe.shape_value_component(i,k,c);
	      }
	  }
    }
}



template<int dim>
void
NAMEPREFIX::u_ni (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input,
  const std::vector<double>& ext_input,
  unsigned int direction,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  double factor)
{
  AssertDimension(int_fe.get_fe().n_components(), ext_fe.get_fe().n_components());
  AssertDimension(int_fetest.get_fe().n_components(), ext_fetest.get_fe().n_components());
  
  if (int_fetest.get_fe().n_components() != 1)
    {
      Assert (int_fetest.get_fe().n_components() == dim,
	      ExcDimensionMismatch(int_fe.get_fe().n_components(), dim));
    }
  else
    {
      Assert (int_fetest.get_fe().n_components() == 1,
	      ExcDimensionMismatch(int_fe.get_fe().n_components(), 1));
    }
  if (ext_fetest.get_fe().n_components() != 1)
    {
      Assert (ext_fetest.get_fe().n_components() == dim,
	      ExcDimensionMismatch(ext_fetest.get_fe().n_components(), dim));
    }
  else
    {
      Assert (ext_fetest.get_fe().n_components() == 1,
	      ExcDimensionMismatch(ext_fetest.get_fe().n_components(), 1));
    }
  
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = factor * int_fe.JxW(k);

      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double ui = int_input[k];
	  const double ue = ext_input[k];
	  
	  const double vi = (int_fetest.get_fe().is_primitive())
			    ? int_fetest.shape_value(i,k)
			    : int_fetest.shape_value_component(i,k,direction);
	  const double ve = (ext_fetest.get_fe().is_primitive())
			    ? ext_fetest.shape_value(i,k)
			    : ext_fetest.shape_value_component(i,k,direction);
	  
	  int_result(i) += dx*.5*n(direction)*ui*vi;
	  ext_result(i) += dx*.5*n(direction)*ui*ve;
	  int_result(i) -= dx*.5*n(direction)*ue*vi;
	  ext_result(i) -= dx*.5*n(direction)*ue*ve;
	}
    }
}



template<int dim>
void
NAMEPREFIX::u_times_n (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input,
  const std::vector<double>& ext_input,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  double factor1, double factor2)
{
  Assert (int_fe.get_fe().n_components() == 1,
	  ExcDimensionMismatch(int_fe.get_fe().n_components(), 1));
  Assert (ext_fe.get_fe().n_components() == 1,
	  ExcDimensionMismatch(ext_fetest.get_fe().n_components(), 1));  
  Assert (int_fetest.get_fe().n_components() == dim,
	  ExcDimensionMismatch(int_fe.get_fe().n_components(), dim));
  Assert (ext_fetest.get_fe().n_components() == dim,
	  ExcDimensionMismatch(ext_fetest.get_fe().n_components(), dim));
  
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  if (factor2 == 0.) factor2 = factor1;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);

      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double ui = int_input[k];
	  const double ue = ext_input[k];
	  
	  for (unsigned int d=0;d<dim;++d)
	    {
	      const double vi = int_fetest.shape_value_component(i,k,d);
	      const double ve = ext_fetest.shape_value_component(i,k,d);
	      
	      int_result(i) += dx*.5*factor1*n(d)*ui*vi;
	      ext_result(i) += dx*.5*factor1*n(d)*ui*ve;
	      int_result(i) -= dx*.5*factor2*n(d)*ue*vi;
	      ext_result(i) -= dx*.5*factor2*n(d)*ue*ve;
	    }
	}
    }
}



template<int dim>
void
NAMEPREFIX::u_dot_n (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  double factor)
{
  Assert(int_input.size() == dim, ExcDimensionMismatch(int_input.size(), dim));
  Assert(ext_input.size() == dim, ExcDimensionMismatch(ext_input.size(), dim));
  Assert (int_fetest.get_fe().n_components() == 1,
	  ExcDimensionMismatch(int_fetest.get_fe().n_components(), 1));
  Assert (ext_fetest.get_fe().n_components() == 1,
	  ExcDimensionMismatch(ext_fetest.get_fe().n_components(), 1));
  
  Assert(int_result.size() == int_fetest.dofs_per_cell,
	 ExcDimensionMismatch(int_result.size(),int_fetest.dofs_per_cell));
  Assert(ext_result.size() == ext_fetest.dofs_per_cell,
	 ExcDimensionMismatch(ext_result.size(),ext_fetest.dofs_per_cell));
  
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = factor * int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);

	  for (unsigned int d=0;d<dim;++d)
	    {
	      const double ui = int_input[d][k];
	      const double ue = ext_input[d][k];
	      
	      int_result(i) += dx*.5*n(d)*ui*vi;
	      ext_result(i) += dx*.5*n(d)*ui*ve;
	      int_result(i) -= dx*.5*n(d)*ue*vi;
	      ext_result(i) -= dx*.5*n(d)*ue*ve;
	    }	      
	}
    }
}

template<int dim>
void
NAMEPREFIX::u_dot_n_scaled (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&,
  const std::vector<double>& int_factor,
  const std::vector<double>& ext_factor)
{
  Assert(int_result.size() == int_fetest.dofs_per_cell,
	 ExcDimensionMismatch(int_result.size(),int_fetest.dofs_per_cell));
  Assert(ext_result.size() == ext_fetest.dofs_per_cell,
	 ExcDimensionMismatch(ext_result.size(),ext_fetest.dofs_per_cell));
//  Assert(int_input_value.size() == int_fe.n_quadrature_points,
//	 ExcDimensionMismatch(int_input_value.size(),
//			      int_fe.n_quadrature_points));
  
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);

	  for (unsigned int d=0;d<dim;++d)
	    {
	      const double ui = int_input_value[d][k];
	      const double ue = ext_input_value[d][k];
	      
	      int_result(i) += dx*.5*n(d)*ui*int_factor[k]*vi;
	      ext_result(i) -= dx*.5*n(d)*ui*int_factor[k]*ve;
	      int_result(i) += dx*.5*n(d)*ue*ext_factor[k]*vi;
	      ext_result(i) -= dx*.5*n(d)*ue*ext_factor[k]*ve;
	    }	      
	}
    }
}


template<int dim>
void
NAMEPREFIX::grad_super (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input_value,
  const std::vector<double>& ext_input_value,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&)
{
  Assert(int_result.size() == dim*int_fetest.dofs_per_cell,
	 ExcDimensionMismatch(int_result.size(),dim*int_fetest.dofs_per_cell));
  Assert(ext_result.size() == dim*ext_fetest.dofs_per_cell,
	 ExcDimensionMismatch(ext_result.size(),dim*ext_fetest.dofs_per_cell));

  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  double C12n = 0.;
  for (unsigned int d=0;d<dim;++d)
    C12n += n(d);
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);

	  const double ui = int_input_value[k];
	  const double ue = ext_input_value[k];

	  for (unsigned int d=0;d<dim;++d)
	    {
	      if (C12n<0)
		{
		  int_result(i+d*t_dofs) += dx*n(d)*ue*vi;
		  ext_result(i+d*t_dofs) -= dx*n(d)*ue*ve;
		}
	      else
		{
		  int_result(i+d*t_dofs) += dx*n(d)*ui*vi;
		  ext_result(i+d*t_dofs) -= dx*n(d)*ui*ve;
		}
	    }	      
	}
    }
}



template<int dim>
void
NAMEPREFIX::div_super (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&)
{
  Assert(int_result.size() == int_fetest.dofs_per_cell,
	 ExcDimensionMismatch(int_result.size(),int_fetest.dofs_per_cell));
  Assert(ext_result.size() == ext_fetest.dofs_per_cell,
	 ExcDimensionMismatch(ext_result.size(),ext_fetest.dofs_per_cell));
//  Assert(int_input_value.size() == int_fe.n_quadrature_points,
//	 ExcDimensionMismatch(int_input_value.size(),
//			      int_fe.n_quadrature_points));
  
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  double C12n = 0.;
  for (unsigned int d=0;d<dim;++d)
    C12n += n(d);
  
  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);

	  for (unsigned int d=0;d<dim;++d)
	    {
	      const double ui = int_input_value[d][k];
	      const double ue = ext_input_value[d][k];
	      
	      if (C12n<0)
		{
		  ext_result(i) += dx*n(d)*ui*ve;
		  ext_result(i) -= dx*n(d)*ue*ve;
		}
	      else
		{
		  int_result(i) += dx*n(d)*ui*vi;
		  int_result(i) -= dx*n(d)*ue*vi;
		}	      
	    }
	}
    }
}


template<int dim>
void
NAMEPREFIX::eps (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&)
{
  Assert (int_input.size() == dim, ExcDimensionMismatch(int_input.size(), dim));
  Assert (ext_input.size() == dim, ExcDimensionMismatch(ext_input.size(), dim));
  
  Assert(int_result.size() == symm_dim(dim)*int_fetest.dofs_per_cell,
	 ExcDimensionMismatch(int_result.size(),symm_dim(dim)*int_fetest.dofs_per_cell));
  Assert(ext_result.size() == symm_dim(dim)*ext_fetest.dofs_per_cell,
	 ExcDimensionMismatch(ext_result.size(),symm_dim(dim)*ext_fetest.dofs_per_cell));

  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);

	  for (unsigned int d=0;d<dim;++d)
	    {
	      const double ui = int_input[d][k];
	      const double ue = ext_input[d][k];
	      
	      int_result(i+d*t_dofs) -= dx*.5*n(d)*ui*vi;
	      ext_result(i+d*t_dofs) += dx*.5*n(d)*ui*ve;
	      int_result(i+d*t_dofs) -= dx*.5*n(d)*ue*vi;
	      ext_result(i+d*t_dofs) += dx*.5*n(d)*ue*ve;
	      Assert (dim==2, ExcNotImplemented());
	      int_result(i+2*t_dofs) -= dx*.5*n(1-d)*ui*vi;
	      ext_result(i+2*t_dofs) += dx*.5*n(1-d)*ui*ve;
	      int_result(i+2*t_dofs) -= dx*.5*n(1-d)*ue*vi;
	      ext_result(i+2*t_dofs) += dx*.5*n(1-d)*ue*ve;	      
	    }	      
	}
    }
}



template<int dim>
void
NAMEPREFIX::eps_div (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&)
{
  Assert (int_input.size() == symm_dim(dim),
	  ExcDimensionMismatch(int_input.size(), symm_dim(dim)));
  Assert (ext_input.size() == symm_dim(dim),
	  ExcDimensionMismatch(ext_input.size(), symm_dim(dim)));
  
  Assert(int_result.size() == dim*int_fetest.dofs_per_cell,
	 ExcDimensionMismatch(int_result.size(),dim*int_fetest.dofs_per_cell));
  Assert(ext_result.size() == dim*ext_fetest.dofs_per_cell,
	 ExcDimensionMismatch(ext_result.size(),dim*ext_fetest.dofs_per_cell));
  
  const unsigned int t_dofs = int_fetest.dofs_per_cell;
  
  const Point<dim>& n = int_fe.normal_vector(0);

  for (unsigned k=0;k<int_fe.n_quadrature_points;++k)
    {
      const double dx = int_fe.JxW(k);
      
      for (unsigned i=0;i<t_dofs;++i)
	{
	  const double vi = int_fetest.shape_value(i,k);
	  const double ve = ext_fetest.shape_value(i,k);


	  for (unsigned int d1=0;d1<dim;++d1)
	    {
	      const double ui1 = int_input[d1][k];
	      const double ue1 = ext_input[d1][k];
	      
	      int_result(i+d1*t_dofs) -= .5*dx*n(d1)*ui1*vi;
	      ext_result(i+d1*t_dofs) += .5*dx*n(d1)*ui1*ve;
	      int_result(i+d1*t_dofs) -= .5*dx*n(d1)*ue1*vi;
	      ext_result(i+d1*t_dofs) += .5*dx*n(d1)*ue1*ve;

	      for (unsigned d2=d1+1;d2<dim;++d2)
		{
		  Assert (dim==2, ExcNotImplemented());
		  const unsigned int d3 = 2;
		  const double ui3 = int_input[d3][k];
		  const double ue3 = ext_input[d3][k];
	      
		  int_result(i+d1*t_dofs) -= .5*dx*vi* n(d2)*ui3;
		  ext_result(i+d1*t_dofs) += .5*dx*ve* n(d2)*ui3;
		  int_result(i+d1*t_dofs) -= .5*dx*vi* n(d2)*ue3;
		  ext_result(i+d1*t_dofs) += .5*dx*ve* n(d2)*ue3;

		  int_result(i+d2*t_dofs) -= .5*dx*vi* n(d1)*ui3;
		  ext_result(i+d2*t_dofs) += .5*dx*ve* n(d1)*ui3;
		  int_result(i+d2*t_dofs) -= .5*dx*vi* n(d1)*ue3;
		  ext_result(i+d2*t_dofs) += .5*dx*ve* n(d1)*ue3;
		}
	    }	      
	}
    }
}

#define dim deal_II_dimension

template
void
NAMEPREFIX::ip<dim> (
      Vector<double>&, Vector<double>&,
      const std::vector<double>&, const std::vector<Tensor<1,dim> >&,
      const std::vector<double>&, const std::vector<Tensor<1,dim> >&,
      const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
      double, double, double);

template
void
NAMEPREFIX::ip<dim> (
      Vector<double>& int_result,
      Vector<double>& ext_result,
      const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& int_input_grad,
      const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& ext_input_grad,
      const FEFaceValuesBase<dim>& int_fe,
      const FEFaceValuesBase<dim>& ext_fe,
      double, double, double);

template
void
NAMEPREFIX::ip_scaled<dim> (
      Vector<double>& int_result,
      Vector<double>& ext_result,
      const std::vector<double>& int_input_value,
      const std::vector<Tensor<1,dim> >& int_input_grad,
      const std::vector<double>& ext_input_value,
      const std::vector<Tensor<1,dim> >& ext_input_grad,
      const FEFaceValuesBase<dim>& int_fe,
      const FEFaceValuesBase<dim>& ext_fe,
      const std::vector<double>& factor,
      double penalty);

template
void
NAMEPREFIX::ip_scaled<dim> (
      Vector<double>& int_result,
      Vector<double>& ext_result,
      const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& int_input_grad,
      const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
      const VectorSlice<const std::vector<std::vector<Tensor<1,dim> > > >& ext_input_grad,
      const FEFaceValuesBase<dim>& int_fe,
      const FEFaceValuesBase<dim>& ext_fe,
      const std::vector<double>& factor,
      double penalty);

template
void
NAMEPREFIX::jump_scaled<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input_value,
  const std::vector<double>& ext_input_value,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<double>& factor,
  const double factor2,
  bool divide);

template
void
NAMEPREFIX::jump<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input_value,
  const std::vector<double>& ext_input_value,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const double factor2);

template
void
NAMEPREFIX::u_ni<dim> (
  Vector<double>&, Vector<double>&,
  const std::vector<double>&, const std::vector<double>&,
  unsigned int,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  double factor);

template
void
NAMEPREFIX::u_times_n<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input_value,
  const std::vector<double>& ext_input_value,
  const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&,
  double, double);

template
void
NAMEPREFIX::u_dot_n<dim> (
  Vector<double>&, Vector<double>&,
  const VectorSlice<const std::vector<std::vector<double> > >&,
  const VectorSlice<const std::vector<std::vector<double> > >&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&,
  const FEFaceValuesBase<dim>&, const FEFaceValuesBase<dim>&, double);

template
void
NAMEPREFIX::u_dot_n_scaled<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&,
  const std::vector<double>& int_factor,
  const std::vector<double>& ext_factor);

template
void
NAMEPREFIX::grad_super<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const std::vector<double>& int_input_value,
  const std::vector<double>& ext_input_value,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&);

template
void
NAMEPREFIX::div_super<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input_value,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input_value,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&);

template
void
NAMEPREFIX::advection<dim> (
  Vector<double>& int_result, Vector<double>& ext_result,
  const std::vector<double>& int_input, const std::vector<double>& ext_input,
  const FEFaceValuesBase<dim>& int_fe, const FEFaceValuesBase<dim>& ext_fe,
  const std::vector<Point<dim> >& velocity, double);

template
void
NAMEPREFIX::advection<dim> (
  Vector<double>& int_result, Vector<double>& ext_result,
  const std::vector<double>& int_input, const std::vector<double>& ext_input,
  const FEFaceValuesBase<dim>& int_fe, const FEFaceValuesBase<dim>& ext_fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double);

template
void
NAMEPREFIX::advection<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>& ext_fe,
  const VectorSlice<const std::vector<std::vector<double> > >& velocity,
  double);

template
void
NAMEPREFIX::eps<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&);

template
void
NAMEPREFIX::eps_div<dim> (
  Vector<double>& int_result,
  Vector<double>& ext_result,
  const VectorSlice<const std::vector<std::vector<double> > >& int_input,
  const VectorSlice<const std::vector<std::vector<double> > >& ext_input,
  const FEFaceValuesBase<dim>& int_fetest,
  const FEFaceValuesBase<dim>& ext_fetest,
  const FEFaceValuesBase<dim>& int_fe,
  const FEFaceValuesBase<dim>&);
