//---------------------------------------------------------------------------
//    $Id: runge_kutta.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <appframe/runge_kutta.h>
#include <base/parameter_handler.h>
#include <lac/block_vector.h>

#include <sstream>
#include <iostream>

using namespace AppFrame;

RungeKuttaExplicit::RungeKuttaExplicit(ApplicationBase& app)
		:
		ApplicationCopy(app)
{
  deallog << "->RK";
}


void
RungeKuttaExplicit::declare_parameters(ParameterHandler& param)
{
  param.enter_subsection("RungeKutta");
  TimestepControl::declare_parameters (param);
  param.declare_entry("Scheme", "RK4",
		      Patterns::Selection("RK2|RK4|Euler|Midpoint"));
  param.leave_subsection();
  ApplicationCopy::declare_parameters(param);
}


void
RungeKuttaExplicit::_initialize (ParameterHandler& param)
{
  param.enter_subsection("RungeKutta");
  control.parse_parameters (param);
  set_scheme(param.get("Scheme"));
  param.leave_subsection ();
}


void
RungeKuttaExplicit::initialize (ParameterHandler& param)
{
  ApplicationCopy::initialize(param);
  _initialize(param);
}


void
RungeKuttaExplicit::resize_vectors(unsigned int n)
{
  a.resize(n);
  b.resize(n);
  c.resize(n);
  for (unsigned int i=0;i<n;++i)
    a[i].resize(i, 0.);
  c[0] = 0.;
}


void
RungeKuttaExplicit::set_scheme(const std::string name)
{
  if (name == "Euler")
    {
      resize_vectors(1);
      b[0] = 1.;
    }
  else if (name == "Midpoint")
    {
      resize_vectors(2);
      a[1][0] = .5;
      b[0] = 0.;
      b[1] = 1.;
      c[1] = .5;
    }
  else if (name == "RK2")
    {
      resize_vectors(2);
      a[1][0] = 2./3.;
      b[0] = .25;
      b[1] = .75;
      c[1] = 2./3.;
    }
  else if (name == "RK4")
    {
      resize_vectors(4);
      a[1][0] = .5;
      a[2][1] = .5;
      a[3][2] = 1.;
      b[0] = 1./6.;
      b[1] = 1./3.;
      b[2] = 1./3.;
      b[3] = 1./6.;
      c[1] = .5;
      c[2] = .5;
      c[3] = 1.;
    }
  else
    {
      AssertThrow(false, ExcUnknownScheme(name))
    }
}


void
RungeKuttaExplicit::solve(FEVector& y, const FEVectors& in_vectors)
{
  deallog.push("RK");
  std::vector<FEVector*> k(b.size());

				   // Allocate auxiliary vectors
  for (unsigned int i=0;i<k.size();++i)
    {
      k[i] = data->block_vector_pool.alloc();
      k[i]->reinit(y, true);
    }
  
  FEVector* aux = data->block_vector_pool.alloc();
  aux->reinit(y, true);
  FEVectors src;
  src.add_vector(*aux, "Runge-Kutta");
  src.merge(in_vectors);
				   // Set up control values
  control.restart();
  bool first = true;
  double previous_step = control.now();
  double t = control.now();
  double sub_time;
  double step = control.step();
  double step_indicator = 0.;
  
  this->data->enter("time", sub_time);
  
  while (t<control.final())
    {
      if (!first)
	{
	  t = control.next(step_indicator);
	  if (t>=control.final()) break;
	  step = control.step();
	}
      
      deallog << "Time step:" << t << std::endl;
//      y.print(std::cout);
      if (control.print())
	{
	  std::ostringstream streamOut;
	  streamOut << t;
	  std::string name = "RK_" + streamOut.str();
	  FEVectors out;
	  out.add_vector(y, "u");
	  app->data_out(name, out);
	}
      
      sub_time = t;
      
      for (unsigned int i=0;i<b.size();++i)
	{
	  sub_time = t + c[i] * step;
	  *aux = y;
	  for (unsigned int j=0;j<i;++j)
	    if (a[i][j] != 0.)
	      aux->add(a[i][j] * step, *k[j]);
	  app->residual(*k[i], src);
	}
      for (unsigned int i=0;i<b.size();++i)
	{
	  y.add(b[i] * step, *k[i]);
//	  deallog << "B " << b[i] << ' ' << step << std::endl;
	}
      
//      y.print(std::cout);
      first = false;
      previous_step = step;
    }
  
  data->erase_scalar("time");
  
  data->block_vector_pool.free(aux);
  for (unsigned int i=0;i<k.size();++i)
    data->block_vector_pool.free(k[i]);
  deallog.pop();
}

