//---------------------------------------------------------------------------
//    $Id: timestepping.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <appframe/timestepping.h>
#include <base/parameter_handler.h>
#include <lac/block_vector.h>

using namespace AppFrame;


const Event ThetaTimestepping::new_time = Event::assign("Theta::time");
const Event ThetaTimestepping::new_step_size = Event::assign("Theta::step");


ThetaTimestepping::ThetaTimestepping(ApplicationBase& app)
		: ApplicationCopy(app),
		  theta(.5),
		  adaptive(false),
		  default_residual(app),
		  compute_residual(&default_residual)
{
  deallog << "->Theta";
}


ThetaTimestepping::ThetaTimestepping(ApplicationBase& app,
				     const Residual& residual)
		: ApplicationCopy(app),
		  theta(.5),
		  adaptive(false),
		  default_residual(app),
		  compute_residual(&residual)
{
  deallog << "->Theta";
}


void
ThetaTimestepping::declare_parameters(ParameterHandler& param)
{
    param.enter_subsection("Theta");
    TimestepControl::declare_parameters (param);
    param.declare_entry("Theta", ".5", Patterns::Double());
    param.declare_entry("Adaptive", "false", Patterns::Bool());
    param.leave_subsection();
    
    ApplicationCopy::declare_parameters(param);
}

void
ThetaTimestepping::_initialize (ParameterHandler& param)
{
    param.enter_subsection("Theta");
    control.parse_parameters (param);
    theta = param.get_double("Theta");
    adaptive = param.get_bool("Adaptive");
    param.leave_subsection ();
}


void
ThetaTimestepping::initialize (ParameterHandler& param)
{
    ApplicationCopy::initialize(param);
    _initialize(param);
}


double
ThetaTimestepping::residual (
  FEVector&, const FEVectors&)
{
  return 0;
}


void
ThetaTimestepping::solve (
  FEVector& u, const FEVectors& in_vectors)
{
  deallog.push ("Theta");
  char fname[30];
  
  FEVector* aux = this->data->block_vector_pool.alloc();
  aux->reinit(u);
  FEVector* u_old = this->data->block_vector_pool.alloc();
  u_old->reinit(u);
  FEVector* res_old = this->data->block_vector_pool.alloc();
  res_old->reinit(u);

  control.restart();
  bool first = true;
  double previous_step = 0;
  double t = control.now();
  double step = control.step();
  double step_indicator = 0.;
  unsigned int count = 0;

				   // The data used to compute the
				   // vector associated with the old
				   // timestep
  FEVectors src1;
  src1.add_vector(u, "previous time solution");
  src1.add_scalar(step, "timestep");
  src1.add_scalar(t, "time");
  src1.add_scalar(theta, "theta");
  src1.merge(in_vectors);
				   // The data provided to the inner
				   // solver
  FEVectors src2;
  src2.add_vector(*aux, "previous time residual");
  src2.add_scalar(step, "timestep");
  src2.add_scalar(t, "time");
  src2.add_scalar(theta, "theta");
  src2.merge(in_vectors);
  
  while (t<control.final())
    {      
      if (first)
	{
	  sprintf(fname, "Step%05d", count);
	  FEVectors out;
	  out.add_vector(u, "u");
	  app->data_out(fname, out);
	}
      
      t = control.next(step_indicator);
      step = control.step();
      deallog << "Time step:" << t << std::endl;
      
				       // Compute
				       // (1./dt + theta A) u
      (*compute_residual)(*aux, src1);
      
      *u_old = u;
      if (step != first)
	app->notify(new_time);
      if (step != previous_step)
	app->notify(new_step_size);
      
      app->solve(u, src2);
      
      if (control.print())
	{
	  sprintf(fname, "Step%05d", ++count);
	  FEVectors out;
	  out.add_vector(u, "u");
	  app->data_out(fname, out);
	}
      
      first = false;
      previous_step = step;
    }
  this->data->block_vector_pool.free(res_old);
  this->data->block_vector_pool.free(u_old);
  this->data->block_vector_pool.free(aux);
  deallog.pop();
}

//----------------------------------------------------------------------//

ThetaResidual::ThetaResidual(ApplicationBase& application)
		:
		app(&application)
{}


double
ThetaResidual::operator() (FEVector& dst, const FEVectors& src) const
{
  FEVectors mysrc;
  double scaling;
  const double theta = src.scalar(2);
  
  mysrc.add_scalar(scaling, "scaling");
  mysrc.merge(src);
  
  const unsigned int ptr = src.find_vector("previous time residual");
  if (ptr != deal_II_numbers::invalid_unsigned_int)
    {
				       // This is the actual residual
				       // for the Newton iteration of
				       // the implicit scheme.
      scaling = theta;
      
      app->residual(dst, mysrc);
      dst.add(-1., src.vector(ptr));
      return dst.l2_norm();
    }
  else
    {
				       // This is the operator
				       // evaluated at the previous timestep
      scaling = theta-1.;
      return app->residual(dst, mysrc);
    }
  return 0.;
}

