//---------------------------------------------------------------------------
//    $Id: block_matrix_step3.cc$
//
//    Copyright (C) 2006-12 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <base/parameter_handler.h>
#include <base/function_lib.h>
#include <base/function.h>
#include <base/quadrature_lib.h>
#include <lac/block_vector.h>
#include <lac/solver_cg.h>
#include <lac/precondition.h>
#include <appframe/block_matrix_application.h>
#include <appframe/mesh_loop.h>
#include <appframe/matrix_shop_cell.h>
#include <appframe/residual_shop_cell.h>
#include <grid/grid_generator.h>
#include <fe/fe_values.h>
#include <numerics/vector_tools.h>
#include <numerics/matrix_tools.h>

using namespace AppFrame;

/**
 * This class is used to test the application 
 * framework to solve a simple finite element problem (Laplace equation).
 * This is done by solving the problem in step-3 in deal.ii using the framework.
 * Then, the results can be compared using 
 * @verbatim
 * $diff matrix_step3_result deal_II_result
 * @endverbatim
 *
 * @author Marc Secanell
*/
template <int dim>
class TestApplication :
  public BlockMatrixApplication<dim>
{
public:
  TestApplication ();
  void _initialize();
  virtual void initialize(ParameterHandler& param);
  virtual void cell_matrix(MatrixVector& cell_matrices,
			   const typename DoFApplication<dim>::CellInfo& cell);  
  virtual void cell_residual(FEVector& cell_vector,
			     const typename DoFApplication<dim>::CellInfo& cell);
  virtual void dirichlet_bc(std::map<unsigned int, double>& boundary_values) const;
  /*virtual void dirichet_bc(FEVector& rhs,
    FEVector& sol);*/
  virtual void solve(FEVector& start, const FEVectors& rhs);
  virtual double evaluate (const FEVectors& src);
};


template <int dim>
TestApplication<dim>::TestApplication()
{
  deallog << "->Step3-" << dim;
  this->boundary_fluxes = false;
  this->interior_fluxes = false;
}


template <int dim>
void
TestApplication<dim>::initialize(ParameterHandler& param)
{
  BlockMatrixApplication<dim>::initialize(param);
  _initialize();
}


template <int dim>
void
TestApplication<dim>::_initialize()
{
  // Specify the coupling between equations and variables
  this->cell_couplings.reinit(1,1);
  this->cell_couplings(0,0) = DoFTools::always;
  this->interior_fluxes = false;
  this->boundary_fluxes = false;
  
  // Generate grid and refine
  GridGenerator::hyper_cube(*this->tr, -1., 1.);
  this->tr->refine_global(5);

  // Initialize dof per cell for the system of equations and information on
  //block sizes and indices
  this->remesh_dofs();
  // Initialize matrices and spartisity pattern for the whole system
  this->remesh_matrices();
}


template <int dim>
void
TestApplication<dim>::cell_matrix(
  MatrixVector& cell_matrices,
  const typename DoFApplication<dim>::CellInfo& info)
{
  Assert(cell_matrices.size() == 1,
	 ExcDimensionMismatch(cell_matrices.size(), 1));
  Assert(cell_matrices[0].row == 0,
	 ExcDimensionMismatch(cell_matrices[0].row, 0));
  Assert(cell_matrices[0].column == 0,
	 ExcDimensionMismatch(cell_matrices[0].column, 0));
  AppShop::Matrix::Cell::laplacian(cell_matrices[0].matrix, info.fe(0));
  //AppShop::Matrix::Cell::mass(cell_matrices[0].matrix, info.fe(0), 1.);
}



template <int dim>
void
TestApplication<dim>::cell_residual(
  FEVector& cell_vector,
  const typename DoFApplication<dim>::CellInfo& info)
{
  const FEValuesBase<dim>& fe = info.fe(0);
  std::vector<double> f(fe.n_quadrature_points, 1.0); //creates vector with n_quadrature_point components
  //with their value initialized to one.
  //solution.value_list(fe.get_quadrature_points(), f);
  AppShop::Residual::Cell::mass(cell_vector.block(0), f, fe, 1.0);//dim*M_PI*M_PI/4.+1.);
}

template <int dim>
void
TestApplication<dim>::dirichlet_bc(std::map<unsigned int, double>& boundary_values) const
{
  VectorTools::interpolate_boundary_values (*this->dof,
					    0,
					    ConstantFunction<dim>(0.0),
					    boundary_values);
}

template <int dim>
void
TestApplication<dim>::solve(FEVector& start,
			    const FEVectors& rhss)
{
  const FEVector& rhs = rhss.vector(0);
  if (this->notifications.any())
    {
      this->assemble(rhss);
      this->clear_events();
    }
  SolverControl control;
  SolverCG<FEVector > solver(control, this->data->block_vector_pool);
  solver.solve(this->filtered_matrix, start, rhs, PreconditionIdentity());

   //Return the solution to all the nodes taking into accout the hanging nodes
  this->hanging_node_constraints.distribute(start);
  
  //Plot solution out (this is so that I can run tests on the solution):
  deallog<<"Solution: "<<std::endl;
  for (unsigned int i=0; i<start.block(0).size(); ++i)
    deallog<<"component "<<i<<" : "<<start.block(0)(i)<<std::endl;
}



template <int dim>
double
TestApplication<dim>::evaluate(const FEVectors&)
{
  return 0;
}


int
main(int argc, char** argv)
{
  // initialize file_utils routines and deallog
  AppFrame::initialize(argc, argv);
  AppFrame::open_logfile();
  deallog << "Program:" << argv[0] << std::endl;
  
  ParameterHandler param;
  TestApplication<2> app;
  deallog << std::endl;
  
  AppFrame::declare_parameter_files(param);
  app.declare_parameters(param);
  AppFrame::read_parameter_files(param);
  app.initialize(param);
  
  FEVector u;
  FEVector rhs;
  app.init_vector(u);
  app.init_vector(rhs);

  FEVectors data;
  data.add_vector(rhs, "main");
  
  // build system of eq
  app.assemble(data);
  // build rhs
  app.residual(rhs, data);
  // apply dirichet bc
  //app.dirichet_bc(rhs,u);
  
  try
    {
      app.solve(u, data);
//      app.data_out("sol-step-3",u);
    }
  catch (const std::exception& e)
    {
      deallog << e.what() << std::endl;
    }
}
