//---------------------------------------------------------------------------
//    $Id: block_matrix_step6.cc 1415 2013-08-30 16:14:41Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <base/parameter_handler.h>
#include <base/function_lib.h>
#include <base/function.h>
#include <base/quadrature_lib.h>

#include <lac/block_vector.h>
#include <lac/solver_cg.h>
#include <lac/precondition.h>

#include <appframe/block_matrix_application.h>
#include <appframe/mesh_loop.h>
#include <appframe/matrix_shop_cell.h>
#include <appframe/residual_shop_cell.h>

#include <grid/grid_generator.h>
#include <grid/tria_accessor.h>
#include <grid/tria_iterator.h>
#include <grid/tria_boundary_lib.h>

#include <fe/fe_values.h>

#include <numerics/vector_tools.h>
#include <numerics/matrix_tools.h>
#include <numerics/error_estimator.h>

#include<fstream>
#include<iostream>
#include<sstream>

using namespace AppFrame;

/**
 * This class is used to test adaptive refinement using the application 
 * framework.
 * This is done by solving the problem in step-6 in deal.ii using the framework.
 * Then, the results can be compared using 
 * @verbatim
 * $diff matrix_step6_result deal_II_result
 * @endverbatim
 *
 * @author Marc Secanell
*/
template <int dim>
class TestApplication :
  public BlockMatrixApplication<dim>
{
public:
  TestApplication ();
  void _initialize();
  virtual void initialize(ParameterHandler& param);
  virtual void cell_matrix(MatrixVector& cell_matrices,
			   const typename DoFApplication<dim>::CellInfo& cell);  
  virtual void cell_residual(FEVector& cell_vector,
			     const typename DoFApplication<dim>::CellInfo& cell);
  virtual void dirichlet_bc(std::map<unsigned int, double>& boundary_values) const;
  /*virtual void dirichlet_bc(FEVector& rhs,
    FEVector& sol);*/
  virtual void solve(FEVector& start, const FEVectors& rhs);
  virtual double estimate (const FEVectors& sol);
  virtual double evaluate (const FEVectors& src);
};


template <int dim>
TestApplication<dim>::TestApplication()
{
  deallog << "->Step6-" << dim<<"d"<<std::endl;
  this->boundary_fluxes = false;
  this->interior_fluxes = false;
}


template <int dim>
void
TestApplication<dim>::initialize(ParameterHandler& param)
{
  BlockMatrixApplication<dim>::initialize(param);
  _initialize();
}


template <int dim>
void
TestApplication<dim>::_initialize()
{
  // Specify the coupling between equations and variables
  this->cell_couplings.reinit(1,1);
  this->cell_couplings(0,0) = DoFTools::always;
  this->interior_fluxes = false;
  this->boundary_fluxes = false;
  
  // Generate grid and refine
  GridGenerator::hyper_ball (*this->tr);
  static const HyperBallBoundary<dim> boundary;
  this->tr->set_boundary(0, boundary);
  this->tr->refine_global(1);

  // Initialize dof per cell for the system of equations and information on
  //block sizes and indices
  this->remesh_dofs();
  // Initialize matrices and spartisity pattern for the whole system
  this->remesh_matrices();
}


template <int dim>
void
TestApplication<dim>::cell_matrix(
  MatrixVector& cell_matrices,
  const typename DoFApplication<dim>::CellInfo& info)
{
  Assert(cell_matrices.size() == 1,
	 ExcDimensionMismatch(cell_matrices.size(), 1));
  Assert(cell_matrices[0].row == 0,
	 ExcDimensionMismatch(cell_matrices[0].row, 0));
  Assert(cell_matrices[0].column == 0,
	 ExcDimensionMismatch(cell_matrices[0].column, 0));
  
  //compute factors:
  std::vector<double> factor(info.fe(0).n_quadrature_points);
  std::vector<Point<dim> > q_points = info.fe(0).get_quadrature_points();
  for (unsigned k=0;k<info.fe(0).n_quadrature_points;++k)
    {
      if (q_points[k].square() < 0.5*0.5)
	factor[k] = 20;
      else
	factor[k] = 1;
    }
  
  AppShop::Matrix::Cell::laplacian_scaled(cell_matrices[0].matrix, info.fe(0), factor, false);
  //AppShop::Matrix::Cell::mass(cell_matrices[0].matrix, info.fe(0), 1.);
}



template <int dim>
void
TestApplication<dim>::cell_residual(FEVector& cell_vector,
									const typename DoFApplication<dim>::CellInfo& info)
{
  const FEValuesBase<dim>& fe = info.fe(0);
  std::vector<double> f(fe.n_quadrature_points, 1.0); //creates vector with n_quadrature_point components
  //with their value initialized to one.
  //solution.value_list(fe.get_quadrature_points(), f);
  AppShop::Residual::Cell::mass(cell_vector.block(0), f, fe, 1.0);//dim*M_PI*M_PI/4.+1.);
}

template <int dim>
void
TestApplication<dim>::dirichlet_bc(std::map<unsigned int, 
								   double>& boundary_values) const
{
  VectorTools::interpolate_boundary_values (*this->dof,
					    0,
					    ZeroFunction<dim>(),
					    boundary_values);
}

template <int dim>
void
TestApplication<dim>::solve(FEVector& start, const FEVectors& rhs)
{
  if (this->notifications.any())
    {
      this->assemble(rhs);
      this->clear_events();
    }
  
  // Now, solve:
  SolverControl control(1000, 1e-12);
  SolverCG<FEVector > cg (control, this->data->block_vector_pool);
  
  PreconditionJacobi<BlockSparseMatrix<double> > preconditioner;
  preconditioner.initialize(this->matrix, 1.2);
  cg.solve(this->filtered_matrix,
	   start, rhs.vector(0), preconditioner);

  //Return the solution to all the nodes taking into accout the hanging nodes
  this->hanging_node_constraints.distribute(start);

  //Plot solution out if necessary:
  deallog<<"Solution: "<<std::endl;
  for (unsigned int i=0; i<start.block(0).size(); ++i)
    deallog<<"component "<<i<<" : "<<start.block(0)(i)<<std::endl;
}

template <int dim>
double
TestApplication<dim>::estimate(const FEVectors& sols)
{
  const FEVector& sol = sols.vector(0);
  this->cell_errors.reinit(this->tr->n_active_cells());
  typename FunctionMap<dim>::type newmann_boundary;
  KellyErrorEstimator<dim>::estimate(*this->dof,
				     QGauss<dim-1>(3),
				     newmann_boundary,
				     sol.block(0),
				     this->cell_errors);
  
  return 0;
}

template <int dim>
double
TestApplication<dim>::evaluate(const FEVectors&)
{
  
  return 0;
  
}


int
main(int argc, char** argv)
{
  // initialize file_utils routines and deallog
  AppFrame::initialize(argc, argv);
  AppFrame::open_logfile();
  deallog << "Program:" << argv[0] << std::endl;
  
  ParameterHandler param;
  TestApplication<2> app;
  deallog << std::endl;
  
  AppFrame::declare_parameter_files(param);
  app.declare_parameters(param);
  AppFrame::read_parameter_files(param);
  app.initialize(param);

  FEVector u;
  FEVector rhs;
  FEVectors vecs;
  vecs.add_vector(u, "main");
  
  //unsigned int cycle=0;
  for (unsigned int cycle=0; cycle<5; ++cycle)
    {
      deallog<< "Cycle "<<cycle<<" : "<<std::endl;
      std::ostringstream streamOut;
      streamOut << cycle;
      std::string filename = "step-6-grid-"+streamOut.str();
      std::string sol_filename = "sol-step-6-grid-"+streamOut.str();
      
      if (cycle == 0)
	{
	  app.remesh_dofs();
	  app.remesh_matrices();
	}
      else
	{
	  app.estimate(vecs);
	  app.remesh();
	}
      
      // initialize vectors with new grid
      app.init_vector(u);
      app.init_vector(rhs);
      FEVectors data;
      data.add_vector(rhs, "main");
      
      // build system of eq
      app.assemble(data);
      // build rhs
      app.residual(rhs,data);
      
      try
	{
	  app.solve(u, data);
	  app.grid_out(filename);
//	  app.data_out(sol_filename,u);
	}
      catch (const std::exception& e)
	{
	  deallog << e.what() << std::endl;
	}
    }
}
