//---------------------------------------------------------------------------
//    $Id: block_matrix_step8.cc 1415 2013-08-30 16:14:41Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <base/parameter_handler.h>
#include <base/function_lib.h>
#include <base/function.h>
#include <base/quadrature_lib.h>

#include <lac/block_vector.h>
#include <lac/solver_cg.h>
#include <lac/solver_bicgstab.h>
#include <lac/solver_gmres.h>
#include <lac/precondition.h>
#include <lac/precondition_block.h>
#include <lac/block_matrix_array.h>
#include <lac/filtered_matrix.h>
#include <lac/sparse_ilu.h>

#include <appframe/block_matrix_application.h>
#include <appframe/mesh_loop.h>
#include <appframe/matrix_shop_cell.h>
#include <appframe/residual_shop_cell.h>

#include <grid/grid_generator.h>
#include <grid/tria_accessor.h>
#include <grid/tria_iterator.h>
#include <grid/tria_boundary_lib.h>

#include <fe/fe_values.h>

#include <numerics/vector_tools.h>
#include <numerics/matrix_tools.h>
#include <numerics/error_estimator.h>

#include<fstream>
#include<iostream>
#include<sstream>

using namespace AppFrame;

/**
 * This class is used to test to solve a system of PDE using the ApplicationFramework and to thes
 * adaptive refinement for systems of PDEs using the framework.
 * This is done by solving the problem in step-8 in deal.ii using the framework.
 * Then, the results can be compared using 
 * @verbatim
 * $diff matrix_step8_result deal_II_result
 * @endverbatim
 *
 * @author Marc Secanell
*/
template <int dim>
class TestApplication :
  public BlockMatrixApplication<dim>
{
public:
  TestApplication ();
  void _initialize();
  virtual void initialize(ParameterHandler& param);
  virtual void cell_matrix(MatrixVector& cell_matrices,
			   const typename DoFApplication<dim>::CellInfo& cell);  
  virtual void cell_residual(FEVector& cell_vector,
			     const typename DoFApplication<dim>::CellInfo& cell);
  virtual void dirichlet_bc(std::map<unsigned int, double>& boundary_values) const;
  virtual void solve(FEVector& start, const FEVectors& rhs);
  virtual double estimate(const FEVector& sol);
  virtual double evaluate (const FEVectors& src);

  void elasticity (FullMatrix<double>& M,
		   const FEValuesBase<dim>& fe,
		   const double factor);
};


template <int dim>
TestApplication<dim>::TestApplication()
{
  deallog << "->Step8-" << dim<<"d"<<std::endl;
  this->boundary_fluxes = false;
  this->interior_fluxes = false;
}


template <int dim>
void
TestApplication<dim>::initialize(ParameterHandler& param)
{
  BlockMatrixApplication<dim>::initialize(param);
  _initialize();
}


template <int dim>
void
TestApplication<dim>::_initialize()
{
  // Specify the coupling between equations and variables
  this->cell_couplings.reinit(2,2);
  this->cell_couplings(0,0) = DoFTools::always;
  this->cell_couplings(1,1) = DoFTools::always;
  this->cell_couplings(0,1) = DoFTools::always;
  this->cell_couplings(1,0) = DoFTools::always;
  this->interior_fluxes = false;
  this->boundary_fluxes = false;

  // Generate grid and refine
  GridGenerator::hyper_cube (*this->tr, -1, 1);
  this->tr->refine_global(2);

  // Initialize dof per cell for the system of equations and information on
  //block sizes and indices
  this->remesh_dofs();
  // Initialize matrices and spartisity pattern for the whole system
  this->remesh_matrices();
  //
}


template <int dim>
void
TestApplication<dim>::cell_matrix(
  MatrixVector& cell_matrices,
  const typename DoFApplication<dim>::CellInfo& info)
{

  /* Assert(cell_matrices[0].row == 0,
	 ExcDimensionMismatch(cell_matrices[0].row, 0));
  Assert(cell_matrices[0].column == 0,
	 ExcDimensionMismatch(cell_matrices[0].column, 0));*/
  
  //Compute cell_matrix for the first equation:
  const FEValuesBase<dim>& fe = info.fe(0);
  const unsigned int n_dofs = fe.dofs_per_cell;
  
  std::vector<double> lambda_values(fe.n_quadrature_points, 1.0); //creates vector with n_quadrature_point components
  std::vector<double> mu_values(fe.n_quadrature_points, 1.0); //creates vector with n_quadrature_point components

  //Guido: Here I wanted to call div_div, however I cannot because I need a matrix of size dim*dof x dim*dof, however,
  //matrices.matrix is of size dof x dof
  /*  
  std::vector<unsigned int> cell_matrix_indices(dim*dim);
  for (unsigned int i=0; i< cell_matrix_indices.size(); ++i) 
    cell_matrix_indices[i] = i;
  AppShop::Matrix::Cell::div_div(cell_matrices,
				 cell_matrix_indices,
				 fe);
  */
  
  //deallog<<"DOF PER CELL: "<<n_dofs<<std::endl;

  for (unsigned int m=0; m<cell_matrices.size(); ++m)
    for (unsigned i=0;i<n_dofs;++i)
      for (unsigned j=0;j<n_dofs;++j)
	for (unsigned k=0;k<fe.n_quadrature_points;++k)
	  {
	    cell_matrices[m].matrix(i,j) += (
					     (fe.shape_grad(i,k)[cell_matrices[m].row]*   //in this case where n_comp == dim, cell_matrices[m].row == component_i
 					      fe.shape_grad(j,k)[cell_matrices[m].column]* //in this case where n_comp == dim, cell_matrices[m].column == component_j
					      lambda_values[k]) 
					     +
					     (fe.shape_grad(i,k)[cell_matrices[m].column]*
					      fe.shape_grad(j,k)[cell_matrices[m].row]*
					      mu_values[k]) 
					     +
					     ((cell_matrices[m].row == cell_matrices[m].column) ?
					      (fe.shape_grad(i,k)*
					       fe.shape_grad(j,k)*
					       mu_values[k]) :
					      0)
					     )*fe.JxW(k);
	  }	      
}

template <int dim>
void
TestApplication<dim>::elasticity (FullMatrix<double>& M,
				  const FEValuesBase<dim>& fe,
				  const double factor)
{
  
}

template <int dim>
void
TestApplication<dim>::cell_residual(FEVector& cell_vector,
				    const typename DoFApplication<dim>::CellInfo& info)
{
  const FEValuesBase<dim>& fe = info.fe(0);
  const unsigned int n_quad = fe.n_quadrature_points;
  std::vector<Point<dim> > point(n_quad); //creates vector with n_quadrature_point components
  point = fe.get_quadrature_points();
  std::vector<double> values_0(n_quad);
  std::vector<double> values_1(n_quad);
  
  
  for (unsigned int p=0; p<n_quad; ++p)
    {
      Point<dim> point_1, point_2;
      point_1(0) = 0.5;
      point_2(0) = -0.5;
      
      if (((point[p]-point_1).square() < 0.2*0.2) ||
	  ((point[p]-point_2).square() < 0.2*0.2))
	values_0[p] = 1;
      else
	values_0[p] = 0;

      if (point[p].square() < 0.2*0.2)
	values_1[p] = 1;
      else
	values_1[p] = 0;
    }
      
  //
  Assert (cell_vector.n_blocks() == dim,
	  ExcDimensionMismatch (cell_vector.n_blocks(), dim));
  
  //create the BlockVector solution:
  AppShop::Residual::Cell::mass(cell_vector.block(0), values_0, fe, 1.0);
  AppShop::Residual::Cell::mass(cell_vector.block(1), values_1, fe, 1.0);
  
}

template <int dim>
void
TestApplication<dim>::dirichlet_bc(std::map<unsigned int, double>& boundary_values) const
{
  /* std::vector<bool> comp_mask (this->element->n_blocks());
  comp_mask[0]= true; 
  comp_mask[1]= true; */
  VectorTools::interpolate_boundary_values (*this->dof,
					    0,
					    ZeroFunction<dim>(this->element->n_blocks()),
					    boundary_values);
}

template <int dim>
void
TestApplication<dim>::solve(FEVector& start, const FEVectors& rhs)
{
  if (this->notifications.any())
    {
      this->assemble(rhs);
      this->clear_events();
    }
  /*for (unsigned int i = 0; i<this->matrix.m(); ++i)
    deallog<<"Diagonal element in row "<<i<<" is "<<this->matrix.diag_element(i)<<std::endl;*/
  
  //--------------------------------
  // System matrix that will be used to solve the large system of eq.
  BlockMatrixArray<double> system_matrix;
  system_matrix.initialize(2,2, this->data->vector_pool);
  
  std::vector<PreconditionSSOR< > >
    preconditioner_matrix(system_matrix.n_block_rows());
				   // our preconditioners do not have
				   // vmult_add, so we need an
				   // auxiliary class
  std::vector<PointerMatrixAux<PreconditionSSOR< >, Vector<double> > >
    preconditioner_pointer(system_matrix.n_block_rows());
  for (unsigned int i=0; i<system_matrix.n_block_rows(); ++i)
    {
      preconditioner_matrix[i].initialize(this->matrix.block(i,i));
      preconditioner_pointer[i].set_memory(&this->data->vector_pool);
      preconditioner_pointer[i] = &preconditioner_matrix[i];
    }
  
  BlockTrianglePrecondition<double> preconditioner;
  preconditioner.initialize(system_matrix.n_block_rows(),
			    this->data->vector_pool);
  
				   // The use of system matrix seems
				   // obsolete now, but maybe we can
				   // save assembling a few blocks?
  for (unsigned int i=0; i<system_matrix.n_block_rows(); ++i)
    for (unsigned int j=0; j<system_matrix.n_block_cols(); ++j)
      {
	system_matrix.enter(this->matrix.block(i,j), i, j);
	if (i==j)
	  preconditioner.enter(preconditioner_pointer[i], i, i);
	else
	  preconditioner.enter(this->matrix.block(i,j), i, j);
      }
  
  // Now, solve:
  SolverControl control(1000, 1e-16);
  SolverBicgstab<FEVector > gmres (control, this->data->block_vector_pool);
  gmres.solve(this->filtered_matrix, start, rhs.vector(0), preconditioner);
  //Return the solution to all the nodes taking into accout the hanging nodes
  this->hanging_node_constraints.distribute(start);

  //Plot solution out if necessary:
  deallog<<"Solution: "<<std::endl;
  unsigned int index(0);
  for (unsigned int i=0; i<start.block(0).size(); ++i)
	  for (unsigned int j = 0; j<start.n_blocks(); ++j)
	  {
		  deallog<<"U displacement at dof "<<index<<" : "<<start.block(j)(i)<<std::endl;
		  index = index + 1;
	  }
}

template <int dim>
double
TestApplication<dim>::estimate(const FEVector& sol)
{
  
  this->cell_errors.reinit(this->tr->n_active_cells());
  typename FunctionMap<dim>::type newmann_boundary;
  KellyErrorEstimator<dim>::estimate(*this->dof,
				     QGauss<dim-1>(3),
				     newmann_boundary,
				     sol,
				     this->cell_errors);
  
  return 0;
}

template <int dim>
double
TestApplication<dim>::evaluate(const FEVectors&)
{
  
  return 0;
  
}


int
main(int argc, char** argv)
{
  // initialize file_utils routines and deallog
  AppFrame::initialize(argc, argv);
  AppFrame::open_logfile();
  deallog << "Program:" << argv[0] << std::endl;
  
  ParameterHandler param;
  TestApplication<2> app;
  deallog << std::endl;

  AppFrame::declare_parameter_files(param);
  app.declare_parameters(param);
  AppFrame::read_parameter_files(param);
  app.initialize(param);

  FEVector u;
  FEVector rhs;
  
  //unsigned int cycle=0;
  for (unsigned int cycle=0; cycle<8; ++cycle)
    {
      deallog<< "Cycle "<<cycle<<" : "<<std::endl;
      std::ostringstream streamOut;
      streamOut << cycle;
      std::string filename = "step-8-grid-"+streamOut.str();
      std::string sol_filename = "sol-step-8-grid-"+streamOut.str();
      
      if (cycle == 0)
	{
	  app.remesh_dofs();
	  app.remesh_matrices();
	}
      else
	{
	  app.estimate(u);
	  app.remesh();
	}
      
      // initialize vectors with new grid
      app.init_vector(u);
      app.init_vector(rhs);
      FEVectors data;
      data.add_vector(rhs, "main");
      
      // build system of eq
      app.assemble(data);
      // build rhs
      app.residual(rhs, data);
      
      try
	{
	  app.solve(u, data);
	  app.grid_out(filename);
	  deallog << "u:" << u.n_blocks() << ':';
	  for (unsigned int b=0;b<u.n_blocks();++b)
	    deallog << ' ' << u.block(b).size();
	  deallog << std::endl;
	  
//	  app.data_out(sol_filename,u);
	}
      catch (const std::exception& e)
	{
	  deallog << e.what() << std::endl;
	}
    }
}
