//---------------------------------------------------------------------------
//    $Id: matrix_02.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#define APPNAME Matrix02

// Test if cell blocks are properly assembled into matrix blocks

#include <appframe/matrix_application.h>
#include <appframe/mesh_loop.h>
#include <base/parameter_handler.h>
#include <lac/block_vector.h>
#include <grid/grid_generator.h>

using namespace AppFrame;

template <int dim>
class APPNAME :
  public MatrixApplication<dim>
{
  public:
    APPNAME ();
    void _initialize();
    virtual void initialize(ParameterHandler& param);
    virtual void cell_matrix(MatrixVector& cell_matrices,
			     const typename DoFApplication<dim>::CellInfo& cell);  
    virtual void cell_residual(FEVector& cell_vector,
			       const typename DoFApplication<dim>::CellInfo& cell);
    virtual void solve(FEVector& start, const FEVectors& rhs);
    virtual double estimate (const FEVectors& src);
    virtual double evaluate (const FEVectors& src);
};


template <int dim>
APPNAME<dim>::APPNAME()
{
  deallog << "->Matrix02-" << dim;
  this->cell_couplings.reinit(3, 3);
  this->cell_couplings(0,0) = DoFTools::always;
  this->cell_couplings(2,0) = DoFTools::always;
  this->cell_couplings(2,1) = DoFTools::always;
  this->boundary_fluxes = false;
  this->interior_fluxes = false;

  this->add_matrix(0,0);
  this->add_matrix(2,0);
  this->add_matrix(2,0);
  this->add_matrix(2,1);
}


template <int dim>
void
APPNAME<dim>::initialize(ParameterHandler& param)
{
  MatrixApplication<dim>::initialize(param);
  _initialize();
}


template <int dim>
void
APPNAME<dim>::_initialize()
{
  GridGenerator::hyper_cube(*this->tr);
  this->remesh_dofs();
  this->remesh_matrices();
}


template <int dim>
void
APPNAME<dim>::cell_matrix(
  MatrixVector& cell_matrices,
  const typename DoFApplication<dim>::CellInfo&)
{
  for (unsigned int m=0;m<cell_matrices.size();++m)
    {
      Table<2, double>& M = cell_matrices[m].matrix;
      const unsigned int iblock = cell_matrices[m].row;
      const unsigned int jblock = cell_matrices[m].column;
      M.fill((double)(iblock+1)*(jblock+1));
    }
}



template <int dim>
void
APPNAME<dim>::cell_residual(
  FEVector& cell_vector,
  const typename DoFApplication<dim>::CellInfo&)
{
  for (unsigned int b=0;b<cell_vector.n_blocks();++b)
    {
      cell_vector.block(b) = b+1;
    }
}


template <int dim>
void
APPNAME<dim>::solve(FEVector& start, const FEVectors& rhs)
{
  if (this->notifications.any())
    {
      this->assemble(rhs);
      this->clear_events();
    }
  
  start = rhs.vector(0);
}



template <int dim>
double
APPNAME<dim>::evaluate(const FEVectors& srcs)
{
  const FEVector& src = srcs.vector(0);
  for (unsigned int m=0;m<this->matrices.size();++m)
    {
      deallog << "Matrix " << m << "  ("
	      << this->matrices[m]->row << ','
	      << this->matrices[m]->column << ')' << std::endl;
      this->matrices[m]->matrix.print_formatted(std::cout, 0, false, 4);
    }
  deallog << std::endl;
  src.print(std::cout, 0, false, false);
  return 0;
}


template <int dim>
double
APPNAME<dim>::estimate(const FEVectors&)
{
  return 0;
}


int
main(int argc, char** argv)
{
  AppFrame::initialize(argc, argv);
  AppFrame::open_logfile();
  deallog << "Program:" << argv[0] << std::endl;
  
  ParameterHandler param;
  APPNAME<2> mapp;
  MeshLoop app(mapp);
  deallog << std::endl;

  app.declare_parameters(param);
  read_parameter_files(param);
  app.initialize(param);
  
  FEVector u;
  FEVectors dummy;
  app.start_vector(u, "Test");
  
  try
    {
      app.solve(u, dummy);
    }
  catch (const std::exception& e)
    {
      deallog << e.what() << std::endl;
    }
}
