//---------------------------------------------------------------------------
//    $Id: matrix_03.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#define APPNAME Matrix03

// Test if cells and faces are properly assembled into matrix blocks

#include <appframe/matrix_application.h>
#include <appframe/mesh_loop.h>
#include <base/parameter_handler.h>
#include <lac/block_vector.h>
#include <grid/grid_generator.h>
#include <fe/fe_values.h>

using namespace AppFrame;

template <int dim>
class APPNAME :
  public MatrixApplication<dim>
{
  public:
    APPNAME ();
    void _initialize();
    virtual void initialize(ParameterHandler& param);
    virtual void cell_matrix(MatrixVector& cell_matrices,
			     const typename DoFApplication<dim>::CellInfo& cell);  
    virtual void cell_residual(FEVector& cell_vector,
			       const typename DoFApplication<dim>::CellInfo& cell);
    virtual void bdry_residual(FEVector& cell_vector,
			       const typename DoFApplication<dim>::FaceInfo& cell);
    virtual void face_residual(FEVector& face_vector1,
			       FEVector& face_vector2,
			       const typename DoFApplication<dim>::FaceInfo& face1,
			       const typename DoFApplication<dim>::FaceInfo& face2);
    virtual void solve(FEVector& start, const FEVectors& rhs);
    virtual double evaluate (const FEVectors& src);
};


template <int dim>
APPNAME<dim>::APPNAME()
{
  deallog << "->Matrix03-" << dim;
  this->cell_couplings.reinit(3, 3);
  this->cell_couplings(0,0) = DoFTools::always;
  this->cell_couplings(2,0) = DoFTools::always;
  this->cell_couplings(2,1) = DoFTools::always;
  this->flux_couplings.reinit(3, 3);
  this->cell_couplings(0,0) = DoFTools::nonzero;
  this->cell_couplings(2,0) = DoFTools::nonzero;
  this->cell_couplings(2,1) = DoFTools::nonzero;

  this->add_matrix(0,0);
  this->add_matrix(2,0);
  this->add_matrix(2,0);
  this->add_matrix(2,1);
}


template <int dim>
void
APPNAME<dim>::initialize(ParameterHandler& param)
{
  MatrixApplication<dim>::initialize(param);
  _initialize();
}


template <int dim>
void
APPNAME<dim>::_initialize()
{
  GridGenerator::hyper_cube(*this->tr);
  this->remesh_dofs();
  this->remesh_matrices();
}


template <int dim>
void
APPNAME<dim>::cell_matrix(
  MatrixVector& cell_matrices,
  const typename DoFApplication<dim>::CellInfo& info)
{
  ((Table<2,double>&) cell_matrices[0].matrix).fill((double)info.cell->index());
  ((Table<2,double>&) cell_matrices[1].matrix).fill((double)info.cell->index());
}



template <int dim>
void
APPNAME<dim>::cell_residual(
  FEVector& cell_vector,
  const typename DoFApplication<dim>::CellInfo& info)
{
  cell_vector.block(0) = info.cell->index();
}


template <int dim>
void
APPNAME<dim>::bdry_residual(
  FEVector& cell_vector,
  const typename DoFApplication<dim>::FaceInfo& info)
{
  const FiniteElement<dim>& fe = info.fe(0).get_fe();
  for (unsigned int i=0;i<fe.dofs_per_cell;++i)
    if (fe.has_support_on_face(i, info.face_number))
      {
	cell_vector.block(1)(i) += info.face->index();
	cell_vector.block(2)(i) += 1;
      }
}


template <int dim>
void
APPNAME<dim>::face_residual(FEVector& face_vector1,
				 FEVector& face_vector2,
				 const typename DoFApplication<dim>::FaceInfo& face1,
				 const typename DoFApplication<dim>::FaceInfo& face2)
{
  const FiniteElement<dim>& fe1 = face1.fe(0).get_fe();
  const FiniteElement<dim>& fe2 = face2.fe(0).get_fe();
  for (unsigned int i=0;i<fe1.dofs_per_cell;++i)
    if (fe1.has_support_on_face(i, face1.face_number))
      {
	  face_vector1.block(1)(i) += face1.face->index();
	  face_vector1.block(2)(i) += 1;
      }
  for (unsigned int i=0;i<fe2.dofs_per_cell;++i)
    if (fe2.has_support_on_face(i, face2.face_number))
      {
	face_vector2.block(1)(i) += face2.face->index();
	face_vector2.block(2)(i) += 1;
      }
}


template <int dim>
void
APPNAME<dim>::solve(FEVector& start, const FEVectors& rhs)
{
  if (this->notifications.any())
    {
      this->assemble(rhs);
      this->clear_events();
    }
  start = rhs.vector(0);
}



template <int dim>
double
APPNAME<dim>::evaluate(const FEVectors& srcs)
{
  const FEVector& src = srcs.vector(0);
  for (unsigned int m=0;m<this->matrices.size();++m)
    {
      deallog << "Matrix " << m << "  ("
	      << this->matrices[m]->row << ','
	      << this->matrices[m]->column << ')' << std::endl;
      const SparseMatrix<double>& M = this->matrices[m]->matrix;
      
      if (M.m() <= 25)
	M.print_formatted(std::cout, 0, false, 4);
      else
	M.print(std::cout);
    }
  deallog << std::endl;
  src.print(std::cout, 0, false, false);
  return 0;
}


int
main(int argc, char** argv)
{
  AppFrame::initialize(argc, argv);
  AppFrame::open_logfile();
  deallog << "Program:" << argv[0] << std::endl;
  
  ParameterHandler param;
  APPNAME<2> mapp;
  MeshLoop app(mapp);
  deallog << std::endl;

  app.declare_parameters(param);
  read_parameter_files(param);
  app.initialize(param);
  
  FEVector u;
  FEVectors dummy;
  app.start_vector(u, "Test");
  
  try
    {
      app.solve(u, dummy);
    }
  catch (const std::exception& e)
    {
      deallog << e.what() << std::endl;
    }
}
