//---------------------------------------------------------------------------
//    $Id: matrix_04.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#define APPNAME Matrix04

// Test filling of local function values, scalar case

#include <appframe/matrix_application.h>
#include <appframe/mesh_loop.h>
#include <base/parameter_handler.h>
#include <base/function_lib.h>
#include <lac/block_vector.h>
#include <grid/grid_generator.h>
#include <fe/fe_values.h>
#include <numerics/vectors.h>

using namespace AppFrame;

template <int dim>
class APPNAME :
  public MatrixApplication<dim>
{
  public:
    APPNAME ();
    void _initialize();
    virtual void initialize(ParameterHandler& param);
    virtual void init_vector (FEVector& v) const;
    virtual void cell_matrix(MatrixVector&, const typename DoFApplication<dim>::CellInfo&);  
    virtual void bdry_matrix(MatrixVector&, const typename DoFApplication<dim>::FaceInfo&);
    virtual void cell_residual(FEVector& cell_vector,
			       const typename DoFApplication<dim>::CellInfo& cell);
    virtual void bdry_residual(FEVector& cell_vector,
			       const typename DoFApplication<dim>::FaceInfo& cell);
    virtual void face_residual(FEVector& face_vector1,
			       FEVector& face_vector2,
			       const typename DoFApplication<dim>::FaceInfo& face1,
			       const typename DoFApplication<dim>::FaceInfo& face2);
    virtual void solve(FEVector& start, const FEVectors& rhs);
    virtual double evaluate (const FEVectors& src);
};


template <int dim>
APPNAME<dim>::APPNAME()
{
  deallog << "->Matrix02-" << dim;
  this->cell_couplings.reinit(1, 1);
  this->cell_couplings(0,0) = DoFTools::always;
  this->interior_fluxes = false;  
  this->add_matrix(0,0);
}


template <int dim>
void
APPNAME<dim>::initialize(ParameterHandler& param)
{
  MatrixApplication<dim>::initialize(param);
  _initialize();
}


template <int dim>
void
APPNAME<dim>::init_vector (
  FEVector& v) const
{
  DoFApplication<dim>::init_vector(v);
  Functions::SquareFunction<dim> solution;
  VectorTools::interpolate(*this->dof, solution, v);
}

  
template <int dim>
void
APPNAME<dim>::_initialize()
{
  GridGenerator::hyper_cube(*this->tr);
  this->remesh_dofs();
  this->remesh_matrices();
}


template <int dim>
void
APPNAME<dim>::cell_matrix(MatrixVector&, const typename DoFApplication<dim>::CellInfo&)
{}



template <int dim>
void
APPNAME<dim>::bdry_matrix(MatrixVector&, const typename DoFApplication<dim>::FaceInfo&)
{}



template <int dim>
void
APPNAME<dim>::cell_residual(
  FEVector&,
  const typename DoFApplication<dim>::CellInfo& info)
{
  const FEValuesBase<dim>& fe = info.fe(0);
  
  Functions::SquareFunction<dim> solution;
  std::vector<double> f(fe.n_quadrature_points);
  solution.value_list(fe.get_quadrature_points(), f);
  
  const std::vector<std::vector<double> >& u = info.values[0];
  
  for (unsigned int i=0;i<fe.n_quadrature_points;++i)
    deallog << "qpt-cell "
//	    << fe.quadrature_point(i)
	    << '\t' << 1.e-6 + f[i] - u[0][i]
	    << std::endl;  
}


template <int dim>
void
APPNAME<dim>::bdry_residual(
  FEVector&,
  const typename DoFApplication<dim>::FaceInfo& info)
{
  const FEValuesBase<dim>& fe = info.fe(0);
  
  Functions::SquareFunction<dim> solution;
  std::vector<double> f(fe.n_quadrature_points);
  solution.value_list(fe.get_quadrature_points(), f);
  
  const std::vector<std::vector<double> >& u = info.values[0];
  
  for (unsigned int i=0;i<fe.n_quadrature_points;++i)
    deallog << "qpt-bdry "
//	    << fe.quadrature_point(i)
	    << '\t' << 1.e-6 + f[i] - u[0][i]
	    << std::endl;  
}


template <int dim>
void
APPNAME<dim>::face_residual(FEVector&,
				 FEVector&,
				 const typename DoFApplication<dim>::FaceInfo& face1,
				 const typename DoFApplication<dim>::FaceInfo& face2)
{
  const FEValuesBase<dim>& fe1 = face1.fe(0);
//  const FEValuesBase<dim>& fe2 = face2.fe(0);

  Functions::SquareFunction<dim> solution;
  std::vector<double> f(fe1.n_quadrature_points);
  solution.value_list(fe1.get_quadrature_points(), f);
  
  const std::vector<std::vector<double> >& u1 = face1.values[0];
  const std::vector<std::vector<double> >& u2 = face2.values[0];
  
  for (unsigned int i=0;i<fe1.n_quadrature_points;++i)
    deallog << "qpt-face "
//	    << fe1.quadrature_point(i)
	    << '\t' << 1.e-6 + f[i] - u1[0][i]
	    << '\t' << 1.e-6 + f[i] - u2[0][i]
	    << std::endl;
}


template <int dim>
void
APPNAME<dim>::solve(FEVector& start, const FEVectors& rhs)
{
  if (this->notifications.any())
    {
      this->assemble(rhs);
      this->clear_events();
    }
  start = rhs.vector(0);
}



template <int dim>
double
APPNAME<dim>::evaluate(const FEVectors&)
{
  return 0;
}


int
main(int argc, char** argv)
{
  AppFrame::initialize(argc, argv);
  AppFrame::open_logfile();
  deallog << "Program:" << argv[0] << std::endl;
  
  ParameterHandler param;
  APPNAME<2> mapp;
  MeshLoop app(mapp);
  deallog << std::endl;

  app.declare_parameters(param);
  read_parameter_files(param);
  app.initialize(param);
  
  FEVector u;
  FEVectors dummy;
  app.start_vector(u, "Test");
  app.get_data()->enter("Test", u);
  
  try
    {
      app.solve(u, dummy);
    }
  catch (const std::exception& e)
    {
      deallog << e.what() << std::endl;
    }
}
