//---------------------------------------------------------------------------
//    $Id: shop_cell_01.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2007 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

// Test cell matrices and residuals for grad and div operators

#include <appframe/residual_shop_cell.h>
#include <appframe/residual_shop_bdry.h>
#include <appframe/matrix_shop_cell.h>
#include <appframe/matrix_shop_bdry.h>
#include <appframe/matrix_base.h>
#include <base/quadrature_lib.h>
#include <lac/block_vector.h>
#include <grid/grid_generator.h>
#include <grid/tria_iterator.h>
#include <grid/tria_accessor.h>
#include <fe/fe_dgp.h>
#include <fe/fe_q.h>
#include <fe/fe_system.h>
#include <fe/fe_raviart_thomas.h>

#include <iostream>

using namespace AppFrame;


template<int dim>
void div(const FEValuesBase<dim>& fe,
	 const FEValuesBase<dim>& fe_test,
	 FEFaceValues<dim>& face,
	 FEFaceValues<dim>& face_test)
{
  deallog.push("div");
  deallog << fe.get_fe().get_name() << " <-> "
	  << fe_test.get_fe().get_name() << std::endl;
  
  const unsigned int n_dofs = fe.get_fe().dofs_per_cell;
  
				   // Index vector needed for
				   // get_function_values etc.
  std::vector<unsigned int> indices(n_dofs*dim/fe.get_fe().n_components());
  for (unsigned int i=0;i<indices.size();++i)
    indices[i] = i;

				   // The local finite element vector
  Vector<double> function(indices.size());
  Vector<double> result(fe_test.get_fe().dofs_per_cell);
  Vector<double> result2(fe_test.get_fe().dofs_per_cell);

  std::vector<std::vector<double> >
    values(dim, std::vector<double>(fe.n_quadrature_points));
  std::vector<std::vector<Tensor<1,dim> > >
    gradients(dim, std::vector<Tensor<1,dim> >(fe.n_quadrature_points));
  std::vector<std::vector<double> >
    face_values(dim, std::vector<double>(face.n_quadrature_points));
  
  deallog << "matrix" << std::endl;
  std::vector<MatrixBlock<FullMatrix<double> > > M(dim);
  std::vector<unsigned int> Mi(dim/fe.get_fe().n_components());
  for (unsigned int d=0;d<dim;++d)
    {
      Mi[d] = d;
      M[d].column = d;
      M[d].matrix.reinit(result.size(), n_dofs);
    }
				   // Compute divergence matrix
  AppShop::Matrix::Cell::div(M, Mi, fe, fe_test);
				   // Since the cell term is in weak
				   // form, we need the boundaries as
				   // well
  for (unsigned int f=0;f<GeometryInfo<dim>::faces_per_cell;++f)
    {
      face.reinit(face.get_cell(), f);
      face_test.reinit(face.get_cell(), f);
      AppShop::Matrix::Boundary::u_dot_n(M, Mi, face, face_test);
    }
				   // Write matrix, blocks on top of
				   // each other
  for (unsigned int d=0;d<Mi.size();++d)
    for (unsigned int i=0;i<n_dofs;++i)
      {
	for (unsigned int j=0;j<result.size();++j)
	  deallog << ' ' << M[d].matrix(j,i);
	deallog << std::endl;
      }

				   // Now check residuals. We compute
				   // the divergence of each basis
				   // function tested with the other
				   // space. In the end, this yields a
				   // matrix which should be the same
				   // as above.

				   // The output is the difference to
				   // the matrix, row by row, and
				   // should be zero for weak and strong
  deallog << "Residuals" << std::endl;
  
  for (unsigned int i=0;i<function.size();++i)
    {
      function = 0.;
      function(i) = 1.;
				       // The values of  the finite
				       // element function in the
				       // quadrature points
      fe.get_function_values(function, indices, values, true);
      fe.get_function_grads(function, indices, gradients, true);

				       // First, the strong divergence
      result = 0.;
      AppShop::Residual::Cell::div(result, make_slice(gradients, 0, dim), fe, fe_test);
      for (unsigned int j=0;j<result.size();++j)
	deallog << " " << result(j);

				       // The weak divergence
      result2 = 0.;
      AppShop::Residual::Cell::div(result2, values, fe, fe_test);
      for (unsigned int f=0;f<GeometryInfo<dim>::faces_per_cell;++f)
	{
	  face.reinit(face.get_cell(), f);
	  face_test.reinit(face.get_cell(), f);
	  face.get_function_values(function, indices, face_values, true);
	  AppShop::Residual::Boundary::u_dot_n(result2, face_values, face, face_test);
	}
	result.add(-1., result2);
      deallog << "    strong-weak " << result.l2_norm();
      deallog << std::endl;
    }
  deallog.pop();
}


template<int dim>
void grad(const FEValuesBase<dim>& fe,
	  const FEValuesBase<dim>& fe_test,
	  FEFaceValues<dim>& face,
	  FEFaceValues<dim>& face_test)
{
  deallog.push("grad");
  deallog << fe.get_fe().get_name() << " <-> "
	  << fe_test.get_fe().get_name() << std::endl;
  
  const unsigned int n_dofs = fe.get_fe().dofs_per_cell;
  
				   // Index vector needed for
				   // get_function_values etc.
  std::vector<unsigned int> indices(n_dofs);
  for (unsigned int i=0;i<n_dofs;++i)
    indices[i] = i;

				   // The local finite element vector
  Vector<double> function(n_dofs);
  Vector<double> result(fe_test.get_fe().dofs_per_cell * dim/fe_test.get_fe().n_components());

  std::vector<double> values(fe.n_quadrature_points);
  std::vector<Tensor<1,dim> > gradients(fe.n_quadrature_points);
  std::vector<double> face_values(face.n_quadrature_points);
  
  for (unsigned int i=0;i<n_dofs;++i)
    {
      function = 0.;
      function(i) = 1.;
      deallog << "InDofs  ";
      for (unsigned int i=0;i<function.size();++i)
	deallog << ' ' << function(i);
      deallog << std::endl;
				       // The values of  the finite
				       // element function in the
				       // quadrature points
      fe.get_function_values(function, indices, values);
      fe.get_function_grads(function, indices, gradients);
      
				       // First, the strong gradient
      result = 0.;
      AppShop::Residual::Cell::grad(result, gradients, fe, fe_test);
      deallog << "strong  ";
      for (unsigned int i=0;i<result.size();++i)
	deallog << ' ' << result(i);
      deallog << std::endl;

				       // The weak gradient
      result = 0.;
      AppShop::Residual::Cell::grad(result, values, fe, fe_test);
      for (unsigned int f=0;f<GeometryInfo<dim>::faces_per_cell;++f)
	{
	  face.reinit(face.get_cell(), f);
	  face_test.reinit(face.get_cell(), f);
	  face.get_function_values(function, indices, face_values);
	  AppShop::Residual::Boundary::u_times_n(result, face_values, face, face_test);
	}
      
      deallog << "weak    ";
      for (unsigned int i=0;i<result.size();++i)
	deallog << ' ' << result(i);
      deallog << std::endl;
    }
  deallog.pop();
}


template<int dim>
void test_element(const typename Triangulation<dim>::cell_iterator cell,
		  FiniteElement<dim>& el,
		  FiniteElement<dim>& el_test)
{
  QGauss<dim> quadrature(el.degree+1);
  QGauss<dim-1> face_quadrature(el.degree+1);
  FEValues<dim> fe(el, quadrature,
		   UpdateFlags(update_values | update_gradients));
  FEValues<dim> fe_test(el_test, quadrature,
			UpdateFlags(update_values | update_gradients));
  FEFaceValues<dim> fe_face(el, face_quadrature,
			    UpdateFlags(update_values
					| update_JxW_values
					| update_normal_vectors));
  FEFaceValues<dim> fe_face_test(el_test, face_quadrature,
				 UpdateFlags(update_values
					     | update_JxW_values
					     | update_normal_vectors));
  fe.reinit(cell);
  fe_test.reinit(cell);
  fe_face.reinit(cell, 0);
  fe_face_test.reinit(cell, 0);
  
  grad(fe, fe_test, fe_face, fe_face_test);
  div(fe_test, fe, fe_face_test, fe_face);
}

template<int dim>
void test()
{
  deallog.threshold_double(1.e-12);
  
  Triangulation<dim> tr;
  GridGenerator::hyper_cube(tr, 0, 1);
  
  FE_Q<dim> q1(1);
  
  FE_DGP<dim> dgp1(1);
  FE_DGP<dim> dgp2(2);
  FE_DGP<dim> dgp3(3);

  FESystem<dim> sq1(q1,dim);
  FESystem<dim> sdgp1(dgp1,dim);
  FESystem<dim> sdgp2(dgp2,dim);

  FE_RaviartThomas<dim> rt1(1);
  
  test_element(tr.begin_active(), dgp1, sdgp1);
  test_element(tr.begin_active(), dgp2, sdgp1);
  test_element(tr.begin_active(), dgp1, sdgp2);
  test_element(tr.begin_active(), q1, sq1);
  test_element(tr.begin_active(), q1, rt1);
}


int main()
{
  test<2>();
}
