//---------------------------------------------------------------------------
//    $Id: shop_face_01.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

// Test face matrices and residuals for normal jump operators

#include <appframe/residual_shop_face.h>
#include <appframe/matrix_shop_face.h>
#include <appframe/matrix_base.h>
#include <base/quadrature_lib.h>
#include <lac/block_vector.h>
#include <grid/grid_generator.h>
#include <grid/tria_iterator.h>
#include <grid/tria_accessor.h>
#include <fe/fe_dgp.h>
#include <fe/fe_dgq.h>
#include <fe/fe_system.h>
#include <fe/fe_raviart_thomas.h>

#include <iostream>

using namespace AppFrame;


// Check consistency of
//   Matrix::Face::u_ni
//   Residual::Face::u_ni
//   Residual::Face::u_dot_n
template<int dim>
void
test_scalar_face(const FEFaceValues<dim>& fe1,
		 const FEFaceValues<dim>& fe2,
		 const FEFaceValues<dim>& fetest1,
		 const FEFaceValues<dim>& fetest2)
{  
  const unsigned int n_dofs = fe1.get_fe().dofs_per_cell;
  const unsigned int t_dofs = fetest1.get_fe().dofs_per_cell;
  const unsigned int n_q = fe1.n_quadrature_points;
  
  std::vector<std::vector<FullMatrix<double> > > m(dim);
  for (unsigned int d=0;d<dim;++d)
    {
      m[d].resize(4);
      for (unsigned int i=0;i<4;++i)
	m[d][i].reinit(t_dofs, n_dofs);
    }
  std::vector<double> qu1(n_q);
  std::vector<double> qu2(n_q);

  Vector<double> u1(n_dofs);
  Vector<double> u2(n_dofs);
  Vector<double> v1(t_dofs);
  Vector<double> v2(t_dofs);
  Vector<double> w1(t_dofs);
  Vector<double> w2(t_dofs);
  
  std::vector<unsigned int> indices(n_dofs);
  for (unsigned int i=0;i<n_dofs;++i)
    indices[i] = i;
  std::vector<unsigned int> indicestest(t_dofs);
  for (unsigned int i=0;i<t_dofs;++i)
    indicestest[i] = i;
  
//  std::vector<double> factor(fe1.n_quadrature_points, 1.);
  
  for (unsigned int d=0;d<dim;++d)
    {
      AppShop::Matrix::Face::u_ni(m[d][0], m[d][1], m[d][2], m[d][3],
				  fe1, fe2, fetest1, fetest2, d);
    }
  for (unsigned int d=0;d<dim;++d)
    for (unsigned int i1=0;i1<n_dofs;++i1)
      for (unsigned int i2=0;i2<n_dofs;++i2)
	{
	  u1.reinit(n_dofs);
	  u1(i1) = 9.;
	  u2.reinit(n_dofs);
	  u2(i2) = 9.;
	  
	  m[d][0].vmult(v1, u1);
	  m[d][3].vmult(v2, u2);
	  m[d][1].vmult_add(v1, u2);
	  m[d][2].vmult_add(v2, u1);
	  
//	  std::cout << "mult" << std::endl;
//	  v1.print("%14.5e");
//	  v2.print("%14.5f");
	  
	  fe1.get_function_values(u1, indices, qu1);
	  fe2.get_function_values(u2, indices, qu2);

	  AppShop::Residual::Face::u_ni(w1, w2, qu1, qu2, d, fe1, fe2, fetest1, fetest2, -1.);
	  w1 += v1;
	  w2 += v2;
	  if (w1.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << v2.linfty_norm() << std::endl;
	  if (w2.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << w2.linfty_norm() << std::endl;

	  w1 = 0.;
	  w2 = 0.;
	  
	  std::vector<std::vector<double> > dqu1(dim, std::vector<double>(n_q, 0.));
	  std::vector<std::vector<double> > dqu2(dim, std::vector<double>(n_q, 0.));
	  for (unsigned int k=0;k<n_q;++k)
	    {
	      dqu1[d][k] = qu1[k];
	      dqu2[d][k] = qu2[k];	      
	    }
	  AppShop::Residual::Face::u_dot_n(w1, w2, dqu1, dqu2, fe1, fe2, fetest1, fetest2, -1.);
	  w1 += v1;
	  w2 += v2;

	  if (w1.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << v2.linfty_norm() << std::endl;
	  if (w2.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << w2.linfty_norm() << std::endl;	  
	}
}


// Check consistency of
//   Matrix::Face::u_ni
//   Residual::Face::u_ni
//   Residual::Face::u_dot_n
template<int dim>
void
test_vector_face1(const FEFaceValues<dim>& fe1,
		  const FEFaceValues<dim>& fe2,
		  const FEFaceValues<dim>& fetest1,
		  const FEFaceValues<dim>& fetest2)
{  
  const unsigned int n_dofs = fe1.get_fe().dofs_per_cell;
  const unsigned int t_dofs = fetest1.get_fe().dofs_per_cell;
  const unsigned int n_q = fe1.n_quadrature_points;
  
  std::vector<std::vector<FullMatrix<double> > > m(dim);
//  std::vector<FullMatrix<double> > msys(4, FullMatrix<double>(t_dofs, n_dofs));
  for (unsigned int d=0;d<dim;++d)
    {
      m[d].resize(4);
      for (unsigned int i=0;i<4;++i)
	m[d][i].reinit(t_dofs, n_dofs);
    }
  std::vector<std::vector<double> > qu1(dim, std::vector<double>(n_q));
  std::vector<std::vector<double> > qu2(dim, std::vector<double>(n_q));

  Vector<double> u1(n_dofs);
  Vector<double> u2(n_dofs);
  Vector<double> v1(t_dofs);
  Vector<double> v2(t_dofs);
  Vector<double> w1(t_dofs);
  Vector<double> w2(t_dofs);
  
  std::vector<unsigned int> indices(n_dofs);
  for (unsigned int i=0;i<n_dofs;++i)
    indices[i] = i;
  std::vector<unsigned int> indicestest(t_dofs);
  for (unsigned int i=0;i<t_dofs;++i)
    indicestest[i] = i;
  
//  std::vector<double> factor(fe1.n_quadrature_points, 1.);
  
  for (unsigned int d=0;d<dim;++d)
    AppShop::Matrix::Face::u_ni(m[d][0], m[d][1], m[d][2], m[d][3],
				fe1, fe2, fetest1, fetest2, d);
  
  for (unsigned int i1=0;i1<n_dofs;++i1)
    for (unsigned int i2=0;i2<n_dofs;++i2)
      {
	u1.reinit(n_dofs);
	u1(i1) = 9.;
	u2.reinit(n_dofs);
	u2(i2) = 9.;

	v1 = 0.;
	v2 = 0.;
	for (unsigned int d=0;d<dim;++d)
	  {
	    m[d][0].vmult_add(v1, u1);
	    m[d][3].vmult_add(v2, u2);
	    m[d][1].vmult_add(v1, u2);
	    m[d][2].vmult_add(v2, u1);
	  }
//	  std::cout << "mult" << std::endl;
//	  v1.print("%14.5e");
//	  v2.print("%14.5f");
	  
	  fe1.get_function_values(u1, indices, qu1, true);
	  fe2.get_function_values(u2, indices, qu2, true);
	  
	  for (unsigned int d=0;d<dim;++d)
	    AppShop::Residual::Face::u_ni(w1, w2, qu1[d], qu2[d], d, fe1, fe2, fetest1, fetest2, -1.);

	  deallog << v1.size() << std::endl;
// 	  for (unsigned int d=0;d<n_dofs;++d)
// 	    deallog << v1(d) << '\t' << w1(d) << std::endl;
	  
	  w1 += v1;
	  w2 += v2;
	  if (w1.linfty_norm() > 1.e-15)
	    deallog << "Error norm 1 " << v1.linfty_norm() << std::endl;
	  if (w2.linfty_norm() > 1.e-15)
	    deallog << "Error norm 2 " << v1.linfty_norm() << std::endl;

	  w1 = 0.;
	  w2 = 0.;
	  
	  AppShop::Residual::Face::u_dot_n(w1, w2, qu1, qu2, fe1, fe2, fetest1, fetest2, -1.);
	  w1 += v1;
	  w2 += v2;
	  if (w1.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << w1.linfty_norm() << std::endl;
	  if (w2.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << w2.linfty_norm() << std::endl;
	}
}


// Check consistency of
//   Matrix::Face::u_ni
//   Residual::Face::u_ni
//   Residual::Face::u_dot_n
template<int dim>
void
test_vector_face2(const FEFaceValues<dim>& fe1,
		  const FEFaceValues<dim>& fe2,
		  const FEFaceValues<dim>& fetest1,
		  const FEFaceValues<dim>& fetest2)
{  
  const unsigned int n_dofs = fe1.get_fe().dofs_per_cell;
  const unsigned int t_dofs = fetest1.get_fe().dofs_per_cell;
  const unsigned int n_q = fe1.n_quadrature_points;
  
  std::vector<std::vector<FullMatrix<double> > > m(dim);
  for (unsigned int d=0;d<dim;++d)
    {
      m[d].resize(4);
      for (unsigned int i=0;i<4;++i)
	m[d][i].reinit(t_dofs, n_dofs);
    }
  std::vector<double> qu1(n_q);
  std::vector<double> qu2(n_q);

  Vector<double> u1(n_dofs);
  Vector<double> u2(n_dofs);
  Vector<double> v1(t_dofs);
  Vector<double> v2(t_dofs);
  Vector<double> w1(t_dofs);
  Vector<double> w2(t_dofs);
  
  std::vector<unsigned int> indices(n_dofs);
  for (unsigned int i=0;i<n_dofs;++i)
    indices[i] = i;
  std::vector<unsigned int> indicestest(t_dofs);
  for (unsigned int i=0;i<t_dofs;++i)
    indicestest[i] = i;
  
//  std::vector<double> factor(fe1.n_quadrature_points, 1.);
  
  for (unsigned int d=0;d<dim;++d)
    {
      AppShop::Matrix::Face::u_ni(m[d][0], m[d][1], m[d][2], m[d][3],
				  fe1, fe2, fetest1, fetest2, d);
    }
  
  for (unsigned int d=0;d<dim;++d)
    for (unsigned int i1=0;i1<n_dofs;++i1)
      for (unsigned int i2=0;i2<n_dofs;++i2)
	{
	  u1.reinit(n_dofs);
	  u1(i1) = 9.;
	  u2.reinit(n_dofs);
	  u2(i2) = 9.;
	  
	  m[d][0].vmult(v1, u1);
	  m[d][3].vmult(v2, u2);
	  m[d][1].vmult_add(v1, u2);
	  m[d][2].vmult_add(v2, u1);
	  
//	  std::cout << "mult" << std::endl;
//	  v1.print("%14.5e");
//	  v2.print("%14.5f");
	  
	  fe1.get_function_values(u1, indices, qu1);
	  fe2.get_function_values(u2, indices, qu2);

	  AppShop::Residual::Face::u_ni(w1, w2, qu1, qu2, d, fe1, fe2, fetest1, fetest2, -1.);
	  w1 += v1;
	  w2 += v2;
	  if (w1.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << w1.linfty_norm() << std::endl;
	  if (w2.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << w2.linfty_norm() << std::endl;
	}
  
    for (unsigned int i1=0;i1<n_dofs;++i1)
      for (unsigned int i2=0;i2<n_dofs;++i2)
	{
	  u1.reinit(n_dofs);
	  u1(i1) = 9.;
	  u2.reinit(n_dofs);
	  u2(i2) = 9.;

	  v1 = 0.;
	  v2 = 0.;
	  for (unsigned int d=0;d<dim;++d)
	    {
	      m[d][0].vmult_add(v1, u1);
	      m[d][3].vmult_add(v2, u2);
	      m[d][1].vmult_add(v1, u2);
	      m[d][2].vmult_add(v2, u1);
	    }
	  
//	  std::cout << "mult" << std::endl;
//	  v1.print("%14.5e");
//	  v2.print("%14.5f");
	  
	  fe1.get_function_values(u1, indices, qu1);
	  fe2.get_function_values(u2, indices, qu2);

	  for (unsigned int d=0;d<fetest1.get_fe().n_components();++d)
	    AppShop::Residual::Face::u_ni(w1, w2, qu1, qu2, d, fe1, fe2, fetest1, fetest2, -1.);
	  w1 += v1;
	  w2 += v2;
	  if (w1.linfty_norm() > 1.e-15)
	    deallog << "Error norm " << v1.linfty_norm() << std::endl;

	  w1 = 0.;
	  w2 = 0.;
	  
	  AppShop::Residual::Face::u_times_n(w1, w2, qu1, qu2, fe1, fe2, fetest1, fetest2, -1.);
	  w1 += v1;
	  w2 += v2;
	  if (w1.linfty_norm() > 1.e-15)
	    {
	      deallog << "Error w1" << std::endl;
	      for (unsigned int i=0;i<w1.size();++i)
		deallog << " " << w1(i);
	      deallog << std::endl;
	      for (unsigned int i=0;i<v1.size();++i)
		deallog << " " << v1(i);
	      deallog << std::endl;
	    }
	  if (w2.linfty_norm() > 1.e-15)
	    {
	      deallog << "Error w2" << std::endl;	  
	      for (unsigned int i=0;i<w2.size();++i)
		deallog << " " << w2(i);
	      deallog << std::endl;
	    }
	}
}


template <int dim>
void
test_element(Triangulation<dim>& tr,
	     FiniteElement<dim>& el,
	     FiniteElement<dim>& el_test)
{
  deallog << "Ansatz " << el.get_name() << std::endl;
  deallog << "Test   " << el_test.get_name() << std::endl;
  
  UpdateFlags flags = UpdateFlags(update_values
				  | update_JxW_values
				  | update_normal_vectors);
  
  QGauss<dim-1> face_quadrature(el.degree+1);
  FEFaceValues<dim> fe1(el, face_quadrature, flags);
  FEFaceValues<dim> fe2(el, face_quadrature, flags);
  FEFaceValues<dim> fetest1(el_test, face_quadrature, flags);
  FEFaceValues<dim> fetest2(el_test, face_quadrature, flags);

  for (typename Triangulation<dim>::active_cell_iterator cell = tr.begin_active(); cell != tr.end(); ++cell)
    {
      for (unsigned int face=0;face<GeometryInfo<dim>::faces_per_cell;++face)
	{
	  if (cell->at_boundary(face)) continue;
	  typename Triangulation<dim>::active_cell_iterator neighbor = cell->neighbor(face);
	  unsigned int neighbor_face = cell->neighbor_of_neighbor(face);

	  fe1.reinit(cell, face);
	  fetest1.reinit(cell, face);
	  fe2.reinit(neighbor, neighbor_face);
	  fetest2.reinit(neighbor, neighbor_face);

	  if (el.n_components() == 1)
	    {
	      if (el_test.n_components() == 1)
		test_scalar_face(fe1, fe2, fetest1, fetest2);
	      else
		test_vector_face2(fe1, fe2, fetest1, fetest2);
	    }
	  else
	      if (el_test.n_components() == 1)
		test_vector_face1(fe1, fe2, fetest1, fetest2);
	      else
		std::cerr << "tbd";	    
	}
    }
}



template <int dim>
void
test()
{
  Triangulation<dim> tr;
  GridGenerator::hyper_ball(tr);

  FE_DGQ<dim> q0(0);
  FE_DGQ<dim> q1(1);
  FE_DGQ<dim> q2(2);
  FE_RaviartThomasNodal<dim> rt0(0);
  FE_RaviartThomasNodal<dim> rt1(1);
  test_element(tr, q1, q1);
  test_element(tr, q1, q2);
  test_element(tr, q2, q1);
  test_element(tr, rt0, q0);
  test_element(tr, q0, rt0);
  test_element(tr, rt1, q1);
  test_element(tr, q1, rt1);
}


int main()
{
  test<2>();
}
