//---------------------------------------------------------------------------
//    $Id: test_app.cc 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008, 2009 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include "test_app.h"

#include <lac/block_vector.h>

#include <cmath>
#include <iostream>


DummyApplication::DummyApplication ()
{
  deallog << "->Dummy";
}


void
DummyApplication::declare_parameters (ParameterHandler&)
{}


void
DummyApplication::initialize (ParameterHandler&)
{
  deallog << __FUNCTION__ << std::endl;
  count = 0;
}


void
DummyApplication::remesh()
{
  deallog << __FUNCTION__ << std::endl;
  count = 0;
}


void
DummyApplication::init_vector (FEVector& dst) const
{
  deallog << __FUNCTION__ << std::endl;
  dst.reinit(1, 1);
}


void
DummyApplication::start_vector (FEVector& dst,
				 std::string caller) const
{
  deallog << __FUNCTION__ << '(' << caller << ')' << std::endl;
  dst.reinit(1, 1);
}


double
DummyApplication::residual(FEVector&, const FEVectors&)
{
  deallog << __FUNCTION__ << std::endl;
  return 1./count/count;
}


void
DummyApplication::solve(FEVector& start, const FEVectors& rhs)
{
  this->clear_events();
  deallog << __FUNCTION__ << std::endl;
  count++;
  start = rhs.vector(0);
}


void
DummyApplication::Tsolve(FEVector& start, const FEVectors& rhs)
{
  deallog << __FUNCTION__ << std::endl;
  count++;
  start = rhs.vector(0);
}


double
DummyApplication::estimate(const FEVectors&)
{
  deallog << __FUNCTION__ << std::endl;
  return 1./count;
}


double
DummyApplication::evaluate(const FEVectors&)
{
  deallog << __FUNCTION__ << std::endl;
  return count;
}


void
DummyApplication::grid_out(const std::string& filename) const
{
  deallog << __FUNCTION__
	  << ' ' << filename
	  << std::endl;
}


void
DummyApplication::data_out(const std::string& filename,
	      const FEVectors&)
{
  deallog << __FUNCTION__
	  << ' ' << filename
	  << std::endl;
}


void
DummyApplication::set_time (double t)
{
  deallog << __FUNCTION__
	  << ' ' << t
	  << std::endl;
}


void
DummyApplication::set_time_step (double t)
{
  deallog << __FUNCTION__
	  << ' ' << t
	  << std::endl;
}


bool
DummyApplication::autonomous () const
{
  deallog << __FUNCTION__ << std::endl;
  return false;
}

//----------------------------------------------------------------------//

RandomApplication::RandomApplication ()
{
  deallog << "->RandomApplication";
}


void
RandomApplication::declare_parameters (ParameterHandler&)
{}


void
RandomApplication::init_vector (FEVector& dst) const
{
  dst.reinit(1, 1);
}


void
RandomApplication::start_vector (FEVector& dst,
				 std::string) const
{
  init_vector(dst);
  dst.block(0)(0) = 1.;
}


void
RandomApplication::solve(FEVector&, const FEVectors&)
{
  this->clear_events();
}


double
RandomApplication::residual(FEVector&, const FEVectors&)
{
  return 0.;
}


double
RandomApplication::evaluate(const FEVectors& src)
{
  deallog << "Value " << src.vector(0).block(0)(0) << std::endl;
  return 0.;
}

//----------------------------------------------------------------------//

SquareApplication::SquareApplication()
    : target(2.),
      derivative(0.)
{
  deallog << "->Square";
}


void
SquareApplication::declare_parameters (ParameterHandler&)
{}


void
SquareApplication::initialize (ParameterHandler&)
{}


void
SquareApplication::remesh()
{}


void
SquareApplication::init_vector (FEVector& dst) const
{
  dst.reinit(1, 1);
}


void
SquareApplication::start_vector (FEVector& dst,
				 std::string) const
{
  init_vector(dst);
  dst.block(0)(0) = 100.;
}


double
SquareApplication::residual(FEVector& dst, const FEVectors& src)
{
    const double u = src.vector(0).block(0)(0);
    const double r = u*u-target;
    dst.block(0)(0) = r;
    return std::fabs(r);
}


void
SquareApplication::solve(FEVector& start, const FEVectors& rhs)
{
  if (this->notifications.any())
    {
      deallog << "assembling" << std::endl;
      derivative = 2. * rhs.vector(1).block(0)(0);
      this->clear_events();
    }
  
  start.block(0)(0) = rhs.vector(0).block(0)(0) / derivative;
}


double
SquareApplication::evaluate (const FEVectors& dst)
{
  deallog << "Value " << dst.vector(0).block(0)(0) << std::endl;
  return dst.vector(0).block(0)(0);
}


//----------------------------------------------------------------------//

ODEApplication::ODEApplication()
{
  deallog << "->ODE";
}


void
ODEApplication::declare_parameters (ParameterHandler&)
{}


void
ODEApplication::initialize (ParameterHandler&)
{}


void
ODEApplication::remesh()
{}


void
ODEApplication::init_vector (FEVector& dst) const
{
  dst.reinit(4, 2);
}


void
ODEApplication::start_vector (FEVector& dst,
				 std::string) const
{
  init_vector(dst);
  for (unsigned int block=0;block<dst.n_blocks();++block)
    dst.block(block) = 1.;
}


double
ODEApplication::residual(FEVector& dst, const FEVectors& in)
{
  const FEVector& src = in.vector(0);
  double time = *data->scalar("time");
  for (unsigned int i=0;i<dst.n_blocks();++i)
    {
      dst.block(i)(0) = (i+1.) * src.block(i)(0) / time;
      dst.block(i)(1) = (i+1.) * src.block(i)(0) * src.block(i)(0)
			/ std::pow(time, i+2.);
      
    }
//  dst.print(std::cout);
  return dst.l2_norm();
}


void
ODEApplication::solve(FEVector&, const FEVectors&)
{
  this->clear_events();
  deallog << "not solving" << std::endl;
}


double
ODEApplication::evaluate (const FEVectors& dst)
{
  double sum = 0.;
  for (unsigned int i=0;i<dst.vector(0).n_blocks();++i)
    {
      const double e0 = std::fabs(dst.vector(0).block(i)(0)/std::pow(10., i+1.)-1.);
      const double e1 = std::fabs(dst.vector(0).block(i)(1)/std::pow(10., i+1.)-1.);
      deallog << "Error("<< i << ",0): " << e0 << std::endl
	      << "Error("<< i << ",1): " << e1 << std::endl;
      sum += e0 + e1;
    }
  return sum;
}


