//---------------------------------------------------------------------------
//    $Id: timestep_control.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__appframe__timestep_control_h
#define __deal2__appframe__timestep_control_h

#include <base/subscriptor.h>
#include <base/smartpointer.h>
#include <lac/vector_memory.h>
#include <cstdio>


using namespace dealii;

namespace dealii
{
  class ParameterHandler;
}

namespace AppFrame
{
/**
 * Control class for timestepping schemes. Its main task is
 * determining the size of the next time step and the according point
 * in the time interval. Additionally, it controls writing the
 * solution to a file.
 */
class TimestepControl : public Subscriptor
{
  public:
				     /**
				      * The time stepping
				      * strategies. These controlled
				      * by the value of
				      * TimestepControl::tolerance()
				      * and
				      * TimestepControl::start_step().
				      */
    enum Strategy
    {
					   /**
					    * Choose a uniform time
					    * step.
					    */
	  uniform,
					   /**
					    * Start with the time step
					    * size given and double it
					    * in every step.
					    */
	  doubling,
	  
					   /**
					    * Modifies the time step
					    * according to the start
					    * residual of the inner
					    * iteration of the
					    * previous time step. This
					    * is meant to generate a
					    * sequence of time steps
					    * approaching infinity
					    * fast (for computing a
					    * stationary limit), while
					    * keeping an inner Newton
					    * iteration in its domain
					    * of convergence.
					    *
					    * If this residual was
					    * larger than the
					    * tolerance, divide the
					    * time step by two.
					    *
					    * If the residual was
					    * smaller than one half of
					    * tolerance, the time step
					    * is doubled.
					    */
	  inner_residual
    };

				     /**
				      * Constructor setting default
				      * values
				      */
    TimestepControl (double start = 0.,
		     double final = 1.,
		     double tolerance = 1.e-2,
		     double start_step = 1.e-2,
		     double print_step = -1.,
		     bool autonomous = false);
    
    static void declare_parameters (ParameterHandler& param);
    void parse_parameters (ParameterHandler& param);

				     /**
				      * The left end of the time interval.
				      */
    double start () const;
				     /**
				      * The right end of the time
				      * interval. The control
				      * mechanism ensures that the
				      * final time step ends at this
				      * point.
				      */
    double final () const;
				     /**
				      * The tolerance value
				      * controlling the time steps.
				      */
    double tolerance () const;
				     /**
				      * The size of the current time
				      * step.
				      */
    double step () const;
				     /**
				      * The current time.
				      */
    double now () const;
				     /**
				      * Compute the next time step
				      * according to tolerance() and
				      * the strategy(). Takes the
				      * control object of the inner
				      * iteration for input.
				      */
    double next (const double indicator);

				     /**
				      * Returns true if the problem is
				      * autonomous (and therefore, no
				      * assembling is required in the
				      * time steps).
				      */
    bool autonomous () const;

				     /**
				      * Set start value.
				      */
    void start (double);
				     /**
				      * Set final time value.
				      */
    void final (double);
				     /**
				      * Set tolerance
				      */
    void tolerance (double);
				     /**
				      * Set strategy.
				      */
    void strategy (Strategy);
    
				     /**
				      * Set size of the first
				      * step. This may be overwritten
				      * by the time stepping strategy.
				      */
    void start_step (double);
    
				     /**
				      * Set size of the maximum
				      * step size.
				      */
    void max_step (double);
    
				     /**
				      * Set the autonomous flag.
				      */
    void autonomous (bool);

				     /**
				      * Set now() equal to
				      * start(). Initialize step() and
				      * print() to their initial
				      * values.
				      */
    void restart ();
				     /**
				      * Return true if this timestep
				      * should be written to disk.
				      */
    bool print ();
				     /**
				      * Set the output name template.
				      */
    void file_name_format (const char*);
    const char* file_name_format ();
  private:
    
    double start_val;
    double final_val;
    double tolerance_val;
    Strategy strategy_val;
    double start_step_val;
    double max_step_val;
    double min_step_val;
				     /**
				      * The size of the current time
				      * step. This may differ from
				      * #step_val, if we aimed at a
				      * #print_step.
				      */
    double current_step_val;
    double step_val;
    bool is_autonomous;
    
    double now_val;
    double print_step;
    double next_print_val;
    
    char format[30];
};


inline double
TimestepControl::start () const
{
  return start_val;
}


inline double
TimestepControl::final () const
{
  return final_val;
}


inline double
TimestepControl::step () const
{
  return current_step_val;
}


inline double
TimestepControl::tolerance () const
{
  return tolerance_val;
}


inline double
TimestepControl::now () const
{
  return now_val;
}


inline void
TimestepControl::start (double t)
{
  start_val = t;
}


inline void
TimestepControl::final (double t)
{
  final_val = t;
}


inline void
TimestepControl::tolerance (double t)
{
  tolerance_val = t;
}


inline void
TimestepControl::strategy (Strategy t)
{
  strategy_val = t;
}


inline void
TimestepControl::start_step (double t)
{
  start_step_val = t;
}


inline void
TimestepControl::max_step (double t)
{
  max_step_val = t;
}


inline void
TimestepControl::restart ()
{
  now_val = start_val;
  step_val = (strategy_val == uniform)
	     ? tolerance_val
	     : start_step_val;
  current_step_val = step_val;
  if (print_step > 0.)
    next_print_val = now_val + print_step;
  else
    next_print_val = now_val - 1.;
}


inline void
TimestepControl::file_name_format (const char* fmt)
{
  strcpy(format, fmt);
}


inline const char*
TimestepControl::file_name_format ()
{
  return format;
}


inline bool
TimestepControl::autonomous () const
{
  return is_autonomous;
}


inline void
TimestepControl::autonomous (bool b)
{
  is_autonomous = b;
}
 
}

#endif
