//---------------------------------------------------------------------------
//    $Id: timestepping.h 1348 2013-08-16 00:45:07Z secanell $
//
//    Copyright (C) 2006, 2007, 2008 by Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __deal2__appframe__timestepping_h
#define __deal2__appframe__timestepping_h

#include <appframe/base.h>
#include <appframe/residual.h>
#include <appframe/timestep_control.h>

namespace dealii
{
  class ParameterHandler;
  template<typename T> class BlockVector;
}

namespace AppFrame
{
/**
 * Application class performing the theta timestepping scheme.
 *
 * The theta scheme is an abstraction of implicit and explicit Euler
 * schemes, the Crank-Nicholson scheme and linear combinations of
 * those. The choice of the actual scheme is controlled by the
 * parameter #theta as follows.
 * <ul>
 * <li> #theta=0: explicit Euler scheme
 * <li> #theta=1: implicit Euler scheme
 * <li> #theta=½: Crank-Nicholson scheme
 * </ul>
 *
 * For fixed #theta, the Crank-Nicholson scheme is the only second
 * order scheme. Nevertheless, further stability may be achieved by
 * choosing #theta larger than ½, thereby introducing a first order
 * error term. In order to avoid a loss of convergence order, the
 * adaptive theta scheme can be used, where <i>#theta=½+c dt</i>.
 *
 * The scheme sets the parameters <tt>theta</tt> <tt>time</tt> and
 * <tt>timestep</tt> in ApplicationData. The residual is called with
 * the solution at the previous time as first vector and should
 * compute <i>1/dt + (1-theta) A</i>. The function
 * ApplicationBase::solve() is called with this result as first
 * vector. It is then up to the underlying application to make sure
 * that the right residual is computed.
 *
 * A class that might help you implementing the right residuals is
 * ThetaResidual. Refer to the examples Tests::RotatingCone1 and
 * Tests::RotatingCone1 for the use of this class with linear and
 * nonlinear operators.
 *
 * @author Guido Kanschat, 2006
 */
  class ThetaTimestepping : public ApplicationCopy
  {
    public:
				       /**
					* ThetaTimestepping notifies
					* of this Event if the next
					* point in time is reached.
					*/
      static const Event new_time;
      
				       /**
					* ThetaTimestepping notifies
					* of this Event if the time
					* step size has changed
					*/
      static const Event new_step_size;
      
				       /**
					* Constructor, receiving the
					* application computing the
					* residual in each timestep.
					*/
      ThetaTimestepping (ApplicationBase& app);
      
				       /**
					* Constructor, receiving the
					* application computing the
					* residual in each timestep
					* and a Residual integrator
					* for computing the vector
					* <i>1/dt + (1-theta) A</i>.
					*/
      ThetaTimestepping (ApplicationBase& app, const Residual& res);
      
				       /**
					* Does nothing and avoids
					* residual function of the
					* base application being
					* called.
					*/
      virtual double residual(FEVector& start, const FEVectors& rhs);
      
				       /**
					* The actual timestep solver.
					*/
      virtual void solve(FEVector& start, const FEVectors& rhs);
      virtual void declare_parameters (ParameterHandler& param);
      virtual void initialize (ParameterHandler& param);
      
				       /**
					* Read the parameters.
					*/
      void _initialize (ParameterHandler& param);

    protected:
      TimestepControl control;
      
    private:
				       /**
					* The control parameter in the
					* range <tt>[0,1]</tt>.
					*/
      double theta;
				       /**
					* Use adaptive #theta if
					* <tt>true</tt>.
					*/
      bool adaptive;

				       /**
					* Default residual object if
					* the function
					* ApplicationBase::residual()
					* shoud be used.
					*/
      ApplicationResidual default_residual;
      
				       /**
					* The Residual object.
					*/
      SmartPointer<const Residual> compute_residual;      
  };


/**
 * A more sophisticated residual class that splits jobs between the
 * two kinds of residual needed for ThetaTimestepping.
 *
 * In each step of ThetaTimestepping, knowing the solution of the
 * previous step <i>u<sub>p</sub></i>, the following system is solved
 * for the solution at the new timestep <i>u</i>:
 *
 * \f{eqnarray*}
 * \left(\frac1{\delta t} u + \vartheta A(u) \right) &= r_p
 * \\
 * r_p &= \frac1{\delta t} u_p + (\vartheta-1) A(u_p).
 * \f}
 *
 * The vector <i>r<sub>1</sub></i> is computed using a Residual.  If
 * <i>A</i> is a nonlinear operator, a Newton scheme is used to solve
 * the first line. This Newton scheme will use a Residual as well,
 * computing the residual of the first equation.
 *
 * Although both residuals are different, they have the same
 * structure. in particular, the form <i>A</i> computed by a
 * LocalIntegrator is the same, just scaled differently.
 *
 * This class now allows you to use the same LocalIntegrator for both
 * residuals. Just use the value named <tt>scaling</tt> instead of
 * <tt>theta</tt> when computing local forms. The subtraction of
 * <i>r<sub>1</sub></i> is done by this class and thus no concern
 * anymore.
 *
 * For an example how to use this class, refer to Tests::RotatingCone2.
 *
 * @author Guido Kanschat
 */
  class ThetaResidual : public Residual
  {
    public:
      ThetaResidual(ApplicationBase& application);
      
      virtual double operator() (FEVector& dst, const FEVectors& src) const;
      
    private:
				       /**
					* The Application class
					* providing the function
					* Application::residual().
					*/
      SmartPointer<ApplicationBase> app;
  };



}


#endif
