//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: app_cathode.cc
//    - Description: 
//    - Developers: M. Secanell
//    - $Id: app_cathode.h 1417 2013-08-30 16:27:42Z secanell $
//
//---------------------------------------------------------------------------

#ifndef _FUELCELL__APP_CATHODE__H
#define _FUELCELL__APP_CATHODE__H

// Include deal.II classes
#include "base/parameter_handler.h"
#include "base/function_lib.h"
#include "base/function.h"
#include "base/quadrature_lib.h"

#include "lac/block_vector.h"
#include "lac/full_matrix.h"
#include "lac/solver_cg.h"
#include "lac/solver_gmres.h"
#include "lac/precondition.h"
#include "lac/precondition_block.h"
#include "lac/block_matrix_array.h"
#include "lac/filtered_matrix.h"
#include "lac/sparse_ilu.h"
#include "lac/sparse_direct.h"

#include "grid/grid_generator.h"
#include "grid/tria_accessor.h"
#include "grid/tria_iterator.h"
#include "grid/tria_boundary_lib.h"

#include "fe/fe_values.h"

#include "numerics/vector_tools.h"
#include "numerics/matrix_tools.h"
#include "numerics/error_estimator.h"

#include "boost/shared_ptr.hpp"

// Include appframe classes
#include "appframe/base.h"
#include "appframe/block_matrix_application.h"
#include "appframe/matrix_shop_cell.h"
#include "appframe/residual_shop_cell.h"
#include "appframe/matrix_shop_bdry.h"
#include "appframe/residual_shop_bdry.h"

// Include FuelCell classes
#include "fcst_constants.h"
#include "system_management.h"
#include "optimization_block_matrix_application.h"
#include "solver_utils.h"

#include "operating_conditions.h"
#include "geometries.h"

#include "PureGas.h"
#include "BinaryDiffusion.h"
#include "platinum.h"
#include "nafion.h"
#include "carbon.h"

#include "gas_diffusion_layer.h"
#include "micro_porous_layer.h"
#include "catalyst_layer.h"

#include "electron_transport_equation.h"
#include "proton_transport_equation.h"
#include "new_ficks_transport_equation.h"
#include "reaction_source_terms.h"

#include "linear_solvers.h"

//Include STL
#include "fstream"
#include "iostream"
#include "sstream"
#include <algorithm>

// Use namespace of deal.II
using namespace dealii;

namespace FuelCell
{
    
    namespace Application
    {
        //---------------------------------------------------------------------------
        //---------------------------------------------------------------------------
        //---------------------------------------------------------------------------
        /**
         * 
         * This class is used to solve a system of equations similar to the one presented in the journal article
         * M. Secanell et al. "Numerical Optimization of Proton Exchange Membrane Fuel Cell Cathode Electrodes",
         * Electrochimica Acta, 52(7):2668-2682, February 2007 and also presented at the ECCM 2006 conference under
         * the abstract titled M. Secanell et al, "A PEM fuel cell electrode model for gradient-based optimization", June 2006.
         *
         *
         * The cathode catalyst layer and the gas diffusion layer are modelled using 
         * a set of equations found in the journal article: 
         * 
         * - M. Secanell et al. "Numerical Optimization of Proton Exchange Membrane Fuel Cell Cathode Electrodes", 
         * Electrochimica Acta, 52(7):2668-2682, February 2007. 
         * 
         * The simulation is isothermal, isotropic, models
         * the electrochemical reactions using tafel kinetics, oxygen diffusion using Ficks Law and electronic
         * and protonic transport using Ohms Law. 
         * 
         * The equations solved are written as follows:
         * \f[
         * R_1(\vec{u}) = \nabla \cdot (c_{total}D^{eff}_{O_2} \nabla x_{O_2} ) = \frac{1}{nF}\nabla\cdot i
         * \f]
         * \f[
         * R_2(\vec{u}) = \nabla \cdot (\sigma^{eff}_{m}\nabla\phi_m) = \nabla \cdot i
         * \f]
         * \f[
         * R_3(\vec{u}) = \nabla \cdot (\sigma^{eff}_{s}\nabla\phi_s) = -\nabla \cdot i
         * \f]
         * where, for the case of a macro-homogeneous catalyst layer,
         * \f[
         * \nabla \cdot i = A_v i^{ref}_0 \left( \frac{c_{O_2}} {c^{ref}_{O_2}} \right) \mbox{exp} 
         * \left( \frac{\alpha F}{RT}(\phi_m - \phi_s) \right)
         * \f]
         * The solution variables, \f$ \vec{u} \f$ are the protonic potential, \f$\phi_m\f$, the electronic potential, 
         * \f$\phi_s\f$ and, instead of the oxygen concentration, we use the oxygen molar fraction, 
         * \f$x_{O_2}\f$, that also accounts for the oxygen dissolving in the ionomer by using Henrys Law. 
         * 
         * The govering equations above are nonlinear and therefore they cannot be solved directly.
         * 
         * In OpenFCST, we have decided to solve the system of equations using a nonlinear Newton solver. Therefore,
         * instead of implemeting the equations above, #cell_matrix and #cell_residual
         * implement a linearization of the governing equations above, i.e.
         * \f[
         * \frac{\partial R_i(\vec{u^{n-1}})}{\partial u_j} u_j = R_i(\vec{u^{n-1}})
         * \f]
         * 
         * For more information on the governing equations, discretization and solution methodology
         * please read the following references:
         * - M. Secanell et al. "Numerical Optimization of Proton Exchange Membrane Fuel Cell Cathode Electrodes", 
         * Electrochimica Acta, 52(7):2668-2682, February 2007. 
         * 
         * Also, the folder /data/cathode/Secanell_EA07_Numerical_Optimization_PEMFC_Cathode_Electrodes contains
         * a sample datafile to run this application with a macro-homogeneous model. 
         * 
         * The application can be used to solve a cathode with and without an MPL. You can select the appropriate
         * geometry by changing the following in the input file:
         * 
         * @code
         * subsection Grid generation
         *   set Type of mesh = Cathode OR CathodeMPL
         *   subsection Internal mesh generator parameters
         *      subsection Dimensions
         *       set Cathode current collector width [cm] = 0.1 # Thickness of the rib of the bipolar plates (BPP)
         *       set Cathode channel width [cm] = 0.1           # Thickness of the channel on the BPP
         *       set Cathode CL thickness [cm] = 1E-3           # Thickness of the cathode catalyst layer [cm]
         *       set Cathode GDL thickness [cm]  = 2E-2         # Thickness of the cathode gas diffusion layer [cm]
         *       set Cathode MPL thickness [cm] = 2E-3          # Thickness of the cathode microporous layer [cm]
         *      end
         *   end
         * end
         * @endcode
         * 
         * The model can also be used to solve an agglomerate catalyst layer model. The governing equations are similar to the
         * ones outlined above, however, the volumetric current density source, i.e. \f$ \nabla \cdot i \f$ is obtained as
         * specified in the following article:
         * -  M. Secanell, K. Karan, A. Suleman and N. Djilali, "Multi-Variable Optimization of PEMFC Cathodes using 
         * an Agglomerate Model ", Electrochimica Acta, 52(22):6318-6337, 2007.
         * 
         * A sample dataset used to solve an agglomerate cathode catalyst layer model is given in 
         * data/cathode/Secanell_EA07_MultiVariable_Optimization_PEMFC_Cathodes_Agglomerate_Model
         * 
         * If using this function please cite the articles above as references.
         * 
         * @author Marc Secanell, 2006-13
         * 
         */
        template <int dim>
        class AppCathode
        :
        public AppFrame::OptimizationBlockMatrixApplication<dim>
        {
        public:
            
            ///@name Constructors, destructor, and initalization
            //@{
            /**
             * Constructor.
             * @note the pointer data is initialized to boost::shared_ptr<> (), this means that
             * the pointer is empty and when we do data.get() it will return 0. This is good because at ApplicationBase
             * constructor an ApplicationData will be constructed.
             */
            AppCathode ( boost::shared_ptr<AppFrame::ApplicationData> data =
            boost::shared_ptr<AppFrame::ApplicationData> () );
            
            /**
             * Destructor
             */
            ~AppCathode();
            /**
             * Declare all parameters that are needed for:
             *   - the computation of the equation coefficients
             *   - the control of the linear system solution
             *   - ...
             */
            virtual void declare_parameters ( ParameterHandler& param );
            
            /**
             * Function called by optimization loop in order to set the values in the
             * ParameterHandler to the new design parameters.
             * Since ParameterHandler depends on the problem we are solving, set_parameters() is set
             * at the most inner loop of the application.
             */
            virtual void set_parameters ( const std::vector<std::string>& name_dvar,
                                          const std::vector<double>& value_dvar,
                                          ParameterHandler& param );
            
            /**
             * Set up how many equations are needed and
             * read in parameters for the parameter handler in order to initialize data
             */
            void _initialize ( ParameterHandler& param );
            
            /**
             * Call the other initialize routines from the inherited classes
             */
            virtual void initialize ( ParameterHandler& param );
            
            /**
             * Initilize and create the grid
             * Determines whether reading from a grid file or creating new
             */
            void initialize_grid ( ParameterHandler& param );
            
            /**
             * Initialize nonlinear solution
             * @note This routine is not working well yet. In fact, I don't know how to do it...
             */
            void init_solution ( AppFrame::FEVector& src ) ;
            
            //@}
            
            ///@name Local CG FEM based assemblers
            //@{
            /**
             * Integration of local bilinear form. Here we loop over the quadrature
             * points and over degrees of freedom in order to compute the matrix for the cell
             * This routine depends on the problem at hand and is called by assemble() in DoF_Handler
             * class
             * The matrix to be assembled is:
             *     \f[
             *       \begin{array}{l}
             *       M(i,j).block(0) = \int_{\Omega} a \nabla \phi_i \nabla \phi_j d\Omega + \int_{\Omega} \phi_i \frac{\partial f}{\partial u_0}\Big|_n \phi_j d\Omega \\
             *       M(i,j).block(1) = \int_{\Omega} \phi_i \frac{\partial f}{\partial u_1}\Big|_n \phi_j d\Omega \\
             *       M(i,j).block(2) = \int_{\Omega} \phi_i \frac{\partial f}{\partial u_2}\Big|_n \phi_j d\Omega
             *       \end{array}
             *     \f]
             */
            virtual void cell_matrix ( MatrixVector& cell_matrices,
                                       const typename DoFApplication<dim>::CellInfo& cell );
            
            virtual void bdry_matrix ( MatrixVector& bdry_matrices,
                                       const typename DoFApplication<dim>::FaceInfo& bdry);
            /**
             * Integration of the rhs of the equations. Here we loop over the quadrature
             * points and over degrees of freedom in order to compute the right
             * hand side for each cell
             * This routine depends on the problem at hand and is called by residual() in DoF_Handler
             * class
             * @note This function is called residual because in the case of nonlinear systems
             * the rhs is equivalent to the residual
             */
            virtual void cell_residual ( AppFrame::FEVector& cell_vector,
                                         const typename DoFApplication<dim>::CellInfo& cell );
            
            virtual void bdry_residual ( AppFrame::FEVector& bdry_vector,
                                         const typename DoFApplication<dim>::FaceInfo& bdry);
            
            
            /**
             * Member function used to set dirichlet boundary conditions.
             * This function is application specific and it only computes the boundary_value
             * values that are used to constraint the linear system of equations that is being
             * solved
             */
            virtual void dirichlet_bc ( std::map<unsigned int, double>& boundary_values ) const;
            //@}
            
            
            /**
             * Solve the linear system of equations
             */
            virtual void solve ( AppFrame::FEVector& start,
                                 const AppFrame::FEVectors& rhs );
            
            /**
             * Estimate error per cell
             */
            virtual double estimate ( const AppFrame::FEVectors& sol );
            
            /**
             * Post-processing. Evaluate a functional such as the objective function of an
             * optimization problem
             */
            virtual double evaluate ( const AppFrame::FEVectors& src );
            
            /**
             * Reimplementation of the routine in the base class BaseApplication in namespace AppFrame so
             * that the right labels are outputed and so that I can compute and output the source term.
             */
            virtual void data_out ( const std::string &basename,
                                    const AppFrame::FEVectors &src );
            
            /**
             * Check to see that the solution read from file can be used for the new simulation conditions.
             * Uses member functions to compare the initial conditions in the parameter file to those in
             * the loaded solution.
             */
            void parse_init_solution ( AppFrame::FEVector& src,
                                       FuelCell::OperatingConditions& OC,
                                       bool &good_solution );
            
            /**
             * Check to see that the initial solution for the solid phase potential is usable.
             * If the variable varies by more than 30% at the boundary, the solution is considered unusable,
             * and a default solution will be generated.
             */
            bool parse_init_solution_phi_s ( AppFrame::FEVector& src, double& V_cell );
            
            /**
             * Check to see that the initial solution for the oxygen mole fraction is usable.
             * If the variable varies by more than 30% at the boundary, the solution is considered unusable,
             * and a default solution will be generated.
             */
            bool parse_init_solution_x_O2 ( AppFrame::FEVector& src, double& x_O2 );
            
            
            ///@name Optimization Member Functions
            //@{
            /**
             * Integration of local bilinear form.
             */
            virtual void cell_dresidual_dlambda ( std::vector<AppFrame::FEVector >& cell_vector,
                                                  const typename DoFApplication<dim>::CellInfo& cell,
                                                  std::vector<std::vector<double> >& src );
            
            /**
             * This class is called by responses to make sure that all responses requested are
             * implemented in either cell_responses, global_responses or face_responses.
             * @note Every time we add a new response that we can calculate we need to update this
             * file.
             */
            virtual void check_responses();
            
            /**
             * Compute the value of all objective function and constraints
             */
            virtual void cell_responses ( std::vector<double>& resp,
                                          const typename DoFApplication<dim>::CellInfo& info,
                                          const AppFrame::FEVector& sol );
            /**
             * This class is used to evaluate all responses that do not require looping over cells.
             * An example of one of this types of constraints is the solid volume fraction.
             */
            virtual void global_responses ( std::vector<double>& resp,
                                            const AppFrame::FEVector& sol );
            
            /**
             * This class is used to evaluate the derivative of all the functionals that require looping over cells
             * with respect to the design variables.
             * This class is called by responses to evaluate the response at each cell.
             */
            virtual void cell_dresponses_dl ( std::vector<std::vector<double> >& cell_df_dl,
                                              const typename DoFApplication<dim>::CellInfo& info,
                                              const AppFrame::FEVector& sol );
            
            /**
             * This class is used to evaluate the sensitivities of all responses that do not require looping over cells
             * with respect of the design variables.
             * An example of one of this types of constraints is the solid volume fraction.
             */
            virtual void global_dresponses_dl ( std::vector<std::vector<double> >& df_dl,
                                                const AppFrame::FEVector& sol );
            /**
             * This class is used to evaluate the derivative of all the functionals that require looping over cells
             * with respect of the unknowns of the system of governing equations.
             * This class is called by responses to evaluate the response at each cell.
             */
            virtual void cell_dresponses_du ( std::vector<AppFrame::FEVector >& cell_df_du,
                                              const typename DoFApplication<dim>::CellInfo& info,
                                              std::vector<std::vector<double> >& src );
            
            /**
             * This class is used to evaluate the sensitivities of all responses that do not require looping over cells
             * with respecto of the unknowns of the system of governing equations.
             * An example of one of this types of constraints is the solid volume fraction.
             */
            virtual void global_dresponses_du ( std::vector<AppFrame::FEVector >& df_du,
                                                const AppFrame::FEVector& src );
            //@}
            
            
    protected:
        ///@name For debugging purposes only
        //@{
        /**
         * Member function used to print a cell matrix into a file so that it can be studied.
         * This routine can be called as follows:
         * @code
         *          print_cell_matrix(cell_matrices[0].matrix, "marc.dat");    
         * @endcode
         * This will store the data in the first block in the MatrixVector  in a file named marc.dat.
         */
        void print_cell_matrix(FullMatrix<double> matrix, std::string filename = "cell_matrix.dat", bool exit = true) const
        {
            std::ofstream marc;
            marc.open(filename);
            matrix.print(marc);
            marc.close();
            if (exit)
            {
                deallog<<"Fins aviat..."<<std::endl;
                abort();
            }
        }
        
        /**
         * Testing file:
         */
        void print_cell_matrix(ParameterHandler param, std::string filename = "parameters_sample.prm", bool exit = true) const
        {
            std::ofstream marc(filename);
            param.print_parameters(marc,
                                   ParameterHandler::Text);
            if (exit)
            {
                deallog<<"Fins aviat..."<<std::endl;
                abort();
            }
            
        }
        
        //@}
        /**
         * Class used to store all information related to the system of equations that is
         * being solved such as variable_names, equation_names and couplings between
         * equations.
         */
        FuelCell::SystemManagement system_management;
        
        ///@name Pre-processor and operating condition classes
        //@{
        /** Create an object to generate the mesh*/
        boost::shared_ptr< FuelCellShop::Geometry::GridBase<dim> > grid;
        
        /** Operating conditions */
        FuelCell::OperatingConditions OC;
        //@}
        
        ///@name Material classes
        //@{       
        /** The cathode contains water vapour, so we need to create an object water
         * in order to compute viscosity, density, etc. for water
         */
        FuelCellShop::Material::Water water;
        
        /** The cathode contains water vapour, so we need to create an object water
         * in order to compute viscosity, density, etc. for water
         */
        FuelCellShop::Material::Oxygen oxygen;
        
        /** The cathode contains water vapour, so we need to create an object water
         * in order to compute viscosity, density, etc. for water
         */
        FuelCellShop::Material::Nitrogen nitrogen;
        //@}
        
        ///@name Layer classes
        //@{
        /** The object GDL layer will contain all the information relevant to the
         * the GDL. We can request any effective property from this class
         */
        boost::shared_ptr<FuelCellShop::Layer::GasDiffusionLayer<dim> > CGDL;
        
        /** The object CMPL layer will contain all the information relevant to the
         * the micro-porous layer. We can request any effective property from this class
         */
        boost::shared_ptr<FuelCellShop::Layer::MicroPorousLayer<dim> > CMPL;
        
        /** The object CCL layer will contain all the information relevant to the
         * the catalyst layer. We can request any effective property from this class
         */
        boost::shared_ptr<FuelCellShop::Layer::CatalystLayer<dim> > CCL;

        //@}
               
        ///@name Physics base classes
        //@{
        /**
         * Equation class used to assemble the cell matrix for electron transport.
         */
        FuelCellShop::Equation::ElectronTransportEquation<dim> electron_transport_equation;
        /**
         * Equation class used to assemble the cell matrix for Fick's mass transport.
         * @note The order in the .h file is important here. Objects that need to be used
         * in the initialization of classes need to be declared before this class. In this case
         * FicksTransportEquation takes in two material classes. Therefore, they
         * need to be defined before FicksTransportEquation.
         * 
         */
        FuelCellShop::Equation::NewFicksTransportEquation<dim> ficks_transport_equation;
        /**
         * Equation class used to assemble the cell matrix for proton transport.
         */
        FuelCellShop::Equation::ProtonTransportEquation<dim> proton_transport_equation;
        /**
         * Equation class used to asemble the cell matrix and cell_residual for
         * the reaction source terms
         */
        FuelCellShop::Equation::ReactionSourceTerms< dim > reaction_source;
        
        //@}
        
        ///@name Other internal data
        //@{
        /**
         * Structure where we store the problem we want to solve.
         * Each vector component contains a string with the
         * name of the equation we want to solve
         * Then, the number of components is equation_names.size()
         */
        std::vector<std::string> equation_names;
        /**
         * Structure where we store the name of each component
         * in our problem. The component names are stored in the same
         * way as they are stored in the solution.
         */
        std::vector<std::string> component_names;
        
        /** Stores the design variable names so that the name can be appended to the .vtk file name. */
        std::vector<std::string> design_var;
        
        /** Stores the values of the design variables so that the number can be appended to the .vtk file name. */
        std::vector<double> design_var_value;
        //@}
            
        };
        
    }
    
}

#endif //_FUELCELL__APP_CATHODE_H
