//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: app_read_mesh.cc
//    - Description: 
//    - Developers: ?????????
//    - $Id: app_read_mesh.h 1417 2013-08-30 16:27:42Z secanell $
//
//---------------------------------------------------------------------------

#ifndef _FUELCELL__APP_READ_MESH__H
#define _FUELCELL__APP_READ_MESH__H

// Include deal.II classes
#include "base/parameter_handler.h"
#include "base/function_lib.h"
#include "base/function.h"
#include "base/quadrature_lib.h"

#include "lac/block_vector.h"
#include "lac/solver_cg.h"
#include "lac/solver_gmres.h"
#include "lac/precondition.h"
#include "lac/precondition_block.h"
#include "lac/block_matrix_array.h"
#include "lac/filtered_matrix.h"
#include "lac/sparse_ilu.h"
#include "lac/sparse_direct.h"

#include "grid/grid_generator.h"
#include "grid/tria_accessor.h"
#include "grid/tria_iterator.h"
#include "grid/tria_boundary_lib.h"

#include "fe/fe_values.h"

#include "numerics/vector_tools.h"
#include "numerics/matrix_tools.h"
#include "numerics/error_estimator.h"
#include "base/data_out_base.h"
#include "numerics/data_out.h"
#include "boost/shared_ptr.hpp"

// Include appframe classes
#include "appframe/base.h"
#include "appframe/block_matrix_application.h"
#include "appframe/matrix_shop_cell.h"
#include "appframe/residual_shop_cell.h"
#include "appframe/matrix_shop_bdry.h"
#include "appframe/residual_shop_bdry.h"

// Include FuelCell classes
#include "optimization_block_matrix_application.h"
#include "geometries.h"

//Include STL
#include "fstream"
#include "iostream"
#include "sstream"
#include <algorithm>

// Use namespace of deal.II
using namespace dealii;

namespace FuelCell
{
/**
 * This namespace is used for all those auxiliary classes that are used by AppCathode.
 * These are mainly classes that inherit Function<dim> and that are necessary in order to call
 * some subroutine from deal.II. For example InitialSolution is created in order to use the
 * deal.II class VectorInterpolate which in turn is used to set up the initial solution to the problem.
 */
namespace InitialSolution
{
/**
 * This class is used when solving the problem using Newton's method to provide an initial solution.
 * This function is called in VectorTools::interpolate(..,..,InitialSolution<dim> marc,...)
 * It provides a solution that satisfies Dirichlet boundaries and has a gradient.
 */
template <int dim>
//class InitialSolution
class AppReadMeshIC
            :
            public Function<dim>
{
public:
    /**
     * Constructor
     */
    AppReadMeshIC (std::vector<std::string> );
    /**
     * Destructor
     */
    ~AppReadMeshIC ();
    /**
     * This is the member function that computes the value of the initial
     * solution for a given point.
     */
    void vector_value ( const Point<dim> &p,
                        Vector<double> &v ) const;
								
    double value (const Point<dim> &/*p*/, const unsigned int) const;

    /**
     * Function to set the component names of the initial solution.
     */
    void set_solution_names ( std::vector<std::string> names )
    {
        component_names = names;
    }

private:

    /**
     * List of solution variables
     */
    std::vector<std::string> component_names;
};
} //end namespace InitialSolution


namespace Application
{
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
/**
 *
 * WHAT DOES THIS APPLICATION DO?
 *
 * @author ???
 */
template <int dim>
class AppReadMesh
            :
            public AppFrame::OptimizationBlockMatrixApplication<dim>
{
public:

    /**
     * Constructor.
     * @note the pointer data is initialized to boost::shared_ptr<> (), this means that
     * the pointer is empty and when we do data.get() it will return 0. This is good because at ApplicationBase
     * constructor an ApplicationData will be constructed.
     */
    AppReadMesh ( boost::shared_ptr<AppFrame::ApplicationData> data =
                     boost::shared_ptr<AppFrame::ApplicationData> () );

    /**
     * Destructor
     */
    ~AppReadMesh();
    /**
     * Declare all parameters that are needed for:
     *   - the computation of the equation coefficients
     *   - the control of the linear system solution
     *   - ...
     */
    virtual void declare_parameters ( ParameterHandler& param );

    /**
     * Set up how many equations are needed and
     * read in parameters for the parameter handler in order to initialize data
     */
    void _initialize ( ParameterHandler& param );

    /**
     * Call the other initialize routines from the inherited classes
     */
    virtual void initialize ( ParameterHandler& param );

    /**
    * Initilize and create the grid
    * Determines whether reading from a grid file or creating new
    */
    void initialize_grid ( ParameterHandler& param );

    /**
     * Initialize nonlinear solution
     * @note This routine is not working well yet. In fact, I don't know how to do it...
     */
    void init_solution ( AppFrame::FEVector& src ) ;
	 
    /**
     * Integration of the rhs of the equations. Here we loop over the quadrature
     * points and over degrees of freedom in order to compute the right
     * hand side for each cell
     * This routine depends on the problem at hand and is called by residual() in DoF_Handler
     * class
     * @note This function is called residual because in the case of nonlinear systems
     * the rhs is equivalent to the residual
     */
    void cell_residual ( AppFrame::FEVector&,
                                 const typename DoFApplication<dim>::CellInfo& )
	 {}
	 
    /**
     * Compute the value of all objective function and constraints
     */
    void cell_responses ( std::vector<double>& ,
                                  const typename DoFApplication<dim>::CellInfo&,
                                  const AppFrame::FEVector& )
	 {}
	 
    /**
     * This class is used to evaluate all responses that do not require looping over cells.
     * An example of one of this types of constraints is the solid volume fraction.
     */
    void global_responses ( std::vector<double>&,
                                    const AppFrame::FEVector& )
	 {}

	 
    /**
     * Reimplementation of the routine in the base class BaseApplication in namespace AppFrame so
     * that the right labels are outputed and so that I can compute and output the source term.
     */
    virtual void data_out ( const std::string &basename,
                            const AppFrame::FEVectors &src );

protected:
    /** Create an object to generate the mesh*/
    boost::shared_ptr< FuelCellShop::Geometry::GridBase<dim> > grid;

    /**
     * Structure where we store the problem we want to solve.
     * Each vector component contains a string with the
     * name of the equation we want to solve
     * Then, the number of components is equation_names.size()
     */
    std::vector<std::string> equation_names;
    /**
     * Structure where we store the name of each component
     * in our problem. The component names are stored in the same
     * way as they are stored in the solution.
     */
    std::vector<std::string> component_names;
    
	/** Stores the design variable names so that the name can be appended to the .vtk file name. */
	std::vector<std::string> design_var;
	
	/** Stores the values of the design variables so that the number can be appended to the .vtk file name. */
	std::vector<double> design_var_value;

};

}
}

#endif