//---------------------------------------------------------------------------
//    $Id: optimization_block_matrix_application.h 1354 2013-08-17 00:01:22Z secanell $
//
//    Copyright (C) 2006 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef __appframe__optimization_block_application_h
#define __appframe__optimization_block_application_h

#include <base/parameter_handler.h>
#include <lac/vector.h>
#include <lac/block_vector.h>
#include <grid/grid_refinement.h>
#include <grid/grid_tools.h>
#include <grid/persistent_tria.h>
#include <dofs/dof_handler.h>
#include <dofs/dof_renumbering.h>
#include <fe/fe.h>
#include <fe/fe_values.h>
#include <numerics/solution_transfer.h>
#include <boost/shared_ptr.hpp>
//#include <numerics/mesh_worker.h>
//#include <numerics/mesh_worker_loop.h>

#include <appframe/base.h>
#include <appframe/dof_application.h>
#include <appframe/matrix_base.h>
#include <appframe/matrix_application.h>
#include <appframe/block_matrix_application.h>
#include <appframe/mesh_worker_loop.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>

namespace AppFrame
{

  /** Note: I need to use this class in order to make sure 
  * that the compiler knows that it will be defined in the file
  * because LocalIntegrator needs this class
  */
  template<int dim> class OptimizationBlockMatrixApplication; 

/**
*
* The Integrator computing the current density in the CL.
*
* Since this class will be used by MeshWorker::AssemblingIntegrator, it
* needs the functions for cell, boundary and interior face integration specified
* exactly as below.
*
* The base class Subscriptor (deal.ii class) is needed so that MeshWorker::AssemblingIntegrator
* can store a SmartPointer to an object of this class
*
* @author Marc Secanell, 2009 
*/

  template <int dim>
  class LocalResponse : 
        public MeshWorker::Assembler::Functional<double>,
	public MeshWorker::WorkerObjects::IntegrationWorker<dim>
  {
  private:
      SmartPointer<OptimizationBlockMatrixApplication<dim> > app;

  public:
  
    /**
    * Initialize
    */
    void initialize(OptimizationBlockMatrixApplication<dim>* a,
                  unsigned int m)
    {
      app = a;
      MeshWorker::Assembler::Functional<double>::initialize(m);
    }
    
    /**
    * Do the work on a cell.
    */
    void cell(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info);

    /**
    * Do the work on a boundary face.
    */
    void bdry(const typename AppFrame::DoFApplication<dim>::FaceInfo& face);

     /**
     * Do the work on an interior face.
     */
     void face(const typename AppFrame::DoFApplication<dim>::FaceInfo& face1,
               const typename AppFrame::DoFApplication<dim>::FaceInfo& face2);
  };

 /**
 * Application handling matrices and assembling the linear system
 * to solve the sensitivity equations.
 *
 * This class inherits the information on Triangulation and DoFHandler
 * from its base class BlockMatrixApplication. It adds information on the
 * linear system for sensitivity analysis
 *
 * @author Marc Secanell, 2006
 */
  template <int dim>
    class OptimizationBlockMatrixApplication :
  public AppFrame::BlockMatrixApplication<dim>
  {
  public:
     /**
      * The Event set by OptimizationMatrixApplication if
      * if we want to compute the sensitivities
      * and a new matrix should be
      * assembled.
      */
    static const AppFrame::Event sensitivity_analysis; 
    /**
     * Constructor for an object,
     * owning its own mesh and dof handler.
     *
     * If <tt>data</tt> is null,
     * a new handler for
     * application data is
     * generated.
     *
     * see constructor of DoFApplication.
     */
    OptimizationBlockMatrixApplication(boost::shared_ptr<AppFrame::ApplicationData> data =
				       boost::shared_ptr<AppFrame::ApplicationData>());

    /**
     * Constructor for an object,
     * borrowing mesh and dof
     * handler from another
     * object. If
     * <tt>triangulation_only</tt>
     * is true, only the
     * triangulation is borrowed.
     *
     * see constructor of DoFApplication.
     */
    OptimizationBlockMatrixApplication(AppFrame::DoFApplication<dim>&,
                                       bool triangulation_only);
	
	/** Destructor */
	~OptimizationBlockMatrixApplication(){};
	
    /**
     * Declare all parameters that are needed for: 
     *   - the computation of the equation coefficients
     *   - the control of the linear system solution
     *   - ...
     */
    virtual void declare_parameters(ParameterHandler& param);
    
    /**
     * Function called by optimization loop in order to set the values in the 
     * ParameterHandler to the new design parameters.
     * Since ParameterHandler depends on the problem we are solving, set_parameters() is set
     * at the most inner loop of the application.
     */
    virtual void set_parameters(const std::vector<std::string>& name_dvar,
                                const std::vector<double>& value_dvar,
                                ParameterHandler& param);

    /**
     */
    virtual void initialize(ParameterHandler& param);
   
    /**
     * Initialize nonlinear solution
     */
    virtual void init_solution(AppFrame::FEVector& src) = 0;

    /**
     * Post-processing. Evaluate all responses, usually functionals, necessary
     * such as the objective function and the constraints of the
     * optimization problem
     * @note size of f is num_objectives + num_constraints
     */
    virtual void responses (std::vector<double>& f,
			    const AppFrame::FEVectors& vectors);
    /**
     * This class is called by responses to make sure that all responses requested are
     * implemented in either cell_responses, global_responses or face_responses.
     */
    virtual void check_responses();

    /**
     * This class is used to evaluate all the functionals that require looping over cells.
     * This class is called by responses to evaluate the response at each cell.
     * 
     */
    virtual void cell_responses (std::vector<double>& resp,
				 const typename DoFApplication<dim>::CellInfo& info,
				 const AppFrame::FEVector& sol);

    /**
     * This class is used to evaluate all responses that do not require looping over cells.
     * An example of one of this types of constraints is the solid volume fraction.
     */
    virtual void global_responses (std::vector<double>& resp,
				   const AppFrame::FEVector& sol);
    
    /** This function is used to print the responses to screen (deallog) */
    virtual void print_responses (std::vector<double>& resp);


    /**
     * Post-processing. Evaluate the partial derivative
     * of all functionals necessary
     * such as the objective function and the constraints of the
     * optimization problem
     * In this member class we obtain the term:
     * \f[
     * \frac{\partial f_m}{\partial \lambda_k}
     * \f]
     * @note size of df_dl is for f num_objectives + num_constraints
     * and for lambda = num_design_var
     */
    virtual void dresponses_dl (std::vector<std::vector<double> >& df_dl,
				const AppFrame::FEVectors& src);
    /**
     * This class is used to evaluate the derivative of all the functionals that require looping over cells
     * with respect to the design variables.
     * This class is called by responses to evaluate the response at each cell.
     */
    virtual void cell_dresponses_dl(std::vector<std::vector<double> >& cell_df_dl,
				    const typename DoFApplication<dim>::CellInfo& info,
				    const AppFrame::FEVector& src);
    /**
     * This class is used to evaluate the sensitivities of all responses that do not require looping over cells
     * with respect of the design variables.
     * An example of one of this types of constraints is the solid volume fraction.
     */
    virtual void global_dresponses_dl(std::vector<std::vector<double> >& df_dl,
				      const AppFrame::FEVector& src);
    /**
     * Loop over all cells
     * and assemble the vector 
     * \f[
     * \frac{\partial f_m}{\partial u_i}
     * \f]
     * that is used to solve the sensitivity
     * equations by using the local
     * matrix integration functions
     * cell_dfunctional_du(), bdry_dfunctional_dlu()
     * and edge_dfunctinal_du() provided
     * by the derived class.
     * The vector index refers to m and the BlockVector
     * index refers to i.
     */
    virtual void dresponses_du(std::vector<AppFrame::FEVector>& dst,
			       const AppFrame::FEVectors& src);
    
    /**
     * Integration of local
     * bilinear form.
     */
    virtual void cell_dresponses_du(std::vector<AppFrame::FEVector>& df_du,
				    const typename DoFApplication<dim>::CellInfo& info,
				    std::vector<std::vector<double> >& src);

    /**
     * This class is used to evaluate the sensitivities of all responses that do not require looping over cells
     * with respect of the design variables.
     * An example of one of this types of constraints is the solid volume fraction.
     */
    virtual void global_dresponses_du(std::vector<AppFrame::FEVector>& df_du,
				      const AppFrame::FEVector& src);
    
    /**
     * Loop over all cells
     * and assemble the vector 
     * \f[
     * \frac{\partial R_n}{\partial \lambda_k}
     * \f]
     * that is used to solve the sensitivity
     * equations by using the local
     * matrix integration functions
     * cell_dresidual_dlambda(), bdry_dresidual_dlambda()
     * and edge_dresidual_dlambda() provided
     * by the derived class.
     * The vector index refers to k and the FEVector
     * index refers to n.
     */
    virtual void dresidual_dlambda(std::vector<AppFrame::FEVector>& dst,
				   const AppFrame::FEVectors& src);
    /**
     * Integration of local
     * bilinear form.
     */
    virtual void cell_dresidual_dlambda(std::vector<AppFrame::FEVector>& cell_vector,
					const typename DoFApplication<dim>::CellInfo& cell,
					std::vector<std::vector<double> >& src);
                                         
    /**
     * Solver in order to obtain the
     * analytical sensitivities for the
     * given objective and constraints using
     * the direct method. This member function
     * calls the different assemble member functions and solves the 
     * system of equations
     * @note The direct method is the most computationally efficient
     * method if the number of objectives and constraints
     * is larger than the number of design variables				       
     */
    void solve_direct (std::vector<std::vector<double> >& df_dl,
		       const AppFrame::FEVectors& sol);
    
    /**
     * Solver in order to obtain the
     * analytical sensitivities for the
     * given objective and constraints using
     * the adjoint method. This member function
     * calls the different assemble member functions and solves the 
     * system of equations
     * @note The adjoint method is the most computationally efficient
     * method if the number of number of design variables
     * is larger than the number of objectives and constraints
     */
    void solve_adjoint (std::vector<std::vector<double> >& df_dl,
			const AppFrame::FEVector& sol);
    /**
     * Member function that returns the number of responses
     */
    unsigned int get_n_resp() const;
    /**
     * Member function that returns the number of design variables
     */
    unsigned int get_n_dvar() const;
    /**
     * Member function that returns the name of the design variables
     */
    inline std::vector<std::string> get_name_dvar() const
      { return name_design_var;}

    /**
     * Member function that returns the name of the responses
     */
    inline std::vector<std::string> get_name_responses() const
      { return name_responses;}

    /**
     *
     */
    /*void enter_data_scalar(std::string name,
			   const double& scalar)
    {
      this->get_data()->enter(name, scalar);
      }*/
    
    /**
     * Print the default parameter handler file
     */
    void print_default_parameter_file()
    {
      ParameterHandler prm;
      this->declare_parameters(prm);
      std::ofstream file;
      file.open("Default_parameter_file.prm");
      prm.print_parameters (file, ParameterHandler::Text);
      file.close();
    }

    /** Function to see if we are transferring a solution on a refined grid to the initial coarse grid */
    bool get_bool_transfer_solution()
 	{return output_coarse_solution;}

	virtual void set_read_initial_solution(bool &read)
	{
		read_in_initial_solution = read;
	}
	
	/**
	 * This routine assigns the value of n_dvar, n_resp, name_design_var and name_responses to the
	 * internal variables n_dvar, n_resp, name_design_var and name_response
	 * 
	 */
	void set_optimization_parameters(unsigned int &n_dvar, unsigned int &n_resp, 
								std::vector<std::string> &name_design_var, std::vector<std::string> &name_responses);

	/**
	 * 
	 */
	void set_output_variables(std::vector<std::string> &dakota_name_responses)
	{
		name_responses = dakota_name_responses;
	};
	
	
  protected:
    /** Auxiliary routine to print the values df_dl 
     * This routine should be called once df_df is assembled. This is done
     * on solve_direct
     */
    void print_dresponses_dl(std::vector<std::vector<double> > pdf_pdl)
    {
      for (unsigned int i=0; i<n_resp; ++i)
	for (unsigned int j=0; j<n_dvar; ++j)
	  std::cout<<"Df"<<i<<"/Dl"<<j<<" is equal to "<<pdf_pdl[i][j]<<std::endl;
    }

    /** Auxiliary routine to print the values of df_du
     * This routine should be called once df_du is assembled. This is done
     * on solve_direct
     */
    void print_dresponses_du(std::vector<AppFrame::FEVector> df_du)
    {
      // Write partial df_du:
      
      for (unsigned int r=0; r<n_resp; ++r)
	{  
	  // Output Du block by block 
	  unsigned int n_blocks = df_du[r].n_blocks();
	  unsigned int size = df_du[r].size();
	  unsigned int comp_per_block = size/n_blocks;
	  for (unsigned int i=0; i<n_blocks; ++i)
	    for (unsigned int j=0; j<comp_per_block; ++j)
	      {
		std::cout<<"Element "<<j<<" in block "<<i<<" in STEP (Du) is "<<df_du[r].block(i)(j)<<std::endl;
	      }
	  std::cout<<"NORM OF Df"<<r<<"/Du = "<<df_du[r].l2_norm()<<std::endl;
	}
    }

    /**
     * Definition of a pointer to a function that is called in a loop over
     * cells to compute the derivatives of either residual or responses with
     * respect to design variables.
     */
    typedef void (*CELL_Dvalues) (std::vector<AppFrame::FEVector>& ,
				  const typename DoFApplication<dim>::CellInfo& ,
				  std::vector<std::vector<double> >& );
    
    /**
     * This is an auxiliary function called by dresidual_dlambda and
     * dfunctional_du. Since both functions have to do the same thing,
     * loop over the cells, but they call a different member function for cell assembly,
     * I use a flag to change from one to the other and then I call this
     * generic function.
     * With this dresidual_dlambda and dfunctional_du have the same interface
     * and the changes happen to a protected class.
     */
    void dfunction (std::vector<AppFrame::FEVector>& dst,
		    const AppFrame::FEVectors& src,
		    bool dfunctional_du,
		    bool dresidual_dlambda);

    
    /**
     * Number of design variables
     */
    unsigned int n_dvar;
    /**
     * Number of objective functions
     */
    unsigned int n_obj;
    /**
     * Number of responses = n_obj + n_con
     */
    unsigned int n_resp;
    /**
     * Member that stores the name of the design variables
     */
    std::vector<std::string> name_design_var;
    /**
     * Member that stores the name of the responses, i.e. objectives and constraints.
     * @note first we put the name of all objectives and then all constraints
     */
    std::vector<std::string> name_responses;
    
    /**
	* Member that stores the name of the output variables, 
	* These names will be written to name_responses if optimization is not being used and ignored otherwise.
	*/
    std::vector<std::string> name_output_var;

    /** Decision variable for whether the solution is to be output for transfer */
    bool output_coarse_solution;

    /** Decision variable for whether the initial solution should be read from file */
    bool read_in_initial_solution;
    
    /** Decision variable for whether the application is being used for optimization */
    bool optimization; 
  };

}

#endif
