//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: ficks_transport_equation.h
//    - Description: Base Catalyst Layer Class. It implements the interface for other catalyst layer class
//        and some common methods.
//    - Developers: M. Secanell
//    - $Id: ficks_transport_equation.h 1354 2013-08-17 00:01:22Z secanell $
//
//---------------------------------------------------------------------------

#ifndef dimension
#if   deal_II_dimension == 1
#define _1D_
#elif deal_II_dimension == 2
#define _2D_
#elif deal_II_dimension == 3
#define _3D_
#endif
#endif

#ifndef _FCST_FUELCELLSHOP_EQUATION_FICKS_EQUATION_H_
#define _FCST_FUELCELLSHOP_EQUATION_FICKS_EQUATION_H_

// FCST includes
#include "equation_base.h"
#include "fcst_constants.h"
#include "gas_diffusion_layer.h"
#include "SGL_24_BA.h"
#include "micro_porous_layer.h"
#include "SGL_24_BC.h"
#include "catalyst_layer.h"

// Appframe includes
#include "appframe/matrix_shop_cell.h"

// STD
#include <sstream>
#include <string>

namespace FuelCellShop
{
    namespace Equation
    {
        
        /**
         * 
         * This class implements Fick's law of diffusion. Please note that this equation works for any gas. The gas type is
         * not selected in the class. Instead, an object of type BaseLayer is used to pass the information regarding the type
         * of gas and effective properties.
         * 
         *TODO Dynamic type casting has to be modified so that the base class type_id is used in the comparision in assemble_cell_matrix
         *
         * \author Marc Secanell, 2012
         */
        
        template<int dim>
        class FicksTransportEquation : public EquationBase<dim>
        {
        public:
            
            /**
             * Constructor. If using this constructor, use member function set_solute_and_solvent to setup 
             * the gases and name of the equation
             */
            FicksTransportEquation(FuelCell::SystemManagement& system_management);
            
            /**
             * Constructor.
             */
            FicksTransportEquation(FuelCell::SystemManagement& system_management,
                                   FuelCellShop::Material::PureGas* solute,
                                   FuelCellShop::Material::PureGas* solvent);
            
            /**
             * Method to set solute and solve if other constructor (not passing solute and solvent in the constructor)
             * is being used.
             * It will also setup name_equation and name_solution
             * 
             */
            inline void set_solute_and_solvent( FuelCellShop::Material::PureGas* solute, FuelCellShop::Material::PureGas* solvent1 )
            {
                this->gas = solute;
                this->solvent = solvent1;
                
                std::stringstream s;
                s <<"Ficks Transport Equation - "<<gas->get_name();
                this->name_equation = s.str();
                
                std::stringstream ss;
                ss <<gas->get_name()<<"_molar_fraction";
                this->name_solution = ss.str();
            }
            
            /**
             * Destructor.
             */
            virtual ~FicksTransportEquation();
            
            /**
             * Declare parameters.
             */
            virtual void declare_parameters(ParameterHandler& param);
            
            /**
             * Initialize parameters.
             */
            virtual void initialize(ParameterHandler& param);
            
            /**
             * Assemble local cell matrix.
             */
            virtual void assemble_cell_matrix(AppFrame::MatrixVector&                                 cell_matrices,
                                              const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                              FuelCellShop::Layer::BaseLayer<dim>* const              layer);
            
            /**
             * Assemble local boundary matrix.
             */
            virtual void assemble_bdry_matrix(AppFrame::MatrixVector&                                 bdry_matrices,
                                              const typename AppFrame::DoFApplication<dim>::FaceInfo& bdry_info,
                                              FuelCellShop::Layer::BaseLayer<dim>* const              layer){};
                                              
           /**
            * Assemble local cell residual.
            */
           virtual void assemble_cell_residual(AppFrame::FEVector&                                     cell_rhs,
                                               const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                               FuelCellShop::Layer::BaseLayer<dim>* const              layer);
           
           /**
            * Assemble local boundary residual.
            */
           virtual void assemble_bdry_residual(AppFrame::FEVector&                                     bdry_rhs,
                                               const typename AppFrame::DoFApplication<dim>::FaceInfo& bdry_info,
                                               FuelCellShop::Layer::BaseLayer<dim>* const              layer){};
                                               
           /**
            * The function returns
            * \p internal_cell_couplings
            * of this class,
            */
           virtual const couplings_map& get_internal_cell_couplings() const
           {
               return internal_cell_couplings;
           }
           
           /**
            * The function printing out
            * the equations info.
            */
           virtual void print_equation_info() const;
           
           /**
            * */
           DeclException1(GasDoesNotExistInGDL,
                          std::string,
                          << "Gas "  << arg1 << " does not exist in the Gas diffusion layer");
           
           /**
            * Member function used to test the functionality of the class.
            * It create an object of this class and test functionality.
            * 
            */
           void class_test();
                                                                                  
        protected:
            /**
             * Gas for which the equation is setup.
             */
            FuelCellShop::Material::PureGas* gas;
            /**
             * Solvent gas for which the equation is setup. The diffusion coefficient that we would use is given by
             * D_gas,solvent.
             */
            FuelCellShop::Material::PureGas* solvent;
            /**
             * Name of the equation that we are solving for. The name depends on the type of
             * gas we use. It is set to "Ficks Transport Equation - "<<gas.get_name();
             */
            std::string name_equation;
            /**
             * Name of the variable we are trying to solve for. This will be the molar fraction of the
             * variable we are solving for. For example, gas.get_name()<<" molar fraction"
             */
            std::string name_solution;
            /**
             * This function fills out \p internal_cell_couplings
             * of this class based on the structure of the equations at hand.
             * The names of variables and equations must match those defined in SystemManagement class
             * 
             */
            virtual void make_internal_cell_couplings();
            
            /**
             * \p internal_cell_couplings
             * object.
             */
            couplings_map internal_cell_couplings;
            
            /**
             * The system matrix block
             * indices related to the
             * equations implemented
             * in this class
             * and to the entire problem at hand.
             */
            std::vector<unsigned int> matrix_block_indices;
            
            /**
             * The component
             * indices related to the
             * equations implemented
             * in this class
             * and to the entire problem at hand.
             */
            std::vector<unsigned int> component_indices;
        };
        
    } // Equation
    
} // FuelCellShop
    
#endif