//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: reaction_source_terms.h
//    - Description: This class is used to assemble cell matrix and cell residual for reaction
//      source terms in the catalyst layer for various equation classes.
//    - Developers: Madhur Bhaiya and M. Secanell
//    - $Id: reaction_source_terms.h 1459 2013-08-30 22:27:10Z secanell $
//
//---------------------------------------------------------------------------

#ifndef _FCST_FUELCELLSHOP_EQUATION_REACTION_SOURCE_TERMS_H_
#define _FCST_FUELCELLSHOP_EQUATION_REACTION_SOURCE_TERMS_H_

#include "utils/fcst_constants.h"

#include "equations/equation_base.h"

#include "layers/catalyst_layer.h"

#include "reactions/tafel_kinetics.h"
#include "reactions/dual_path_kinetics.h"


namespace FuelCellShop
{
    namespace Equation
    {
        
        ///@name Exceptions
        //@{
        /**
         * Exception thrown when a particular "Required" solution variable
         * is not found for \p cathode or \p anode kinetics,
         * in \p user-defined solution variables.
         */
        DeclException2(VariableNotFoundForKinetics,
                       std::string,
                       std::string,
                       << "For " << arg1 << " kinetics source terms, \"" << arg2 << "\" is not found as one of the solution variables.");
        
        
        //@}	
        
        /**
         * This class is used to assemble reaction source terms in the catalyst layer. Usage details are similar to the SorptionSourceTerms class.
         * 
         * \author Madhur Bhaiya, 2013
         */
        template<int dim>
        class ReactionSourceTerms : public EquationBase<dim>
        {
        public:
            
            ///@name Constructors, destructor, and initalization
            //@{
            
            /**
             * Constructor.
             */
            ReactionSourceTerms(FuelCell::SystemManagement& system_management);
            
            /**
             * Destructor.
             */
            virtual ~ReactionSourceTerms();
            
            /**
             * Declare parameters.
             */
            virtual void declare_parameters(ParameterHandler& param) const;
            
            /**
             * Initialize parameters.
             */
            virtual void initialize(ParameterHandler& param);
            
            /**
             * Set the pointer to cathode kinetics in the object.
             * \note If an application uses cathode kinetics model, this method should be called neccessarily before initializing this object (using #initialize method).
             */
            inline void set_cathode_kinetics(FuelCellShop::Kinetics::BaseKinetics* kinetics)
            {
                cathode_kinetics = kinetics;
            }
            
            /**
             * Set the pointer to anode kinetics in the object.
             * \note If an application uses anode kinetics model, this method should be called neccessarily before initializing this object (using #initialize method).
             */
            inline void set_anode_kinetics(FuelCellShop::Kinetics::BaseKinetics* kinetics)
            {
                anode_kinetics = kinetics;
            }
            
            //@}
            
            ///@name Local CG FEM based assemblers
            //@{
            
            /**
             * Assemble local cell matrix.
             */
            virtual void assemble_cell_matrix(AppFrame::MatrixVector& cell_matrices,
                                              const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                              FuelCellShop::Layer::BaseLayer<dim>* const layer);
            
            /**
             * Assemble local cell residual.
             */
            virtual void assemble_cell_residual(AppFrame::FEVector& cell_residual,
                                                const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                FuelCellShop::Layer::BaseLayer<dim>* const layer);
            
            //@}
            
            ///@name Accessors & Info
            //@{
            
            /**
             * This function is used to adjust \p std::vector \p < \p internal_cell_couplings \p >, which is generated after getting \p internal_cell_couplings
             * from all the equations being used in the application.
             * 
             * \note It's very important to use this function, if we are considering source terms due to reaction in the catalyst layers. It's also noteworthy that
             * this function should be used after whole vector of \p internal_cell_couplings is created, and \b BEFORE \p make_cell_couplings of \b SystemManagement
             * is called.
             */
            virtual void adjust_internal_cell_couplings(std::vector< couplings_map >& ) const;
            
            /**
             * This function prints out
             * the \p info for this class.
             */
            virtual void print_equation_info() const;

            //@}
            
        protected:
            
            ///@name Kinetics Objects
            //@{
            /**
             * Pointer to Cathode Kinetics object, initialized in the constructor.
             */
            FuelCellShop::Kinetics::BaseKinetics* cathode_kinetics;
            
            /**
             * Pointer to Anode Kinetics object, initialized in the constructor.
             */
            FuelCellShop::Kinetics::BaseKinetics* anode_kinetics;
            //@}
            
            
            ///@name Local CG FEM based assemblers - make_ functions
            //@{
            
            /**
             * This function computes Local CG FEM based
             * assemblers - constant data (generic).
             * 
             * \warning <b>For the DEVELOPERS:</b> \p block_index for VariableInfo are not filled here, but \p indices_exist flag is set to \b TRUE (as it
             * is needed at other places before \p cell_matrix or \p cell_residual assembly).
             * \p Developers need to be wary of this fact that the \p block_indices are still not filled yet, and they need to <b>fill them before</b> doing \p cell_matrix assembly.
             */
            virtual void make_assemblers_generic_constant_data();
            
            /**
             * This function computes
             * <p> Local CG FEM based assemblers - constant data (cell) </p>
             * and allocates the memory for
             * shape functions, shape function gradients, and 
             * \p JxW_cell in
             * <p> Local CG FEM based assemblers - variable data (cell) </p>.
             */
            virtual void make_assemblers_cell_constant_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info);
            
            /**
             * This function computes
             * <p> Local CG FEM based assemblers - variable data (cell) </p>.
             */
            virtual void make_assemblers_cell_variable_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                            FuelCellShop::Layer::BaseLayer<dim>* const layer);
            
            /**
             * This function is specifically created for assembly of cell matrices for the following equations, \em viz.
             * - <b>"Proton Transport Equation"</b>
             * - <b>"Electron Transport Equation"</b>
             * - <b>"Ficks Transport Equation - oxygen"</b>
             * - <b>"Ficks Transport Equation - water"</b>
             * 
             * Cell matrix terms for the abovementioned equations follow a similar pattern, using different factors; hence this function avoids
             * using repeated lines of code for each of these equations in  #assemble_cell_matrix method. The argument list is defined as follows:
             * - first corresponds to \p AppFrame::MatrixVector which we need to fill.
             * - second corresponds to typename AppFrame::DoFApplication<dim>::CellInfo object; used to access fe elements for variables other than test function.
             * - second corresponds to \p std::string for the name of the equation for which we are doing the assembly. It should match with names mentioned above.
             * - third corresponds to \p FEValuesBase (fe discretization) used for the test function \em i.e. equation for which assembly is being done.
             * - fourth corresponds to shape functions for the test function, filled already by #make_assemblers_cell_variable_data.
             * - fifth corresponds to the factor used to multiply with the current density in source term for a particular equation.
             * 
             * \warning <b>For the DEVELOPERS:</b> This function should only be used after #make_assemblers_cell_variable_data has been called in #assemble_cell_matrix, otherwise it is bound to give wrong results.
             */            
            virtual void assemble_matrix_for_equation(AppFrame::MatrixVector& cell_matrices,
                                                      const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                      const std::string& eq_name,
                                                      const FEValuesBase<dim>& test_fe,
                                                      const std::vector< std::vector<double> >& test_shape_functions,
                                                      const double& sourceterm_factor);
            //@}
            
            ///@name Generic Constant Data
            //@{
            /**
             * VariableInfo structure corresponding to \p "oxygen_molar_fraction".
             */
            VariableInfo x_oxygen;
            
            /**
             * VariableInfo structure corresponding to \p "water_molar_fraction".
             */
            VariableInfo x_water;
            
            /**
             * VariableInfo structure corresponding to \p "electronic_electrical_potential".
             */
            VariableInfo phi_s;
            
            /**
             * VariableInfo structure corresponding to \p "protonic_electrical_potential".
             */
            VariableInfo phi_m;
            
            /**
             * VariableInfo structure corresponding to \p "temperature_of_REV".
             */
            VariableInfo t_rev;
            
            /**
             * Universal Faraday's constant
             */
            double F;
            
            //@}
            
            ///@name Local CG FEM based assemblers - variable data (cell)
            //@{
            
            /**
             * \f$ \mathbf{T} \f$ shape functions.
             * 
             * \p phi_T_cell \p[ \p q \p] \p[ \p k \p] denotes
             * \f$ k \f$-th \f$ \mathbf{T} \f$ shape function
             * computed in \f$ q \f$-th quadrature point of the cell.
             */
            std::vector< std::vector<double> > phi_T_cell;
            
            /**
             * \f$ \mathbf{x_{O_2}} \f$ shape functions.
             * 
             * \p phi_xOxygen_cell \p[ \p q \p] \p[ \p k \p] denotes
             * \f$ k \f$-th \f$ \mathbf{x_{O_2}} \f$ shape function
             * computed in \f$ q \f$-th quadrature point of the cell.
             */
            std::vector< std::vector<double> > phi_xOxygen_cell;
            
            /**
             * \f$ \mathbf{\phi_s} \f$ shape functions.
             * 
             * \p phi_phiS_cell \p[ \p q \p] \p[ \p k \p] denotes
             * \f$ k \f$-th \f$ \mathbf{\phi_s} \f$ shape function
             * computed in \f$ q \f$-th quadrature point of the cell.
             */
            std::vector< std::vector<double> > phi_phiS_cell;
            
            /**
             * \f$ \mathbf{\phi_m} \f$ shape functions.
             * 
             * \p phi_phiM_cell \p[ \p q \p] \p[ \p k \p] denotes
             * \f$ k \f$-th \f$ \mathbf{\phi_m} \f$ shape function
             * computed in \f$ q \f$-th quadrature point of the cell.
             */
            std::vector< std::vector<double> > phi_phiM_cell;
            
            /**
             * \f$ \mathbf{x_{H_2O}} \f$ shape functions.
             * 
             * \p phi_xWater_cell \p[ \p q \p] \p[ \p k \p] denotes
             * \f$ k \f$-th \f$ \mathbf{x_{H_2O}} \f$ shape function
             * computed in \f$ q \f$-th quadrature point of the cell.
             */
            std::vector< std::vector<double> > phi_xWater_cell;
            
            /**
             * Current produced [\p A/cm^3] at all quadrature points
             * in the cell.
             */
            std::vector<double> current_cell;
            
            /**
             * Derivative of current w.r.t. \f$ \mathbf{\phi_s} \f$
             * at all quadrature points in the cell.
             */
            std::vector<double> dcurrent_dphiS_cell;
            
            /** 
             * Derivative of current w.r.t. \f$ \mathbf{\phi_m} \f$
             * at all quadrature points in the cell.
             */
            std::vector<double> dcurrent_dphiM_cell;
            
            /** Derivative of current w.r.t. \f$ \mathbf{T} \f$
             * at all quadrature points in the cell.
             */
            std::vector<double> dcurrent_dT_cell;
            
            /**
             * Derivative of current w.r.t. \f$ \mathbf{x_{O_2}} \f$
             * at all quadrature points in the cell.
             */
            std::vector<double> dcurrent_dxOxygen_cell;
            
            /**
             * Derivative of current w.r.t. \f$ \mathbf{x_{H_2O}} \f$
             * at all quadrature points in the cell.
             */
            std::vector<double> dcurrent_dxWater_cell;
            
            /**
             * Active area [\p cm^2/cm^3] of the cell.
             * This is specifically required for heat source terms, as ReactionHeat return
             * heat terms in [\p W/cm^2].
             */
            double active_area_cell;
            
            /**
             * Factor for <p>Proton Transport Equation</p> source term, depending on whether the current cell
             * is in the Anode catalyst layer or Cathode catalyst layer.
             */
            double factor_protontranseq_cell;
            
            /**
             * Factor for <p>Electron Transport Equation</p> source term, depending on whether the current cell
             * is in the Anode catalyst layer or Cathode catalyst layer.
             */
            double factor_electrontranseq_cell;
            
            /**
             * Factor for <p>Ficks Transport Equation - oxygen</p> source term, depending on whether the current cell
             * is in the Anode catalyst layer or Cathode catalyst layer.
             */
            double factor_oxygentranseq_cell;
            
            /**
             * Factor for <p>Ficks Transport Equation - water</p> source term, depending on whether the current cell
             * is in the Anode catalyst layer or Cathode catalyst layer.
             */
            double factor_watertranseq_cell;
            
            //@}
            
            ///@name Counters
            //@{
            
            /**
             * Counter set to \em TRUE when \p cell_matrix is being assembled.
             * This ensures that only \b derivatives of source terms are computed in the cell, in #make_assemblers_cell_variable_data.
             * 
             * \note <b>For developers:</b> Other counter #cell_residual_counter should be set to \em FALSE, when this counter is set to \em TRUE.
             */
            bool cell_matrix_counter;
            
            /**
             * Counter set to \em TRUE when \p cell_residual is being assembled.
             * This ensures that only source terms, \em e.g. \p Current or \p Reaction \p Heat, are computed in the cell, in #make_assemblers_cell_variable_data.
             * 
             * \note <b>For developers:</b> Other counter #cell_matrix_counter should be set to \em FALSE, when this counter is set to \em TRUE.
             */
            bool cell_residual_counter;
            
            //@}
            
            /**
             * Variable used to store the index in cell_info->global_data of the previous Newton solution
             * The solution at the previous iteration is used to compute cell_matrix and cell_residual
             */  
             unsigned int last_iter_cell;            
            
        };
             
    } // Equation
             
} // FuelCellShop

#endif