//---------------------------------------------------------------------------
// C++ Interface: agglomerate_CL.h
//
// Description: Class characterizing the catalyst layer and defining effective properties
//
// Authors: Marc Secanell <secanell@ualberta.ca,
//		 Peter Dobson <pdobson@ualberta.ca>, (C) 2011
//		University of Alberta
//
// Copyright: See COPYING file that comes with this distribution
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__LAYER__AGGLOMERATE_CL__LAYER_H
#define _FUELCELLSHOP__LAYER__AGGLOMERATE_CL__LAYER_H

#include <boost/smart_ptr.hpp>

// Include deal.II classes
#include<base/parameter_handler.h>
#include<base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#include "base/timer.h"

// Include FCST classes
#include "utils/fcst_constants.h"
#include "utils/fcst_utilities.h"
#include "utils/fcst_db.h"

#include "layers/conventional_CL.h"
#include "layers/agglomerate_base.h"
#include "layers/agglomerate_ionomer_sun.h"


//Include STL
#include<cmath>
#include<iostream>
#include <stdexcept>
#include <map>

namespace FuelCellShop
{
    namespace Layer
    {
        /**
         * This class characterizes a catalyst layer and uses this information
         * to compute effective transport properties and interfacial areas for phase
         * change or electrochemical reactions.
         * 
         * @author M. Secanell, 2009-13
         * @author P. Dobson, 2009-11
         * @author M. Moore, 2010-12
         * @author M. Bhaiya, 2011-13
         * @author P. Wardlaw, 2012-13 
         * 
         */
        template <int dim>
        class AgglomerateCL :
        public ConventionalCL<dim>
        {
        public:
            
            /**
             * Concrete name used for objects of this class. This name is used when
             * setting up the subsection where the data is stored in the input file.
             * 
             * The data will be store under
             * \code
             * subsection name_specified_in_constructor
             *    set Material id = 2
             *    set Catalyst layer type = DummyCL # <-here I select the type of object of type CatalystLayer
             *    subsection DummyCL # <- this is the concrete_name for this class
             *       set all info relevant to this object
             *    end
             * end
             * \endcode
             */
            static const std::string concrete_name;
            
            ///@name Constructors, destructor, and initalization
            //@{    
            /**
             * Prototype Constructor
             * 
             * \warning For internal use only
             */
            AgglomerateCL ( );
            
            /**
             * Constructor
             * 
             * \warning To be private in release 1.1
             * 
             * \deprecated Use create_CatalystLayer
             */
            AgglomerateCL ( std::string name );
            
            /**
             * Destructor
             */
            ~AgglomerateCL();
            
            /**
             * Declare all necessary parameters in order to compute the coefficients
             * 
             * \deprecated Use declare_all_CatalystLayer_parameters
             */            
            void declare_parameters ( ParameterHandler &param ) const
            {
               declare_parameters(this->name, param);                
            }
            
            /**
             * Member function used to set new parameters values in the optimization loop
             * 
             * \deprecated Use set_all_CatalystLayer_parameters
             */
            void set_parameters ( const std::vector<std::string>& name_dvar, 
                                  const std::vector<double>& value_dvar, 
                                  ParameterHandler &param ) const
            {
               set_parameters(name_dvar, value_dvar, this->name, param);                
            }

            /**
             * Member function used to read in data and initialize the necessary data
             * to compute the coefficients.
             */
            void initialize ( ParameterHandler &param );

            //@}
            
            ///@name Accessors and info
            //@{
            /**
             * Inline member function that returns the number of agglomerates in the catalyst layer
             */
            inline double get_n () const
            {
                return n_agg;
            }
            
            /**
             * This member function will use a FuelCellShop::Kinetics class in order to compute the current density
             * production in the CL
             * Returns the current density at each quadrature point.  If the current is averaged over the cell,
             * the average current is assigned to each quadrature point in the cell.
             *
             * @param current is an empty vector to be filled with current density values.  Will be resized to the size of
             *              the solution given in set_solution()
             *
             */
            virtual void current_density ( std::vector<double>& current );
            
            /**
             * This member function will use a FuelCellShop::Kinetics class in order to compute the current density
             * production in the CL
             * Returns the current density at each quadrature point.  If the current is averaged over the cell,
             * the average current is assigned to each quadrature point in the cell.
             *
             * @param current is an empty vector to be filled with current density values.  Will be resized to the size of
             * 				the solution given in set_solution()
             * 
             * @param effectiveness is an empty vector to be filled with agglomerate effectiveness values.
             *
             */
            virtual void current_density (std::vector<double>& current, std::vector<double>& effectiveness );           
            
            /**
             * This member function will use a FuelCellShop::Kinetics class in order to compute the derivative of the current density
             * with respect to the variables setup in \ref set_derivative_flags (std::vector< std::string > &flags)
             *
             */
            virtual void derivative_current_density ( std::map< VariableNames, std::vector<double> >& );
            
            /**
             * Print out composition and micro-structural properties of the catalyst layer
             */
            virtual void print_layer_properties() const;
            
            inline void return_mesh_final(std::vector<double>& temp, std::vector<double>& temp1)
            { 
                temp = mesh_final;
                temp1 = co2_final;
                mesh_final.clear();
            }	 
            
            inline void set_mesh(std::vector<double>& temp)
            { mesh_final = temp;}	
            //@}
             

           
        protected:
            ///@name Constructors and declarations
            //@{
            /**
            * Constructor
            */
            //AgglomerateCL(const std::string& cl_section_name);

            /**
            * Constructor that also intialises the pointers to the catalyst, catalyst
            * support and electrolyte objects.
            */
            /*AgglomerateCL(const std::string& name, 
                          FuelCellShop::Material::PolymerElectrolyteBase*,
                          FuelCellShop::Material::CatalystSupportBase*,
                          FuelCellShop::Material::CatalystBase*);   
            */
            /**
             * Declare parameters for a parameter file.
             * 
             * Parameters that can be declared are defined in:
             * 
             * @code 
             * subsection Fuel cell data
             *  (...)
             *   subsection Cathode catalyst layer          <- This is the name of the CL subsection specified in cl_section_name
             *     (...)
             *     NOTE: Here you will need info on ConventionalCL since information from ConventionalCL is used for AgglomerateCL
             *     (...)
             *     subsection AgglomerateCL                 <- This is the subsection specified by concrete_name 
             *      set Agglomerate type = spherical        #["spherical|cylindrical"]
             *      set Agglomerate composition = ionomer   # [ionomer | water], Compositon of the agglomerate core
             *      set Agglomerate solver = numerical      # [analytical | numerical], Method to solve the agglomerate problem
             *      set Average current in cell = false     # Decide whether to take the average current density in the cell
             *      set Radius of the agglomerate [nm] = 200
             *      set Agglomerate porosity = 0.25
             *      set Agglomerate Loading Profile = 1     # A list of doubles of the form "1,2.2,3...", used to wait the Pt. loading profile
             *      set Conductivity factor = 1.0           # Allows the user to directly multiplay pronton conductivity, used for model observations
             *      set Initial condition tolerance factor = 0.05 # Tolerance factor for numerical agglomerate initial conditions
             *     end
             *   end
             * end
             * @endcode
             */
            void declare_parameters (const std::string& cl_section_name, 
                                     ParameterHandler &param) const;
                                     
            /**
             * Member function used to set new parameters values in the optimization loop
             *             
             */
            void set_parameters ( const std::vector<std::string>& name_dvar, 
                                  const std::vector<double>& value_dvar, 
                                  const std::string& cl_section_name, 
                                  ParameterHandler &param ) const;
            //@}
            
            ///@name Instance Delivery (Replica creator)
            //@{
            /**
             * This member function is used to create an object of type gas diffusion layer
             * 
             * \warning This class MUST be redeclared in every child.
             */
            virtual boost::shared_ptr<FuelCellShop::Layer::CatalystLayer<dim> > create_replica (const std::string &cl_section_name)
            {
                return boost::shared_ptr<FuelCellShop::Layer::CatalystLayer<dim> > (new FuelCellShop::Layer::AgglomerateCL<dim> (cl_section_name));
            }   
            //@}           
            ///@name Instance Delivery (Prototype)
            //@{
            /**
             *
             */            
            static AgglomerateCL<dim> const* PROTOTYPE;
            //@}
            
            ///@name Internal member functions
            //@{
            /**
             * Compute porosity and volume fraction of solid and ionomer in the catalyst layer
             */
            void compute_void_fraction();
            
            /**
             * Creates the agg objects populating the aggs object
             */
            void initialize_aggs();          
            
            /**
             * Function for creating shared pointer object of an unique (multithreadable) agglomerate object.
             */
            boost::shared_ptr<FuelCellShop::MicroScale::AgglomerateBase>  get_agg();
            
            /**
             * Member function to compute the number of agglomerates. The number of agglomerates is computed differently depending
             * on the type of agglomerate and depends on the agglomerate radius, amount of ionomer per agglomerate and the
             * solid phase volume fraction in the CL.
             */
            double compute_n() const;
            
            /**
             * Member function used to calculate the sensitivities of the number of agglomerates with respect to the
             * dependent variables.
             * \todo2 NOT WORKING YET -- Might need some thinking...
             *  Note that ideally I would have a "get" and a "compute"
             */
            //double derivative_n_agg() const;
            
            /** Member function to compute the thickness of the agglomerate thin film based on the radius and structure */
            double compute_thickness_agg ();
            
            /**
             * Inline function used to compute the amount of electrolyte in the catalyst layer. This is
             * given by the agglomerate sturcture and the amount of solid phase. In particular,
             * \f[
             * \epsilon_N^{cat} = n \frac{4.0}{3.0}3.1416({r_agg*1e-7})^{3.0}*(\epsilon_agg-1) + (r_{agg}1e-4 + \delta_{agg}1e-4)^{3.0};
             * \f]
             * where n is computed by another function. For details on this equation and other see, M.Secanell et al., "",...
             *
             */
            double epsilon_N_agg ( const double delta_agg ) const;
            
            /**
             * Inline function to compute
             *      \f[
             *      \frac{\partial \epsilon_N^{cat}}{\partial thickness_{agg}}
             *      \f]
             * This function should only be used to calculate the thin film thickness using
             * Newton's method.  Physically, the thickness of the thin film should not have an effect
             * on the volume fractions defined in ConventionalCL
             */
            double depsilonN_dthickness_agg ( const double thickness_agg ) const;
            
            /**
             * Private member functions for solving current density given an agglomerate.
             */
            double micro_scale_current( boost::shared_ptr <FuelCellShop::MicroScale::AgglomerateBase> &_agg ,  std::map<VariableNames ,SolutionVariable>& solutionMap, const unsigned int& index, double& E_r);
            
            /**
             * Private member functions for choosing agglomerate type.
             */
            void choose_agglomerate(boost::shared_ptr <FuelCellShop::MicroScale::AgglomerateBase> &agg_);
            
            /**
             * Private member functions for solving for current derivatives in an averaging approach.
             */
            void solve_current_derivatives_average(std::map< VariableNames, std::vector<double> >& Dcurrent);
            
            /**
             * Private member functions for solving for current derivatives in a per node approach.
             */
            void solve_current_derivatives_at_each_node(std::map< VariableNames, std::vector<double> >& Dcurrent);
            
            /*
             * Function for checking if our initial guess data is "up to date" within a certain tolerance.
             */
            bool check_micro_scale_initial_guess(const double &x_O2, const double &phi_m, const double &phi_s, 
                                                 const double lambda, const double radius, const double film_thickness, 
                                                 const double porosity);

            
            /*
             * Function for updating micro_scale_initial_guess, returns true if successful.
             */
            bool get_micro_scale_initial_guess(const double &x_O2, const double &phi_m, const double &phi_s, const double lambda, 
                                               boost::shared_ptr <FuelCellShop::MicroScale::AgglomerateBase> &_agg, const double radius,
                                               const double film_thickness, const double porosity);
            
            /*
             * Function for updating the database with new data.
             */
            void commit_data_to_db(std::vector<std::vector<double>> agg_data, std::vector<std::string> agg_column_titles, const double &x_R,
                                   const double &phi_m, const double &phi_s, const double lambda, const double radius, const double film_thickness, const double porosity);
            //@}
            
            ///@name Internal variables
            //@{

            /**  Bool for noting if the agglomerates have been initialized.*/
            bool aggs_initialized;
                     
            /** Type of agglomerate */
            std::string type_agg;
            
            /** Composition of the agglomerate, e.g. ionomer or water */
            std::string comp_agg;
            
            /**Double vector for storing agglomerate loading profiles.*/
            std::vector<double> agglomerateLoadingWeigths;
            
            /** Check to see if the agglomerate is water-filled */
            bool water_agg;
            
            /** Choose the analytical or numerical solver for the agglomerate problem */
            std::string solver_agg;
            
            /** Boolean value to choose whether to average the current over the cell */
            bool average_cell_current;

            /** Radius of the agglomerate */
            double r_agg;

            /** Thickness electrolyte film */
            double thickness_agg;

            /** Porosity of the agglomerate */
            double epsilon_agg;
            
            /** Number of agglomerates (computed from known quantities) */
            double n_agg;

            /** Pi */
            double pi;

            /** Stores the node locations used to solve the numerical agglomerate */
            std::vector<double> mesh_final;
            
            /** Stores the oxygen concentration profile across the agglomerate */
            std::vector<double> co2_final;
            
            /** Factor to be applied to the proton conductivity */
            double cond_factor;
            
            /** 
             * Vector of shared_ptr agg objects used for calculating current density and current density derivatives
             */
            std::vector<boost::shared_ptr<FuelCellShop::MicroScale::AgglomerateBase>> aggs; //Up to three aggs needed for parallel computation
            
            /** 
             * Storage for micro scale object (i.e. agglomerate) initial data
             */
            std::vector<std::vector<double>> micro_scale_initial_guess;
            
            /** 
             * Database object for getting and storing initial guess data
             */
            FcstUtilities::FCSTdatabase db;          
            
            //** Data for monitoring the relevance our current initial guess */
            double IC_x_R;
            double IC_phi_m;
            double IC_phi_s;
            double IC_lambda;
            double IC_tolerance;
            double IC_radius;
            double IC_film_thickness;
            double IC_porosity;
            //@}

        };
        
    } // Layer
    
}  // FuelCellShop

#endif // _FUELCELLSHOP__LAYER__AGGLOMERATE_CL__LAYER_H
    
