//---------------------------------------------------------------------------
// C++ Interface: agglomerate_base.h
//
// Description: Class characterizing the catalyst layer and defining effective properties
//
// Authors: Marc Secanell <secanell@ualberta.ca,
//		 Peter Dobson <pdobson@ualberta.ca>, (C) 2011
//		University of Alberta
//
// Copyright: See COPYING file that comes with this distribution
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__LAYER__AGGLOMERATE_BASE_H
#define _FUELCELLSHOP__LAYER__AGGLOMERATE_BASE_H

#include "fcst_constants.h"
#include "catalyst_base.h"
#include "catalyst_support_base.h"
#include "polymer_electrolyte_material_base.h"
#include "base_kinetics.h"
#include "tafel_kinetics.h"

#include "platinum.h"
#include "nafion.h"

#include <vector>
#include <string.h>

//Liquid water
#include "PureLiquid.h"

class AnalyticalAgglomerateTest;

namespace FuelCellShop
{
namespace MicroScale
{
/**
* \brief Class that serves as a base class for all agglomerate calculations
*/
	class AgglomerateBase
	{

	friend class ::AnalyticalAgglomerateTest;

	public:
		/** Constructor */
		AgglomerateBase()
		{
			//Constants
			R = Constants::R();
			F = Constants::F();
			permittivity_0 = Constants::E0() *1e-2; // cm
			pi = Constants::Pi();
			maxRadialDimension = 1.0;
			IC_loaded = false;
		}

		/** Destructor */
		virtual ~AgglomerateBase() {};

		/** Set the composition and structure of the agglomerate */
		virtual void set_structure ( FuelCellShop::Material::CatalystSupportBase*,
								FuelCellShop::Material::CatalystBase*,
								FuelCellShop::Material::PolymerElectrolyteBase*,
								double A, double r, double delta, double epsilon ) = 0;

		void set_kinetics ( FuelCellShop::Kinetics::BaseKinetics* kin )
		{
			kinetics = kin;
		};

		/**
		* Main function of the class used to compute the current over the whole agglomerate
		* at the local operating conditions
		*/
		virtual double compute_current (double &E_r) = 0;

		/**
		 * Function for setting the solution variables (reactant concentration, phi_s, phi_m, etc) before we compute current
		 */
		void set_solution(const std::map<VariableNames,SolutionVariable>&,const VariableNames&, const int&);

		/**
		 * Function to compute the derivative of the current density at the local operating conditions;
		 */

		virtual std::vector<double> compute_derivative_current ()
		{
		  Assert(false, ExcPureFunctionCalled());
		  return std::vector<double>(3, 0.0);
		}

		/** Set the pressure for the agglomerate */
		inline void set_P ( double P_Tot )
		{
			P = P_Tot;
		}

		/** Set the temperature for the agglomerate */
		inline void set_T ( double Temp )
		{
			T = Temp;
		}

		inline void setLoadingWeigths(std::vector<double> weigths)
		{
			loadingWeigths = weigths;
		}
		
		inline void set_conductivity_factor ( double temp )
		{
			cond_factor = temp;
		}
		
		inline bool analytical_derivatives() 
		{
			return analyticalDerivatives;
		}
		
		inline void load_initial_guess(std::vector<std::vector<double>> IC){
		    initial_data = IC;
		    IC_loaded = true;
		}

		inline bool guess_not_loaded(){
		    return !IC_loaded;
		}

		void deep_copy_pointers();



		std::vector<std::vector<double>> results;
		std::vector<std::string> column_titles;

    protected:



		inline void setAV(double newAV){
                AV = newAV;
        }

        double getAV(double location);

        void setUpLoadings();

        void interpolate_initial_data(double z[], double &x, double left_pos, double right_pos, std::vector<double> left_data, std::vector<double> right_data);

        bool use_initial_data(double z[], double &x);

		// Constants & V.ariables
		double permittivity_0;
		double pi;
		double F;
		double R;

		// Operating Conditions
		double P;
		double T;

		// Kinetic parameters
		std::vector<double> loadingWeigths;
		std::vector<double> actualLoadings;
		std::vector<double> loadingRadii;
		double AV;

		double maxRadialDimension;



		bool IC_loaded;

		/** Variable to check if analytical derivatives are implemented */
		bool analyticalDerivatives;

		/**
		* Object used to calculate the carbon black conductivity in the catalyst layer.
		*/
		FuelCellShop::Material::CatalystSupportBase *catalyst_support;

		/**
		* Object used to store the properties of the catalyst used in the layer.
		*/
		FuelCellShop::Material::CatalystBase *catalyst;

		/**
		* Object used to calculate the properties of the electrolyte in the catalyst layer.
		* In this case we assume is Nafion.
		*/
		FuelCellShop::Material::PolymerElectrolyteBase *electrolyte;

		/** Create the kinetics object */
		FuelCellShop::Kinetics::BaseKinetics *kinetics;
		
		/** 
		 * Vector storing the oxygen concentration profile across the agglomerate. Used for 
		 * to ensure the numerical agglomerate matches the analytical agglomerate. 
		*/
		std::vector<double> co2_final;
		/** 
		 * Vector storing the location of the nodes of the mesh used in the numerical agglomerate. 
		 * This can then be passed to the analytical agglomerate so that the oxygen concentration
		 * can be found at the same locations. 
		*/
		std::vector<double> mesh_final;
		
		/** Used to modify the proton conductivity in agglomerate */
		double cond_factor;

		/** Radius of the agglomerate */
        double r_agg;

        /** Thickness of the ionomer thin film */
        double delta_agg;

        /** Thickness of the ionomer thin film */
        double epsilon_agg;

        /** Concentration of oxygen at the boundary */
        double c_R;

        /** Electrolyte (Membrane) phase potential at the boundary */
        double phi_M;

        /** Solid phase potential through the agglomerate */
        double phi_S;

        /** Value of the boundary between the thin film and agglomerate domain */
        double interface;

        /** Oxygen Diffusion */
        double D_R_N;

        /** Henry's Constant */
        double H_R_N;

        /** absolute tolerance in reactant concentration */
        double R_tol;

        std::vector<VariableNames> sol_names;

        std::map<VariableNames, SolutionVariable> solutions;
        VariableNames reactant;
        int sol_index;

        std::vector< std::vector<double>> initial_data;
	};
}
}

#endif
