//---------------------------------------------------------------------------
// C++ Interface: agglomerate_ionomer_analytical.h
//
// Description: Used to solve a system of equations representing a
// 			spherical ionomer-filled agglomerate.
//
// Author: Peter Dobson <pdobson@ualberta.ca>, (C) 2011
//			University of Alberta
//
// Copyright: See COPYING file that comes with this distribution
//
//---------------------------------------------------------------------------

#ifndef FUEL_CELL__IONOMER_AGGLOMERATE_SUN__H
#define FUEL_CELL__IONOMER_AGGLOMERATE_SUN__H

//------------------------------
// STD LIBRARY DECLARATIONS
//------------------------------
#include <cmath>
#include <iostream>
#include <fstream>

//------------------------------
// DEAL.II DECLARATIONS
//------------------------------
#include "base/logstream.h"

//------------------------------
// FUEL CELL DECLARATIONS
//-----------------------------
#include "catalyst_base.h"
#include "catalyst_support_base.h"
#include "polymer_electrolyte_material_base.h"
#include "agglomerate_base.h"

namespace FuelCellShop
{
    namespace MicroScale
    {
        /**
         * \brief Analytical solution to an ionomer-filled agglomerate problem in 1D
         *
         * Adapted from Source:
         * W. Sun, B. A. Peppley, K. Karan, An improved two-dimensional agglomerate
         * cathode model to study the influence of catalyst layer structural parameters,
         * Electrochimica Acta 50 (16-17) (2005) 3347–3358.
         */
    class IonomerAgglomerateSun : public AgglomerateBase
        {
            public:
                /** Constructor */
                IonomerAgglomerateSun ();

                /** Set the composition and structure of the agglomerate */
                virtual void set_structure ( FuelCellShop::Material::CatalystSupportBase* CS,
                        FuelCellShop::Material::CatalystBase* Cat,
                        FuelCellShop::Material::PolymerElectrolyteBase* PEM,
                        double A, double r, double delta, double epsilon);

                /**
                 * Main function of the class used to compute the current over the whole agglomerate
                 * at the local operating conditions
                 */
                virtual double compute_current ( double &E_r);

                /**
                 * Function to compute the derivative of the current density at the local operating conditions;
                 */
                virtual std::vector<double> compute_derivative_current ();




            private:

                void _initialize();

                /** Function to compute the effectiveness of the agglomerate core */
                double compute_Er (const double k_c, const double D);

                /** Function to compute the derivative of the effectiveness of the agglomerate core */
                double compute_dEr (const double k_c,const double dk_c, const double D);


        };// class IonomerAgglomerateSun

    } // namespace Layer

} // namespace FuelCellShop

#endif

