//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: conventional_cl.h
//    - Description: Class characterizing the conventional catalyst layer and methods for computing effective properties. It also provides interface to various material classes used in catalyst layer.
//    - Developers: Peter Dobson (2011) and Madhur Bhaiya (2013)
//    - Id: $Id: conventional_CL.h 1465 2013-08-30 22:42:25Z madhur $ 
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__CONVENTIONAL_CL__H
#define _FUELCELLSHOP__CONVENTIONAL_CL__H

// Include deal.II classes
#include<base/parameter_handler.h>
#include<base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

// Include FCST classes
#include "catalyst_layer.h"
#include "polymer_electrolyte_material_base.h"
#include "catalyst_support_base.h"
#include "catalyst_base.h"
#include "base_kinetics.h"
#include "platinum.h"
#include "nafion.h"
#include "carbon.h"
#include "geometries.h"
#include "boost/shared_ptr.hpp"

//Include STL
#include <cmath>
#include <iostream>
#include <algorithm>


namespace FuelCellShop
{
    namespace Layer
    {
        /**
         * This class characterizes a catalyst layer and uses this information 
         * to compute effective transport properties and interfacial areas for phase
         * change or electrochemical reactions.
         * 
         * This class implements a macrohomogeneous catalyst layer.
         *
         */
        template <int dim>
        class ConventionalCL :
        public CatalystLayer<dim>
        {
        public:
            
            /**
             * Concrete name used for objects of this class. This name is used when
             * setting up the subsection where the data is stored in the input file.
             * 
             * The data will be store under
             * \code
             * subsection name_specified_in_constructor
             *    set Material id = 2
             *    set Catalyst layer type = DummyCL # <-here I select the type of object of type CatalystLayer
             *    subsection DummyCL # <- this is the concrete_name for this class
             *       set all info relevant to this object
             *    end
             * end
             * \endcode
             */
            static const std::string concrete_name;
            
            ///@name Constructors, destructor, and initalization
            //@{                
            /** \warning For internal use only.
             * 
             * Constructor used only to create a prototype. Do not use
             * in general since this will not include the name of the section
             * in the parameter file you need. 
             */
            ConventionalCL();
            
            /**
             * Destructor
             */
            ~ConventionalCL();
                       
            //@}
            
            ///@name Accessors and info
            //@{
            
            /**
             * Print out the volume fraction in the catalyst layer
             */
            virtual void print_layer_properties() const;
            
            /** Get the volume fractions in the catalyst layer */
            virtual inline void get_volume_fraction(double &epsilon_S_out, 
                                                    double &epsilon_V_out, 
                                                    double &epsilon_N_out) const
            {
                epsilon_S_out = epsilon_S;
                epsilon_V_out = epsilon_V;
                epsilon_N_out = epsilon_N;
            }
            
            /** Return the platinum loading per cm3 catalyst layer */
            virtual inline double get_V_Pt() const
            {
                return V_Pt;
            }
            
            /** Get the active area of platinum per unit volume of CL */
            inline double get_active_area_Pt() const
            {
                return Av;
            }
            
            //@}
            ///@name Effective property calculators
            //@{           
            /**
             * Compute the effective diffusivty in the CL. This routine takes the 
             * gas diffusivity from FuelCellShop::BinaryDiffusion and transforms
             * it into an effective property taking into account the porosity and 
             * structure of the CL
             */
            virtual void effective_gas_diffusivity(const double& , double& ) const;
            
            /**
            * Compute the effective diffusivty in the CL. This routine takes the 
            * gas diffusivity from FuelCellShop::BinaryDiffusion and transforms
            * it into an effective property taking into account the porosity and 
            * structure of the CL
            */
            virtual void effective_gas_diffusivity(Table< 2, Tensor< 2, dim > >&) const;
            
            /**
             * Compute the effective electron conductivity in the CL
             */
            virtual void effective_electron_conductivity(double& ) const;
            
            /**
             * Compute the effective electron conductivity in the CL as an
             * anisotropic tensor.
             */
            virtual void effective_electron_conductivity(Tensor<2,dim>& ) const;
            
            /**
             * Compute the derivative of the effective electron conductivity in the GDL
             * with respect to either the solution or design parameters. The parameters with respect to
             * which the derivatives are computed are setup in FuelCellShop::Layer::set_derivative_flags()
             */
            virtual void derivative_effective_electron_conductivity(std::vector<double>& ) const;
            
            /**
             * Compute the effective proton conductivity in the CL.
             */
            virtual void effective_proton_conductivity(double& ) const;
            virtual void effective_proton_conductivity(std::vector<double>& ) const;
            /**
             * Compute the derivative of the effective proton conductivity in the CL
             * with respect to either the solution or design parameters. The parameters with respect to
             * which the derivatives are computed are setup in FuelCellShop::Layer::set_derivative_flags()
             */
            virtual void derivative_effective_proton_conductivity(std::map< VariableNames, std::vector<double> >& ) const;
            
            /**
             * Compute the effective water diffusivity (lambda diffusivity) in the CL.
             */
            virtual void effective_water_diffusivity(double& ) const;
            virtual void effective_water_diffusivity(std::vector<double>& ) const;
            /**
             * Compute the derivative of the effective water diffusivity (lambda diffusivity) in the CL
             * with respect to either the solution or design parameters. The parameters with respect to
             * which the derivatives are computed are setup in FuelCellShop::Layer::set_derivative_flags()
             */
            virtual void derivative_effective_water_diffusivity(std::map< VariableNames, std::vector<double> >& ) const;

            //@}
            
        protected:
            ///@name Constructors
            //@{
                /**
             * Constructor
             */
            ConventionalCL(std::string name);
            
            /**
             * Declare parameters for a parameter file.
             * 
             * @code 
             * subsection Fuel cell data
             *   (...)
             *   subsection Cathode Catalyst Layer #<- This is the name fo the subsection that you specify in cl_section_name
             *     (...)
             *     subsection ConventionalCL #<- This is the name in (this->concrete_name)
             *       set Platinum loading on support (%wt) = 0.46                    # Mass percentage of platinum catalyst on the support carbon black
             *       set Platinum loading per unit volume (mg/cm3) = 400             # Catalyst platinum mass loading per unit volume of CL
             *       set Electrolyte loading (%wt)  = 0.3                            # Electrode loading is the weight percentage of ionomer per gram of CL
             *       //-- Network characteristics
             *       set Method effective transport properties in pores = Bruggemann # OPTIONS: Given|Bruggemann|Percolation -- Method used to compute effective transport properties in the void phase.
             *       set Porosity threshold  = 0.12                                  # Threshold value of the volume fraction of void space in the CL. If the porosity is less than this value transport does not occur
             *       set Porosity network constant = 2.0                             # Parameter used when using percolation theory      
             *       set Porosity gamma network constant = 0.0                       # Parameter used when using percolation theory to account for extra diffusion
             *       //--
             *       set Method effective transport properties in solid phase = Bruggemann # OPTIONS: Given|Bruggemann|Percolation --- Method used to compute effective transport properties in pores
             *       set Solid network threshold = 0.12                              # Threshold value of the volume fraction of solid (electron conductive) phase in the CL. If the solid phase is less than this value transport in the fibre network does not occur
             *       set Solid network constant = 2.0                                # Parameter used when using percolation theory
             *       set Method effective transport properties in electrolyte phase = Bruggemann # OPTIONS: Given|Bruggemann|Percolation|Iden11 -- Method used to compute effective transport properties in pores
             *       set Electrolyte network threshold = 0.12                        # Threshold value of the volume fraction of electrolyte (proton conductive) phase in the CL. If the electrolyte phase is less than this value transport in the network does not occur
             *       set Electrolyte network constant = 2.0                          # Parameter used when using percolation theory
             *       set Method to compute active area = given                       # OPTIONS: given|Marr|ETEK06|ETEK07 -- 
             *       set Active area [cm^2/cm^3] = 2.0e5
             *       set Method to compute porosity = marc                           #OPTIONS: marc
             *     end
             *   end 
             * end
             * @endcode
             */
            void declare_parameters (const std::string& cl_section_name, 
                                     ParameterHandler &param) const;
                                     
            /**
             * Member function used to set new parameters values in the optimization loop
             *             
             */
            void set_parameters ( const std::vector<std::string>& name_dvar, 
                                  const std::vector<double>& value_dvar, 
                                  const std::string& cl_section_name, 
                                  ParameterHandler &param ) const;
                                  
            /**
             * Member function used to read in data and initialize the necessary data
             * to compute the coefficients.
             */
            void initialize (ParameterHandler &param);
            //@}
            /**
             * Compute porosity and volume fraction of solid and ionomer in the catalyst layer
             */
            void compute_volume_fraction();
            
            /**
             * Compute the active area of catalyst in the layer by the specified method
             */ 
            void compute_Av();
            /**
             * Compute the derivative of the effective proton conductivity w.r.t. the electrolyte
             * loading.
             */
            void derivative_effective_proton_conductivity_wrt_electrolyte_loading(double& ) const;
            
            /**
             * Function to compute the partial derivative of the volume fraction 
             * the different phases in the catalyst layer with respect to 
             * the design variables of the optimization problem
             */
            void derivative_volume_fractions(double &Depsilon_S, 
                                             double &Depsilon_V, 
                                             double &Depsilon_N) const;
                                             
            /**
            * Get the effective transport method in the pores
            */
            void get_method_transport_property_pores(std::string& method)
            {
                method = method_eff_property_pores;
            };
            
            /**
            * Get the effective transport method in the electrolyte
            */
            void get_method_transport_property_electrolyte(std::string& method)
            {
                method = method_eff_property_electrolyte;
            };
            
            /**
            * Get the effective transport method in the solid phase
            */
            void get_method_transport_property_solid(std::string& method)
            {
                method = method_eff_property_solid;
            };
            
            /**
            * Inline function to compute
            * \f[
            * \frac{\partial \epsilon_S^{cat}}{\partial \%Pt}
            * \f]
            */
            inline double depsilon_S_cat_dprc_Pt(const double V_Pt, 
                                                const double prc_Pt) const
            {
                return -(V_Pt*1e-3)/(rho_c*pow(prc_Pt,2.0));
            }
            
            /**
            * Inline function to compute
            * \f[
            * \frac{\partial \epsilon_S^{cat}}{\partial m_{Pt}}
            * \f]
            */
            inline double depsilon_S_cat_dVPt(const double prc_Pt) const
            {
                return  (1/rho_Pt + (1-prc_Pt)/(prc_Pt*rho_c))*(1e-3);
            }
            
            /**
            * Inline function to compute
            * \f[
            * \frac{\partial \epsilon_V^{cat}}{\partial \epsilon_S^{cat}}
            * \f]
            */
            inline double depsilon_V_cat_depsilon_S_cat() const
            {
                return -1;
            }
            
            /**
            * Inline function to compute
            * \f[
            * \frac{\partial \epsilon_V^{cat}}{\partial \epsilon_S^{cat}}
            * \f]
            */
            inline double depsilon_V_cat_depsilon_N_cat() const
            {
                return -1;
            }
            
            //-- Composition
            /** Volume fraction of Nafion in the cathode catalyst layer */
            double epsilon_N;
            /** Void volume fraction (Porosity) of the catalyst layer */
            double epsilon_V;
            /** Solid volume fraction in the catalyst layer */
            double epsilon_S;
            /** Volume fraction of water in the cathode catalyst layer */
            double epsilon_W;
            
            //-- Catalyst properties
            /** Density of platinum */
            double rho_Pt;
            /** Density of support material */
            double rho_c;
            /** Percentage of platinum per carbon on the catalyst layer */
            double prc_Pt;
            /** Platinum loading at the catalyst layer per unit volume */
            double V_Pt;
            /** Platinum loading at the catalyst layer per unit area */
            double M_Pt;
            /** Active area of catalyst per unit volume of catalyst layer */
            double Av;
            /** Method to compute active area */
            std::string method_Av;
            /** Method to compute porosity */
            std::string method_porosity;
            /** Layer thickness*/
            double L_CL;
            
            //-- Electrolyte properties
            /** Density of electrolyte */
            double rho_N;
            /** Electrolyte loading. Electrode loading is the weight percentage 
            * of ionomer per gram of CL
            * loading_N = weight electrolyte / (weight Pt + weight C + weight electrolyte)
            */
            double loading_N;
            /** Percentage (mass fraction) of electrolyte in the catalyst layer */
            double prc_N;
            
            // Network characteristics
            /** Method used to compute effective properties -- Type of network */
            std::string method_eff_property_pores;
            /** Porous network threshold */
            double porosity_th;
            /** Porous network constant */
            double porosity_mu;
            /** */
            double porosity_gamma;
            
            /** Method used to compute effective properties -- Type of network */
            std::string method_eff_property_solid;
            /** Input electron conductivity */
            double solid_electron_conductivity;
            /** Solid phase network threshold */
            double solid_th;
            /** Solid phase network constant */
            double solid_mu;
            /** Method used to compute effective properties -- Type of network */
            std::string method_eff_property_electrolyte;
            /** Input electrolyte proton conductivity */
            double electrolyte_proton_conductivity;
            /** Electrolyte network threshold */
            double electrolyte_th;
            /** Electrolyte network constant */
            double electrolyte_mu;
            /** Electrolyte proton conductivity */
            double sigma_p;
            
            /** Theoretical open-circuit voltage under the operating conditions */
            double E0;
        };
    }
}

#endif