//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: design_fibrous_GDL.h
//    - Description: Class used to represent a fibrous GDL where effective properties are computed based on the porosity etc.
//    - Developers: M. Secanell and Madhur Bhaiya
//    - $Id: design_fibrous_GDL.h 1460 2013-08-30 22:28:49Z madhur $ 
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__DESIGN_FIBROUS_GDL_H
#define _FUELCELLSHOP__DESIGN_FIBROUS_GDL_H

// FCST classes
#include <base_material.h>
#include <gas_diffusion_layer.h>

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

//Include STL
#include<cmath>
#include<iostream>

using namespace dealii;

namespace FuelCellShop
{
    namespace Layer
    {
        /**
         * This class defines a GDL made of fibres. Due to this, the conductivity in the X and Y direction is 
         * extremely anisotropic since the contact resistance is very high. Given this situation, I decided to
         * use a different value for the bulk resistance in the X and Y direction.
         *
         * @author M. Secanell and M. Bhaiya
         * 
         */
        template <int dim>
        class DesignFibrousGDL : 
        public  GasDiffusionLayer<dim>
        {
        public:

            /**
             * Concrete name used for objects of this class. This name is used when
             * setting up the subsection where the data is stored in the input file.
             * 
             * The data will be store under
             * \code
             * subsection name_specified_in_constructor
             *    set Material id = 2
             *    set Gas diffusion layer type = DummyGDL # <-here I select the type of object of type GasDiffusionLayer
             *    subsection DummyGDL # <- this is the concrete_name for this class
             *       set all info relevant to this object
             *    end
             * end
             * \endcode
             */
            static const std::string concrete_name;
            
            ///@name Destructor, and initalization
            //@{    
            /** 
             * Replica Constructors
             * 
             * \warning For internal use only.
             * 
             * Constructor used only to create a prototype. Do not use
             * in general since this will not include the name of the section
             * in the parameter file you need. 
             * 
             */
            DesignFibrousGDL();
            
            /** 
             * \note Eventually, I would like to make this private.
             * 
             * \deprecated Use create_GasDiffusionLayer
             */
            DesignFibrousGDL(const std::string& name);
            
            /** Destructor */
            ~DesignFibrousGDL()
            {}
            
            /**
             * Declare parameters for a parameter file
             * 
             * \deprecated Use declare_all_GasDiffusionLayer_parameters
             */
            void declare_parameters (ParameterHandler &param) const
            {
                declare_parameters(this->name, param);
            };
            
            /**
             * Member function used to set new parameters values in the optimization loop
             * 
             * \deprecated Use set_all_GasDiffusionLayer_parameters
             */
            virtual void set_parameters (const std::vector<std::string>& name_dvar,
                                         const std::vector<double>& value_dvar,
                                         ParameterHandler &param) const {}
                                 
            /** Member function used to read in data and initialize the necessary data
             * to compute the coefficients. */
            virtual void initialize (ParameterHandler &param);
            //@}
            
             ///@name Effective properties
            //@{               
            /**
             * Compute effective diffusivity of the media.
             * \deprecated
             */
            virtual void effective_gas_diffusivity(const double& prop, double& prop_eff) const;
            /**
             * Compute effective diffusivity of the media.
             * \deprecated
             */
            virtual void effective_gas_diffusivity(const double& prop, Tensor<2,dim>& prop_eff) const;
            
            /**
             * Compute effective diffusivity of the media.
             * 
             */
            virtual void effective_gas_diffusivity(Table< 2, Tensor<2,dim> > &D_eff) const;
            
            /**
             * Compute the electrical conductivity of the media.
             */
            virtual void effective_electron_conductivity(double&) const;
            
            /**
             * Compute the electrical conductivity of the media.
             */
            virtual void effective_electron_conductivity(Tensor<2,dim>&) const;
            
            //@}

        private:
            ///@name Constructors, destructor, and initalization
            //@{    
            /** Declare parameters for a parameter file. The parameters that need to be declared are 
             * The parameters that need to be declared are
             * - Porosity (default : 0.3) Represents *the porosity in the GDL
             * - Method effective transport properties in pores (default: Bruggemann) Other options "Given|Bruggemann|Percolation|Mezedur"
             * - Method effective transport properties in solid (default: Bruggemann) Other options "Given|Bruggemann|Percolation|Mezedur"
             *
             * - Electrical conductivity X
             * - Electrical conductivity Y
             * - Electrical conductivity Z
             * 
             * An example input file would look as follows:
             * 
             * @code
             * subsection Cathode gas diffusion layer
             *    set Material id = 2
             *    set Gas diffusion layer type = DesignFibrousGDL
             *    subsection DesignFibrousGDL
             *      ######### Composition: #########
             *      set Porosity = 0.6
             *      ######### Gas transport #########
             *      ## Anisotropy
             *      set Anisotropic transport = true                  # (default) false
             *      set Method effective transport properties in pores = Tomadakis    # (default) Bruggemann | Given | Percolation | Tomadakis | Mezedur
             *      set Method effective transport properties in solid = Percolation  # (default) Bruggemann | Given | Percolation
             *      ## XX
             *      set Porosity threshold X = 0.11               # (default) 0.12 | 0.118 (Peter's Thesis) | 0.11
             *      set Porosity network constant X = 0.785           # (default) 2.0  |0.785 (Peter's Thesis)  [Page 69]
             *      set Porosity gamma network constant X = 0.0           # (default) 0.0 | 
             *      #
             *      set Electrical conductivity X = 16.03         # [S/cm]
             *      set Solid network threshold X = 0.0               # (default) 0.12 | 
             *      set Solid network constant X = 1.5                # (default) 2.0 |
             *      ## YY
             *      set Porosity threshold Y = 0.11               # (default) 0.12 | 0.118 (Peter's Thesis) | 0.11
             *      set Porosity network constant Y = 0.521           # (default) 2.0 |
             *      set Porosity gamma network constant Y = 0.0           # (default) 0.0 |
             *      #
             *      set Electrical conductivity Y = 272.78            # [S/cm]
             *      set Solid network threshold Y = 0.0               # (default) 0.12 |
             *      set Solid network constant Y = 1.0                # (default) 2.0 |
             *    end
             *  end
             * @endcode
             * 
             */
            virtual void declare_parameters (const std::string& name, 
                                             ParameterHandler &param) const;
                                             
             /**
             * Member function used to set new parameters values in the optimization loop
             * 
             * \deprecated Use set_all_GasDiffusionLayer_parameters
             */
            virtual void set_parameters (const std::vector<std::string>& name_dvar,
                                         const std::vector<double>& value_dvar,
                                         const std::string& name,
                                         ParameterHandler &param) const 
            {
                const std::type_info& info = typeid(*this);
                deallog << "Pure function " << __FUNCTION__
                << " called in Class "
                << info.name() << std::endl;                            
                                             
            }
            //@}                     
            ///@name Instance delivery function
            //@{
            /**
             * This member function is used to create an object of type gas diffusion layer
             * 
             * \warning This class MUST be redeclared in every child.
             */
            virtual boost::shared_ptr<FuelCellShop::Layer::GasDiffusionLayer<dim> > create_replica (const std::string &name)
            {
                return boost::shared_ptr<FuelCellShop::Layer::GasDiffusionLayer<dim> > (new FuelCellShop::Layer::DesignFibrousGDL<dim> (name));
            }
            //@}           
            
            ///@name Internal variables
            //@{
            //--------------------------            
            /** Porosity of the GDL */
            double porosity;
            /** Volume fraction of solid phase, i.e. solid */
            double solid_phase;
            /** Method used to compute the effective properties in the pores */
            std::string method_eff_property_pores;
            /** Method used to compute the effective properties in the solid */
            std::string method_eff_property_solid;
            /** Electrical conductivity "bulk" solid material */
            std::vector<double> sigma_e;
            
            //----------------------
            /** Anisotropy ? */
            bool anisotropy;
            /** Anisotropic properties */
            /** Porosity of the GDL threshold */
            std::vector<double> porosity_th;
            /** Network constant */
            std::vector<double> porosity_mu;
            /** Network constant gamma */
            std::vector<double> porosity_gamma;
            /** Solid network of the GDL threshold */
            std::vector<double> solid_th;
            /** Solid network constant */
            std::vector<double> solid_mu;
            //@}
            
            ///@name Instance delivery object
            //@{
            /**
             * Prototype declaration
             */             
            static DesignFibrousGDL<dim> const* PROTOTYPE;
            //@}
        };
    }
}

#endif
