//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: nafion_membrane.h
//    - Description: Class representing Nafion membrane layer class - returning effective transport properties
//    - Developers: Madhur Bhaiya (2012-13)
//    - Id: $Id: nafion_membrane.h 1460 2013-08-30 22:28:49Z madhur $ 
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__NAFION_MEMBRANE_H
#define _FUELCELLSHOP__NAFION_MEMBRANE_H

// Include FCST classes
#include "layers/membrane_layer.h"
#include "materials/polymer_electrolyte_material_base.h"
#include "materials/nafion.h"

#include "boost/shared_ptr.hpp"

namespace FuelCellShop
{
    namespace Layer
    {
        /**
         * This class implements the necessary information for a Nafion membrane.
         * 
         * It contains information regarding the material_id used in the mesh to assign cells
         * with that belong to the membrane and provides an interface to compute
         * bulk membrane properties based on the FuelCellShop::Material::Nafion class
         *
         * <h3>Usage Details:</h3>
         * 
         * For usage information please see MembraneLayer documentation.
         * 
         * \author M. Secanell, 2009-13
         * \author M. Bhaiya, 2013 
         */
        template <int dim>
        class NafionMembrane : public MembraneLayer<dim>
        {
        public:
            /**
             * Concrete name used for objects of this class. This name is used when
             * setting up the subsection where the data is stored in the input file.
             * 
             * The data will be store under
             * \code
             * subsection name_specified_in_constructor
             *    set Material id = 2
             *    set Membrane type =  NafionMembrane # <-here I select the type of object of type GasDiffusionLayer
             *    subsection NafionMembrane # <- this is the concrete_name for this class
             *       set all info relevant to this object
             *    end
             * end
             * \endcode
             */
            static const std::string concrete_name;
            
            ///@name Constructors, destructor, and initalization
            //@{    
            /** 
             * Replica Constructors
             * 
             * \warning For internal use only.
             * 
             * Constructor used only to create a prototype. Do not use
             * in general since this will not include the name of the section
             * in the parameter file you need. 
             * 
             */
            NafionMembrane();
                        
            /**
            * Destructor
            */
            ~NafionMembrane();          
            
            //@}
            
            ///@name Accessors and info
            //@{
            /**
            * Compute the constant effective proton conductivity of the membrane.
            */
            virtual void effective_proton_conductivity(double&) const;
            /**
            * Compute the effective proton conductivity of the membrane at every quadrature point,
            * as a function of solution variables.
            */
            virtual void effective_proton_conductivity(std::vector<double>&) const;
            /**
            * Compute the derivative of effective proton conductivity of the membrane with respect to
            * the flags set by set_derivative_flags, at every quadrature point.
            */
            virtual void derivative_effective_proton_conductivity(std::map< VariableNames, std::vector<double> >&) const;
            
            /**
            * Compute the constant effective water diffusivity of the membrane.
            */
            virtual void effective_water_diffusivity(double&) const;
            /**
            * Compute the effective water diffusivity of the membrane at every quadrature point,
            * as a function of solution variables.
            */
            virtual void effective_water_diffusivity(std::vector<double>&) const;
            /**
            * Compute the derivative of effective water diffusivity of the membrane with respect to
            * the flags set by set_derivative_flags, at every quadrature point.
            */
            virtual void derivative_effective_water_diffusivity(std::map< VariableNames, std::vector<double> >&) const;

            /**
            * Compute the constant effective oxygen diffusivity of the membrane.
            */
            virtual void effective_oxygen_diffusivity(double&) const;
            /**
            * Compute the effective oxygen diffusivity of the membrane at every quadrature point,
            * as a function of Temperature.
            * \note This method should only be used when Temperature is one of the solution variables.
            */
            virtual void effective_oxygen_diffusivity(std::vector<double>&) const;
            /**
            * Compute the derivative of effective oxygen diffusivity of the membrane with respect to
            * the flags set by set_derivative_flags, at every quadrature point.
            * \note This method should only be used when Temperature is one of the solution variables.
            */
            virtual void derivative_effective_oxygen_diffusivity(std::map< VariableNames, std::vector<double> >&) const;
            
            //@}
        
        protected:
            ///@name Construtor and initalization (Private)
            //@{
            /**
             * Constructor
             * 
             */
            NafionMembrane(std::string name);
            
            /**
             * Declare parameters for a parameter file.
             * 
             * Nothing to declare here other than the declaration of the base layer. 
             * However, the member function must be implemented so that the layer has the
             * same structure as other layers.
             * 
             */
            void declare_parameters (const std::string& name, 
                                     ParameterHandler &param) const;
                                     
            /**
             * Member function used to set new parameters values in the optimization loop
             * 
             */
            void set_parameters (const std::vector<std::string>& name_dvar,
                                 const std::vector<double>& value_dvar, 
                                 const std::string& name, 
                                 ParameterHandler &param) const;
                                 
            /**
            * Member function used to read in data and initialize the necessary data
            * to compute the coefficients.
            */
            void initialize (ParameterHandler &param);
            
            //@}
            ///@name Instance Delivery Functions
            //@{         
            /**
             * This member function is used to create an object of type gas diffusion layer
             * 
             * \warning This class MUST be redeclared in every child.
             */
            virtual boost::shared_ptr<FuelCellShop::Layer::MembraneLayer<dim> > create_replica (std::string &name)
            {
                return boost::shared_ptr<FuelCellShop::Layer::MembraneLayer<dim> > (new FuelCellShop::Layer::NafionMembrane<dim> (name));
            }            
            //@}                        
            ///@name Instance Delivery Data
            //@{         
            /**
             * Create prototype for the layer
             */            
            static NafionMembrane<dim> const* PROTOTYPE;
            //@}                 
        };
    }
}

#endif
