//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: porous_layer.h
//    - Description: Child of base layer that implements common functionality for handling gases.
//    - Developers: M. Secanell and Madhur Bhaiya
//    - $Id: porous_layer.h 1373 2013-08-21 00:34:53Z madhur $
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__POROUS__LAYER_H
#define _FUELCELLSHOP__POROUS__LAYER_H

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

//Include STL
#include "cmath"
#include "iostream"
#include "algorithm"

// FCST
#include "base_layer.h"
#include "PureGas.h"
#include "BinaryDiffusion.h"

using namespace dealii;

namespace FuelCellShop
{
    namespace Layer
    {
        /**
         * Virtual class used to implement properties that are characteristic of a void layer or a porous layer.
         * This layer is used to setup information needed to compute molecular diffusivities and Knudsen diffusivities (pending implementation)
         * for any gases. Therefore, it provides member functions to specify the gases that exist within the layer as well as their
         * temperature and pressure.
         * 
         * It also inherits all properties of BaseLayer.
         * 
         * Note you cannot make an object of this class since it still lacks many member functions that are necessary for a full
         * layer implementation.
         * 
         * \author M. Secanell, 2013
         */
        template <int dim>
        class PorousLayer 
        : 
        public BaseLayer<dim>
        {
        public:

            ///@name Initalization
            //@{    
            /**
             * Member function used to store all the gases that are in the pore space in the gas diffusion layer as well as
             * their temperature [\p Kelvin] and total pressure [\p atm].
             * Then, when computing diffusion, the class will return each one of the diffusion coefficients for the gases.
             * For example, if we pass a vector with for gases, say oxygen and nitrogen, effective_gas_diffusivity will return
             * \f$ D_{N_2,O_2} \f$
             */
            void set_gases_and_compute (std::vector<FuelCellShop::Material::PureGas*>& gases_in,
                                        const double& pressure_in,
                                        const double& temperature_in);
            
            /**
             * Member function used to compute bulk diffusion coefficients and derivatives w.r.t temperature
             * for non-isothermal case and store inside the layer. This function takes solute and solvent gas as an input argument (<b>Fick's diffusion</b> model).
             * Before calling this function, pressure [\p atm] must be set inside the layer, and
             * temperature [\p Kelvin] must be set using #set_temperature method.
             */
            void compute_gas_diffusion (FuelCellShop::Material::PureGas* solute_gas,
                                        FuelCellShop::Material::PureGas* solvent_gas);
            
            /**
             * Member function used to store all the gases that are in the pore space in the porous layer.
             * Besides this, it takes total pressure [\p atm] as input.
             * This method can be particulary used for non-isothermal applications when temperature changes
             * in every cell, hence this method doesn't compute the diffusion coefficients but store the relevant constant data.
             * \note Please ensure that in the case of <b>Fick's diffusion</b> model, \p solvent \p gas should be
             * the last object inside the input vector. On not ensuring this may not give any errors, as this function may also
             * be used for multi-component gas transport but for fickian diffusion, it may give wrong results.
             */
            inline void set_gases (std::vector<FuelCellShop::Material::PureGas*>& gases_in,
                            const double& pressure_in)
            {
                Assert(gases_in.size() >= 2, ExcMessage("Number of gases should be more than or equal to two in PorousLayer::set_gases method."));
                this->gases = gases_in;
                this->pressure = pressure_in;
            }
            
            /**
             * Member function used to set the temperature ]\p Kelvin] at every quadrature point
             * inside the cell. This function should particulary be used in the case of non-isothermal application.
             */
            inline void set_temperature (const SolutionVariable& T_in)
            {
                Assert( T_in.get_variablename() == temperature_of_REV, ExcMessage("Wrong solution variable passed in PorousLayer::set_temperature.") );
                this->T_vector = T_in;
            }
            
            //@}
            
            ///@name Accessors and info
            //@{
            /**
             * Return the FuelCellShop::Material::PureGas  pointer that is stored inside the class in the ith position.
             * The diffusion coefficient D[i][j] would be the diffusion coefficient of gas type i with j.
             */
            FuelCellShop::Material::PureGas* get_gas_pointer(int index) const
            {
                Assert(index > 0 || index < gases.size(), ExcIndexRange(index,0,gases.size()));
                return gases[index];
            };
            
            /**
             * Returns the vector of FuelCellShop::Material::PureGas pointers stored in the porous layer.
             * This method is useful to access all gases which are being transported inside a particular porous layer object.
             */
            std::vector<FuelCellShop::Material::PureGas*> get_gases() const
            {
                return this->gases;
            }
            
            /**
             * Return the gas index in the GDL class. If the gas does not exist in the GDL it will return -1 as the index.
             */
            void get_gas_index(FuelCellShop::Material::PureGas* gas_type,
                               int& index) const;
            /**
             * Return the constant temperature [\p Kelvin] and constant pressure [\p atm] inside the layer.
             */
            void get_T_and_p(double &T, double &p) const
            {
                T = this->temperature;
                p = this->pressure;
            }
            
            /**
             * Return the constant pressure [\p atm] inside the layer.
             */
            void get_p(double& p) const
            {
                p = this->pressure;
            }
            
            /**
             * This member function is a virtual class that can be used to output to screen information from the layer.
             *
             * This function should be re-implemented in each layer with data that is relevant in each case.
             */
            virtual void print_layer_properties() const
            {
                const std::type_info& info = typeid(*this);
                deallog << "Pure function " << __FUNCTION__
                << " called in Class "
                << info.name()  << std::endl;
            };
            
            /**
             * This virtual class should be used for any derived class to be able to test the functionality of the class.
             */
            virtual bool test_layer() const;
            
            //@}
            
            
            
        protected:
            ///@name Constructors, destructor, and initalization
            //@{    
            /**
             * Constructor
             */
            PorousLayer(const std::string& name)
            : FuelCellShop::Layer::BaseLayer<dim> (name)
            {
                this->derivative_flags.push_back(total_pressure);
                this->derivative_flags.push_back(temperature_of_REV);
            };
            
            /**
             * Constructor
             * 
             * \warning For internal use only
             */
            PorousLayer()
            : FuelCellShop::Layer::BaseLayer<dim> ()
            {};
            
            /**
             * Destructor
             */
            virtual ~PorousLayer()
            {};
            /**
             * Declare parameters for a parameter file
             */
            virtual void declare_parameters (const std::string &name, ParameterHandler &param) const
            {
                FuelCellShop::Layer::BaseLayer<dim>::declare_parameters(name,param);
                
                param.enter_subsection("Fuel cell data");
                {
                    param.enter_subsection(name);
                    {
                        
                    }
                    param.leave_subsection();
                }
                param.leave_subsection();
            }
            
            /**
             * Declare parameters for a parameter file
             * 
             * \note Since there are two declare parameters, this one is hidden by the former, so it has to be
             * redefined otherwise it cannot be called :( 
             */
            virtual void declare_parameters (ParameterHandler &param) const
            {
                declare_parameters(this->name, param);
            };
            
            /**
             * Member function used to read in data and initialize the necessary data
             * to compute the coefficients.
             */
            virtual void initialize (ParameterHandler &param);
            
            /**
             * Member function used to change the values in the parameter file for a given list
             * of parameters.
             * - object_name is the name of the subsection that will be used to declare all the parameters for an object of this type.
             * - name_param should ideally contain the string as it would appear in the input file, i.e. "Thickness of the agglomerate radius"
             * - value_param contains the value that the variable should be set at
             * - param is the ParameterHandler that contains all the information that has been read from file
             * 
             * Note that this is a static function, therefore it requires as the first argument the string with the name
             * of the section you would like to create for the object of this class.
             */
            void set_parameters(const std::string &object_name, 
                                       const std::vector<std::string>& name_dvar,
                                       const std::vector<double>& value_dvar,
                                       ParameterHandler &param)
            {
                FuelCellShop::Layer::BaseLayer<dim>::set_parameters(object_name,name_dvar,value_dvar,param);
                
                param.enter_subsection("Fuel cell data");
                {
                    param.enter_subsection(object_name);
                    {
                        // ADD VARIABLES IF NEEDED
                        //param.set(name_var, value_param);
                    }
                    param.leave_subsection();
                }
                param.leave_subsection();
            };
            
            /**
             * Member function used to change the values in the parameter file for a given list
             * of parameters.
             * - name_param should ideally contain the string as it would appear in the input file, i.e. "Thickness of the agglomerate radius"
             * - value_param contains the value that the variable should be set at
             * - param is the ParameterHandler that contains all the information that has been read from file
             */
            virtual void set_parameters(const std::vector<std::string>& name_dvar,
                                        const std::vector<double>& value_dvar,
                                        ParameterHandler &param)
            {};
            //@}            
            ///@name Internal accessors and info
            //@{
            /**
             * Return the molecular diffusivty all the gases assigned to the
             * layer using set_gases_and_compute. This routine uses FuelCellShop::Mixture::ChapmanEnskog to compute
             * the binary diffusivity for each gas with respect to each other. If the layer contains three gases the 
             * vector returns D12, D13, D23. For 4 gases, it returns D12, D13, D14, D23, D24, D34.
             */
            virtual void gas_diffusion_coefficients(Table< 2, double > &) const;
            
            /**
             * Return the derivative of the molecular diffusion coefficient
             * with respect to the derivative flags for all the gases assigned to the
             * layer using set_gases_and_compute. This routine uses FuelCellShop::Mixture::ChapmanEnskog to compute
             * the binary diffusivity for each gas with respect to each other. If the layer contains three gases the 
             * vector returns D12, D13, D23. For 4 gases, it returns D12, D13, D14, D23, D24, D34 . 
             * Derivatives are provided in the vector w.r.t following parameters in the same order as below:
             * - #total_pressure
             * - #temperature_of_REV
             */
            virtual void derivative_gas_diffusion_coefficients(std::vector< Table< 2, double > >&) const;  
            //@}                        
            ///@name Internal variables
            //@{
            /** If GDL properties are stored inside the class (\em e.g DummyGDL) then, return the property
             * stored under coefficient_name name
             */
            std::string diffusion_species_name; 

            /** Gases inside the layer*/
            std::vector<FuelCellShop::Material::PureGas*> gases;
            
            /** Temperature [\p Kelvin] used to compute gas diffusivity */
            double temperature;
            
            /** Total pressure [\p atm] used to compute gas diffusivity */
            double pressure;
            
            /** Temperature at every quadrature point inside the cell. */
            SolutionVariable T_vector;
            
            /** Tensor of diffusion coefficients 
             *-- This are computed with setting up the gas so that they do not need to be recomputed all the time.*/
            Table< 2, double > D_ECtheory; 
            
            /** Vector of tensors for the derivative of the diffusion coefficients 
             *-- This are computed with setting up the gas so that they do not need to be recomputed all the time.*/
            std::vector< Table< 2, double > > dD_ECtheory_dx;
            
            /**
             * Vector of bulk diffusion coefficients at every quadrature point inside the cell.
             * This particulary stores non-isothermal diffusion coefficients computed using #compute_gas_diffusion method.
             */
            std::vector<double> D_bulk;
            /**
             * Vector of derivative of bulk diffusion coefficients w.r.t temperature, at every quadrature
             * point inside the cell. These are computed using #compute_gas_diffusion method.
             */
            std::vector<double> dD_bulk_dT;
            
            //@}                                    
        };
        
    } // Layer
    
}  // FuelCellShop
    
#endif // _FUELCELLSHOP__GENERIC__LAYER_H