//---------------------------------------------------------------------------
//    $Id: solid_layer.h 1354 2013-08-17 00:01:22Z secanell $
//
//    Copyright (C) 2009 by Marc Secanell, University of Alberta
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__SOLID_LAYER_H
#define _FUELCELLSHOP__SOLID_LAYER_H

// FCST classes
#include <base_material.h>
#include <base_layer.h>
#include <carbon.h>
#include "fiber_base.h"
#include "carbon_fiber.h"

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

//Include STL
#include<cmath>
#include<iostream>

using namespace dealii;

namespace FuelCellShop
{
  namespace Layer
  {
    /**
     * This class characterizes a gas diffusion layer and uses this information to compute
	* effective transport properties. It is use to store an identification number for the
	* layer and its boundary conditions as well as the geometry of the layer.
     *
     * !!!!! UNDER CONSTRUCTION !!!!!!!
     * 
     */
    template <int dim>
      class SolidLayer : 
      public  BaseLayer<dim>
      {
      public:
	/**
	 * Constructor
	 */
	SolidLayer(std::string name);
	
	/**
	 * Destructor
	 */
	~SolidLayer();
	
	/**
	 * Declare parameters for a parameter file. The parameters that need to be declared are
	 - Porosity (default : 0.3) Represents the porosity in the GDL
	 - Method effective transport properties in pores (default: Bruggemann) Other options "Given|Bruggemann|Percolation|Mezedur"
	 - Method effective transport properties in solid (default: Bruggemann) Other options "Given|Bruggemann|Percolation|Mezedur"
	*/
	void declare_parameters (ParameterHandler &param) const;

	/**
	 * Member function used to set new parameters values in the optimization loop
	 */
	void set_parameters (const std::vector<std::string>& name_dvar,
			     const std::vector<double>& value_dvar,
			     ParameterHandler &param);
	
	/**
	 * Member function used to read in data and initialize the necessary data
	 * to compute the coefficients.
	 */
	void initialize (ParameterHandler &param);
	
	/**
	 * Member function that sets the material in the layer and the temperature
	 * */
	void set_solid_and_compute (FuelCellShop::Material::FiberBase* ,
				    double );
	
	/**
	* Compute the effective electron conductivity in the Solid
	*/
	virtual void effective_electron_conductivity(double& ) const;
	
	
	
	/**
	* Compute the effective electron conductivity in the solid
	*/
	virtual void effective_electron_conductivity(Tensor<2,dim>& ) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	}
		/**
	* Compute the effective electron conductivity in the solid
	*/
	virtual void effective_electron_conductivity(const double& /*prop*/, Tensor<2,dim>& /*prop_eff*/) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	}
	

	/**
	* Compute the derivative of the effective electron conductivity in the GDL
	* with respect to either the solution or design parameters. The parameters with respect to
	* which the derivatives are computed are setup in FuelCellShop::Layer::set_derivative_flags()
	*/
	virtual void derivative_effective_electron_conductivity(std::vector<Tensor<2,dim> >& ) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	};
	/**
	* Compute the effective thermal conductivity (isotropic) in the GDL
	*/
	virtual void effective_thermal_conductivity(double& ) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	};
	/**
	* Compute the effective thermal conductivity (anisotropic) in the GDL
	*/
	virtual void effective_thermal_conductivity(Tensor<2,dim>& ) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	};
	/**
	* Compute the effective thermal conductivity (isotropic) in the GDL
	*/
	virtual void effective_thermal_conductivity(const double&, const double&, double&) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	};
	/**
	* Compute the effective thermal conductivity (anisotropic) in the GDL, dependent on various solution variables, eg: Temperature
	*/
	virtual void effective_thermal_conductivity(std::vector< Tensor<2,dim> >&) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	};
	/**
	* Compute the derivative of the effective thermal conductivity in the GDL
	* with respect to either the solution or design parameters. The parameters with respect to
	* which the derivatives are computed are setup in FuelCellShop::Layer::set_derivative_flags()
	*/
	virtual void derivative_effective_thermal_conductivity(std::vector<Tensor<2,dim> >& ) const
	{
	  const std::type_info& info = typeid(*this);
	  deallog << "Pure function " << __FUNCTION__
		  << " called in Class "
		  << info.name() << std::endl;
	};
	
    /**
     * This member function returns a type_info object with the name of the 
     * base layer type the inherited class belongs to, i.e.
     * - GasDiffusionLayer
     * - MicroPorousLayer
     * - CatalystLayer
     * - MembraneLayer
     * - SolidLayer
     * - Channel
     * 
     * Note that this is necessary if we want to find out not the name of the actual class which can be obtain using
     * @code const std::type_info& name = typeid(*this) @endcode
     * but the name of the parent class.
     * 
     * @note Do not re-implement this class in children classes
     */
    const std::type_info& get_base_type() const
    {
        return typeid(SolidLayer<dim>);
    }
    
    /**
	 * Class test
	 */
	virtual void test_class();
	
      protected:
	
	/** Boundary ids */
	std::vector<unsigned int> boundary_ids;
	
	/** Temperature used to compute gases */
	double temperature;
	
	/** Data member that stores the solid material the layer is made of*/
	FuelCellShop::Material::FiberBase* solid;
	 
	/** Variable to store electcical conducivity */
	double electrical_conductivity;
	};
  }
}  // FuelCellShop

#endif // _FUELCELLSHOP__GAS_DIFFUSION_LAYER_H
