//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: BinaryDiffusion.h
//    - Description: Class storing methods compute binary diffusion coefficients
//    - Developers: Ali MalekpourKoupaei, M. Secanell and Madhur Bhaiya
//    - $Id: BinaryDiffusion.h 1354 2013-08-17 00:01:22Z secanell $
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__BINARYDIFFUSION__H
#define _FUELCELLSHOP__BINARYDIFFUSION__H

#define __Del(p) {if(p) delete p; p=__null;}
#define __printf __builtin_printf
#define __scanf  __builtin_scanf
#define __exp    __builtin_exp
#define __pow    __builtin_pow
#define __DUMMY  -9999

//Include STL and boost libraries
#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <cctype>
#include <boost/lexical_cast.hpp>

//Include deal.ii classes
#include <base/exceptions.h>

// openFCST classes
#include "PureGas.h"
#include "fcst_variables.h"


enum enDiffusionModelID {dmChapmanEnskog=1, dmWilkeChang, dmHanBartelsInterpolation , dmFerrellHimmelblau, dmNone};

namespace FuelCellShop
{
  namespace Mixture
  {


  /**
     BinaryDiffusion as an abstract class
  */
  class BinaryDiffusion
  {
  protected:
    FuelCellShop::Material::PureGas *species1, *species2;
  public:
    
    //class constructors
    BinaryDiffusion();
    BinaryDiffusion(FuelCellShop::Material::PureGas& species1, 
		    FuelCellShop::Material::PureGas& species2);
    
    //class distructor
    virtual ~BinaryDiffusion();
    
    // the method's specs.
    virtual char* get_name()			       	               =0;
    virtual char* get_description()		       	               =0;
    virtual enDiffusionModelID get_ID()            	               =0;
    
    /**
     * Comupute the diffusion coefficient [in m^2/s] at a given temperature [K] and pressure [atm].
     * To specify tempertaure and pressure use the member 
     * function  set_T_and_p(). The given temperature and pressure should be in Kelvin and atmospheres.
     */
    virtual void compute_diffusion (double& D)		  	       =0;
    
    /**
     * Compute a vector of diffusion coefficients [in m^2/s] for a non-isothermal, isobaric case.
     * To specify temperatures, use the member function #set_T_vector and to specify pressure, use the member function #set_p.
     * The given temperatures and pressure should be in Kelvin and atmospheres.
     */
    virtual void compute_diffusion(std::vector<double>& D_vec)             =0;
    
    virtual double compute_diffusion (const double& T,const double& p)     =0;
    virtual void compute_diffusion (double& D,std::vector<double>& dD)     =0;
    virtual void compute_diffusion (const double& T,const double& p,
				    double& D,std::vector<double>& dD)     =0;
    virtual void compute_dD (std::vector<double>& dD)			   =0;	
    virtual double compute_dD_dP (const double& T,const double& p) 	   =0;
    virtual double compute_dD_dT (const double& T,const double& p) 	   =0;
    
    /**
     * Method to compute derivative of diffusion coefficient with respect to temperature [in m^2/(s-K)]
     * for a non-isothermal, isobaric case, at all quadrature points of the cell.
     * To specify temperatures, use the member function #set_T_vector and to specify pressure, use the member function #set_p.
     * The given temperatures and pressure should be in Kelvin and atmospheres.
     */
    virtual void compute_dD_dT(std::vector<double>& dD_dT_vec)             =0;
    
    /**
     * Method used to set temperature (T) and pressure (p) for diffusion coefficient computation.
     * - Unit for T: Kelvin (K)
     * - Unit for p: atmospheres (atm)
     */
    void set_T_and_p(const double& T,const double& p);
    
    /**
     * Method to set pressure (p) for isobaric diffusion coefficient case.
     * Unit for p: atmospheres (atm)
     */
    void set_p(const double& p_atm);
    
    /**
     * Method to set a vector of temperatures (T) for nonisothermal diffusion
     * coefficient computation.
     * - Unit for T: Kelvin (K)
     */
    void set_temperature(const FuelCellShop::SolutionVariable& T_vec);
    
    void set_species(FuelCellShop::Material::PureGas& species1, 
		     FuelCellShop::Material::PureGas& species2);
    
    virtual void scan();
    virtual void report() ;
    virtual void report2file(char* filename);
    
  protected:
    /** Temperature [in K] used to compute the diffusion coefficient, unless specified as 
     * an input to the member function
     */
    double T;
    /** Pressure [in atm] used to compute the diffusion coefficient, unless specified as 
     * an input to the member function
     */
    double p;
    
    /**
     * Temperature [in K] used to compute the diffusion coefficient.
     * This container is used in the case of nonisothermal applications. This container is filled using
     * #set_temperature method.
     */
    FuelCellShop::SolutionVariable T_vector;
  };
  //---------------------------------------------------------------------------
  
  /*
    VaporDiffusion as a publicly derived class of BinaryDiffusion
  */
  class VaporDiffusion : public BinaryDiffusion
  {
  public:
    
    //class constructors
    VaporDiffusion();
    VaporDiffusion(FuelCellShop::Material::PureGas& species1, 
		   FuelCellShop::Material::PureGas& species2);
    
    //class distructor
    virtual ~VaporDiffusion();
    
    virtual double getSikmaAB(const double& sikma_A,
			      const double& sikma_B) 					  =0;
    virtual double getTstarAB(const double& Tstar_A,
			      const double& Tstar_B) 				   	  =0;
    virtual double getOmegaAB(const double& Tstar) 	  =0;
    
    virtual void scan();
    virtual void report() ;
    virtual void report2file(char* filename);
  };
  //---------------------------------------------------------------------------
  
  /**
   * This class uses the Chapman Enskog theory for mono-atomic gases to estimate the binary diffusion diffusivity between
   * two gases. The gases are input using the member function set_species(). Using member functions such as compute_diffusion() 
   * and compute_dD() the diffusion coefficient in units of m/s and the analytical derivatives of the diffusion coefficient
   * with respect of temperature and pressure respectively can be obtained.
   * Note that this class requires two objects of type FuelCellShop::Material::PureGas. In order to compute the binary diffusion
   * coefficient for a gas that has not been implemented, a new child of FuelCellShop::Material::PureGas needs to be implemented with 
   * the properties specified in the base class obtained from experimental data or from the literature.
   * 
   * This class is publicly derived class of VaporDiffusion
   *
   * <h3>Usage details</h3>
   * In order to use this class, you will need to specify the two gases for which you would like to compute the diffusion diffusion coefficient. 
   * For example, in the sample code below, we will compute the binary diffusion coefficient for oxygen in nitrogen. So, first we create an
   * object oxygen and nitrogen gas from FuelCellShop::Material::PureGas. Then, we set these gases as the gases for which we will compute the coefficients.
   * The diffusion coefficient is a function of temperature and pressure, therefore, we next set these values. The class takes these values in 
   * units of temperature [K] and pressure [atm] (Note: If you need different units please use the class in fcst_units.h).
   * Using compute_diffusion() the diffusion coefficient in units fo m2/s is returned. Note that the analytical derivatives of the diffusion coefficient
   * with respect of temperature and pressure can also be obtained.
   * 
   * @code
   *    double Diff_coef(0);
   *    FuelCellShop::Mixture::ChapmanEnskog CE;
   *    FuelCellShop::Material::Oxygen O2;
   *    FuelCellShop::Material::Nitrogen N2;
   *    CE.set_species(O2,N2);
   *    double temperature(353); //Temperature in K
   *    double pressure(1); //Pressure in atm
   *    CE.set_T_and_p(temperature,pressure);
   *    CE.compute_diffusion(Diff_coef);
   *    std::cout<<"Diffusion coeff: "<<Diff_coef<<" in m2/s"<<std::endl;
   * @endcode 
   * 
   * @author A. MalekpourKoupaei and M. Secanell
   * @date 2009-2013
   */
  
  class ChapmanEnskog : public VaporDiffusion
  {
  public:
    
    //class constructors
    ChapmanEnskog();
    ChapmanEnskog(FuelCellShop::Material::PureGas& species1, 
		  FuelCellShop::Material::PureGas& species2);
    
    //class distructor
    virtual ~ChapmanEnskog();
    
    virtual char* get_name();
    virtual char* get_description();
    virtual enDiffusionModelID get_ID();
    
    virtual double getSikmaAB(const double& sikma_A,
			      const double& sikma_B);
    virtual double getTstarAB(const double& Tstar_A,
			      const double& Tstar_B);
    virtual double getOmegaAB(const double& Tstar);
    /**
     * Comupute the diffusion coefficient [in m^2/s] at a given temperature [K] and pressure [atm].
     * To specify tempertaure and pressure use the member 
     * function  set_T_and_p(). The given temperature and pressure should be in Kelvin and atmospheres.
     */
    virtual void compute_diffusion (double& D);
    
    /**
     * Compute a vector of diffusion coefficients [in m^2/s] for a non-isothermal, isobaric case.
     * To specify temperatures, use the member function set_T_vector and to specify pressure, use the member function set_p.
     * The given temperatures and pressure should be in Kelvin and atmospheres.
     */
    virtual void compute_diffusion(std::vector<double>& D_vec);
    
    /**
     * Comupute the diffusion coefficient [in m^2/s] at a given temperature [K] and pressure [atm]
     */
    virtual double compute_diffusion (const double& T,const double& p);
     /**
     * Comupute diffusion coefficient [in m^2/s] and the sensitivity of the diffusion coefficient
     * with respect  to temperature and pressure. To specify tempertaure and pressure use the member 
     * function  set_T_and_p(). The given temperature and pressure should be in Kelvin and atmospheres
     */   
    virtual void compute_diffusion (double& D,std::vector<double>& dD) ;
    /**
     * Comupute diffusion coefficient [in m^2/s] and the sensitivity of the diffusion coefficient
     * with respect  to temperature and pressure at a given temperature [K] and pressure [atm]
     */
    virtual void compute_diffusion (const double& T,const double& p,
				    double& D,std::vector<double>& dD);
    virtual void compute_dD (std::vector<double>& dD);	
    virtual double compute_dD_dP (const double& T,const double& p);
    virtual double compute_dD_dT (const double& T,const double& p);
    
    /**
     * Method to compute derivative of diffusion coefficient with respect to temperature [in m^2/(s-K)]
     * for a non-isothermal, isobaric case, at all quadrature points of the cell.
     * To specify temperatures, use the member function #set_T_vector and to specify pressure, use the member function #set_p.
     * The given temperatures and pressure should be in Kelvin and atmospheres.
     */
    virtual void compute_dD_dT(std::vector<double>& dD_dT_vec);
    
    virtual void scan();
    virtual void report() ;
    virtual void report2file(char* filename);
    
  private:
    double M_A,M_B;
    double Tnb_A,Tnb_B;
    double Vl_A,Vl_B;
    double sikma_A,sikma_B;
    double Tstar_A,Tstar_B;
    double sikma_AB,Tstar_AB;
    double omega_DAB;
  };
  //---------------------------------------------------------------------------
  
  /*
    LiquidDiffusion as a publicly derived class of BinaryDiffusion
  */
  class LiquidDiffusion : public BinaryDiffusion
  {
  public:
    
    //class constructors
    LiquidDiffusion();
    LiquidDiffusion(FuelCellShop::Material::PureGas& species1, 
		    FuelCellShop::Material::PureGas& species2);
    
    //class distructor
    virtual ~LiquidDiffusion();
    
    virtual void scan();
    virtual void report() ;
    virtual void report2file(char* filename);
  };
  //---------------------------------------------------------------------------
  
  /*
    Wilke-Chang as a publicly derived class of BinaryDiffusion
  */
  class WilkeChang : public LiquidDiffusion
  {
  public:
    
    //class constructors
    WilkeChang();
    WilkeChang(FuelCellShop::Material::PureGas& species1, 
	       FuelCellShop::Material::PureGas& species2);
    
    //class distructor
    virtual ~WilkeChang();
    
    virtual char* get_name();
    virtual char* get_description();
    virtual enDiffusionModelID get_ID();
    
    virtual void compute_diffusion (double& D)		  		   ;
    virtual double compute_diffusion (const double& T,const double& p) ;
    virtual void compute_diffusion (double& D,std::vector<double>& dD) ;
    virtual void compute_diffusion (const double& T,const double& p,
				    double& D,std::vector<double>& dD)                             ;
    virtual void compute_dD (std::vector<double>& dD)			   ;	
    virtual double compute_dD_dP (const double& T,const double& p) 	   ;
    virtual double compute_dD_dT (const double& T,const double& p) 	   ;
    
    virtual void scan();
    virtual void report() ;
    virtual void report2file(char* filename);
    
  private:
    double mu, phi;
    void get_phi(double& phi);
    
  };
  //---------------------------------------------------------------------------
  
  /*
    HanBartelsInterpolation as a publicly derived class of BinaryDiffusion
  */
  class HanBartelsInterpolation : public LiquidDiffusion
  {
  public:
    
    //class constructors
    HanBartelsInterpolation();
    HanBartelsInterpolation(FuelCellShop::Material::PureGas& species1, 
			    FuelCellShop::Material::PureGas& species2);
    
    //class distructor
    virtual ~HanBartelsInterpolation();
    
    virtual char* get_name();
    virtual char* get_description();
    virtual enDiffusionModelID get_ID();
    
    virtual void compute_diffusion (double& D)		  		   ;
    virtual double compute_diffusion (const double& T,const double& p) ;
    virtual void compute_diffusion (double& D,std::vector<double>& dD) ;
    virtual void compute_diffusion (const double& T,const double& p,
				    double& D,std::vector<double>& dD)                             ;
    virtual void compute_dD (std::vector<double>& dD)			   ;	
    virtual double compute_dD_dP (const double& T,const double& p) 	   ;
    virtual double compute_dD_dT (const double& T,const double& p) 	   ;
    
    virtual void scan();
    virtual void report() ;
    virtual void report2file(char* filename);
    
  private:
    double mu, phi;
    void get_phi(double& phi);
    
  };
  //---------------------------------------------------------------------------
  
  /*
  Ferrell-Himmelblau as a publicly derived class of BinaryDiffusion
  */
  class FerrellHimmelblau : public LiquidDiffusion
  {
  public:
    
    //class constructors
    FerrellHimmelblau();
    FerrellHimmelblau(FuelCellShop::Material::PureGas& species1, 
		      FuelCellShop::Material::PureGas& species2);
    
    //class distructor
    virtual ~FerrellHimmelblau();
    
    virtual char* get_name();
    virtual char* get_description();
    virtual enDiffusionModelID get_ID();
    
    virtual void compute_diffusion (double& D)		  		   ;
    virtual double compute_diffusion (const double& T,const double& p) ;
    virtual void compute_diffusion (double& D,std::vector<double>& dD) ;
    virtual void compute_diffusion (const double& T,const double& p,
				    double& D,std::vector<double>& dD)                             ;
    virtual void compute_dD (std::vector<double>& dD)			   ;	
    virtual double compute_dD_dP (const double& T,const double& p) 	   ;
    virtual double compute_dD_dT (const double& T,const double& p) 	   ;
    
    virtual void scan();
    virtual void report() ;
    virtual void report2file(char* filename);
    
  private:
    
  };
  //---------------------------------------------------------------------------
  }
}

#endif
