//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: PureGas.h
//    - Description: Class storing pure gas properties.
//    - Developers: Ali MalekpourKoupaei (2009-10) and Madhur Bhaiya (2013)
//    - Id: $Id: PureGas.h 1354 2013-08-17 00:01:22Z secanell $ 
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__PUREGAS__H
#define _FUELCELLSHOP__PUREGAS__H

#define __Del(p) {if(p) delete p; p=__null;}
#define __printf __builtin_printf
#define __scanf  __builtin_scanf
#define __exp    __builtin_exp
#define __pow    __builtin_pow
#define __DUMMY  -9999

//Include STL
#include<cmath>
#include<iostream>
#include <vector>
#include <fstream>
#include <deal.II/base/subscriptor.h>

#include "base_material.h"

//enumerations
enum enMaterialID {matOxygen=1, matNitrogen, matWater, matHydrogen, matNone};

namespace FuelCellShop

{
  namespace Material
  {

    /**
     * Virtual class used to describe different gases and pure materials
     * for which viscority, diffusivity, etc. can be computed
     */
    class PureGas : public dealii::Subscriptor, BaseMaterial
    {
    public:
      
        /** Critical temperature in Kelvin */
        double Tc;
        /** Critical pressure in bar */
        double Pc;
        /** Critical volume in Cm3/mol */
        double Vc;
        /** Critical gas compressibility factor (Z).
        * It is a useful thermodynamic property for modifying the ideal gas law to account for the real gas behavior
        */
        double Zc;      
        /** Material other critical properties */
        double w;
        /** Molar mass in g/mol or kg/kmol */
        double M;
        /** in Kelvin */ 
        double Tnb;
        /** in m3/kmol */
        double Vl;
        /** dynamic viscosity at STP*/
        double Mu_0;
        /** Hard sphere diameter in Angstrom */
        double hard_sphere_diameter;
        /** epsilon/k */
        double e_k;

        PureGas();
        virtual ~PureGas();

        /** Lennard-Jones parameters */
        virtual double getCollisionDiameter();
        virtual double getDimensionlessTemperature(const double& T);

        virtual enMaterialID get_ID()=0; 
        virtual char* get_name()     =0; 
        virtual char* get_formula()  =0; 

        virtual void get_viscosity(double& mu,const double& T) =0;
        virtual void get_viscosity_Chapman_Enskog(double& mu, const double& T, const double& MW, const double& hard_sphere_diameter);
        virtual void get_viscosity_Chapman_Enskog(double& mu, const double& T);
        virtual void get_dMu_d_T_Chapman_Enskog(double& dmu, const double& T, 
                                                const double& MW, const double& hard_sphere_diameter);
        virtual void get_dMu_d_T_Chapman_Enskog(double& dmu, const double& T);
        
        /**
         * Abstract method for defining interface to return molar enthalpy [\p J/mol] of an ideal gas.
         * For an ideal gas, it's just a function of temperature. The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_molar_enthalpy(const double& T) const = 0;
        
        /**
         * Abstract method for defining interface to return derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K], 
         * for an ideal gas. The method takes temperature (in \p Kelvins) as an input argument.
         */
        virtual double get_dmolar_enthalpy_dT(const double& T) const = 0;
        
        /**
         * Abstract method for defining interface to return second derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K], 
         * for an ideal gas. The method takes temperature (in \p Kelvins) as an input argument.
         */
        virtual double get_d2molar_enthalpy_dT2(const double& T) const = 0;
        
        
    };
//---------------------------------------------------------------------------

    /**
     * Oxygen as a publicly derived class of Material
     */
    class Oxygen : public PureGas
    {
    public:
        
        Oxygen();
        virtual ~Oxygen();
        
        virtual enMaterialID get_ID(); 
        virtual char* get_name()     ; 
        virtual char* get_formula()  ; 
        
        virtual void get_viscosity(double& mu,const double& T) ;
        
        /**
         * Method to return molar enthalpy [\p J/mol] of Oxygen gas \f$ (O_2) \f$
         * (assuming ideal gas behavior), as a function of temperature [\p K].
         * The method takes temperature (in \p Kelvins) as an input argument.
         */
        virtual double get_molar_enthalpy(const double&) const;
        
        /**
         * Method to return derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Oxygen gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_dmolar_enthalpy_dT(const double& T) const;
        
        /**
         * Method to return second derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Oxygen gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_d2molar_enthalpy_dT2(const double& T) const;
    };
//---------------------------------------------------------------------------
    
    /**
     * Nitrogen as a publicly derived class of Material
     */
    class Nitrogen : public PureGas
    {
    public:
        
        Nitrogen();
        virtual ~Nitrogen();
        
        virtual enMaterialID get_ID(); 
        virtual char* get_name()     ; 
        virtual char* get_formula()  ; 
        
        virtual void get_viscosity(double& mu,const double& T) ;
        
        /**
         * Method to return molar enthalpy [\p J/mol] of Nitrogen gas \f$ (N_2) \f$
         * (assuming ideal gas behavior), as a function of temperature [\p K].
         * The method takes temperature (in \p Kelvins) as an input argument.
         */
        virtual double get_molar_enthalpy(const double&) const;
                
        /**
         * Method to return derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Nitrogen gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_dmolar_enthalpy_dT(const double& T) const;
        
        /**
         * Method to return second derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Nitrogen gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_d2molar_enthalpy_dT2(const double& T) const;
    };
    
//---------------------------------------------------------------------------
    
    /**
     * Water as a publicly derived class of Material
     */
    class Water : public PureGas
    {
    public:
      
        Water();
        virtual ~Water();

        virtual enMaterialID get_ID(); 
        virtual char* get_name()     ; 
        virtual char* get_formula()  ; 

        virtual void get_viscosity(double& mu,const double& T) ;
        
        /**
         * Method to return molar enthalpy [\p J/mol] of Water gas \f$ (H_2O) \f$
         * (assuming ideal gas behavior), as a function of temperature [\p K].
         * The method takes temperature (in \p Kelvins) as an input argument.
         */
        virtual double get_molar_enthalpy(const double&) const;
                
        /**
         * Method to return derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Water gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_dmolar_enthalpy_dT(const double& T) const;
        
        /**
         * Method to return second derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Water gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_d2molar_enthalpy_dT2(const double& T) const;
        
        /** Static method to return Water vapor saturation pressure in Pascals for an input temperature in Kelvin. */
        static double satPressure(const double&);

        /** Static method to return derivative of water vapor saturation pressure with respect to Temperature (units: Pa/K) for an input temperature in Kelvin. */
        static double derivSatPressure(const double&);
        
        /**
         * Static method to return molar weight of water in grams/mole.
         */
        static double molarWeight()
        {
            return 18.015;
        }
    };
    
//---------------------------------------------------------------------------
    
    /**
     * Hydrogen as a publicly derived class of Material
     */
    class Hydrogen : public PureGas
    {
    public:
      
        Hydrogen();
        virtual ~Hydrogen();

        virtual enMaterialID get_ID(); 
        virtual char* get_name()     ; 
        virtual char* get_formula()  ; 

        virtual void get_viscosity(double& mu,const double& T) ;
        
        /**
         * Method to return molar enthalpy [\p J/mol] of Hydrogen gas \f$ (H_2) \f$
         * (assuming ideal gas behavior), as a function of temperature [\p K].
         * The method takes temperature (in \p Kelvins) as an input argument.
         */
        virtual double get_molar_enthalpy(const double&) const;
                
        /**
         * Method to return derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Hydrogen gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_dmolar_enthalpy_dT(const double& T) const;
        
        /**
         * Method to return second derivative of molar enthalpy [\p J/mol] w.r.t Temperature [\p K]
         * for Hydrogen gas (assuming ideal gas behavior). The method takes temperature (in \p Kelvins)
         * as an input argument.
         */
        virtual double get_d2molar_enthalpy_dT2(const double& T) const;
    };
    
    //---------------------------------------------------------------------------
  }
}

#endif
