//---------------------------------------------------------------------------
//    $Id: PureLiquid.h 1354 2013-08-17 00:01:22Z secanell $
//
//    Copyright (C) 2012 by Madhur Bhaiya
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------
#ifndef _FUELCELLSHOP__PURELIQUID__H
#define _FUELCELLSHOP__PURELIQUID__H

//Include STL
#include <cmath>
#include <iostream>
#include <cstdlib>
#include <vector>
#include <fstream>

#include <deal.II/base/parameter_handler.h>
using namespace dealii;


namespace FuelCellShop

{
  namespace Material
  {

	/**
	* Virtual class used to describe different liquids and pure materials
	* for which viscority, diffusivity, etc. can be computed
	*/
	class PureLiquid : public dealii::Subscriptor
	{
	public:
		
		/** Molar mass in g/mol or kg/kmol */
		double M;
		/** dynamic viscosity at STP*/
		double Mu_0;

		/** Constructor */
		PureLiquid(){};
		/** Destructor */
		virtual ~PureLiquid(){};
		
		virtual char* get_name()     =0; 
		virtual char* get_formula()  =0; 
		
		virtual void get_viscosity(double&, const double&) =0;

    };
    //---------------------------------------------------------------------------
    
    /**
       [ LiquidWater as a publicly derived class of Pureliquid ]
	*/
	class LiquidWater : public PureLiquid
	{
	public:
		


		LiquidWater();
		virtual ~LiquidWater();
		
		virtual char* get_name()     ; 
		virtual char* get_formula()  ; 
		
		virtual void get_viscosity(double&, const double&) ;
		

		/**
		* Declare all necessary parameters in order to compute the coefficients
		*/
		void declare_parameters ( ParameterHandler &param ) const;

		/**
		* Member function used to read in data and initialize the necessary data
		* to compute the coefficients.
		*/
		void initialize ( ParameterHandler &param );



		/** Function to compute latent heat of vaporization of water (J/mol) as a function of temperature (K).
		 * Ref: Simple Forumulas for Thermophysical Properties of Liquid Water for Heat Transfer Calculations (from 0C to 150C), C. O. Popiel and K. Wojtkowiak,
		 * Heat Transfer Engineering, Vol.19, No.3, 1998.
		 */
		static double latentVap_heat(const double&);
		/** Function to compute derivative of latent heat of vaporization of water with respect to temperature. */
		static double deriv_latentVap_heat(const double&);

		double get_DO2();
		double get_DH();
		double get_Relative_Permittivity();

	private:

		double oxygen_diffusion_coeff;
		double proton_diffusion_coeff;
		double relative_permittivity;

	};

    //---------------------------------------------------------------------------
  }
}

#endif
