//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2011-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: base_material.h
//    - Description: Base material class
//    - Developers: M. Secanell and Madhur Bhaiya
//    - $Id: base_material.h 1373 2013-08-21 00:34:53Z madhur $
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__BASE_MATERIAL__H
#define _FUELCELLSHOP__BASE_MATERIAL__H

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>
#include <deal.II/base/subscriptor.h>

// Include STL
#include<cmath>
#include<iostream>

// Include OpenFCST routines:
#include "utils/fcst_variables.h"
#include "utils/system_management.h"

using namespace dealii;

namespace FuelCellShop
{
    namespace Material
    {
        /**
         * Virtual class used to provide the interface for all material classes.
         * 
         * No object of type BaseMaterial should ever be created. 
         * 
         * \todo create a proper tree for materials. I think it should go, material_base, then childs: electrolyte, catalyst support, catalyst (...)
         */
        class BaseMaterial
        {
        public:
            
            ///@name Initalization
            //@{    
            /**
             * Member function used to read in data and initialize the necessary data
             * to compute the coefficients.
             * 
             * \warning This is a PureFunction and it does not initialize anything, so please do not call this function in the children.
             * 
             */
            virtual void initialize(ParameterHandler&)
            {
                const std::type_info& info = typeid(*this);
                deallog << "Pure function " << __FUNCTION__
                << " called in Class "
                << info.name()  << std::endl;
            }

            /**
             * Set the variables for which you would like to compute the derivatives.
             */
            void set_derivative_flags(const std::vector<VariableNames>& flags)
            {
                derivative_flags = flags;
            }

            //@}
            
            ///@name Accessors and info
            //@{
            /**
             * Return the name of the layer
             */
            inline const std::string& name_material()
            {
                return name;
            }
            
            /**
             * This function prints out
             * the material properties.
             */
            virtual void print_material_properties() const
            {
                const std::type_info& info = typeid(*this);
                deallog << "Pure function " << __FUNCTION__
                << " called in Class "
                << info.name()  << std::endl;
            }
            //@}
        protected:
            
            ///@name Constructors, destructor, and parameter initalization
            //@{    
            /**
             * Constructor
             */
            BaseMaterial()
            :
            name("no_name")
            { }

            /**
             * Constructor
             * 
             * \deprecated Material classes should not take any string. The string is not
             * used for anything.
             */
            BaseMaterial(const std::string& name)
            : name(name)
            {}

            /**
             * Destructor
             */
            virtual ~BaseMaterial()
            {}
            
            /**
             * Declare parameters for a parameter file
             * 
             * \warning This is a PureFunction and it does not declare anything, so please do not call this function in the children.
             */
            virtual void declare_parameters(ParameterHandler&) const
            {
                const std::type_info& info = typeid(*this);
                deallog << "Pure function " << __FUNCTION__
                << " called in Class "
                << info.name()  << std::endl;
            }
            
            /**
             * Member function used to change the values in the parameter file for a given list
             * of parameters.
             * - name_param should ideally contain the string as it would appear in the input file, i.e. "Thickness of the agglomerate radius"
             * - value_param contains the value that the variable should be set at
             * - param is the ParameterHandler that contains all the information that has been read from file
             * 
             * \warning This is a PureFunction and it does not declare anything, so please do not call this function in the children.
             */
            virtual void set_parameters(const std::vector<std::string>& ,
                                        const std::vector<double>& ,
                                        ParameterHandler &)
            {
                const std::type_info& info = typeid(*this);
                deallog << "Pure function " << __FUNCTION__
                << " called in Class "
                << info.name()  << std::endl;
            }
            //@}
            
            ///@name Internal variables
            //@{            
            /** Name of the layer. This value is used as a header in a subsection for the parameter file
             * where all the information concerning this layer is held*/
            const std::string name;
            
            /** Flags for derivatives: These flags are used to request derivatives*/
            std::vector<VariableNames> derivative_flags;
            //@}
        };
        
    } // Material
    
}  // FuelCellShop
    
    #endif