//---------------------------------------------------------------------------
//
//    Copyright (C) 2012 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__IDEAL_GAS_H
#define _FUELCELLSHOP__IDEAL_GAS_H

namespace FuelCellShop
{
	namespace Material
	{
		/**
		 * Implementation of the ideal gas law. This class once initialized with either a pressure and temperature
		 * or a concentration and temperature will return the additional variable. Enjoy!
		 * 
		 * <h3>Usage details</h3>
		 * In order to use this class, you first need to use one of the set_* options in order to initialize all the variables
		 * that define an ideal gas. In the set_* member function all properties are computed. If you would like to change, temperature, type
		 * of gas, etc. you will need to call this routine again.
		 * Once the set routine has been called, then you can request any gas property using get_* routines.
		 * 
		 * @code
		 * //Create a gas object
		 * FuelCellShop::Material::Oxygen oxygen; 
		 * //Create an object IdealGas
		 * FuelCellShop::Material::IdealGas marc;
		 * // Set all gas properties. Make sure you are entering the info using the correct units.
		 * marc.set_pT_and_gas(101325, 298, &oxygen);
		 * // You can now request info from your class.
		 * double temperature = marc.get_T();
		 * //Print to screen all properties
		 * marc.print_properties_to_screen();
		 * @endcode 
		 * 
		 * @author M. Secanell
		 * @date 2012
		 */
		class IdealGas
		{
			public:
			/** Constructor */
			IdealGas()
			:
			initialized(false)
			{}
			
			/** Destructor */
			~IdealGas()
			{}
			
			/**
			 * Function to set the temperature in K and pressure in Pa of the gas
			 */
			void set_pT_and_gas(const double p, const double T, FuelCellShop::Material::PureGas* gas)
			{
				initialized = true;
				double R = Constants::R();
				press = p;
				temp = T;
				conc = press / (R*temp);
				gas_type = gas;
				density = (press * (gas_type->M*1.0e-3))/ (R*temp);
			}
			/**
			 * Function to modify the temperature in K and concentration in mol/m3 of the gas
			 */
			void set_cT_and_gas(const double c, const double T, FuelCellShop::Material::PureGas* gas)
			{
				initialized = true;
				double R = Constants::R();
				conc = c;
				temp = T;
				press = conc*R*temp;
				gas_type = gas;
				density = (press * (gas_type->M*1.0e-3))/ (R*temp);
			}
			
			/**
			* Member function used to return the temperature of the gas in Kelvin.
			*/
			double get_T()
			{
				Assert (initialized, ExcNotInitialized());
				return temp;
			}
			/**
			 * Member function used to return the pressure of the gas in Pascal.
			 */
			double get_p()
			{
				Assert (initialized, ExcNotInitialized());
				return press;
			}
			/**
			 * Member function used to return the pressure of the gas in atm.
			 */
			double get_p_atm()
			{
				Assert (initialized, ExcNotInitialized());
				return press/101325.0;
			}
			/**
			 * Member function used to return the concentration of the gas in mol/m3.
			 */
			double get_c()
			{
				Assert (initialized, ExcNotInitialized());
				return conc;
			}
			/**
			 * Member function used to return the density of the gas in kg/m3.
			 */
			double get_density()
			{
				Assert (initialized, ExcNotInitialized());
				return density;
			}
			/**
			 * Member function used to return the density of the gas in kg/m3.
			 */
			FuelCellShop::Material::PureGas* get_gas()
			{
				Assert (initialized, ExcNotInitialized());
				return gas_type;
			}
			/**
			 * Member function that prints to screen p, T and c
			 */
			void print_properties_to_screen()
			{
				Assert (initialized, ExcNotInitialized());
				std::cout<<"Pressure is: "<< press <<" Pa"<<std::endl;
				std::cout<<"Temperature is: "<< temp <<" K"<<std::endl;
				std::cout<<"Concentration is: "<< conc <<" mol/m3"<<std::endl;
				std::cout<<"Density is: "<< density <<" kg/m3"<<std::endl;
			}
			/** Member function use to make sure the class is returning appropriate values. 
			 * In this case I simply test that for oxygen it returns the appropriate values for STP.
			 */
			void test()
			{
    			FuelCellShop::Material::Oxygen oxygen;
				FuelCellShop::Material::IdealGas marc;
				marc.set_pT_and_gas(101325, 298, &oxygen);
				std::cout<<"======== TESTING set_pT_and_gas ========"<<std::endl;
				if (marc.get_T() - 298 < 1e-4)
					std::cout<<"======== Temp OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Temp BUG !!!!!!!"<<std::endl;
				// test pressure
				if (marc.get_p() - 101325 < 1e-4)
					std::cout<<"======== Pressure OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Pressure BUG !!!!!!!"<<std::endl;
				// test concentration
				if (marc.get_c() - 40.8949 < 1e-4)
					std::cout<<"======== Concentration OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Concentration BUG !!!!!!!"<<std::endl;
				// test density
				if (marc.get_density() - 1.308592 < 1e-4)
					std::cout<<"======== Density OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Density BUG !!!!!!!"<<std::endl;
				
				marc.set_cT_and_gas(40.8949, 298, &oxygen);
				std::cout<<"======== TESTING set_cT_and_gas ========"<<std::endl;
				if (marc.get_T() - 298 < 1e-4)
					std::cout<<"======== Temp OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Temp BUG !!!!!!!"<<std::endl;
				// test pressure
				if (marc.get_p() - 101325 < 1e-4)
					std::cout<<"======== Pressure OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Pressure BUG !!!!!!!"<<std::endl;
				// test concentration
				if (marc.get_c() - 40.8949 < 1e-4)
					std::cout<<"======== Concentration OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Concentration BUG !!!!!!!"<<std::endl;
				// test density
				if (marc.get_density() - 1.308592 < 1e-4)
					std::cout<<"======== Density OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Density BUG !!!!!!!"<<std::endl;
			}
		private:
			/** Bool element that let's me know if the class has been initialized */
			bool initialized;
			/** Temperature of the gas, [K] */
			double temp;
			/** Pressure of the gas, [Pa] */
			double press;
			/** Concentration of the gas, [mol/m3] */
			double conc;
			/** Density of the gas, [kg/m3] */
			double density;
			/** Gas species */
			SmartPointer<FuelCellShop::Material::PureGas> gas_type;

		}; //class
	} //namespace
	
	namespace Mixture
	{
		/**
		 * Implementation of the ideal gas law for mixtures. This class once initialized using several of the set_xx options
		 * will return partial pressures, concentrations for each species and mole fractions.
		 * 
		 * <h3>Usage details</h3>
		 * In order to use this class, you first need to use one of the set_* options in order to initialize all the variables
		 * that define an ideal gas. In the set_* member function all properties are computed. If you would like to change, temperature, type
		 * of gas, etc. you will need to call this routine again.
		 * Once the set routine has been called, then you can request any gas property using get_* routines.
		 * 
		 * @code
		 * FuelCellShop::Material::Oxygen oxygen;
		 * FuelCellShop::Material::Nitrogen nitrogen;
		 * std::vector<FuelCellShop::Material::PureGas* > gasv;
		 * gasv.push_back(&oxygen);
		 * gasv.push_back(&nitrogen);
		 * // NOTE: The order is important. The FuelCellShop::Material::PureGas MUST be created before the
		 * // object of type IdealGasMixuture otherwise the gases would be destroyed first at the end of the
		 * // program leading to SmartPointer errors!!!
		 * FuelCellShop::Mixture::IdealGasMixture marc;
		 * std::vector<double> xv;
		 * xv.push_back(0.5);
		 * xv.push_back(0.5);
		 * marc.set_pT_xi_and_gases(101325, 298, xv, gasv);
		 * marc.print_properties_to_screen();
		 * 
		 * @endcode 
		 * 
		 * @author M. Secanell
		 * @date 2012
		 */
		class IdealGasMixture
		{
			public:
			/** Constructor */
			IdealGasMixture()
			:
			initialized(false)
			{}
			
			/** Destructor */
			~IdealGasMixture()
			{}
			/**
			 * Function to set the temperature in K, pressure in Pa of the gas and the molar fractions
			 */
			void set_pT_xi_and_gases(const double p, const double T, const std::vector<double> xv, std::vector<FuelCellShop::Material::PureGas* > gases)
			{
				Assert (xv.size() == gases.size(),ExcDimensionMismatch(xv.size(),gases.size()));
				initialized = true;
				double R = Constants::R();
				press = p;
				temp = T;
				conc =  press / (R*temp);
				concv.resize(xv.size());
				pressv.resize(xv.size());
				gas_types.resize(xv.size());
				densityv.resize(xv.size());
				M_mix = 0.0;
				for (unsigned int i = 0; i<xv.size(); i++)
				{
					gas_types[i] = gases[i];
					concv[i] = xv[i]*(press / (R*temp));
					pressv[i] = xv[i]*press;
					densityv[i] = (pressv[i] * (gas_types[i]->M*1.0e-3))/ (R*temp);
					M_mix += xv[i]*gas_types[i]->M;
				}
				density = (press * (M_mix*1.0e-3))/ (R*temp);
				
			}
			/**
			 * Function to modify the temperature in K and concentration in mol/m3 of the gas
			 */
			void set_cT_xi_and_gases(const double c, const double T, const std::vector<double> xv, std::vector<FuelCellShop::Material::PureGas* > gases)
			{
				Assert (xv.size() == gases.size(),ExcDimensionMismatch(xv.size(),gases.size()));
				initialized = true;
				double R = Constants::R();
				conc = c;
				temp = T;
				press = c*R*T;
				concv.resize(xv.size());
				pressv.resize(xv.size());
				gas_types.resize(xv.size());
				densityv.resize(xv.size());
				M_mix = 0.0;
				for (unsigned int i = 0; i<xv.size(); i++)
				{
					gas_types[i] = gases[i];
					concv[i] = xv[i]*(press / (R*temp));
					pressv[i] = xv[i]*press;
					densityv[i] = (pressv[i] * (gas_types[i]->M*1.0e-3))/ (R*temp);
					M_mix += xv[i]*gas_types[i]->M;
				}
				density = (press * (M_mix*1.0e-3))/ (R*temp);

			}
			/**
			 * Function to set the temperature in K, pressure in Pa of the gas and the molar fractions
			 */
			void set_pv_T_and_gases(const std::vector<double> pv, const double T, std::vector<FuelCellShop::Material::PureGas* > gases)
			{
				Assert (pv.size() == gases.size(),ExcDimensionMismatch(pv.size(),gases.size()));
				initialized = true;
				double R = Constants::R();
				pressv = pv;
				temp = T;
				concv.resize(pv.size());
				gas_types.resize(pv.size());
				densityv.resize(pv.size());
				press = 0;
				for (unsigned int i = 0; i<pv.size(); i++)
				{
					gas_types[i] = gases[i];
					press += pv[i];
				}
				conc = press / (R*temp);
				for (unsigned int i = 0; i<pv.size(); i++)
				{
					concv[i] = pv[i] / (R*temp);
					densityv[i] = (pressv[i] * (gas_types[i]->M*1.0e-3))/ (R*temp);
					M_mix += (pressv[i]/press)*gas_types[i]->M;
				}
				density = (press * (M_mix*1.0e-3))/ (R*temp);

			}
			/**
			 * Function to modify the temperature in K and concentration in mol/m3 of the gas
			 */
			void set_cv_T_and_gases(const std::vector<double> cv, const double T, std::vector<FuelCellShop::Material::PureGas* > gases)
			{
				Assert (cv.size() == gases.size(),ExcDimensionMismatch(cv.size(),gases.size()));
				initialized = true;
				double R = Constants::R();
				temp = T;
				concv = cv;
				pressv.resize(cv.size());
				gas_types.resize(cv.size());
				densityv.resize(cv.size());
				conc = 0;
				for (unsigned int i = 0; i<cv.size(); i++)
				{
					gas_types[i] = gases[i];
					conc += cv[i];
				}
				press = conc*R*temp;
				for (unsigned int i = 0; i<cv.size(); i++)
				{
					pressv[i] = concv[i]*R*temp;
					densityv[i] = (pressv[i] * (gas_types[i]->M*1.0e-3))/ (R*temp);
					M_mix += (pressv[i]/press)*gas_types[i]->M;
				}
				density = (press * (M_mix*1.0e-3))/ (R*temp);
			}
			/**
			* Member function used to return the temperature of the gas in Kelvin.
			*/
			double get_T()
			{
				Assert (initialized, ExcNotInitialized());
				return temp;
			}
			/**
			 * Member function used to return the total pressure of the gas mixture in Pascal.
			 */
			double get_p()
			{
				Assert (initialized, ExcNotInitialized());
				return press;
			}
			
			/**
			 * Member function used to return the partial pressure of the components in the gas in Pascal.
			 */
			std::vector<double> get_pi()
			{
				Assert (initialized, ExcNotInitialized());
				return pressv;
			}
			
			/**
			 * Member function used to return the pressure of the gas mixture in atm.
			 */
			double get_p_atm()
			{
				Assert (initialized, ExcNotInitialized());
				return press/101325.0;
			}
			/**
			 * Member function used to return the concentration of the gas in mol/m3.
			 */
			double get_c()
			{
				Assert (initialized, ExcNotInitialized());
				return conc;
			}
			/**
			 * Member function used to return the concentration of the components in the gas in mol/m3.
			 */
			std::vector<double> get_ci()
			{
				Assert (initialized, ExcNotInitialized());
				return concv;
			}
			/**
			 * Member function used to return the density of the gas in kg/m3.
			 */
			double get_density()
			{
				Assert (initialized, ExcNotInitialized());
				return density;
			}
			/**
			 * Member function used to return the partial density of the gas in kg/m3.
			 */
			std::vector<double> get_densityi()
			{
				Assert (initialized, ExcNotInitialized());
				return densityv;
			}
			/**
			 * Member function used to return the mole fraction of the components in the gas mixture.
			 */
			std::vector<double> get_xi()
			{
				Assert (initialized, ExcNotInitialized());
				std::vector<double> xi;
				xi.resize(pressv.size());
				for (unsigned int i = 0; i<pressv.size(); ++i)
					xi[i] = pressv[i]/press;
				
				return xi;
			}
			/**
			 * Member function used to return a pointer to the gas type stored in position index.
			 */
			dealii::SmartPointer<FuelCellShop::Material::PureGas> get_gas(const unsigned int index)
			{
				Assert (initialized, ExcNotInitialized());
				return gas_types[index];
			}
			/**
			 * Member function that prints to screen p, T and c
			 */
			void print_properties_to_screen()
			{
				Assert (initialized, ExcNotInitialized());
				std::cout<<"Pressure is: "<< press <<" Pa"<<std::endl;
				std::cout<<"Temperature is: "<< temp <<" K"<<std::endl;
				std::cout<<"Concentration is: "<< conc <<" mol/m3"<<std::endl;
				std::cout<<"Density is: "<< density <<" kg/m3"<<std::endl;
				
				for (unsigned int i = 0; i<concv.size(); i++)
				{
					std::cout<<"Parial pressure for species "<<i<<" is: "<< pressv[i] <<" Pa"<<std::endl;
					std::cout<<"Concentration for species: "<<i<<" is: "<< concv[i] <<" mol/m3"<<std::endl;
					std::cout<<"Density for species: "<<i<<" is: "<< densityv[i] <<" kg/m3"<<std::endl;
				}
			}
			/**
			 * Member function used to test class is working
			 * NOTE: The test is not complete at this point since it does not test all set_* options.
			 */
			void test()
			{
				FuelCellShop::Material::Oxygen oxygen;
				FuelCellShop::Material::Nitrogen nitrogen;
				std::vector<FuelCellShop::Material::PureGas* > gasv;
				FuelCellShop::Mixture::IdealGasMixture marc;
				gasv.push_back(&oxygen);
				gasv.push_back(&nitrogen);
				std::vector<double> v;
				v.push_back(0.5*40.8949);
				v.push_back(0.5*40.8949);
				marc.set_cv_T_and_gases(v, 298, gasv);
				// test temp
				if (marc.get_T() - 298 < 1e-4)
					std::cout<<"======== Temp OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Temp BUG !!!!!!!"<<std::endl;
				// test pressure
				if (marc.get_p() - 101325 < 1e-4)
					std::cout<<"======== Pressure OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Pressure BUG !!!!!!!"<<std::endl;
				// test concentration
				if (marc.get_c() - 40.8949 < 1e-4)
					std::cout<<"======== Concentration OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Concentration BUG !!!!!!!"<<std::endl;
				// test density
				if (marc.get_density() - 1.22711 < 1e-4)
					std::cout<<"======== Density OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Density BUG !!!!!!!"<<std::endl;
				// test gases
				for (unsigned int i=0; i<marc.get_ci().size(); ++i)
				if (marc.get_gas(i)->M - gasv[i]->M < 1e-4)
					std::cout<<"======== Gas index "<<i<<" is OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Gas index "<<i<<" has a BUG !!!!!!!"<<std::endl;
				// test gases
				for (unsigned int i=0; i<marc.get_ci().size(); ++i)
				if (marc.get_ci()[i] - marc.get_c()*0.5 < 1e-4)
					std::cout<<"======== Gas index "<<i<<" is OK ========"<<std::endl;
				else
					std::cout<<"!!!!!!! Gas index "<<i<<" has a BUG !!!!!!!"<<std::endl;
			}
		private:
			/** Bool element that let's me know if the class has been initialized */
			bool initialized;
			/** Temperature of the gas, [K] */
			double temp;
			/** Pressure of the gas, [Pa] */
			double press;
			/** Concentration of the gas, [mol/m3] */
			double conc;
			/** Density of the gas, [kg/m3] */
			double density;
			/** Molar mass of the mixture [g/mol mix] */
			double M_mix;
			/** Pressure of the gas, [Pa] */
			std::vector<double> pressv;
			/** Concentration of the gas, [mol/m3] */
			std::vector<double> concv;
			/** Density of the gas, [kg/m3] */
			std::vector<double> densityv;
			/** Vector containing pointers to all gas species */
			std::vector<SmartPointer<FuelCellShop::Material::PureGas> > gas_types;

		}; //class
	} //namespacie
	
} //namespacie


#endif