//---------------------------------------------------------------------------
//
//    Copyright (C) 2011 by Marc Secanell, University of Alberta
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP_MATERIAL_PLATE_GRAPHITE__H
#define _FUELCELLSHOP_MATERIAL_PLATE_GRAPHITE__H

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/point.h>
#include <base/function.h>
#include <lac/vector.h>

// Include FCST classes
#include "fcst_constants.h"
#include "material_plate_base.h"

//Include STL
#include<cmath>
#include<iostream>

namespace FuelCellShop
{
  namespace Material
  { 
    /**
     * Class to compute the properties of graphite used in bipolar plates.
     * class, this classs is only for setting up the interface.
     * This class returns:
     *  - thermal conductivity tensor and derivatives
     *  - electron (electrical) conductivity tensort and derivatives
     *  - Young's modulus tensor and derivatives
     *  - Poisson's ratio
     *  - expansion coefficient
     *  - surface properties (contact angle)
     * 
     * @TODO Implement a liquid gas so that it can be passed to the member function that computes contact angles
     */
    class MaterialPlateGraphite
    :
    public MaterialPlateBase
    {
      public:
	/**
	* Constructor
	*/
	MaterialPlateGraphite(std::string name);
	/**
	* Destructor
	*/
	~MaterialPlateGraphite();
	/**
	* Declare parameters
	*/
	void declare_parameters (ParameterHandler &param) const;
	/**
	* Set parameters on the parameter handler object.
	* (Only for optimization with DAKOTA)
	*/
	void set_parameters (const std::vector<std::string>& name_dvar, 
			     const std::vector<double>& value_dvar, 
			     ParameterHandler& param);
	/**
	* Member function used to read in data and initialize the necessary data
	* to compute the coefficients.
	*/
	void initialize (ParameterHandler &param);
	
	/**
	* Member function to compute the electron conductivity (Isotropic properties). If the electron conductivity depends on the solution, then the solution is set in member function set_solution(std::vector< std::vector< double > * >, std::vector< std::string >)
	*/
	double get_electron_conductivity() const;
	
	/**
	* Member function to compute the derivatives of the electron conductivity (Isotropic properties) with respect to the solution. The derivative is computed with respect to the variables that you have indicted using the member function set_derivative_flags(std::vector< std::string > &flags).
	*/
	void get_electron_conductivity_derivative(double &, std::vector<double>&) const;
	/**
	* Member function to compute the thermal conductivity (Isotropic properties). If the thermal conductivity depends on the solution, then the solution is set in member function set_solution(std::vector< std::vector< double > * >, std::vector< std::string >)
	*/
	//double get_thermal_conductivity() const;
	
	/**
	* Member function to compute the derivatives of the thermal conductivity (Isotropic properties) with respect to the solution. The derivative is computed with respect to the variables that you have indicted using the member function set_derivative_flags(std::vector< std::string > &flags).
	*/
	//void get_thermal_conductivity_derivative(double &, std::vector<double>&) const;
	/**
	* Member function to compute the Young's modulus (Isotropic properties). If the thermal conductivity depends on the solution, then the solution is set in member function set_solution(std::vector< std::vector< double > * >, std::vector< std::string >)
	*/
	//double get_youngs_modulus() const;
	
	/**
	* Member function to compute the derivatives of the Young's modulus (Isotropic properties) with respect to the solution. The derivative is computed with respect to the variables that you have indicted using the member function set_derivative_flags(std::vector< std::string > &flags).
	*/
	//void get_youngs_modulus_derivative(double &, std::vector<double>&) const;
	/**
	* Member function to compute the Poisson's ratio (Isotropic properties). If the thermal conductivity depends on the solution, then the solution is set in member function set_solution(std::vector< std::vector< double > * >, std::vector< std::string >)
	*/
	//double get_poissons_ratio() const;
	
	/**
	* Member function to compute the derivatives of the Poisson's ratio (Isotropic properties) with respect to the solution. The derivative is computed with respect to the variables that you have indicted using the member function set_derivative_flags(std::vector< std::string > &flags).
	*/
	//void get_poissons_modulus_derivative(double &, std::vector<double>&) const;
	/**
	* Member function to compute the expansion coefficient (Isotropic properties). If the thermal conductivity depends on the solution, then the solution is set in member function set_solution(std::vector< std::vector< double > * >, std::vector< std::string >)
	*/
	//double get_expansion_coefficient() const;
	
	/**
	* Member function to compute the derivatives of the expansion coefficient (Isotropic properties) with respect to the solution. The derivative is computed with respect to the variables that you have indicted using the member function set_derivative_flags(std::vector< std::string > &flags).
	*/
	//void get_expansion_coefficient_derivative(double &E, std::vector<double>& dE) const;
	};
  }
}


#endif
