//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: nafion.h
//    - Description:  Class to represent and return bulk properties for Nafion (Polymer Electrolyte Membrane)
//    - Developers: M. Secanell (2009-13) and Madhur Bhaiya (2012-13)
//    - Id: $Id: nafion.h 1460 2013-08-30 22:28:49Z madhur $ 
//
//---------------------------------------------------------------------------


#ifndef _FUELCELLSHOP__NAFION_H
#define _FUELCELLSHOP__NAFION_H

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

#include "polymer_electrolyte_material_base.h"
#include "PureGas.h"

//Include STL
#include<cmath>
#include<iostream>

class NafionTest;

namespace FuelCellShop
{
    namespace Material
    {
        /**
         * Class used to store data and provide information regarding the electrolyte. In this case
         * the electrolyte that is setup is Nafion. In particular the data coded in this class is obtained
         * from the literature from Nafion 1100, usually from NRE 211 properties. 
         * 
         * Literature sources are:
         * - XXX
         * - XXX
         * 
         * \TODO (Madhur please fill this part -- Add references to each section too.)
         * 
         * In the parameter file the following parameters can be specified within the subsections specified below:
         * 
         * @code
         * subsection Fuel cell data
         *   subsection Materials
         *     subsection Nafion
         *       set Method for sorption isotherm = Previous        # Options are Previous | Liu09 -- Method to compute equilibrium lambda from sorption isotherm - Depends on Water molar fraction and Temperature
         *       set Method to compute proton conductivity = NRE211 # Options are: Constant|Springer|NRE211|Iden11 -- Method used to compute proton conductivity inside the membrane - Depends on lambda and T
         *       set Proton conductivity [S/cm] = 0.1                      # Proton conductivity inside the membrane, S/cm
         *       set Method to compute water diffusion = Springer   # Options are: "Constant|Springer|Motupally" -- Method used to compute diffusion of water inside the membrane - Depends on lambda and T
         *       set Water diffusion coefficient = 2e-              # Water diffusion coefficient inside the membrane
         *       set Electro-osmotic drag method = Springer         # Options are: Constant|Springer -- Method to compute nd. Springer depends on water content
         *       set Electro-osmotic drag coefficient = 1.0         # Number of water molecules dragged by one proton
         *       set Oxygen diffusion coefficient [cm^2/s] = 9.726e-6    # From J. Peron et al., “Properties of Nafion NR-211 membranes for PEMFCs”
         *       set Proton diffusion coefficient [cm^2/s] = 9.2e-5      # Peter Dobsons thesis page 42
         *       set Henry's Law Constant for Oxygen [Pa cm^3/mol] = 3.1664e10
         *       set Henry's Law Constant for Hydrogen [Pa cm^3/mol] = 6.69e10
         *     end
         *   end
         * end
         * 
         * @endcode
         * 
         * <h3>Usage Details:</h3>         
         * 
         * As with most routines you need to first declare_parameters, then set_parameters (only if using optimization/parameter estimation), then initialize. After this the class is ready for use.
         * 
         * See for example FuelCell::OperatingConditions class for more details.
         * 
         * \authors M. Secanell and M. Bhaiya
         * 
         */
        class Nafion :
        public PolymerElectrolyteBase
        {
        public:
            
            /** 
             * Name of the class. This name is used to select the layer.
             */
            static const std::string concrete_name;
            
            ///@name Friend class for Unit Testing
            //@{                     
            
            /**
             * Friend class for testing purposes.
             */
            friend class ::NafionTest;
            //@}
            
            ///@name Constructors, destructor, and parameter initalization
            //@{
            /** PROTOTYE Constructor 
             * 
             * \warning For internal use only.
             */
            Nafion(const bool);
            
            /** Constructor 
             * The constructor initialize parameters using the default values. This is
             * so that if I do not want to call declare_parameters and initialize, I can
             * still use the routine with the hard coded values.
             * 
             * \todo1 Make this private and remove name.
             */
            Nafion(std::string name = "Nafion");
            
            /**
             * Destructor
             */
            ~Nafion();
            
            /** Declare parameters.
             * 
             * \todo1 Make it protected
             * 
             * In the parameter file the following parameters can be specified within the subsections specified below:
             * 
             * @code
             * subsection Fuel cell data
             *   subsection Materials
             *     subsection Nafion
             *       set Method for sorption isotherm = Previous        # Options are Previous | Liu09 -- Method to compute equilibrium lambda from sorption isotherm - Depends on Water molar fraction and Temperature
             *       set Method to compute proton conductivity = NRE211 # Options are: Constant|Springer|NRE211|Iden11 -- Method used to compute proton conductivity inside the membrane - Depends on lambda and T
             *       set Proton conductivity [S/cm] = 0.1                      # Proton conductivity inside the membrane, S/cm
             *       set Method to compute water diffusion = Springer   # Options are: "Constant|Springer|Motupally" -- Method used to compute diffusion of water inside the membrane - Depends on lambda and T
             *       set Water diffusion coefficient = 2e-              # Water diffusion coefficient inside the membrane
             *       set Electro-osmotic drag method = Springer         # Options are: Constant|Springer -- Method to compute nd. Springer depends on water content
             *       set Electro-osmotic drag coefficient = 1.0         # Number of water molecules dragged by one proton
             *       set Oxygen diffusion coefficient [cm^2/s] = 9.726e-6    # From J. Peron et al., “Properties of Nafion NR-211 membranes for PEMFCs”
             *       set Proton diffusion coefficient [cm^2/s] = 9.2e-5      # Peter Dobsons thesis page 42
             *       set Henry's Law Constant for Oxygen [Pa cm^3/mol] = 3.1664e10
             *       set Henry's Law Constant for Hydrogen [Pa cm^3/mol] = 6.69e10
             *     end
             *   end
             * end
             * 
             * @endcode
             */
            virtual void declare_parameters(ParameterHandler &param) const;
            
            /**
            * Member function used to change the values in the parameter file for a given list
            * of parameters. 
            * - \p name_dvar should ideally contain the string as it would appear in the input parameter file.
            * - \p value_dvar contains the value that the variable should be set at.
            * - \p param is the ParameterHandler that contains all the information that has been read from file.
            * 
            * List of available std::string that can be changed, \em i.e. \p \b name_dvar are given below:
            * - \b "sigma_p" : protonic conductivity [\p S/cm] of Nafion.
            * - \b "diffusion_w" : water diffusion coefficient [\p cm^2/s] of Nafion.
            * - \b "D_O2_N" : oxygen diffusion coefficient [\p cm^2/s] of Nafion.
            * - \b "H_O2" : Henry's law constant for oxygen dissolution in Nafion [\p Pa-cm^3/mol].
            * - \b "H_H2" : Henry's law constant for hydrogen dissolution in Nafion [\p Pa-cm^3/mol].
            * 
            * \todo1 Make it protected
            */
            virtual void set_parameters (const std::vector<std::string>& name_dvar, 
                                 const std::vector<double>& value_dvar, 
                                 ParameterHandler& param);
            
            /**
             * Member function used to read in data and initialize the necessary data from the parameter file to compute the coefficients.
             */
            virtual void initialize (ParameterHandler &param);
            
            //@}
            
            ///@name Transport properties and derivatives accessor methods
            //@{
            
            /**
             * Compute the equilibrium water content, \f$ \lambda_{eq} \f$, inside the Nafion for vapor-equilibriated case, at
             * every quadrature point in the cell. It takes a vector as an input argument and values are passed by reference.
             * It is required to call #set_water_molar_fraction method atleast before computing sorption isotherm values.
             */
            virtual void sorption_isotherm(std::vector<double>&) const;
            /**
             * Compute the derivatives for water sorption source terms, \f$ \frac{\partial \lambda_{eq}}{\partial u} \f$, at every quadrature 
             * point in the cell. The derivatives are computed based on the flags set by the #set_derivative_flags method. It takes map as an input argument by reference, in 
             * which \p Key corresponds to the variable about which derivative is being computed, \em i.e., #VariableNames and \p Value corresponds to
             * the vector storing the derivative values. It is required to call #set_water_molar_fraction method atleast before computing sorption isotherm values. Also, this 
             * method should only be called when atleast there is one derivative flag.
             */
            virtual void sorption_isotherm_derivative(std::map < VariableNames, std::vector<double> >&) const;
            
            /**
             * Compute the proton conductivity, \f$ \sigma_{H^+} \f$ [\p S/cm], inside the nafion for a case of constant \f$ \lambda \f$ and 
             * \f$ T \f$. Before calling this method, #set_T method should be used for setting the constant temperature value. If the constant \f$ \lambda \f$ 
             * value is required to be different from the default value \b 12.0, then #set_lambda method should be used. It takes a double as an input argument and value is passed by reference. 
             * \remarks This method is recommended only for multi-scale models. In case of macrohomogeneous models, constant values should always be set before the variable ones,
             * otherwise it may give erroneous results.
             */
            virtual void proton_conductivity(double&) const;
            /**
             * Compute the proton conductivity, \f$ \sigma_{H^+} \f$ [\p S/cm], inside the nafion, at every quadrature point in the cell. It is computed 
             * as a function of \f$ \lambda \f$ and \f$ T \f$, hence solution setting methods should be called accordingly for them, whether constant or variable. It is recommended to set the
             * constant ones in the initialization of the application, unless you are solving for multi-scale model. It takes vector as an input argument and values are passed by reference.
             * \warning Precaution must be used while using two different solution setting methods, \em i.e., either constant or variable for a single solution variable on the same electrolyte object. You 
             * cannot use constant solution setting methods for this object, if variable solution setting methods are already used for a single solution variable. In case 
             * of constant \f$ \lambda \f$ and \f$ T \f$, vector size should be set equal to the number of quadrature points, before passing to this function. 
             */
            virtual void proton_conductivity(std::vector<double>&) const;
            /**
             * Compute the derivatives of proton conductivity at every quadrature point in the cell. The derivatives are computed based on the flags
             * set by the #set_derivative_flags method. It takes map as an input argument by reference, in  which \p Key corresponds to the variable about 
             * which derivative is being computed, \em i.e., #VariableNames and \p Value corresponds to the vector storing the derivative values. Note that 
             * this method should not be used when both \f$ \lambda \f$ and \f$ T \f$ are constant. Hence, atleast #set_membrane_water_content or #set_temperature should be 
             * called before using this method. Also, this method should only be called when atleast there is one derivative flag.
             */
            virtual void proton_conductivity_derivative(std::map< VariableNames, std::vector<double> >&) const;
            
            /**
             * Compute the water diffusivity, \f$ D_{\lambda} \f$ [\p cm^2/s], inside the nafion for a case of constant \f$ \lambda \f$ and 
             * \f$ T \f$. Before calling this method, #set_T method should be used for setting the constant temperature value. If the constant \f$ \lambda \f$ 
             * value is required to be different from the default value \b 12.0, then #set_lambda method should be used. It takes a double as an input argument and value is passed by reference. 
             * \remarks This method is recommended only for multi-scale models. In case of macrohomogeneous models, it's not required in particular because \f$ D_{\lambda} \f$ is 
             * required only when \f$ \lambda \f$ is one of the solution variables in the application.
             */
            virtual void water_diffusivity(double&) const;
            /**
             * Compute the water diffusivity, \f$ D_{\lambda} \f$ [\p cm^2/s], inside the nafion at every quadrature point in the cell. It is computed 
             * as a function of \f$ \lambda \f$ and \f$ T \f$, hence solution setting methods should be called accordingly for them, whether constant or variable. It is recommended to set the
             * constant ones in the initialization of the application. Note that this method should not be used when both \f$ \lambda \f$ and \f$ T \f$ are constant. Hence, 
             * atleast #set_membrane_water_content or #set_temperature should be called before using this method. It takes vector as an input argument and values are passed by reference.
             */
            virtual void water_diffusivity(std::vector<double>&) const;
            /**
             * Compute the derivatives of water diffusivity at every quadrature point in the cell. The derivatives are computed based on the flags
             * set by the #set_derivative_flags method. It takes map as an input argument by reference, in  which \p Key corresponds to the variable about 
             * which derivative is being computed, \em i.e., #VariableNames and \p Value corresponds to the vector storing the derivative values. Note that 
             * this method should not be used when both \f$ \lambda \f$ and \f$ T \f$ are constant. Hence, atleast #set_membrane_water_content or #set_temperature should be 
             * called before using this method. Also, this method should only be called when atleast there is one derivative flag.
             */
            virtual void water_diffusivity_derivative(std::map< VariableNames, std::vector<double> >&) const;
            
            /**
             * Compute the electro-osmotic drag coefficient inside the nafion at every quadrature point in
             * the cell. It takes vector as an input argument and values are passed by reference. Note that the double return method is not implemented for electro-osmotic drag 
             * coefficients because it is required only when \f$ \lambda \f$ is one of the solution variables in the application. #set_membrane_water_content 
             * should be called before using this method.
             */
            virtual void electroosmotic_drag(std::vector<double>&) const;
            /**
             * Compute the derivatives of electro-osmotic drag coefficient, at every quadrature point in the cell. The derivatives are computed based on the flags
             * set by the #set_derivative_flags method. It takes map as an input argument by reference, in  which \p Key corresponds to the variable about 
             * which derivative is being computed, \em i.e., #VariableNames and \p Value corresponds to the vector storing the derivative values. #set_membrane_water_content 
             * should be called before using this method. Also, this method should only be called when atleast there is one derivative flag.
             */
            virtual void electroosmotic_drag_derivative(std::map< VariableNames, std::vector<double> >&) const;
            
            /**
             * Compute the oxygen diffusivity, [\p cm^2/s], inside the nafion for a constant \f$ T \f$. Before calling this method, #set_T method should be used for 
             * setting the constant temperature value. It takes a double as an input argument and value is passed by reference.
             */
            virtual void oxygen_diffusivity(double&) const;
            /**
             * Compute the oxygen diffusivity, [\p cm^2/s], inside the nafion at every quadrature point in the cell. It takes
             * vector as an input argument and values are passed by reference. #set_temperature should be called before using this method.
             */
            virtual void oxygen_diffusivity(std::vector<double>&) const;
            /**
             * Compute the derivatives of oxygen diffusivity, at every quadrature point in the cell. The derivatives are computed based on the flags
             * set by the #set_derivative_flags method. It takes map as an input argument by reference, in  which \p Key corresponds to the variable about 
             * which derivative is being computed, \em i.e., #VariableNames and \p Value corresponds to the vector storing the derivative values. #set_temperature should 
             * be called before using this method. Also, this method should only be called when atleast there is one derivative flag.
             */
            virtual void oxygen_diffusivity_derivative(std::map< VariableNames, std::vector<double> >&) const;
            
            /**
             * Compute the proton diffusivity, [\p cm^2/s], inside the nafion for constant case. It takes a double as an 
             * input argument and value is passed by reference. It basically returns a constant value specified using the parameter file.
             */
            virtual void proton_diffusivity(double&) const;

            //@}
            
        private:
            
            ///@name Instance Delivery
            //@{
             /**
             * This member function is used to create an object of type Nafion material.
             * 
             * \warning This class MUST be redeclared in every child.
             */
            virtual boost::shared_ptr<FuelCellShop::Material::PolymerElectrolyteBase > create_replica ()
            {
                return boost::shared_ptr<FuelCellShop::Material::PolymerElectrolyteBase > (new FuelCellShop::Material::Nafion ());
            }
            /**
             * Create prototype for the layer
             */            
            static Nafion const* PROTOTYPE;
            //@}
            
            /**
             * Method to modify parameters, without declaring and initializing the parameter file. It takes two arguments, first one (unsigned int) corresponds 
             * to a particular type of method and second one (std::string) corresponds to various implementations for that method type. \b index corresponds to following methods:
             *  - \b 1: protonic conductivity methods
             *  - \b 2: water diffusivity methods
             *  - \b 3: electro-osmotic drag methods
             *  - \b 4: sorption isotherm methods
             * \note This function is used by the unit testing for this class. For \b method strings, look at declare_parameters() method.
             */
            inline void modify_parameters(const unsigned int& index, const std::string& method)
            {
                Assert(index>=1 && index<=6, ExcMessage("Wrong index input in Nafion::modify_parameters."));
                
                if (index == 1)
                    method_conductivity = method;
                else if (index == 2)
                    method_diffusivity = method;
                else if (index == 3)
                    method_electroosmotic_drag = method;
                else if (index == 4)
                    method_sorption = method;
            }
            
        };
    }
}
#endif


      
      
    
