//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2009-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: adaptive_refinement.h
//    - Description: Child of ApplicationCopy used to implement adaptive refinement
//    - Developers: M. Secanell
//    - $Id: adaptive_refinement.h 1389 2013-08-27 23:29:33Z zingan $
//
//---------------------------------------------------------------------------

#ifndef _FUELCELL__ADAPTIVEREFINEMENT_H_
#define _FUELCELL__ADAPTIVEREFINEMENT_H_

// Deal.II include files:
#include <base/parameter_handler.h>
#include <base/smartpointer.h>
#include <grid/tria.h>

// Appframe include files:
#include <appframe/base.h>
#include <appframe/newton.h>
//#include <appframe/fe_vectors.h>

// Fuel cell include files
#include "optimization_block_matrix_application.h"

// STD include files
#include <string>
#include <stdio.h>
#include <stdlib.h>


namespace AppFrame
{
/**
 * This class is initialized with an application that describes the linearization of the problem that
 * we would like to solve and the nonlinear solved that drives the process (usually a Newton loop).
 * Then, this class implements the adaptive refinement loop for the application.
 *
 * This class needs two inputs:
 * - Refinement levels
 * - Analytical gradients
 *
 * \author M. Secanell, 2006-13
 */
template <int dim>
class AdaptiveRefinement
{
public:
    /**
     * Constructor
     */
    AdaptiveRefinement ( AppFrame::OptimizationBlockMatrixApplication<dim>& app_lin,
                         AppFrame::ApplicationCopy& app,
                         const AppFrame::FEVector& solution = AppFrame::FEVector() );

    /**
     *
     */
    ~AdaptiveRefinement();

    /**
     * Declare all parameters that are needed for:
     *   - control of adaptive refinement loop
     *   - control of output options during the loop, e.g. return the solution at each cycle
     *
     * Currently the options implemented are:
     *
     * @code
     * subsection Adaptive refinement
     *   set Number of Refinements = 1                   # Number of adaptive refinements used
     *   set Output initial mesh = false                 # Set flag to true if you want to output a EPS figure of the initial mesh using the value in file initial mesh
     *   set Output initial mesh filename = initial_mesh # File where the initial mesh will be output
     *   set Output initial solution = false             # Set flag to true if you want to output the initial solution to file
     *   set Initial solution filename = initial_sol     # File where the initial solution will be output
     *   set Output intermediate solutions = true        # Output the solution after each adaptive refinement level
     *   set Output intermediate responses = true        # Output any functional such as current density at each refinement level
     *   set Output final solution = true                # Output final solution to a file named fuel_cell-sol-cycleX where X is the cycle name.
     *   set Output solution for transfer = false        # Check whether a solution on a refined grid should be output to file on a coarse mesh
     *   set Read in initial solution from file = false  # Check whether a stored solution should be read from file and applied to the grid
     * end
     * @endcode
     *
     */
    void declare_parameters ( ParameterHandler& param ) const;

    /**
     * Reset the parameters used to compute the linear solver equations.
     *
     * @note Mainly used for optimization
     */
    void set_parameters ( const std::vector<std::string>& name_dvar,
                          const std::vector<double>& value_dvar,
                          ParameterHandler& param );
    /**
     * Set up how many equations are needed and
     * read in parameters for the parameter handler in order to initialize data
     */
    void initialize ( ParameterHandler& param );

    /**
     * Solve the nonlinear problem
     */
    void solve ( const std::string param_file, ParameterHandler& param );

    /**
     * Member function used to test the derivatives
     */
    void test_derivatives ( const std::string input_file,
                            const std::string dvar,
                            const double value,
                            std::vector<double>& resp,
                            std::vector<std::vector<double> >& dresp,
                            const bool gradient = true );

    /**
     * Print parameters:
     */
    void print_parameters() const;

    /**
     * Run Newton application
     */
    void run_Newton();

    /**
     * Run Newton application
     */
    void run_Newton ( std::vector<double>& resp );

    /**
     * Run Newton application
     */
    void run_Newton ( std::vector<double>& resp, //std::vector<double> (),
                      std::vector<std::vector<double> >& dresp_dl ); //std::vector<std::vector<double> > (0, std::vector<double> ()));

    /**
     * This function returns
     * \p solution.
     */
    const AppFrame::FEVector& get_solution() const
    {
           return solution;
    }

private:
    /** Flag to output the initial grid to be solved */
    bool output_initial_mesh;
    /** Flag to output the initial grid to be solved */
    bool output_initial_sol;
    /** Filename where to output the initial grid */
    std::string filename_initial_mesh;
    /** Filename where to output the initial grid */
    std::string filename_initial_sol;
    /** */
    bool output_intermediate_sol;
    /** */
    bool output_final_sol;
    /** */
    bool output_intermediate_resp;

    bool output_coarse_solution;

    bool read_in_initial_solution;

    /**
     * Number of initial refinements for the original mesh
     *
     * \todo3 Change the loop so that zero refinements means once through the loop
     */
    unsigned int n_ref;

    /** Compute the gradients?*/
    bool gradients;

    /** Pointer to application */
    AppFrame::OptimizationBlockMatrixApplication<dim> *app_linear;

    /** Poiner to nonlinear application */
    AppFrame::ApplicationCopy *app;

    /**
     * Global FE solution at the support points
     * of the computational domain.
     */
    AppFrame::FEVector solution;
};
}

#endif