//---------------------------------------------------------------------------
//    $Id: newton_base.h 1354 2013-08-17 00:01:22Z secanell $
//
//    Copyright (C) 2007 by Jason Boisvert
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------


#ifndef __deal2__appframe__newton_base_h
#define __deal2__appframe__newton_base_h

#include <appframe/base.h>
#include <appframe/residual.h>
#include <lac/solver_control.h>

using namespace dealii;
using namespace AppFrame;

namespace AppFrame
{
/**
 * Base class for all classes peforming Newton's iteration.  Childeren of this class largley differ
 * by global seacrh method employed
 *
 * @author Jason Boisvert, 2009
 */

	

	class newtonBase:public ApplicationCopy
		{
		public:
			/**
			 * The Event set by Jason if
			 * convergence is becoming bad
			 * and a new matrix should be
			 * assembled.
			 */
			static const AppFrame::Event bad_derivative; 
			/**
			 * Constructor, receiving the application computing the residual and
			 * solving the linear problem.
			 */
			newtonBase(ApplicationBase& app);
			
			/**
			 * Instead of assembling, this function only sets a flag, such that
			 * the inner application will be required to assemble a new derivative
			 * matrix next time solve() is called.
			 */
			virtual void assemble();
			
			virtual double residual(AppFrame::FEVector& dst,
									const AppFrame::FEVectors& rhs);
			
			/**
			 * The actual Newton solver.
			 */
			virtual void solve(AppFrame::FEVector& u,
							   const AppFrame::FEVectors& in_vectors);
			/**
			 * Declare the input parameters. The following are the parameters used and their default values:
			 * In section: "Newton"
			 * <ul>
			 * - "Assemble threshold", Default: "0."
			 * - "Debug level", Default: "0",
			 * - "Debug solution", Default: "false", Patterns::Bool()
			 * - "Debug update", Default: "false", Patterns::Bool()
			 * - "Debug residual", "false", Patterns::Bool()
			 * </ul>
			 */
			virtual void declare_parameters (ParameterHandler& param);
			
			/**
			 * Read the parameters local to newtonBase.
			 */
			void _initialize (ParameterHandler& param);
			
			/**
			 * Read the parameters.
			 */
			virtual void initialize (ParameterHandler& param);
			
			/**
			 * Set the maximal residual reduction allowed without triggering
			 * assembling in the next step. Return the previous value.
			 */
			double threshold(double new_value);
			
			/**
			 * Control object for the Newton iteration.
			 */
			void initialize_initial_guess(BlockVector<double>& dst);

			/**
	 		* Control object for the Newton iteration.
	 		*/
			ReductionControl control;
			
			// Output data
			//int numIterations;
			double numIter;
		protected:
			/**
			 * Default residual object if
			 * the function
			 * ApplicationBase::residual()
			 * shoud be used.
			 */
			ApplicationResidual default_residual;
			
			/**
			 * The Residual object.
			 */
			SmartPointer<const Residual> compute_residual;
			/**
			 * This flag is set by the
			 * function assemble(),
			 * indicating that the matrix
			 * must be assembled anew upon
			 * start.
			 */
			bool assemble_now;
			
			/**
			 * Threshold for re-assembling matrix.
			 *
			 * If the quotient of two
			 * consecutive residuals is
			 * smaller than this threshold,
			 * the system matrix is not
			 * assembled in this step.
			 *
			 * @note This parameter should be
			 * adjusted to the residual gain
			 * of the inner solver.
			 */
			double assemble_threshold;
			/**
			 * Print updated solution after
			 * each step into file
			 * <tt>Newton_uNNN</tt>?
			 */
			bool debug_solution;
			/**
			 * Print Newton update after
			 * each step into file
			 * <tt>Newton_dNNN</tt>?
			 */
			bool debug_update;
			/**
			 * Print Newton residual after
			 * each step into file
			 * <tt>Newton_rNNN</tt>?
			 */
			bool debug_residual;
			/**
			 * Write debug output to
			 * #deallog; the higher the
			 * number, the more output.
			 */
			unsigned int debug;	
			/**
			 * Value used to multiply \f$\delta u\f$, the solution update, from the Newton algorithm
			 * during the first overrelax_steps iterations. We will multiply the first
			 * iteration by \f$\alpha \delta u\f$, where \f$ \alpha \f$ is the value of overrelax and then
			 * the following iterations by \f$it \alpha \delta u\f$ where \f$ it \f$ is the iteration number
			 */

		};
}

#endif

		


