//---------------------------------------------------------------------------
//    $Id: solver_utils.h 1354 2013-08-17 00:01:22Z secanell $
//
//    Copyright (C) 2008 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------


#ifndef _SOLVER_UTILS__H
#define _SOLVER_UTILS__H

// Include deal.II classes
#include <lac/block_sparse_matrix.h>
#include <lac/full_matrix.h>
#include "lac/block_vector.h"
#include "appframe/base.h"

// Include STD classes
#include <vector>
#include <fstream>
#include <iostream>
#include <sstream>

// Use namespace of deal.II
using namespace dealii;

/**
 * This class is used to include routines that are used
 * in the solve() routine of several applications such as
 * AppCathode and AppPemfc.
 *
 * @author Marc Secanell, 2008
 */
class SolverUtils
{
 public:
  /**
   * Constructor
  */
  SolverUtils();
  
  /**
   * Check that the diagonal of the matrix does not have any zeros. If it does, 
   * the code issues a warning on optimized mode or throws an exception on debug mode
   */
  void check_diagonal(const BlockSparseMatrix<double>& A) const;

  /**
   * Output diagonal elements of the stiffness matrix to the screen
   */
  void output_diagonal(const BlockSparseMatrix<double>& A) const;
  
  /**
   * Print the diagonal elements of a matrix to a file named diag_matrix.dat
   *
   */
  void print_diagonal(const BlockSparseMatrix<double>& A, 
		      const std::string& file = std::string("diag_matrix.dat")) const;
  /**
   * This member function is used to make sure that the BlockSpareMatrix
   * has no zeros in the diagonal. If it has zeros, then, the diagonal is filled
   * with the average value between the largest and the smallest number in the
   * diagonal.
   *
   * In most problems you will need to use this class to remove zeros in the diagonal
   * for the equations that are not physical in certain domains. For example, when solving
   * the catalyst layer, the membrane potential inside the GDL is NOT a physical quantity. 
   * Instead of adding a small number during assembly which would affect the solution,
   * we add zeros. Then, we modify the diagonal matrix so that the value is zero everywhere
   * on that domain and the fluxes at the boundaries between domains are not accounted for.
   */
  void repair_diagonal (BlockSparseMatrix<double>& A);

  /**
   * This member function is used to make sure that the BlockSpareMatrix
   * has no zeros in the diagonal. If it has zeros, then we rewrite the equation
   * so that we ensure the solution at that node has a value of zero. This is done
   * by passing the solution vector to the function and setting the RHS to this value.
   * The zero in the system matrix is set to 1, so that we have \f$ 1(-\delta u) = \delta u \f$
   */  
  void repair_diagonal (BlockSparseMatrix<double>& A, AppFrame::FEVector& ,  const AppFrame::FEVector& );
};


#endif //_SOLVER_UTILS__H
