//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: fcst_utilities.h
//    - Description: Namespace carrying a lot of utility functions
//    - Developers: Madhur Bhaiya, Valentin N. Zingan
//    - $Id: fcst_utilities.h 1380 2013-08-22 02:05:45Z zingan $
//
//---------------------------------------------------------------------------

#ifndef _FUELCELLSHOP__FCST_UTILITIES_H
#define _FUELCELLSHOP__FCST_UTILITIES_H

//Include STL and boost libraries
#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <cctype>
#include <boost/lexical_cast.hpp>
#include <fstream>

//Include deal.ii classes
#include <base/utilities.h>
#include <base/exceptions.h>
#include <base/parameter_handler.h>


using namespace dealii;

namespace FcstUtilities
{
    /**
     * Exception shown when a particular string can't be converted to Number type.
     */
    DeclException1 (ExcWrongString,
                    std::string,
                    << "Can't convert the string " << arg1
                    << " to the desired Number type");
    
    /**
     * Function to convert a std::string into Number template. Currently, it is templated only for
     * \p double and \p unsigned \p int types. But it can be easily extended for other numeric data types.
     *
     * Input string should be free from blank spaces at the beginning and end.
     *
     * \note There is a similar function in \p deal.ii \em Utilities namespace, but it has certain limitations
     * as it doesn't do effective error checking for wrong input arguments. For that purpose, a new function is written here using boost::lexical_cast.
     */
    template <typename NumType>
    NumType string_to_number(const std::string&);
    
    /**
     * Function to convert a std::vector< std::string > into std::vector < Number template >. Currently, it is templated only for
     * \p double and \p unsigned \p int types. But it can be easily extended for other numeric data types.
     *
     * Input strings should be free from blank spaces at the beginning and end.
     *
     * \note Before extending its capabilities to deal with other numeric data types, capability of #string_to_number should be extended.
     */
    template <typename NumType>
    std::vector<NumType> string_to_number(const std::vector<std::string>&);
    
    
    /**
     * Function to convert a std::vector< std::string > into std::map< KeyType, ValueType >.
     *
     * This function is used to convert a string like "2 : 3.86" into a \p std::map where Key is 2 and Value is 3.86.
     * This function is useful to convert a \p Patterns::Map (of \p ParameterHandler class of \p deal.ii) into a useful data map.
     *
     * \warning String should contain ":" to distinguish between key and corresponding value. Also a key should not be repeated in the
     * parameter strings. Also Input strings should be free from blank spaces at the beginning and end.
     *
     * \note Currently, it is templated only to have \p KeyType as \p unsigned \p int and \p ValueType as \p double. But,
     * the capabilities can be extended.
     */

    template <typename KeyType, typename ValueType>
    std::map< KeyType, ValueType > string_to_map(const std::vector<std::string>&);
       


    /**
     * This function is the same as the previous one.
     * The only difference is that this function takes
     * std::string instead of std::vector<std::string>.
     */
    template< typename KeyType, typename ValueType >
    std::map< KeyType, ValueType > string_to_map(const std::string& name);
    


    /**
     * This function is a particular reimplementation
     * of the previous one.
     */
    template<>
    std::map< unsigned int, std::string > string_to_map<>(const std::string& name);
    



    /**
     * This function takes a std::map<KeyType, std::string>, and std::string contains text separated by a \p delimiter.
     * It splits the std::string into std::vector<ValueType>, removing all the leading and trailing spaces.
     * It returns std::map< KeyType, std::vector<ValueType> >. The default value of the \p delimiter is \p semi-colon <b>";"</b>.
     * \note It is necessary to have the delimiter "atleast once" in the \p Value string.
     */
    template< typename KeyType, typename ValueType >
    std::map< KeyType, std::vector<ValueType> > split_mapvalue_list(const std::map< KeyType, std::string >&, const char delimiter = ';');
    
    /**
     * This function prepares
     * an XML file with ALL
     * your data inside.
     */
    void print_parameter_file_XML(ParameterHandler& param);
    
    /**
     * This function returns the address of the fcst root directory.
     * It presumes that you are executing from within the data folder - an exception will be thrown otherwise.
     */
    std::string find_fcst_root();
    
    
    /**
     * This function runs external python scripts. User must provide the script name and argument.
     * It is presumed that the script resides in the python folder.
     */
    void run_python(std::string script_name,std::string arg);
    
    /**
     * Function for casting from double to string. Uses boost lexical cast.
     */
    template <typename NumType>
    std::string number_to_string(const NumType &num);


} //FcstUtilities
#endif //_FUELCELLSHOP__FCST_UTILITIES_H
