// ----------------------------------------------------------------------------
//
// FCST: Fuel Cell Simulation Toolbox
//
// Copyright (C) 2006-2013 by Energy Systems Design Laboratory, University of Alberta
//
// This software is distributed under the MIT License
// For more information, see the README file in /doc/LICENSE
//
// - Class: fcst_variables.h
// - Description: Header file storing list of fcst variables and solution variable structures etc.
// - Developers: Phil Wardlaw,    University of Alberta
//               Madhur Bhaiya,   University of Alberta
// - $Id: base_layer.h 1267 2013-08-07 17:22:53Z madhur $
//
// ----------------------------------------------------------------------------

#ifndef _FUELCELLSHOP_VARIABLES_H
#define _FUELCELLSHOP_VARIABLES_H

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/point.h>
#include <base/function.h>
#include <lac/vector.h>
#include <fe/fe_values.h>

//Include STL
#include <cmath>
#include <iostream>

// Include OpenFCST routines:
#include "utils/system_management.h"

using namespace dealii;

namespace FuelCellShop
{    
    /**
     * This structure is used to store values for a particular solution variables, at all quadrature points in the cell. It is
     * best utilized while setting previous Newton iteration values from application to the layer classes etc. It is recommended that
     * those classes which require certain old solution values to compute effective properties etc. should contain this structure as their 
     * data member. This structure has four constructors. Default constructor doesn't set any value. It also sets the boolean member #initialized to \p \b false. This
     * can be checked by using #is_initialized member function and is helpful to determine whether any values are stored or not inside the object. The other constructor
     * takes std::vector<double>* corresponding to the solution variable values and enumeration #VariableNames representing the name of the
     * solution variable. The #VariableNames of the solution variable can be accessed by get_variablename() method. In order
     * to access value at a particular quadrature point, <b>operator [ ]</b> is provided. Other two constructors are useful in the case when we set some default values
     * for a particular variable (which is not being solved in the current application).
     * 
     * <h3>Sample Usage:</h3>
     * @code
     *  // Inside the application, let's pass temperature vector from old solution to the layer class.
     *  // Old solution is accessed from AppFrame::DoFApplication<dim>::CellInfo cell_info object.
     * 
     *  SolutionVariable temperature_old( &cell_info.values[last_iter][temp_index], VariableNames::temperature_of_REV );
     *  layer.set_solution( temperature_old );
     * 
     *  // Let us compute the temperature in Celsius at first quadrature point in the layer/kinetics etc. class:
     * 
     *  SolutionVariable temperature_cell;
     *  // Inside the layer/kinetics etc. class, temperature values are set into the temperature_cell structure using the set_solution method from the application.
     * 
     *  if (temperature_cell.is_initialized() && (temperature_cell.get_variablename() == VariableNames::temperature_of_REV)
     *  {
     *     double temp_celsius = temperature_cell[0] - 273.15;
     *  } 
     * @endcode
     * 
     * \author Philip Wardlaw
     * \author Madhur Bhaiya
     * \author Marc Secanell
     */
    
    struct SolutionVariable
    {
        ///@name Initialization
        //@{
        
        /** Default Constructor */
        SolutionVariable()
        {
            data = NULL;
            initialized = false;
        }
        
        /**
         * Constructor for setting up the pointer to solution variable values and name of the solution variable.
         */
        SolutionVariable(const std::vector<double>* data_in, const VariableNames& name_in)
        {
            data = data_in;
            name = name_in;
            initialized = true;
        }
        
        /**
         * Constructor to initialize the solution variable values, taking a default value and size of the vector as an input argument.
         * \note This constructor is recommended when the solution variable values are not being solved for in the application, while
         * it is a required variable to compute certain properties. Then, it can be used to set some default values.
         */
        
        SolutionVariable(const double& value, const unsigned int& length, const VariableNames& name_in)
        {
            data = NULL;
            default_data = std::vector<double>(length, value);
            name = name_in;
            initialized = true;
        }

        /**
         * Constructor to initialize the solution variable values, taking values as an input vector argument.
         * \note This constructor is recommended when the solution variable values are not being solved for in the application, while
         * it is a required variable to compute certain properties. Then, it can be used to set some default values.
         */
        SolutionVariable(const std::vector<double>& data_in, const VariableNames& name_in)
        {
            data = NULL;
            default_data = data_in;
            name = name_in;
            initialized = true;
        }

        //@}
        
        ///@name Accessors
        //@{
        /**
         * Operator to access the value at a particular quadrature point in the cell.
         * \note Solution values vector should be initialized before using this operator. Also, the input index should not be out of range.
         */
        const double& operator[](const unsigned int& i) const
        {
            Assert( initialized, ExcMessage("SolutionVariables struct is not initialized !!!") );
            
            if (data != NULL)
            {
                Assert( i < data->size(), ExcMessage("Index is out of range in operator[] for SolutionVariables struct.") );
                return data->at(i);
            }
            else if (data == NULL)
            {
                Assert( i < default_data.size(), ExcMessage("Index is out of range in operator[] for SolutionVariables struct.") );
                return default_data.at(i);
            }
        }

        /**
         * Function to determine whether the structure is initialized or not.
         */
        bool is_initialized() const
        {
            return initialized;
        }
          
        /**
         * Function to the length of the internal data element.
         */
        unsigned int size() const
        {
            unsigned int answer = 0;
            //If initialized and pointer is not NULL
            if (initialized && (data != NULL))
                answer = data->size();
            else if (initialized && (data == NULL))
                answer = default_data.size();
            
            return answer;
        }

        /**
         * Function to get the #VariableNames enumeration corresponding to this struct.
         */
        VariableNames get_variablename() const
        {
            Assert( initialized, ExcMessage("SolutionVariable not initialized !!!") );
            return name;
        }
        //@}
        
        
    private:
        /** Default data member */
        std::vector<double>           default_data;
        
        /** Pointer storing the solution variable values at all quadrature points in the cell. */
        const std::vector<double>*    data;
        
        /** #VariableNames enumeration */
        VariableNames                  name;
        
        /** Boolean member determining initialized or not. */
        bool                          initialized;
    };

    ///@name Unary Predicate Functions for SolutionVariable structure
    //@{
    
    /**
    * Unary Predicate to return true if a SolutionVariable object belongs to #protonic_electrical_potential.
    */
    static bool is_phiM(const SolutionVariable& sol_var)
    {
        return (sol_var.get_variablename() == protonic_electrical_potential);
    }
    
    /**
    * Unary Predicate to return true if a SolutionVariable object belongs to #electronic_electrical_potential.
    */
    static bool is_phiS(const SolutionVariable& sol_var)
    {
        return (sol_var.get_variablename() == electronic_electrical_potential);
    }
    
    //@}

    
}  // FuelCellShop
    
#endif // _FUELCELLSHOP_VARIABLES_H
