// ----------------------------------------------------------------------------
//
// FCST: Fuel Cell Simulation Toolbox
//
// Copyright (C) 2006-2013 by Energy Systems Design Laboratory, University of Alberta
//
// This software is distributed under the MIT License
// For more information, see the README file in /doc/LICENSE
//
// - Class: initial_and_boundary_data.h
// - Description: This namespace contains data and methods
//                that handle initial and boundary data of
//                a problem at hand
// - Developers: Valentin N. Zingan, University of Alberta
// - Id: $Id: initial_and_boundary_data.h 1384 2013-08-22 21:40:07Z zingan $
//
// ----------------------------------------------------------------------------

#ifndef _FCST_FUELCELL_INITIAL_AND_BOUNDARY_DATA_H_
#define _FCST_FUELCELL_INITIAL_AND_BOUNDARY_DATA_H_

#define _IS_NOT_CONSTANT_ 1.e300

#include <boost/lexical_cast.hpp>

#include <base/point.h>
#include <base/function.h>

#include <lac/vector.h>
#include <lac/block_vector.h>
#include <lac/block_sparse_matrix.h>

#include <dofs/function_map.h>

#include <fe/mapping_q1.h>
#include <fe/mapping_q.h>

#include <numerics/vector_tools.h>
#include <numerics/matrix_tools.h>

#include "system_management.h"
#include "solver_utils.h"

using namespace dealii;
using namespace AppFrame;

/**
 * The typedef for the std::map
 * that reflects the following
 * structure:
 *
 * - \p first  \p argument : name of the solution component,
 * - \p second \p argument : material id,
 * - \p third  \p argument : value of the solution component.
 */
typedef std::map<   std::string , std::map<types::material_id, double>   > component_materialID_value_map;

/**
 * The typedef for the std::map
 * that reflects the following
 * structure:
 *
 * - \p first  \p argument : name of the solution component,
 * - \p second \p argument : boundary id,
 * - \p third  \p argument : value of the solution component.
 */
typedef std::map<   std::string , std::map<types::boundary_id, double>   > component_boundaryID_value_map;

/**
 * This namespace contains data and methods
 * that handle initial and boundary data of
 * a problem at hand.
 *
 * \author Valentin N. Zingan, 2013
 */

namespace FuelCell
{
namespace InitialAndBoundaryData
{

  /**
   * This function does
   * some checkings on
   * its argument \p maps.
   */
  template<typename COMPONENT_xxxID_VALUE_MAP>
  const bool check(const
                   std::vector< COMPONENT_xxxID_VALUE_MAP >& maps)
  {
         // --- types info ---

         const std::type_info& material = typeid(component_materialID_value_map);
         const std::type_info& boundary = typeid(component_boundaryID_value_map);

         const std::type_info& info = typeid(COMPONENT_xxxID_VALUE_MAP);

         // --- checkings ---

         // --- 1 ---

         if( maps.size() == 0 )
         {
                std::cerr << "The argument is empty" << std::endl;
                return false;
         }

         // --- 2 ---

         for(unsigned int i = 0; i < maps.size(); ++i)
                if( maps[i].empty() )
                {
                       std::cerr << "Outer map/maps of the argument is/are empty" << std::endl;
                       return false;
                }

         // --- 3 ---

         for(unsigned int i = 0; i < maps.size(); ++i)
                for(typename COMPONENT_xxxID_VALUE_MAP::const_iterator iter  = maps[i].begin();
                                                                       iter != maps[i].end();
                                                                     ++iter)
                       if( iter->second.empty() )
                       {
                              std::cerr << "Inner map/maps of the argument is/are empty" << std::endl;
                              return false;
                       }

         // --- 4 ---

         if( info == material )
                for(unsigned int i = 0; i < maps.size(); ++i)
                       for(typename COMPONENT_xxxID_VALUE_MAP::const_iterator iter  = maps[i].begin();
                                                                              iter != maps[i].end();
                                                                            ++iter)
                              if( iter->second.find(numbers::invalid_material_id) != iter->second.end() )
                              {
                                     std::cerr << "Invalid material id/ids of the argument" << std::endl;
                                     return false;
                              }

         if( info == boundary )
                for(unsigned int i = 0; i < maps.size(); ++i)
                       for(typename COMPONENT_xxxID_VALUE_MAP::const_iterator iter  = maps[i].begin();
                                                                              iter != maps[i].end();
                                                                            ++iter)
                              if( iter->second.find(numbers::invalid_boundary_id) != iter->second.end() )
                              {
                                     std::cerr << "Invalid boundary id/ids of the argument" << std::endl;
                                     return false;
                              }

         return true;
  }

  /**
   * This function makes
   * piece wise constant initial data.
   */
  template<typename VECTOR, typename DH>
  void make_piece_wise_constant_initial_data(VECTOR&                                              dst,
                                             const Mapping<DH::dimension, DH::space_dimension>&   mapping,
                                             const DH&                                            dof,
                                             const FuelCell::SystemManagement&                    system_management,
                                             const std::vector< component_materialID_value_map >& maps)
  {
         // --- checking ---

         AssertThrow( FuelCell::InitialAndBoundaryData::check< component_materialID_value_map >(maps),
                      ExcMessage("The last argument in the FuelCell::InitialAndBoundaryData::make_piece_wise_constant_initial_data function "
                                 "is wrong : check the previous message") );

         // --- main ---

         const unsigned int n_components = system_management.get_number_of_solution_names();
         AssertThrow( n_components == dof.get_fe().n_components(),
                      ExcDimensionMismatch( n_components , dof.get_fe().n_components() ) );

         for(unsigned int i = 0; i < maps.size(); ++i)
                for(component_materialID_value_map::const_iterator iter  = maps[i].begin();
                                                                   iter != maps[i].end();
                                                                 ++iter)
                {
                       std::map<types::material_id, double> tmp = iter->second;

                       for(std::map<types::material_id, double>::const_iterator iter2  = tmp.begin();
                                                                                iter2 != tmp.end();
                                                                              ++iter2)
                       {
                              if( iter2->second != _IS_NOT_CONSTANT_ )
                              {
                                     typename FunctionMap<DH::space_dimension>::type2 function_map;

                                     const ConstantFunction<DH::space_dimension> constant_function(iter2->second, n_components);
                                     function_map[boost::lexical_cast<unsigned int>(iter2->first)] = &constant_function;

                                     std::vector<bool> component_mask(n_components, false);
                                     component_mask[system_management.solution_name_to_index(iter->first)] = true;

                                     VectorTools::interpolate( dst,
                                                               mapping,
                                                               dof,
                                                               function_map,
                                                               component_mask );
                              }
                       }
                }
  }

  /**
   * This function applies
   * piece wise constant Dirichlet BCs.
   */
  template<typename VECTOR, typename DH>
  void apply_piece_wise_constant_DirichletBCs(VECTOR&                                              dst,
                                              const Mapping<DH::dimension, DH::space_dimension>&   mapping,
                                              const DH&                                            dof,
                                              const FuelCell::SystemManagement&                    system_management,
                                              const std::vector< component_boundaryID_value_map >& maps)
  {
         // --- checking ---

         AssertThrow( FuelCell::InitialAndBoundaryData::check< component_boundaryID_value_map >(maps),
                      ExcMessage("The last argument in the FuelCell::InitialAndBoundaryData::apply_piece_wise_constant_DirichletBCs function "
                                 "is wrong : check the previous message") );

         // --- main ---

         const unsigned int n_components = system_management.get_number_of_solution_names();
         AssertThrow( n_components == dof.get_fe().n_components(),
                      ExcDimensionMismatch( n_components , dof.get_fe().n_components() ) );

         for(unsigned int i = 0; i < maps.size(); ++i)
                for(component_boundaryID_value_map::const_iterator iter  = maps[i].begin();
                                                                   iter != maps[i].end();
                                                                 ++iter)
                {
                       std::map<types::boundary_id, double> tmp = iter->second;

                       for(std::map<types::boundary_id, double>::const_iterator iter2  = tmp.begin();
                                                                                iter2 != tmp.end();
                                                                              ++iter2)
                       {
                              if( iter2->second != _IS_NOT_CONSTANT_ )
                              {
                                     typename FunctionMap<DH::space_dimension>::type function_map;

                                     const ConstantFunction<DH::space_dimension> constant_function(iter2->second, n_components);
                                     function_map[boost::lexical_cast<unsigned int>(iter2->first)] = &constant_function;

                                     std::vector<bool> component_mask(n_components, false);
                                     component_mask[system_management.solution_name_to_index(iter->first)] = true;

                                     std::map<unsigned int, double> boundary_values;

                                     VectorTools::interpolate_boundary_values( mapping,
                                                                               dof,
                                                                               function_map,
                                                                               boundary_values,
                                                                               component_mask );

                                     for(std::map<unsigned int, double>::const_iterator iter3  = boundary_values.begin();
                                                                                        iter3 != boundary_values.end();
                                                                                      ++iter3)
                                     {
                                            dst(iter3->first) = iter3->second;
                                     }
                              }
                       }
                }
  }

  /**
   * This function makes
   * zero boundary values.
   */
  template<typename DH>
  void make_zero_boundary_values(std::map<unsigned int, double>&                      dst,
                                 const Mapping<DH::dimension, DH::space_dimension>&   mapping,
                                 const DH&                                            dof,
                                 const FuelCell::SystemManagement&                    system_management,
                                 const std::vector< component_boundaryID_value_map >& maps)
  {
         // --- checking ---

         AssertThrow( FuelCell::InitialAndBoundaryData::check< component_boundaryID_value_map >(maps),
                      ExcMessage("The last argument in the FuelCell::InitialAndBoundaryData::make_zero_boundary_values function "
                                 "is wrong : check the previous message") );

         // --- main ---

         const unsigned int n_components = system_management.get_number_of_solution_names();
         AssertThrow( n_components == dof.get_fe().n_components(),
                      ExcDimensionMismatch( n_components , dof.get_fe().n_components() ) );

         for(unsigned int i = 0; i < maps.size(); ++i)
                for(component_boundaryID_value_map::const_iterator iter  = maps[i].begin();
                                                                   iter != maps[i].end();
                                                                 ++iter)
                {
                       std::map<types::boundary_id, double> tmp = iter->second;

                       for(std::map<types::boundary_id, double>::const_iterator iter2  = tmp.begin();
                                                                                iter2 != tmp.end();
                                                                              ++iter2)
                       {
                              typename FunctionMap<DH::space_dimension>::type function_map;

                              const ZeroFunction<DH::space_dimension> zero_function(n_components);
                              function_map[boost::lexical_cast<unsigned int>(iter2->first)] = &zero_function;

                              std::vector<bool> component_mask(n_components, false);
                              component_mask[system_management.solution_name_to_index(iter->first)] = true;

                              VectorTools::interpolate_boundary_values( mapping,
                                                                        dof,
                                                                        function_map,
                                                                        dst,
                                                                        component_mask );
                       }
                }
  }

  /**
   * This function applies
   * zero boundary values to
   * the linear system of equations.
   */
  template<typename MATRIX, typename VECTOR, typename DH>
  void apply_zero_boundary_values_to_linear_system(MATRIX&                                              matrix,
                                                   VECTOR&                                              solution,
                                                   VECTOR&                                              rhs,
                                                   const Mapping<DH::dimension, DH::space_dimension>&   mapping,
                                                   const DH&                                            dof,
                                                   const FuelCell::SystemManagement&                    system_management,
                                                   const std::vector< component_boundaryID_value_map >& maps,
                                                   const bool&                                          repair_diagonal = false)
  {
         if( repair_diagonal )
         {
                SolverUtils solver_utils;
                solver_utils.repair_diagonal( matrix,
                                              solution,
                                              rhs );
         }

         std::map<unsigned int, double> boundary_values;
         FuelCell::InitialAndBoundaryData::make_zero_boundary_values<DH>( boundary_values,
                                                                          mapping,
                                                                          dof,
                                                                          system_management,
                                                                          maps );
         MatrixTools::apply_boundary_values( boundary_values,
                                             matrix,
                                             solution,
                                             rhs );
  }

/**
 * This class is a means
 * to make variable
 * initial or boundary data.
 *
 * \author Valentin N. Zingan, 2013
 */
template<int dim>
class InitialOrBoundaryDataBase : public Function<dim>
{
public:

///@name Service functions
//@{

  /**
   * This function calls
   * the \p math_expression() function of
   * this class
   * and has NOT to be overriden
   * in the derived classes.
   */
  virtual double value(const Point<dim>&  point,
                       const unsigned int no_component = 0) const;

//@}

protected:

///@name Constructors, destructor, and initialization
//@{

  /**
   * Constructor.
   *
   * - \p n_components is the total number of
   *                   the solution components.
   */
  InitialOrBoundaryDataBase(const unsigned int n_components = 1);

  /**
   * Destructor.
   */
 ~InitialOrBoundaryDataBase();

//@}

///@name Service functions
//@{

  /**
   * This function implements
   * the main functionality of
   * this class
   * and has to be overriden
   * in the derived classes.
   */
  virtual double math_expression(const Point<dim>&  point,
                                 const unsigned int no_component = 0) const;

//@}

///@name Minor functions
//@{

  /**
   * This function is used
   * to print out the name of another function
   * that has been declared in the scope of this class,
   * but not yet been implemented.
   */
  void print_caller_name(const std::string& caller_name) const;

//@}

};

} // InitialAndBoundaryData

} // FuelCell

#endif