//---------------------------------------------------------------------------
//    $Id: operating_conditions.h 1354 2013-08-17 00:01:22Z secanell $
//
//    Copyright (C) 2009 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#ifndef _FUELCELL_OPERATING_CONDITIONS__H
#define _FUELCELL_OPERATING_CONDITIONS__H

// Include deal.II classes
#include <base/parameter_handler.h>
#include <base/logstream.h>

#include "fcst_constants.h"

using namespace dealii;

namespace FuelCell
{
    
    /**
     * Class used to store, read from file and define the operating conditions for a fuel cell.
     * It stores the following:
     * - cell temperature
     * - cell voltage
     * - oxygen mole fraction at the inlet of the channel
     * - relative humidity at anode and cathode
     * - total pressure at anode and cathode
     * 
     * This information, in conjunction with the water saturation equation is used to compute the
     * mole fractions for each component in the mixture.
     * 
     * If the inlet is humidified air, please set the oxygen mole fraction to 0.21 (default).
     * 
     * In the input file, the following parameters can be specified (see declare_parameters ):
     * @code
     * subsection Fuel cell data
     * (...)
     *   subsection Operating conditions
     *     set Temperature cell = 353
     *     set Cathode pressure = 101325
     *     set Cathode initial oxygen mole fraction (prior to humidification) = 0.21
     *     set Cathode relative humidity = 0.7
     *     set Anode pressure = 101325
     *     set Anode relative humidity = 0.7
     *     set Voltage cell = 0.6
     *     set Voltage drop in the anode = 0.015
     *     set Open circuit voltage = 1.23
     *   end
     * end
     * @endcode 
     * 
     * <h3>Usage Details:</h3>
     * 
     * In order to use this class, first an object of the class needs to be created. Usually, one such objects exists in every application. To create the object,
     * include the .h file in the include application file and in the application data member section add the object. For example:
     * @code
     * #include "operating_conditions.h"
     * 
     * // Then in the data member declaration (usually a private member)
     * FuelCell::OperatingConditions OC;
     * @endcode
     * 
     * Once the object is created, the section where the input data will be specified in the input file needs to be delcared. To do so, in the declare_parameters section 
     * of your application call the following:
     * 
     * @code
     * //--------- IN DECLARE_PARAMETERS ------------------------------------------------------
     * template <int dim>
     * void 
     * NAME::AppCathode<dim>::declare_parameters(ParameterHandler& param)
     * {
     *   (...)
     *   OC.declare_parameters(param);
     *   (...)
     * }
     * @endcode
     *          
     * If you intent to run an optimization/parameteric study, then you might have some variables that you wish to modify during the simulation. This is
     * achieved in section set_parameters in the application file:
     * @code
     * //--------- IN SET_PARAMETERS ------------------------------------------------------
     * template <int dim>
     * void 
     * NAME::AppCathode<dim>::set_parameters(const std::vector<std::string>& name_dvar,
     *                                      const std::vector<double>& value_dvar,
     *                                      ParameterHandler& param)
     * {
     *   (...)
     *   // Set any parameters in the parameter file related to the layers.
     *    OC.set_parameters(name_dvar, value_dvar, param);  
     *   (...)
     * }
     * @endcode
     * 
     * Finally, once the input file has been read by our application, your class needs to be initialized. This is achieved using the function initialize()
     * @code
     * //--------- IN INITIALIZE ------------------------------------------------------
     * template <int dim>
     * void
     * NAME::AppCathode<dim>::_initialize(ParameterHandler& param)
     * {   
     *  (...) 
     *  OC.initialize(param);
     * }
     * @endcode
     * 
     * You are now ready to use your OperatingConditions object!.
     * 
     * 
     * @author M. Secanell, 2009-2013
     * 
     */
    class OperatingConditions
    {
    public:
        /**
         * Constructor
         */
        OperatingConditions();
        
        /**
         * Destructor
         */
        ~OperatingConditions();
        
        /**
         * Declare all necessary parameters in order to compute the coefficients
         * 
         * The parameters that can be specified in the input file are as follows:
         * 
         * @code
         * subsection Fuel cell data
         * (...)
         *   subsection Operating conditions
         *     set Temperature cell = 353
         *     set Cathode pressure = 101325
         *     set Cathode initial oxygen mole fraction (prior to humidification) = 0.21
         *     set Cathode relative humidity = 0.7
         *     set Anode pressure = 101325
         *     set Anode relative humidity = 0.7
         *     set Voltage cell = 0.6
         *     set Voltage drop in the anode = 0.015
         *     set Open circuit voltage = 1.23
         *   end
         * end
         * @endcode 
         */
        void declare_parameters (ParameterHandler &param) const;
        /**
         * This class is used in order to create an interface with DAKOTA. Each parameter that could be used for optimization
         * is assigned a design variable name that is related to the name on the parameter file. DAKOTA uses the design variable
         * name to request a new analysis with a different design variable.
         * This routine changes the ParameterHandler to take into account the changes on the design variable that DAKOTA requested
         * In this routine we create the following design variables (NOTE: Derivatives not implemented for this values):
         * - dV = "Voltage across MEA" in subsections Fuel cell data >> Operating conditions
         * - x_o2 = "Cathode oxygen molar fraction" in Fuel cell data >> Operating conditions
         * - x_wv = "Cathode water vapour molar fraction" in Fuel cell data >> Operating conditions
         */
        void set_parameters(const std::vector<std::string>& name_dvar,
                            const std::vector<double>& value_dvar,
                            ParameterHandler& param) const;
        /**
         * Class used to read in data and initialize the necessary data
         * to compute the coefficients.
         */
        void initialize (ParameterHandler& param);
        
        
        /**
         * Get the water vapour saturation pressure in atmospheres (atm) using cell temperature. 
         */
        double saturation_pressure();
        
        /**
         * Get the mole fraction of water at the cathode channel from pressure, temperature, and relative humidity
         *   note that the anode mole fraction of water is not required, as there are only two species
         *   so x_wv_a = 1 - x_h2 
         */
        double get_x_wv();
        
        /**
         * Get the mole fraction of oxygen at the cathode channel from pressure, temperature, and relative humidity
         */
        double get_x_o2();
        
        /**
         * Get the mole fraction of hydrogen at the anode channel from pressure, temperature, and relative humidity
         */       
        double get_x_h2();
        
        /**
         * NOTE: This function is redefined in base_kinetics class, considering variable temperature and gas pressures. The function here should only be used for defining Initial condition values.
         * Get the theoretical cell voltage using the cell temperature, and the reactant gas pressures
         */
        double voltage_cell_th();
        
        void print_operating_conditions();
        
        /**
         * Get the total gas concentration in the cathode. 
         * NOTE: This is a constant value. Use only if the total gas concentration
         * is assumed to be constant
         */
        inline double get_c_c()
        {	return c_c; }
        /** Return cell temperture as input in Operating Conditions subsection */
        inline double get_T()
        { return T_cell; }
        /** Return cell voltage as input in Operating Conditions subsection */
        inline double get_V()
        { return V_cell; }
        /** Return the voltage drop in the anode 
         * 
         * \note Can be used as boundary condition for anode model or initial condition
         * 		in a full MEA model
         */
        inline double get_dV_a()
        { return dV_a; }
        /** Return cathode pressure as input in Operating Conditions subsection */
        inline double get_pc_Pa()
        { return p_c; }
        /** Return cathode pressure as input in Operating Conditions subsection */
        inline double get_pc_atm()
        { return p_c/101325; }
        /** Get the total gas concentration in the anode. */
        inline double get_c_a()
        {	return c_a; }
        /** Return anode pressure as input in Operating Conditions subsection */
        inline double get_pa_Pa()
        { return p_a; }
        /** Return anode pressure as input in Operating Conditions subsection */
        inline double get_pa_atm()
        { return p_a/101325; }
        /** Return anode relative humidity as input in Operating Conditions subsection */
        inline double get_RH_a()
        {return RH_a;}
        /** Return cathode relative humidity as input in Operating Conditions subsection */
        inline double get_RH_c()
        {return RH_c;}
        /** Get the open circuit voltage for the cell */
        inline double get_OCV()
        {return OCV;}
        
    private:
        //------------ CONSTANTS -----------------------------------------
        double R; //Gas constant 8.3144 J / (mol K)
        
        /** Initial amount of oxygen in channel prior to humidification */
        double channel_oxygen_mole_fraction;
        
        //------------ CELL DATA -----------------------------------------
        /** Operating temperature of the cell */
        double T_cell;
        /** Operating voltage of the cell */
        double V_cell;
        /** Voltage drop in the anode */
        double dV_a;
        /** Open circuit voltage for the cell */
        double OCV;
        /** Theoretical voltage for the cell */
        double E_th;
        
        //------------- ANODE DATA -------------------------------------
        /** Pressure of the gas mixture in the anode B.C. */
        double p_a; 
        /** Concentration of the gas mixture in the anode B.C. */
        double c_a;
        /** Relative humidity of the gas mixture in the anode B.C */
        double RH_a;
        
        //------------- CATHODE DATA -------------------------------------
        /** Pressure of the gas mixture in the cathode B.C. */
        double p_c;
        /** Concentration of the gas mixture in the cathode B.C. */
        double c_c;
        /** Relative humidity of the gas mixture in the anode B.C */
        double RH_c;
        
    };
    
}

#endif // _FUELCELL_OPERATING_CONDITIONS__H
