//---------------------------------------------------------------------------
// C++ Interface: simulation_selector.h
//
// Description: Used to instantiate optimization routines
//
// Author: Peter Dobson <pdobson@ualberta.ca>, (C) 2011
//		University of Alberta
//
// Copyright: See COPYING file that comes with this distribution
//
//---------------------------------------------------------------------------

#ifndef _SIMULATION_TOOLS_H
#define _SIMULATION_TOOLS_H

#include <string>
#include <iostream>

#include <boost/shared_ptr.hpp>

#include "optimization_block_matrix_application.h"
#include "adaptive_refinement.h"

#include "newton_basic.h"
#include "newton_w_line_search.h"
#include "newton_w_3pp.h"
#include "newton_w_3ppC.h"

#include "app_cathode.h"
#include "app_pemfc.h"
#include "app_laplace.h"
#include "app_read_mesh.h"

//---------------------------------------------------------
/**
 * This class is used to select the different applications within FCST.
 *
 * @author P. Dobson, M. Secanell, Ali M. Koupaei, Valentin N. Zingan and Madhur Bhaiya
 *
 * $Header$
 */
template <int dim>
class SimulationSelector
{
public:

     /** Constructor */
     SimulationSelector();

     /** Destructor */
     ~SimulationSelector();

     /** Declare the necessary parameters on the handler object */
     void declare_parameters(ParameterHandler& param) const;

     /** Initialize the local variables from the handler object */
     void initialize(ParameterHandler& param);

     /**
      * Assign the application you would like to solve. All applications should be inherited from
      * AppFrame::OptimizationBlockMatrixApplication<dim>.
      *
      * If you develop a new application, you would assign the application an application name and add
      * the following to this section:
      * \code
       else if (name_application.compare("cathode") == 0)
       {
         deallog<<" YOU ARE CURRENTLY SOLVING A CATHODE MODEL                "<<std::endl;
         app_lin = new FuelCell::AppCathode<dim>;
       }
        \endcode
      */
     boost::shared_ptr<AppFrame::OptimizationBlockMatrixApplication<dim> > select_application();

     /**
      * For linear, non-linear and transient problems, select nothing, a non-linear or a transient solver that you would like to use
      */
     //boost::shared_ptr<AppFrame::newtonBase >	select_solver(AppFrame::OptimizationBlockMatrixApplication<dim>* app_lin);
     boost::shared_ptr<AppFrame::ApplicationCopy >	select_solver(AppFrame::OptimizationBlockMatrixApplication<dim>* app_lin);

     /**
      * Select the solution method that you would like to use, eg. Adaptive refinement
      */
     boost::shared_ptr<AppFrame::AdaptiveRefinement<dim> >
     select_solver_method(AppFrame::OptimizationBlockMatrixApplication<dim>* app_lin,
                          AppFrame::ApplicationCopy* newton_solver,
                          const AppFrame::FEVector& solution = AppFrame::FEVector());
     //select_solver_method(AppFrame::OptimizationBlockMatrixApplication<dim>* app_lin, AppFrame::newtonBase* newton_solver);


protected:
     /**
      * Variable storing the name of the application to be solved
      */
     std::string name_application;

     /**
      * Variable storing the name of the non-linear solver
      */
     std::string name_solver;

     /**
      * Variable storing the name of the solution method
      */
     std::string name_solve_method;
};

#endif
