//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2011-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: simulation_builder.cc
//    - Description:
//    - Developers: M. Secanell, P. Dobson, Valentin N. Zingan
//    - $Id: simulator_builder.h 1444 2013-08-30 21:03:06Z wardlawp $
//
//---------------------------------------------------------------------------

#ifndef SIMULATOR_BUILDER_H
#define SIMULATOR_BUILDER_H

// STL
#include <string>
#include <fstream>
#include <iostream>
#include <vector>
#include <ctime>
#include <stdlib.h>

// Boost
#include <boost/smart_ptr.hpp>

#include <base/timer.h>
#include <base/utilities.h>
#include <appframe/base.h>
#include <appframe/block_matrix_application.h>

// FCST
#include "optimization_block_matrix_application.h"
#include "geometries.h"
#include "geometry.h"
#include "simulation_selector.h"
#include "adaptive_refinement.h"
#include "tafel_kinetics.h"
#include "FCST_TEST_SUITE.h"

#include "dakota_interface.h"
// These files can only be used if DAKOTA is linked to our package:
#ifdef _WITH_DAKOTA
#include "dakota_direct_interface.h"
#include "dakota_application.h"
#endif

//---------------------------------------------------------

/**
 * This class is used to initialize and launch simulations or optimization routines.
 *
 * @author P. Dobson, M. Secanell, Ali M. Koupaei
 *
 * and the concept of sequential applications by
 *
 * @author Valentin N. Zingan
 *
 * $Header$
 */
/**
 * This class is used to read from the command line the input file and any other flags. Based on
 * the flags it will print instructions to screen or setup a simulation.
 *
 * This is the only class called in main and it manages the problem to be solved.
 */

template <int dim>
class SimulatorBuilder
{
public:

    /**
     * Constructor
     */
    SimulatorBuilder();

    /**
     * Destructor
     */
    ~SimulatorBuilder();

    /**
     * Message that is printed to screen if the program is run without any flags and input
     * files. The usage message explain the flabs avaiable and how they work
     */
    void usage(int exit_value);

    /**
     * Member function that is used to print information about the program. The info
     * in print logo appears everytime the program is executed.
     */
    void print_logo() const;

    /**
     * Print \p timer info.
     */
    void print_timer_info() const;

    /**
     * Member function used to read in the command line and initialize the necessary data
     */
    void parse_inputs(int argc, char *argv[]);

    /**
     * Member function used to call the declaration of the parameters, initialize data
     * and set the application and solver objects to those specified in the input file
     */
    void scan();

    /**
     * Run the simulation.
     * Determines whether optimization is being used and runs a single simulation
     * or optimization routines
     */
    virtual void run();

protected:

    /**
     * Declare all necessary parameters to read the input files
     * The following parameters are set here:
     * - "Dakota direct": Set to true if you would like to solve an optimization problem using Dakota. Transfer of information between the application and Dakota is via the class DakotaDirectInterface
     * - "simulator parameter file name": Specify the file that contains all the application (See the application that you would like to run, e.g. AppCathode)
     * - "optimization parameter file name" : Name of the file that specifies all the information related to the optimization class (See DakotaApplication for parameters)
     * - "Run tests" : Set to true if you would like to run a single function. This routine can be re-implemented to test different objects without running a full simulation
     * - "save transfer files": If set to false the mesh and solution from a parametric study will be deleted.
     * TO DO: Create filters for the input parameters
     */
    void declare_parameters(ParameterHandler& param) const;

    /**
     * Initialize the local variables declared by the parameter handler
     */
    void initialize(ParameterHandler& param);

    /**
     *
     */
    void open_logfile (const std::string&);

    /**
     * Set up the required optimization objects (i.e.
     * - Dakota::ParallelLibrary: Is where you specify what type of Message Passing Interface (MPI) you would like to run. E.g. Running in "serial mode" single processor, or in "parallel" on multiple processors.
     * - Dakota::ProblemDescDB: Stores optimization problem description and optimization strategy
     * - DakotaApplication: Manages inputs and sends data to application via DakotaDirectInterface (See DakotaApplication for more detials)
     * - DirectApplicInterface: Communicates Dakota parameters to application, sets up application (manages input from Dakota to Application),
     * runs application and results from the Application to Dakota.
     * It must be a child of Dakota::DirectApplicInterface
     *
     *  http://dakota.sandia.gov/docs/dakota/5.2/html-dev/DakLibrary.html
     *
     * \todo3 Print results to file from direct interface?
     */
    void run_optimization();

    /**
     * This routine is mainly used for testing single member functions in classes.
     * It will run in isolation from the main simulation framework, so that you are not
     * solving a finite element problem.
     * */
    void run_test();

    /**
     * Returns a random number in val argument. The number will be between min and max.
     */
    inline void random_number(double &val, double min, double max)
    {
        double rand_num = drand48();
        val = min + rand_num*(max - min);
    }

    /**
     * Variable set in parse_inputs which determines whether the program is
     * being called from DAKOTA
     */
    bool dakota_use;

    /**
     * Decision varible for using the direct dakota interface
     */
    bool dakota_direct;

    /**
     * Decision varible for running tests.
     */
    bool run_tests;

    /** Stores the name of the input file with the application selector data*/
    std::string input_file;

    /** Stores the name of the log file*/
    boost::shared_ptr<std::ofstream> log_file;

    /** Stores the name of the parameter file containing the physical data
     * for the simulation.  I.e. Fuel cell data.
     */
    std::string simulator_parameter_file_name;

    /**
     * This vector stores the names of the parameter files
     * containing the physical data for the same simulation.
     *
     * Sometimes it might happen that you need to go
     * through the sequential list of the parameter files
     * containing the physical data for the same simulation.
     *
     * The final solution for the previous run
     * will form the initial guess for the next one.
     *
     * For example, some fluid flow applications lack the
     * good initial guess which heavily influences the iterative
     * process itself. More often it occurs for high \p Re numbers.
     *
     * What we do in this case is:
     *
     * - we form the parameter file for some fluid flow app
     *   that includes the relatively high dynamic viscosity
     *   and therefore relatively low \p Re number,
     * - we use, say, zero initial guess for all solution
     *   variables like pressure and velocity,
     * - we use the final solution for this problem as the
     *   initial guess for the next problem whose parameter
     *   file will have lower dynamic viscosity and therefore
     *   higher \p Re number,
     * - we repeat till we reach the desirable dynamic
     *   viscosity (\p Re number) of the fluid for which
     *   the simulation itself was designed.
     *
     * Other examples are a steep gradient of membrane water
     * content, high overpotential, and other.
     *
     * In other words, it may be extremely useful for nonlinear problems
     * to go through the list of conditions from light or moderate
     * towards heavy smartly forming initial guesses for each of those
     * conditions.
     */
    std::vector<std::string> simulator_parameter_file_names;

    /** Stores the name of the parameter file containing the optimization data
     */
    std::string optimization_parameter_file_name;

    /**
     * Variable set in parse_inputs which stores the name of the results (responses) file
     * if the simulation is being called from DAKOTA
     */
    std::string dakota_results;

    /**
     * Variable set in parse_inputs which stores the name of the parameter file
     * if the simulation is being called from DAKOTA
     */
    std::string dakota_parameters;

    /** Local varible for the program name */
    std::string program_name;

    /** Local variable for the program version */
    std::string program_version;

    /** Set to true if you want to keep the mesh and solution from the simulation. */
    bool save_transfer_files;

    /**
     * Parameter handler object that will be used to store all input
     * data for the application
     */
    ParameterHandler param;

    /** Pointer where linear application is stored */
    boost::shared_ptr <AppFrame::OptimizationBlockMatrixApplication<dim> >app_lin;

    /** Pointer to the non-linear solver */
    //boost::shared_ptr <AppFrame::newtonBase> newton;
    boost::shared_ptr <AppFrame::ApplicationCopy> newton;

    /**
     * Pointer to a solver application which applies adaptive refinement to the grid
     */
    boost::shared_ptr <AppFrame::AdaptiveRefinement<dim> > solver;

    /**
     * Object which stores the name of the application and solver
     * Allows the user to select the application and set objects to the empty pointers
     */
    boost::shared_ptr< SimulationSelector<dim> > sim_selector;

    /** Object used to calculate the CPU and Run time for the simulation */
    Timer timer;
};

#endif
