// ----------------------------------------------------------------------------
//
// FCST: Fuel Cell Simulation Toolbox
//
// Copyright (C) 2006-2013 by Energy Systems Design Laboratory, University of Alberta
//
// This software is distributed under the MIT License
// For more information, see the README file in /doc/LICENSE
//
// - Class: system_management.h
// - Description: This class manages systems of equations
// - Developers: Valentin N. Zingan,    University of Alberta
//                       Marc Secanell Gallart, University of Alberta
//                       Madhur Bhaiya, University of Alberta
// - Id: $Id: system_management.h 1449 2013-08-30 21:18:58Z madhur $
//
// ----------------------------------------------------------------------------

#ifndef _FCST_FUELCELL_SYSTEM_MANAGEMENT_H_
#define _FCST_FUELCELL_SYSTEM_MANAGEMENT_H_

#include "optimization_block_matrix_application.h"

using namespace dealii;
using namespace AppFrame;

/**
 * The typedef for the map of
 * cell or flux (DG FEM only)
 * couplings stored in the
 * actual equation classes.
 *
 * The first std::string means
 * the name of available FCST equation.
 *
 * The second std::string means
 * the name of available FCST solution variable.
 */
typedef std::map<   std::string , std::map<std::string, DoFTools::Coupling>   > couplings_map;

/**
 * The enumeration containing
 * the names of some of the available FCST solution variables
 * and their derivatives to be used in
 * kinetics, materials, and other related classes.
 *
 * The enumeration is usually used in conjunction with FuelCellShop::SolutionVariable class.
 * For example, if the only available FCST solution variable is \p electronic_electrical_potential
 * then the respective piece of code will look like:
 *
 * @code
 * FuelCellShop::SolutionVariable solution;
 * (...)
 * if( solution.get_variablename() == VariableNames::electronic_electrical_potential )
 * {
 * (...)
 * }
 * @endcode
 *
 * In this case, if \p solution is initialized such that it contains \p electronic_electrical_potential
 * then the \p if statement will return \p TRUE.
 */
enum VariableNames
{
  nothing = 0,

  oxygen_molar_fraction,
  water_molar_fraction,
  hydrogen_molar_fraction,
  nitrogen_molar_fraction,
  protonic_electrical_potential,
  electronic_electrical_potential,
  membrane_water_content,
  temperature_of_solid_phase,
  temperature_of_fluid_phase,
  temperature_of_REV,

  oxygen_concentration,
  water_concentration,
  hydrogen_concentration,
  nitrogen_concentration,
  proton_concentration,

  total_pressure
};

namespace FuelCell
{

/**
 * IMPORTANT: Add all new solution variables
 * and equations here !
 *
 * This class contains
 * all the information on available FCST solution variables
 * and equations and feeds the actual equation classes with
 * different types of information on the linear system
 * structure.
 *
 * Available FCST solution variables:
 *
 * -  GROUP 1
 *
 * - "oxygen_molar_fraction"           \f$ \left( x_{O_2}          \right) \f$
 * - "water_molar_fraction"            \f$ \left( x_{H_2O}         \right) \f$
 * - "hydrogen_molar_fraction"         \f$ \left( x_{H_2}          \right) \f$
 * - "nitrogen_molar_fraction"         \f$ \left( x_{N_2}          \right) \f$
 * - "protonic_electrical_potential"   \f$ \left( \Phi_m           \right) \f$
 * - "electronic_electrical_potential" \f$ \left( \Phi_s           \right) \f$
 * - "membrane_water_content"          \f$ \left( \lambda          \right) \f$
 *
 * Available FCST equations:
 *
 * -  GROUP 1
 *
 * - "Ficks Transport Equation - oxygen"
 * - "Ficks Transport Equation - water"
 * - "Ficks Transport Equation - hydrogen"
 * - "Ficks Transport Equation - nitrogen"
 * - "Proton Transport Equation"
 * - "Electron Transport Equation"
 * - "Membrane Water Content Transport Equation"
 *
 * \author Valentin N. Zingan, 2012-13
 * \author Marc Secanell Gallart, 2012-13
 */

class SystemManagement : public Subscriptor
{
public:

///@name Constructors, destructor, and initialization
//@{

  /**
   * Constructor assembling
   *
   * - \p all_solution_names,
   * - \p all_equation_names.
   */
  SystemManagement();

  /**
   * Constructor assembling
   *
   * - \p all_solution_names,
   * - \p all_equation_names,
   * - \p block_info     (AppFrame structure) will be initialized by the AppFrame itself,
   * - \p cell_couplings (AppFrame structure) will be initialized by the \p make_cell_couplings function of this class called by a user defined application,
   * - \p flux_couplings (AppFrame structure) will be initialized by the \p make_flux_couplings function of this class called by a user defined application (DG FEM only).
   */
  SystemManagement(MeshWorker::BlockInfo&        block_info,
                   Table<2, DoFTools::Coupling>& cell_couplings,
                   Table<2, DoFTools::Coupling>& flux_couplings);

  /**
   * Destructor.
   */
 ~SystemManagement();

  /**
   * This function assembles
   *
   * - \p block_info     (AppFrame structure) will be initialized by the AppFrame itself,
   * - \p cell_couplings (AppFrame structure) will be initialized by the \p make_cell_couplings function of this class called by a user defined application,
   * - \p flux_couplings (AppFrame structure) will be initialized by the \p make_flux_couplings function of this class called by a user defined application (DG FEM only)
   *
   * if the empty constructor was used.
   */
  void initialize(MeshWorker::BlockInfo&        rblock_info,
                  Table<2, DoFTools::Coupling>& rcell_couplings,
                  Table<2, DoFTools::Coupling>& rflux_couplings)
  {
    block_info     = &rblock_info;
    cell_couplings = &rcell_couplings;
    flux_couplings = &rflux_couplings;
  }

  /**
   * Declare parameters.
   */
  void declare_parameters(ParameterHandler& param) const;

  /**
   * Initialize parameters.
   */
  void initialize(ParameterHandler& param);

//@}

///@name Service functions
//@{

  /**
   * This function returns
   * a boolean stating
   * whether an available FCST solution variable
   * with name \p name stored in \p solution_names.
   */
  const bool solution_in_userlist(const std::string& name) const;

  /**
   * This function returns
   * the index of
   * a user defined solution variable
   * with name \p name stored in \p solution_names.
   */
  const unsigned int solution_name_to_index(const std::string& name) const;

  /**
   * This function returns
   * the index of
   * a user defined equation
   * with name \p name stored in \p equation_names.
   */
  const unsigned int equation_name_to_index(const std::string& name) const;

  /**
   * This function returns
   * the index of
   * a system matrix block
   * based on
   *
   * - \p equation_name (line),
   * - \p solution_name (column),
   * - \p cell_couplings.
   *
   * \note A system matrix block
   * with
   * \p cell_couplings(\p equation_name, \p solution_name) = DoFTools::none
   * is missed.
   */
  const unsigned int matrix_block_index(const std::string& equation_name,
                                        const std::string& solution_name) const;

  /**
   * This function fills out the
   * \p cell_couplings object
   * based on the information
   * drawn from the actual equation classes
   * in use.
   */
  void make_cell_couplings(const std::vector<couplings_map>& src);

  /**
   * This function fills out the
   * \p flux_couplings (DG FEM only) object
   * based on the information
   * drawn from the actual equation classes
   * in use.
   */
  void make_flux_couplings(const std::vector<couplings_map>& src);

//@}

///@name Accessors and info
//@{

  /**
   * This function returns
   * \p all_solution_names.
   */
  const std::vector<std::string>& get_all_solution_names() const
  {
    return all_solution_names;
  }

  /**
   * This function returns
   * \p all_equation_names.
   */
  const std::vector<std::string>& get_all_equation_names() const
  {
    return all_equation_names;
  }

  /**
   * This function returns
   * \p solution_names.
   */
  const std::vector<std::string>& get_solution_names() const
  {
    return solution_names;
  }

  /**
   * This function returns
   * \p equation_names.
   */
  const std::vector<std::string>& get_equation_names() const
  {
    return equation_names;
  }

  /**
   * This function returns
   * \p n_solution_names.
   */
  const unsigned int& get_number_of_solution_names() const
  {
    return n_solution_names;
  }

  /**
   * This function prints out
   * the information on
   * available FCST solution variables
   * and equations
   * and what you have picked out.
   *
   * It also displays
   * the basic linear
   * system structure.
   */
  void print_system_info() const;

//@}

///@name Exceptions
//@{

  /**
   * Exception thrown when
   * a user defined solution variable
   * is not found among available FCST solution variables.
   */
  DeclException2(VariableNotFoundInFCSTVariables,
                 std::string,
                 std::string,
                 << "A "  << arg1 << " with name \"" << arg2 << "\" is not stored in available FCST solution variables");

  /**
   * Exception thrown when
   * a user defined equation
   * is not found among available FCST equations.
   */
  DeclException2(EquationNotFoundInFCSTEquations,
                 std::string,
                 std::string,
                 << "An " << arg1 << " with name \"" << arg2 << "\" is not stored in available FCST equations");

  /**
   * Exception thrown when
   * a \p name in \p solution_name_to_index function
   * is not found among user defined solution variables.
   */
  DeclException2(VariableNotFoundInUserVariables,
                 std::string,
                 std::string,
                 << "A "  << arg1 << " with name \"" << arg2 << "\" is not stored in user defined solution variables");

  /**
   * Exception thrown when
   * a \p name in \p equation_name_to_index function
   * is not found among user defined equations.
   */
  DeclException2(EquationNotFoundInUserEquations,
                 std::string,
                 std::string,
                 << "An " << arg1 << " with name \"" << arg2 << "\" is not stored in user defined equations");

  /**
   * Exception thrown when
   * a requested system matrix block
   * does not exist.
   *
   * \note Such a block has
   * \p cell_couplings(\p equation_name, \p solution_name) = DoFTools::none.
   */
  DeclException5(SystemMatrixBlockDoesNotExist,
                 std::string,
                 unsigned int,
                 unsigned int,
                 std::string,
                 std::string,
                 << "A "  << arg1 << " (" << arg2 << "," << arg3 << ") " << "does not exist, because the variable \"" << arg4 << "\" is not coupled with the equation \"" << arg5 << "\"");

//@}

  //////////
  // DATA //
  //////////

///@name External data pointers
//@{

  /**
   * Pointer to the external
   * YourApplication<dim>::block_info object.
   * Direct access is provided.
   */
  MeshWorker::BlockInfo* block_info;

  /**
   * Pointer to the external
   * YourApplication<dim>::cell_couplings object.
   * Direct access is provided.
   */
  Table<2, DoFTools::Coupling>* cell_couplings;

  /**
   * Pointer to the external
   * YourApplication<dim>::flux_couplings (DG FEM only) object.
   * Direct access is provided.
   */
  Table<2, DoFTools::Coupling>* flux_couplings;

//@}

protected:

///@name Other functions
//@{

  /**
   * This function fills out \p all_solution_names.
   *
   * If you have a name of a new FCST solution variable, add it here.
   */
  void set_all_solution_names();

  /**
   * This function fills out \p all_equation_names.
   *
   * If you have a name of a new FCST equation, add it here.
   */
  void set_all_equation_names();

  /**
   * This function throws an exception if
   * a user defined solution variable
   * is not found among available FCST solution variables.
   */
  void check_solution_names() const;

  /**
   * This function throws an exception if
   * a user defined equation
   * is not found among available FCST equations.
   */
  void check_equation_names() const;

//@}

  //////////
  // DATA //
  //////////

///@name System properties
//@{

  /**
   * Vector storing the names of available FCST solution variables.
   */
  std::vector<std::string> all_solution_names;

  /**
   * Vector storing the names of available FCST equations.
   */
  std::vector<std::string> all_equation_names;

  /**
   * Vector storing the names of user defined solution variables.
   */
  std::vector<std::string> solution_names;

  /**
   * Vector storing the names of user defined equations.
   */
  std::vector<std::string> equation_names;

  /**
   * The number of user defined solution variables.
   * The number of user defined equations
   * is implicitly assumed to be the same.
   */
  unsigned int n_solution_names;

//@}

};

} // FuelCell

#endif