//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: app_cathode.cc
//    - Description: Application to solve a cathode catalyst layer including CL and GDL.
//    - Developers: M. Secanell
//    - $Id: app_cathode.cc 1422 2013-08-30 16:53:57Z secanell $
//
//---------------------------------------------------------------------------

#include "app_cathode.h"

namespace NAME = FuelCell::Application;

using namespace AppFrame;

//---------------------------------------------------------------------------
template <int dim>
NAME::AppCathode<dim>::AppCathode(boost::shared_ptr<AppFrame::ApplicationData> data)
:
OptimizationBlockMatrixApplication<dim>(data),
system_management(this->block_info, this->cell_couplings, this->flux_couplings),
ficks_transport_equation(system_management,&oxygen, &nitrogen),
electron_transport_equation(system_management),
proton_transport_equation(system_management),
reaction_source(system_management)
{
    deallog << "->FuelCell::Application::AppCathode_test-" << dim <<"d"<<std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
NAME::AppCathode<dim>::~AppCathode()
{

}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::declare_parameters(ParameterHandler& param)
{
    OptimizationBlockMatrixApplication<dim>::declare_parameters(param);

    // Declare parameters in system management:
    system_management.declare_parameters(param);

    // Declare grid generator parameters:
    FuelCellShop::Geometry::GridBase<dim>::declare_GridGenerator_parameters(param);

    // Declare parameters for operating conditions
    OC.declare_parameters(param);

    // Declare parameters for layer classes:
    FuelCellShop::Layer::GasDiffusionLayer<dim>::declare_GasDiffusionLayer_parameters("Cathode gas diffusion layer", param);
    FuelCellShop::Layer::MicroPorousLayer<dim>::declare_MicroPorousLayer_parameters("Cathode microporous layer", param);
    FuelCellShop::Layer::CatalystLayer<dim>::declare_CatalystLayer_parameters("Cathode catalyst layer", param);

    // Declare parameters for physics classes:
    electron_transport_equation.declare_parameters(param);
    proton_transport_equation.declare_parameters(param);
    ficks_transport_equation.declare_parameters(param);
    reaction_source.declare_parameters(param);

}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::set_parameters(const std::vector<std::string>& name_dvar,
                                      const std::vector<double>& value_dvar,
                                      ParameterHandler& param)
{
    // Set any parameters in the parameter file related to operating conditions and mesh
    OC.set_parameters(name_dvar, value_dvar, param);
    FuelCellShop::Geometry::GridBase<dim>::set_GridGenerator_parameters (name_dvar, value_dvar, param);

    // Set any parameters in the parameter file related to the layers.
    FuelCellShop::Layer::GasDiffusionLayer<dim>::set_GasDiffusionLayer_parameters(name_dvar, value_dvar, "Cathode gas diffusion layer", param);
    FuelCellShop::Layer::MicroPorousLayer<dim>::set_MicroPorousLayer_parameters(name_dvar, value_dvar, "Cathode microporous layer", param);
    FuelCellShop::Layer::CatalystLayer<dim>::set_CatalystLayer_parameters(name_dvar, value_dvar, "Cathode catalyst layer", param);

    //Store names and values of the design variables
    design_var = name_dvar;
    design_var_value = value_dvar;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::_initialize(ParameterHandler& param)
{
    // NOT NEEDED:
    component_names.clear();
    component_names.push_back("Oxygen molar fraction");
    component_names.push_back("Protonic electrical potential");
    component_names.push_back("Electronic electrical potential");

    // Initialize parameters in system management:
    system_management.initialize(param);
    
    // Make sure that the number of input finite elements matches with the problem:
    Assert (this->element->n_blocks() == system_management.get_number_of_solution_names(),
              ExcDimensionMismatch(this->element->n_blocks(), system_management.get_number_of_solution_names()));

    // Initialize coefficients
    OC.initialize(param);
    
    // Generate grid and refine
    this->initialize_grid(param);

    // Initialize gases and material classes:
    std::vector< FuelCellShop::Material::PureGas * > gases;
    gases.push_back(&oxygen);
    gases.push_back(&nitrogen);

    // Initialize layer classes:
    CGDL = FuelCellShop::Layer::GasDiffusionLayer<dim>::create_GasDiffusionLayer("Cathode gas diffusion layer",param);
    CGDL->set_gases_and_compute(gases, OC.get_pc_atm (), OC.get_T());

    CMPL = FuelCellShop::Layer::MicroPorousLayer<dim>::create_MicroPorousLayer("Cathode microporous layer",param);
    CMPL->set_gases_and_compute(gases, OC.get_pc_atm (), OC.get_T());

    CCL = FuelCellShop::Layer::CatalystLayer<dim>::create_CatalystLayer("Cathode catalyst layer", param);
    CCL->set_gases_and_compute (gases, OC.get_pc_atm (), OC.get_T());

    // Initialise the necessary kinetics parameters in CCL.
    std::string name = "ORR";
    CCL->set_reaction_kinetics(name);
    CCL->set_constant_solution(OC.get_pc_Pa(),  VariableNames::total_pressure);   //Note: Pressure and temp. can only be set after setting kinetics.
    CCL->set_constant_solution(OC.get_T(),  VariableNames::temperature_of_REV);

    // Setting kinetics in the reaction source terms object.
    reaction_source.set_cathode_kinetics(CCL->get_kinetics());

    // Initialize parameters for physics classes:
    electron_transport_equation.initialize(param);
    proton_transport_equation.initialize(param);
    ficks_transport_equation.initialize(param);
    reaction_source.initialize(param);

    // Initialize system of equations based on the physical phenomena being solved for:
    std::vector<couplings_map> tmp;
    tmp.push_back(ficks_transport_equation.get_internal_cell_couplings() );
    tmp.push_back(proton_transport_equation.get_internal_cell_couplings() );
    tmp.push_back(electron_transport_equation.get_internal_cell_couplings() );
    reaction_source.adjust_internal_cell_couplings(tmp);
    system_management.make_cell_couplings(tmp);

    // Initialize dof per cell for the system of equations and information on
    //block sizes and indices
    this->remesh_dofs();

    // Initialize matrices and spartisity pattern for the whole system
    this->remesh_matrices();

    // Some info for debugging purposes:
    //system_management.print_system_info ();
    CCL->print_layer_properties ();
    //abort();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::initialize(ParameterHandler& param)
{
    OptimizationBlockMatrixApplication<dim>::initialize(param);
    _initialize(param);
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::initialize_grid(ParameterHandler& param)
{

    grid = FuelCellShop::Geometry::GridBase<dim>::create_GridGenerator (param);
    // Make sure that grid is appropriate for application
    AssertThrow (( (grid->get_mesh_type()).compare("Cathode") == 0 || (grid->get_mesh_type()).compare("CathodeMPL") == 0 
                   || (grid->get_mesh_type()).compare("GridExternal") == 0 || (grid->get_mesh_type()).compare("GridTest") == 0),
             ExcMessage("Only GridExternal, Cathode and CathodeMPL meshes are permitted for AppCathode. Check entry Type of mesh in subsection Grid generation in the parameter file"));

    // Check to see whether we are reading the grid from a file
    // Must read grid if we are reading a stored solution
    bool read_grid_from_file = grid->read_from_file;

    // If not reading from a grid from a previous solution,
    //	check whether we can use one from external source
    if (this->read_in_initial_solution || read_grid_from_file)
    {
        // TODO: Read grid_file name and format from parameter handler
        std::ifstream grid_file("._mesh_to_be_transfered.msh");
        if(!grid_file.fail())
        {
            deallog << "Reading grid from file" << std::endl;
            GridIn<dim> grid_in;
            grid_in.attach_triangulation(*this->tr);
            grid_in.read_msh(grid_file);
            grid_file.close();
            read_grid_from_file = true;
        }
        else
        {
            // if grid fails to read, generate standard grid, don't read initial solution
            deallog << "Grid file not available for reading" << std::endl;
            read_grid_from_file = false;
            this->read_in_initial_solution = false;
        }
    }
    if(!read_grid_from_file)
    {
        deallog << "Generating new grid" << std::endl;
        grid->generate_grid(*this->tr);
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::init_solution(AppFrame::FEVector& dst)
{
    bool solution_read_success = false;
    bool use_transfer_solution = false;

    // resize vector:
    dst.reinit(this->block_info.global);

    // Read initial solution from file if read_in_initial_solution set to true:
    if (this->read_in_initial_solution)
    {
        this->read_init_solution(dst,solution_read_success);
    }

    // If the initial solution is read in properly,
    //  check to see that the initial solution (i.e. Boundary Conditions)
    //  is within reasonable range for this simulation
    if (solution_read_success)
    {
        parse_init_solution(dst, OC, use_transfer_solution);
    }

    // If we either did not read an initial solution or the initial solution was not read correcty
    // generate an initial solution
    if (!use_transfer_solution)
    {
        // Obtain the oxygen concentration in the channel and the
        // cell potential:
        double x_o2   = OC.get_x_o2();
        double V_cell = - OC.get_V();

        if(V_cell < 0.0)
            V_cell = - 1.0 * V_cell;

        // Resize the solution vector:
        dst.reinit(this->block_info.global);

        // Create the function map for each variable and the component mask
        std::map<types::material_id, const Function<dim>* > function_map;
        std::vector<bool> component_mask(3, false);

        // --- xo2 ---
        const ConstantFunction<dim> f_xo2_GDL(x_o2, 3);
        const ConstantFunction<dim> f_xo2_MPL(x_o2/2.0, 3);
        const ConstantFunction<dim> f_xo2_CL(x_o2/4.0, 3);
        function_map[CGDL->get_material_id()] = &f_xo2_GDL;
        function_map[CMPL->get_material_id()] = &f_xo2_MPL;
        function_map[CCL->get_material_id()] = &f_xo2_CL;

        std::vector< std::string > solution_names( system_management.get_solution_names () );

        for (unsigned int i = 0; i<solution_names.size(); i++)
        {
            if (solution_names[i] == "oxygen_molar_fraction")
                component_mask[i] = true;
            else
                component_mask[i] = false;
        }

        VectorTools::interpolate( dst,
                                  *this->mapping,
                                  *this->dof,
                                  function_map,
                                  component_mask );

        // --- fp ---
        const ConstantFunction<dim> f_fp(0.0, 3);
        function_map[CGDL->get_material_id()] = &f_fp;
        function_map[CMPL->get_material_id()] = &f_fp;
        function_map[CCL->get_material_id()] = &f_fp;

        for (unsigned int i = 0; i<solution_names.size(); i++)
        {
            if (solution_names[i] == "protonic_electrical_potential")
                component_mask[i] = true;
            else
                component_mask[i] = false;
        }

        VectorTools::interpolate( dst,
                                  *this->mapping,
                                  *this->dof,
                                  function_map,
                                  component_mask );

        // --- fe ---

        const ConstantFunction<dim> f_fe(V_cell, 3);

        function_map[CGDL->get_material_id()] = &f_fe;
        function_map[CMPL->get_material_id()] = &f_fe;
        function_map[CCL->get_material_id()] = &f_fe;

        for (unsigned int i = 0; i<solution_names.size(); i++)
        {
            if (solution_names[i] == "electronic_electrical_potential")
                component_mask[i] = true;
            else
                component_mask[i] = false;
        }

        VectorTools::interpolate( dst,
                                  *this->mapping,
                                  *this->dof,
                                  function_map,
                                  component_mask );
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::cell_matrix(MatrixVector& cell_matrices,
                                   const typename DoFApplication<dim>::CellInfo& info)
{

    if (CGDL->belongs_to_material(info.cell->material_id()))
    {
        ficks_transport_equation.assemble_cell_matrix(cell_matrices, info, CGDL.get());
        electron_transport_equation.assemble_cell_matrix(cell_matrices, info, CGDL.get());
    }
    else if (CMPL->belongs_to_material(info.cell->material_id()))
    {
        ficks_transport_equation.assemble_cell_matrix(cell_matrices, info, CMPL.get());
        electron_transport_equation.assemble_cell_matrix(cell_matrices, info, CMPL.get());
    }
    else if (CCL->belongs_to_material(info.cell->material_id()))
    {
        ficks_transport_equation.assemble_cell_matrix(cell_matrices, info, CCL.get());
        electron_transport_equation.assemble_cell_matrix(cell_matrices, info, CCL.get());
        proton_transport_equation.assemble_cell_matrix(cell_matrices, info, CCL.get());
        reaction_source.assemble_cell_matrix(cell_matrices, info, CCL.get());
    }
    else
    {
        deallog<<"Material id: "    <<info.cell->material_id()<<" does not correspond to any layer"<<std::endl;
    }

}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::bdry_matrix(MatrixVector& bdry_matrices,
                                   const typename DoFApplication<dim>::FaceInfo& bdry)
{}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::cell_residual(AppFrame::FEVector& cell_vector,
                                     const typename DoFApplication<dim>::CellInfo& info)
{
    // -- Assertion before starting routine:
    // Make sure vectors are the right size
    Assert (cell_vector.n_blocks() == this->element->n_blocks(),
            ExcDimensionMismatch (cell_vector.n_blocks(), this->element->n_blocks()));


    if (CGDL->belongs_to_material(info.cell->material_id()))
    {
        ficks_transport_equation.assemble_cell_residual(cell_vector, info, CGDL.get());
        electron_transport_equation.assemble_cell_residual(cell_vector, info, CGDL.get());
    }
    else  if (CMPL->belongs_to_material(info.cell->material_id()))
    {
        ficks_transport_equation.assemble_cell_residual(cell_vector, info, CMPL.get());
        electron_transport_equation.assemble_cell_residual(cell_vector, info, CMPL.get());
    }
    else if (CCL->belongs_to_material(info.cell->material_id()))
    {
        ficks_transport_equation.assemble_cell_residual(cell_vector, info, CCL.get());
        electron_transport_equation.assemble_cell_residual(cell_vector, info, CCL.get());
        proton_transport_equation.assemble_cell_residual(cell_vector, info, CCL.get());
        reaction_source.assemble_cell_residual(cell_vector, info, CCL.get());
    }
    else
    {
        deallog<<"Material id: "    <<info.cell->material_id()<<" does not correspond to any layer"<<std::endl;
    }

}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::bdry_residual(AppFrame::FEVector& bdry_vector,
                                     const typename DoFApplication<dim>::FaceInfo& bdry)
{}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::dirichlet_bc(std::map<unsigned int, double>& boundary_values) const
{
    std::vector<bool> comp_mask (this->element->n_blocks(), false);
    unsigned int i;

    // -- boundary component 1: membrane/catalyst layer interface

    for (i = 0; i < component_names.size(); ++i)
    {
        if (component_names[i] == "Protonic electrical potential")
        {
            comp_mask[i] = true;
            break;
        }
    }
    VectorTools::interpolate_boundary_values (*this->dof,
                                              grid->get_boundary_id("c_CL/Membrane"),
                                              ZeroFunction<dim>(this->element->n_blocks()),
                                              boundary_values,
                                              comp_mask);
    // -- boundary component 2: area underneath ribs
    comp_mask[i] = false;
    if (this->boundary_fluxes == false)
    {
        for (i = 0; i < component_names.size(); ++i)
        {
            if (component_names[i] == "Electronic electrical potential")
            {
                comp_mask[i] = true;
                break;
            }
        }
        VectorTools::interpolate_boundary_values (*this->dof,
                                                  grid->get_boundary_id("c_BPP/GDL"),
                                                  ZeroFunction<dim>(this->element->n_blocks()),
                                                  boundary_values,
                                                  comp_mask);
    }

    // -- boundary component 3: area underneath gas channel
    // No Dirichlet boundary conditions for phases but we have Dirichlet BC for gas composition. Since for the linear
    //system we solve for delta_u, the dirichlet BC for delta_u are to have the value to be zero

    comp_mask[i] = false;

    for (i = 0; i < component_names.size(); ++i)
    {
        if (component_names[i] == "Oxygen molar fraction")
        {
            comp_mask[i] = true;
            break;
        }
    }
    VectorTools::interpolate_boundary_values (*this->dof,
                                              grid->get_boundary_id("c_Ch/GDL"),
                                              ZeroFunction<dim>(this->element->n_blocks()),
                                              boundary_values,
                                              comp_mask);
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::solve(AppFrame::FEVector& start,
                             const AppFrame::FEVectors& rhs)
{  
    
    // Assemble global matrix if necessary:
    bool assemble_numerically = false;
    
    if (this->notifications.any())//if (this->assemble_reasons)
    {
        AppFrame::FEVectors sol;
        unsigned int ind = rhs.find_vector("Newton iterate");
        sol.add_vector(rhs.vector(ind),"Newton iterate");
        if (assemble_numerically)
        {
            this->assemble_numerically(rhs);
        }
        else
        {
            this->assemble(sol); //Note the second component of
        }            
        this->notifications.clear();//this->clear_assemble();
    }
   
    // Eliminate zeros in the diagonal
    SolverUtils diagonal;
    unsigned int index_Newton_iterate = rhs.find_vector("Newton iterate");
    diagonal.repair_diagonal(this->matrix, start, rhs.vector(index_Newton_iterate));

    // Copy RHS to solution vector
    unsigned int index_Newton_residual = rhs.find_vector("Newton residual");
    start = rhs.vector(index_Newton_residual);

    // Make sure that the solution is of the same size as the RHS
    Assert(start.size() == rhs.vector(index_Newton_iterate).size(),
           ExcDimensionMismatch(start.size(), rhs.vector(index_Newton_iterate).size()));

    std::string solver_name ("UMFPACK"); //("UMFPACK") | ("ILU-GMRES");

    //*********** GMRES with ILU preconditioner *****
    if (solver_name.compare("ILU-GMRES") == 0) //Could also use if (strcmp (solver_name,"ILU-GMRES") == 0)
    {
           LinearSolvers::ILUPreconditioner prec(this->matrix);

           LinearSolvers::GMRESSolver solver;
           solver.solve( this->filtered_matrix,
                         start,
                         rhs.vector(index_Newton_residual),
                         10000,
                         1e-12,
                         prec.preconditioner );
    }
    //*********** UMFPACK *****
    else if (solver_name.compare("UMFPACK") == 0)
    {
        // Apply BC to matrix, rhs and sol
        std::map<unsigned int, double> boundary_values;
        this->dirichlet_bc(boundary_values);

        MatrixTools::apply_boundary_values(boundary_values,
                                           this->matrix,
                                           start,
                                           start);

        LinearSolvers::SparseDirectUMFPACKSolver solver;
        solver.solve(this->matrix,
                     start,
                     start);
    }
    else
    {
        const std::type_info& info = typeid(*this);
        deallog<<"Solver not implemented in Class "<<info.name()<<". Member function solve()"<<std::endl;
        abort();
    }

    //////////////////////////////////////////////////////////////////////
    // Distribute solution:
    //////////////////////////////////////////////////////////////////////
    //Return the solution to all the nodes taking into accout the hanging nodes
    this->hanging_node_constraints.distribute(start);

}

//---------------------------------------------------------------------------
template <int dim>
double
NAME::AppCathode<dim>::estimate(const AppFrame::FEVectors& vectors)
{
    // Reinitialize error estimator vector and set Newmann boundaries:
    this->cell_errors.reinit(this->tr->n_active_cells());

    // Extract soluton from FEVectors
    unsigned int ind = vectors.find_vector("Solution");
    const BlockVector<double>& sol = vectors.vector(ind);

    //component_mask denote which components of the finite element space shall be used to
    //compute the error estimator. If it is left as specified by the default value (i.e. an empty array), all components are
    //used.
    std::vector<bool> comp_mask (this->element->n_blocks());

    // Set refinement flags for material_id = 1
    // -- material_id 1: catalyst layer
    comp_mask[0] = true; // - comp 1 = x02
    comp_mask[1] = true; //true;
    comp_mask[2] = true; //true; // - comp 2 = phi_m (membrane phase potential)

    KellyErrorEstimator<dim>::estimate(*this->dof,
                                       QGauss<dim-1>(3),
                                       typename FunctionMap<dim>::type(),
                                       sol,
                                       this->cell_errors,
                                       comp_mask);

    return 0.0;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::parse_init_solution(AppFrame::FEVector& dst,
                                           FuelCell::OperatingConditions& OC,
                                           bool &good_initial_solution)
{
    deallog << "Parsing initial solution" << std::endl;
    // Check for degrees of freedom matching in the vector and triangulations

    //Declare the operating conditions
    double V_cell = -OC.get_V();
    if (V_cell < 0.0)
        V_cell = -1*V_cell;
    double x_O2 = OC.get_x_o2();

    if (parse_init_solution_phi_s(dst, V_cell)
        && parse_init_solution_x_O2(dst, x_O2))
    {
        good_initial_solution = true;
    }
    else
    {
        deallog << "The loaded initial solution differs significantly from the current conditions" << std::endl;
        good_initial_solution = false;
    }
}

//---------------------------------------------------------------------------
template <int dim>
bool
NAME::AppCathode<dim>::parse_init_solution_phi_s(AppFrame::FEVector& dst, double& V_cell)
{
    unsigned int n_components = component_names.size();
    std::vector<bool> component_selection (n_components, false);
    //checking for phi_s
    for (unsigned int i = 0; i<n_components; ++i)
    {
        if (component_names[i] == "Electronic electrical potential")
            component_selection[i].flip();
    }

    // Check to see that the value of phi_s at the boundary is within reasonable range.
    {
        //The BPP/GDL boundary is defined as 2 in the data file
        unsigned char boundary[] = {2};
        std::set<unsigned char> boundary_indicators(boundary,boundary+1);
        std::vector<bool> selected_dofs (dst.size());
        DoFTools::extract_boundary_dofs(*this->dof,component_selection,selected_dofs,boundary_indicators);

        for(unsigned int i=0; i<dst.size(); i++)
        {
            if (selected_dofs[i])
            {
                double delta_phi = sqrt(pow((dst(i) - V_cell),2.)); // absolute value
                //deallog << "delta_phi " << delta_phi << std::endl;
                if (delta_phi > 0.2)
                    return false;
                else
                    dst(i) = V_cell;
            }
        }
    }
    return true;
}


//---------------------------------------------------------------------------
template <int dim>
bool
NAME::AppCathode<dim>::parse_init_solution_x_O2(AppFrame::FEVector& dst, double& x_O2)
{
    unsigned int n_components = component_names.size();
    std::vector<bool> component_selection (n_components, false);
    //checking for x_O2
    for (unsigned int i = 0; i<n_components; ++i)
    {
        if (component_names[i] == "Oxygen molar fraction")
            component_selection[i].flip();
    }

    // Check to see that the value of x_O2 at the cathode gas channel inlet s within reasonable range.
    {
        //The channel/GDL boundary is defined as 3 in the data file
        unsigned char boundary[] = {3};
        std::set<unsigned char> boundary_indicators(boundary,boundary+1);
        std::vector<bool> selected_dofs (dst.size());
        DoFTools::extract_boundary_dofs(*this->dof,component_selection,selected_dofs,boundary_indicators);

        for(unsigned int i=0; i<dst.size(); i++)
        {
            if (selected_dofs[i])
            {
                double delta_x_O2 = sqrt(pow((dst(i) - x_O2),2.))/x_O2; // relative change
                if (delta_x_O2 > 0.30) // 30% change
                    return false;
                else
                    dst(i) = x_O2;
            }
        }
    }
    return true;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//-----------    OPTIMIZATION ROUTINES     ----------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::cell_dresidual_dlambda(std::vector<AppFrame::FEVector >&,
                                              const typename DoFApplication<dim>::CellInfo&,
                                              std::vector<std::vector<double> >& /*src*/)
{
    // This has not yet been implemented. See app_cathode in the archive folder for
    // the code that used the old coding system.
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name()  << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::check_responses()
{
    // This has not yet been implemented. See app_cathode in the archive folder for
    // the code that used the old coding system.
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name()  << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::cell_responses (std::vector<double>& resp,
                                       const typename DoFApplication<dim>::CellInfo& info,
                                       const AppFrame::FEVector& /*src*/)
{
    const unsigned int last_iteration = 0;
    // -- Find out what material is the cell made of, i.e. MEA layer)
    const unsigned int material_id = info.dof_active_cell->material_id();

    // Use quadratue associated with the first element for the integration (this is arbitrary):
    const FEValuesBase<dim>& feValues = info.fe(0);
    // get integration points
    const unsigned int n_quad = feValues.n_quadrature_points;

    // Create a vector where the values of coefficients at quadrature points are stored:
    std::vector<double> values(n_quad);

    // Storing solution indices
    const unsigned int xO2_index = system_management.solution_name_to_index("oxygen_molar_fraction");
    const unsigned int phiM_index = system_management.solution_name_to_index("protonic_electrical_potential");
    const unsigned int phiS_index = system_management.solution_name_to_index("electronic_electrical_potential");

    for (unsigned int r = 0; r < this->n_resp; ++r)
    {
        if (this->name_responses[r] == "current")
        {
            if (CCL->belongs_to_material(material_id)) //the material is the catalyst layer
            {
                // Creating solution variable vector to passed to catalyst layer classes
                std::vector< FuelCellShop::SolutionVariable > solution_variables;
                solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][xO2_index], oxygen_molar_fraction) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][phiM_index], protonic_electrical_potential) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][phiS_index], electronic_electrical_potential) );

                CCL->set_solution(solution_variables);
                CCL->current_density(values);

                for (unsigned int q=0; q<n_quad; ++q)
                {
                    //integrate RR = \nabla*i at the quadrature points. Since all components of f_values have
                    //the value RR stored, I just use the first one.
                    resp[r] += -(values[q] * feValues.JxW(q)) / (grid->L_channel_c()/2.0 + grid->L_land_c()/2.0);
                }
            }
        }
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::global_responses(std::vector<double>&,
                                        const AppFrame::FEVector& /*src*/)
{
    // This has not yet been implemented.
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name()  << std::endl;

    //This is called during an non optimisation run, however has no effect on the solution. Commented
    //out so that the volume fraction object could be removed.
    /*  //-- Initialize var
     d ouble epsilon_S = 0.0;  *
     double epsilon_V = 0.0;
     double epsilon_N = 0.0;

     for (unsigned int r = 0; r < this->n_resp; ++r)
            {
                if (this->name_responses[r] == "epsilon_V_cat_c")
            {
                volume_fraction.epsilon_cat_c(epsilon_S, epsilon_V, epsilon_N);
                resp[r] = epsilon_V;
            }
            else if (this->name_responses[r] == "epsilon_S_cat_c")
            {
                volume_fraction.epsilon_cat_c(epsilon_S, epsilon_V, epsilon_N);
                resp[r] = epsilon_S;
            }
            }*/
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::cell_dresponses_dl(std::vector<std::vector<double> >&,
                                          const typename DoFApplication<dim>::CellInfo&,
                                          const AppFrame::FEVector& /*src*/)
{
    // This has not yet been implemented. See app_cathode in the archive folder for
    // the code that used the old coding system.
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name()  << std::endl;
}


//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::global_dresponses_dl(std::vector<std::vector<double> >&,
                                            const AppFrame::FEVector& )
{
    // This has not yet been implemented. See app_cathode in the archive folder for
    // the code that used the old coding system.
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name()  << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::cell_dresponses_du(std::vector<AppFrame::FEVector >&,
                                          const typename DoFApplication<dim>::CellInfo&,
                                          std::vector<std::vector<double> >& /*src*/)
{
    // This has not yet been implemented. See app_cathode in the archive folder for
    // the code that used the old coding system.
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name()  << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::global_dresponses_du(std::vector<AppFrame::FEVector >&,
                                            const AppFrame::FEVector& /*src*/)
{
    // This has not yet been implemented. See app_cathode in the archive folder for
    // the code that used the old coding system.
    const std::type_info& info = typeid(*this);
    deallog << "Pure function " << __FUNCTION__
    << " called in Class "
    << info.name()  << std::endl;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//-----------    POSTPROCESSING ROUTINES     --------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
template <int dim>
double
NAME::AppCathode<dim>::evaluate (const AppFrame::FEVectors& vectors)
{
    std::vector<double> test(this->n_resp, 0.0);
    this->responses(test,
                    vectors);
    return -test[0];
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppCathode<dim>::data_out(const std::string &basename,
                                const AppFrame::FEVectors &vectors)
{
    // Read the solution vector out of the many vectors:
    unsigned int i = vectors.find_vector("Solution");
    const BlockVector<double>& src = vectors.vector(i);

    //deallog<<"Number of vectors :"<<vectors.n_vectors()<<std::endl;
    //vectors.list(deallog);
    //std::ofstream marc;
    //marc.open("solution_out.dat");
    //vectors.vector(0).print(marc);


    // Output solution with its corresponent names:
    std::vector<std::string> solution_names;
    solution_names.push_back("x_O_2");
    solution_names.push_back("phi_m");
    solution_names.push_back("phi_s");
    AppFrame::DoFApplication<dim>::data_out(basename,vectors,solution_names);

    const std::string suffix = this->d_out.default_suffix();
    if (suffix == "") return;

    // -------------------------------------------------------------------------------------
    // Obtain the reaction rate over the catalyst layer and output values:
    if (this->element->n_base_elements() == 1)
    {
        AppFrame::FEVector source;
        source.reinit(src);

        AppFrame::FEVector reaction_rate(this->element->n_blocks(),
                                         this->element->base_element(this->element->block_to_base_index(0).first).dofs_per_cell);


        // Quadrature with points at the position of the dof of the base_element(0). I can choose any base element I like since the
        //main reason I do this is so that I can use add_data_vector which needs the vector to be of size n_dof.
        const Quadrature<dim> dummy_qr (this->element->base_element(0).get_generalized_support_points());
        std::vector<FEValues<dim>*> fe(this->element->n_base_elements());
        for (unsigned int i=0; i<fe.size(); ++i)
            fe[i] = new FEValues<dim> (*this->mapping,
                                       this->element->base_element(0),
                                       dummy_qr,
                                       UpdateFlags(update_values    |    update_q_points  |    update_JxW_values));

            const unsigned int dofs_per_cell = this->element->dofs_per_cell;

        // Create vector that related degrees of freedom of the cell with dof of the global system (to assembly global matrix)
        std::vector<unsigned int> indices_org (dofs_per_cell);
        std::vector<unsigned int> indices (dofs_per_cell);

        // loop over cells
        typename DoFHandler<dim>::active_cell_iterator cell;
        for (cell = this->dof->begin_active(); cell != this->dof->end(); ++cell)
        {
            // initialize the vector that relates local element dof to global dof
            cell->get_dof_indices (indices_org);
            for (unsigned int i=0; i<indices.size(); ++i)
                indices[this->block_info.local_renumbering[i]] = indices_org[i];

            //initialize shape function values for the cell
            for (unsigned int i=0; i<fe.size(); ++i)
                fe[i]->reinit ((typename Triangulation<dim>::cell_iterator) cell);

            // Since all fe are the same use fe(0) for all elements:
            const FEValuesBase<dim>& fe_values = *fe[0];

            // get integration points
            const unsigned int n_quad = fe_values.n_quadrature_points;
            std::vector<Point<dim> > point(n_quad);
            point = fe_values.get_quadrature_points();

            if (CCL->belongs_to_material(cell->material_id()))
            {
                // Create a vector to store solution in the cell:
                std::vector<std::vector<double> > solution(this->element->n_blocks(), std::vector<double> (n_quad, 0.));
                for (unsigned int i=0; i<this->element->n_blocks(); ++i)
                    fe_values.get_function_values(src,
                                                  make_slice(indices, this->block_info.local.block_start(i), fe_values.dofs_per_cell),
                                                  solution[i]);
                // [BC] calculate source term contributions at each quadrature point.
                std::vector<double> values(n_quad);

                // Creating solution variable vector to be passed to catalyst layer classes
                std::vector< FuelCellShop::SolutionVariable > solution_variables;
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("oxygen_molar_fraction")], oxygen_molar_fraction) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("protonic_electrical_potential")], protonic_electrical_potential) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("electronic_electrical_potential")], electronic_electrical_potential) );

                CCL->set_solution(solution_variables);
                CCL->current_density(values);

                for (unsigned int q=0; q<n_quad; ++q)
                {
                    reaction_rate.block(0)(q) = values[q]/(4.*Constants::F());  //Note I loop over the indices of the first component which
                    reaction_rate.block(1)(q) = values[q];
                    reaction_rate.block(2)(q) = -values[q];
                }
            }

            // Store in the global vector:
            //std::cout<<"SIZE SOURCE"<<source.size()<<"and SIZE REACTION_RATE "<<reaction_rate(3)<<std::endl;
            for (unsigned int i=0; i<this->element->dofs_per_cell; ++i)
                source(indices[i]) = reaction_rate(i);

            //Have to clear the reaction rate objects to ensure that the values from a previous
            //cell aren't used for current one (i.e. the values from a CCL cell used in a GDL one)
            for (unsigned int q=0; q<n_quad; ++q)
            {
                reaction_rate.block(0)(q) = 0.0;  //Note I loop over the indices of the first component which
                reaction_rate.block(1)(q) = 0.0;
                reaction_rate.block(2)(q) = 0.0;
            }
        }

        // Delete all FEValues. This is necessary in order to clean the subscriptor. Otherwise, when we
        // finish the subscriptor class issues an error:
        for (unsigned int i=0; i<fe.size(); ++i)
            delete fe[i];

        // names for source terms
        std::vector<std::string> source_names;
        source_names.push_back ("source_o2");
        source_names.push_back ("source_phi_m");
        source_names.push_back ("source_phi_s");

        this->d_out.add_data_vector(source, source_names, DataOut_DoFData<DoFHandler<dim>, dim>::type_dof_data, this->data_interpretation);
        this->d_out.build_patches();

    } //end if

    // Create string that will be appended to the .vtk filename. This string will contain the
    //name and value of the design variables.
    std::string filename1;

    //loop over all design variables
    for (i=0; i<design_var.size(); i++)
    {
        //create object that can print out numbers
        std::ostringstream streamOut;
        //Used to set how many digits after the decimal point are used. Insert the required number into the brackets.
        streamOut.precision();
        //If there are fewer digits in the design variable than are requested, the remaining digits can be filled with zeroes.
        //Use std::fixed to force this: streamOut <<  std::fixed << design_var_value[design_var.size()-i-1];
        //This can be used to get paraview to group the vtk files correctly.

        //Note that the design variables are printed in reverse, so that file managers will group the vtk files by the last design
        // variable first before moving to the second last etc.
        streamOut <<  design_var_value[design_var.size()-i-1];

        //Append design variable name and value to filename1
        filename1 += "__" + design_var[design_var.size()-i-1] + "_" + streamOut.str();
    }
    //construct the full filename, where basename is given by the adaptive refinement loop and suffix is the file extension (i.e. .vtk)
    const std::string filename = basename + filename1 + suffix;
    deallog << "Datafile:" << filename << std::endl;
    std::ofstream of (filename.c_str ());

    this->d_out.write (of);
    this->d_out.clear();
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations.
template class NAME::AppCathode<deal_II_dimension>;
