//---------------------------------------------------------------------------
//
//    Copyright (C) 2012 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <app_laplace.h>
#include <grid/grid_generator.h>

namespace NAME = FuelCell::Application;
using namespace AppFrame;

//---------------------------------------------------------------------------
template <int dim>
NAME::AppLaplace<dim>::AppLaplace(boost::shared_ptr<AppFrame::ApplicationData> data)
:
OptimizationBlockMatrixApplication<dim>(data),  
system_management(this->block_info, this->cell_couplings, this->flux_couplings),
FicksLaw(system_management,&oxygen,&nitrogen)
{
  deallog << "->FuelCell::Application::Laplace" << dim <<"d"<<std::endl;
 }
 
//---------------------------------------------------------------------------
template <int dim>
NAME::AppLaplace<dim>::~AppLaplace()
{}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppLaplace<dim>::declare_parameters(ParameterHandler& param)
{
  OptimizationBlockMatrixApplication<dim>::declare_parameters(param);
 
  system_management.declare_parameters(param);
  // Declare image:
    FuelCellShop::Geometry::GridBase<dim>::declare_GridGenerator_parameters(param);
  // Declare operating conditions:
  OC.declare_parameters(param);
  // Declare materials:
  FuelCellShop::Layer::GasDiffusionLayer<dim>::declare_GasDiffusionLayer_parameters("Cathode gas diffusion layer", param);
  // Declare equations:
  FicksLaw.declare_parameters(param);
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppLaplace<dim>::set_parameters(const std::vector<std::string>& /*name_dvar*/,
                                      const std::vector<double>& /*value_dvar*/,
                                      ParameterHandler& /*param*/)
{

}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppLaplace<dim>::_initialize(ParameterHandler& param)
{
  //Initialize the problem data:
  // This specifies the type of problem we have
  system_management.initialize(param);
  
  // Generate grid and refine
    grid = FuelCellShop::Geometry::GridBase<dim>::create_GridGenerator (param);
    grid->generate_grid(*this->tr);
  // Initialize operating conditions
  OC.initialize(param);
  // Initialize materials and layers
  CGDL = FuelCellShop::Layer::GasDiffusionLayer<dim>::create_GasDiffusionLayer("Cathode gas diffusion layer",param);
  std::vector< FuelCellShop::Material::PureGas * > gases;
  gases.push_back(&oxygen);
  gases.push_back(&nitrogen);
  CGDL->set_gases_and_compute(gases, OC.get_pc_atm (), OC.get_T());

  // Initialize equations
  FicksLaw.initialize(param);

  // Create couplings between equations and solution variables
  std::vector<couplings_map> tmp;
  tmp.push_back(FicksLaw.get_internal_cell_couplings() );
  system_management.make_cell_couplings(tmp);
  
  // Initialize dof per cell for the system of equations and information on
  //block sizes and indices
  this->remesh_dofs();
  // Initialize matrices and spartisity pattern for the whole system
  this->remesh_matrices();

  //system_management.print_system_info();
  // Print default parameter file:
  //this->print_parameters_to_file(param,"default_parameters.xml",ParameterHandler::XML); 
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppLaplace<dim>::initialize(ParameterHandler& param)
{
  OptimizationBlockMatrixApplication<dim>::initialize(param);
  _initialize(param);
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppLaplace<dim>::init_solution(AppFrame::FEVector& dst)
{
  // resize vector:
  dst.reinit(this->block_info.global);
  FuelCell::InitialSolution::AppLaplaceIC<dim> initial_solution(&OC);

  VectorTools::interpolate (*this->dof, 
			    initial_solution, 
			    dst);  
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppLaplace<dim>::cell_matrix(MatrixVector& cell_matrices,
				       const typename DoFApplication<dim>::CellInfo& info)
{
  FicksLaw.assemble_cell_matrix(cell_matrices,info,CGDL.get());
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppLaplace<dim>::cell_residual(AppFrame::FEVector& cell_vector,
				  const typename DoFApplication<dim>::CellInfo& info)
{ 
  FicksLaw.assemble_cell_residual(cell_vector,info,CGDL.get());
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppLaplace<dim>::dirichlet_bc(std::map<unsigned int, double>& boundary_values) const
{
  std::vector<bool> comp_mask (this->element->n_blocks(), false);
  comp_mask[0]=true;
  
  // -- boundary component 1: Dirichlet BCs
  ConstantFunction<dim> LHS(0.5);
  VectorTools::interpolate_boundary_values (*this->dof,
					    1,
					    LHS,
					    boundary_values,
					    comp_mask);
  /*
  VectorTools::interpolate_boundary_values (*this->dof,
					    1,
					    LHS,
					    boundary_values,
					    comp_mask);
					    */
  ConstantFunction<dim> RHS(2);
  VectorTools::interpolate_boundary_values (*this->dof,
					    2, //3
					    RHS,
					    boundary_values,
					    comp_mask);
  VectorTools::interpolate_boundary_values (*this->dof,
					    5,
					    RHS,
					    boundary_values,
					    comp_mask);
} 

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppLaplace<dim>::solve(AppFrame::FEVector& start,
				 const AppFrame::FEVectors& rhs)
  
{ 
    // COMPUTE THE RHS:
    AppFrame::FEVector residual;
    
    // I AM USING A LINEAR APPLICATION -> NO NEWTON LOOP, so No Newton Iterate avaiable:  
    residual.reinit(start);  
    double res = this->residual(residual, rhs, false);
    
    // Copy RHS to solution vector
    /*
    unsigned int index_Newton_residual = rhs.find_vector("Newton residual");
    residual = rhs.vector(index_Newton_residual); 
    start = rhs.vector(index_Newton_residual);  
    unsigned int index_Newton_iterate = rhs.find_vector("Newton iterate");
    */      
    // FOR DEBUG MODE ONLY:
    /*
     * std::ofstream fp_out;
     * fp_out.open("residual_vector.txt");
     * residual.print(fp_out);
     * fp_out.close();
     */
    
    // ASSEMBLY SYSTEM MATRIX:
    this->assemble(rhs); // rhs is useless since I am not  in a Newton loop.
    
    // FIX SYSTEM and RESIDUAL
    //SolverUtils diagonal;
    //diagonal.repair_diagonal(this->matrix);
    
    // FOR DEBUG MODE ONLY:
//     std::ofstream file_out;
//     file_out.open("system_matrix.txt");
//     this->matrix.print_formatted(file_out,3,true,0,0);
//     file_out.close();
//     abort();

    // Apply BC to matrix, rhs and sol
    std::map<unsigned int, double> boundary_values;
    this->dirichlet_bc(boundary_values);
    MatrixTools::apply_boundary_values(boundary_values,
                                       this->matrix,
                                       start, 
                                       residual);
    
    std::string solver("GMRES");  //UMFPACK OR GMRES
    
    if (solver.compare("UMFPACK") == 0)
    {
    //*********** UMFPACK *****

    
    // UMFPACK can only receive a Vector, so transform the assign the BlockVector to a normal
    // vector
    Vector<double> start_copy;
    start_copy = residual;
    // Facorize matrix and solve:
    SparseDirectUMFPACK solver;
    solver.initialize(this->matrix);
    solver.solve(start_copy);
    start = start_copy;
    //**********************************************
    }
    else if (solver.compare("GMRES")  == 0)
    {
    //************************
    // GMRES SOLVER:   
    LinearSolvers::GMRESSolver solver;
    solver.solve(this->matrix,start,residual,1000,1e-10,PreconditionIdentity());
    //**********************************************
    }
    
    // Distribute solution:
    //Return the solution to all the nodes taking into accout the hanging nodes
    this->hanging_node_constraints.distribute(start);
}

//---------------------------------------------------------------------------
template <int dim>
double 
NAME::AppLaplace<dim>::estimate(const AppFrame::FEVectors& vectors)
{ 
//   // Reinitialize error estimator vector and set Newmann boundaries:
  this->cell_errors.reinit(this->tr->n_active_cells());

  // Extract soluton from FEVectors
  unsigned int ind = vectors.find_vector("Solution");
  const BlockVector<double>& sol = vectors.vector(ind);

  //component_mask denote which components of the finite element space shall be used to 
  //compute the error estimator. If it is left as specified by the default value (i.e. an empty array), all components are 
  //used.
  std::vector<bool> comp_mask (this->element->n_blocks());
  
  // Set refinement flags for material_id = 1
  // -- material_id 1: catalyst layer
  comp_mask[0] = true; // - comp 1 = x02
      
  this->cell_errors.reinit(this->tr->n_active_cells());
  //typename FunctionMap<dim>::type newmann_boundary;
  KellyErrorEstimator<dim>::estimate(*this->dof,
				     QGauss<dim-1>(3),
				     typename FunctionMap<dim>::type(),
				     sol,
				     this->cell_errors,
				     comp_mask);
  
  return 0.0;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//-----------    POSTPROCESSING ROUTINES     --------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
template <int dim>
double 
NAME::AppLaplace<dim>::evaluate (const AppFrame::FEVectors& vectors)
{ 
	return 0.0;
}

//---------------------------------------------------------------------------
/*template <int dim>
void 
NAME::AppLaplace<dim>::data_out(const std::string &basename, 
				    const AppFrame::FEVectors &vectors)
{  

}
*/

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//-----------    OPTIMIZATION ROUTINES     ----------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template <int dim>
FuelCell::InitialSolution::AppLaplaceIC<dim>::AppLaplaceIC(FuelCell::OperatingConditions* OC)
  :
  Function<dim> (1),
  OC(OC)
{}

//---------------------------------------------------------------------------
template <int dim>
FuelCell::InitialSolution::AppLaplaceIC<dim>::~AppLaplaceIC()
{}

//---------------------------------------------------------------------------
template <int dim>
double
FuelCell::InitialSolution::AppLaplaceIC<dim>::value (const Point<dim> &p, 
						     unsigned int i) const
{
  return 0.0;
}

//---------------------------------------------------------------------------
template <int dim>
void
FuelCell::InitialSolution::AppLaplaceIC<dim>::vector_value (const Point<dim> &p, 
								Vector<double> &v) const
{
  if (p(0) == 0.0) 
  {
    for(unsigned int i = 0; i<v.size(); ++i)
      this->value(p,i);
  }
  else if (p(0) == 1.0)
  {
    for(unsigned int i = 0; i<v.size(); ++i)
      this->value(p,i);
  }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations. 
template class NAME::AppLaplace<deal_II_dimension>;
template class FuelCell::InitialSolution::AppLaplaceIC<deal_II_dimension>;