//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2009-2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: app_pemfc_multiscale.cc
//    - Description: 
//    - Developers: M. Secanell, P. Dobson, M. Bhaiya
//    - $Id: app_pemfc.cc 1383 2013-08-22 17:34:50Z wardlawp $
//
//---------------------------------------------------------------------------

#include "applications/app_pemfc.h"

namespace NAME = FuelCell::Application;

//---------------------------------------------------------------------------
template <int dim>
NAME::AppPemfc<dim>::AppPemfc(boost::shared_ptr<AppFrame::ApplicationData> data)
  :
  OptimizationBlockMatrixApplication<dim>(data),
  system_management(this->block_info, this->cell_couplings, this->flux_couplings),
  proton_transport(system_management),
  lambda_transport(system_management),
  electron_transport(system_management),
  reaction_source_terms(system_management),
  sorption_source_terms(system_management),
  ficks_oxygen_nitrogen(system_management, &oxygen, &nitrogen),
  ficks_water_nitrogen(system_management, &water, &nitrogen),
  ficks_water_hydrogen(system_management, &water, &hydrogen)
{
    deallog << "FuelCell::Application::AppPemfc-" << dim<<"d"<<std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
NAME::AppPemfc<dim>::~AppPemfc()
{

}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::declare_parameters(ParameterHandler& param)
{
    OptimizationBlockMatrixApplication<dim>::declare_parameters(param);

    // Declare parameters in system management:
    system_management.declare_parameters(param);
    
    // Declare grid generator parameters:
    FuelCellShop::Geometry::GridBase<dim>::declare_GridGenerator_parameters(param);

    // Declare parameters for operating conditions
    OC.declare_parameters(param);
        
    // Declare parameters for layer classes:
    FuelCellShop::Layer::GasDiffusionLayer<dim>::declare_GasDiffusionLayer_parameters("Anode gas diffusion layer", param);
    FuelCellShop::Layer::MicroPorousLayer<dim>::declare_MicroPorousLayer_parameters("Anode microporous layer", param);
    FuelCellShop::Layer::CatalystLayer<dim>::declare_CatalystLayer_parameters("Anode catalyst layer", param);
    FuelCellShop::Layer::MembraneLayer<dim>::declare_MembraneLayer_parameters("Membrane layer", param);
    FuelCellShop::Layer::CatalystLayer<dim>::declare_CatalystLayer_parameters("Cathode catalyst layer", param);
    FuelCellShop::Layer::MicroPorousLayer<dim>::declare_MicroPorousLayer_parameters("Cathode microporous layer", param);
    FuelCellShop::Layer::GasDiffusionLayer<dim>::declare_GasDiffusionLayer_parameters("Cathode gas diffusion layer", param);  
   
    // Declare parameters for physics classes:
    proton_transport.declare_parameters(param);
    lambda_transport.declare_parameters(param);
    electron_transport.declare_parameters(param);
    reaction_source_terms.declare_parameters(param);
    sorption_source_terms.declare_parameters(param);
    ficks_oxygen_nitrogen.declare_parameters(param);
    ficks_water_hydrogen.declare_parameters(param);
    ficks_water_nitrogen.declare_parameters(param);    
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::set_parameters(const std::vector<std::string>& name_dvar,
                                        const std::vector<double>& value_dvar,
                                        ParameterHandler& param)
{
    // Set any parameters in the parameter file related to operating conditions and mesh  
    FuelCellShop::Geometry::GridBase<dim>::set_GridGenerator_parameters (name_dvar, value_dvar, param);
    OC.set_parameters(name_dvar, value_dvar, param);
    
    // Set any parameters in the parameter file related to the layers.
    FuelCellShop::Layer::GasDiffusionLayer<dim>::set_GasDiffusionLayer_parameters(name_dvar, value_dvar, "Anode gas diffusion layer", param);
    FuelCellShop::Layer::MicroPorousLayer<dim>::set_MicroPorousLayer_parameters(name_dvar, value_dvar, "Anode microporous layer", param);
    FuelCellShop::Layer::CatalystLayer<dim>::set_CatalystLayer_parameters(name_dvar, value_dvar, "Anode catalyst layer", param);
    FuelCellShop::Layer::MembraneLayer<dim>::set_MembraneLayer_parameters(name_dvar, value_dvar, "Membrane layer", param);
    FuelCellShop::Layer::CatalystLayer<dim>::set_CatalystLayer_parameters(name_dvar, value_dvar, "Cathode catalyst layer", param);
    FuelCellShop::Layer::MicroPorousLayer<dim>::set_MicroPorousLayer_parameters(name_dvar, value_dvar, "Cathode microporous layer", param);
    FuelCellShop::Layer::GasDiffusionLayer<dim>::set_GasDiffusionLayer_parameters(name_dvar, value_dvar, "Cathode gas diffusion layer", param);    
    
    //Store names and values of the design variables
    design_var = name_dvar;
    design_var_value = value_dvar;
    
    // For debug purposes only:
    //   std::ofstream file;
    //   file.open("Modified_parameter_file.prm");
    //   param.print_parameters (file, ParameterHandler::XML);
    //   file.close();
    
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppPemfc<dim>::_initialize(ParameterHandler& param)
{
    //Initialize the problem data: 
    component_names.clear();
    component_names.push_back("Oxygen molar fraction");
    component_names.push_back("Water molar fraction");
    component_names.push_back("Protonic electrical potential");
    component_names.push_back("Electronic electrical potential");
    component_names.push_back("Membrane water content");
    
    // Initialize parameters in system management:
    system_management.initialize(param);
    
    // Make sure that the number of input finite elements matches with the problem:
    Assert (this->element->n_blocks() == system_management.get_number_of_solution_names(),
            ExcDimensionMismatch(this->element->n_blocks(), system_management.get_number_of_solution_names()));
    Assert (this->element->n_blocks() == system_management.get_number_of_solution_names(),
            ExcDimensionMismatch(this->element->n_blocks(), system_management.get_number_of_solution_names()));  
    
    initialize_grid(param);
    
    // Initialize coefficients (NOTE: grid already initialized)
    OC.initialize(param);
    
    // Initialize gases and material classes:
    std::vector< FuelCellShop::Material::PureGas* > cathode_gases;
    cathode_gases.push_back(&oxygen);
    cathode_gases.push_back(&water);
    cathode_gases.push_back(&nitrogen);
    std::vector< FuelCellShop::Material::PureGas* > anode_gases;
    anode_gases.push_back(&water);
    anode_gases.push_back(&hydrogen); 
    
    // Initialize layer classes:    
    AGDL = FuelCellShop::Layer::GasDiffusionLayer<dim>::create_GasDiffusionLayer("Anode gas diffusion layer",param);
    AGDL->set_gases_and_compute(anode_gases, OC.get_pa_atm (), OC.get_T());
    
    AMPL = FuelCellShop::Layer::MicroPorousLayer<dim>::create_MicroPorousLayer("Anode microporous layer",param);
    AMPL->set_gases_and_compute(anode_gases, OC.get_pa_atm (), OC.get_T());

    ACL = FuelCellShop::Layer::CatalystLayer<dim>::create_CatalystLayer("Anode catalyst layer", param);
    ACL->set_gases_and_compute (anode_gases, OC.get_pa_atm (), OC.get_T());
    
    ML = FuelCellShop::Layer::MembraneLayer<dim>::create_MembraneLayer("Membrane layer", param);
    
    CCL = FuelCellShop::Layer::CatalystLayer<dim>::create_CatalystLayer("Cathode catalyst layer", param);
    CCL->set_gases_and_compute (cathode_gases, OC.get_pc_atm (), OC.get_T());
    
    CMPL = FuelCellShop::Layer::MicroPorousLayer<dim>::create_MicroPorousLayer("Cathode microporous layer",param);
    CMPL->set_gases_and_compute(cathode_gases, OC.get_pc_atm (), OC.get_T());
    
    CGDL = FuelCellShop::Layer::GasDiffusionLayer<dim>::create_GasDiffusionLayer("Cathode gas diffusion layer",param);
    CGDL->set_gases_and_compute(cathode_gases, OC.get_pc_atm (), OC.get_T());
    
    // Initialize necessary kinetics parameters in catalyst layers.
    std::string name = "HOR";
    ACL->set_reaction_kinetics(name);
    name = "ORR";
    CCL->set_reaction_kinetics(name);
    
    // Setting required constant solutions, viz., temperature and pressure in the catalyst layer and membrane layer objects.    
    ACL->set_constant_solution(OC.get_pa_Pa(), total_pressure);   //Note: Pressure and temp. can only be set after setting kinetics.
    ACL->set_constant_solution(OC.get_T(), temperature_of_REV);
    CCL->set_constant_solution(OC.get_pc_Pa(), total_pressure);   //Note: Pressure and temp. can only be set after setting kinetics.
    CCL->set_constant_solution(OC.get_T(), temperature_of_REV);
    ML->set_constant_solution(OC.get_T(), temperature_of_REV);
    
    // Setting kinetics in the reaction source terms object.
    reaction_source_terms.set_cathode_kinetics(CCL->get_kinetics());
    reaction_source_terms.set_anode_kinetics(ACL->get_kinetics());
    
    // Initialize parameters for physics classes:
    ficks_oxygen_nitrogen.initialize(param);
    ficks_water_hydrogen.initialize(param);
    ficks_water_nitrogen.initialize(param);
    proton_transport.initialize(param);
    lambda_transport.initialize(param);
    electron_transport.initialize(param);
    reaction_source_terms.initialize(param);
    sorption_source_terms.initialize(param);

    // Initialize system of equations based on the physical phenomena being solved for:
    std::vector<couplings_map> tmp;
    tmp.push_back(ficks_water_nitrogen.get_internal_cell_couplings() );
    tmp.push_back(ficks_water_hydrogen.get_internal_cell_couplings() );
    tmp.push_back(ficks_oxygen_nitrogen.get_internal_cell_couplings() );
    tmp.push_back(proton_transport.get_internal_cell_couplings() );
    tmp.push_back(electron_transport.get_internal_cell_couplings() );
    tmp.push_back(lambda_transport.get_internal_cell_couplings() );
    reaction_source_terms.adjust_internal_cell_couplings(tmp);
    sorption_source_terms.adjust_internal_cell_couplings(tmp);
    system_management.make_cell_couplings(tmp);  
    
    // Initialize dof per cell for the system of equations and information on
    //block sizes and indices
    this->remesh_dofs();
    // Initialize matrices and spartisity pattern for the whole system
    this->remesh_matrices();
 
    // Output CL properties
    OC.print_operating_conditions();
    ACL->print_layer_properties();
    CCL->print_layer_properties();
    
    // FOR DEBUGGING PURPOSES ONLY:
    //   std::ofstream file;
    //   file.open("Modified_parameter_file.prm");
    //   param.print_parameters (file, ParameterHandler::XML);
    //   file.close();
    
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppPemfc<dim>::initialize(ParameterHandler& param)
{
  OptimizationBlockMatrixApplication<dim>::initialize(param);
  _initialize(param);
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppPemfc<dim>::initialize_grid(ParameterHandler& param)
{
  grid = FuelCellShop::Geometry::GridBase<dim>::create_GridGenerator (param);

  // Check to see whether we are reading the grid from a file
  // Must read grid if we are reading a stored solution
  bool read_grid_from_file = grid->read_from_file;

  // If not reading from a grid from a previous solution, 
  //	check whether we can use one from external source
  if (this->read_in_initial_solution || read_grid_from_file)
  {
	// TODO: Read grid_file name and format from parameter handler
	std::ifstream grid_file("._mesh_to_be_transfered.msh");
	if(!grid_file.fail())
	{
		deallog << "Reading grid from file" << std::endl;
		GridIn<dim> grid_in;
		grid_in.attach_triangulation(*this->tr);
		grid_in.read_msh(grid_file);
		grid_file.close();
		read_grid_from_file = true;
	}
	else
	{
		// if grid fails to read, generate standard grid, don't read initial solution
		deallog << "Grid file not available for reading" << std::endl;
		read_grid_from_file = false;
		this->read_in_initial_solution = false;
	}
  }
  if(!read_grid_from_file)
  {
  	deallog << "Generating new grid" << std::endl;	
	grid->generate_grid(*this->tr);
  }
}


//---------------------------------------------------------------------------
template <int dim>
void
NAME::AppPemfc<dim>::init_solution(AppFrame::FEVector& dst)
{
    bool solution_read_success = false;
    bool use_transfer_solution = false;
    // resize vector:
    dst.reinit(this->block_info.global);
    
    if (this->read_in_initial_solution)
        this->read_init_solution(dst,solution_read_success);
    
    // If the initial solution is read in properly,
    // 	check to see that the initial solution (i.e. Boundary Conditions)
    //	is within reasonable range for this simulation
    if (solution_read_success)
        parse_init_solution(dst, OC, use_transfer_solution);
    
    if (!use_transfer_solution)
    {
        FuelCell::InitialSolution::AppPemfcIC<dim> initial_solution(&OC, grid);
        // resize vector:
        deallog << "Generating default initial solution" << std::endl;
        dst.reinit(this->block_info.global);
        VectorTools::interpolate (*this->dof, initial_solution, dst);
    }
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::cell_matrix(MatrixVector& cell_matrices,
                                      const typename DoFApplication<dim>::CellInfo& info)
{

    const unsigned int material_id = info.cell->material_id();
    
    if ( CGDL->belongs_to_material(material_id) )
    {
        ficks_oxygen_nitrogen.assemble_cell_matrix(cell_matrices, info, CGDL.get());
        ficks_water_nitrogen.assemble_cell_matrix(cell_matrices, info, CGDL.get());
        electron_transport.assemble_cell_matrix(cell_matrices, info, CGDL.get());
    }
    else if ( CMPL->belongs_to_material(material_id) )
    {
        ficks_oxygen_nitrogen.assemble_cell_matrix(cell_matrices, info, CMPL.get());
        ficks_water_nitrogen.assemble_cell_matrix(cell_matrices, info, CMPL.get());
        electron_transport.assemble_cell_matrix(cell_matrices, info, CMPL.get());
    }
    else if ( CCL->belongs_to_material(material_id) )
    {
        ficks_oxygen_nitrogen.assemble_cell_matrix(cell_matrices, info, CCL.get());
        ficks_water_nitrogen.assemble_cell_matrix(cell_matrices, info, CCL.get());
        proton_transport.assemble_cell_matrix(cell_matrices, info, CCL.get());
        lambda_transport.assemble_cell_matrix(cell_matrices, info, CCL.get());
        electron_transport.assemble_cell_matrix(cell_matrices, info, CCL.get());
        reaction_source_terms.assemble_cell_matrix(cell_matrices, info, CCL.get());
        sorption_source_terms.assemble_cell_matrix(cell_matrices, info, CCL.get());
    }
    else if ( ML->belongs_to_material(material_id) )
    {
        proton_transport.assemble_cell_matrix(cell_matrices, info, ML.get());
        lambda_transport.assemble_cell_matrix(cell_matrices, info, ML.get());
    }
    else if ( ACL->belongs_to_material(material_id) )
    {
        ficks_water_hydrogen.assemble_cell_matrix(cell_matrices, info, ACL.get());
        proton_transport.assemble_cell_matrix(cell_matrices, info, ACL.get());
        lambda_transport.assemble_cell_matrix(cell_matrices, info, ACL.get());
        electron_transport.assemble_cell_matrix(cell_matrices, info, ACL.get());
        reaction_source_terms.assemble_cell_matrix(cell_matrices, info, ACL.get());
        sorption_source_terms.assemble_cell_matrix(cell_matrices, info, ACL.get());
    }
    else if ( AMPL->belongs_to_material(material_id) )
    {
        ficks_water_hydrogen.assemble_cell_matrix(cell_matrices, info, AMPL.get());
        electron_transport.assemble_cell_matrix(cell_matrices, info, AMPL.get());
    }
    else if ( AGDL->belongs_to_material(material_id) )
    {
        ficks_water_hydrogen.assemble_cell_matrix(cell_matrices, info, AGDL.get());
        electron_transport.assemble_cell_matrix(cell_matrices, info, AGDL.get());
    }
    else
        deallog<<"Material id: "    <<info.cell->material_id()<<" does not correspond to any layer"<<std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::cell_residual(AppFrame::FEVector& cell_vector,
                                        const typename DoFApplication<dim>::CellInfo& info)
{ 

    // -- Assertion before starting routine:
    // Make sure vectors are the right size
    Assert (cell_vector.n_blocks() == this->element->n_blocks(),
                        ExcDimensionMismatch (cell_vector.n_blocks(), this->element->n_blocks()));
    
    // -- Find out what material is the cell made of, i.e. MEA layer)
    const unsigned int material_id = info.cell->material_id();
    
    if ( CGDL->belongs_to_material(material_id) )
    {
        ficks_oxygen_nitrogen.assemble_cell_residual(cell_vector, info, CGDL.get());
        ficks_water_nitrogen.assemble_cell_residual(cell_vector, info, CGDL.get());
        electron_transport.assemble_cell_residual(cell_vector, info, CGDL.get());
    }
    else if ( CMPL->belongs_to_material(material_id) )
    {
        ficks_oxygen_nitrogen.assemble_cell_residual(cell_vector, info, CMPL.get());
        ficks_water_nitrogen.assemble_cell_residual(cell_vector, info, CMPL.get());
        electron_transport.assemble_cell_residual(cell_vector, info, CMPL.get());
    }
    else if ( CCL->belongs_to_material(material_id) )
    {
        ficks_oxygen_nitrogen.assemble_cell_residual(cell_vector, info, CCL.get());
        ficks_water_nitrogen.assemble_cell_residual(cell_vector, info, CCL.get());
        proton_transport.assemble_cell_residual(cell_vector, info, CCL.get());
        lambda_transport.assemble_cell_residual(cell_vector, info, CCL.get());
        electron_transport.assemble_cell_residual(cell_vector, info, CCL.get());
        reaction_source_terms.assemble_cell_residual(cell_vector, info, CCL.get());
        sorption_source_terms.assemble_cell_residual(cell_vector, info, CCL.get());
    }
    else if ( ML->belongs_to_material(material_id) )
    {
        proton_transport.assemble_cell_residual(cell_vector, info, ML.get());
        lambda_transport.assemble_cell_residual(cell_vector, info, ML.get());
    }
    else if ( ACL->belongs_to_material(material_id) )
    {
        ficks_water_hydrogen.assemble_cell_residual(cell_vector, info, ACL.get());
        proton_transport.assemble_cell_residual(cell_vector, info, ACL.get());
        lambda_transport.assemble_cell_residual(cell_vector, info, ACL.get());
        electron_transport.assemble_cell_residual(cell_vector, info, ACL.get());
        reaction_source_terms.assemble_cell_residual(cell_vector, info, ACL.get());
        sorption_source_terms.assemble_cell_residual(cell_vector, info, ACL.get());
    }
    else if ( AMPL->belongs_to_material(material_id) )
    {
        ficks_water_hydrogen.assemble_cell_residual(cell_vector, info, AMPL.get());
        electron_transport.assemble_cell_residual(cell_vector, info, AMPL.get());
    }
    else if ( AGDL->belongs_to_material(material_id) )
    {
        ficks_water_hydrogen.assemble_cell_residual(cell_vector, info, AGDL.get());
        electron_transport.assemble_cell_residual(cell_vector, info, AGDL.get());
    }
    else
        AssertThrow(false, ExcNotImplemented());

}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::dirichlet_bc(std::map<unsigned int, double>& boundary_values) const
{
    std::vector<bool> comp_mask (this->element->n_blocks());
    // -- boundary component 1: GDL/land at cathode
    comp_mask[0] = false; 
    comp_mask[1] = false;
    comp_mask[2] = false; 
    comp_mask[3] = true;
    comp_mask[4] = false;
    VectorTools::interpolate_boundary_values (*this->dof,
                                              1,
                                              ZeroFunction<dim>(this->element->n_blocks()),
                                              boundary_values,
                                              comp_mask);
    
    // -- boundary component 2: gas channel at cathode
    comp_mask[0] = true; 
    comp_mask[1] = true;
    comp_mask[2] = false; 
    comp_mask[3] = false;
    comp_mask[4] = false;
    VectorTools::interpolate_boundary_values (*this->dof,
                                              2,
                                              ZeroFunction<dim>(this->element->n_blocks()),
                                              boundary_values,
                                              comp_mask);
    // -- boundary component 3: GDL/land at anode
    comp_mask[0] = false; 
    comp_mask[1] = false;
    comp_mask[2] = false; 
    comp_mask[3] = true;
    comp_mask[4] = false;
    VectorTools::interpolate_boundary_values (*this->dof,
                                              3,
                                              ZeroFunction<dim>(this->element->n_blocks()),
                                              boundary_values,
                                              comp_mask);
    
    // -- boundary component 4: gas channel at anode
    comp_mask[0] = false; 
    comp_mask[1] = true;
    comp_mask[2] = false; 
    comp_mask[3] = false;
    comp_mask[4] = false;
    VectorTools::interpolate_boundary_values (*this->dof,
                                              4,
                                              ZeroFunction<dim>(this->element->n_blocks()),
                                              boundary_values,
                                              comp_mask);
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::solve(AppFrame::FEVector& start,
                                const AppFrame::FEVectors& rhs)
{
    Assert(start.size() == rhs.vector(0).size(),
           ExcDimensionMismatch(start.size(), rhs.vector(0).size()));
    
    
    //  if (this->notifications.any())
    if (this->notifications.test(AppFrame::Newton::bad_derivative) ||
        this->notifications.test(AppFrame::NewtonLineSearch::bad_derivative) ||
        this->notifications.test(AppFrame::Newton3pp::bad_derivative) ||
        this->notifications.test(AppFrame::Newton3ppC::bad_derivative) ||
        //this->notifications.test(AppFrame::NewtonDamp::bad_derivative) ||
        this->notifications.test(OptimizationBlockMatrixApplication<dim>::sensitivity_analysis) )
    {
        deallog<<"Assembling stiffness matrix"<<std::endl;
        AppFrame::FEVectors sol;
        unsigned int ind = rhs.find_vector("Newton iterate");
        sol.add_vector(rhs.vector(ind),rhs.vector_name(ind));
        //sol.add_vector(rhs.vector(1),"Newton iterate");
        this->assemble(sol); //Note the second component of
        this->notifications.clear();//this->clear_assemble();
    }
    
    
    // Eliminate zeros in the diagonal
    SolverUtils diagonal;
    unsigned int index_Newton_iterate = rhs.find_vector("Newton iterate");
    diagonal.repair_diagonal(this->matrix, start, rhs.vector(index_Newton_iterate));
          
    // Copy RHS to solution vector
    unsigned int index_Newton_residual = rhs.find_vector("Newton residual");
    start = rhs.vector(index_Newton_residual);     

    //  FOR DEBUG PURPOSES ONLY     
    /*
     * diagonal.print_diagonal(this->matrix);
     * 
     * std::cout<<"== MATRIX BLOCK 1 =="<<std::endl; 
     * std::ofstream file;
     * file.open("matrix_file_00.dat");
     * this->matrix.block(0,0).print_formatted(file, 3, true, 0, " 0 ");
     * file.close();
     * file.open("matrix_file_11.dat");
     * this->matrix.block(1,1).print_formatted(file, 3, true, 0, " 0 ");
     * file.close();
     * file.open("matrix_file_22.dat");
     * this->matrix.block(2,2).print_formatted(file, 3, true, 0, " 0 ");
     * file.close();
     * file.open("matrix_file_33.dat");
     * this->matrix.block(3,3).print_formatted(file, 3, true, 0, " 0 ");
     * file.close();
     * 
     * std::cout<<"============================"<<std::endl;
     * std::cout<<"== RHS TERMS =="<<std::endl;
     * // Output RHS in the dof order:
     * for (unsigned int i = 0; i<rhs.size(); ++i)
     *   std::cout<<"RHS value in row "<<i<<" is "<<rhs(i)<<std::endl;
     * // Output RHS in by block order:
     * unsigned int index = 0;
     * for (unsigned int i = 0; i<this->matrix.n_block_rows(); ++i)
     *   for (unsigned int j = 0; j<this->matrix.n_block_cols(); ++j)
     *     {
     * std::cout<<"Block in position ("<<i<<", "<<j<<") has "<<this->matrix.block(i,j).n_actually_nonzero_elements()<<" actual nonzero elements."<<std::endl;
     * index += 1; }
     */
        
    // Make sure that the solution is of the same size as the RHS
    Assert(start.size() == rhs.vector(index_Newton_iterate).size(),
           ExcDimensionMismatch(start.size(), rhs.vector(index_Newton_iterate).size()));
    
    std::string solver_name ("UMFPACK"); //("UMFPACK") | ("ILU-GMRES");
    
    //*********** GMRES with ILU preconditioner *****
    if (solver_name.compare("ILU-GMRES") == 0) //Could also use if (strcmp (solver_name,"ILU-GMRES") == 0)
    { 
        deallog<<"-- Solving problem using ILU-GMRES solver --"<<std::endl;
        /*
        
        LinearSolvers::ILUPreconditioner ilu_preconditioner(this->matrix);
        
        LinearSolvers::GMRESSolver solver;
        solver.solve(this->filtered_matrix,
                     start,
                     rhs.vector(index_Newton_residual),
                     1000,
                     1.e-10,
                     ilu_preconditioner.preconditioner);
        */
              
        // ******* SETUP PRECONDITIONER **********************
        // Setup preconditioner vector (one per block):
        //std::vector<PreconditionSSOR< > > preconditioner_matrix(system_matrix.n_block_rows());
        std::vector<SparseILU<double> > preconditioner_matrix(this->matrix.n_block_rows());
        
        // The preconditioners do not have vmult_add, so we need an auxiliary class
        //  std::vector<PointerMatrixAux<PreconditionSSOR< >, Vector<double> > > preconditioner_pointer(system_matrix.n_block_rows());
        std::vector<PointerMatrixAux<SparseILU<double>, Vector<double> > > preconditioner_pointer(this->matrix.n_block_rows());
        
        // Initialize the preconditioner for each block:
        for (unsigned int i=0; i<this->matrix.n_block_rows(); ++i)
        {
            //preconditioner_matrix[i].initialize(this->matrix.block(i,i));
            preconditioner_matrix[i].initialize(this->matrix.block(i,i), SparseILU<double>::AdditionalData());
            preconditioner_pointer[i].set_memory(&this->data->vector_pool);
            preconditioner_pointer[i] = &preconditioner_matrix[i];
        }
        
        // 
        BlockTrianglePrecondition<double> preconditioner;
        preconditioner.initialize(this->matrix.n_block_rows(),
                                  this->data->vector_pool);
        
        // 
        for (unsigned int i=0; i<this->matrix.n_block_rows(); ++i)
        {
            for (unsigned int j=0; j<this->matrix.n_block_cols(); ++j)
            {
                //system_matrix.enter(this->matrix.block(i,j), i, j);
                if (i==j)
                    preconditioner.enter(preconditioner_pointer[i], i, i);
                else
                    preconditioner.enter(this->matrix.block(i,j), i, j);
            }
        }
        //**************
        
        //****** Setup a system matrix:    
        // Nothing to be done. It's all done in block_matrix application for matrix and RHS.
        // Summary of things done:
        // In block_matrix_application.cc assembly:
        //   this->hanging_node_constraints.condense(this->matrix);
        //   filtered_matrix.clear_constraints();
        //   std::map<unsigned int, double> boundary_values;
        //   this->dirichlet_bc(boundary_values);
        //   filtered_matrix.initialize(this->matrix);
        //   filtered_matrix.add_constraints(boundary_values);
        // And then, in residual_constraints (where dst = rhs)
        //   this->hanging_node_constraints.condense(dst);
        //   filtered_matrix.apply_constraints(dst,false); 
        
        //****** Solve sytem:
        SolverControl control(10000, 1e-12);
        SolverGMRES<AppFrame::FEVector > gmres (control, 
                                                this->data->block_vector_pool,
                                                SolverGMRES<AppFrame::FEVector >::AdditionalData (100));
        gmres.solve(this->filtered_matrix, start, rhs.vector(index_Newton_residual), preconditioner);      
        //*************************
        
    }
    //*********** UMFPACK *****
    else if (solver_name.compare("UMFPACK") == 0)
    {

        deallog<<"-- Solving problem using UMFPACK solver --"<<std::endl;
        
        // Apply BC to matrix, rhs and sol
        std::map<unsigned int, double> boundary_values;
        this->dirichlet_bc(boundary_values);
        
        MatrixTools::apply_boundary_values(boundary_values,
                                           this->matrix,
                                           start,
                                           start);
        // UMFPACK can only receive a Vector, so transform the assign the BlockVector to a normal
        // vector
        Vector<double> start_copy;
        start_copy = start;
        // Factorize matrix and solve:
        SparseDirectUMFPACK solver;
        solver.initialize(this->matrix);
        solver.solve(start_copy);
        start = start_copy;   
        //*************************
    }
    else
    {
        const std::type_info& info = typeid(*this);
        deallog<<"Solver not implemented in Class "<<info.name()<<". Member function solve()"<<std::endl;
        abort();
    }
    
    //////////////////////////////////////////////////////////////////////
    // Distribute solution:
    //////////////////////////////////////////////////////////////////////
    //Return the solution to all the nodes taking into accout the hanging nodes
    this->hanging_node_constraints.distribute(start);
    
}

//---------------------------------------------------------------------------
template <int dim>
double 
NAME::AppPemfc<dim>::estimate(const AppFrame::FEVectors& vectors)
{ 
    // Solution:
    //const BlockVector<double>& sol = *(vectors.vector(0));
    
    //component_mask denote which components of the finite element space shall be used to 
    //compute the error estimator. If it is left as specified by the default value (i.e. an empty array), all components are 
    //used.
    std::vector<bool> comp_mask (this->element->n_blocks());
    
    // Set refinement flags for material_id = 1
    // -- material_id 1: catalyst layer
    comp_mask[0] = true; // - comp 1 = x02
    comp_mask[1] = true;
    comp_mask[2] = true; // - comp 2 = phi_m (membrane phase potential)
    comp_mask[3] = true; // - comp 3 = phi_s (solid phase potential)
    comp_mask[4] = true;
    
    this->cell_errors.reinit(this->tr->n_active_cells());
    typename FunctionMap<dim>::type newmann_boundary;
    KellyErrorEstimator<dim>::estimate(*this->dof,
                                       QGauss<dim-1>(3),
                                       newmann_boundary,
                                       vectors.vector(0), //sol,
                                       this->cell_errors,
                                       comp_mask);
  
    return 0.0; 
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::parse_init_solution(AppFrame::FEVector& dst,
                                              FuelCell::OperatingConditions& OC,
                                              bool &good_initial_solution)
{
    deallog << "Parsing initial solution" << std::endl;
    // Check for degrees of freedom matching in the vector and triangulations
    
    //Declare the operating conditions
    double V_cell = OC.get_V();
    double x_wv_a = 1.0 - OC.get_x_h2();
    double x_wv_c = OC.get_x_wv();
    double x_O2 = OC.get_x_o2();
    
    if (parse_init_solution_phi_s(dst, V_cell) 
        && parse_init_solution_x_O2(dst, x_O2) 
        && parse_init_solution_x_wv(dst, x_wv_a, x_wv_c))
        // If these tests pass, the values on the boundary are replaced in the source vector.
    {
        good_initial_solution = true;
    }
    else
    {
        deallog << "The loaded initial solution differs significantly from the current conditions" << std::endl;
        good_initial_solution = false;
    }
}

//---------------------------------------------------------------------------
template <int dim>
bool  
NAME::AppPemfc<dim>::parse_init_solution_phi_s(AppFrame::FEVector& dst, double& V_cell)
{
    unsigned int n_components = component_names.size();
    std::vector<bool> component_selection (n_components, false);
    //checking for phi_s = component[3]
    component_selection[3].flip();
    
    // Check to see that the value of phi_s at the boundary is within reasonable range.
    {
        unsigned char boundary[] = {1};
        std::set<unsigned char> boundary_indicators(boundary,boundary+1);
        std::vector<bool> selected_dofs (dst.size());
        DoFTools::extract_boundary_dofs(*this->dof,component_selection,selected_dofs,boundary_indicators);
        
        for(unsigned int i=0; i<dst.size(); i++)
        {
            if (selected_dofs[i])
            {
                double delta_phi = sqrt(pow((dst(i) - V_cell),2.)); // absolute value
                if (delta_phi > 0.2)
                    return false;
                else
                    dst(i) = V_cell;
            }
        }
    }
    return true;
}


//---------------------------------------------------------------------------
template <int dim>
bool  
NAME::AppPemfc<dim>::parse_init_solution_x_O2(AppFrame::FEVector& dst, double& x_O2)
{
    unsigned int n_components = component_names.size();
    std::vector<bool> component_selection (n_components, false);
    //checking for x_O2 = component[0]
    component_selection[0].flip();
    
    // Check to see that the value of x_O2 at the cathode gas channel inlet s within reasonable range.
    {
        unsigned char boundary[] = {2};
        std::set<unsigned char> boundary_indicators(boundary,boundary+1);
        std::vector<bool> selected_dofs (dst.size());
        DoFTools::extract_boundary_dofs(*this->dof,component_selection,selected_dofs,boundary_indicators);
        
        for(unsigned int i=0; i<dst.size(); i++)
        {
            if (selected_dofs[i])
            {
                double delta_x_O2 = sqrt(pow((dst(i) - x_O2),2.))/x_O2; // relative change
                if (delta_x_O2 > 0.30) // 30% change 
                    return false;
                else
                    dst(i) = x_O2;
            }
        }
    }
    return true;
}

//---------------------------------------------------------------------------
template <int dim>
bool  
NAME::AppPemfc<dim>::parse_init_solution_x_wv(AppFrame::FEVector& dst, double& x_wv_a, double& x_wv_c)
{
    unsigned int n_components = component_names.size();
    std::vector<bool> component_selection (n_components, false);
    //checking for x_wv = component[1]
    component_selection[1].flip();
    
    // Check to see that the value of x_wv at the cathode gas channel inlet s within reasonable range.
    {
        unsigned char boundary[] = {2};
        std::set<unsigned char> boundary_indicators(boundary,boundary+1);
        std::vector<bool> selected_dofs (dst.size());
        DoFTools::extract_boundary_dofs(*this->dof,component_selection,selected_dofs,boundary_indicators);
        
        for(unsigned int i=0; i<dst.size(); i++)
        {
            if (selected_dofs[i])
            {
                double delta_x_wv_c = sqrt(pow((dst(i) - x_wv_c),2.))/x_wv_c; // relative change
                if (delta_x_wv_c > 0.3) // 30% change
                    return false;
                else
                    dst(i) = x_wv_c;
            }
        }
    }
    
    // Check to see that the value of x_wv at the anode gas channel inlet s within reasonable range.
    {
        unsigned char boundary[] = {4};
        std::set<unsigned char> boundary_indicators(boundary,boundary+1);
        std::vector<bool> selected_dofs (dst.size());
        DoFTools::extract_boundary_dofs(*this->dof,component_selection,selected_dofs,boundary_indicators);
        
        for(unsigned int i=0; i<dst.size(); i++)
        {
            if (selected_dofs[i])
            {
                double delta_x_wv_a = sqrt(pow((dst(i) - x_wv_a),2.))/x_wv_a; // relative change
                if (delta_x_wv_a > 0.3) // 30% change 
                    return false;
                else
                    dst(i) = x_wv_a;
            }
        }
    }
    return true;
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//-----------    OPTIMIZATION ROUTINES     ----------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::check_responses()
{
//   for (unsigned int r = 0; r < this->n_resp; ++r)
//     {
//       if (this->name_responses[r] != "current" &&
// 	  this->name_responses[r] != "m_Pt_c" &&
// 	  this->name_responses[r] != "epsilon_V_cat_c" &&
// 	  this->name_responses[r] != "epsilon_N_cat_c" &&
// 	  this->name_responses[r] != "epsilon_S_cat_c" &&
// 	  this->name_responses[r] != "m_Pt_a" &&
// 	  this->name_responses[r] != "epsilon_V_cat_a" &&
// 	  this->name_responses[r] != "epsilon_N_cat_a" &&
// 	  this->name_responses[r] != "epsilon_S_cat_a" &&
// 	  this->name_responses[r] != "t_m_Pt" &&
// 	  this->name_responses[r] != "l_cat_c" &&
// 	  this->name_responses[r] != "l_cat_a")
// 	{
// 	  std::cout <<"Response "<<this->name_responses[r]<<" is not implemented"
// 		    <<"Error in file "<<__FILE__<<"line "<<__LINE__<< std::endl;
// 	  // Throw an exception:
// 	  ExcNotImplemented ();
// 	}
//     }
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::cell_responses (std::vector<double>& resp,
                                          const typename DoFApplication<dim>::CellInfo& info,
                                          const AppFrame::FEVector& /*src*/)
{
    // TODO: ISSUE HERE
    // Info needed during integration. Note this might be a problem when the geometry is supplied from input file
    l_channel = grid->L_channel_c();
    l_land = grid->L_land_c();
    double time_k = sorption_source_terms.get_time_constant();
    
    const unsigned int last_iteration = 0;
    // -- Find out what material is the cell made of, i.e. MEA layer)
    const unsigned int material_id = info.dof_active_cell->material_id();
    
    // Use quadratue associated with the first element for the integration (this is arbitrary):
    const FEValuesBase<dim>& feValues = info.fe(0);
    // get integration points
    const unsigned int n_quad = feValues.n_quadrature_points;
    
    // Storing solution indices
    const unsigned int xO2_index = system_management.solution_name_to_index("oxygen_molar_fraction");
    const unsigned int xH2O_index = system_management.solution_name_to_index("water_molar_fraction");
    const unsigned int phiM_index = system_management.solution_name_to_index("protonic_electrical_potential");
    const unsigned int phiS_index = system_management.solution_name_to_index("electronic_electrical_potential");
    const unsigned int lambda_index = system_management.solution_name_to_index("membrane_water_content");
    
    for (unsigned int r = 0; r < this->n_resp; ++r)
    {
        if ( (this->name_responses[r] == "cathode_current") && (CCL->belongs_to_material(material_id)) )
        {
            // Creating solution variable vector to passed to catalyst layer classes
            std::vector< FuelCellShop::SolutionVariable > solution_variables;
            solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][xO2_index], oxygen_molar_fraction) );
            solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][phiM_index], protonic_electrical_potential) );
            solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][phiS_index], electronic_electrical_potential) );
            
            std::vector<double> response_vec(n_quad, 0.0);
            CCL->set_solution(solution_variables);
            CCL->current_density(response_vec);
            for (unsigned int q = 0; q < n_quad; ++q)
                resp[r] += -(response_vec[q] * feValues.JxW(q)) / (l_channel/2.0 + l_land/2.0);
        }
        
        else if ( (this->name_responses[r] == "anode_current") && (ACL->belongs_to_material(material_id)) )
        {
           
            std::vector<double> x_H2(n_quad, 0.0);
            for(unsigned int q = 0; q < n_quad; ++q)
                x_H2[q] = 1.0 - info.values[last_iteration][xH2O_index][q];
            
            // Creating solution variable vector to passed to catalyst layer classes
            std::vector< FuelCellShop::SolutionVariable > solution_variables;
            solution_variables.push_back( FuelCellShop::SolutionVariable(&x_H2, hydrogen_molar_fraction) );
            solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][phiM_index], protonic_electrical_potential) );
            solution_variables.push_back( FuelCellShop::SolutionVariable(&info.values[last_iteration][phiS_index], electronic_electrical_potential) );
            
            std::vector<double> response_vec(n_quad, 0.0);
            ACL->set_solution(solution_variables);

            ACL->current_density(response_vec);
            for (unsigned int q = 0; q < n_quad; ++q)
                resp[r] += (response_vec[q] * feValues.JxW(q)) / (l_channel/2.0 + l_land/2.0);
        }
        
        else if ( (this->name_responses[r] == "water_cathode") && (CCL->belongs_to_material(material_id)) )
        {
            double rhoDry = CCL->get_electrolyte()->get_density();
            double EW = CCL->get_electrolyte()->get_EW();
            CCL->get_electrolyte()->set_water_molar_fraction( FuelCellShop::SolutionVariable(&info.values[last_iteration][xH2O_index],
                                                                  water_molar_fraction) );
            std::vector<double> lambdaValue( info.values[last_iteration][lambda_index] );
            std::vector<double> lambdaEq(n_quad, 0.0);
            CCL->get_electrolyte()->sorption_isotherm(lambdaEq);
            for (unsigned int q = 0; q < n_quad; ++q)
                resp[r] += ( ( ((time_k*rhoDry)/EW)*(lambdaEq[q] - lambdaValue[q]) ) * feValues.JxW(q) ) / (l_channel/2.0 + l_land/2.0);
        }
        
        else if ( (this->name_responses[r] == "water_anode") && (ACL->belongs_to_material(material_id)) )
        {
            double rhoDry = ACL->get_electrolyte()->get_density();
            double EW = ACL->get_electrolyte()->get_EW();
            ACL->get_electrolyte()->set_water_molar_fraction( FuelCellShop::SolutionVariable(&info.values[last_iteration][xH2O_index],
                                                                  water_molar_fraction) );
            std::vector<double> lambdaValue( info.values[last_iteration][lambda_index] );
            std::vector<double> lambdaEq(n_quad, 0.0);
            ACL->get_electrolyte()->sorption_isotherm(lambdaEq);
            for (unsigned int q = 0; q < n_quad; ++q)
                resp[r] += ( ( ((time_k*rhoDry)/EW)*(lambdaEq[q] - lambdaValue[q]) ) * feValues.JxW(q) ) / (l_channel/2.0 + l_land/2.0);
        }
    }
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::global_responses(std::vector<double>& resp,
						  const AppFrame::FEVector& /*src*/)
{
//   //-- Initialize var
//   double epsilon_S = 0.0;
//   double epsilon_V = 0.0;
//   double epsilon_N = 0.0;
//   
//   for (unsigned int r = 0; r < this->n_resp; ++r)
//     {
//       if (this->name_responses[r] == "epsilon_V_cat_c")
// 	{
// 	  volume_fraction.epsilon_cat_c(epsilon_S, epsilon_V, epsilon_N);
// 	  resp[r] = epsilon_V;
// 	}
//       else if (this->name_responses[r] == "epsilon_N_cat_c")
// 	{
// 	  volume_fraction.epsilon_cat_c(epsilon_S, epsilon_V, epsilon_N);
// 	  resp[r] = epsilon_N;
// 	}
//       else if (this->name_responses[r] == "epsilon_S_cat_c")
// 	{
// 	  volume_fraction.epsilon_cat_c(epsilon_S, epsilon_V, epsilon_N);
// 	  resp[r] = epsilon_S;
// 	}
//       else if (this->name_responses[r] == "m_Pt_c")
// 	{
// 	  resp[r] = CCL.get_V_Pt()/volume_fraction.get_l_cat_c();
// 	}
//       else if (this->name_responses[r] == "epsilon_V_cat_a")
// 	{
// 	  volume_fraction.epsilon_cat_a(epsilon_S, epsilon_V, epsilon_N);
// 	  resp[r] = epsilon_V;
// 	}
//       else if (this->name_responses[r] == "epsilon_N_cat_a")
// 	{
// 	  volume_fraction.epsilon_cat_a(epsilon_S, epsilon_V, epsilon_N);
// 	  resp[r] = epsilon_N;
// 	}
//       else if (this->name_responses[r] == "epsilon_S_cat_a")
// 	{
// 	  volume_fraction.epsilon_cat_a(epsilon_S, epsilon_V, epsilon_N);
// 	  resp[r] = epsilon_S;
// 	}
//       else if (this->name_responses[r] == "m_Pt_a")
// 	{
// 	  resp[r] = volume_fraction.get_m_Pt_a();
// 	}
//       else if (this->name_responses[r] == "t_m_Pt")
// 	{
// 	  resp[r] = volume_fraction.get_m_Pt_a() 
// 	    + volume_fraction.get_m_Pt_c();
// 	}
//       else if (this->name_responses[r] == "l_cat_c")
// 	{
// 	  resp[r] = volume_fraction.get_m_Pt_c()/volume_fraction.get_l_cat_c();
// 	}
//       else if (this->name_responses[r] == "l_cat_a")
// 	{
// 	  resp[r] = volume_fraction.get_m_Pt_a()/volume_fraction.get_l_cat_a();
// 	}
//     }
} 

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::global_dresponses_dl(std::vector<std::vector<double> >& df_dl,
					      const AppFrame::FEVector& )
{
//   //-- Initialize var
//   double Depsilon_S = 0.0;
//   double Depsilon_V = 0.0;
//   double Depsilon_N = 0.0;
//   
//   for (unsigned int i = 0; i<this->name_design_var.size(); ++i)
//     for (unsigned int r = 0; r < this->n_resp; ++r)
//       {
// 	if (this->name_responses[r] == "epsilon_V_cat_c")
// 	  {
// 	    volume_fraction.DepsilonDl_cat_c(this->name_design_var[i], Depsilon_S, Depsilon_V, Depsilon_N);
// 	    df_dl[r][i] = Depsilon_V;
// 	  }
// 	else if (this->name_responses[r] == "epsilon_S_cat_c")
// 	{
// 	  volume_fraction.DepsilonDl_cat_c(this->name_design_var[i], Depsilon_S, Depsilon_V, Depsilon_N);
// 	  df_dl[r][i] = Depsilon_S;
// 	}
// 	else if (this->name_responses[r] == "epsilon_N_cat_c")
// 	{
// 	  volume_fraction.DepsilonDl_cat_c(this->name_design_var[i], Depsilon_S, Depsilon_V, Depsilon_N);
// 	  df_dl[r][i] = Depsilon_N;
// 	}
// 	else if (this->name_responses[r] == "m_Pt_c")
// 	  {
// 	    double derivative = 0.0;
// 	    volume_fraction.get_Dm_Pt_c(this->name_design_var[i], derivative);
// 	    df_dl[r][i] = derivative;
// 	  }
// 	else if (this->name_responses[r] == "epsilon_V_cat_a")
// 	  {
// 	    volume_fraction.DepsilonDl_cat_a(this->name_design_var[i], Depsilon_S, Depsilon_V, Depsilon_N);
// 	    df_dl[r][i] = Depsilon_V;
// 	  }
// 	else if (this->name_responses[r] == "epsilon_S_cat_a")
// 	{
// 	  volume_fraction.DepsilonDl_cat_a(this->name_design_var[i], Depsilon_S, Depsilon_V, Depsilon_N);
// 	  df_dl[r][i] = Depsilon_S;
// 	}
// 	else if (this->name_responses[r] == "epsilon_N_cat_a")
// 	{
// 	  volume_fraction.DepsilonDl_cat_a(this->name_design_var[i], Depsilon_S, Depsilon_V, Depsilon_N);
// 	  df_dl[r][i] = Depsilon_N;
// 	}
// 	else if (this->name_responses[r] == "m_Pt_a")
// 	  {
// 	    double derivative = 0.0;
// 	    volume_fraction.get_Dm_Pt_a(this->name_design_var[i], derivative);
// 	    df_dl[r][i] = derivative;
// 	  }
// 	else if (this->name_responses[r] == "t_m_Pt")
// 	  {
// 	    double derivative_c = 0.0;
// 	    double derivative_a = 0.0;
// 	    volume_fraction.get_Dm_Pt_a(this->name_design_var[i], derivative_a);
// 	    volume_fraction.get_Dm_Pt_c(this->name_design_var[i], derivative_c);
// 	    df_dl[r][i] = derivative_c + derivative_a;
// 	  }
//       }
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::global_dresponses_du(std::vector<AppFrame::FEVector >& df_du,
                                               const AppFrame::FEVector& /*src*/)
{
//   for (unsigned int r = 0; r < this->n_resp; ++r)
//     {
//       if (this->name_responses[r] == "epsilon_V_cat_c")
// 	{
// 	  for (unsigned int i = 0; i<this->element->n_blocks(); ++i)
// 	    for (unsigned int j=0; j<df_du[r].block(i).size(); ++j)
// 	      df_du[r].block(i)(j) = 0;
// 	}
//       else if (this->name_responses[r] == "epsilon_S_cat_c")
// 	{
// 	  for (unsigned int i = 0; i<this->element->n_blocks(); ++i)
// 	    for (unsigned int j=0; j<df_du[r].block(i).size(); ++j)
// 	      df_du[r].block(i)(j) = 0;
// 	}
//       else if (this->name_responses[r] == "epsilon_N_cat_c")
// 	{
// 	  for (unsigned int i = 0; i<this->element->n_blocks(); ++i)
// 	    for (unsigned int j=0; j<df_du[r].block(i).size(); ++j)
// 	      df_du[r].block(i)(j) = 0;
// 	}
//       else if (this->name_responses[r] == "epsilon_V_cat_a" ||
// 	       this->name_responses[r] == "epsilon_S_cat_a" ||
// 	       this->name_responses[r] == "epsilon_N_cat_a" ||
// 	       this->name_responses[r] == "m_Pt_a" ||
// 	       this->name_responses[r] == "t_m_Pt")
// 	{
// 	  for (unsigned int i = 0; i<this->element->n_blocks(); ++i)
// 	    for (unsigned int j=0; j<df_du[r].block(i).size(); ++j)
// 	      df_du[r].block(i)(j) = 0;
// 	}
//     }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//-----------    POSTPROCESSING ROUTINES     --------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
template <int dim>
double 
NAME::AppPemfc<dim>::evaluate (const AppFrame::FEVectors& vectors)
{
    std::vector<double> test(this->n_resp, 0.0);
    this->responses(test, vectors);

    for (unsigned int r = 0; r < this->n_resp; ++r)
        std::cout << "Response " << this->name_responses[r] << " is: :" << test[r] << std::endl;
    
    return 0.0;
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::AppPemfc<dim>::data_out(const std::string &basename, 
                                   const AppFrame::FEVectors &vectors)
{
    this->solution_interpretations.clear();
    this->solution_interpretations.resize(this->element->n_blocks(),
                                        DataComponentInterpretation::component_is_scalar);
    
    unsigned int last_iter = vectors.find_vector("Solution");
    const BlockVector<double>& src = vectors.vector(last_iter);

    // -------------------------------------------------------------------------------------
    // Post processing Functionals -- they "have to be" equal to no_of_blocks
    AppFrame::FEVector functionals;
    functionals.reinit(src);
    
    std::vector<std::string> functional_names;
    
    if (this->element->n_base_elements() == 1)
    {
        AppFrame::FEVector functionals_cell( this->element->n_blocks(),
                                            this->element->base_element(this->element->block_to_base_index(0).first).dofs_per_cell );
        
        // Quadrature with points at the position of the dof of the base_element(0). I can choose any base element I like since the
        //main reason I do this is so that I can use add_data_vector which needs the vector to be of size n_dof.
        const Quadrature<dim> dummy_qr (this->element->base_element(0).get_generalized_support_points());
        std::vector<FEValues<dim>*> fe(this->element->n_base_elements());
        for (unsigned int i=0; i<fe.size(); ++i)
            fe[i] = new FEValues<dim> (*this->mapping, this->element->base_element(0), dummy_qr, 
                                    UpdateFlags(update_values | update_gradients | update_quadrature_points  | update_JxW_values));
            
        const unsigned int dofs_per_cell = this->element->dofs_per_cell;
        
        // Create vector that related degrees of freedom of the cell with dof of the global system (to assemble global matrix)
        std::vector<unsigned int> indices_org (dofs_per_cell);
        std::vector<unsigned int> indices (dofs_per_cell);
        
        // loop over cells
        typename DoFHandler<dim>::active_cell_iterator cell;
        for (cell = this->dof->begin_active(); cell != this->dof->end(); ++cell)
        {
            // initialize the vector that relates local element dof to global dof
            cell->get_dof_indices (indices_org);
            for (unsigned int i=0; i<indices.size(); ++i)
                indices[this->block_info.local_renumbering[i]] = indices_org[i];
            
            //initialize shape function values for the cell
            for (unsigned int i=0; i<fe.size(); ++i)
                fe[i]->reinit ((typename Triangulation<dim>::cell_iterator) cell);
            
            // set material id
            unsigned char materialId = cell->material_id ();
            
            // Since all fe are the same use fe(0) for all elements:
            const FEValuesBase<dim>& fe_values = *fe[0];
            // get no. of quadrature points
            const unsigned int n_quad = fe_values.n_quadrature_points;
            // Create a vector where the values of coefficients at quadrature points are stored:
            std::vector<double> current_density(n_quad,0.0);
            std::vector<double> values_effectiveness(n_quad,0.0);
            std::vector<double> relative_humidity(n_quad,0.0);
            
            // Create a vector to store solution:
            std::vector<std::vector<double> > solution(this->element->n_blocks(), std::vector<double> (n_quad, 0.0));
            //
            for (unsigned int i=0; i<this->element->n_blocks(); ++i)
                fe_values.get_function_values(src, make_slice(indices, this->block_info.local.block_start(i), fe_values.dofs_per_cell), solution[i]);
                                                                            
            // ---- RELATIVE HUMIDITY -----------------------------------------------------
            if (CGDL->belongs_to_material(materialId) || CMPL->belongs_to_material(materialId) || CCL->belongs_to_material(materialId))
                for (unsigned int i=0; i<n_quad; ++i)
                    relative_humidity[i] = (OC.get_pc_Pa()*solution[system_management.solution_name_to_index("water_molar_fraction")][i])/FuelCellShop::Material::Water::satPressure(OC.get_T());
                
            else if (AGDL->belongs_to_material(materialId) || AMPL->belongs_to_material(materialId) || ACL->belongs_to_material(materialId))
                for (unsigned int i=0; i<n_quad; ++i)
                    relative_humidity[i] = (OC.get_pa_Pa()*solution[system_management.solution_name_to_index("water_molar_fraction")][i])/FuelCellShop::Material::Water::satPressure(OC.get_T());
                    
            // ---- Looping over Layers for other Functionals-------------------------------
                    
            if ( CCL->belongs_to_material(materialId) )
            {
                // Creating solution variable vector to passed to catalyst layer classes
                std::vector< FuelCellShop::SolutionVariable > solution_variables;
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("oxygen_molar_fraction")], oxygen_molar_fraction) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("protonic_electrical_potential")], protonic_electrical_potential) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("electronic_electrical_potential")], electronic_electrical_potential) );

                CCL->set_solution(solution_variables);
                CCL->current_density(current_density, values_effectiveness);
            }
                
            else if ( ACL->belongs_to_material(materialId) )
            {
                
                // Creating solution vector to be passed to kinetics classes...
                std::vector<double> x_h2 (n_quad, 0.0);
                for (unsigned int k = 0; k < n_quad; ++k)
                    x_h2[k] = 1.0 - solution[system_management.solution_name_to_index("water_molar_fraction")][k];
                
                // Creating solution variable vector to passed to catalyst layer classes
                std::vector< FuelCellShop::SolutionVariable > solution_variables;
                solution_variables.push_back( FuelCellShop::SolutionVariable(&x_h2, hydrogen_molar_fraction) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("protonic_electrical_potential")], protonic_electrical_potential) );
                solution_variables.push_back( FuelCellShop::SolutionVariable(&solution[system_management.solution_name_to_index("electronic_electrical_potential")], electronic_electrical_potential) );
                
                ACL->set_solution(solution_variables);
                ACL->current_density(current_density, values_effectiveness);
            }
            
            for (unsigned int q=0; q<n_quad; ++q)
            {
                //Note I loop over the indices of the first component which
                functionals_cell.block(0)(q) = current_density[q];
                functionals_cell.block(1)(q) = values_effectiveness[q];
                functionals_cell.block(2)(q) = relative_humidity[q];
                //functionals_cell.block(3)(q) = //to be filled
                //functionals_cell.block(4)(q) = //to be filled
            }
                    
            // Store in the global vector:
            for (unsigned int j=0; j<this->element->dofs_per_cell; ++j)
                functionals(indices[j]) = functionals_cell(j);
        }
        
        // Delete all FEValues. This is necessary in order to clean the subscriptor. Otherwise, when we
        // finish the subscriptor class issues an error:
        for (unsigned int i=0; i<fe.size(); ++i)
            delete fe[i];
    } //end if
    
    // names for source terms
    functional_names.push_back ("current_density");
    functional_names.push_back ("agglomerate_effectiveness");
    functional_names.push_back ("relative_humidity");
    functional_names.push_back ("__nothing__");
    functional_names.push_back ("__nothing__");
    
    this->postprocessing_interpretations.clear();
    this->postprocessing_interpretations.resize( this->element->n_blocks(),
                                                DataComponentInterpretation::component_is_scalar ); 

    // Create string that will be appended to the .vtk filename. This string will contain the name and value of the design variables. 
    std::string filename1;
    
    //loop over all design variables
    for (unsigned int i=0; i<design_var.size(); i++)
    {
        //create object that can print out numbers
        std::ostringstream streamOut;
        //Used to set how many digits after the decimal point are used. Insert the required number into the brackets.
        streamOut.precision(); 
        //If there are fewer digits in the design variable than are requested, the remaining digits can be filled with zeroes.
        //Use std::fixed to force this: streamOut <<  std::fixed << design_var_value[design_var.size()-i-1];
        //This can be used to get paraview to group the vtk files correctly.
        
        //Note that the design variables are printed in reverse, so that file managers will group the vtk files by the last design
        // variable first before moving to the second last etc. 
        streamOut << std::fixed <<  design_var_value[design_var.size()-i-1];
        
        //Append design variable name and value to filename1
        filename1 += "__" + design_var[design_var.size()-i-1] + "_" + streamOut.str();
    }
    //construct the full filename, where basename is given by the adaptive refinement loop and suffix is the file extension (i.e. .vtk)
    const std::string fileName = basename + filename1;
    deallog << "Datafile:" << fileName + this->d_out.default_suffix() << std::endl;
    
    // --- data out ---
    DoFApplication<dim>::data_out(fileName,
                                  src,
                                  system_management.get_solution_names(),
                                  functionals,
                                  functional_names);
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template <int dim>
FuelCell::InitialSolution::AppPemfcIC<dim>::AppPemfcIC(FuelCell::OperatingConditions* OC,
                                                                 boost::shared_ptr< FuelCellShop::Geometry::GridBase<dim> > grid)
  :
  Function<dim> (5),
  OC(OC),
  grid(grid)
{

}

//---------------------------------------------------------------------------
template <int dim>
FuelCell::InitialSolution::AppPemfcIC<dim>::~AppPemfcIC()
{}

//---------------------------------------------------------------------------
template <int dim>
void
FuelCell::InitialSolution::AppPemfcIC<dim>::vector_value (const Point<dim> &p,
                                                               Vector<double> &v) const
{
    // size checking for vectors
    Assert(v.size() == this->n_components,
           ExcDimensionMismatch (v.size(), this->n_components));
    
    double RH_c = OC->get_RH_c();
    double RH_a = OC->get_RH_a();
    double V_cell = std::fabs(OC->get_V());
    double x_wv = OC->get_x_wv();
    double x_o2 = OC->get_x_o2();
    double x_h2 = OC->get_x_h2();
    double OCV = OC->get_OCV();
    
    double l_mem = grid->L_mem();
    double l_cat_a = grid->L_cat_a();
    double l_mpl_a = grid->L_mpl_a();
    double l_gdl_a = grid->L_gdl_a();
    
    double x = p(0);
    double x_norm = (x - l_gdl_a - l_cat_a - l_mpl_a)/l_mem;  
    double delta = -0.0001;
    
    if (x <= l_gdl_a + l_cat_a + l_mpl_a) //Anode
    {
        v(0) = 0.0;//0.5*x_o2; 
        v(1) = 1 - x_h2;  // x_wv_a - Dirichlet B.C. 
        v(2) = delta; // phi_m
        v(3) = 0;        // phi_s - Dirichlet B.C.
        if(RH_a <= 0.5)
            v(4) = 4.0;
        else
            v(4) = 7.0; //lamda
    }
    else if (x > l_gdl_a + l_cat_a + l_mpl_a && x < l_gdl_a + l_cat_a + l_mpl_a + l_mem) //membrane
    {
        v(0) = 0.0;//.5*x_o2;
        v(1) = x_wv;
        v(2) = ((V_cell-OCV)/2.0 - (delta))*x_norm + delta; 
        v(3) = V_cell * x_norm;
        if(RH_a <= 0.5 && RH_c <= 0.5)
            v(4) = 4.0;
        else if(RH_a <= 0.5 && RH_c > 0.5)
            v(4) = 3.0*x_norm + 4.0;
        else if(RH_a > 0.5 && RH_c <= 0.5)
            v(4) = -3.0*x_norm + 7.0;
        else
            v(4) = 7.0;
    }
    else if (x >= l_gdl_a + l_cat_a + l_mpl_a + l_mem) //Cathode
    {
        v(0) = x_o2;  //Dirichlet B.C.
        v(1) = x_wv;  //Dirichlet B.C.
        v(2) = (V_cell-OCV)/2.0;  // phi_m - i.e. 1/2 of the losses through the CCL
        v(3) = V_cell;   // phi_s - Dirichlet B.C.
        // lamda
        if (RH_c <= 0.5)
            v(4) = 4.0;
        else
            v(4) = 7.0;
    }
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations. 
template class NAME::AppPemfc<deal_II_dimension>;
template class FuelCell::InitialSolution::AppPemfcIC<deal_II_dimension>;
