//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: electron_transport_equation.cc
//    - Description: Implementation of electron_transport_equation.h. This
//      class is used to setup the matrix and rhs for electron transport based
//      on Ohms law assuming a constant conductivity
//    - Developers: M. Secanell and Madhur Bhaiya
//    - $Id: electron_transport_equation.cc 1354 2013-08-17 00:01:22Z secanell $
//
//---------------------------------------------------------------------------

#include "equations/electron_transport_equation.h"

namespace NAME = FuelCellShop::Equation;

// ---             ---
// --- Constructors ---
// ---             ---
template<int dim>
NAME::ElectronTransportEquation<dim>::ElectronTransportEquation(FuelCell::SystemManagement& system_management)
:
EquationBase<dim>(system_management),
name_base_variable("electronic_electrical_potential")
{
    this->equation_name = "Electron Transport Equation";
    deallog << "->FuelCellShop::Equation::ElectronTransportEquation"<<std::endl;
    
    //----Initializing VariableInfo Structs------------------------------------------
    //----Setting indices_exist to false --------------------------------------------
    phi_s.indices_exist = false;
    
    //-- Counter to store if generic_data and constant_data need to be initialized.
    this->counter.resize(3, true);
}
// ---            ---
// --- Destructor ---
// ---            ---

template<int dim>
NAME::ElectronTransportEquation<dim>::~ElectronTransportEquation()
{ }

// ---                    ---
// --- declare_parameters ---
// ---                    ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::declare_parameters(ParameterHandler& param) const
{
    param.enter_subsection("Equations");
    {
        param.enter_subsection("Electron Transport Equation");
        {
            param.enter_subsection("Boundary conditions");
            {
                param.declare_entry("Dirichlet Boundary Conditions",
                                "",
                                Patterns::Map( Patterns::Integer(0), Patterns::Double() ),
                                "A comma-separated list of Dirichlet boundary_id(s) with prescribed values of Potential [V]."
                                "\n"
                                "Correct format is of a map, given as ''id1: value1, id2: value2, id3: value3, ...'' where "
                                "each id must be an unsigned integer, and each value can be a double corresponding to potential value.");

                param.declare_entry("Constant Electron Current Flux Boundary Conditions",
                                "",
                                Patterns::Map( Patterns::Integer(0), Patterns::Double() ),
                                "A comma-separated list of boundaries with Galvanostatic b.c., with prescribed values of electronic current fluxes [A/cm^2]."
                                "\n"
                                "Correct format is of a map, given as ''id1: value1, id2: value2, id3: value3, ...'' where "
                                "each id must be an unsigned integer, and each value can be a double (positive for electronic current flux leaving out and negative for current coming in).");
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

// ---            ---
// --- initialize ---
// ---            ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::initialize(ParameterHandler& param)
{
    param.enter_subsection("Equations");
    {
        param.enter_subsection("Electron Transport Equation");
        {
            param.enter_subsection("Boundary conditions");
            {
                dirichlet_bdry_map = FcstUtilities::string_to_map<unsigned int, double>( Utilities::split_string_list( param.get("Dirichlet Boundary Conditions") ) );
                electron_current_flux_map = FcstUtilities::string_to_map<unsigned int, double>( Utilities::split_string_list( param.get("Constant Electron Current Flux Boundary Conditions")) );
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
    
    this->make_internal_cell_couplings();
    
    this->make_boundary_types();
}

// ---                      ---
// --- assemble_cell_matrix ---
// ---                      ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::assemble_cell_matrix(AppFrame::MatrixVector&                                 cell_matrices,
                                                           const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                           FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{   
    if ( this->counter[0] )
    {
        this->make_assemblers_generic_constant_data();
        this->counter[0] = false;
    }
    
    if ( this->counter[1] )
    {
        this->make_assemblers_cell_constant_data(cell_info);
        this->counter[1] = false;
    }
    
    this->make_assemblers_cell_variable_data(cell_info, layer);
    
    
    //-------- Looping over Quadrature points ----------------------------
    for (unsigned int q = 0; q < this->n_q_points_cell; ++q)
    {
        //---------------LOOP over i -----------------------------------------------------------------
        for (unsigned int i=0; i < (cell_info.fe(phi_s.fetype_index)).dofs_per_cell; ++i)
        {
            //--------------LOOP(s) over j-------------------------------------------------------------
            
            //-----------Assembling Matrix for terms corresponding to "phi_s" BLOCK------------------------
            for (unsigned int j=0; j < (cell_info.fe(phi_s.fetype_index)).dofs_per_cell; ++j)
            {
                cell_matrices[phi_s.block_index].matrix(i,j) += (this->JxW_cell[q] * (grad_phi_phiS_cell[q][i] * sigmaSeff_cell * grad_phi_phiS_cell[q][j]));
            }
        }
    } 
}

// ---                        ---
// --- assemble_cell_residual ---
// ---                        ---
template<int dim>
void
NAME::ElectronTransportEquation<dim>::assemble_cell_residual(AppFrame::FEVector&                                     cell_residual,
                                                             const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                             FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{   
    if ( this->counter[0] )
    {
        this->make_assemblers_generic_constant_data();
        this->counter[0] = false;
    }
    
    if ( this->counter[1] )
    {
        this->make_assemblers_cell_constant_data(cell_info);
        this->counter[1] = false;
    }
    
    this->make_assemblers_cell_variable_data(cell_info, layer);
    
    
    for (unsigned int q=0; q < this->n_q_points_cell; ++q)
    {
        for (unsigned int i=0; i < (cell_info.fe(phi_s.fetype_index)).dofs_per_cell; ++i)
        {
            cell_residual.block(phi_s.solution_index)(i) += ( this->JxW_cell[q] * (grad_phi_phiS_cell[q][i] * sigmaSeff_cell * cell_info.gradients[last_iter_cell][phi_s.solution_index][q]) );
        }
    }
    
}


// ---                      ---
// --- assemble_bdry_matrix ---
// ---                      ---
template<int dim>
void
NAME::ElectronTransportEquation<dim>::assemble_bdry_matrix(AppFrame::MatrixVector&                                 bdry_matrices,
                                                           const typename AppFrame::DoFApplication<dim>::FaceInfo& bdry_info,
                                                           FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{
    // Since effective electron conductivity is not a function of any of the solution variables.
    // Hence, the contribution to bdry_matrix due to an electron current flux is "zero".
}

// ---                        ---
// --- assemble_bdry_residual ---
// ---                        ---
template<int dim>
void
NAME::ElectronTransportEquation<dim>::assemble_bdry_residual(AppFrame::FEVector&                                     bdry_residual,
                                                           const typename AppFrame::DoFApplication<dim>::FaceInfo& bdry_info,
                                                           FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{
    if ( this->counter[0] )
    {
        this->make_assemblers_generic_constant_data();
        this->counter[0] = false;
    }
    
    if ( this->counter[2] )
    {
        this->make_assemblers_bdry_constant_data(bdry_info);
        this->counter[2] = false;
    }
    
    //------ Constant Electron Current Flux boundaries --------------------------------------------------------------
    std::map<unsigned int, double>::const_iterator iter = electron_current_flux_map.find( bdry_info.dof_face->boundary_indicator() );
    if (iter != electron_current_flux_map.end() )
    {
        const double& electron_current_value = iter->second;
        
        this->make_assemblers_bdry_variable_data(bdry_info, layer);
        
        //-------- Looping over Quadrature points ----------------------------
        for (unsigned int q = 0; q < this->n_q_points_bdry; ++q)
        {
            for (unsigned int i = 0; i < (bdry_info.fe(phi_s.fetype_index)).dofs_per_cell; ++i)
            {
                bdry_residual.block(phi_s.solution_index)(i) += ( this->JxW_bdry[q] * phi_phiS_bdry[q][i] * (-1.0) * electron_current_value );
            }
        }
    }
}

// ---                              ---
// --- make_internal_cell_couplings ---
// ---                              ---
template<int dim>
void
NAME::ElectronTransportEquation<dim>::make_internal_cell_couplings()
{
    AssertThrow(this->system_management->solution_in_userlist(name_base_variable), VariableShouldExistForEquation(name_base_variable, this->equation_name) );
    AssertThrow(this->system_management->equation_name_to_index(this->equation_name) == this->system_management->solution_name_to_index(name_base_variable), 
              IndexDoNotMatch(name_base_variable, this->equation_name) );
    
    std::map< std::string, DoFTools::Coupling > tmp;
    
    std::vector< std::string> sol_names = this->system_management->get_solution_names();

    for (unsigned int i=0; i < sol_names.size(); ++i)
    {
        if (sol_names[i] == name_base_variable)
            tmp[name_base_variable] = DoFTools::always;
        
        else
            tmp[sol_names[i]] = DoFTools::none;
    }
    
    this->internal_cell_couplings[this->equation_name] = tmp;
}

// ---                     ---
// --- make_boundary_types ---
// ---                     ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::make_boundary_types()
{
    unsigned int i = 0;
    for ( std::map<unsigned int,double>::const_iterator iter = dirichlet_bdry_map.begin(); iter != dirichlet_bdry_map.end(); ++iter)
    {
        i++;
        BoundaryType temp_dirich;
        
        std::ostringstream streamOut;
        streamOut << i;
        std::string temp_name = "Dirichlet_" + streamOut.str();
        
        temp_dirich.boundary_name = temp_name;
        temp_dirich.boundary_id = iter->first;
        temp_dirich.boundary_condition = "Dirichlet";
        
        this->boundary_types.push_back(temp_dirich);
    }
}

// ---                     ---
// --- print_equation_info ---
// ---                     ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::print_equation_info() const
{
    deallog << std::endl;
    deallog << "-------------------------------------------------------------------------------" << std::endl;
    deallog << std::endl;
    deallog << "INTERNAL CELL COUPLINGS FOR \" Electron Transport Equation \":" << std::endl;
    deallog << std::endl;
    
    couplings_map::const_iterator iter;
    
    for( iter = this->internal_cell_couplings.begin(); iter != this->internal_cell_couplings.end(); ++iter )
    {
        deallog << "\"" << iter->first << "\"" << ":";
        
        std::map<std::string, DoFTools::Coupling> int_map = iter->second;
        std::map<std::string, DoFTools::Coupling>::const_iterator int_iter;
        
        for( int_iter = int_map.begin(); int_iter != int_map.end(); ++int_iter )
            deallog << "\"" << int_iter->first << "\"" << " is coupled as " << int_iter->second << std::endl;
        
        deallog << std::endl;
    }
    
    deallog << std::endl;
    deallog << "-------------------------------------------------------------------------------" << std::endl;
    deallog << std::endl;    
}

/////////////////////////////////////////////////////
/////////////////////////////////////////////////////
// LOCAL CG FEM BASED ASSEMBLERS - make_ FUNCTIONS //
/////////////////////////////////////////////////////
/////////////////////////////////////////////////////
       
// ---                                       ---
// --- make_assemblers_generic_constant_data ---
// ---                                       ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::make_assemblers_generic_constant_data()
{
    //-----------Filling VariableInfo structures------------------------------------------
    phi_s.solution_index = this->system_management->solution_name_to_index("electronic_electrical_potential");
    phi_s.block_index = this->system_management->matrix_block_index(this->equation_name, "electronic_electrical_potential");
    phi_s.fetype_index = this->system_management->block_info->base_element[phi_s.solution_index];
    phi_s.indices_exist = true;   
}

// ---                                    ---
// --- make_assemblers_cell_constant_data ---
// ---                                    ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::make_assemblers_cell_constant_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info)
{
    Assert( phi_s.indices_exist, ExcMessage("make_assemblers_generic_constant_data function not called before.") );
    
    this->n_q_points_cell = (cell_info.fe(phi_s.fetype_index)).n_quadrature_points;
    last_iter_cell = cell_info.global_data->find_vector("Newton iterate");
    
    //-------------Allocation------------------------------------------
    grad_phi_phiS_cell.resize( this->n_q_points_cell, std::vector< Tensor<1,dim> >( (cell_info.fe(phi_s.fetype_index)).dofs_per_cell ) );
    
    //-----------------------------------------------------------------
    this->JxW_cell.resize(this->n_q_points_cell);
}

// ---                                    ---
// --- make_assemblers_bdry_constant_data ---
// ---                                    ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::make_assemblers_bdry_constant_data(const typename AppFrame::DoFApplication<dim>::FaceInfo& bdry_info)
{
    Assert( phi_s.indices_exist, ExcMessage("make_assemblers_generic_constant_data function not called before.") );
    
    this->n_q_points_bdry = (bdry_info.fe(phi_s.fetype_index)).n_quadrature_points;
    last_iter_bdry = bdry_info.global_data->find_vector("Newton iterate");
    
    //-------------Allocation------------------------------------------
    phi_phiS_bdry.resize( this->n_q_points_bdry, std::vector<double>( (bdry_info.fe(phi_s.fetype_index)).dofs_per_cell ) );
    
    //-----------------------------------------------------------------
    this->JxW_bdry.resize(this->n_q_points_bdry);
}

// ---                                    ---
// --- make_assemblers_cell_variable_data ---
// ---                                    ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::make_assemblers_cell_variable_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                                         FuelCellShop::Layer::BaseLayer<dim>* const layer)
{
    Assert( this->n_q_points_cell != 0, ExcMessage("make_assemblers_cell_constant_data function not called before.") );
    
    //---------------Effective Transport Properties---------------------------------------------------------------
    // ----- type infos -------------
    const std::type_info& GasDiffusionLayer = typeid(FuelCellShop::Layer::GasDiffusionLayer<dim>);
    const std::type_info& MicroPorousLayer  = typeid(FuelCellShop::Layer::MicroPorousLayer<dim>);
    const std::type_info& CatalystLayer = typeid(FuelCellShop::Layer::CatalystLayer<dim>);
    const std::type_info& SolidLayer = typeid(FuelCellShop::Layer::SolidLayer<dim>);
    
    const std::type_info& base_layer = layer->get_base_type();  
    
    // ----- Clearing and initializing the containers -----------------------------------------
    // ----- All containers intialized to zero by default -------------------------------------
    
    sigmaSeff_cell.clear(); // This automatically initializes it to zero Tensor.
    
    // ----- dynamic cast and filling the containers -----------------
    try
    {
        if (base_layer == GasDiffusionLayer)
        {
            FuelCellShop::Layer::GasDiffusionLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::GasDiffusionLayer<dim>* >(layer);
            ptr->effective_electron_conductivity(sigmaSeff_cell);     
        }
        
        else if (base_layer == MicroPorousLayer)
        {
            FuelCellShop::Layer::MicroPorousLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::MicroPorousLayer<dim>* >(layer);
            ptr->effective_electron_conductivity(sigmaSeff_cell);         
        }
        
        else if (base_layer == CatalystLayer)
        {
            FuelCellShop::Layer::CatalystLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::CatalystLayer<dim>* >(layer);
            ptr->effective_electron_conductivity(sigmaSeff_cell);
        }
        
        else if (base_layer == SolidLayer)
        {    
            FuelCellShop::Layer::SolidLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::SolidLayer<dim>* >(layer);
            ptr->effective_electron_conductivity(sigmaSeff_cell);            
        }
        else
            AssertThrow( false, ExcNotImplemented() );
    }
    catch(const std::bad_cast& e)
    {
        const std::type_info& info = typeid(*layer);  
        std::cerr << "Object of type "<<info.name()<<" not implemented"<< std::endl;
        std::cerr << e.what() << std::endl;
    }
    
    //---------------------------------------------------------------------------------------------------------------
    //------------Looping over quadrature points in the cell --------------------------------------------------------
    for (unsigned int q = 0; q < this->n_q_points_cell; ++q)
    {
        //-------JxW----------
        this->JxW_cell[q] = (cell_info.fe(phi_s.fetype_index)).JxW(q);
        
        //------ Filling shape functions etc ----------------------------------------------------------------------
        //------ This avoids recalculating shape functions etc for efficiency -------------------------------------
        for (unsigned int k=0; k < (cell_info.fe(phi_s.fetype_index)).dofs_per_cell; ++k)
        {
            grad_phi_phiS_cell[q][k] = (cell_info.fe(phi_s.fetype_index)).shape_grad(k,q);
        }
    }
}

// ---                                    ---
// --- make_assemblers_bdry_variable_data ---
// ---                                    ---

template<int dim>
void
NAME::ElectronTransportEquation<dim>::make_assemblers_bdry_variable_data(const typename AppFrame::DoFApplication<dim>::FaceInfo& bdry_info,
                                            FuelCellShop::Layer::BaseLayer<dim>* const layer)
{
    Assert( this->n_q_points_bdry != 0, ExcMessage("make_assemblers_bdry_constant_data function not called before.") );
    //---------------------------------------------------------------------------------------------------------------
    //------------Looping over quadrature points in the cell --------------------------------------------------------
    for (unsigned int q=0; q < this->n_q_points_bdry; ++q)
    {
        //-------JxW & normal_vectors----------------------------------------------------------------------------
        this->JxW_bdry[q] = (bdry_info.fe(phi_s.fetype_index)).JxW(q);
        
        //------ Filling shape functions etc --------------------------------------------------------------------
        //------ This avoids recalculating shape functions etc for efficiency -----------------------------------
        
        for (unsigned int k=0; k < (bdry_info.fe(phi_s.fetype_index)).dofs_per_cell; ++k)
        {
            phi_phiS_bdry[q][k] = (bdry_info.fe(phi_s.fetype_index)).shape_value(k,q);
        }
    }
    
    
}

/////////////////////////////////////////////////////
/////////////////////////////////////////////////////
// CLASS TEST
/////////////////////////////////////////////////////
/////////////////////////////////////////////////////


template<int dim>
void 
NAME::ElectronTransportEquation<dim>::class_test()
{
    
    MeshWorker::BlockInfo block_info;
    Table< 2, DoFTools::Coupling > cell_couplings;
    Table< 2, DoFTools::Coupling > flux_couplings;
    FuelCell::SystemManagement sys(block_info, cell_couplings, flux_couplings);
    deallog<<"Create object ElectronTransportEquation:"<<std::endl;
    FuelCellShop::Equation::ElectronTransportEquation<dim> test(sys);
    test.print_equation_info();
    
}

// ---                         ---
// --- explicit instantiations ---
// ---                         ---

template class NAME::ElectronTransportEquation<deal_II_dimension>;