// ----------------------------------------------------------------------------
//
// FCST: Fuel Cell Simulation Toolbox
//
// Copyright (C) 2006-2013 by Energy Systems Design Laboratory, University of Alberta
//
// This software is distributed under the MIT License
// For more information, see the README file in /doc/LICENSE
//
// - Class: equation_base.cc
// - Description: This is a base class for all available FCST equations
// - Developers: Valentin N. Zingan,    University of Alberta
//               Marc Secanell Gallart, University of Alberta
// - Id: $Id: equation_base.cc 1354 2013-08-17 00:01:22Z secanell $
//
// ----------------------------------------------------------------------------

#include "equation_base.h"

namespace NAME = FuelCellShop::Equation;

       //////////////////////////////////////////////////
       //////////////////////////////////////////////////
       // CONSTRUCTORS, DESTRUCTOR, AND INITIALIZATION //
       //////////////////////////////////////////////////
       //////////////////////////////////////////////////

// ---             ---
// --- Constructor ---
// ---             ---

template<int dim>
NAME::EquationBase<dim>::EquationBase(FuelCell::SystemManagement& system_management)
:
system_management(&system_management)
{ }

// ---            ---
// --- Destructor ---
// ---            ---

template<int dim>
NAME::EquationBase<dim>::~EquationBase()
{ }

       ////////////////
       ////////////////
       // CONVERTERS //
       ////////////////
       ////////////////

// ---                        ---
// --- standard_to_block_wise ---
// ---                        ---

template<int dim>
void
NAME::EquationBase<dim>::standard_to_block_wise(FullMatrix<double>& target) const
{
  const std::vector<unsigned int> local_renumbering( system_management->block_info->local_renumbering );

  FullMatrix<double> tmp(target.m(), target.n());

  for(unsigned int i = 0; i < target.m(); ++i)
    for(unsigned int j = 0; j < target.n(); ++j)
      tmp(local_renumbering[i], local_renumbering[j]) = target(i,j);

  target = tmp;
}

// ---                        ---
// --- standard_to_block_wise ---
// ---                        ---

template<int dim>
void
NAME::EquationBase<dim>::standard_to_block_wise(Vector<double>& target) const
{
  const std::vector<unsigned int> local_renumbering( system_management->block_info->local_renumbering );

  Vector<double> tmp(target.size());

  for(unsigned int i = 0; i < target.size(); ++i)
    tmp(local_renumbering[i]) = target(i);

  target = tmp;
}

// ---                    ---
// --- dealII_to_appframe ---
// ---                    ---

template<int dim>
void
NAME::EquationBase<dim>::dealII_to_appframe(AppFrame::MatrixVector&          dst,
                                            const FullMatrix<double>&        src,
                                            const std::vector<unsigned int>& matrix_block_indices) const
{
  FullMatrix<double> block_wise(src);
  this->standard_to_block_wise(block_wise);

  BlockIndices local = system_management->block_info->local;
  std::vector<unsigned int> sizes(local.size());
  for(unsigned int i = 0; i < sizes.size(); ++i)
    sizes[i] = local.block_size(i);

  unsigned int number = 0;

  unsigned int i_from = 0;
  unsigned int i_to   = sizes[0] - 1;

  for(unsigned int line = 0; line < sizes.size(); ++line)
  {
    unsigned int j_from = 0;
    unsigned int j_to   = sizes[0] - 1;

    for(unsigned int column = 0; column < sizes.size(); ++column)
    {
           // ---

           const std::vector<unsigned int>::const_iterator
             iter = std::find(matrix_block_indices.begin(), matrix_block_indices.end(), number);

           if( iter == matrix_block_indices.end() )
           {
                  if( (*(system_management->cell_couplings))(line,column) != 0 )
                  {
                         if( number != dst.size() - 1 )
                         {
                                number++;
                         }
                  }
           }
           else
           {
                  if( (*(system_management->cell_couplings))(line,column) != 0 )
                  {
                         const unsigned int n_lines   = i_to - i_from + 1;
                         const unsigned int n_columns = j_to - j_from + 1;

                         FullMatrix<double> tmp(n_lines, n_columns);

                         unsigned int line_index = i_from;

                         for(unsigned int i = 0; i < n_lines; ++i)
                         {
                                unsigned int column_index = j_from;

                                for(unsigned int j = 0; j < n_columns; ++j)
                                {
                                       tmp(i,j) = block_wise(line_index, column_index);
                                       column_index++;
                                }

                                line_index++;
                         }

                         dst[number].matrix = tmp;

                         if( number != dst.size() - 1 )
                         {
                                number++;
                         }
                  }
           }

           if( column != sizes.size() - 1 )
           {
                  j_from = j_to + 1;
                  j_to   = j_to + sizes[column+1];
           }

           // ---
    }

    if( line != sizes.size() - 1 )
    {
           i_from = i_to + 1;
           i_to   = i_to + sizes[line+1];
    }
  }
}

// ---                    ---
// --- dealII_to_appframe ---
// ---                    ---

template<int dim>
void
NAME::EquationBase<dim>::dealII_to_appframe(AppFrame::FEVector&              dst,
                                            const Vector<double>&            src,
                                            const std::vector<unsigned int>& residual_indices) const
{
  AssertThrow( dst.size() == src.size() , ExcDimensionMismatch( dst.size() , src.size() ) );

  Vector<double> block_wise(src);
  this->standard_to_block_wise(block_wise);

  BlockIndices local = system_management->block_info->local;
  std::vector<unsigned int> sizes(local.size());
  for(unsigned int i = 0; i < sizes.size(); ++i)
    sizes[i] = local.block_size(i);

  unsigned int number = 0;

  unsigned int i_from = 0;
  unsigned int i_to   = sizes[0] - 1;

  for(unsigned int line = 0; line < sizes.size(); ++line)
  {
         const std::vector<unsigned int>::const_iterator
           iter = std::find(residual_indices.begin(), residual_indices.end(), number);

         if( iter != residual_indices.end() )
         {
                const unsigned int n_lines = i_to - i_from + 1;

                Vector<double> tmp(n_lines);

                unsigned int line_index = i_from;

                for(unsigned int i = 0; i < n_lines; ++i)
                {
                       tmp(i) = block_wise(line_index);
                       line_index++;
                }

                dst.block(line) = tmp;
         }

         if( number != sizes.size() - 1 )
         {
                number++;
         }

         if( line != sizes.size() - 1 )
         {
                i_from = i_to + 1;
                i_to   = i_to + sizes[line+1];
         }
  }
}

       /////////////////////
       /////////////////////
       // MINOR FUNCTIONS //
       /////////////////////
       /////////////////////

// ---                   ---
// --- print_caller_name ---
// ---                   ---

template<int dim>
void
NAME::EquationBase<dim>::print_caller_name(const std::string& caller_name) const
{
  const std::type_info& info = typeid(*this);
  deallog << "Pure function " << caller_name << " called in Class " << info.name() << std::endl;
}

       /////////////////////////////
       /////////////////////////////
       // EXPLICIT INSTANTIATIONS //
       /////////////////////////////
       /////////////////////////////

// ---              ---
// --- EquationBase ---
// ---              ---

template class NAME::EquationBase<deal_II_dimension>;