// ------------------------------------------------------------------
//
// flicks_transport_equation.cc 27-10-2012
//
// Copyright (C) 2012 by Marc Secanell, University of Alberta.
//
// This file is subject to QPL and may not be  distributed
// without copyright and license information. Please refer
// to the file deal.II/doc/license.html for the  text  and
// further information on this license.
//
// ------------------------------------------------------------------

#include "ficks_transport_equation.h"

namespace NAME = FuelCellShop::Equation;

// ---             ---
// --- Constructors ---
// ---             ---
template<int dim>
NAME::FicksTransportEquation<dim>::FicksTransportEquation(FuelCell::SystemManagement& system_management,
							  FuelCellShop::Material::PureGas* solute,
							  FuelCellShop::Material::PureGas* solvent)
:
EquationBase<dim>(system_management),
gas(solute),
solvent(solvent)
{
  std::stringstream s;
  s <<"Ficks Transport Equation - "<<gas->get_name();
  this->name_equation = s.str();
  
  std::stringstream ss;
  ss <<gas->get_name()<<"_molar_fraction";
  this->name_solution = ss.str();
  
  deallog << "->FuelCellShop::Equation::FicksTransportEquation-" << dim << "D"<<std::endl;
}

template<int dim>
NAME::FicksTransportEquation<dim>::FicksTransportEquation(FuelCell::SystemManagement& system_management)
:
EquationBase<dim>(system_management)
{	
	
}
  
// ---            ---
// --- Destructor ---
// ---            ---

template<int dim>
NAME::FicksTransportEquation<dim>::~FicksTransportEquation()
{ }

// ---                    ---
// --- declare_parameters ---
// ---                    ---

template<int dim>
void
NAME::FicksTransportEquation<dim>::declare_parameters(ParameterHandler& param)
{

	    param.enter_subsection("Equations");
	    {
		  
	    }
	    param.leave_subsection();
}

// ---            ---
// --- initialize ---
// ---            ---

template<int dim>
void
NAME::FicksTransportEquation<dim>::initialize(ParameterHandler& param)
{ param.enter_subsection("System management");
  {
    param.enter_subsection("Equations");
    {
	if (this->name_equation.compare("")!=0)
		deallog<<"Default Constructor was used"<<std::endl;
	else
	{
	    std::string name = "Equation 1";
	    std::string eq_name;
	    std::stringstream var_name;
	    this->name_equation = param.get(name.c_str());

	    if (this->name_equation.compare("Ficks Transport Equation - oxygen") == 0)
	    {
		    var_name<<"oxygen_molar_fraction";
		    this->name_solution = var_name.str();
	    }
	    else if(this->name_equation.compare("Ficks Transport Equation - water") == 0)
	    {
		    var_name<<"water_molar_fraction";
		    this->name_solution = var_name.str();
	    }
	}
    }
    param.leave_subsection();
  }
  param.leave_subsection();
  
  this->make_internal_cell_couplings();
}

// ---                      ---
// --- assemble_cell_matrix ---
// ---                      ---

template<int dim>
void
NAME::FicksTransportEquation<dim>::assemble_cell_matrix(AppFrame::MatrixVector&                                 cell_matrices,
                                                        const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                        FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{
    
    // Acquire typeid for the base layer we have:
    const std::type_info& base_layer = layer->get_base_type();  
    
    // -- Choose the correct finite element for the equation
    //deallog<<"Name equation is: "<<this->name_equation<<std::endl;
    unsigned int eq_index = this->system_management->equation_name_to_index(this->name_equation);
    unsigned int fe_type = this->system_management->block_info->base_element[eq_index];
    const FEValuesBase<dim>& fe = cell_info.fe(fe_type);
    
    if (base_layer == typeid(FuelCellShop::Layer::GasDiffusionLayer<dim>))
    {
        try
        {
            // --- fluid properties via dynamic casting ---
            FuelCellShop::Layer::GasDiffusionLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::GasDiffusionLayer<dim>* >(layer);
            if (ptr->belongs_to_material(cell_info.cell->material_id()))
            {
                Table<2, double> D_eff;
                ptr->effective_gas_diffusivity (D_eff);
                int index;
                int index_2;
                ptr->get_gas_index(this->gas, index);
                ptr->get_gas_index(this->solvent, index_2);
                
                AppShop::Matrix::Cell::laplacian(cell_matrices[this->system_management->matrix_block_index(this->name_equation, this->name_solution)].matrix, fe, D_eff(index,index_2));
            }
        }
        catch(const std::bad_cast& e)
        {
            std::cerr << e.what() << std::endl;
            std::cerr << "The object is not of type FuelCellShop::Layer::GasDiffusionLayer<dim>" << std::endl;
        }
    }
    /*
    else if (info == typeid(FuelCellShop::Layer::MicroPorousLayer< dim >))
    {
        try
        {
            // --- fluid properties via dynamic casting ---
            FuelCellShop::Layer::MicroPorousLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::MicroPorousLayer<dim>* >(layer);
            if (ptr->belongs_to_material(cell_info.cell->material_id()))
            {
                std::vector<std::vector<Tensor<2,dim> > > D_eff;
                ptr->effective_gas_diffusivity (D_eff);
                int index;
                int index_2;
                ptr->get_gas_index(this->gas, index);
                ptr->get_gas_index(this->solvent, index_2);
                
                AppShop::Matrix::Cell::laplacian(cell_matrices[this->system_management->matrix_block_index(this->name_equation, this->name_solution)].matrix, fe, D_eff[index][index_2]);
            }
        }
        catch(const std::bad_cast& e)
        {
            std::cerr << e.what() << std::endl;
            std::cerr << "The object is not of type FuelCellShop::Layer::GasDiffusionLayer<dim>" << std::endl;
        }
        
        else if (info == typeid(FuelCellShop::Layer::FuelCellShop::Layer::CatalystLayer< dim >))
        {
            try
            {
                // --- fluid properties via dynamic casting ---
                FuelCellShop::Layer::CatalystLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::CatalystLayer<dim>* >(layer);
                if (ptr->belongs_to_material(cell_info.cell->material_id()))
                {
                    std::vector<std::vector<Tensor<2,dim> > > D_eff;
                    ptr->effective_gas_diffusivity (D_eff);
                    int index;
                    int index_2;
                    ptr->get_gas_index(this->gas, index);
                    ptr->get_gas_index(this->solvent, index_2);
                    
                    AppShop::Matrix::Cell::laplacian(cell_matrices[this->system_management->matrix_block_index(this->name_equation, this->name_solution)].matrix, fe, D_eff[index][index_2]);
                }
            }
            catch(const std::bad_cast& e)
            {
                std::cerr << e.what() << std::endl;
                std::cerr << "The object is not of type FuelCellShop::Layer::GasDiffusionLayer<dim>" << std::endl;
            }
        }
       */ 
        else
        {
            Assert( false , ExcNotImplemented() );
        }
    }
// ---                        ---
// --- assemble_cell_residual ---
// ---                        ---

template<int dim>
void
NAME::FicksTransportEquation<dim>::assemble_cell_residual(AppFrame::FEVector&                                     cell_rhs,
                                                             const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                             FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{
  cell_rhs.block(this->system_management->equation_name_to_index(this->name_equation)) = 0;
         
}

// ---                     ---
// --- print_equation_info ---
// ---                     ---

template<int dim>
void
NAME::FicksTransportEquation<dim>::print_equation_info() const
{
  deallog << std::endl;
  deallog << "Parameters For "<<this->name_equation<<" : "<< std::endl;
  deallog << "Name of the variable "<<this->name_solution<<std::endl;
  couplings_map::const_iterator iter;

  for( iter = internal_cell_couplings.begin(); iter != internal_cell_couplings.end(); ++iter )
  {
    deallog << "\"" << iter->first << "\"" << ":";

    std::map<std::string, DoFTools::Coupling> int_map = iter->second;
    std::map<std::string, DoFTools::Coupling>::const_iterator int_iter;

    for( int_iter = int_map.begin(); int_iter != int_map.end(); ++int_iter )
      deallog << "\"" << int_iter->first << "\"" << " is coupled as " << int_iter->second << std::endl;

    deallog << std::endl;
  }
}

// ---                              ---
// --- make_internal_cell_couplings ---
// ---                              ---
template<int dim>
void
NAME::FicksTransportEquation<dim>::make_internal_cell_couplings()
{
  std::map< std::string, DoFTools::Coupling > tmp;
  
  tmp[this->name_solution] = DoFTools::always;
  internal_cell_couplings[this->name_equation]       = tmp;
  tmp.clear();
}

// ---                     ---
// --- class_test ---
// ---                     ---

template<int dim>
void 
NAME::FicksTransportEquation<dim>::class_test()
{
  FuelCellShop::Material::Hydrogen H2;
  FuelCellShop::Material::Oxygen O2;
  MeshWorker::BlockInfo block_info;
  Table< 2, DoFTools::Coupling > cell_couplings;
  Table< 2, DoFTools::Coupling > flux_couplings;
  FuelCell::SystemManagement sys(block_info, cell_couplings, flux_couplings);
  deallog<<"Create object FicksTransportEquation:"<<std::endl;
  FuelCellShop::Equation::FicksTransportEquation<dim> test(sys,&O2,&H2);
  test.print_equation_info();
  
}


/*
// ---                               ---
// --- draw_out_matrix_block_indices ---
// ---                               ---

template<int dim>
void
NAME::FicksTransportEquation<dim>::draw_out_matrix_block_indices()
{
 std::cout<<"Not yet implemented"<<std::endl;
}

// ---                            ---
// --- draw_out_component_indices ---
// ---                            ---

template<int dim>
void
NAME::FicksTransportEquation<dim>::draw_out_component_indices()
{
 std::cout<<"Not yet implemented"<<std::endl;
}
*/
// ---                         ---
// --- explicit instantiations ---
// ---                         ---

template class NAME::FicksTransportEquation<deal_II_dimension>;