//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: new_ficks_transport_equation.cc
//    - Description: Implementation of Ficks diffusion model.
//    - Developers: M. Secanell and Madhur Bhaiya
//    - $Id: new_ficks_transport_equation.cc 1460 2013-08-30 22:28:49Z madhur $
//
//---------------------------------------------------------------------------

#include "new_ficks_transport_equation.h"

namespace NAME = FuelCellShop::Equation;

// ---             ---
// --- Constructors ---
// ---             ---
template<int dim>
NAME::NewFicksTransportEquation<dim>::NewFicksTransportEquation(FuelCell::SystemManagement& system_management,
                                                                FuelCellShop::Material::PureGas* solute,
                                                                FuelCellShop::Material::PureGas* solvent)
:
EquationBase<dim>(system_management),
gas(solute),
solvent(solvent)
{
    std::stringstream s;
    s <<"Ficks Transport Equation - "<<gas->get_name();
    this->equation_name = s.str();
    
    std::stringstream ss;
    ss <<gas->get_name()<<"_molar_fraction";
    this->name_base_variable = ss.str();
    
    deallog << "->FuelCellShop::Equation::NewFicksTransportEquation-" << dim << "D"<<std::endl;
    
    xi.indices_exist = false;
    t_rev.indices_exist = false;
    
    this->counter.resize(2, true);
}

template<int dim>
NAME::NewFicksTransportEquation<dim>::NewFicksTransportEquation(FuelCell::SystemManagement& system_management)
:
EquationBase<dim>(system_management)
{ 
}

// ---            ---
// --- Destructor ---
// ---            ---

template<int dim>
NAME::NewFicksTransportEquation<dim>::~NewFicksTransportEquation()
{ }

// ---                    ---
// --- declare_parameters ---
// ---                    ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::declare_parameters(ParameterHandler& param)
{
    
    param.enter_subsection("Equations");
    {
        param.enter_subsection(this->equation_name);
        {
            param.enter_subsection("Boundary conditions");
            {
                param.declare_entry("Dirichlet Boundary Indicators",
                                    "",
                                    Patterns::List( Patterns::Integer(0) ),
                                    "Boundary_id(s) corresponding to Dirichlet boundaries (where molar fraction is to be specified), "
                                    "provided by comma-separated list of unsigned integers.");
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

// ---            ---
// --- initialize ---
// ---            ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::initialize(ParameterHandler& param)
{ 
    param.enter_subsection("Equations");
    {
        param.enter_subsection(this->equation_name);
        {
            param.enter_subsection("Boundary conditions");
            {
                dirichlet_bdry_ids = FcstUtilities::string_to_number<unsigned int>( Utilities::split_string_list( param.get("Dirichlet Boundary Indicators") ) );
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
    
    this->make_internal_cell_couplings();
    
    this->make_boundary_types();
}

// ---                      ---
// --- assemble_cell_matrix ---
// ---                      ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::assemble_cell_matrix(AppFrame::MatrixVector& cell_matrices,
                                                           const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                           FuelCellShop::Layer::BaseLayer<dim>* const layer)
{
    if ( this->counter[0] )
    {
        this->make_assemblers_generic_constant_data();
        this->counter[0] = false;
    }
    
    if ( this->counter[1] )
    {
        this->make_assemblers_cell_constant_data(cell_info);
        this->counter[1] = false;
    }
    
    this->make_assemblers_cell_variable_data(cell_info, layer);
    
    //-------- Looping over Quadrature points ----------------------------
    for (unsigned int q = 0; q < this->n_q_points_cell; ++q)
    {
        //---------------LOOP over i -----------------------------------------------------------------
        for (unsigned int i=0; i < (cell_info.fe(xi.fetype_index)).dofs_per_cell; ++i)
        {
            //--------------LOOP(s) over j-------------------------------------------------------------
            
            //-----------Assembling Matrix for terms corresponding to "xi" BLOCK------------------------
            for (unsigned int j=0; j < (cell_info.fe(xi.fetype_index)).dofs_per_cell; ++j)
            {
                cell_matrices[xi.block_index].matrix(i,j) += ( this->JxW_cell[q] * (grad_phi_xi_cell[q][i] * conc_Deff_cell[q] * grad_phi_xi_cell[q][j]) );
            }
        }
    }         
}
// ---                        ---
// --- assemble_cell_residual ---
// ---                        ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::assemble_cell_residual(AppFrame::FEVector&                                     cell_rhs,
                                                             const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                             FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{
  if ( this->counter[0] )
    {
        this->make_assemblers_generic_constant_data();
        this->counter[0] = false;
    }
    
    if ( this->counter[1] )
    {
        this->make_assemblers_cell_constant_data(cell_info);
        this->counter[1] = false;
    }
    
    this->make_assemblers_cell_variable_data(cell_info, layer);
   
    for (unsigned int q=0; q < this->n_q_points_cell; ++q)
    {
        for (unsigned int i=0; i < (cell_info.fe(xi.fetype_index)).dofs_per_cell; ++i)
        {         
            cell_rhs.block(xi.solution_index)(i) += ( this->JxW_cell[q] * (grad_phi_xi_cell[q][i] * conc_Deff_cell[q] * cell_info.gradients[last_iter_cell][xi.solution_index][q]) );
        }
    }
    
}

// ---                     ---
// --- print_equation_info ---
// ---                     ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::print_equation_info() const
{
  deallog << std::endl;
  deallog << "Parameters For "<<this->equation_name<<" : "<< std::endl;
  deallog << "Name of the variable "<<this->name_base_variable<<std::endl;
  couplings_map::const_iterator iter;

  for( iter = this->internal_cell_couplings.begin(); iter != this->internal_cell_couplings.end(); ++iter )
  {
    deallog << "\"" << iter->first << "\"" << ":";

    std::map<std::string, DoFTools::Coupling> int_map = iter->second;
    std::map<std::string, DoFTools::Coupling>::const_iterator int_iter;

    for( int_iter = int_map.begin(); int_iter != int_map.end(); ++int_iter )
      deallog << "\"" << int_iter->first << "\"" << " is coupled as " << int_iter->second << std::endl;

    deallog << std::endl;
  }
}

// ---                              ---
// --- make_internal_cell_couplings ---
// ---                              ---
template<int dim>
void
NAME::NewFicksTransportEquation<dim>::make_internal_cell_couplings()
{
    AssertThrow(this->system_management->solution_in_userlist(name_base_variable), VariableShouldExistForEquation(name_base_variable, this->equation_name) );
    AssertThrow(this->system_management->equation_name_to_index(this->equation_name) == this->system_management->solution_name_to_index(name_base_variable), 
                IndexDoNotMatch(name_base_variable, this->equation_name) );
    
    std::map< std::string, DoFTools::Coupling > tmp;
    
    std::vector< std::string> sol_names = this->system_management->get_solution_names();
    
    for (unsigned int i = 0; i < sol_names.size(); ++i)
    {
        if (sol_names[i] == name_base_variable)
            tmp[name_base_variable] = DoFTools::always;
        
        else if (sol_names[i] == "temperature_of_REV")
            tmp["temperature_of_REV"] = DoFTools::always;
        
        else
            tmp[sol_names[i]] = DoFTools::none;
    }

    this->internal_cell_couplings[this->equation_name] = tmp;
}

// ---                     ---
// --- make_boundary_types ---
// ---                     ---
template<int dim>
void
NAME::NewFicksTransportEquation<dim>::make_boundary_types()
{
    for (unsigned int index = 1; index <= dirichlet_bdry_ids.size(); ++index)
    {
        BoundaryType temp_dirich;
        
        std::ostringstream streamOut;
        streamOut << index;
        std::string temp_name = "Dirichlet_" + streamOut.str();
        
        temp_dirich.boundary_name = temp_name;
        temp_dirich.boundary_id = dirichlet_bdry_ids[index-1];
        temp_dirich.boundary_condition = "Dirichlet";
        
        this->boundary_types.push_back(temp_dirich);
    }
}

        /////////////////////////////////////////////////////
        /////////////////////////////////////////////////////
        // LOCAL CG FEM BASED ASSEMBLERS - make_ FUNCTIONS //
        /////////////////////////////////////////////////////
        /////////////////////////////////////////////////////
       
// ---                                       ---
// --- make_assemblers_generic_constant_data ---
// ---                                       ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::make_assemblers_generic_constant_data()
{
    //-----------Filling VariableInfo structures------------------------------------------
    xi.solution_index = this->system_management->solution_name_to_index(name_base_variable);
    xi.block_index = this->system_management->matrix_block_index(this->equation_name, name_base_variable);
    xi.fetype_index = this->system_management->block_info->base_element[xi.solution_index];
    xi.indices_exist = true;
    
    //----------temperature_of_solid_phase----------------------------------------------------
    if ( this->system_management->solution_in_userlist("temperature_of_REV") )
    {
        t_rev.solution_index = this->system_management->solution_name_to_index("temperature_of_REV");
        t_rev.block_index = this->system_management->matrix_block_index(this->equation_name, "temperature_of_REV");
        t_rev.fetype_index = this->system_management->block_info->base_element[t_rev.solution_index];
        t_rev.indices_exist = true;
    }
}


// ---                                    ---
// --- make_assemblers_cell_constant_data ---
// ---                                    ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::make_assemblers_cell_constant_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info)
{
    Assert( xi.indices_exist, ExcMessage("make_assemblers_generic_constant_data function not called before.") );
    
    this->n_q_points_cell = (cell_info.fe(xi.fetype_index)).n_quadrature_points;
    this->last_iter_cell = cell_info.global_data->find_vector("Newton iterate");
    
    //-------------Allocation------------------------------------------
    phi_xi_cell.resize( this->n_q_points_cell, std::vector<double>( (cell_info.fe(xi.fetype_index)).dofs_per_cell ) );
    grad_phi_xi_cell.resize( this->n_q_points_cell, std::vector< Tensor<1,dim> >( (cell_info.fe(xi.fetype_index)).dofs_per_cell ) );
    
    if (t_rev.indices_exist)
        grad_phi_T_cell.resize( this->n_q_points_cell, std::vector< Tensor<1,dim> >( (cell_info.fe(t_rev.fetype_index)).dofs_per_cell ) );

    this->JxW_cell.resize(this->n_q_points_cell);
    
    //-----------------------------------------------------------------
    conc_Deff_cell.resize(this->n_q_points_cell);
}

// ---                                    ---
// --- make_assemblers_cell_variable_data ---
// ---                                    ---

template<int dim>
void
NAME::NewFicksTransportEquation<dim>::make_assemblers_cell_variable_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                            FuelCellShop::Layer::BaseLayer<dim>* const layer)
{
    Assert( this->n_q_points_cell != 0, ExcMessage("make_assemblers_cell_constant_data function not called before.") );
    
    //---------------Effective Transport Properties---------------------------------------------------------------
    // ----- type infos -------------
    const std::type_info& GasDiffusionLayer = typeid(FuelCellShop::Layer::GasDiffusionLayer<dim>);
    const std::type_info& MicroPorousLayer  = typeid(FuelCellShop::Layer::MicroPorousLayer<dim>);
    const std::type_info& CatalystLayer = typeid(FuelCellShop::Layer::CatalystLayer<dim>);
    
    const std::type_info& base_layer = layer->get_base_type();    
    
    // Creating some internal variables -- to be used commonly by all the dynamically casted layers
    Table< 2, Tensor<2,dim> > Deff_iso;
    double p, T;
    int index_gas, index_solvent;
   
    // ----- dynamic cast and filling the containers -----------------                
    try
    {
        if (base_layer == GasDiffusionLayer)
        {
            FuelCellShop::Layer::GasDiffusionLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::GasDiffusionLayer<dim>* >(layer);

            ptr->get_T_and_p(T,p);
            double concentration = ((p*Units::convert(1.,Units::ATM_to_PA))/(Constants::R()*T))*Units::convert(1.,Units::PER_C_UNIT3, Units::PER_UNIT3);       // mol/cm^3
            ptr->effective_gas_diffusivity(Deff_iso);               // m^2/s
            ptr->get_gas_index(this->gas, index_gas);
            ptr->get_gas_index(this->solvent, index_solvent);
            for (unsigned int q=0; q<this->n_q_points_cell; ++q)
                conc_Deff_cell[q] = concentration * Deff_iso(index_gas,index_solvent)*Units::convert(1.,Units::C_UNIT2, Units::UNIT2);                         // mol/(cm-s)
        }
        
        else if (base_layer == MicroPorousLayer)
        {
            FuelCellShop::Layer::MicroPorousLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::MicroPorousLayer<dim>* >(layer);

            ptr->get_T_and_p(T,p);
            double concentration = ((p*Units::convert(1.,Units::ATM_to_PA))/(Constants::R()*T))*Units::convert(1.,Units::PER_C_UNIT3, Units::PER_UNIT3);       // mol/cm^3
            ptr->effective_gas_diffusivity(Deff_iso);               // m^2/s
            ptr->get_gas_index(this->gas, index_gas);
            ptr->get_gas_index(this->solvent, index_solvent);
            for (unsigned int q=0; q<this->n_q_points_cell; ++q)
                conc_Deff_cell[q] = concentration * Deff_iso(index_gas,index_solvent)*Units::convert(1.,Units::C_UNIT2, Units::UNIT2);                         // mol/(cm-s)
        }
        
        else if (base_layer == CatalystLayer)
        {
            FuelCellShop::Layer::CatalystLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::CatalystLayer<dim>* >(layer);

            ptr->get_T_and_p(T,p);
            double concentration = ((p*Units::convert(1.,Units::ATM_to_PA))/(Constants::R()*T))*Units::convert(1.,Units::PER_C_UNIT3, Units::PER_UNIT3);       // mol/cm^3
            ptr->effective_gas_diffusivity(Deff_iso);               // m^2/s
            ptr->get_gas_index(this->gas, index_gas);
            ptr->get_gas_index(this->solvent, index_solvent);
            for (unsigned int q=0; q<this->n_q_points_cell; ++q)
                conc_Deff_cell[q] = concentration * Deff_iso(index_gas,index_solvent)*Units::convert(1.,Units::C_UNIT2, Units::UNIT2);                         // mol/(cm-s)

        }
        else
            AssertThrow( false, ExcNotImplemented() );
    }
    catch(const std::bad_cast& e)
    {
        const std::type_info& info = typeid(*layer);  
        std::cerr << "Object of type "<<info.name()<<" not implemented"<< std::endl;
        std::cerr << e.what() << std::endl;
    }
    
    //---------------------------------------------------------------------------------------------------------------
    //------------Looping over quadrature points in the cell --------------------------------------------------------
    for (unsigned int q = 0; q < this->n_q_points_cell; ++q)
    {
        //-------JxW----------
        this->JxW_cell[q] = (cell_info.fe(xi.fetype_index)).JxW(q);
        
        //------ Filling shape functions etc ----------------------------------------------------------------------
        //------ This avoids recalculating shape functions etc for efficiency -------------------------------------
        for (unsigned int k=0; k < (cell_info.fe(xi.fetype_index)).dofs_per_cell; ++k)
        {
            phi_xi_cell[q][k] = (cell_info.fe(xi.fetype_index)).shape_value(k,q);
            grad_phi_xi_cell[q][k] = (cell_info.fe(xi.fetype_index)).shape_grad(k,q);
        }
        
        if (t_rev.indices_exist)
            for (unsigned int k=0; k < (cell_info.fe(t_rev.fetype_index)).dofs_per_cell; ++k)
                grad_phi_T_cell[q][k] = (cell_info.fe(t_rev.fetype_index)).shape_grad(k,q);
    }
}
// ---                     ---
// --- class_test ---
// ---                     ---

template<int dim>
void 
NAME::NewFicksTransportEquation<dim>::class_test()
{
  FuelCellShop::Material::Hydrogen H2;
  FuelCellShop::Material::Oxygen O2;
  MeshWorker::BlockInfo block_info;
  Table< 2, DoFTools::Coupling > cell_couplings;
  Table< 2, DoFTools::Coupling > flux_couplings;
  FuelCell::SystemManagement sys(block_info, cell_couplings, flux_couplings);
  deallog<<"Create object NewFicksTransportEquation:"<<std::endl;
  FuelCellShop::Equation::NewFicksTransportEquation<dim> test(sys,&O2,&H2);
  test.print_equation_info();
  
}

// ---                         ---
// --- explicit instantiations ---
// ---                         ---

template class NAME::NewFicksTransportEquation<deal_II_dimension>;