//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: reaction_source_terms.cc
//    - Description: This class is used to assemble cell matrix and cell residual for reaction
//      source terms in the catalyst layer for various equation classes.
//    - Developers: Madhur Bhaiya and M. Secanell
//    - $Id: reaction_source_terms.cc 1452 2013-08-30 21:49:59Z madhur $
//
//---------------------------------------------------------------------------

#include "equations/reaction_source_terms.h"

namespace NAME = FuelCellShop::Equation;

// ---             ---
// --- Constructor ---
// ---             ---

template<int dim>
NAME::ReactionSourceTerms<dim>::ReactionSourceTerms(FuelCell::SystemManagement& system_management)
:
NAME::EquationBase<dim>(system_management)
{
    deallog << "Creating " << "FuelCellShop::Equation::ReactionSourceTerms-" << dim << "D object" << std::endl;
    
    anode_kinetics = NULL;
    cathode_kinetics = NULL;
    
    //----Initializing VariableInfo Structs------------------------------------------
    //----Setting indices_exist to false --------------------------------------------
    x_oxygen.indices_exist = false;
    x_water.indices_exist = false;
    phi_s.indices_exist = false;
    phi_m.indices_exist = false;
    t_rev.indices_exist = false;
    
    this->counter.resize(2, true);
}

// ---            ---
// --- Destructor ---
// ---            ---

template<int dim>
NAME::ReactionSourceTerms<dim>::~ReactionSourceTerms()
{}

// ---                    ---
// --- declare_parameters ---
// ---                    ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::declare_parameters(ParameterHandler& param) const
{}

// ---            ---
// --- initialize ---
// ---            ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::initialize(ParameterHandler& param)
{   
    Assert( cathode_kinetics!=NULL || anode_kinetics!=NULL, ExcMessage("Atleast one of either cathode/anode kinetics should be set before ReactionSourceTerms::initialize.") );
    
    //-------Assertion checks for "Required" solution variables for Cathode/Anode kinetics --------------------
    AssertThrow( this->system_management->solution_in_userlist("electronic_electrical_potential"), VariableNotFoundForKinetics("Cathode/Anode", "electronic_electrical_potential") );
    AssertThrow( this->system_management->solution_in_userlist("protonic_electrical_potential"), VariableNotFoundForKinetics("Cathode/Anode", "protonic_electrical_potential") );
    
    if ( cathode_kinetics != NULL )
        AssertThrow( this->system_management->solution_in_userlist("oxygen_molar_fraction"), VariableNotFoundForKinetics("Cathode", "oxygen_molar_fraction") );
    
    if ( anode_kinetics != NULL )
        AssertThrow( this->system_management->solution_in_userlist("water_molar_fraction"), VariableNotFoundForKinetics("Anode", "water_molar_fraction") );	
    
    //----Setting cell_matrix and cell_residual counters to FALSE -------------------
    cell_matrix_counter = false;
    cell_residual_counter = false;
}

// ---                      ---
// --- assemble_cell_matrix ---
// ---                      ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::assemble_cell_matrix(AppFrame::MatrixVector&                                 cell_matrices,
                                                     const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                     FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{
    if ( this->counter[0] )
    {
        this->make_assemblers_generic_constant_data();
        this->counter[0] = false;
    }
    
    if ( this->counter[1] )
    {
        this->make_assemblers_cell_constant_data(cell_info);
        this->counter[1] = false;
    }
    
    cell_matrix_counter = true;
    cell_residual_counter = false;
    this->make_assemblers_cell_variable_data(cell_info, layer);
    
    // --- Matrix assembly for equations ----------------------------------------------------
    assemble_matrix_for_equation(cell_matrices, cell_info, "Proton Transport Equation", cell_info.fe(phi_m.fetype_index), phi_phiM_cell, factor_protontranseq_cell);
    assemble_matrix_for_equation(cell_matrices, cell_info, "Electron Transport Equation", cell_info.fe(phi_s.fetype_index), phi_phiS_cell, factor_electrontranseq_cell);
    if ( x_oxygen.indices_exist )
        assemble_matrix_for_equation(cell_matrices, cell_info, "Ficks Transport Equation - oxygen", cell_info.fe(x_oxygen.fetype_index), phi_xOxygen_cell, factor_oxygentranseq_cell);
    if ( x_water.indices_exist )
        assemble_matrix_for_equation(cell_matrices, cell_info, "Ficks Transport Equation - water", cell_info.fe(x_water.fetype_index), phi_xWater_cell, factor_watertranseq_cell);
}

// ---                        ---
// --- assemble_cell_residual ---
// ---                        ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::assemble_cell_residual(AppFrame::FEVector&                                     cell_residual,
                                                       const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                       FuelCellShop::Layer::BaseLayer<dim>* const              layer)
{
    if ( this->counter[0] )
    {
        this->make_assemblers_generic_constant_data();
        this->counter[0] = false;
    }
    
    if ( this->counter[1] )
    {
        this->make_assemblers_cell_constant_data(cell_info);
        this->counter[1] = false;
    }
    
    cell_residual_counter = true;
    cell_matrix_counter = false;
    this->make_assemblers_cell_variable_data(cell_info, layer);
    
    for (unsigned int q=0; q < this->n_q_points_cell; ++q)
    {
        // ---- Proton Transport Equation ------------------------------
        for (unsigned int i=0; i < (cell_info.fe(phi_m.fetype_index)).dofs_per_cell; ++i)
            cell_residual.block(phi_m.solution_index)(i) += ( this->JxW_cell[q] * factor_protontranseq_cell * current_cell[q] * phi_phiM_cell[q][i] );
        
        // ---- Electron Transport Equation ------------------------------
        for (unsigned int i=0; i < (cell_info.fe(phi_s.fetype_index)).dofs_per_cell; ++i)
            cell_residual.block(phi_s.solution_index)(i) += ( this->JxW_cell[q] * factor_electrontranseq_cell * current_cell[q] * phi_phiS_cell[q][i] );
        
        // ---- Ficks Transport Equation - oxygen ------------------------
        if ( x_oxygen.indices_exist )
            for (unsigned int i=0; i < (cell_info.fe(x_oxygen.fetype_index)).dofs_per_cell; ++i)
                cell_residual.block(x_oxygen.solution_index)(i) += ( this->JxW_cell[q] * factor_oxygentranseq_cell * current_cell[q] * phi_xOxygen_cell[q][i] );
        
        // ---- Ficks Transport Equation - water ------------------------
        if ( x_water.indices_exist )
            for (unsigned int i=0; i < (cell_info.fe(x_water.fetype_index)).dofs_per_cell; ++i)
                cell_residual.block(x_water.solution_index)(i) += ( this->JxW_cell[q] * factor_watertranseq_cell * current_cell[q] * phi_xWater_cell[q][i] );
            
    } // End Loop Over Quadrature Points
}

// ---                              ---
// --- adjust_internal_cell_couplings ---
// ---                              ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::adjust_internal_cell_couplings(std::vector< couplings_map >& equation_map) const
{
    Assert( equation_map.size() != 0, ExcMessage("Vector size should be greater than zero in ReactionSourceTerms::adjust_internal_cell_couplings.") );
    
    for (unsigned int i=0; i<equation_map.size(); ++i)
    {
        for ( couplings_map::iterator iter = equation_map[i].begin(); iter != equation_map[i].end(); ++iter )
        {
            if ( (iter->first == "Proton Transport Equation") || (iter->first == "Electron Transport Equation") )
            {
                (iter->second)["electronic_electrical_potential"] = DoFTools::always;
                (iter->second)["protonic_electrical_potential"] = DoFTools::always;
                
                if ( this->system_management->solution_in_userlist("temperature_of_REV") )
                    (iter->second)["temperature_of_REV"] = DoFTools::always;
                
                if (cathode_kinetics != NULL)
                    (iter->second)["oxygen_molar_fraction"] = DoFTools::always;
                
                if (anode_kinetics != NULL)
                    (iter->second)["water_molar_fraction"] = DoFTools::always;
            }
            
            else if ( iter->first == "Ficks Transport Equation - oxygen" )
            {
                (iter->second)["oxygen_molar_fraction"] = DoFTools::always;
                (iter->second)["electronic_electrical_potential"] = DoFTools::always;
                (iter->second)["protonic_electrical_potential"] = DoFTools::always;
                
                if (anode_kinetics != NULL)
                    (iter->second)["water_molar_fraction"] = DoFTools::always;
                
                if ( this->system_management->solution_in_userlist("temperature_of_REV") )
                    (iter->second)["temperature_of_REV"] = DoFTools::always;
            }
            
            else if ( iter->first == "Ficks Transport Equation - water" )
            {
                (iter->second)["water_molar_fraction"] = DoFTools::always;
                (iter->second)["electronic_electrical_potential"] = DoFTools::always;
                (iter->second)["protonic_electrical_potential"] = DoFTools::always;
                
                if (cathode_kinetics != NULL)
                    (iter->second)["oxygen_molar_fraction"] = DoFTools::always;
                
                if ( this->system_management->solution_in_userlist("temperature_of_REV") )
                    (iter->second)["temperature_of_REV"] = DoFTools::always;
            }
        }
    }
    
}

// ---                     ---
// --- print_equation_info ---
// ---                     ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::print_equation_info() const
{
    deallog << std::endl;
    deallog << "-------------------------------------------------------------------------------" << std::endl;
    deallog << std::endl;
    deallog << "No parameters in Reaction Source Terms class." << std::endl;
    deallog << std::endl;
    deallog << "-------------------------------------------------------------------------------" << std::endl;
    deallog << std::endl;
}


/////////////////////////////////////////////////////
/////////////////////////////////////////////////////
// LOCAL CG FEM BASED ASSEMBLERS - make_ FUNCTIONS //
/////////////////////////////////////////////////////
/////////////////////////////////////////////////////

// ---                                       ---
// --- make_assemblers_generic_constant_data ---
// ---                                       ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::make_assemblers_generic_constant_data()
{
    //--------Faradays constant --------------------------------------------------------------------------
    F = Constants::F();
    
    //--------Block indices can't be filled here, as they depend on what equation we are in -------------------------------------------
    //--------While doing cell_matrix assembly, developers need to be wary of the fact that block_indices are still not filled yet-----
    
    //-----------Filling VariableInfo structures----------------------------------------------------------
    //-----------electronic_electrical_potential----------------------------------------------------------
    phi_s.solution_index = this->system_management->solution_name_to_index("electronic_electrical_potential");
    phi_s.fetype_index = this->system_management->block_info->base_element[phi_s.solution_index];
    phi_s.indices_exist = true;
    
    //----------protonic_electrical_potential-------------------------------------------------------------
    phi_m.solution_index = this->system_management->solution_name_to_index("protonic_electrical_potential");
    phi_m.fetype_index = this->system_management->block_info->base_element[phi_m.solution_index];
    phi_m.indices_exist = true;
    
    if ( cathode_kinetics != NULL ) // It indirectly checks whether oxygen_molar_fraction is in user-defined list or not.
    {
        //----------oxygen_molar_fraction-------------------------------------------------------------
        x_oxygen.solution_index = this->system_management->solution_name_to_index("oxygen_molar_fraction");
        x_oxygen.fetype_index = this->system_management->block_info->base_element[x_oxygen.solution_index];
        x_oxygen.indices_exist = true;
    }
    
    if ( anode_kinetics != NULL) // It indirectly checks whether water_molar_fraction is in user-defined list or not.
    {
        //----------water_molar_fraction--------------------------------------------------------------
        x_water.solution_index = this->system_management->solution_name_to_index("water_molar_fraction");
        x_water.fetype_index = this->system_management->block_info->base_element[x_water.solution_index];
        x_water.indices_exist = true;
    }
}

// ---                                    ---
// --- make_assemblers_cell_constant_data ---
// ---                                    ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::make_assemblers_cell_constant_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info)
{
    Assert( (phi_s.indices_exist && phi_m.indices_exist), ExcMessage("make_assemblers_generic_constant_data function not called before.") );
    
    this->n_q_points_cell = (cell_info.fe(phi_s.fetype_index)).n_quadrature_points;
    last_iter_cell = cell_info.global_data->find_vector("Newton iterate");
    
    //-------Allocation------------------------------------------------------------------------
    // ----- All containers intialized to zero by default -------------------------------------
    phi_phiS_cell.resize( this->n_q_points_cell, std::vector<double>( (cell_info.fe(phi_s.fetype_index)).dofs_per_cell ) );
    phi_phiM_cell.resize( this->n_q_points_cell, std::vector<double>( (cell_info.fe(phi_m.fetype_index)).dofs_per_cell ) );
    
    current_cell.resize( this->n_q_points_cell );
    
    if (t_rev.indices_exist)
    {
        phi_T_cell.resize( this->n_q_points_cell, std::vector<double>( (cell_info.fe(t_rev.fetype_index)).dofs_per_cell ) );
    }
    
    if (x_oxygen.indices_exist)
        phi_xOxygen_cell.resize( this->n_q_points_cell, std::vector<double>( (cell_info.fe(x_oxygen.fetype_index)).dofs_per_cell ) );
    
    if (x_water.indices_exist)
        phi_xWater_cell.resize( this->n_q_points_cell, std::vector<double>( (cell_info.fe(x_water.fetype_index)).dofs_per_cell ) );
    
    //-----------------------------------------------------------------
    this->JxW_cell.resize(this->n_q_points_cell);
}

// ---                                    ---
// --- make_assemblers_cell_variable_data ---
// ---                                    ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::make_assemblers_cell_variable_data(const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                                   FuelCellShop::Layer::BaseLayer<dim>* const layer)
{
    Assert( this->n_q_points_cell != 0, ExcMessage("make_assemblers_cell_constant_data function not called before.") );
    Assert( cell_matrix_counter != cell_residual_counter, ExcMessage("cell_matrix_counter should not match with cell_residual_counter, code has some bugs.") );
    
    // ----- type infos -------------
    const std::type_info& CatalystLayer = typeid(FuelCellShop::Layer::CatalystLayer<dim>);
    
    const std::type_info& base_layer = layer->get_base_type();
    
    // ----- dynamic cast and filling the containers -----------------
    try
    {
        if ( base_layer == CatalystLayer )
        {
            FuelCellShop::Layer::CatalystLayer<dim>* ptr = dynamic_cast< FuelCellShop::Layer::CatalystLayer<dim>* >(layer);
            active_area_cell = ptr->get_active_area_Pt();       //--- Storing active area (cm^2/cm^3) for the catalyst layer ---------------

            std::vector<FuelCellShop::SolutionVariable> solution_variables;
            std::vector<VariableNames> deriv_flags;

            solution_variables.push_back( FuelCellShop::SolutionVariable(&cell_info.values[last_iter_cell][phi_s.solution_index], electronic_electrical_potential) );
            deriv_flags.push_back(electronic_electrical_potential);
            
            solution_variables.push_back( FuelCellShop::SolutionVariable(&cell_info.values[last_iter_cell][phi_m.solution_index], protonic_electrical_potential) );
            deriv_flags.push_back(protonic_electrical_potential);

            if (ptr->get_kinetics() == cathode_kinetics)
            {
                solution_variables.push_back( FuelCellShop::SolutionVariable(&cell_info.values[last_iter_cell][x_oxygen.solution_index], oxygen_molar_fraction) );
                deriv_flags.push_back(oxygen_molar_fraction);
                
                //----- Filling in factors for the source terms--------------------------
                factor_protontranseq_cell = 1.0;
                factor_electrontranseq_cell = (-1.0);
                factor_oxygentranseq_cell = 1.0/(4.0*F);
                factor_watertranseq_cell = (-1.0)/(2.0*F);
            }
            else if (ptr->get_kinetics() == anode_kinetics)
            {
                //---- Converting x_water into x_hydrogen ---------------------------
                std::vector<double> x_hydrogen( (cell_info.values[last_iter_cell][x_water.solution_index]).size() );
                for (unsigned int i=0; i < x_hydrogen.size(); ++i)
                    x_hydrogen[i] = 1.0 - (cell_info.values[last_iter_cell][x_water.solution_index][i]);
                
                solution_variables.push_back( FuelCellShop::SolutionVariable(x_hydrogen, hydrogen_molar_fraction) );
                deriv_flags.push_back(hydrogen_molar_fraction);
                
                //----- Filling in factors for the source terms--------------------------
                factor_protontranseq_cell = (-1.0);
                factor_electrontranseq_cell = 1.0;
                factor_oxygentranseq_cell = 0.0;
                factor_watertranseq_cell = 0.0;
            }
            else
                AssertThrow(false, ExcMessage("Anode/Cathode kinetics are not set in the catalyst layer in the initialize method of the application.") );
            
            //------ Setting solution in the catalyst layer -----------------------------------------------
            ptr->set_solution(solution_variables);
            ptr->current_density( current_cell );
            
            //-------- Filling derivatives of source terms for cell_matrix assembly -----------------------
            if (cell_matrix_counter)
            {
                std::map< VariableNames, std::vector<double> > Dcurrent;
                ptr->set_derivative_flags(deriv_flags);
                ptr->derivative_current_density(Dcurrent);
                
                dcurrent_dphiS_cell = Dcurrent.at(electronic_electrical_potential);
                dcurrent_dphiM_cell = Dcurrent.at(protonic_electrical_potential);
                
                if (ptr->get_kinetics() == cathode_kinetics)
                {
                    dcurrent_dxOxygen_cell = Dcurrent.at(oxygen_molar_fraction);
                    dcurrent_dxWater_cell.assign(this->n_q_points_cell, 0.);
                }
                else if (ptr->get_kinetics() == anode_kinetics)
                {
                    dcurrent_dxOxygen_cell.assign(this->n_q_points_cell, 0.);
                    dcurrent_dxWater_cell.resize(this->n_q_points_cell, 0.);
                    for (unsigned int k=0; k < Dcurrent.at(hydrogen_molar_fraction).size(); ++k)
                        dcurrent_dxWater_cell[k] = Dcurrent.at(hydrogen_molar_fraction)[k] * (-1.0);            // Accounting for the fact that Dj/DxH2 = -Dj/DxH2O
                }
            }
        }
        
        else
            AssertThrow( false, ExcNotImplemented() );
        
    }
    
    catch(const std::bad_cast& e)
    {
        const std::type_info& info = typeid(*layer);  
        std::cerr << "Object of type " << info.name() << " not implemented" << std::endl;
        std::cerr << e.what() << std::endl;
    }
    
    //---------------------------------------------------------------------------------------------------------------
    //------------Looping over quadrature points in the cell --------------------------------------------------------
    for (unsigned int q = 0; q < this->n_q_points_cell; ++q)
    {
        //-------JxW----------
        this->JxW_cell[q] = (cell_info.fe(phi_s.fetype_index)).JxW(q);
        
        //------ Filling shape functions etc ----------------------------------------------------------------------
        //------ This avoids recalculating shape functions etc for efficiency -------------------------------------
        
        for (unsigned int k=0; k < (cell_info.fe(phi_s.fetype_index)).dofs_per_cell; ++k)
            phi_phiS_cell[q][k] = (cell_info.fe(phi_s.fetype_index)).shape_value(k,q);
        
        for (unsigned int k=0; k < (cell_info.fe(phi_m.fetype_index)).dofs_per_cell; ++k)
            phi_phiM_cell[q][k] = (cell_info.fe(phi_m.fetype_index)).shape_value(k,q);
        
        //------- Checking based on boolean flags for other fe elements--------------------------------------------
        if ( x_oxygen.indices_exist )
            for (unsigned int k=0; k < (cell_info.fe(x_oxygen.fetype_index)).dofs_per_cell; ++k)
                phi_xOxygen_cell[q][k] = (cell_info.fe(x_oxygen.fetype_index)).shape_value(k,q);
            
        if ( x_water.indices_exist )
            for (unsigned int k=0; k < (cell_info.fe(x_water.fetype_index)).dofs_per_cell; ++k)
                phi_xWater_cell[q][k] = (cell_info.fe(x_water.fetype_index)).shape_value(k,q);
                
        if ( t_rev.indices_exist )
            for (unsigned int k=0; k < (cell_info.fe(t_rev.fetype_index)).dofs_per_cell; ++k)
                phi_T_cell[q][k] = (cell_info.fe(t_rev.fetype_index)).shape_value(k,q);
    }
    
}

// ---                              ---
// --- assemble_matrix_for_equation ---
// ---                              ---

template<int dim>
void
NAME::ReactionSourceTerms<dim>::assemble_matrix_for_equation(AppFrame::MatrixVector& cell_matrices,
                                                            const typename AppFrame::DoFApplication<dim>::CellInfo& cell_info,
                                                            const std::string& eq_name,
                                                            const FEValuesBase<dim>& test_fe,
                                                            const std::vector< std::vector<double> >& test_shape_functions,
                                                            const double& sourceterm_factor)
{
    Assert( ((cell_matrix_counter == true) && (cell_residual_counter == false)), ExcInternalError() );
    Assert( this->n_q_points_cell != 0, ExcMessage("make_assemblers_cell_constant_data function not called before.") );
    Assert( ((eq_name == "Proton Transport Equation") || (eq_name == "Electron Transport Equation") || 
            (eq_name == "Ficks Transport Equation - oxygen") || (eq_name == "Ficks Transport Equation - water")), ExcNotImplemented() );
    
    // --- Filling block indices --------------------------------------------
    phi_s.block_index = this->system_management->matrix_block_index(eq_name,"electronic_electrical_potential");
    phi_m.block_index = this->system_management->matrix_block_index(eq_name,"protonic_electrical_potential");
    if (x_oxygen.indices_exist)
        x_oxygen.block_index = this->system_management->matrix_block_index(eq_name,"oxygen_molar_fraction");
    if (x_water.indices_exist)
        x_water.block_index = this->system_management->matrix_block_index(eq_name,"water_molar_fraction");
    if (t_rev.indices_exist)
        t_rev.block_index = this->system_management->matrix_block_index(eq_name,"temperature_of_REV");
    
    std::vector<double> *dsource_dphiS, *dsource_dphiM, *dsource_dxOxygen, *dsource_dxWater, *dsource_dT;
    dsource_dphiS = &dcurrent_dphiS_cell;
    dsource_dphiM = &dcurrent_dphiM_cell;
    dsource_dxOxygen = &dcurrent_dxOxygen_cell;
    dsource_dxWater = &dcurrent_dxWater_cell;
    dsource_dT = &dcurrent_dT_cell;

    
    //-------- Looping over Quadrature points ----------------------------
    for (unsigned int q = 0; q < this->n_q_points_cell; ++q)
    {
        //---------------LOOP over i -----------------------------------------------------------------
        for (unsigned int i=0; i < test_fe.dofs_per_cell; ++i)
        {
            //--------------LOOP(s) over j-------------------------------------------------------------
            for (unsigned int j=0; j < (cell_info.fe(phi_s.fetype_index)).dofs_per_cell; ++j)
                cell_matrices[phi_s.block_index].matrix(i,j) += ( this->JxW_cell[q] * test_shape_functions[q][i] * phi_phiS_cell[q][j] * 
                sourceterm_factor * dsource_dphiS->at(q) );
            
            for (unsigned int j=0; j < (cell_info.fe(phi_m.fetype_index)).dofs_per_cell; ++j)
                cell_matrices[phi_m.block_index].matrix(i,j) += ( this->JxW_cell[q] * test_shape_functions[q][i] * phi_phiM_cell[q][j] * 
                sourceterm_factor * dsource_dphiM->at(q) );
            
            if ( x_oxygen.indices_exist )
                for (unsigned int j=0; j < (cell_info.fe(x_oxygen.fetype_index)).dofs_per_cell; ++j)
                    cell_matrices[x_oxygen.block_index].matrix(i,j) += ( this->JxW_cell[q] * test_shape_functions[q][i] * phi_xOxygen_cell[q][j] * 
                    sourceterm_factor * dsource_dxOxygen->at(q) );
                
            if ( x_water.indices_exist )
                for (unsigned int j=0; j < (cell_info.fe(x_water.fetype_index)).dofs_per_cell; ++j)
                    cell_matrices[x_water.block_index].matrix(i,j) += ( this->JxW_cell[q] * test_shape_functions[q][i] * phi_xWater_cell[q][j] * 
                    sourceterm_factor * dsource_dxWater->at(q) );
            
            if ( t_rev.indices_exist )
                for (unsigned int j=0; j < (cell_info.fe(t_rev.fetype_index)).dofs_per_cell; ++j)
                    cell_matrices[t_rev.block_index].matrix(i,j) += ( this->JxW_cell[q] * test_shape_functions[q][i] * phi_T_cell[q][j] * 
                    sourceterm_factor * dsource_dT->at(q) );
        } // End Loop over "i"
    } // End Loop over "q"
}

// ---                           ---
// ---  EXPLICIT INSTANTIATIONS  ---
// ---                           ---

template class NAME::ReactionSourceTerms<deal_II_dimension>;