//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: agglomerate_CL.cc
//    - Description: Class characterizing the catalyst layer and defining effective properties
//    - Developers: M. Secanell, Peter Dobson
//    - $Id: agglomerate_CL.cc 1466 2013-08-30 22:46:21Z wardlawp $
//
//---------------------------------------------------------------------------

#include "agglomerate_CL.h"
//#include <chrono>
#ifdef _OPENMP
#include <omp.h>
#define PARALLEL 1
#else
#define omp_get_thread_num() 0
#define PARALLEL 1
#endif

namespace NAME = FuelCellShop::Layer;

template<int dim>
const std::string NAME::AgglomerateCL<dim>::concrete_name("AgglomerateCL");

template<int dim>
NAME::AgglomerateCL<dim> const* NAME::AgglomerateCL<dim>::PROTOTYPE =
        new NAME::AgglomerateCL<dim>();

//---------------------------------------------------------------------------
template<int dim>
NAME::AgglomerateCL<dim>::AgglomerateCL() :
NAME::ConventionalCL<dim>() 
{

    this->get_mapFactory()->insert(
            std::pair<std::string, FuelCellShop::Layer::CatalystLayer<dim>*>(
                    concrete_name, this));

    aggs_initialized = false;
    IC_x_R = 0.0;
    IC_phi_m = 0.0;
    IC_phi_s = 0.0;
    IC_lambda = 0.0;
    IC_radius = 0.0;
    IC_film_thickness = 0.0;
    IC_porosity = 0.0;

}

//---------------------------------------------------------------------------
template<int dim>
NAME::AgglomerateCL<dim>::AgglomerateCL(std::string cl_section_name) :
NAME::ConventionalCL<dim>(cl_section_name) 
{
    aggs_initialized = false;
    IC_x_R = 0.0;
    IC_phi_m = 0.0;
    IC_phi_s = 0.0;
    IC_lambda = 0.0;
    IC_radius = 0.0;
    IC_film_thickness = 0.0;
    IC_porosity = 0.0;

}

//---------------------------------------------------------------------------
template<int dim>
NAME::AgglomerateCL<dim>::~AgglomerateCL() {
    aggs.clear();
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::declare_parameters(
        const std::string& cl_section_name, ParameterHandler &param) const {

    NAME::ConventionalCL<dim>::declare_parameters(cl_section_name, param);

    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection(cl_section_name);
        {
            param.enter_subsection(this->concrete_name);
            {
                //-- Agglomerate Structure
                param.declare_entry("Agglomerate type", "spherical",
                        Patterns::Selection("spherical|cylindrical"),
                        "Shape of the agglomerate");
                param.declare_entry("Agglomerate composition", "ionomer",
                        Patterns::Selection("ionomer"),
                        "Compositon of the agglomerate core");
                param.declare_entry("Agglomerate solver", "analytical",
                        Patterns::Selection("analytical"),
                        "Method to solve the agglomerate problem");
                param.declare_entry("Average current in cell", "false",
                        Patterns::Bool(),
                        "Decide whether to take the average current density in the cell");
                param.declare_entry("Radius of the agglomerate [nm]", "200",
                        Patterns::Double());
                param.declare_entry("Agglomerate porosity", "0.25",
                        Patterns::Double());
                param.declare_entry("Agglomerate Loading Profile", "1",
                        Patterns::List(Patterns::Double()));
                param.declare_entry("Conductivity factor", "1.0",
                        Patterns::Double());
                param.declare_entry("Initial condition tolerance factor",
                        "0.05", Patterns::Double(0.0005, 0.5));
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::set_parameters(
        const std::vector<std::string>& name_dvar,
        const std::vector<double>& value_dvar,
        const std::string& cl_section_name, ParameterHandler &param) const {

    NAME::ConventionalCL<dim>::set_parameters(name_dvar, value_dvar,
            cl_section_name, param);

    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection(cl_section_name);
        {
            param.enter_subsection(this->concrete_name);
            {
                for (unsigned int i = 0; i < name_dvar.size(); ++i) {
                    if (this->name == "Cathode catalyst layer") {
                        if (name_dvar[i] == "r_agg_c")
                            param.set("Radius of the agglomerate [nm]",
                                    value_dvar[i]);
                        else if (name_dvar[i] == "r_agg")
                            param.set("Radius of the agglomerate [nm]",
                                    value_dvar[i]);
                        else if (name_dvar[i] == "epsilon_agg_c")
                            param.set("Agglomerate porosity", value_dvar[i]);
                        else if (name_dvar[i] == "epsilon_agg")
                            param.set("Agglomerate porosity", value_dvar[i]);
                    } else if (this->name == "Anode catalyst layer") {
                        if (name_dvar[i] == "r_agg_a")
                            param.set("Radius of the agglomerate [nm]",
                                    value_dvar[i]);
                        else if (name_dvar[i] == "r_agg")
                            param.set("Radius of the agglomerate [nm]",
                                    value_dvar[i]);
                        else if (name_dvar[i] == "epsilon_agg_a")
                            param.set("Agglomerate porosity", value_dvar[i]);
                        else if (name_dvar[i] == "epsilon_agg")
                            param.set("Agglomerate porosity", value_dvar[i]);
                    }
                }
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::initialize(ParameterHandler &param) {
    NAME::ConventionalCL<dim>::initialize(param);

    pi = Constants::Pi();


    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection(this->name);
        {
            param.enter_subsection(this->concrete_name);
            {
                type_agg = param.get("Agglomerate type");
                comp_agg = param.get("Agglomerate composition");
                solver_agg = param.get("Agglomerate solver");
                average_cell_current = param.get_bool(
                        "Average current in cell");
                r_agg = param.get_double("Radius of the agglomerate [nm]");
                //thickness_agg = param.get_double("Thickness electrolyte film");
                epsilon_agg = param.get_double("Agglomerate porosity");
                agglomerateLoadingWeigths = FcstUtilities::string_to_number<double>
                    ( Utilities::split_string_list( param.get("Agglomerate Loading Profile")));
                cond_factor = param.get_double("Conductivity factor");
                IC_tolerance = param.get_double(
                        "Initial condition tolerance factor");
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();

    // Initialize other quantities
    water_agg = (comp_agg == "water");
    n_agg = compute_n();
    thickness_agg = compute_thickness_agg();
    compute_void_fraction();



}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::initialize_aggs() {

    // Check to see that the problem is set properly in the parameter file
    if (solver_agg == "analytical") {
        if (water_agg) {
            deallog
            << "Cannot solve the water filled agglomerate problem analytically."
            << "Please modify the parameter file" << std::endl;
            abort();
        }

        std::vector<VariableNames> names;
        if(this->reactant == oxygen_molar_fraction)
            names.push_back(oxygen_concentration);
        else {
            deallog << "Agglomerate CL cannot solve for this reactant type!" << std::endl;
            abort();

        }

        std::map<VariableNames, double> gamma_map;

        this->catalyst->reaction_order( names,  gamma_map);
        double reaction_order = gamma_map[names[0]];
        if (reaction_order < 0.999999 || reaction_order > 1.000001) {
            deallog
            << "Cannot solve the agglomerate problem analytically with a reaction order other than 1."
            << "Please modify the parameter file" << std::endl;
            abort();
        }
    }
    //Initialize aggregates
    aggs.push_back(get_agg());
    aggs.push_back(get_agg());
    aggs.push_back(get_agg());
    aggs_initialized = true;
}

//---------------------------------------------------------------------------
template<int dim>
boost::shared_ptr<FuelCellShop::MicroScale::AgglomerateBase>
NAME::AgglomerateCL<dim>::get_agg() {
    // Declare an agglomerate object and choose its type.
    boost::shared_ptr<FuelCellShop::MicroScale::AgglomerateBase> _agg;
    choose_agglomerate(_agg);

    double Av_agg = this->Av / (1.0 - this->epsilon_V);
    _agg->set_structure(this->catalyst_support.get(), this->catalyst.get(),
            this->electrolyte.get(), Av_agg, r_agg, thickness_agg, epsilon_agg);
    _agg->set_kinetics(this->kinetics.get());
    _agg->set_T(this->constant_solutions.at(temperature_of_REV));
    _agg->set_P(this->constant_solutions.at(total_pressure));
    _agg->setLoadingWeigths(agglomerateLoadingWeigths);
    _agg->set_conductivity_factor(cond_factor);

#ifdef _OPENMP
    _agg->deep_copy_pointers();
#endif

    return _agg;
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::current_density(std::vector<double>& current) {
    std::vector<double> effectiveness(current.size());

    //call the current density function for the class
    current_density(current, effectiveness);
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::current_density(std::vector<double>& current,
        std::vector<double>& Er) {

    // Declare an agglomerate object and choose its type.
    if (!aggs_initialized)
        initialize_aggs();

    current.clear();
    current.resize(this->solutions[this->reactant].size());
    Er.resize(this->solutions[this->reactant].size());

    std::vector<double> derivatives(3, 0.0);

    if (average_cell_current)
    {
        // ##### average the solution variables over the all quadrature points, use average to compute single value for current density, apply to all points

        std::map<VariableNames, SolutionVariable > averagedSol;
        double cell_current;


        double x_R = 0.;
        double Vm = 0.;
        double Vs = 0.;
        double lambda_cell = 0;
        double Eff = 0.;
        //Take a simple average of the values in the cell
        for (unsigned int j = 0; j < this->solutions[this->reactant].size(); ++j) {
            x_R += this->solutions[this->reactant][j] / this->solutions[this->reactant].size();

            Vm +=  this->solutions[protonic_electrical_potential][j]
                            / this->solutions[protonic_electrical_potential].size();

            Vs +=  this->solutions[electronic_electrical_potential][j]
                            / this->solutions[electronic_electrical_potential].size();

            lambda_cell += this->solutions[membrane_water_content][j]
                    / this->solutions[membrane_water_content].size();
        }

        averagedSol[this->reactant] = SolutionVariable(x_R, 1, this->reactant);
        averagedSol[protonic_electrical_potential] = SolutionVariable(Vm, 1, protonic_electrical_potential);
        averagedSol[electronic_electrical_potential] = SolutionVariable(Vs, 1, electronic_electrical_potential);
        averagedSol[membrane_water_content] = SolutionVariable(lambda_cell, 1, membrane_water_content);
        cell_current = micro_scale_current(aggs[0], averagedSol, 0, Eff);

        for (unsigned int i = 0; i < current.size(); ++i) {
            current[i] = cell_current;
            Er[i] = Eff;
            if (isnan(current[i]))
                current[i] = 0.0;
        }
    } else {
        for (unsigned int i = 0; i < current.size(); ++i) {
            current[i] = micro_scale_current(aggs[0], this->solutions, i, Er[i]);

            if (isnan(current[i]))
                current[i] = 0.0;
        }
    }
}

//---------------------------------------------------------------------------
template<int dim>
double NAME::AgglomerateCL<dim>::micro_scale_current(boost::shared_ptr<FuelCellShop::MicroScale::AgglomerateBase> &_agg, std::map<VariableNames ,SolutionVariable>& solutionMap, const unsigned int& index, double& E_r) {

    //TODO:: refactor this function so that IC fetching is not run in parallel
    //TODO: Clean up these terribly ugly interfaces
    bool initial_guess_obtained = false; //Should we improve the databases knowledge of this model?
    if (solver_agg != "analytical") {
        if (!check_micro_scale_initial_guess(solutionMap[this->reactant][index], solutionMap[protonic_electrical_potential][index],
                solutionMap[electronic_electrical_potential][index], solutionMap[membrane_water_content][index], r_agg,
                thickness_agg, epsilon_agg))
            initial_guess_obtained = get_micro_scale_initial_guess(solutionMap[this->reactant][index], solutionMap[protonic_electrical_potential][index],
                    solutionMap[electronic_electrical_potential][index], solutionMap[membrane_water_content][index],_agg, r_agg,
                    thickness_agg, epsilon_agg); //from db
        else
            initial_guess_obtained = true; //our initial guess was already good enough
    }

    if (initial_guess_obtained
            or (_agg->guess_not_loaded()
                    and (micro_scale_initial_guess.size() != 0)))
        _agg->load_initial_guess(micro_scale_initial_guess);

    //Compute current
    _agg->set_solution(solutionMap, this->reactant, index);

    double answer = ((_agg->compute_current( E_r))
            * (1.0 - this->epsilon_V));

    //TODO:: refactor this function so that IC fetching is not run in parallel

    if (!initial_guess_obtained and (solver_agg != "analytical")) {
        IC_x_R = solutionMap[this->reactant][index];
        IC_phi_m = solutionMap[protonic_electrical_potential][index];
        IC_phi_s = solutionMap[electronic_electrical_potential][index];
        IC_lambda = solutionMap[membrane_water_content][index];
        IC_radius = r_agg;
        IC_film_thickness = thickness_agg;
        IC_porosity = epsilon_agg;
        _agg->load_initial_guess(_agg->results);
        commit_data_to_db(_agg->results, _agg->column_titles, IC_x_R, IC_phi_m,
                IC_phi_s, IC_lambda, r_agg, thickness_agg, epsilon_agg);
    }

    return answer;

}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::commit_data_to_db(
        std::vector<std::vector<double>> agg_data,
        std::vector<std::string> agg_column_titles, const double &x_R,
        const double &phi_m, const double &phi_s, const double lambda,
        const double radius, const double film_thickness,
        const double porosity) {
    FcstUtilities::DatabaseOC OC;

    OC.add_param("x_O2", x_R);
    OC.add_param("phi_m", phi_m);
    OC.add_param("phi_s", phi_s);
    OC.add_param("lambda", lambda);
    OC.add_param("radius", radius);
    OC.add_param("film_thickness", film_thickness);
    OC.add_param("porosity", porosity);

    if (db.connect()) {
        deallog << "Adding a database entry for model " + solver_agg + comp_agg
                << std::endl;
        db.commit_data(solver_agg + comp_agg, OC, agg_column_titles, agg_data);
        db.disconnect();

    }

}

//---------------------------------------------------------------------------
template<int dim>
bool NAME::AgglomerateCL<dim>::check_micro_scale_initial_guess(
        const double &x_R, const double &phi_m, const double &phi_s,
        const double lambda, const double radius, const double film_thickness,
        const double porosity) {
    bool answer = false;

    //Compare new conditions whit old ones we have already got data for within certain tolerance
    double tolerance = IC_tolerance;

    if ((std::abs(x_R - IC_x_R) / IC_x_R) < tolerance)
        if ((std::abs(phi_m - IC_phi_m) / IC_phi_m) < tolerance)
            if ((std::abs(phi_s - IC_phi_s) / IC_phi_s) < tolerance)
                if ((std::abs(lambda - IC_lambda) / IC_lambda) < tolerance)
                    if ((std::abs(radius - IC_radius) / IC_radius) < tolerance)
                        if ((std::abs(porosity - IC_porosity) / IC_porosity)< tolerance)
                            if ((std::abs(film_thickness - IC_film_thickness) / IC_film_thickness) < tolerance)
                                answer = true;

    return answer;
}

//---------------------------------------------------------------------------
template<int dim>
bool NAME::AgglomerateCL<dim>::get_micro_scale_initial_guess(const double &x_R,
        const double &phi_m, const double &phi_s, const double lambda,
        boost::shared_ptr<FuelCellShop::MicroScale::AgglomerateBase> &_agg,
        const double radius, const double film_thickness,
        const double porosity) {
    //Returns true if an IC within a certain tolerance was found
    bool answer = false;

    //Try get "more accurate" data
    double tolerance = IC_tolerance * 0.8;

    if (db.connect()) {
        FcstUtilities::DatabaseOC OC;
        OC.add_param("x_O2", x_R);
        OC.add_param("phi_m", phi_m);
        OC.add_param("phi_s", phi_s);
        OC.add_param("lambda", lambda);
        OC.add_param("radius", radius);
        OC.add_param("film_thickness", film_thickness);
        OC.add_param("porosity", porosity);

        if (db.has_data(solver_agg + comp_agg, OC, tolerance)) {
            //deallog << "Getting micro-scale initial condition from database." << std::endl;
            micro_scale_initial_guess = db.get_data(solver_agg + comp_agg, OC,
                    tolerance, _agg->column_titles[0]);
            answer = true;

            IC_x_R = x_R;
            IC_phi_m = phi_m;
            IC_phi_s = phi_s;
            IC_lambda = lambda;
            IC_radius = r_agg;
            IC_film_thickness = thickness_agg;
            IC_porosity = epsilon_agg;
        }

        //Try increase the tolerance
        else {
            tolerance = IC_tolerance * 3;
            //deallog << "Getting micro-scale initial condition from database (Increased tolerance)." << std::endl;
            if (db.has_data(solver_agg + comp_agg, OC, tolerance))
                micro_scale_initial_guess = db.get_data(solver_agg + comp_agg,
                        OC, tolerance, _agg->column_titles[0]);

            //Try increase the tolerance
            else {
                tolerance *= 1.5;
                if (db.has_data(solver_agg + comp_agg, OC, tolerance ))
                    micro_scale_initial_guess = db.get_data(
                            solver_agg + comp_agg, OC, tolerance,
                            _agg->column_titles[0]);
                else
                    deallog << "Unable to get initial condition for "
                            << solver_agg + comp_agg << " from database."
                            << std::endl;
            }

        }

        db.disconnect();
    } else
        std::cout << "Could not connect to db!" << std::endl;

    return answer;
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::solve_current_derivatives_average(
        std::map< VariableNames, std::vector<double> >& Dcurrent) {
    //TODO: refactor this function


    std::vector<double> dcurrent_cell(3, 0.0);
    double cell_current = 0;
    double Er_dummy;
    double x_R = 0.;
    double Vm = 0.;
    double Vs = 0.;
    double lambda_cell = 0;

    //Take a simple average of the values in the cell
    for (unsigned int j = 0; j < this->solutions[this->reactant].size(); ++j) {
        x_R += this->solutions[this->reactant][j] / this->solutions[this->reactant].size(); //this->x_O2->at(j)/this->x_O2->size();

        Vm += this->solutions[protonic_electrical_potential][j]
                / this->solutions[protonic_electrical_potential].size(); //this->phi_m->at(j)/this->phi_m->size();

        Vs +=
                this->solutions[electronic_electrical_potential][j]
                        / this->solutions[electronic_electrical_potential].size(); //this->phi_s->at(j)/this->phi_s->size();

        lambda_cell += this->solutions[membrane_water_content][j]
                / this->solutions[membrane_water_content].size(); //this->lambda->at(j)/this->lambda->size();
    }

    std::map<VariableNames, SolutionVariable > averagedSol;
    averagedSol[this->reactant]                                 = SolutionVariable(x_R, 1, this->reactant);
    averagedSol[protonic_electrical_potential]    = SolutionVariable(Vm, 1, protonic_electrical_potential);
    averagedSol[electronic_electrical_potential]  = SolutionVariable(Vs, 1, electronic_electrical_potential);
    averagedSol[membrane_water_content]           = SolutionVariable(lambda_cell, 1, membrane_water_content);

    //Derivative Pertubations
    double h = 1.0e-4; //TODO: This should not be hard coded
    double x_R_h = x_R + h;
    double Vm_h = Vm + h;
    double Vs_h = Vs + h;
    double x_R_h2 = x_R - h;
    double Vm_h2 = Vm - h;
    double Vs_h2 = Vs - h;

    //Used to store the location of the protonic and electronic derivatives.
    unsigned int location_phi_m = 0;
    unsigned int location_phi_s = 0;

    if (aggs[0]->analytical_derivatives()) {
        aggs[0]->set_solution(averagedSol, this->reactant, 0);
        dcurrent_cell = aggs[0]->compute_derivative_current(); //Derivatives are in the order x02, phi_s, phi_m
        VariableNames name;
        for (unsigned int i = 0; i < this->derivative_flags.size(); ++i) {
           if(i==0)
               name = oxygen_molar_fraction;
           else if(i==1)
               name = protonic_electrical_potential;
           else if(i==2)
               name = electronic_electrical_potential;


            for (unsigned int j = 0; j < this->solutions[this->reactant].size(); ++j) {
                Dcurrent[name][j] = dcurrent_cell[i] * (1.0 - this->epsilon_V); //*(x_O2[j]/x);//*(kinetics_dcurrent[i][j]/ave_dcurrent[i]);// / Dcurrent[i].size();
                if (isnan(Dcurrent[name][j]))
                    Dcurrent[name][j] = 0.0;
            }
        }
    }
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::solve_current_derivatives_at_each_node(
        std::map< VariableNames, std::vector<double> >& Dcurrent) {
    //TODO: Implement in Parallel

    std::vector<double> cell_current(this->solutions[this->reactant].size(), 0.0);
    std::vector<std::vector<double> > d_current(this->derivative_flags.size(),
            std::vector<double>(this->solutions[this->reactant].size(), 0.0));


    std::vector<double> x_R_h(this->solutions[this->reactant].size(), 0.0);
    std::vector<double> phi_m_h(this->solutions[protonic_electrical_potential].size(), 0.0);
    std::vector<double> phi_s_h(this->solutions[electronic_electrical_potential].size(), 0.0);

    double Er_dummy;
    //Compute Analytical Derivatives
    if (aggs[0]->analytical_derivatives()) {
        for (unsigned int j = 0; j < this->solutions[this->reactant].size(); ++j) {
            aggs[0]->set_solution(this->solutions, this->reactant, j);
            std::vector<double> dcurrent_q_point = aggs[0]->compute_derivative_current();
            for (unsigned int i = 0; i < this->derivative_flags.size(); ++i) {
                if (this->derivative_flags[i] == oxygen_molar_fraction)
                    Dcurrent[oxygen_molar_fraction][j] = dcurrent_q_point[0]
                            * (1.0 - this->epsilon_V);
                else if (this->derivative_flags[i] == protonic_electrical_potential)
                    Dcurrent[protonic_electrical_potential][j] = dcurrent_q_point[1]
                            * (1.0 - this->epsilon_V);
                else if (this->derivative_flags[i] == electronic_electrical_potential)
                    Dcurrent[electronic_electrical_potential][j] = dcurrent_q_point[2]
                            * (1.0 - this->epsilon_V);
                else
                    Dcurrent[this->derivative_flags[i]][j] = 0.0;

                if (isnan(Dcurrent[this->derivative_flags[i]][j]))
                    Dcurrent[this->derivative_flags[i]][j] = 0.0;
            }

        }
    }

}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::derivative_current_density(
        std::map< VariableNames, std::vector<double> >& Dcurrent) {


    if (!aggs_initialized)
        initialize_aggs();

    for( unsigned int i =0; i < this->derivative_flags.size(); i++){
        Dcurrent[this->derivative_flags[i]] = std::vector<double>(this->solutions[this->reactant].size(), 0.0);
    }

    if (average_cell_current)
        solve_current_derivatives_average(Dcurrent);
    else
        solve_current_derivatives_at_each_node(Dcurrent);

}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::compute_void_fraction() {
    this->epsilon_W =
            (water_agg) ?
                    (this->epsilon_S * epsilon_agg / (1 - epsilon_agg)) : 0.0;

    // Porosity
    this->epsilon_V = 1 - this->epsilon_S - this->epsilon_N - this->epsilon_W;
}

//---------------------------------------------------------------------------
template<int dim>
inline double NAME::AgglomerateCL<dim>::compute_n() const {
    double n_agg = 0.0;
    if (type_agg == "spherical")
        n_agg =
                this->epsilon_S
                        / ((4.0 / 3.0) * pi * pow(r_agg * 1e-7, 3.0)
                                * (1 - epsilon_agg)); //# of agglomerates per unit volume
    else if (type_agg == "cylindrical")
        n_agg = this->epsilon_S / (pi * pow(r_agg * 1e-7, 2.0));

    return n_agg;
}

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
template<int dim>
double NAME::AgglomerateCL<dim>::compute_thickness_agg() {
    //set default and starting values
    double thickness_0 = r_agg / 10.;
    const double tol = 1e-6;
    const double eps_tol = 1e-6;
    const double iterations = 1000;

    for (int i = 0; i < iterations; ++i) {

        double thickness_i = thickness_0;
        thickness_i = thickness_0
                - ((epsilon_N_agg(thickness_0) - this->epsilon_N)
                        / depsilonN_dthickness_agg(thickness_0));

        if (fabs((thickness_i - thickness_0) / thickness_0) < tol
                || fabs(epsilon_N_agg(thickness_i) - epsilon_N_agg(thickness_0))
                        < eps_tol) {
            if (thickness_i < 0)
                return -1;
            else
                return thickness_i;
        } else
            thickness_0 = thickness_i;

    }
    //if value does not converge, return value that will drive epsilon_N_cat above 1
    //error handling will set porosity to a minimum
    //BEWARE: program will run as normal - however, deallog output will output this negative value
    return -1;
}

//---------------------------------------------------------------------------
template<int dim>
double NAME::AgglomerateCL<dim>::epsilon_N_agg(const double delta_agg) const {
    double value = 0.0;

    // if the agglomerate is water-filled, there is no ionomer within the agglomerate.
    double porosity = (water_agg) ? 0.0 : epsilon_agg;

    if (type_agg == "spherical") {
        value = n_agg * (4.0 / 3.0) * pi
                * (pow(r_agg * 1e-7, 3.0) * (porosity - 1)
                        + pow(r_agg * 1e-7 + delta_agg * 1e-7, 3.0));
    } else if (type_agg == "cylindrical") {
        value = n_agg * pi
                * (pow(r_agg * 1e-7 + delta_agg * 1e-7, 2.0)
                        - (1.0 - porosity) * pow(r_agg * 1e-7, 2.0));
    }
    return value;
}

//---------------------------------------------------------------------------
template<int dim>
double NAME::AgglomerateCL<dim>::depsilonN_dthickness_agg(
        const double thickness_agg) const {
    double value = 0.0;

    if (type_agg == "spherical")
        value = n_agg * (4.0) * pi
                * (pow(r_agg * 1e-7 + thickness_agg * 1e-7, 2.0)) * 1e-7;

    else if (type_agg == "cylindrical")
        value = n_agg * pi * 2 * (r_agg * 1e-7 + thickness_agg * 1e-7) * 1e-7;

    return value;
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::choose_agglomerate(
        boost::shared_ptr<FuelCellShop::MicroScale::AgglomerateBase> &agg_) {

    bool error_state = false;

    if (solver_agg == "analytical"){

        if (comp_agg == "ionomer"){
            agg_ = boost::shared_ptr<FuelCellShop::MicroScale::IonomerAgglomerateSun>(
                    new FuelCellShop::MicroScale::IonomerAgglomerateSun);
        }
        else
            error_state = true;

    }


    if(error_state){
        deallog << "Invalid selection for agglomerate type and solver"
                << std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template<int dim>
void NAME::AgglomerateCL<dim>::print_layer_properties() const {
    FuelCellShop::Layer::ConventionalCL<dim>::print_layer_properties();

    deallog << "======= CATALYST LAYER MICROSTRUCTURE ======" << std::endl;
    deallog << "=== " << this->name << " ===" << std::endl;
    deallog << "Agglomerate Radius: " << r_agg << " nm" << std::endl;
    deallog << "Agglomerate Type: " << solver_agg << " " << comp_agg
            << std::endl;
    deallog << "Agglomerate Porosity: " << epsilon_agg << std::endl;
    deallog << "Agglomerate Thin Film: " << thickness_agg << " nm" << std::endl;
    deallog << "=======================================" << std::endl;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations.
template class NAME::AgglomerateCL<deal_II_dimension>;
