//---------------------------------------------------------------------------
// C++ Interface: agglomerate_ionomer_analytical.cc
//
// Description: Used to solve a system of equations representing a
// 			spherical ionomer-filled agglomerate.
//
// Author: Peter Dobson <pdobson@ualberta.ca>, (C) 2011
//			University of Alberta
//
// Copyright: See COPYING file that comes with this distribution
//
//---------------------------------------------------------------------------

#include "agglomerate_ionomer_sun.h"

namespace NAME = FuelCellShop::MicroScale;

//---------------------------------------------------------------------------
NAME::IonomerAgglomerateSun::IonomerAgglomerateSun()
{
	this->analyticalDerivatives= true;
}


//---------------------------------------------------------------------------
void
NAME::IonomerAgglomerateSun::set_structure(FuelCellShop::Material::CatalystSupportBase *CS,
					FuelCellShop::Material::CatalystBase *Cat,
					FuelCellShop::Material::PolymerElectrolyteBase* PEM,
					double A, double r, double delta, double epsilon)
{
	catalyst_support = CS;
	catalyst = Cat;
	electrolyte = PEM;
	AV = A;
	r_agg = r*1e-7; // convert to cm
	delta_agg = delta*1e-7; // convert to cm
	epsilon_agg = epsilon;
	interface = r_agg/(r_agg+delta_agg);
}


//---------------------------------------------------------------------------
void
NAME::IonomerAgglomerateSun::_initialize()
{

	if(this->reactant == oxygen_molar_fraction){
	    electrolyte->oxygen_diffusivity(D_R_N);
	    H_R_N = electrolyte->get_H_O2();
	}
	else if(this->reactant == hydrogen_molar_fraction){//TODO: change this!!!!!!!!!!!
	    deallog << "Analytical agglomerate cannot solve for VariableName::Hydrogen_molar_fraction" << std::endl;
	    abort();
	}

}


//---------------------------------------------------------------------------
double
NAME::IonomerAgglomerateSun::compute_current (double &E_r)
{
	E_r = 0.0;
	double I = 0.0;
	if(this->solutions[this->reactant][this->sol_index]< 0.0)
	    return 0.0; // Quit on non-real values
    else
        _initialize();

	double D_eff = pow(epsilon_agg,1.5) * D_R_N;

	// Set the oxygen concentration and membrane potential at the boundary
	// Set the solid phase potential for the domain
	c_R = (this->solutions[this->reactant][this->sol_index]*P)/H_R_N;
	phi_S = this->solutions[electronic_electrical_potential][this->sol_index];
	phi_M = this->solutions[protonic_electrical_potential][this->sol_index];

	double unit_size = 1;
	double agg_size = interface/unit_size;
	double volume = (4.0/3.0)*pi*pow(unit_size,3.0);
	double agg_volume = (4.0/3.0)*pi*pow(agg_size,3.0);

	//get the solution at the quadrature points
	std::vector<double> r_rate(1,0.0);

    std::vector<SolutionVariable> c_reactants;

    if(this->reactant == oxygen_molar_fraction)
        c_reactants.push_back(SolutionVariable (c_R,1, oxygen_concentration));
    else if (this->reactant == hydrogen_molar_fraction)
        c_reactants.push_back(SolutionVariable (c_R,1,hydrogen_concentration));

    SolutionVariable v_membrane(phi_M, 1, protonic_electrical_potential);
    SolutionVariable v_solid(phi_S, 1, electronic_electrical_potential);


    this->kinetics->set_reactant_concentrations(c_reactants);
    this->kinetics->set_electrolyte_potential(v_membrane);
    this->kinetics->set_solid_potential(v_solid);

	//this->kinetics->molar_reaction_rate(r_rate, "Oxygen molar fraction");
	this->kinetics->current_density(r_rate);
	r_rate[0]/=c_R;
	double k_c = AV*r_rate[0] / (4.0*F);
	double E_agg = compute_Er(k_c, D_eff);
	
	//compute the concentration at the boundary
	double cR_boundary= c_R * pow(1.0 + ((pow(r_agg,2)*delta_agg*E_agg*k_c)/(3*(r_agg + delta_agg)*D_R_N)), -1);
	
	co2_final.clear();
	co2_final.resize(mesh_final.size(),0.0);
	//Compute oxygen concentration profile across the agglomerate
	double temp = sqrt(k_c/D_eff);
	
	for(int i =0; i<mesh_final.size();i++)
	{
		
		co2_final[i]=cR_boundary*(r_agg*sinh(r_agg*mesh_final[i]*temp))/(r_agg*mesh_final[i]*sinh(r_agg*temp));
	}


	I = (4.0*F) * c_R * pow(1/(E_agg*k_c*(1-epsilon_agg)) + ((r_agg + delta_agg)*delta_agg)/(AV*r_agg*D_R_N), -1);


	double I_max = (4.0*F) * c_R * k_c * agg_volume/volume;

	double I_avg = I*agg_volume/volume;
	E_r = I_avg/I_max;

	return I_avg;
}

//---------------------------------------------------------------------------
std::vector<double>
NAME::IonomerAgglomerateSun::compute_derivative_current ()
{
	std::vector<double> dI(3);
	if(this->solutions[this->reactant][this->sol_index]< 0.0 < 0.0)
	    return dI; // Quit on non-real values
	else
	    _initialize();
	double D_eff = pow(epsilon_agg,1.5) * D_R_N;

	// Set the oxygen concentration and membrane potential at the boundary
	// Set the solid phase potential for the domain
	c_R = (this->solutions[this->reactant][this->sol_index]*P)/H_R_N;
    phi_S = this->solutions[electronic_electrical_potential][this->sol_index];
    phi_M = this->solutions[protonic_electrical_potential][this->sol_index];

	double unit_size = 1;
	double agg_size = interface/unit_size;
	double volume = (4.0/3.0)*pi*pow(unit_size,3.0);
	double agg_volume = (4.0/3.0)*pi*pow(agg_size,3.0);

	//get the solution at the quadrature points
	std::vector<double> r_rate(1,0.0);
	std::vector<double> dr_rate_dphi_m(1,0.0);
	std::vector<double> dr_rate_dphi_s(1,0.0);


	std::vector<SolutionVariable> c_reactants;


	if(this->reactant == oxygen_molar_fraction)
	    c_reactants.push_back(SolutionVariable (c_R,1,oxygen_concentration));
	else if (this->reactant == hydrogen_molar_fraction)
	    c_reactants.push_back(SolutionVariable (c_R,1,hydrogen_concentration));

	SolutionVariable v_membrane(phi_M, 1, protonic_electrical_potential);
	SolutionVariable v_solid(phi_S, 1, electronic_electrical_potential);


	this->kinetics->set_reactant_concentrations(c_reactants);
	this->kinetics->set_electrolyte_potential(v_membrane);
	this->kinetics->set_solid_potential(v_solid);



    std::map< VariableNames, std::vector<double> > derivatives;
    //TODO: Derivatives should not be set in this way, change
    std::vector<VariableNames>flags; flags.resize(3);
    flags.push_back(oxygen_molar_fraction);
    flags.push_back(electronic_electrical_potential);
    flags.push_back(protonic_electrical_potential);
    this->kinetics->set_derivative_flags(flags);
    this->kinetics->derivative_current(derivatives);

	double k_c = AV*derivatives[oxygen_molar_fraction][0] /(10*c_R*(4.0*F));
	double dkc_dphi_s = AV*derivatives[electronic_electrical_potential][0] / (c_R*(4.0*F));
	double dkc_dphi_m = AV*derivatives[protonic_electrical_potential][0] / (c_R*(4.0*F));
	double E_agg = compute_Er(k_c, D_eff);
	double dE_agg = 0.0;


    //Original
    //-- dRR_dxo2
    dI[0] =  (4*F*P)/H_R_N*pow(1/(E_agg*k_c*(1-epsilon_agg)) + ((r_agg + delta_agg)*delta_agg)/(AV*r_agg*D_R_N), -1);
    //-- dRR_dphi_m
    dE_agg = compute_dEr(k_c, dkc_dphi_m, D_eff);
    dI[1] = (4*F*c_R)*pow(1/(E_agg*k_c*(1-epsilon_agg)) + ((r_agg + delta_agg)*delta_agg)/(AV*r_agg*D_R_N), -2)
      *((dE_agg*k_c*(1-epsilon_agg) + E_agg*dkc_dphi_m*(1-epsilon_agg) )/pow(E_agg*k_c*(1-epsilon_agg),2.0));
    //-- dRR_dphi_s
    dE_agg = compute_dEr(k_c, dkc_dphi_s, D_eff);
    dI[2] = (4*F*c_R)*pow(1/(E_agg*k_c*(1-epsilon_agg)) + ((r_agg+ delta_agg)*delta_agg)/(AV*r_agg*D_R_N), -2)
      *((dE_agg*k_c*(1-epsilon_agg) + E_agg*dkc_dphi_s*(1-epsilon_agg) )/pow(E_agg*k_c*(1-epsilon_agg),2.0));


	for(unsigned int i=0;i<dI.size();++i)
		dI[i] *= agg_volume/volume;

	return dI;
}


//---------------------------------------------------------------------------
double
NAME::IonomerAgglomerateSun::compute_Er (const double k_c, const double D)
{
    //Works the same as before
	double phi_L = (r_agg / 3.0)*sqrt(k_c/D);
	double E_r = (1/phi_L)*(1/tanh(3*phi_L) - 1/(3*phi_L));  // Spherical agglomerate
	return E_r;
}


//---------------------------------------------------------------------------
double
NAME::IonomerAgglomerateSun::compute_dEr (const double k_c, const double dk_c, const double D)
{
	//-- Compute phi_L
	double phi_L = (r_agg/3.0)*sqrt(k_c/D);
	double dphiL_dkc = (r_agg / 3.0)*pow(k_c/D,-0.5)*(0.5/D);
	
	//std::cout << "phi_L = " << phi_L << " dphiL_dkc = " << dphiL_dkc << std::endl;
	//-- Dependance of E_r on phi_L
	double  dEr_dphiL = -1/(pow(phi_L,2.0)*tanh(3*phi_L)) - 3*(1-pow(tanh(3*phi_L),2.0))/(phi_L*pow(tanh(3*phi_L),2.0)) + 2.0/(3*pow(phi_L,3.0));
	
	double dE_r = dEr_dphiL*dphiL_dkc*dk_c;
	
	return dE_r;
}



