//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: conventional_cl.cc
//    - Description: Class characterizing the conventional catalyst layer and methods for computing effective properties. It also provides interface to various material classes used in catalyst layer.
//    - Developers: Marc Secanell (2006-13), Peter Dobson (2011) and Madhur Bhaiya (2013)
//    - Id: $Id: conventional_CL.cc 1465 2013-08-30 22:42:25Z madhur $ 
//
//---------------------------------------------------------------------------

#include "layers/conventional_CL.h"

namespace NAME = FuelCellShop::Layer;

typedef std::map< VariableNames, std::vector<double> >::iterator dmap_iter; 

template <int dim>
const std::string NAME::ConventionalCL<dim>::concrete_name ("ConventionalCL");

//---------------------------------------------------------------------------
template <int dim>
NAME::ConventionalCL<dim>::ConventionalCL(std::string name)
  : CatalystLayer<dim>(name)
{  
  epsilon_S = 0.0;
  epsilon_V = 0.0;
  epsilon_N = 0.0;
  Av = 0.0;
}

//---------------------------------------------------------------------------
template <int dim>
NAME::ConventionalCL<dim>::ConventionalCL()
  : CatalystLayer<dim>()
{  }

//---------------------------------------------------------------------------
template <int dim>
NAME::ConventionalCL<dim>::~ConventionalCL()
{}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::declare_parameters (const std::string& cl_section_name, 
                                               ParameterHandler &param) const
{
    NAME::CatalystLayer<dim>::declare_parameters(cl_section_name, param);
    
    param.enter_subsection("Fuel cell data"); 
    {
        param.enter_subsection(cl_section_name); 
        {
            param.enter_subsection(this->concrete_name); 
            { 
            //-- Composition
            param.declare_entry ("Platinum loading on support (%wt)", // mass percentage of platinum catalyst on the support carbon black
                                 "0.46", // [-]
                                 Patterns::Double(),
                                 "Mass percentage of platinum catalyst on the support carbon black");
            param.declare_entry ("Platinum loading per unit volume (mg/cm3)",// catalyst platinum mass loading per unit volume of CL
                                 "400",
                                 Patterns::Double(),
                                 "Catalyst platinum mass loading per unit volume of CL");
            param.declare_entry ("Electrolyte loading (%wt)",  //
                                 "0.3", // [-]
                                 Patterns::Double(),
                                 "Electrode loading is the weight percentage of ionomer per gram of CL");
            
            //-- Network characteristics
            param.declare_entry ("Method effective transport properties in pores",
                                 "Bruggemann",
                                 Patterns::Selection("Given|Bruggemann|Percolation"),
                                 "Method used to compute effective transport properties in the void phase.");
            param.declare_entry ("Porosity threshold",  // volume fraction of void space in the CL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of void space in the CL."
                                 "If the porosity is less than this value transport does not occur");
            param.declare_entry ("Porosity network constant",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            param.declare_entry ("Porosity gamma network constant",
                                 "0.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory to account for extra diffusion");
            // ---
            param.declare_entry ("Method effective transport properties in solid phase",
                                 "Bruggemann",
                                 Patterns::Selection("Given|Bruggemann|Percolation"),
                                 "Method used to compute effective transport properties in pores");
            param.declare_entry ("Solid network threshold",  // volume fraction of solid in the CL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of solid (electron conductive) phase in the CL."
                                 "If the solid phase is less than this value transport in the fibre network does not occur");
            param.declare_entry ("Solid network constant",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            param.declare_entry ("Method effective transport properties in electrolyte phase",
                                 "Bruggemann",
                                 Patterns::Selection("Given|Bruggemann|Percolation|Iden11"),
                                 "Method used to compute effective transport properties in pores");
            param.declare_entry ("Electrolyte network threshold",  // volume fraction of electrolyte in the CL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of electrolyte (proton conductive) phase in the CL."
                                 "If the electrolyte phase is less than this value transport in the network does not occur");
            param.declare_entry ("Electrolyte network constant",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            param.declare_entry ("Method to compute active area",
                                 "given",
                                 Patterns::Selection("given|Marr|ETEK06|ETEK07"));
            param.declare_entry ("Active area [cm^2/cm^3]",
                                 "2.0e5",
                                 Patterns::Double());
            param.declare_entry ("Method to compute porosity",
                                 "marc",
                                 Patterns::Selection("marc"));
           }
           param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}


//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::set_parameters (const std::vector<std::string>& name_dvar,
                                           const std::vector<double>& value_dvar,
                                           const std::string& cl_section_name,
                                           ParameterHandler &param) const
{
    NAME::CatalystLayer<dim>::set_parameters(name_dvar, value_dvar, cl_section_name, param);
    
    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection(cl_section_name); 
        {
            param.enter_subsection(this->concrete_name); 
            { 
                for (unsigned int i=0; i<name_dvar.size(); ++i)
                {
                    if(this->name == "Cathode catalyst layer")
                    {
                        if (name_dvar[i] == "V_Pt_c")
                            param.set("Platinum loading per unit volume (mg/cm3)", value_dvar[i]);
                        else if (name_dvar[i] == "prc_Pt_c")
                            param.set("Platinum loading on support (%wt)", value_dvar[i]);
                        else if (name_dvar[i] == "prc_N_c")
                            param.set("Electrolyte loading (%wt)", value_dvar[i]);
                        else if (name_dvar[i] == "Av_c")
                            param.set("Active area [cm^2/cm^3]", value_dvar[i]);
                    }
                    
                    if(this->name == "Anode catalyst layer")
                    {
                        if (name_dvar[i] == "V_Pt_a")
                            param.set("Platinum loading per unit volume (mg/cm3)", value_dvar[i]);
                        else if (name_dvar[i] == "prc_Pt_a")
                            param.set("Platinum loading on support (%wt)", value_dvar[i]);
                        else if (name_dvar[i] == "prc_N_a")
                            param.set("Electrolyte loading (%wt)", value_dvar[i]);
                        else if (name_dvar[i] == "Av_a")
                            param.set("Active area [cm^2/cm^3]", value_dvar[i]);
                    }
                }
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::initialize (ParameterHandler &param)
{
    NAME::CatalystLayer<dim>::initialize(param);
    
    rho_N = this->electrolyte->get_density();
    rho_c = this->catalyst_support->get_density();
    rho_Pt = this->catalyst->get_density();
    
    param.enter_subsection("Fuel cell data"); 
    {
        param.enter_subsection(this->name);
        {
            param.enter_subsection(this->concrete_name); 
            { 
                prc_Pt = param.get_double("Platinum loading on support (%wt)");
                V_Pt = param.get_double("Platinum loading per unit volume (mg/cm3)");
                loading_N = param.get_double("Electrolyte loading (%wt)");
                
                method_eff_property_pores = param.get("Method effective transport properties in pores");
                method_eff_property_solid = param.get("Method effective transport properties in solid phase");
                method_eff_property_electrolyte = param.get("Method effective transport properties in electrolyte phase");
                
                porosity_th = param.get_double("Porosity threshold");
                porosity_mu = param.get_double("Porosity network constant");
                porosity_gamma = param.get_double("Porosity gamma network constant");
                
                solid_th = param.get_double("Solid network threshold");
                solid_mu = param.get_double("Solid network constant");
                
                electrolyte_th = param.get_double("Electrolyte network threshold");
                electrolyte_mu = param.get_double("Electrolyte network constant");
                
                method_Av = param.get("Method to compute active area");
                method_porosity = param.get("Method to compute porosity");
                
                if(method_Av == "given")
                    Av = param.get_double("Active area [cm^2/cm^3]");
                else 
                    compute_Av();
            }
            param.leave_subsection();
        }
        param.leave_subsection();   
    }
    param.leave_subsection();
    
    compute_volume_fraction();
    
    if (epsilon_S > 1 ||  epsilon_V <0)
    {
        epsilon_V = 1e-4;
        epsilon_S = 1 - epsilon_N - epsilon_V;
        deallog << "WARNING: Adjusted volume fractions to physical values" << std::endl;
        this->print_layer_properties();//this->print_volume_fractions();
    }
    else if (epsilon_S < 0 ||  epsilon_V > 1)
    {
        epsilon_S = 1e-4;
        epsilon_V = 1 - epsilon_N - epsilon_S;
        deallog << "WARNING: Adjusted volume fractions to physical values" << std::endl;
        this->print_layer_properties();//this->print_volume_fractions();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::effective_gas_diffusivity(const double& prop,
                                                     double& prop_eff) const
{
    // Initialize variable
    prop_eff = 0.0;
    
    if (method_eff_property_pores == "Bruggemann")
    {
        // Most common method
        prop_eff = prop*pow(epsilon_V, 1.5);
    }
    else if (method_eff_property_pores == "Percolation")
    {
        // Method used by M. Eikerling 
        // Note: For porosity_th[0] I use index 0 for convenience. They all contain the same value
        double aux = fabs(epsilon_V - porosity_th)/(1 - porosity_th);
        double step = 1.0e-3;
        if (epsilon_V >= porosity_th)
            step = 1.0;
        prop_eff = prop*((1-porosity_gamma)*pow(aux,porosity_mu)*step + porosity_gamma);
    }
    else
    {
        std::cout << "Unknown method to compute effective transport in the pores in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template<int dim>
void
NAME::ConventionalCL<dim>::effective_gas_diffusivity(Table< 2, Tensor< 2, dim > >& prop_eff) const
{
    prop_eff.reinit(this->gases.size(),this->gases.size());
    
    for (unsigned int i = 0; i<this->gases.size(); ++i)
    {
        for (unsigned int j = 0; j<this->gases.size(); ++j)
        {
            if ( i != j)        // Not computing self diffusion coefficients
            {
                for (unsigned int d = 0; d<dim; ++d)
                {
                    this->effective_gas_diffusivity( this->D_ECtheory(i,j), prop_eff(i,j)[d][d] );
                }
            }
        }
    }
}
//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::effective_electron_conductivity(double& prop_eff) const
{
    // Initialize variable
    prop_eff = this->catalyst_support->get_electrical_conductivity();
    
    if (method_eff_property_solid == "Bruggemann")
        // Most common method
    {
        prop_eff = prop_eff*pow(epsilon_S, 1.5);
    }
    else if (method_eff_property_solid == "Percolation")
        // Method used by M. Eikerling
    {
        double aux = fabs(epsilon_S - solid_th)/(1 - solid_th);
        double step = 1e-3; //this should be zero, but this would make the code diverge
        if (epsilon_S >= solid_th)
            step = 1.0;
        prop_eff = prop_eff*pow(aux,solid_mu)*step;
    }
    else
    {
        std::cout << "Unknown method to compute effective transport in the solid phase in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::effective_electron_conductivity(Tensor<2,dim>& prop_eff) const
{
    double iso_prop_eff;
    this->effective_electron_conductivity(iso_prop_eff);
    
    for (unsigned int i=0; i<dim; ++i)
        prop_eff[i][i] = iso_prop_eff;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::derivative_effective_electron_conductivity(std::vector<double>& ) const
{
	const std::type_info& info = typeid(*this);
	deallog << "Pure function " << __FUNCTION__
	<< " called in Class "
	<< info.name() << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::effective_proton_conductivity(double& prop_eff) const
{	
    // Initialize variable
    prop_eff = 0.0;
    this->electrolyte->proton_conductivity(prop_eff);
    
    if (method_eff_property_electrolyte == "Bruggemann")
        // Most common method
    {
        prop_eff = prop_eff*pow(epsilon_N, 1.5);
    }    
    else if (method_eff_property_electrolyte == "Percolation")
        // Method used by M. Eikerling
    {
        double aux = fabs(epsilon_N - electrolyte_th)/(1 - electrolyte_th);
        double step = 1e-3; //this should be zero, but this would make the code diverge
        if (epsilon_N >= electrolyte_th)
            step = 1.0;
        prop_eff = prop_eff*pow(aux,electrolyte_mu)*step;
    }
    
    else if (method_eff_property_electrolyte == "Iden11")
    {
        // Mesh structure
        prop_eff = prop_eff*pow(epsilon_N, 1.6);
    }
    else
    {
        std::cout << "Unknown method to compute effective transport in the electrolyte in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::effective_proton_conductivity(std::vector<double>& prop_eff) const
{	
    this->electrolyte->proton_conductivity(prop_eff);
    
    if (method_eff_property_electrolyte == "Bruggemann")
        // Most common method
    {
        for (unsigned int i=0; i<prop_eff.size(); ++i)
        {
            prop_eff[i] = prop_eff[i]*pow(epsilon_N, 1.5);           
        }
    }    
    else if (method_eff_property_electrolyte == "Percolation")
        // Method used by M. Eikerling
    {
        for (unsigned int i=0; i<prop_eff.size(); ++i)
        {
            double aux = fabs(epsilon_N - electrolyte_th)/(1 - electrolyte_th);
            double step = 1e-3; //this should be zero, but this would make the code diverge
            if (epsilon_N >= electrolyte_th)
                step = 1.0;
            prop_eff[i] = prop_eff[i]*pow(aux,electrolyte_mu)*step;
        }
    }
    else if (method_eff_property_electrolyte == "Iden11")
    {
        for (unsigned int i=0; i<prop_eff.size(); ++i)
        {	prop_eff[i] = prop_eff[i] * (std::pow(epsilon_N, 1.6));}
    }
    else
    {
        std::cout << "Unknown method to compute effective transport in the electrolyte in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::derivative_effective_proton_conductivity(std::map< VariableNames, std::vector<double> >& Dsigma_eff) const
{
    Assert (this->derivative_flags.size()!=0, ExcMessage("set_derivative_flags has not been probably called in ConventionalCL::derivative_effective_proton_conductivity."));
    
    this->electrolyte->proton_conductivity_derivative(Dsigma_eff);
    
    if (method_eff_property_electrolyte == "Bruggemann")
        for ( dmap_iter i=Dsigma_eff.begin(); i!=Dsigma_eff.end(); ++i)
            for (unsigned int j=0; j < i->second.size(); ++j)
                i->second[j] = i->second[j]*pow(epsilon_N, 1.5);

    else if (method_eff_property_electrolyte == "Percolation")
        for ( dmap_iter i=Dsigma_eff.begin(); i!=Dsigma_eff.end(); ++i)
            for (unsigned int j=0; j < i->second.size(); ++j)
            {
                double aux = fabs(epsilon_N - electrolyte_th)/(1 - electrolyte_th);
                double step = 1.e-3; //this should be zero, but this would make the code diverge
                if (epsilon_N >= electrolyte_th)
                    step = 1.0;
                i->second[j] = i->second[j]*pow(aux,electrolyte_mu)*step;
            }

    else if (method_eff_property_electrolyte == "Iden11")
        for ( dmap_iter i=Dsigma_eff.begin(); i!=Dsigma_eff.end(); ++i)
            for (unsigned int j=0; j < i->second.size(); ++j)
                i->second[j] = i->second[j]*(std::pow(epsilon_N, 1.6));
   
    else
    {
        std::cout << "Unknown method to compute effective transport in the electrolyte in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::effective_water_diffusivity(double& prop_eff) const
{
    // Initialize variable
    prop_eff = 0.0;
    this->electrolyte->water_diffusivity(prop_eff);
    
    if (method_eff_property_electrolyte == "Bruggemann")
        // Most common method
    {
        prop_eff = prop_eff*pow(epsilon_N, 1.5);
    }    
    else if (method_eff_property_electrolyte == "Percolation")
        // Method used by M. Eikerling
    {
        double aux = fabs(epsilon_N - electrolyte_th)/(1 - electrolyte_th);
        double step = 1e-3; //this should be zero, but this would make the code diverge
        if (epsilon_N >= electrolyte_th)
            step = 1.0;
        prop_eff = prop_eff*pow(aux,electrolyte_mu)*step;
    }
    
    else if (method_eff_property_electrolyte == "Iden11")
    {
        // Mesh structure
        prop_eff = prop_eff*pow(epsilon_N, 1.6);
    }
    else
    {
        std::cout << "Unknown method to compute effective transport in the electrolyte in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::effective_water_diffusivity(std::vector<double>& prop_eff) const
{
    this->electrolyte->water_diffusivity(prop_eff);
    
    if (method_eff_property_electrolyte == "Bruggemann")
        // Most common method
    {
        for (unsigned int i=0; i<prop_eff.size(); ++i)
        {	prop_eff[i] = prop_eff[i]*pow(epsilon_N, 1.5);}
    }    
    else if (method_eff_property_electrolyte == "Percolation")
        // Method used by M. Eikerling
    {
        for (unsigned int i=0; i<prop_eff.size(); ++i)
        {
            double aux = fabs(epsilon_N - electrolyte_th)/(1 - electrolyte_th);
            double step = 1e-3; //this should be zero, but this would make the code diverge
            if (epsilon_N >= electrolyte_th)
                step = 1.0;
            prop_eff[i] = prop_eff[i]*pow(aux,electrolyte_mu)*step;
        }
    }
    else if (method_eff_property_electrolyte == "Iden11")
    {
        for (unsigned int i=0; i<prop_eff.size(); ++i)
        {	prop_eff[i] = prop_eff[i] * (std::pow(epsilon_N, 1.6));}
    }
    else
    {
        std::cout << "Unknown method to compute effective transport in the electrolyte in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::derivative_effective_water_diffusivity(std::map< VariableNames, std::vector<double> >& Dprop_eff) const
{
    Assert (this->derivative_flags.size()!=0, ExcMessage("set_derivative_flags has not been probably called in ConventionalCL::derivative_effective_water_diffusivity."));
    
    this->electrolyte->water_diffusivity_derivative(Dprop_eff);
    
    if (method_eff_property_electrolyte == "Bruggemann")
        for ( dmap_iter i=Dprop_eff.begin(); i!=Dprop_eff.end(); ++i)
            for (unsigned int j=0; j < i->second.size(); ++j)
                i->second[j] = i->second[j]*pow(epsilon_N, 1.5);
   
    else if (method_eff_property_electrolyte == "Percolation")
        for ( dmap_iter i=Dprop_eff.begin(); i!=Dprop_eff.end(); ++i)
            for (unsigned int j=0; j < i->second.size(); ++j)
            {
                double aux = fabs(epsilon_N - electrolyte_th)/(1 - electrolyte_th);
                double step = 1e-3; //this should be zero, but this would make the code diverge
                if (epsilon_N >= electrolyte_th)
                    step = 1.0;
                i->second[j] = i->second[j]*pow(aux,electrolyte_mu)*step;
            }

    else if (method_eff_property_electrolyte == "Iden11")
        for ( dmap_iter i=Dprop_eff.begin(); i!=Dprop_eff.end(); ++i)
            for (unsigned int j=0; j < i->second.size(); ++j)
                i->second[j] = i->second[j]*(std::pow(epsilon_N, 1.6));

    else
    {
        std::cout << "Unknown method to compute effective transport in the electrolyte in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//------------ PROTECTED MEMBER FUNCTIONS:
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::compute_volume_fraction()
{
    if (method_porosity == "marc")
    {
        // Solid phase volume fraction:
        epsilon_S = (1.0/rho_Pt + (1.0-prc_Pt)/(prc_Pt*rho_c))*(V_Pt*1.e-3);
        
        // Electrolyte volume fraction:
        epsilon_N = loading_N/(1.0-loading_N)*(1.0/prc_Pt)*(1.0/rho_N)*(V_Pt*1.e-3);
        
        // Porosity
        epsilon_V = 1.0 - epsilon_S - epsilon_N;
    }
    else if (epsilon_S == 0.0 && epsilon_V == 0.0 && epsilon_N == 0.0)
    {
        deallog << "warning" << std::endl;
    }
}


//---------------------------------------------------------------------------
template <int dim>
void 
NAME::ConventionalCL<dim>::compute_Av()
{
    // Compute the active area of catalyst. Depends on V_Pt and %Pt
    double A_0 = 0.0;
    if (method_Av == "Marr")
    {
        A_0 = 2.2779e6*pow(prc_Pt,3.0) - 1.5857e6*pow(prc_Pt,2.0) - 2.0153e6*prc_Pt + 1.5950e6;
        Av = A_0*(V_Pt*1e-3);
    }
    else if (method_Av == "ETEK06")
    {
        A_0 = -4.5646e5*pow(prc_Pt,3.0) + 1.0618e6*pow(prc_Pt,2.0) - 1.8564e6*prc_Pt + 1.5955e6;
        Av = A_0*(V_Pt*1e-3);
    }
    else if (method_Av == "ETEK07")
    {
        A_0 = 0.7401e7*pow(prc_Pt,4.0) - 1.8105e7*pow(prc_Pt,3.0) + 1.5449e7*pow(prc_Pt,2.0) - 0.6453e7*prc_Pt + 0.2054e7;
        Av = A_0*(V_Pt*1e-3);
    }
    else if (this->Av == 0.0)
    {
        deallog << "warning" << std::endl;
    }
}


//---------------------------------------------------------------------------
template <int dim>
void
NAME::ConventionalCL<dim>::derivative_effective_proton_conductivity_wrt_electrolyte_loading(double& Dsigma_eff) const
{
    if (method_eff_property_electrolyte == "Bruggemann")
    {
        Dsigma_eff = electrolyte_proton_conductivity*1.5*pow(epsilon_N, 0.5);
    }
    else if (method_eff_property_electrolyte == "Percolation")
    {
        double aux = fabs(epsilon_N - electrolyte_th)/(1 - electrolyte_th);
        double Daux_Depsilon = 1/(1 - electrolyte_th);
        double step = 0.0;
        if (epsilon_N >= electrolyte_th)
            step = 1.0;
        Dsigma_eff = electrolyte_proton_conductivity*electrolyte_mu*pow(aux,electrolyte_mu-1)*Daux_Depsilon*step;
    }
    else
    {
        std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}


//---------------------------------------------------------------------------
template <int dim>
inline void
NAME::ConventionalCL<dim>::derivative_volume_fractions(double &Depsilon_S,
							     double &Depsilon_V,
							     double &Depsilon_N) const
{
// for (unsigned int i=0; i<this->derivative_flags.size(); ++i)
// {
// 	if (this->name == "Anode catalyst layer")
// 	{
// 		if (this->derivative_flags[i] == "epsilon_N_cat_a")
// 		{
// 			Depsilon_V = depsilon_V_cat_depsilon_N_cat();
// 			Depsilon_S = 0.0;
// 			Depsilon_N = 1.0;
// 		}
// 		else if (this->derivative_flags[i] == "V_Pt_a")
// 		{
// 			Depsilon_V = depsilon_V_cat_depsilon_S_cat()*depsilon_S_cat_dVPt(prc_Pt);
// 			Depsilon_S = depsilon_S_cat_dVPt(prc_Pt);
// 			Depsilon_N = 0.0; 
// 		}
// 		else if (this->derivative_flags[i] == "prc_Pt_a")
// 		{
// 			Depsilon_V = depsilon_V_cat_depsilon_S_cat()*depsilon_S_cat_dprc_Pt(V_Pt, prc_Pt);
// 			Depsilon_S = depsilon_S_cat_dprc_Pt(V_Pt, prc_Pt);
// 			Depsilon_N = 0.0;
// 		}
// 		else if (this->derivative_flags[i] == "epsilon_V_gdl_a")
// 		{
// 			Depsilon_V = 0.0;
// 			Depsilon_S = 0.0;
// 			Depsilon_N = 0.0;
// 		}
// 		else
// 		{
// 			Depsilon_V = 0.0;
// 			Depsilon_S = 0.0;
// 			Depsilon_N = 0.0;
// 		}
// 	}
// 	if (this->name == "Cathode catalyst layer")
// 	{
// 		if (this->derivative_flags[i] == "epsilon_N_cat_c")
// 		{
// 			Depsilon_V = depsilon_V_cat_depsilon_N_cat();
// 			Depsilon_S = 0.0;
// 			Depsilon_N = 1.0;
// 		}
// 		else if (this->derivative_flags[i] == "V_Pt_c")
// 		{
// 			Depsilon_V = depsilon_V_cat_depsilon_S_cat()*depsilon_S_cat_dVPt(prc_Pt);
// 			Depsilon_S = depsilon_S_cat_dVPt(prc_Pt);
// 			Depsilon_N = 0.0;
// 		}
// 		else if (this->derivative_flags[i] == "prc_Pt_c")
// 		{
// 			Depsilon_V = depsilon_V_cat_depsilon_S_cat()*depsilon_S_cat_dprc_Pt(V_Pt, prc_Pt);
// 			Depsilon_S = depsilon_S_cat_dprc_Pt(V_Pt, prc_Pt);
// 			Depsilon_N = 0.0;
// 		}
// 		else if (this->derivative_flags[i] == "epsilon_V_gdl_c")
// 		{
// 			Depsilon_V = 0.0;
// 			Depsilon_S = 0.0;
// 			Depsilon_N = 0.0;
// 		}
// 		else
// 		{
// 			Depsilon_V = 0.0;
// 			Depsilon_S = 0.0;
// 			Depsilon_N = 0.0;
// 		}
// 	}
// }
}

//---------------------------------------------------------------------------
template <int dim>
inline void
NAME::ConventionalCL<dim>::print_layer_properties() const
{
    deallog<<"======= CATALYST LAYER STRUCTURE ======"<<std::endl;
    deallog<<"=== "<<this->name<<" ==="<<std::endl;
    deallog<<"Solid phase volume fraction:"<<epsilon_S<<std::endl;
    deallog<<"Ionomer in the CL:"<<epsilon_N<<std::endl;
    deallog<<"Porosity in the CL:"<<epsilon_V<<std::endl;
    deallog<<"Active area in the CL:"<<Av<<std::endl;
    deallog<<"---------------------------------------"<<std::endl;
    deallog<<"Method to compute eff_property_electrolyte: "<<method_eff_property_electrolyte<<std::endl;
    deallog<<"======================================="<<std::endl;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations. 
template class NAME::ConventionalCL<deal_II_dimension>;
