//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: design_MPL.cc
//    - Description: Class used to represent a design MPL where effective properties are computed based on the porosity etc.
//    - Developers: Madhur Bhaiya
//    - $Id: design_MPL.cc 1460 2013-08-30 22:28:49Z madhur $ 
//
//---------------------------------------------------------------------------

#include "layers/design_MPL.h"

namespace NAME = FuelCellShop::Layer;

template <int dim>
const std::string NAME::DesignMPL<dim>::concrete_name ("DesignMPL");

template <int dim>
NAME::DesignMPL<dim> const* NAME::DesignMPL<dim>::PROTOTYPE = new NAME::DesignMPL<dim>();

//---------------------------------------------------------------------------
template <int dim>
NAME::DesignMPL<dim>::DesignMPL(std::string name)
  : NAME::MicroPorousLayer<dim>(name)
{
  deallog<<" of type DesignMPL"<<std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
NAME::DesignMPL<dim>::DesignMPL()
  : NAME::MicroPorousLayer<dim>()
{
  //deallog<<" Register DesignMPL MPL to FactoryMap"<<std::endl;
  this->get_mapFactory()->insert(std::pair<std::string, MicroPorousLayer<dim>* > (this->concrete_name, this) ); 
}
//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::declare_parameters (const std::string& mpl_section_name,
                                          ParameterHandler &param) const
{
  
  NAME::MicroPorousLayer<dim>::declare_parameters(param);
  
  param.enter_subsection("Fuel cell data"); 
  {
    param.enter_subsection(mpl_section_name); 
    { 
        param.enter_subsection(concrete_name); 
        { 
            // General data
            param.declare_entry ("Porosity",  // volume fraction of void space in the GDL
                                 "0.3", // [-]
                                 Patterns::Double(),
                                 "Volume fraction of void space in the GDL");
            param.declare_entry ("Electric conductivity", 
                                 "10",
                                 Patterns::Double(),
                                 "Either effective (if Given method used) or solid electric conductivity [S/cm]");
            // Method to compute effective properties
            param.declare_entry ("Method effective transport properties in pores",
                                 "Bruggemann",
                                 Patterns::Selection("Given|Bruggemann|Percolation"),
                                 "Method used to compute effective transport properties in the void phase.");
            param.declare_entry ("Method effective transport properties in solid phase",
                                 "Bruggemann",
                                 Patterns::Selection("Given|Bruggemann|Percolation"),
                                 "Method used to compute effective transport properties in solid phase");
            
            //-- Diffusivity for the isotropic case:
            param.declare_entry ("Porosity threshold",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of void space in the GDL."
                                 "If the porosity is less than this value transport does not occur");
            param.declare_entry ("Porosity network constant",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            param.declare_entry ("Porosity gamma network constant",
                                 "0.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory to account for extra diffusion");
            param.declare_entry ("Oxygen diffusion coefficient",
                                 "0.2741", //1atm 353K
                                 Patterns::Double());
            param.declare_entry ("Water vapour diffusion coefficient",
                                 "0.29676", 
                                 Patterns::Double());
            param.declare_entry ("Solid network threshold",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of solid (electron conductive) phase in the MPL."
                                 "If the solid phase is less than this value transport in the fibre network does not occur");
            param.declare_entry ("Solid network constant",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            //-- Diffusivity for the anisotropic case:
            param.declare_entry ("Anisotropic transport",
                                 "false", // [S/cm]
                                 Patterns::Bool(),
                                 "Boolean variable. Set to true if we want to account for anisotropy of the MPL");
            //--- XX
            param.declare_entry ("Electric conductivity X", 
                                 "10",
                                 Patterns::Double(),
                                 "Either effective (if Given method used) or solid electric conductivity [S/cm]");
            param.declare_entry ("Porosity threshold X",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of void space in the GDL."
                                 "If the porosity is less than this value transport does not occur");
            param.declare_entry ("Porosity network constant X",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            param.declare_entry ("Porosity gamma network constant X",
                                 "0.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory to account for extra diffusion");
            param.declare_entry ("Oxygen diffusion coefficient X",
                                 "0.2741", //1atm 353K
                                 Patterns::Double());
            param.declare_entry ("Water vapour diffusion coefficient X",
                                 "0.29676", 
                                 Patterns::Double());
            param.declare_entry ("Fibre network threshold X",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of fibres in the GDL."
                                 "If the solid phase is less than this value transport in the fibre network does not occur");
            param.declare_entry ("Fibre network constant X",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            // YY
            param.declare_entry ("Electric conductivity Y", 
                                 "10",
                                 Patterns::Double(),
                                 "Either effective (if Given method used) or solid electric conductivity [S/cm]");
            param.declare_entry ("Porosity threshold Y",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of void space in the GDL."
                                 "If the porosity is less than this value transport does not occur");
            param.declare_entry ("Porosity network constant Y",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            param.declare_entry ("Porosity gamma network constant Y",
                                 "0.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory to account for extra diffusion");
            param.declare_entry ("Oxygen diffusion coefficient Y",
                                 "0.2741", //1atm 353K
                                 Patterns::Double());
            param.declare_entry ("Water vapour diffusion coefficient Y",
                                 "0.29676", 
                                 Patterns::Double());
            param.declare_entry ("Fibre network threshold Y",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of fibres in the GDL."
                                 "If the solid phase is less than this value transport in the fibre network does not occur");
            param.declare_entry ("Fibre network constant Y",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            // ZZ
            param.declare_entry ("Electric conductivity Z", 
                                 "10",
                                 Patterns::Double(),
                                 "Either effective (if Given method used) or solid electric conductivity [S/cm]");
            param.declare_entry ("Porosity threshold Z",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of void space in the GDL."
                                 "If the porosity is less than this value transport does not occur");
            param.declare_entry ("Porosity network constant Z",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
            param.declare_entry ("Porosity gamma network constant Z",
                                 "0.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory to account for extra diffusion");
            param.declare_entry ("Oxygen diffusion coefficient Z",
                                 "0.2741", //1atm 353K
                                 Patterns::Double());
            param.declare_entry ("Water vapour diffusion coefficient Z",
                                 "0.29676", 
                                 Patterns::Double());
            param.declare_entry ("Fibre network threshold Z",  // volume fraction of void space in the GDL
                                 "0.12", // [-]
                                 Patterns::Double(),
                                 "Threshold value of the volume fraction of fibres in the GDL."
                                 "If the solid phase is less than this value transport in the fibre network does not occur");
            param.declare_entry ("Fibre network constant Z",
                                 "2.0",
                                 Patterns::Double(),
                                 "Parameter used when using percolation theory");
        }
        param.leave_subsection();
    }
    param.leave_subsection();
  }
  param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::set_parameters (const std::vector<std::string>& name_dvar,
                                      const std::vector<double>& value_dvar,
                                      ParameterHandler &param)
{
    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection(this->name);
        {
            param.enter_subsection(concrete_name);
            {
                for (unsigned int i=0; i<name_dvar.size(); ++i)
                {
                    if (name_dvar[i] == "D_O2_CMPL")
                    {
                        if(this->name == "Cathode microporous layer")
                            param.set("Oxygen diffusion coefficient", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "D_H2O_CMPL")
                    {
                        if(this->name == "Cathode microporous layer")
                            param.set("Water vapour diffusion coefficient", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "D_H2O_AMPL")
                    {
                        if(this->name == "Anode microporous layer")
                            param.set("Water vapour diffusion coefficient", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "sigma_e_MPL_X")
                    {
                        param.set("Electric conductivity X", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "sigma_e_MPL_Y")
                    {
                        param.set("Electric conductivity Y", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "sigma_e_MPL")
                    {
                        param.set("Electric conductivity", value_dvar[i]);
                    }
                }
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::initialize (ParameterHandler &param)
{
    NAME::MicroPorousLayer<dim>::initialize(param);
    
    param.enter_subsection("Fuel cell data"); 
    {
        param.enter_subsection(this->name); 
        { 
            param.enter_subsection(concrete_name); 
            { 
                porosity = param.get_double("Porosity");
                electrical_conductivity = param.get_double("Electric conductivity");
                solid_phase = 1 - porosity;
                method_eff_property_pores = param.get("Method effective transport properties in pores");
                method_eff_property_fibres = param.get("Method effective transport properties in solid phase");
                
                // Anisotropy
                anisotropy = param.get_bool ("Anisotropic transport");
                
                matrix_electrical_conductivity.clear();
                porosity_th.resize(dim);
                porosity_mu.resize(dim);
                porosity_gamma.resize(dim);
                D_O2.resize(dim);
                D_wv.resize(dim);
                fibre_th.resize(dim);
                fibre_mu.resize(dim);
                
                if (anisotropy == true)
                {
                    if (dim == 1)
                    {
                        // X
                        matrix_electrical_conductivity[0][0] = param.get_double("Electric conductivity X");
                        porosity_th[0] = param.get_double("Porosity threshold X");
                        porosity_mu[0] = param.get_double("Porosity network constant X");
                        porosity_gamma[0] = param.get_double("Porosity gamma network constant X");
                        D_O2[0] = param.get_double("Oxygen diffusion coefficient X");
                        D_wv[0] = param.get_double("Water vapour diffusion coefficient X"); 
                        fibre_th[0] = param.get_double("Fibre network threshold X");
                        fibre_mu[0] = param.get_double("Fibre network constant X");
                    }
                    else if (dim == 2)
                    {
                        // X
                        matrix_electrical_conductivity[0][0] = param.get_double("Electric conductivity X");
                        porosity_th[0] = param.get_double("Porosity threshold X");
                        porosity_mu[0] = param.get_double("Porosity network constant X");
                        porosity_gamma[0] = param.get_double("Porosity gamma network constant X");
                        D_O2[0] = param.get_double("Oxygen diffusion coefficient X");
                        D_wv[0] = param.get_double("Water vapour diffusion coefficient X"); 
                        fibre_th[0] = param.get_double("Fibre network threshold X");
                        fibre_mu[0] = param.get_double("Fibre network constant X");
                        // Y
                        matrix_electrical_conductivity[1][1] = param.get_double("Electric conductivity Y");
                        porosity_th[1] = param.get_double("Porosity threshold Y");
                        porosity_mu[1] = param.get_double("Porosity network constant Y");
                        porosity_gamma[1] = param.get_double("Porosity gamma network constant Y");
                        D_O2[1] = param.get_double("Oxygen diffusion coefficient Y");
                        D_wv[1] = param.get_double("Water vapour diffusion coefficient Y"); 
                        fibre_th[1] = param.get_double("Fibre network threshold Y");
                        fibre_mu[1] = param.get_double("Fibre network constant Y");
                    }
                    else if (dim == 3)
                    {
                        // X
                        matrix_electrical_conductivity[0][0] = param.get_double("Electric conductivity X");
                        porosity_th[0] = param.get_double("Porosity threshold X");
                        porosity_mu[0] = param.get_double("Porosity network constant X");
                        porosity_gamma[0] = param.get_double("Porosity gamma network constant X");
                        D_O2[0] = param.get_double("Oxygen diffusion coefficient X");
                        D_wv[0] = param.get_double("Water vapour diffusion coefficient X"); 
                        fibre_th[0] = param.get_double("Fibre network threshold X");
                        fibre_mu[0] = param.get_double("Fibre network constant X");
                        // Y
                        matrix_electrical_conductivity[1][1] = param.get_double("Electric conductivity Y");
                        porosity_th[1] = param.get_double("Porosity threshold Y");
                        porosity_mu[1] = param.get_double("Porosity network constant Y");
                        porosity_gamma[1] = param.get_double("Porosity gamma network constant Y");
                        D_O2[1] = param.get_double("Oxygen diffusion coefficient Y");
                        D_wv[1] = param.get_double("Water vapour diffusion coefficient Y"); 
                        fibre_th[1] = param.get_double("Fibre network threshold Y");
                        fibre_mu[1] = param.get_double("Fibre network constant Y");
                        // Z
                        matrix_electrical_conductivity[2][2] = param.get_double("Electric conductivity Z");
                        porosity_th[2] = param.get_double("Porosity threshold Z");
                        porosity_mu[2] = param.get_double("Porosity network constant Z");
                        porosity_gamma[2] = param.get_double("Porosity gamma network constant Z");
                        D_O2[2] = param.get_double("Oxygen diffusion coefficient Z");
                        D_wv[2] = param.get_double("Water vapour diffusion coefficient Z"); 
                        fibre_th[2] = param.get_double("Fibre network threshold Z");
                        fibre_mu[2] = param.get_double("Fibre network constant Z");
                    }
                }
                else
                {
                    for (unsigned int i =0; i < porosity_th.size(); i++)
                    {
                        matrix_electrical_conductivity[i][i] = param.get_double("Electric conductivity");
                        porosity_th[i] = param.get_double("Porosity threshold");
                        porosity_mu[i] = param.get_double("Porosity network constant");
                        porosity_gamma[i] = param.get_double("Porosity gamma network constant");
                        D_O2[i] = param.get_double("Oxygen diffusion coefficient");
                        D_wv[i] = param.get_double("Water vapour diffusion coefficient"); 
                        fibre_th[i] = param.get_double("Solid network threshold");
                        fibre_mu[i] = param.get_double("Solid network constant");	
                    }
                }
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();    
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::effective_gas_diffusivity(Table< 2, Tensor< 2, dim > > &D_eff) const
{
    D_eff.reinit(this->gases.size(),this->gases.size());    
    
    for (unsigned int i=0; i<this->gases.size(); i++)
    {
        for (unsigned int j=0; j<this->gases.size(); j++)
        {
            if( i != j )        // Not computing self diffusion coefficients
            {
                this->effective_gas_diffusivity( this->D_ECtheory(i,j), D_eff(i,j) );
            }
        }
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::effective_gas_diffusivity(const double& prop,
                                                double& prop_eff ) const
{
  // Assert if anisotropy is false
  AssertThrow(anisotropy == false, ExcMessage("DesignMPL::effective_transport_property_pores with double return value"
					 " can only be used for isotropic GDLs"));
  // Initialize variable
  prop_eff = 0;

  if (method_eff_property_pores == "Bruggemann")
    {
      // Most common method
      prop_eff = prop*pow(porosity, 1.5);
    }
  else if (method_eff_property_pores == "Percolation")
    {
      // Method used by M. Eikerling 
      // Note: For porosity_th[0] I use index 0 for convenience. They all contain the same value
      double aux = fabs(porosity - porosity_th[0])/(1 - porosity_th[0]);
      double step = 0.0;
      if (porosity >= porosity_th[0])
          step = 1.0;
      prop_eff = prop*(pow(aux,porosity_mu[0])*step);
      
    }
  else
    {
      std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
      abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::effective_gas_diffusivity(const double& prop,
                                                Tensor<2,dim>& prop_eff) const
{
    // Initialize variable
    prop_eff.clear();
    
    if (method_eff_property_pores == "Bruggemann")
    {
        // Most common method
        for (unsigned int i=0; i<dim; i++)
            prop_eff[i][i] = prop*pow(porosity, 1.5);
    }
    else if (method_eff_property_pores == "Percolation")
    {
        // Method used by M. Eikerling
        for (unsigned int i=0; i<dim; i++)
        {
            double aux = fabs(porosity - porosity_th[i])/(1 - porosity_th[i]);
            double step = 0.0;
            if (porosity >= porosity_th[i])
                step = 1.0;
            prop_eff[i][i] = prop*pow(aux,porosity_mu[i])*step;
        }
    }
    else 
    {
        std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::effective_electron_conductivity(double& prop_eff) const
{
  effective_transport_property_solid(electrical_conductivity, 
                                     prop_eff);
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::effective_electron_conductivity(Tensor<2,dim>& prop_eff) const
{
  effective_transport_property_solid(matrix_electrical_conductivity, 
                                     prop_eff);
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::effective_transport_property_solid(const double& prop, 
                                                         double& prop_eff) const
{
    // Asset if anisotropy is false
    Assert(anisotropy == false, ExcMessage("DesignMPL::effective_transport_property_pores with double return value"
    " can only be used for isotropic GDLs"));
    // Initialize variable
    prop_eff = 0.0;
    
    // In some papers prop_eff is give directly
    if (method_eff_property_fibres == "Given")
    {
        prop_eff = prop;
    }
    else if (method_eff_property_fibres == "Bruggemann")
        // Most common method
    {
        prop_eff = prop*pow(solid_phase, 1.5);
    }
    else if (method_eff_property_fibres == "Percolation")
        // Method used by M. Eikerling
    {
        unsigned int i = 0;
        double aux = fabs(solid_phase - fibre_th[i])/(1 - fibre_th[i]);
        double step = 1e-3; //this should be zero, but this would make the code diverge
        if (solid_phase >= fibre_th[i])
            step = 1.0;
        prop_eff = prop*pow(aux,fibre_mu[i])*step;
    }
    else
    {
        std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}
                                                         
//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignMPL<dim>::effective_transport_property_solid(const Tensor<2,dim>& prop, 
                                                         Tensor<2,dim>& prop_eff) const
{
    // Initialize variable
    prop_eff = 0;
    
    // In some papers prop_eff is give directly
    if (method_eff_property_fibres == "Given")
    {
        for (unsigned int i=0; i<dim; i++)
            prop_eff[i][i] = prop[i][i];
    }
    if (method_eff_property_fibres == "Bruggemann")
    {
        // Most common method
        for (unsigned int i=0; i<dim; i++)
            prop_eff[i][i] = prop[i][i]*pow(solid_phase, 1.5);
    }
    // Method used by M. Eikerling
    else if (method_eff_property_fibres == "Percolation")
    {
        for (unsigned int i=0; i<dim; i++)
        {
            double aux = fabs(solid_phase - fibre_th[i])/(1 - fibre_th[i]);
            double step = 1e-3; //this should be zero, but this would make the code diverge
            if (solid_phase >= fibre_th[i])
                step = 1.0;
            prop_eff[i][i] = prop[i][i]*pow(aux,fibre_mu[i])*step;
        }
    }
    else
    {
        std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// ------------ PRIVATE MEMBER FUNCTIONS
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
//-------------------- OPTIMIZATION ROUTINES --------------------------------
//---------------------------------------------------------------------------
/*
template <int dim>
Tensor<2,dim>
NAME::MicroPorousLayer<dim>::Deffective_transport_property_pores_Dporosity(const double prop) const
{

}
*/

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations. 
template class NAME::DesignMPL<deal_II_dimension>;
