//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: design_fibrous_GDL.cc
//    - Description: Class used to represent a fibrous GDL where effective properties are computed based on the porosity etc.
//    - Developers: M. Secanell (2009) and Madhur Bhaiya (2012)
//    - Id: $Id: design_fibrous_GDL.cc 1460 2013-08-30 22:28:49Z madhur $ 
//
//---------------------------------------------------------------------------

#include "layers/design_fibrous_GDL.h"

namespace NAME = FuelCellShop::Layer;

template <int dim>
const std::string NAME::DesignFibrousGDL<dim>::concrete_name ("DesignFibrousGDL");

template <int dim>
NAME::DesignFibrousGDL<dim> const* NAME::DesignFibrousGDL<dim>::PROTOTYPE = new NAME::DesignFibrousGDL<dim>();

//---------------------------------------------------------------------------
template <int dim>
NAME::DesignFibrousGDL<dim>::DesignFibrousGDL()
  : NAME::GasDiffusionLayer<dim> ()
{
    deallog<<" Register DesignFibrousGDL GDL to FactoryMap"<<std::endl;
    this->get_mapFactory()->insert(std::pair<std::string, GasDiffusionLayer<dim>* > (this->concrete_name, this) ); 
}

//---------------------------------------------------------------------------
template <int dim>
NAME::DesignFibrousGDL<dim>::DesignFibrousGDL(const std::string& name)
  : NAME::GasDiffusionLayer<dim> (name)
{
  deallog<<" of type DesignFibrousGDL"<<std::endl;
}


//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignFibrousGDL<dim>::declare_parameters (const std::string& name, 
                                                 ParameterHandler &param) const
{            
    
    FuelCellShop::Layer::GasDiffusionLayer<dim>::declare_parameters(name, param);
    
    /// Nothing extra to declare since all properties are hard-coded.
    
    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection(name);
        {
            param.enter_subsection(concrete_name); //-- Transport for the anisotropic case:
            {
                // General data
                param.declare_entry ("Porosity",  // volume fraction of void space in the GDL
                                     "0.3", // [-]
                                     Patterns::Double(),
                                     "Volume fraction of void space in the GDL");
                
                param.declare_entry ("Method effective transport properties in pores",
                                     "Bruggemann",
                                     Patterns::Selection("Given|Bruggemann|Percolation|Tomadakis|Mezedur"),
                                                         "Method used to compute effective transport properties in the void phase."
                                                         "Note the method defines the type of network, therefore the same method"
                                                         "is used even in the GDL is anisotropic");
                                     
                param.declare_entry ("Method effective transport properties in solid",
                                     "Bruggemann",
                                     Patterns::Selection("Given|Bruggemann|Percolation"),
                                                         "Method used to compute effective transport properties in solid"
                                                         "Note the method defines the type of network, therefore the same method"
                                                         "is used even in the GDL is anisotropic");
                
                //Electrical conductivity values                
                param.declare_entry ("Electrical conductivity",
                                     "100.", // [S/cm]
                                     Patterns::Double(),
                                     "Effective cond. if given is used, otherwise conducitivity of the raw material. Units [S/cm]");
                param.declare_entry ("Electrical conductivity X",
                                     "100.", // [S/cm]
                                     Patterns::Double(),
                                     "Component X of the electrical conductivity tensor. Units [S/cm]");
                param.declare_entry ("Electrical conductivity Y",
                                     "100.", // [S/cm]
                                     Patterns::Double(),
                                     "Component Y of the electrical conductivity tensor. Units [S/cm]");
                param.declare_entry ("Electrical conductivity Z",
                                     "100.", // [S/cm]
                                     Patterns::Double(),
                                     "Component Z of the electrical conductivity tensor. Units [S/cm]");
                
                //-- Transport for the isotropic case:
                param.declare_entry ("Porosity threshold",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of void space in the GDL."
                                     "If the porosity is less than this value transport does not occur");
                param.declare_entry ("Porosity network constant",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                param.declare_entry ("Porosity gamma network constant",
                                     "0.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory to account for extra diffusion");
                param.declare_entry ("Solid network threshold",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of solid in the GDL."
                                     "If the solid phase is less than this value transport in the solid network does not occur");
                param.declare_entry ("Solid network constant",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                
                
                //-- Transport for the anisotropic case:
                param.declare_entry ("Anisotropic transport",
                                     "false", // [S/cm]
                                     Patterns::Bool(),
                                     "Boolean variable. Set to true if we want to account for anisotropy of the GDL");
                
                //--- XX
                param.declare_entry ("Porosity threshold X",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of void space in the GDL."
                                     "If the porosity is less than this value transport does not occur");
                param.declare_entry ("Porosity network constant X",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                param.declare_entry ("Porosity gamma network constant X",
                                     "0.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory to account for extra diffusion");
                param.declare_entry ("Solid network threshold X",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of solid in the GDL."
                                     "If the solid phase is less than this value transport in the solid network does not occur");
                param.declare_entry ("Solid network constant X",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                
                // YY
                param.declare_entry ("Porosity threshold Y",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of void space in the GDL."
                                     "If the porosity is less than this value transport does not occur");
                param.declare_entry ("Porosity network constant Y",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                param.declare_entry ("Porosity gamma network constant Y",
                                     "0.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory to account for extra diffusion");
                param.declare_entry ("Solid network threshold Y",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of solid in the GDL."
                                     "If the solid phase is less than this value transport in the solid network does not occur");
                param.declare_entry ("Solid network constant Y",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                
                // ZZ
                param.declare_entry ("Porosity threshold Z",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of void space in the GDL."
                                     "If the porosity is less than this value transport does not occur");
                param.declare_entry ("Porosity network constant Z",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                param.declare_entry ("Porosity gamma network constant Z",
                                     "0.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory to account for extra diffusion");
                param.declare_entry ("Solid network threshold Z",  // volume fraction of void space in the GDL
                                     "0.12", // [-]
                                     Patterns::Double(),
                                     "Threshold value of the volume fraction of solid in the GDL."
                                     "If the solid phase is less than this value transport in the solid network does not occur");
                param.declare_entry ("Solid network constant Z",
                                     "2.0",
                                     Patterns::Double(),
                                     "Parameter used when using percolation theory");
                
            }
            param.leave_subsection();     
        }
        param.leave_subsection();        
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignFibrousGDL<dim>::initialize (ParameterHandler &param)
{
    NAME::GasDiffusionLayer<dim>::initialize(param);
    
    param.enter_subsection("Fuel cell data"); 
    {
        param.enter_subsection(this->name); 
        { 
            param.enter_subsection(concrete_name);
            {
                porosity = param.get_double("Porosity");
                solid_phase = 1 - porosity;
                method_eff_property_pores = param.get("Method effective transport properties in pores");
                method_eff_property_solid = param.get("Method effective transport properties in solid");
                
                // Anisotropy
                anisotropy = param.get_bool ("Anisotropic transport");
                
                porosity_th.resize(3);
                porosity_mu.resize(3);
                porosity_gamma.resize(3);
                solid_th.resize(3);
                solid_mu.resize(3);
                
                if (anisotropy == true)
                {
                    // X
                    porosity_th[0] = param.get_double("Porosity threshold X");
                    porosity_mu[0] = param.get_double("Porosity network constant X");
                    porosity_gamma[0] = param.get_double("Porosity gamma network constant X");
                    solid_th[0] = param.get_double("Solid network threshold X");
                    solid_mu[0] = param.get_double("Solid network constant X");
                    // Y
                    porosity_th[1] = param.get_double("Porosity threshold Y");
                    porosity_mu[1] = param.get_double("Porosity network constant Y");
                    porosity_gamma[1] = param.get_double("Porosity gamma network constant Y");
                    solid_th[1] = param.get_double("Solid network threshold Y");
                    solid_mu[1] = param.get_double("Solid network constant Y");
                    // Z
                    porosity_th[2] = param.get_double("Porosity threshold Z");
                    porosity_mu[2] = param.get_double("Porosity network constant Z");
                    porosity_gamma[2] = param.get_double("Porosity gamma network constant Z");
                    solid_th[2] = param.get_double("Solid network threshold Z");
                    solid_mu[2] = param.get_double("Solid network constant Z");
                }
                else
                {
                    for (unsigned int i =0; i < porosity_th.size(); i++)
                    {
                        porosity_th[i] = param.get_double("Porosity threshold");
                        porosity_mu[i] = param.get_double("Porosity network constant");
                        porosity_gamma[i] = param.get_double("Porosity gamma network constant");
                        solid_th[i] = param.get_double("Solid network threshold");
                        solid_mu[i] = param.get_double("Solid network constant");
                    }
                }
                //electronic properties
                sigma_e.resize(3);
                
                if (anisotropy == true)
                {
                    // X
                    sigma_e[0] = param.get_double("Electrical conductivity X");
                    
                    // Y
                    sigma_e[1] = param.get_double("Electrical conductivity Y");
                    
                    // Z
                    sigma_e[2] = param.get_double("Electrical conductivity Z");
                }
                else
                {
                    for (unsigned int i =0; i < this->porosity_th.size(); i++)
                    {
                        sigma_e[i] = param.get_double("Electrical conductivity");
                    }
                }
            }
            param.leave_subsection(); 
        }
        param.leave_subsection(); 
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignFibrousGDL<dim>::effective_gas_diffusivity(const double& prop,
                                                       double& prop_eff) const
{
    // Asset if anisotropy is false
    AssertThrow(anisotropy == false, ExcMessage("FibrousGDL::effective_gas_diffusivity with double return value"
                                                " can only be used for isotropic GDLs"));
    // Initialize variable
    prop_eff = 0.0;
    // In some papers prop_eff is give directly
    if (method_eff_property_pores == "Given")
    {
        Assert(false, ExcMessage("Given transport values can only be used with DummyGDL::effective_gas_diffusivity"));
    }
    if (method_eff_property_pores == "Bruggemann")
    {
        // Most common method
        prop_eff = prop*pow(porosity, 1.5);
    }
    else if (method_eff_property_pores == "Percolation")
    {
		// Method used by M. Eikerling
		double aux = fabs(porosity - porosity_th[0])/(1 - porosity_th[0]);
		double step = 0.0;
		if (porosity >= porosity_th[0])
			step = 1.0;
		prop_eff = prop*((1-porosity_gamma[0])*pow(aux,porosity_mu[0])*step + porosity_gamma[0]);
	}
	else if (method_eff_property_pores == "Tomadakis")
	{
		// Method  reported by Tomadakis and Sotrichos
		double aux = (porosity - porosity_th[0])/(1 - porosity_th[0]);
		double step = 0.0;
		if (porosity >= porosity_th[0])
			step = 1.0;
		prop_eff = prop*porosity*pow(aux,porosity_mu[0])*step;
	}
	else if (method_eff_property_pores == "Mezedur")
	{
		// Method reported in Mezedur et al., "Effects of poer structure, randomness
		// and size on effective mass diffusivity", AIChE J. 48 (2002) pp15-24
		prop_eff = prop*(1 - pow(1-porosity, 0.46));
	}
	else
	{
		std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
		abort();
	}
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignFibrousGDL<dim>::effective_gas_diffusivity(const double& prop,
                                                       Tensor<2,dim>& prop_eff) const
{
    prop_eff.clear(); // Initialize all entries to zero in the Tensor
    
    // In some papers prop_eff is give directly
    if (method_eff_property_pores == "Given")
    {
        Assert(false, ExcMessage("Given transport values can only be used with DummyGDL::effective_gas_diffusivity"));
    }
    // Most common method
    if (method_eff_property_pores == "Bruggemann")
    {
        for (unsigned int i=0; i<dim; i++)
            prop_eff[i][i] = prop*pow(porosity, 1.5);
    }
    // Method used by M. Eikerling
    else if (method_eff_property_pores == "Percolation")
    {
        for (unsigned int i=0; i<dim; i++)
        {
            double aux = fabs(porosity - porosity_th[i])/(1 - porosity_th[i]);
            double step = 0.0;
            if (porosity >= porosity_th[i])
                step = 1.0;
            prop_eff[i][i] = prop*((1-porosity_gamma[i])*pow(aux,porosity_mu[i])*step + porosity_gamma[i]);
        }
    }
    // Method  reported by Tomadakis and Sotrichos
    else if (method_eff_property_pores == "Tomadakis")
    {
        for (unsigned int i=0; i<dim; i++)
        {
            double aux = (porosity - porosity_th[i])/(1 - porosity_th[i]);
            double step = 0.0;
            if (porosity >= porosity_th[i])
                step = 1.0;
            prop_eff[i][i] = prop*porosity*pow(aux,porosity_mu[i])*step;
        }
    }
    // Method reported in Mezedur et al., "Effects of poer structure, randomness
    // and size on effective mass diffusivity", AIChE J. 48 (2002) pp15-24
    else if (method_eff_property_pores == "Mezedur")
    {
        for (unsigned int i=0; i<dim; i++)
            prop_eff[i][i] = prop*(1 - pow(1-porosity, 0.46));
    }
    else
    {
        std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
        abort();
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DesignFibrousGDL<dim>::effective_gas_diffusivity(Table< 2, Tensor<2,dim> > &D_eff) const
{
    D_eff.reinit(this->gases.size(),this->gases.size());
    
    for (unsigned int i=0; i<this->gases.size(); i++)
    {
        for (unsigned int j=0; j<this->gases.size(); j++)
        {
            if( i != j )        // Not computing self diffusion coefficients
            {
                this->effective_gas_diffusivity( this->D_ECtheory(i,j), D_eff(i,j) );
            }
        }
    }
    
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::DesignFibrousGDL<dim>::effective_electron_conductivity(double& prop_eff) const
{
	// Asset if anisotropy is false
	AssertThrow(anisotropy == false, ExcMessage("FibrousGDL::effective_transport_property_pores with double return value"
	" can only be used for isotropic GDLs"));
	// Initialize variable
	prop_eff = 0.0;
	const double prop = sigma_e[0];
	
	// In some papers prop_eff is give directly
	if (method_eff_property_solid == "Given")
	{
		Assert(false, ExcMessage("Given transport values can only be used with DummyGDL::effective_electron_conductivity"));
	}
	// Most common method
	else if (method_eff_property_solid == "Bruggemann")
	{
		prop_eff = prop*pow(solid_phase, 1.5);
		
	}
	// Method used by M. Eikerling
	else if (method_eff_property_solid == "Percolation")
	{
		double aux = fabs(solid_phase - solid_th[0])/(1 - solid_th[0]);
		double step = 1e-3; //this should be zero, but this would make the code diverge
		if (solid_phase >= solid_th[0])
			step = 1.0;
		prop_eff = prop*pow(aux,solid_mu[0])*step;
	}
	else
	{
		std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
		abort();
	}
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::DesignFibrousGDL<dim>::effective_electron_conductivity(Tensor<2,dim>& prop_eff) const
{
	// Initialize variable
	prop_eff.clear(); // Initialize all entries to zero in the Tensor
	
	const std::vector<double> prop = sigma_e;
	Assert(prop.size() == 3, ExcMessage("Prop should contain the conductivity in the X, Y and Z dimension"));

	// In some papers prop_eff is give directly
	if (method_eff_property_solid == "Given")
	{
		for (unsigned int i=0; i<dim; i++)
			prop_eff[i][i] = prop[i];
	}
	// Most common method
	else if (method_eff_property_solid == "Bruggemann")
	{
		for (unsigned int i=0; i<dim; i++)
			prop_eff[i][i] = prop[i]*pow(solid_phase, 1.5);
	}
	// Method used by M. Eikerling
	else if (method_eff_property_solid == "Percolation")
	{
		for (unsigned int i=0; i<dim; i++)
		{
			double aux = fabs(solid_phase - solid_th[i])/(1 - solid_th[i]);
			double step = 1e-3; //this should be zero, but this would make the code diverge
			if (solid_phase >= solid_th[i])
				step = 1.0;
			prop_eff[i][i] = prop[i]*pow(aux,solid_mu[i])*step;
		}
	}
	else
	{
		std::cout << "Unknown method to compute effective conduction in "<<__FILE__ <<" line "<<__LINE__<<std::endl;
		abort();
	}
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations. 
template class NAME::DesignFibrousGDL<deal_II_dimension>;
