//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: dummy_cl.cc
//    - Description: Class characterizing the conventional catalyst layer and defining constant effective properties
//    - Developers: Marc Secanell (2011) and Madhur Bhaiya (2013)
//    - Id: $Id: dummy_CL.cc 1417 2013-08-30 16:27:42Z secanell $ 
//
//---------------------------------------------------------------------------

#include "layers/dummy_CL.h"

namespace NAME = FuelCellShop::Layer;

template <int dim>
const std::string NAME::DummyCL<dim>::concrete_name ("DummyCL");

template <int dim>
NAME::DummyCL<dim> const* NAME::DummyCL<dim>::PROTOTYPE = new NAME::DummyCL<dim>();

//---------------------------------------------------------------------------
template <int dim>
NAME::DummyCL<dim>::DummyCL()
  : NAME::CatalystLayer<dim> ()
{
    //deallog<<" Register DummyCL CL to FactoryMap"<<std::endl;
    this->get_mapFactory()->insert(std::pair<std::string, FuelCellShop::Layer::CatalystLayer<dim>* >(concrete_name, this));
}

//---------------------------------------------------------------------------
template <int dim>
NAME::DummyCL<dim>::DummyCL(const std::string& name)
  : NAME::CatalystLayer<dim> (name)
{
  deallog<<" of type DummyCL"<<std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
NAME::DummyCL<dim>::~DummyCL()
{}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::initialize (ParameterHandler &param)
{
    NAME::CatalystLayer<dim>::initialize(param);
    
    param.enter_subsection("Fuel cell data"); 
    {
        param.enter_subsection(this->name); 
        {
            param.enter_subsection(concrete_name); //-- Transport for the anisotropic case:
            {
                D_O2 = param.get_double("Oxygen diffusion coefficient, [cm^2/s]");
                D_wv = param.get_double("Water vapour diffusion coefficient, [cm^2/s]");
                sigma_e = param.get_double("Electrical conductivity, [S/cm]");
                sigma_m = param.get_double("Protonic conductivity, [S/cm]");
                Av = param.get_double("Active area [cm^2/cm^3]");
            }
            param.leave_subsection(); 
        }
        param.leave_subsection(); 
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::set_parameters (const std::vector<std::string>& name_dvar,
                                    const std::vector<double>& value_dvar,
                                    const std::string& cl_section_name, 
                                    ParameterHandler &param) const
{
    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection(cl_section_name);
        {
            param.enter_subsection(concrete_name); //-- Transport for the anisotropic case:
            {
                for (unsigned int i=0; i<name_dvar.size(); ++i)
                {
                    if (name_dvar[i] == "D_O2_CCL")
                    {
                        if(this->name == "Cathode catalyst layer")
                            param.set("Oxygen diffusion coefficient, [cm^2/s]", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "D_H2O_CCL")
                    {
                        if(this->name == "Cathode catalyst layer")
                            param.set("Water vapour diffusion coefficient, [cm^2/s]", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "D_H2O_ACL")
                    {
                        if(this->name == "Anode catalyst layer")
                            param.set("Water vapour diffusion coefficient, [cm^2/s]", value_dvar[i]);
                    }
                    else if (name_dvar[i] == "sigma_e_CL")
                    {
                        param.set("Electrical conductivity, [S/cm]", value_dvar[i]);
                    }	
                    else if (name_dvar[i] == "sigma_m_CL")
                    {
                        param.set("Protonic conductivity, [S/cm]", value_dvar[i]);
                    }
                }
            }
            param.leave_subsection();
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::effective_gas_diffusivity(Table< 2, Tensor< 2, dim > >& prop_eff) const
{
    
    prop_eff.reinit(this->gases.size(),this->gases.size());
    
    for (unsigned int i = 0; i<this->gases.size(); i++)
    {
        std::string name_i;
        name_i = this->gases[i]->get_name();
        
        for (unsigned int j = i+1; j<this->gases.size(); j++)
        {  
            std::string name_j;
            name_j = this->gases[j]->get_name();
            
            if (name_i.compare("oxygen") == 0)
            {
                
                if (name_j.compare("nitrogen") == 0)
                {
                    for (unsigned int d = 0; d<dim; d++)
                    {
                        prop_eff(i,j)[d][d] = Units::convert(D_O2, Units::UNIT2, Units::C_UNIT2);
                        prop_eff(j,i)[d][d] = prop_eff(i,j)[d][d];
                    }
                }
                else
                {
                    deallog << "Species "<< name_j << " diffusivity requested in"
                    << "DummyCL<dim>::effective_gas_diffusivity"
                    <<"not implemented"<< std::endl;
                    exit(1);
                }
                
            }
            else if (name_i.compare("water") == 0)
            {
                
                if (name_j.compare("nitrogen") == 0)
                {
                    for (unsigned int d = 0; d<dim; d++)
                    {
                        prop_eff(i,j)[d][d] = Units::convert(D_wv, Units::UNIT2, Units::C_UNIT2);
                        prop_eff(j,i)[d][d] = prop_eff(i,j)[d][d];
                        break;
                    }
                }
                else
                {
                    deallog << "Species "<< name_j << " diffusivity requested in"
                    << "DummyCL<dim>::effective_gas_diffusivity"
                    <<"not implemented"<< std::endl;
                    exit(1);
                }
            }
            else
            {
                deallog << "Species "<< name_i << " diffusivity requested in"
                << "DummyCL<dim>::effective_gas_diffusivity"
                <<"not implemented"<< std::endl;
                exit(1);
            } 
        }
    }
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::effective_electron_conductivity(double& prop_eff) const
{
    prop_eff = sigma_e;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::effective_electron_conductivity(Tensor<2,dim>& prop_eff) const
{
    prop_eff.clear();
    for (unsigned int i=0; i<dim; i++)
        prop_eff[i][i] = sigma_e;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::effective_proton_conductivity(double& prop_eff) const
{
    prop_eff = sigma_m;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::effective_proton_conductivity(std::vector<double>& prop_eff) const
{
    Assert( prop_eff.size() != 0, ExcMessage("Input vector should be initialized before passing to this method.") );
    for (unsigned int q = 0; q < prop_eff.size(); ++q)
        prop_eff[q] = sigma_m;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::derivative_effective_proton_conductivity(std::map< VariableNames, std::vector<double> >& prop_eff) const
{
    Assert( false, ExcNotImplemented());
    // Issue of internal vector size. Normally not required to be initialized inside the conventional CL but how to tackle it here..Madhur
    //Assert( this->derivative_flags.size()!=0, ExcMessage("set_derivative_flags has not been probably called in DummyCL::derivative_effective_proton_conductivity."));    
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::current_density(std::vector<double>& coef)
{
    this->kinetics->current_density(coef);
    for (unsigned int i = 0; i<coef.size(); ++i)
        coef[i] *= Av;
}

//---------------------------------------------------------------------------
template <int dim>
void
NAME::DummyCL<dim>::derivative_current_density(std::map< VariableNames, std::vector<double> >& dcoef_du)
{
    this->kinetics->derivative_current(dcoef_du);
    for (std::map< VariableNames, std::vector<double> >::iterator iter=dcoef_du.begin(); iter!=dcoef_du.end(); ++iter)
        for (unsigned int q=0; q<iter->second.size(); ++q)
            iter->second[q] *= Av;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations. 
template class NAME::DummyCL<deal_II_dimension>;