//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: porous_layer.cc
//    - Description: Child of base layer that implements common functionality for handling gases.
//    - Developers: M. Secanell and Madhur Bhaiya
//    - $Id: porous_layer.cc 1373 2013-08-21 00:34:53Z madhur $
//
//---------------------------------------------------------------------------

#include "layers/porous_layer.h"

namespace NAME = FuelCellShop::Layer;

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::PorousLayer<dim>::initialize (ParameterHandler &param)
{
    NAME::BaseLayer<dim>::initialize(param);
}
//---------------------------------------------------------------------------
template <int dim>
void 
NAME::PorousLayer<dim>::set_gases_and_compute (std::vector<FuelCellShop::Material::PureGas*>& gases_in,
                                               const double& pressure_in,
                                               const double& temperature_in)
{
    Assert(gases_in.size() >= 2, ExcMessage("Number of gases should be more than or equal to two in PorousLayer::set_gases_and_compute method."));
    
    this->gases = gases_in;
    this->pressure = pressure_in;
    this->temperature = temperature_in;

    this->D_ECtheory.reinit(gases_in.size(), gases_in.size());
    FuelCellShop::Mixture::ChapmanEnskog CE;
    CE.set_T_and_p(temperature,pressure);
    
    this->dD_ECtheory_dx.resize(this->derivative_flags.size() );
    for (unsigned int d = 0; d< this->derivative_flags.size(); ++d)
        this->dD_ECtheory_dx[d].reinit(gases_in.size(), gases_in.size());
    
    for  (unsigned int i = 0; i<gases_in.size(); ++i)
    {
        for (unsigned int j = 0; j<gases_in.size(); ++j)
        {
            CE.set_species(*(gases_in[i]),*(gases_in[j]));
            CE.compute_diffusion(this->D_ECtheory(i,j));
            for (unsigned int d = 0; d< this->derivative_flags.size(); ++d)
            {
                if (this->derivative_flags[d] == total_pressure)
                    this->dD_ECtheory_dx[d](i,j) = CE.compute_dD_dP (temperature,pressure);
                else if (this->derivative_flags[d] == temperature_of_REV)
                    this->dD_ECtheory_dx[d](i,j) = CE.compute_dD_dT (temperature,pressure);
                else
                    AssertThrow(false, ExcNotImplemented());
            }
        }
    }
    
    
};

//---------------------------------------------------------------------------
template <int dim>
void
NAME::PorousLayer<dim>::compute_gas_diffusion(FuelCellShop::Material::PureGas* solute_gas,
                                              FuelCellShop::Material::PureGas* solvent_gas)
{
    Assert( this->pressure != 0., ExcMessage("Pressure is not set inside the porous layer. Use set_gases method in initialization of the application.") );
    Assert( this->T_vector.size()!=0, ExcMessage("T_vector is not set inside the porous layer. Use set_T_vector method before calling this function.") );
    Assert( std::find(this->gases.begin(), this->gases.end(), solute_gas)!=this->gases.end(), ExcMessage("Solute gas not set using set_gases method in initialization of the application.") );
    Assert( std::find(this->gases.begin(), this->gases.end(), solvent_gas)!=this->gases.end(), ExcMessage("Solvent gas not set using set_gases method in initialization of the application.") );
       
    FuelCellShop::Mixture::ChapmanEnskog CE;
    CE.set_species(*solute_gas,*solvent_gas);
    CE.set_p(this->pressure);
    CE.set_temperature(this->T_vector);
    CE.compute_diffusion(this->D_bulk);
    CE.compute_dD_dT(this->dD_bulk_dT);
}
//---------------------------------------------------------------------------
template <int dim>
void 
NAME::PorousLayer<dim>::get_gas_index(FuelCellShop::Material::PureGas* gas_type,
                                      int& index) const
{
    Assert( std::find(this->gases.begin(), this->gases.end(), gas_type)!=this->gases.end(), ExcMessage("Gas not set using set_gases_and_compute method in initialization of the application.") );
    
    for (unsigned int i=0; i<this->gases.size(); i++)
    {
        if (this->gases[i]->get_name() == gas_type->get_name())
        {
            index = i;
            break;
        } 
    }
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::PorousLayer<dim>::gas_diffusion_coefficients(Table< 2, double > &D_eff ) const
{
    D_eff = this->D_ECtheory;
}

//---------------------------------------------------------------------------
template <int dim>
void 
NAME::PorousLayer<dim>::derivative_gas_diffusion_coefficients(std::vector<Table< 2, double > > &dD_dx) const
{
    dD_dx = this->dD_ECtheory_dx;
}

//---------------------------------------------------------------------------
template <int dim>
bool 
NAME::PorousLayer<dim>::test_layer() const
{
  double pressure(1.0); //atm 
  double temperature(273); //K
  FuelCellShop::Layer::PorousLayer<dim> test("test");
  FuelCellShop::Material::Oxygen oxygen;
  FuelCellShop::Material::Nitrogen nitrogen;
  deallog<<"Computing diffusion coefficient of: "<<oxygen.get_name()<<" in "<<nitrogen.get_name()<<" at "<<pressure<<" atm and "<<temperature<<" Kelvin"<<std::endl;
  std::vector<FuelCellShop::Material::PureGas*> gases;
  gases.push_back(&oxygen);
  gases.push_back(&nitrogen);
  std::vector<VariableNames> deriv_flags;
  deriv_flags.push_back(total_pressure);
  deriv_flags.push_back(temperature_of_REV);
  test.set_derivative_flags(deriv_flags);
  test.set_gases_and_compute(gases, pressure, temperature);
  Table<2, double> D_eff;
  test.gas_diffusion_coefficients(D_eff);
  std::vector<Table< 2, double > > dD_dx;
  test.derivative_gas_diffusion_coefficients(dD_dx);
  
  deallog<<"The diffusion coefficients are: D_11="<<D_eff(0,0)<<" D_12 = "<<D_eff(0,1)<<"m2/s"<<std::endl;  
  deallog<<"The diffusion coefficients are: D_21="<<D_eff(1,0)<<" D_22 = "<<D_eff(1,1)<<"m2/s"<<std::endl;  
  deallog<<"The experimental value at 273K and 1atm is: 0.181 cm2/s"<<std::endl;
  
  deallog<<"The diffusion coefficients are: dD_11_dp="<<dD_dx[0](0,0)<<" dD_12_dp = "<<dD_dx[0](0,1)<<"m2/s"<<std::endl;  
  deallog<<"The diffusion coefficients are: dD_21_dp="<<dD_dx[0](1,0)<<" dD_22_dp = "<<dD_dx[0](1,1)<<"m2/s"<<std::endl;
  
  deallog<<"The diffusion coefficients are: dD_11_dT="<<dD_dx[1](0,0)<<" dD_12_dT = "<<dD_dx[1](0,1)<<"m2/s"<<std::endl;  
  deallog<<"The diffusion coefficients are: dD_21_dT="<<dD_dx[1](1,0)<<" dD_22_dT = "<<dD_dx[1](1,1)<<"m2/s"<<std::endl;
  
  return true;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantiations.
template class NAME::PorousLayer<deal_II_dimension>;