//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: BinaryDiffusion.cc
//    - Description: Class storing methods compute binary diffusion coefficients
//    - Developers: Ali MalekpourKoupaei, M. Secanell and Madhur Bhaiya
//    - $Id: BinaryDiffusion.cc 1354 2013-08-17 00:01:22Z secanell $
//
//---------------------------------------------------------------------------

#include "materials/BinaryDiffusion.h"

//---------------------------------------------------------------------------
FuelCellShop::Mixture::BinaryDiffusion::BinaryDiffusion()
//---------------------------------------------------------------------------
{
  species1 = species2 = __null;
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::BinaryDiffusion::BinaryDiffusion(FuelCellShop::Material::PureGas& species1,
                                                        FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
  this->species1 = &species1;
  this->species2 = &species2;
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::BinaryDiffusion::~BinaryDiffusion()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void 
FuelCellShop::Mixture::BinaryDiffusion::set_T_and_p(const double& T,const double& p)
//---------------------------------------------------------------------------
{
  this->T = T;
  this->p = p;
}

//---------------------------------------------------------------------------
void 
FuelCellShop::Mixture::BinaryDiffusion::set_p(const double& p_atm)
//---------------------------------------------------------------------------
{
  this->p = p_atm;
}

//---------------------------------------------------------------------------
void
FuelCellShop::Mixture::BinaryDiffusion::set_temperature(const FuelCellShop::SolutionVariable& T_in)
{
  T_vector = T_in;
}

//---------------------------------------------------------------------------
void 
FuelCellShop::Mixture::BinaryDiffusion::set_species(FuelCellShop::Material::PureGas& species1,
                                                    FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
  this->species1 = &species1;
  this->species2 = &species2;
}

//---------------------------------------------------------------------------
void 
FuelCellShop::Mixture::BinaryDiffusion::scan()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void 
FuelCellShop::Mixture::BinaryDiffusion::report()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::BinaryDiffusion::report2file(char* filename)
//---------------------------------------------------------------------------
{
  std::ofstream out(filename);
  out << "Method name: " << get_name() << "\n";
  out << "Method ID: " << get_ID() << std::endl;
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::VaporDiffusion::VaporDiffusion()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::VaporDiffusion::VaporDiffusion
	(FuelCellShop::Material::PureGas& species1, FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::VaporDiffusion::~VaporDiffusion ()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::VaporDiffusion::scan()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::VaporDiffusion::report()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::VaporDiffusion::report2file(char* filename)
//---------------------------------------------------------------------------
{
}


//---------------------------------------------------------------------------
//CHAPMAN-ENSKOG METHOD IMPLEMENTATION---------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
FuelCellShop::Mixture::ChapmanEnskog::ChapmanEnskog()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::ChapmanEnskog::ChapmanEnskog(FuelCellShop::Material::PureGas& species1, 
					   FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
  this->species1 = &species1;
  this->species2 = &species2;
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::ChapmanEnskog::~ChapmanEnskog()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::ChapmanEnskog::get_name()
//---------------------------------------------------------------------------
{
  return "Chapman Enskog";
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::ChapmanEnskog::get_description()
//---------------------------------------------------------------------------
{
  return "this model is for non-polar or polar-nonpolar pairs";
}

//---------------------------------------------------------------------------
enDiffusionModelID 
FuelCellShop::Mixture::ChapmanEnskog::get_ID()
//---------------------------------------------------------------------------
{
  return dmChapmanEnskog;
}

//---------------------------------------------------------------------------
double 
FuelCellShop::Mixture::ChapmanEnskog::getSikmaAB(const double& sikma_A,
					const double& sikma_B)
//---------------------------------------------------------------------------
{
  return 0.5 * (sikma_A + sikma_B);
}

//---------------------------------------------------------------------------
double 
FuelCellShop::Mixture::ChapmanEnskog::getTstarAB(const double& Tstar_A,
					const double& Tstar_B)
//---------------------------------------------------------------------------
{
  return __pow(Tstar_A * Tstar_B,0.5);
}

//---------------------------------------------------------------------------
double 
FuelCellShop::Mixture::ChapmanEnskog::getOmegaAB(const double& Tstar)
//---------------------------------------------------------------------------
{
  return 1.06036/__pow(Tstar,0.15610)+
          0.19300/__builtin_exp(0.47635*Tstar)+
            1.03587/__builtin_exp(1.52996*Tstar)+
              1.76474/__builtin_exp(3.89411*Tstar);
}

//---------------------------------------------------------------------------
void 
FuelCellShop::Mixture::ChapmanEnskog::compute_diffusion (double& D)
{
  D = compute_diffusion (T,p);
}

//---------------------------------------------------------------------------
void
FuelCellShop::Mixture::ChapmanEnskog::compute_diffusion (std::vector<double>& D_vec)
{
  Assert( T_vector.size() != 0, dealii::StandardExceptions::ExcMessage("T_vector has not been initialized yet. Call set_temperature before using this method for computing diffusion coefficients.") );

  D_vec.resize( T_vector.size() );
  
  for (unsigned int i = 0; i < T_vector.size(); ++i)
	  D_vec[i] = compute_diffusion( T_vector[i], p );
}

//---------------------------------------------------------------------------
double 
FuelCellShop::Mixture::ChapmanEnskog::compute_diffusion (const double& T, 
						const double& p)
//---------------------------------------------------------------------------
{

  M_A      = species1->M;
  M_B      = species2->M;

  sikma_A  = species1->getCollisionDiameter();
  sikma_B  = species2->getCollisionDiameter();

  Tstar_A  = species1->getDimensionlessTemperature(T);
  Tstar_B  = species2->getDimensionlessTemperature(T);

  //combining rules
  sikma_AB  = getSikmaAB(sikma_A,sikma_B);
  Tstar_AB  = getTstarAB(Tstar_A, Tstar_B);
  omega_DAB = getOmegaAB(Tstar_AB);

  return 1.8583E-7*__pow(T*T*T*(1.0/M_A+1.0/M_B),0.5)/p/sikma_AB/sikma_AB/omega_DAB;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::ChapmanEnskog::compute_diffusion (double& D,std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void 
FuelCellShop::Mixture::ChapmanEnskog::compute_diffusion (const double& T, const double& p,
						double& D,std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::ChapmanEnskog::compute_dD (std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::ChapmanEnskog::compute_dD_dP (const double& T, const double& p)
//---------------------------------------------------------------------------
{
  set_T_and_p(T,p);

  M_A      = species1->M;
  M_B      = species2->M;

  sikma_A  = species1->getCollisionDiameter();
  sikma_B  = species2->getCollisionDiameter();

  Tstar_A  = species1->getDimensionlessTemperature(T);
  Tstar_B  = species2->getDimensionlessTemperature(T);

  //combining rules
  sikma_AB  = getSikmaAB(sikma_A,sikma_B);
  Tstar_AB  = getTstarAB(Tstar_A, Tstar_B);
  omega_DAB = getOmegaAB(Tstar_AB);

  return -1.8583E-7*__pow(T*T*T*(1.0/M_A+1.0/M_B),0.5)/p/p/
    sikma_AB/sikma_AB/omega_DAB;
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::ChapmanEnskog::compute_dD_dT (const double& T, const double& p)
//---------------------------------------------------------------------------
{
  set_T_and_p(T,p);

  M_A      = species1->M;
  M_B      = species2->M;

  sikma_A  = species1->getCollisionDiameter();
  sikma_B  = species2->getCollisionDiameter();

  Tstar_A  = species1->getDimensionlessTemperature(T);
  Tstar_B  = species2->getDimensionlessTemperature(T);

  //combining rules
  sikma_AB  = getSikmaAB(sikma_A,sikma_B);
  Tstar_AB  = getTstarAB(Tstar_A, Tstar_B);
  omega_DAB = getOmegaAB(Tstar_AB);

  return 1.8583E-7*1.5*__pow(T*(1.0/M_A+1.0/M_B),0.5)/p/sikma_AB/sikma_AB/omega_DAB;
}

//---------------------------------------------------------------------------
void
FuelCellShop::Mixture::ChapmanEnskog::compute_dD_dT(std::vector<double>& dD_dT_vec)
{
  Assert( T_vector.is_initialized(), dealii::StandardExceptions::ExcMessage("T_vector has not been initialized yet. Call set_temperature before using this method for computing dD_dT.") );

  dD_dT_vec.resize( T_vector.size() );
  
  for (unsigned int i = 0; i < T_vector.size(); ++i)
	  dD_dT_vec[i] = compute_dD_dT( T_vector[i], p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::ChapmanEnskog::scan()
//---------------------------------------------------------------------------
{
  __printf("\nplease enter temperature in Kelvin:");   
  fflush(stdin);
  __scanf("%lf",&T);

  __printf("please enter pressure in Atmosphere:");    
  fflush(stdin);
  __scanf("%lf",&p);

  __printf("\n");
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::ChapmanEnskog::report()
//---------------------------------------------------------------------------
{
  __printf("\n");
  __printf("\t-----------------------------------------------------------\n");
  __printf("\t|      |M       |Vl     |Tnb    |sikma  |Tstar   |omega   |\n");
  __printf("\t|------|--------|-------|-------|-------|--------|--------|\n");
  __printf("\t|%s    |%-5.5f|%-5.5f|%-5.4f|%-5.5f|%-5.5f |   -    |\n",species1->get_formula(),M_A,Vl_A,Tnb_A,sikma_A,Tstar_A);
  __printf("\t|------|--------|-------|-------|-------|--------|--------|\n");
  __printf("\t|%s    |%-5.5f|%-5.5f|%-5.4f|%-5.5f|%-5.5f |   -    |\n",species2->get_formula(),M_B,Vl_B,Tnb_B,sikma_B,Tstar_B);
  __printf("\t|------|--------|-------|-------|-------|--------|--------|\n");
  __printf("\t|Binary|   -    |   -   |   -   |%-5.5f|%-5.5f |%-5.5f |\n",sikma_AB,Tstar_AB,omega_DAB);
  __printf("\t-----------------------------------------------------------\n");
  __printf("\tConditions:\n");
  __printf("\t  Temperature:%e[K]\n",T);
  __printf("\t  Pressure:%e[atm]\n\n",p);
  __printf("\tsikma: collision diameter[nm]\n");
  __printf("\tTstar: dimensionless temperature (k*T/e)\n");
  __printf("\tomega: collision integral\n\n");
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::ChapmanEnskog::report2file(char* filename)
//---------------------------------------------------------------------------
{
}



//---------------------------------------------------------------------------
FuelCellShop::Mixture::LiquidDiffusion::LiquidDiffusion()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::LiquidDiffusion::LiquidDiffusion
	(FuelCellShop::Material::PureGas& species1, FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::LiquidDiffusion::~LiquidDiffusion ()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::LiquidDiffusion::scan()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::LiquidDiffusion::report()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::LiquidDiffusion::report2file(char* filename)
//---------------------------------------------------------------------------
{
}



//---------------------------------------------------------------------------
FuelCellShop::Mixture::WilkeChang::WilkeChang()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::WilkeChang::WilkeChang(FuelCellShop::Material::PureGas& species1, 
  FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::WilkeChang::~WilkeChang()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::WilkeChang::get_name()
//---------------------------------------------------------------------------
{
  return "Wilke Change";
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::WilkeChang::get_description()
//---------------------------------------------------------------------------
{
  return "this model predicts the diffusivity in the dilute liquid systems";
}

//---------------------------------------------------------------------------
enDiffusionModelID FuelCellShop::Mixture::WilkeChang::get_ID()
//---------------------------------------------------------------------------
{
  return dmWilkeChang;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::compute_diffusion (double& D)
//---------------------------------------------------------------------------
{
  D = compute_diffusion(T,p);
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::WilkeChang::compute_diffusion (const double& T, const double& p)
//---------------------------------------------------------------------------
{
  get_phi(phi);
  double& Vl = species1->Vl;
  double& Mb = species2->M;
  species2->get_viscosity(mu,T);

#ifdef _DETAILED_REPORT_
  printf("\n[WARNING]:pressure value ignored for %s.\n\n",get_name());
#endif
  
  return 117.3E-18 * __pow(phi * Mb,0.5)
	*T/mu/__pow(Vl,0.6); //m2/sec;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::compute_diffusion (double& D, std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::compute_diffusion (const double& T,const double& p,
    double& D,std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::compute_dD (std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::WilkeChang::compute_dD_dP (const double& T, 
	const double& p)
//---------------------------------------------------------------------------
{
  return 0.0;
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::WilkeChang::compute_dD_dT (const double& T, 
	const double& p)
//---------------------------------------------------------------------------
{
  get_phi(phi);
  double& Vl = species1->Vl;
  double& Mb = species2->M;
  species2->get_viscosity(mu,T);

#ifdef _DETAILED_REPORT_
  printf("\n[WARNING]:pressure value ignored for %s.\n\n",get_name());
#endif
  
  return 117.3E-18 * __pow(phi * Mb,0.5)
	/mu/__pow(Vl,0.6); //m2/sec;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::get_phi(double& phi)
//---------------------------------------------------------------------------
{
  enMaterialID matID = 
	species2->get_ID();
  switch(matID)
  {
    case matWater:
      phi = 2.26;
      break;
    default : phi = __DUMMY;
  }
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::scan()
//---------------------------------------------------------------------------
{
  fflush(stdin);
  __printf("\nplease enter temperature in Kelvin:");   __scanf("%lf",&T);
  __printf("\n");
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::report()
//---------------------------------------------------------------------------
{
  __printf("Conditions:\n");
  __printf("  Temperature:%e[K]\n\n",T);
  __printf("solute name: %s\n",species1->get_name());
  __printf("solute molal volume at normal boiling point(%-5.5fK): %e [kg/m/sec]\n",species1->Tnb,mu);
  __printf("solvent name: %s\n",species2->get_name());
  __printf("solvent molecular weight: %e\n",species2->M);
  __printf("association factor for solvent : %f\n",phi);
  __printf("solution viscosity: %e [kg/m/sec]\n\n",mu);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::WilkeChang::report2file(char* filename)
//---------------------------------------------------------------------------
{
}


//---------------------------------------------------------------------------
FuelCellShop::Mixture::HanBartelsInterpolation::HanBartelsInterpolation()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::HanBartelsInterpolation::HanBartelsInterpolation(FuelCellShop::Material::PureGas& species1, 
  FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::HanBartelsInterpolation::~HanBartelsInterpolation()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::HanBartelsInterpolation::get_name()
//---------------------------------------------------------------------------
{
  return "Han Bartels Interpolation";
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::HanBartelsInterpolation::get_description()
//---------------------------------------------------------------------------
{
  return "this model is just for O2-H2O liquid system";
}

//---------------------------------------------------------------------------
enDiffusionModelID FuelCellShop::Mixture::HanBartelsInterpolation::get_ID()
//---------------------------------------------------------------------------
{
  return dmHanBartelsInterpolation;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::compute_diffusion (double& D)
//---------------------------------------------------------------------------
{
  D = compute_diffusion(T,p);
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::HanBartelsInterpolation::compute_diffusion (const double& T, const double& p)
//---------------------------------------------------------------------------
{

#ifdef _DETAILED_REPORT_
  printf("\n[WARNING]:pressure value ignored for %s.\n\n",get_name());
#endif
  return __pow(10.0,-4.410 + 773.8/T - 256440.96/T/T)*1E-4;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::compute_diffusion (double& D, std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::compute_diffusion (const double& T,const double& p,
    double& D,std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::compute_dD (std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::HanBartelsInterpolation::compute_dD_dP (const double& T, 
	const double& p)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::HanBartelsInterpolation::compute_dD_dT (const double& T, 
	const double& p)
//---------------------------------------------------------------------------
{
#ifdef _DETAILED_REPORT_
  printf("\n[WARNING]:pressure value ignored for %s.\n\n",get_name());
#endif
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::get_phi(double& phi)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::scan()
//---------------------------------------------------------------------------
{
  fflush(stdin);
  __printf("\nplease enter temperature in Kelvin:");   __scanf("%lf",&T);
  __printf("\n");
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::report()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::HanBartelsInterpolation::report2file(char* filename)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::FerrellHimmelblau::FerrellHimmelblau()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::FerrellHimmelblau::FerrellHimmelblau(FuelCellShop::Material::PureGas& species1, 
  FuelCellShop::Material::PureGas& species2)
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
FuelCellShop::Mixture::FerrellHimmelblau::~FerrellHimmelblau()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::FerrellHimmelblau::get_name()
//---------------------------------------------------------------------------
{
  return "Ferrell-Himmelblau (modified Wilke-Chang model)";
}

//---------------------------------------------------------------------------
char* FuelCellShop::Mixture::FerrellHimmelblau::get_description()
//---------------------------------------------------------------------------
{
  return "this model is just for H2-H2O liquid system";
}

//---------------------------------------------------------------------------
enDiffusionModelID FuelCellShop::Mixture::FerrellHimmelblau::get_ID()
//---------------------------------------------------------------------------
{
  return dmFerrellHimmelblau;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::FerrellHimmelblau::compute_diffusion (double& D)
//---------------------------------------------------------------------------
{
  D = compute_diffusion(T,p);
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::FerrellHimmelblau::compute_diffusion (const double& T, const double& p)
//---------------------------------------------------------------------------
{
  double mu;
  double alpha = 0.7217;
  double A_star= 0.8985; 
  double Vm    = species1->Vl*1e+3;
  species2->get_viscosity(mu,T); mu*=1e+3;

#ifdef _DETAILED_REPORT_
  printf("\n[WARNING]:pressure value ignored for %s.\n\n",get_name());
#endif

  return (4.8e-7*T/__pow(mu,alpha)*__pow((1+A_star*A_star)/Vm,0.6))*1e-4;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::FerrellHimmelblau::compute_diffusion (double& D, std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::FerrellHimmelblau::compute_diffusion (const double& T,const double& p,
    double& D,std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  D    = compute_diffusion(T,p);
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::FerrellHimmelblau::compute_dD (std::vector<double>& dD)
//---------------------------------------------------------------------------
{
  dD[0]= compute_dD_dT(T,p);
  dD[1]= compute_dD_dP(T,p);
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::FerrellHimmelblau::compute_dD_dP (const double& T, 
	const double& p)
//---------------------------------------------------------------------------
{
  return 0;
}

//---------------------------------------------------------------------------
double FuelCellShop::Mixture::FerrellHimmelblau::compute_dD_dT (const double& T, 
	const double& p)
//---------------------------------------------------------------------------
{
  double mu;
  double alpha = 0.7217;
  double A_star= 0.8985; 
  double Vm    = species1->Vl*1e+3;
  species2->get_viscosity(mu,T); mu*=1e+3;

#ifdef _DETAILED_REPORT_
  printf("\n[WARNING]:pressure value ignored for %s.\n\n",get_name());
#endif

  return (4.8e-7/__pow(mu,alpha)*__pow((1+A_star*A_star)/Vm,0.6))*1e-4;
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::FerrellHimmelblau::scan()
//---------------------------------------------------------------------------
{
  fflush(stdin);
  __printf("\nplease enter temperature in Kelvin:");   __scanf("%lf",&T);
  __printf("\n");
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::FerrellHimmelblau::report()
//---------------------------------------------------------------------------
{
}

//---------------------------------------------------------------------------
void FuelCellShop::Mixture::FerrellHimmelblau::report2file(char* filename)
//---------------------------------------------------------------------------
{
}
