//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: PureGas.cc
//    - Description: Class storing pure gas properties.
//    - Developers: Ali MalekpourKoupaei (2009-10) and Madhur Bhaiya (2013)
//    - Id: $Id: PureGas.cc 1354 2013-08-17 00:01:22Z secanell $ 
//
//---------------------------------------------------------------------------

#include "materials/PureGas.h"

namespace NAME = FuelCellShop::Material;

//---------------------------------------------------------------------------
NAME::PureGas::PureGas()
: FuelCellShop::Material::BaseMaterial()
{
  Mu_0 = Tc = Pc = Vc = Zc = w = M = Tnb = __DUMMY;
}
//---------------------------------------------------------------------------
NAME::PureGas::~PureGas()
{
}
//---------------------------------------------------------------------------
double NAME::PureGas::
//---------------------------------------------------------------------------
getCollisionDiameter()
{
  return 11.66*__builtin_pow(Vl,1.0/3.0); //in Angestrom
}
//---------------------------------------------------------------------------
double NAME::PureGas::
//---------------------------------------------------------------------------
getDimensionlessTemperature(const double& T)
{
  return T/1.15/Tnb;
}

//---------------------------------------------------------------------------
void NAME::PureGas::
//---------------------------------------------------------------------------
get_viscosity_Chapman_Enskog(double& mu, const double& T, 
                const double& MW, const double& hard_sphere_diameter) 
{
  double Ts =T/e_k;
  if (hard_sphere_diameter == __DUMMY | e_k == __DUMMY)
  {
    std::cout << "hard_sphere_diameter and e_k are unknown for "<< get_name() << "!!" << std::endl;
    mu=__DUMMY;
    return;
  }

  double A=1.16145; //ref. prausnitz
  double B=0.14874; //ref. prausnitz
  double C=0.52487; //ref. prausnitz
  double D=0.77320; //ref. prausnitz
  double E=2.16178; //ref. prausnitz
  double F=2.43787; //ref. prausnitz
  
  if(!(Ts>=0.3 && Ts<=100.))
      std::cout<< "Temperature is OUT OF RANGE." << std::endl;
  
  double omg_v=A*pow(Ts,(-B))+C*exp(-D*Ts)+E*exp(-F*Ts); //ref. prausnitz
  mu = 26.69e-7 * sqrt(MW*T)/omg_v/pow(hard_sphere_diameter,2); // pa : ref. prausnitz
}

//---------------------------------------------------------------------------
void NAME::PureGas::
//---------------------------------------------------------------------------
get_viscosity_Chapman_Enskog(double& mu, const double& T) 
{ 
  get_viscosity_Chapman_Enskog(mu, T, M, hard_sphere_diameter); 
}


//---------------------------------------------------------------------------
void NAME::PureGas::
//---------------------------------------------------------------------------
get_dMu_d_T_Chapman_Enskog(double& dmu, const double& T, 
                const double& MW, const double& hard_sphere_diameter) 
{
  double Ts =T/e_k;
  if (hard_sphere_diameter == __DUMMY | e_k == __DUMMY)
  {
    std::cout << "hard_sphere_diameter and e_k are unknown for "<< get_name() << "!!" << std::endl;
    dmu=__DUMMY;
    return;
  }

  double A=1.16145; //ref. prausnitz
  double B=0.14874; //ref. prausnitz
  double C=0.52487; //ref. prausnitz
  double D=0.77320; //ref. prausnitz
  double E=2.16178; //ref. prausnitz
  double F=2.43787; //ref. prausnitz
  
  if(!(Ts>=0.3 && Ts<=100.))
      std::cout<< "Temperature is OUT OF RANGE." << std::endl;
  
  double omg_v=A*pow(Ts,-B)+C*exp(-D*Ts)+E*exp(-F*Ts); //ref. prausnitz
  double domg_v=-B*A*pow(Ts,-B-1)-C*D*exp(-D*Ts)-F*E*exp(-F*Ts);
  
  dmu = 26.69e-7/pow(hard_sphere_diameter,2)*pow(MW*e_k,.5)*
        (.5*pow(Ts,-.5)*omg_v - domg_v * pow(Ts, .5)); 
}

//---------------------------------------------------------------------------
void NAME::PureGas::
//---------------------------------------------------------------------------
get_dMu_d_T_Chapman_Enskog(double& dmu, const double& T) 
{ 
  get_dMu_d_T_Chapman_Enskog(dmu, T, M, hard_sphere_diameter); 
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//OXYGEN---------------------------------------------------------------------
//---------------------------------------------------------------------------
NAME::Oxygen::Oxygen()
{
  M   = 31.999   ; // gm/mol
  w   = 0.022    ;
  Tc  = 154.6    ; // Kelvin
  Pc  = 50.43    ; // Bar
  Zc  = 0.288    ; //
  Vc  = 73.4     ; // Cm3/mol
  Tnb = 90.2     ; // Kelvin
  Vl  = 25.6E-3  ; // m3/kmol
  hard_sphere_diameter = 3.467; //ref. Treybal
  e_k = 106.7; //ref. Treybal
}
//---------------------------------------------------------------------------
NAME::Oxygen::~Oxygen()
{
}
//---------------------------------------------------------------------------
char* NAME::Oxygen::
//---------------------------------------------------------------------------
get_name()
{
  return "oxygen";
}
//---------------------------------------------------------------------------
char* NAME::Oxygen::
//---------------------------------------------------------------------------
get_formula()
{
  return "O2";
}

//---------------------------------------------------------------------------
enMaterialID NAME::Oxygen::
//---------------------------------------------------------------------------
get_ID()
{
  return matOxygen;
}

//---------------------------------------------------------------------------
void NAME::Oxygen::
//---------------------------------------------------------------------------
get_viscosity(double& mu,const double& T) 
{
  mu = __DUMMY;
}

//---------------------------------------------------------------------------
double NAME::Oxygen::get_molar_enthalpy(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
    
    double Cp_molar = (0.88 - 0.0001*(T/1000.0) + 0.54*std::pow(T/1000.0,2) - 0.33*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double h_molar = 0.0 + Cp_molar*(T - 298.0); // J/mol
    return h_molar;
}

//---------------------------------------------------------------------------
double NAME::Oxygen::get_dmolar_enthalpy_dT(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.

    double Cp_molar = (0.88 - 0.0001*(T/1000.0) + 0.54*std::pow(T/1000.0,2) - 0.33*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double dCp_molar_dT = ( -0.0001 + 0.54*2.0*(T/1000.0) - 0.33*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double dh_molar_dT = dCp_molar_dT*(T - 298.0) + Cp_molar;
    return dh_molar_dT;
}

//---------------------------------------------------------------------------
double NAME::Oxygen::get_d2molar_enthalpy_dT2(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.

    double dCp_molar_dT = ( -0.0001 + 0.54*2.0*(T/1000.0) - 0.33*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double d2Cp_molar_dT2 = ( 0.54*2.0 - 0.33*3.0*2.0*(T/1000.0) )*M*(1.0/1000.0)*(1.0/1000.0);
    double d2h_molar_dT2 = d2Cp_molar_dT2*(T - 298.0) + 2.0*dCp_molar_dT;
    return d2h_molar_dT2;
}

//---------------------------------------------------------------------------
//NITROGEN-------------------------------------------------------------------
//---------------------------------------------------------------------------
NAME::Nitrogen::Nitrogen()
{
  M   = 28.014   ; // gm/mol
  w   = 0.038    ;
  Tc  = 126.2    ; // Kelvin
  Pc  = 34.00    ; // Bar
  Zc  = 0.289    ; //
  Vc  = 89.2     ; // Cm3/mol
  Tnb = 77.3     ; // Kelvin
  Vl  = 31.2e-3  ; // m3/kmol
  hard_sphere_diameter = 3.798;
  e_k = 71.4;
}
//---------------------------------------------------------------------------
NAME::Nitrogen::~Nitrogen()
{
}
//---------------------------------------------------------------------------
char* 
NAME::Nitrogen::get_name()
{
  return "nitrogen";
}
//---------------------------------------------------------------------------
char* 
NAME::Nitrogen::get_formula()
{
  return "N2";
}
//---------------------------------------------------------------------------
enMaterialID NAME::Nitrogen::
//---------------------------------------------------------------------------
get_ID()
{
  return matNitrogen;
}

//---------------------------------------------------------------------------
void NAME::Nitrogen::
//---------------------------------------------------------------------------
get_viscosity(double& mu,const double& T) 
{
  mu = __DUMMY;
}

//---------------------------------------------------------------------------
double NAME::Nitrogen::get_molar_enthalpy(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.

    double Cp_molar = (1.11 - 0.48*(T/1000.0) + 0.96*std::pow(T/1000.0,2) - 0.42*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double h_molar = 0.0 + Cp_molar*(T - 298.0); // J/mol
    return h_molar;
}

//---------------------------------------------------------------------------
double NAME::Nitrogen::get_dmolar_enthalpy_dT(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
    
    double Cp_molar = (1.11 - 0.48*(T/1000.0) + 0.96*std::pow(T/1000.0,2) - 0.42*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double dCp_molar_dT = ( -0.48 + 0.96*2.0*(T/1000.0) - 0.42*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double dh_molar_dT = dCp_molar_dT*(T - 298.0) + Cp_molar;
    return dh_molar_dT;
}

//---------------------------------------------------------------------------
double NAME::Nitrogen::get_d2molar_enthalpy_dT2(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
    
    double dCp_molar_dT = ( -0.48 + 0.96*2.0*(T/1000.0) - 0.42*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double d2Cp_molar_dT2 = ( 0.96*2.0 - 0.42*3.0*2.0*(T/1000.0) )*M*(1.0/1000.0)*(1.0/1000.0);
    double d2h_molar_dT2 = d2Cp_molar_dT2*(T - 298.0) + 2.0*dCp_molar_dT;
    return d2h_molar_dT2;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//WATER-------------------------------------------------------------------
//---------------------------------------------------------------------------
NAME::Water::Water()
{
  M   = 18.015   ; // gm/mol
  w   = 0.345    ;
  Tc  = 647.1    ; // Kelvin
  Pc  = 220.55   ; // Bar
  Zc  = 0.229    ; //
  Vc  = 55.9     ; // Cm3/mol
  Tnb = 373.2    ; // Kelvin
  Vl  = 18.9e-3  ; //m3/kmol
  hard_sphere_diameter = 2.641; // Angstrome : prausnitz appendix B
  e_k = 809.1;/*->epsilon/k*/
}
//---------------------------------------------------------------------------
NAME::Water::~Water()
{
}
//---------------------------------------------------------------------------
char* 
NAME::Water::get_name()
{
  return "water";
}
//---------------------------------------------------------------------------
char* 
NAME::Water::get_formula()
{
  return "H2O";
}
//---------------------------------------------------------------------------
enMaterialID 
NAME::Water::get_ID()
{
  return matWater;
}

//---------------------------------------------------------------------------
void 
NAME::Water::get_viscosity(double& mu,const double& T) 
//---------------------------------------------------------------------------
{
  double a =  8.056e+006;
  double b =    -0.05901;
  double c =       37.23;
  double d =     -0.0133;

  mu = (a*__exp(b*T) + c*__exp(d*T))*1e-3; //kg/m/s
}

//---------------------------------------------------------------------------
double NAME::Water::get_molar_enthalpy(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
 
    double Cp_molar = (1.79 + 0.107*(T/1000.0) + 0.586*std::pow(T/1000.0,2) - 0.20*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double h_molar = (-241826.0) + Cp_molar*(T - 298.0); // J/mol
    return h_molar;
}

//---------------------------------------------------------------------------
double NAME::Water::get_dmolar_enthalpy_dT(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
   
    double Cp_molar = (1.79 + 0.107*(T/1000.0) + 0.586*std::pow(T/1000.0,2) - 0.20*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double dCp_molar_dT = ( 0.107 + 0.586*2.0*(T/1000.0) - 0.2*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double dh_molar_dT = dCp_molar_dT*(T - 298.0) + Cp_molar;
    return dh_molar_dT;
}

//---------------------------------------------------------------------------
double NAME::Water::get_d2molar_enthalpy_dT2(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
   
    double dCp_molar_dT = ( 0.107 + 0.586*2.0*(T/1000.0) - 0.2*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double d2Cp_molar_dT2 = ( 0.586*2.0 - 0.2*3.0*2.0*(T/1000.0) )*M*(1.0/1000.0)*(1.0/1000.0);
    double d2h_molar_dT2 = d2Cp_molar_dT2*(T - 298.0) + 2.0*dCp_molar_dT;
    return d2h_molar_dT2;
}
//---------------------------------------------------------------------------
double
NAME::Water::satPressure(const double& Temp)
{
	double logSatP = -2.1794 + (0.02953*(Temp-273.15)) - (9.1837e-5*std::pow(Temp-273.15, 2.0)) + (1.4454e-7*std::pow(Temp-273.15, 3.0));
	double satP = (std::pow(10.0, logSatP)) * 101325.0; //pressure converted from atm to Pa
	return satP;
}

//---------------------------------------------------------------------------
double
NAME::Water::derivSatPressure(const double& Temp)
{
	double dpdT = (satPressure(Temp))*(std::log(10.0))*(0.02953 - (9.1837e-5*2.0*(Temp-273.15)) + (1.4454e-7*3.0*std::pow(Temp-273.15, 2.0)));
	return dpdT;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//HYDROGEN-------------------------------------------------------------------
//---------------------------------------------------------------------------
NAME::Hydrogen::Hydrogen()
{
  M   = 2.0160   ; // gm/mol
  w   = -0.216   ;
  Tc  = 33.19    ; // Kelvin
  Pc  = 13.13    ; // Bar
  Zc  = 0.305    ; //
  Vc  = 64.1     ; // Cm3/mol
  Tnb = 20.40    ; // Kelvin
  Vl  = 14.3E-3  ; //m3/kmol
  hard_sphere_diameter=2.827; // Angstrome : prausnitz appendix B
  e_k=59.7; //Treybal
}
//---------------------------------------------------------------------------
NAME::Hydrogen::~Hydrogen()
{
}
//---------------------------------------------------------------------------
char* 
NAME::Hydrogen::get_name()
{
  return "Hydrogen";
}
//---------------------------------------------------------------------------
char* 
NAME::Hydrogen::get_formula()
{
  return "H2";
}
//---------------------------------------------------------------------------
enMaterialID 
NAME::Hydrogen::get_ID()
{
  return matHydrogen;
}

//---------------------------------------------------------------------------
void 
NAME::Hydrogen::get_viscosity(double& mu,const double& T) 
//---------------------------------------------------------------------------
{
  get_viscosity_Chapman_Enskog(mu, T) ;
}

//---------------------------------------------------------------------------
double NAME::Hydrogen::get_molar_enthalpy(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.

    double Cp_molar = (13.46 + 4.6*(T/1000.0) - 6.85*std::pow(T/1000.0,2) + 3.79*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double h_molar = 0.0 + Cp_molar*(T - 298.0); // J/mol
    return h_molar;
}

//---------------------------------------------------------------------------
double NAME::Hydrogen::get_dmolar_enthalpy_dT(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
    
    double Cp_molar = (13.46 + 4.6*(T/1000.0) - 6.85*std::pow(T/1000.0,2) + 3.79*std::pow(T/1000.0,3))*M; // J/(mol-K)
    double dCp_molar_dT = ( 4.6 - 6.85*2.0*(T/1000.0) + 3.79*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double dh_molar_dT = dCp_molar_dT*(T - 298.0) + Cp_molar;
    return dh_molar_dT;
}

//---------------------------------------------------------------------------
double NAME::Hydrogen::get_d2molar_enthalpy_dT2(const double& T) const
{
    // Ref: Richard E. Sonntag, Claus Borgnakke, and Gordon J. Van Wylen. Fundamentals of Thermodynamics,
    // Revised 6th Edition. John Wiley & Sons, Inc., 2003.
    
    double dCp_molar_dT = ( 4.6 - 6.85*2.0*(T/1000.0) + 3.79*3.0*std::pow(T/1000.0,2) )*M*(1.0/1000.0);
    double d2Cp_molar_dT2 = ( -6.85*2.0 + 3.79*3.0*2.0*(T/1000.0) )*M*(1.0/1000.0)*(1.0/1000.0);
    double d2h_molar_dT2 = d2Cp_molar_dT2*(T - 298.0) + 2.0*dCp_molar_dT;
    return d2h_molar_dT2;
}