//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2009-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: adaptive_refinement.cc
//    - Description: Child of ApplicationCopy used to implement adaptive refinement
//    - Developers: M. Secanell
//    - $Id: adaptive_refinement.cc 1389 2013-08-27 23:29:33Z zingan $
//
//---------------------------------------------------------------------------

#include "adaptive_refinement.h"

//---------------------------------------------------------------------------
template <int dim>
AppFrame::AdaptiveRefinement<dim>::AdaptiveRefinement
     (AppFrame::OptimizationBlockMatrixApplication<dim>& app_lin,
      AppFrame::ApplicationCopy& app,
      const AppFrame::FEVector& solution)
  :
  app_linear(&app_lin),
  app(&app),
  solution(solution)

{
  deallog <<"->AdaptiveRefinement" << std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
AppFrame::AdaptiveRefinement<dim>::~AdaptiveRefinement()
{}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::declare_parameters(ParameterHandler& param) const
{
    // Initialize to parameters needed for evaluate (note: this is provisional)
    param.enter_subsection("Adaptive refinement");
    {
        param.declare_entry ("Number of Refinements",
                             "1",
                             Patterns::Integer());
        param.declare_entry("Output initial mesh",
                            "false",
                            Patterns::Bool(),
                            "Set flag to true if you want to output a EPS figure of the initial mesh "
                            "using the value in file initial mesh");
        param.declare_entry("Output initial mesh filename",
                            "initial_mesh",
                            Patterns::Anything(),
                            "File where the initial mesh will be output");
        param.declare_entry("Output initial solution",
                            "false",
                            Patterns::Bool(),
                            "Set flag to true if you want to output the initial solution to file");
        param.declare_entry("Initial solution filename",
                            "initial_sol",
                            Patterns::Anything(),
                            "File where the initial solution will be output");
        param.declare_entry("Output intermediate solutions",
                            "true",
                            Patterns::Bool());
        param.declare_entry("Output intermediate responses",
                            "true",
                            Patterns::Bool());
        param.declare_entry("Output final solution",
                            "true",
                            Patterns::Bool());
        param.declare_entry("Output solution for transfer",
                            "false",
                            Patterns::Bool(),
                            "Check whether a solution on a refined grid should be output to file on a coarse mesh");
        param.declare_entry("Read in initial solution from file",
                            "false",
                            Patterns::Bool(),
                            "Check whether a stored solution should be read from file and applied to the grid");

    }
    param.leave_subsection();

    param.enter_subsection("Application");
    {
        param.declare_entry ("Analytical gradients",
                             "false",
                             Patterns::Bool());
    }
    param.leave_subsection();

    //--- Create entries on ParameterHandler object:
    app->declare_parameters(param);

}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::set_parameters(const std::vector<std::string>& name_dvar,
                                                  const std::vector<double>& value_dvar,
                                                  ParameterHandler& param)
{
    app_linear->set_parameters(name_dvar,
                               value_dvar,
                               param);
}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::initialize(ParameterHandler& param)
{
  // Initialize application using ParameterHandler object:
    param.enter_subsection("Adaptive refinement");
    {
        n_ref = param.get_integer("Number of Refinements");

        output_initial_mesh = param.get_bool("Output initial mesh");
        filename_initial_mesh = param.get("Output initial mesh filename");

        output_initial_sol = param.get_bool("Output initial solution");
        filename_initial_sol = param.get("Initial solution filename");

        output_intermediate_sol = param.get_bool("Output intermediate solutions");
        output_final_sol = param.get_bool("Output final solution");

        output_intermediate_resp = param.get_bool("Output intermediate responses");

        output_coarse_solution = param.get_bool("Output solution for transfer");
        read_in_initial_solution = param.get_bool("Read in initial solution from file");

        app_linear->set_read_initial_solution(read_in_initial_solution);
    }
    param.leave_subsection();
    param.enter_subsection("Application");
    {
        gradients = param.get_bool ("Analytical gradients");
    }
    param.leave_subsection();

    app->initialize(param);
}


//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::solve(const std::string param_file, ParameterHandler& param)
{
    AppFrame::declare_parameter_files(param);
    this->declare_parameters(param);

    //--- Read data from file to ParamterHandler object:
    AppFrame::read_parameter_files(param,
                                   param_file);

    //--- Use the data read from file to initialize problem
    this->initialize(param);

    //--- Solve problem
    this->run_Newton();
}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::run_Newton()
{
  std::vector<double> resp(0);
  resp.clear();
  resp.resize(app_linear->get_n_resp());

  run_Newton (resp);
}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::run_Newton(std::vector<double>& resp)
{
  std::vector<std::vector<double> > dresp_dl(0);
  run_Newton (resp, dresp_dl);
}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::run_Newton(std::vector<double>& resp,
                                              std::vector<std::vector<double> >& dresp_dl)
{
  // Output grid:
  if (output_initial_mesh == true)
    app_linear->grid_out(filename_initial_mesh);

  // Vector to store solution and rhs:
  //AppFrame::FEVector solution;
  AppFrame::FEVectors vectors;
  vectors.add_vector(solution,"Solution");

  // Tell app_linear that the solution will be transferred to refined meshes
  app_linear->add_vector_for_transfer(&solution);

  // Create a copy of the triangulation object to store the information of the coarse mesh.
  Triangulation<dim> tr_coarse;
  AppFrame::FEVector coarse_solution;
  if (output_coarse_solution)
      app_linear->store_triangulation(tr_coarse);

  // -- Start adaptive refinement loop:
  for (unsigned int cycle = 0; cycle<n_ref; ++cycle)
  {
      bool last_cycle = false;
      if (cycle == n_ref-1) last_cycle = true;
      deallog.push ("Solver");
      deallog<< "Cycle "<<cycle<<" : "<<std::endl;
      std::ostringstream streamOut;
      streamOut << cycle;
      std::string filename = "fuel_cell-grid-cycle"+streamOut.str();
      std::string sol_filename = "fuel_cell-sol-cycle"+streamOut.str();

      app_linear->get_data()->enter("Refinement", cycle);

      if (cycle == 0)
      {
          // initialize vectors with new grid
          if( solution.size() == 0 || solution.all_zero() )
                 app_linear->init_solution(solution);

          if (output_initial_sol == true)
              app->data_out(filename_initial_sol, vectors);

          if(output_coarse_solution) // if the solution needs to be transferred, set the size of the vector
              coarse_solution.reinit(solution);
      }
      else
      {
          app->estimate(vectors);
          // Remesh and start with an initial "coarse" solution. Note that
          // in order to transfer the soltion, add_vector_for_transfer needs
          // to be initialized (see before loop)
          app->remesh();
      }

      // -- Solve the nonlinear system of equations:
      try
      {
          deallog << "Solving..." << std::endl;
          app->solve(solution, vectors);

          deallog << "Solution:" << solution.n_blocks() << ':';
          for (unsigned int b=0;b<solution.n_blocks();++b)
              deallog << ' ' << solution.block(b).size();
          deallog << std::endl;

          if(output_intermediate_resp || last_cycle)
          {
              //NOT WOKING for now: deallog<< "Current density : "<<app->evaluate(vectors)<<"A/cm^2"<<std::endl;
              //--- Resize the response vector and copy to it all responses
              app->evaluate(vectors);
              //Provisional while evaluate() is not changed
              app_linear->responses (resp,vectors);
              //deallog<< "Current density : "<<-resp[0]<<"A/cm^2"<<std::endl;
          }

          if(output_intermediate_sol || (last_cycle && output_final_sol) )
          {
              app_linear->grid_out(filename);
              app->data_out(sol_filename,vectors);
          }

          if(last_cycle && output_coarse_solution)
          {
              AppFrame::FEVectors transfer_sol;
              app_linear->transfer_solution_to_coarse_mesh(tr_coarse,coarse_solution,solution);
          }
      }
      catch (const std::exception& e)
      {
          deallog << e.what() << std::endl;
      }
      deallog.pop();
  }//for loop

  if (gradients == true)
  {
      try
      {
          dresp_dl.clear();
          dresp_dl.resize(app_linear->get_n_resp(), std::vector<double> (app_linear->get_n_dvar()));
          app_linear->solve_direct(dresp_dl,
                                   vectors);
      }
      catch (const std::exception& e)
      {
          deallog << e.what() << std::endl;
      }
  }

  for (unsigned int i=0; i<resp.size(); ++i)
      deallog<<"Response "<<i<<" is : "<<resp[i]<<std::endl;
}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::test_derivatives(const std::string input_file,
                                                    const std::string dvar,
                                                    const double value,
                                                    std::vector<double>& resp,
                                                    std::vector<std::vector<double> >& dresp,
                                                    const bool grad)
{
    //--- Declare parameters:
    ParameterHandler param;
    this->declare_parameters(param);

    //--- Read data from file to ParamterHandler object:
    AppFrame::read_parameter_files(param,
                                   input_file);
    //--- Modify data just read:
    std::vector<std::string> name_dvar(1);
    std::vector<double> value_dvar(1);
    name_dvar[0] = dvar;
    value_dvar[0] = value;
    set_parameters(name_dvar,
                   value_dvar,
                   param);

    //--- Initialize application:
    this->initialize(param);

    gradients = grad;

    // -- Solve the system of governing equations for the given set of parameters:
    run_Newton(resp, dresp);
}

//---------------------------------------------------------------------------
template <int dim>
void
AppFrame::AdaptiveRefinement<dim>::print_parameters() const
{
    // Initialize Parameter handler
    ParameterHandler param;
    this->declare_parameters(param);

    std::ofstream marc("parameters_sample.prm");
    param.print_parameters(marc,
                           ParameterHandler::Text);
}


//---------------------------------------------------------------------------
// Explicit instantations
template class AppFrame::AdaptiveRefinement<deal_II_dimension>;