// ----------------------------------------------------------------------------
//
// FCST: Fuel Cell Simulation Toolbox
//
// Copyright (C) 2006-2013 by Energy Systems Design Laboratory, University of Alberta
//
// This software is distributed under the MIT License
// For more information, see the README file in /doc/LICENSE
//
// - Class: newton_basic.cc
// - Description: This class performs basic Newton iterations with a constant weight
// - Developers: Valentin N. Zingan, University of Alberta
// - Id: $Id: newton_basic.cc 1354 2013-08-17 00:01:22Z secanell $
//
// ----------------------------------------------------------------------------

#include <newton_basic.h>

const Event NewtonBasic::bad_derivative = Event::assign("Newton");

       //////////////////////////////////////////////////
       //////////////////////////////////////////////////
       // CONSTRUCTORS, DESTRUCTOR, AND INITIALIZATION //
       //////////////////////////////////////////////////
       //////////////////////////////////////////////////

// ---             ---
// --- Constructor ---
// ---             ---

NewtonBasic::NewtonBasic(ApplicationBase& app)
:
newtonBase(app)
{
  deallog << "->NewtonBasic";
}

// ---            ---
// --- Destructor ---
// ---            ---

NewtonBasic::~NewtonBasic()
{ }

// ---                    ---
// --- declare_parameters ---
// ---                    ---

void
NewtonBasic::declare_parameters(ParameterHandler& param)
{
  newtonBase::declare_parameters(param);

  param.enter_subsection("Newton");
  {
         param.declare_entry("Basic Newton constant weight",
                             "1.0",
                              Patterns::Double(),
                             " ");
  }
  param.leave_subsection();
}

// ---            ---
// --- initialize ---
// ---            ---

void
NewtonBasic::initialize(ParameterHandler& param)
{
  newtonBase::initialize(param);

  param.enter_subsection("Newton");
  {
         weight = param.get_double("Basic Newton constant weight");
  }
  param.leave_subsection();
}

       ////////////////////
       ////////////////////
       // SOLVE FUNCTION //
       ////////////////////
       ////////////////////

// ---       ---
// --- solve ---
// ---       ---

void
NewtonBasic::solve(FEVector&        u,
                   const FEVectors& in_vectors)
{
  // initialize "step"
  step = 0;

  // output the L2 norm of the initial guess if needed
  deallog.push("Newton");
  if( debug > 2 )
    deallog << "L2 norm of the initial guess: " << u.l2_norm() << std::endl;

  // "Du" is the solution of a linear system
  FEVector* Du = this->get_data()->block_vector_pool.alloc();
  Du->reinit(u);

  // "res" is the residual of a linear system
  FEVector* res = this->get_data()->block_vector_pool.alloc();
  res->reinit(u);

  // "src1" contains "Solution" and "Newton iterate"
  FEVectors src1;
  src1.add_vector(u, "Newton iterate");
  src1.merge(in_vectors);

  // "src2" contains "Solution", "Newton iterate", and "Newton residual"
  FEVectors src2;
  src2.add_vector(*res, "Newton residual");
  src2.merge(src1);

  // "ApplicationBase::data" contains "Newton"
  this->get_data()->enter("Newton", u);

  // fill "res" using the initial guess info
  double residual     = app->residual(*res, src1);
  double old_residual = residual;

  deallog << "iter  = " << step     << std::endl;
  deallog << "error = " << residual << std::endl; // L2 norm of the residual

  // output the initial guess info
  if(debug_solution || debug_update || debug_residual)
  {
    std::ostringstream streamOut;
    streamOut << step;
    std::string name = "Newton_" + streamOut.str();
    FEVectors out;
    out.add_vector(u, "Solution");
    out.add_vector(*Du, "Update");
    out.add_vector(*res, "Residual");
    app->data_out(name, out);
  }

  // the basic Newton's loop with a constant weight
  while(control.check(step++, residual) == SolverControl::iterate)
  {
    // checking
    if(residual/old_residual >= assemble_threshold)
      app->notify(bad_derivative);

    // reset "Du"
    Du->reinit(u);

    // solve a linear system
    // we pass u^n, Du = 0, res^n
    try
    {
      app->solve(*Du, src2);
    }
    catch(SolverControl::NoConvergence& e)
    {
      deallog << "Inner iteration failed after "
              << e.last_step
              << " steps with residual "
              << e.last_residual
              << std::endl;
    }

    // update "u"
    u.add(weight, *Du);

    // update "old_residual"
    old_residual = residual;

    // reset "res"
    res->reinit(u);

    // fill "res" using "u"
    residual = app->residual(*res, src1);

    deallog << "iter  = " << step     << std::endl;
    deallog << "error = " << residual << std::endl; // L2 norm of the residual

    // output "u" info
    if(debug_solution || debug_update || debug_residual)
    {
      std::ostringstream streamOut;
      streamOut << step;
      std::string name = "Newton_" + streamOut.str();
      FEVectors out;
      out.add_vector(u, "Solution");
      out.add_vector(*Du, "Update");
      out.add_vector(*res, "Residual");
      app->data_out(name, out);
    }
  } // end while

  // deallocation
  get_data()->erase_vector("Newton");
  get_data()->block_vector_pool.free(Du);
  get_data()->block_vector_pool.free(res);

  deallog.pop();

  if( control.last_check() != SolverControl::success )
    throw SolverControl::NoConvergence(control.last_step(),
                                       control.last_value());
  numIter = step - 1;
  get_data()->enter("niters", numIter);
}