//---------------------------------------------------------------------------
//    $Id: newton_w_3pp.cc 1354 2013-08-17 00:01:22Z secanell $
//    Version: $Name:  $
//
//    Copyright (C) 2007 by Jason Boisvert
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------


#include <newton_w_3pp.h>
#include <base/parameter_handler.h>
#include <base/data_out_base.h>
#include <lac/block_vector.h>

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <algorithm>

using namespace AppFrame;

const Event Newton3pp::bad_derivative = Event::assign("Newton");

Newton3pp::Newton3pp(ApplicationBase& app)
: newtonBase(app)

{
	deallog << "->Newton3pp";
}

void
Newton3pp::solve(AppFrame::FEVector& u,
		 const AppFrame::FEVectors& in_vectors)
{
  // Initialize:
  double lambda_trial = 0.0;
  double lambda_previous  = 0.0;
  double lambda_current = 0.0;

//   double residual_trial_norm = 0.0;
//   double residual_norm = 0.0;
//   double residual_hold_norm = 0.0;

  double mf_original = 0.0;
  double mf_current = 0.0;
  double mf_previous = 0.0;

  double lambda_old = 1.0;

  int line_search_iterations = 0;

  step = 0;

  deallog.push ("Newton");
  if (debug>2)
    deallog << "u: " << u.l2_norm() << std::endl;

  FEVector* Du = get_data()->block_vector_pool.alloc();
  FEVector* res = get_data()->block_vector_pool.alloc();


  res->reinit(u);
  FEVectors src1;
  FEVectors src2;
  src1.add_vector(u, "Newton iterate");
  src1.merge(in_vectors);
  src2.add_vector(*res, "Newton residual");
  src2.merge(src1);

  get_data()->enter("Newton", u);

  // fill res with (f(u), v)
  double residual = app->residual(*res, src1);
  double old_residual = residual;

  // Output the solution at the Newton iteration if residual debug is on
  //std::cout << "debug resid: " <<  debug_residual <<"\n";
  if (debug_residual)
    {
      std::ostringstream streamOut;
      streamOut << step;
      std::string name = "Newton_" + streamOut.str();
      FEVectors out;
      out.add_vector(u, "Solution");
      Du->reinit(u);
      out.add_vector(*Du, "update");
      out.add_vector(*res, "residual");
      app->data_out(name, out);
    }

  line_search_iterations = 0;
  lambda_trial = 1.0;
  lambda_previous = 1.0;
  lambda_current = lambda_trial;
  double lambda_2 = 0;

  //Begin Newton Iterations
  while (control.check(step++, residual) == SolverControl::iterate)
    {
	lambda_current = 1.0;
	//std::cout << "Initial lambda: " << lambda_current << "\n";
	//std::cout << "residual: " << residual <<"\n";
	// assemble (Df(u), v)
	if(residual/old_residual >= assemble_threshold)
	  app->notify (bad_derivative);
	Du->reinit(u);
	//Solver Linear System
	try
	  {
	    app->solve (*Du, src2);
	  }
	catch (SolverControl::NoConvergence& e)
	  {
	    deallog << "Inner iteration failed after "
		    << e.last_step << " steps with residual "
		    << e.last_residual << std::endl;
	  }

	////////////////////////////////////////////////////////////
	//-- Three Point Parabolic Line Search Method
	////////////////////////////////////////////////////////////
	/*
	//initalize paremters
	*/
	//Depends in which refinement
	const double* ref = get_data()->scalar("Refinement");
	//std::cout << "Ref: "<< *ref <<"\n";
	if (*ref == 0)
	{
	  this->max_reduction_factor = 0.90;
	  this->min_reduction_factor = 0.01;
	  this->alpha=.01;
	}
	else
	{
	  this->max_reduction_factor = 0.90;
	  this->min_reduction_factor = 0.50;
	  this->alpha=.001;
	}
	// get coefficients
	mf_original = sum_of_squares(old_residual);
	//start search

	//Try with full newton step;
 	u.add(-lambda_current,*Du);
	old_residual =residual;
	residual = app->residual(*res, src1);

	//Get coefficients with full Newton step
	mf_current = sum_of_squares(residual);
	mf_previous = mf_current;
	int max_line_search_reductions =  25;

	while (std::isnan(residual) || !armijo(lambda_current,old_residual,residual)  )
	{
	  residual = app->residual(*res, src1);
	  //std::cout << "Starting line search at: " <<  line_search_iterations << " for a max of: "<<max_line_search_reductions << "\n";
	  //std::cout<<"I am in search loop.\n";

	  u.add(lambda_current,*Du);
	  //compute new step
	  lambda_trial = three_point_step (line_search_iterations,lambda_current,lambda_previous,mf_original,mf_current,mf_previous);
	  //std::cout<< "Computed the step: " << lambda_trial << "\n";

	  if (std::isnan(lambda_trial))
	    lambda_trial = lambda_current * (max_reduction_factor+min_reduction_factor)*0.5;
	  //Now the safe Gurads.
	  lambda_trial = std::max(lambda_trial,min_reduction_factor * lambda_current);
	  lambda_trial = std::min(lambda_trial,max_reduction_factor * lambda_current);
	  std::cout << "Step after safe guads: "<< lambda_trial << std::endl;

	  //update lambda
	  lambda_previous = lambda_current;
	  lambda_current = lambda_trial;

	  //Try Stepsize
	  u.add(-lambda_current,*Du);

	  //calcuate residual_trial
	  residual = (*compute_residual)(*res, src1);

	  //new coefficents
	  mf_previous = mf_current;
	  mf_current = sum_of_squares(residual);

	  //incrment line searc iterations and check for mac converg reach
	  line_search_iterations++;
	  //std::cout<<"IN LOOP: "<< line_search_iterations <<  " \n";
	  //std::cout<<"Residual: "<<residual <<"\n";
	  if(line_search_iterations > max_line_search_reductions)
	  {
	    std::cout << "Reached maximum line search iterations: \nExiting and restarting.\n";
	    break;
	  }
	  //std::cout << "residual:" << residual << "\n";
	  //std::cout << "old residual:" << old_residual << "\n";

	  //Update lambda_old
	}
	if(step > 2)
	  lambda_2= lambda_old;
	lambda_old = lambda_current;
	line_search_iterations = 0;

	if (debug_solution || debug_update || debug_residual)
	  {
	    std::ostringstream streamOut;
	    streamOut << step;
	    std::string name = "Newton_" + streamOut.str();
	    FEVectors out;
	    out.add_vector(u, "Solution");
	    out.add_vector(*Du, "update");
	    out.add_vector(*res, "residual");
	    app->data_out(name, out);
	  }
    }


  get_data()->erase_vector("Newton");
  get_data()->block_vector_pool.free(Du);
  get_data()->block_vector_pool.free(res);
  //get_data()->block_vector_pool.free(u_trial);
  deallog.pop();

				   // in case of failure: throw
				   // exception
  if (control.last_check() != SolverControl::success)
    throw SolverControl::NoConvergence (control.last_step(),
					control.last_value());
				   // otherwise exit as normal

 //step=6;
 	numIter= step - 1;
	get_data()->enter("niters",numIter);
}

double
Newton3pp::sum_of_squares (double residual_norm)
{
	return (0.5 * residual_norm * residual_norm);
}

double
Newton3pp::three_point_step (int line_search_iterations,double lambda_current, double lambda_previous, double mf_original, double mf_current, double mf_previous)
{

	if (line_search_iterations == 0 ) return max_reduction_factor * lambda_current;

        // Compute coefficients of interpolation polynomial
	double c_2 = 0.0;
	c_2 = lambda_previous * (mf_current - mf_original) - lambda_current * (mf_previous - mf_original);

	if (c_2 >= 0) return (max_reduction_factor * lambda_current);

	double c_1 = 0.0;
	c_1 = lambda_current * lambda_current * (mf_previous - mf_original) - lambda_previous * lambda_previous * (mf_current - mf_original);
	return -c_1 * 0.5 / c_2;

}

bool
Newton3pp::armijo(double step_length,double residual_norm,double new_residual_norm)
{

	//set acceptance factor
	double fact = (1.0 - alpha*step_length);
	//std::cout <<"alpha: " << alpha <<"acc: "<< fact << "step: " << step_length << "\n";
	return ( new_residual_norm <= ( fact * residual_norm ));

}
