//---------------------------------------------------------------------------
//    $Id: newton_w_line_search.cc 1354 2013-08-17 00:01:22Z secanell $
//    Version: $Name:  $
//
//    Copyright (C) 2007 by Marc Secanell and Guido Kanschat
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include <newton_w_line_search.h> //If I put it in appframe #include<appframe/newton_w_line_search.h>
#include <base/parameter_handler.h>
#include <base/data_out_base.h>
#include <lac/block_vector.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <iomanip>


using namespace AppFrame;

const Event NewtonLineSearch::bad_derivative = Event::assign("Newton");

NewtonLineSearch::NewtonLineSearch(ApplicationBase& app)
    : newtonBase(app)
{
  deallog << "->NewtonLineSearch";
}


void
NewtonLineSearch::declare_parameters(ParameterHandler& param)
{
	param.enter_subsection("Newton");
	{
		param.declare_entry("Line search","false", Patterns::Bool());
		param.declare_entry("Initial Overrelaxation", "1.", Patterns::Double());
		param.declare_entry("Number of iterations with overrelaxation","1",Patterns::Integer());
	}
	param.leave_subsection();
	
	newtonBase::declare_parameters(param);
}

void
NewtonLineSearch::_initialize (ParameterHandler& param)
{
    param.enter_subsection("Newton");
	line_search = param.get_bool("Line search");
	overrelax = param.get_double("Initial Overrelaxation");
	overrelax_steps = param.get_integer("Number of iterations with overrelaxation");
    param.leave_subsection ();
}

void
NewtonLineSearch::initialize (ParameterHandler& param)
{
    newtonBase::initialize(param);
    _initialize(param);
}

void
NewtonLineSearch::solve (AppFrame::FEVector& u, const AppFrame::FEVectors& in_vectors)
{
  deallog.push ("Newton");
  if (debug>2)
    deallog << "u: " << u.l2_norm() << std::endl;

  FEVector* Du_debug = get_data()->block_vector_pool.alloc();
  FEVector* res = get_data()->block_vector_pool.alloc();

  res->reinit(u);
  Du_debug->reinit(u);
  FEVectors src1;
  FEVectors src2;
  src1.add_vector(u, "Newton iterate");
  src1.merge(in_vectors);
  src2.add_vector(*res, "Newton residual");
  src2.merge(src1);

  unsigned int step = 0;
				   // fill res with (f(u), v)
  double residual = app->residual(*res, src1);
  double old_residual = residual;

  // Output the solution at the Newton iteration if residual debug is on
  if (debug_residual)
    {
      std::ostringstream streamOut;
      streamOut << step;
      std::string name = "Newton_" + streamOut.str();
      FEVectors out;
      out.add_vector(u, "Solution");
      out.add_vector(*Du_debug, "update");
      out.add_vector(*res, "residual");
      app->data_out(name, out);
    }
  get_data()->block_vector_pool.free(Du_debug);

  //
  while (control.check(step++, residual) == SolverControl::iterate)
    {
				       // assemble (Df(u), v)
	if (residual/old_residual >= assemble_threshold)
	  app->notify (bad_derivative);//app->assemble_reason (reason_nonlinear); //Changed after Guido's changes: app->assemble (src1);
	
	FEVector* Du = get_data()->block_vector_pool.alloc();
	Du->reinit(u);
	try 
	  {
	    app->solve (*Du, src2);
	  }
	catch (SolverControl::NoConvergence& e)
	  {
	    deallog << "Inner iteration failed after "
		    << e.last_step << " steps with residual "
		    << e.last_residual << std::endl;
	  }

	////////////////////////////////////////////////////////////
	//-- Step size control
	////////////////////////////////////////////////////////////
	if (line_search)
	  {
	    deallog << "Line Search" << std::endl;
	    // -- Use a line search to increase robustness
	    double min_residual = residual;
	    double opt_alpha = -0.25;
	    unsigned int num_points = 1;
	    const double* ref = get_data()->scalar("Refinement");	
	    if (step < 10 &&  *ref < 1)
	      num_points = 20;
	    else if ((step >=10 && step < 20) &&  *ref < 1)
	      num_points = 10;
	    else
	      num_points = 5;
	    	    
	    // ------------ GEOMETRIC LINE SEARCH
	    // Scale: I will look at 2^0, 2^-1, 2^-2, ..., 2^{-num_points}
	    double scale = 2;
	    for (unsigned int alpha = 1; alpha < num_points + 1; ++alpha)
	      {
		// scale u
		u.add(-pow(scale,-double(alpha-1)), *Du);
		residual = app->residual(*res, src1);
		
		if (residual < min_residual &&
		    !std::isnan(residual))
		  {
		    opt_alpha = -pow(scale,-double(alpha-1));
		    min_residual = residual;
		  }
		//  std::cout<<"Residual for "<<pow(scale,-double(alpha-1))<<" is: "<<residual<<std::endl;
		// unscale u
		u.add(pow(scale,-double(alpha-1)), *Du);
	      }
	    //std::cout<<"Norm of change in solution: "<<Du->l2_norm()<<std::endl;
	    //std::cout<<"Line search with "<<num_points<<" number of points. Optimal alpha: "<<opt_alpha<<std::endl;
	    deallog << "Step size = " << -opt_alpha << std::endl;
	    u.add(opt_alpha, *Du);
	    residual = app->residual(*res, src1);
	    deallog << "Residual = " << residual << std::endl;
		for (unsigned int i = 0; i<res->n_blocks(); i++)
			deallog << "Residual for equation "<<i<<" is: "<<res->block(i).l2_norm() << std::endl;

	    // ------ LINEAR LINE SEARCH -------------
	    /*
     	    unsigned int interval = 1;	
	    for (unsigned int alpha = 1; alpha < num_points*interval + 1; ++alpha)
	      {
		u.add(-1.0/num_points, *Du);
		residual = app->residual(*res, u);
		if (residual < min_residual &&
		    !std::isnan(residual))
		  {
		    opt_alpha = 1 - 1.0/num_points*alpha;
		    min_residual = residual;
		  }
		std::cout<<"Residual for "<<1.0/num_points*alpha<<" is: "<<residual<<std::endl;
	      }
	    std::cout<<"Norm of change in solution: "<<Du->l2_norm()<<std::endl;
	    std::cout<<"Optimal alpha: "<<opt_alpha<<std::endl;
	    u.add(opt_alpha, *Du);
	    residual = app->residual(*res, u);
	    */
	  }
	else
	  {
	    ////////////////////////////////////////////////////////////
	    // a) First, use overrelaxation during the initial iterations because Newton's algorithm
	    //tends to overshoot:
	    double alpha = 1.0;
	    const double* ref = get_data()->scalar("Refinement");
	    if (step < overrelax_steps &&  *ref < 1)
		{
	      alpha = overrelax*step;
		  deallog << "Step size = " << alpha << std::endl;
		}
	    
	    // Update solution
	    // If no overrelaxation used: 
	    //u.add(-1, *Du);
	    // using overrelaxation:
	    u.add(-alpha,*Du);
	    
	    // Update residual
	    old_residual = residual;
	    residual = app->residual(*res, src1);	

	    // b) Next, make sure that the chosen step reduces the residual overwise further
	    // reduce the residual:
	    unsigned int step_size = 0;
	    while (residual >= old_residual || std::isnan(residual))
	      {
			  ++step_size;
			  if (control.log_history())
				  deallog << "Step size: 2^{-" << step_size << "}" << std::endl;
			  if (step_size > 10)
			  {
				  if (control.log_history())
					  deallog << "Step size too small!" << std::endl;
				  break;
				  
			}
		u.add(1./(1<<step_size), *Du);
		residual = (*compute_residual)(*res, src1);
	      }
   		 deallog << "Residual = " << residual << std::endl;
		 for (unsigned int i = 0; i<res->n_blocks(); i++)
			deallog << "Residual for equation "<<i<<" is: "<<res->block(i).l2_norm() << std::endl;
	  }
	
	////////////////////////////////////////////////////////////
	// Debug output options:
	// - Debug solution: Outputs the solution at Newton iteration
	if (debug_solution || debug_update || debug_residual)
	  {
	    std::ostringstream streamOut;
	    streamOut << step;
	    std::string name = "Newton_" + streamOut.str();
	    FEVectors out;
	    out.add_vector(u, "Solution");
	    out.add_vector(*Du, "update");
	    out.add_vector(*res, "residual");
	    app->data_out(name, out);
	  }

	get_data()->block_vector_pool.free(Du);
    }

  get_data()->block_vector_pool.free(res);
  deallog.pop();

				   // in case of failure: throw
				   // exception
  if (control.last_check() != SolverControl::success)
    throw SolverControl::NoConvergence (control.last_step(),
					control.last_value());
				   // otherwise exit as normal

}
