#include <fcst_db.h>
namespace NAME =  FcstUtilities;

//------------------------------------------------------------------------------------//
NAME::DatabaseOC::DatabaseOC(){
    num_param = 0;
    max_param = 10;
}

//-----------------------------------------------------------------------------------//
bool
NAME::DatabaseOC::add_param(const std::string&  name, const double& value){
    return  add_param(name, std::to_string(value));
}

//------------------------------------------------------------------------------------//
bool
NAME::DatabaseOC::add_param(const std::string&  name, const std::string&  value){
    bool answer = false;

    if(num_param < max_param){
        std::vector<std::string> line;
        line.push_back(name);
        line.push_back(value);
        param_data.push_back(line);
        answer = true;
        num_param++;
    }

    return answer;
}

//------------------------------------------------------------------------------------//
NAME::FCSTdatabase::FCSTdatabase() {
    connected = false;

    //set default db path;
    db_path = FcstUtilities::find_fcst_root() + "main_db";

}

//------------------------------------------------------------------------------------//
NAME::FCSTdatabase::~FCSTdatabase() {
    if (connected)
        disconnect();
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::connect(){
    //db path is set to fcst_dir/main_db in constructor
    return connect(db_path);
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::connect(const std::string&  db_name){
    db_path = db_name;
    int status;

    if(connected){
        std::cerr << "FCSTdatabase:Database connection already established." << std::endl;
        return false;
    }

    std::ifstream ifile(db_name);
    if (ifile) {
        status = sqlite3_open(db_name.c_str(), &db); //sqlite returns true if it does not connect
        // The file exists, and is open for input
        if(status == SQLITE_OK)
                   connected =true;
    }
    else{
        std::cerr << "FCSTdatabase:Database not found; connection not established." << std::endl;
        connected = false;
    }


    //unable to connect or exception occured
    return connected;
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::disconnect(){
    int rc;

    if (connected){
        rc = sqlite3_close(db); //sqlite returns true if it remains connected

        if(rc == SQLITE_OK)
           connected =false;
    }

    //unable to disconnect or exception occured
    return !connected; //logic: if we are not connected then we have disconnected
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::has_data(const std::string&  model_name, const DatabaseOC& OC){
    bool answer = false;

    if (find_table(model_name, OC) != "NONE")
        answer = true; //we have found data for given args

    return answer;
}

//------------------------------------------------------------------------------------//
std::vector<std::vector<double>>
NAME::FCSTdatabase::get_data(const std::string&  model_name, const DatabaseOC& OC, const double& tolerance){

    return get_data(model_name, OC, tolerance, "");

}

//------------------------------------------------------------------------------------//
std::vector<std::vector<double>>
NAME::FCSTdatabase::get_data(const std::string&  model_name, const DatabaseOC& OC, const double& tolerance, const std::string&  orderby){

    std::vector<std::vector<double>> answer;
    std::vector<double> temp_line;
    //Check for data
    std::string table_name = find_table(model_name, OC, tolerance);
    if(table_name !="NONE"){
        //Get table name

        if(orderby != "")
            table_name += " ORDER BY " + orderby;
        //Get everything from the table
        std::vector<std::vector<std::string>> string_answer = request("SELECT * FROM " + table_name);

        //Change type from string to double
        for(int i = 0; i != string_answer.size(); i++){
            temp_line.clear();
            for(int j = 0; j != string_answer[i].size(); j++){

                try
                {
                    temp_line.push_back(std::atof(string_answer[i][j].c_str()));
                }
                catch (std::exception e)
                {
                    std::cerr << "FCSTdatabase:function get_data() cannot return value of double type./n";
                    std::cerr << e.what() << std::endl;
                    exit(1);
                }

            }
            answer.push_back(temp_line);
        }

    }
    return answer;

}

//------------------------------------------------------------------------------------//
std::vector<std::vector<double>>
NAME::FCSTdatabase::get_data(const std::string&  model_name, const DatabaseOC& OC){
    return get_data(model_name, OC, 0.00000001);
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::request(const std::string&  sqlCmd, const bool& useCallBack, int lockCounter){

    //The main difference between this function and request is that here callback is simply 0, i.e. we do not read the returned data

    bool answer = false;
    if(!connected){
        std::cerr << "FCSTdatabase:Database connection not established." << std::endl;;
        throw std::exception();
    }

    temp.clear(); //clear the temporary data

    int status;
    char *zErrMsg = 0;

    if(useCallBack)
        status = sqlite3_exec(db, sqlCmd.c_str(), callback, this, &zErrMsg);
    else
        status = sqlite3_exec(db, sqlCmd.c_str(), 0, this, &zErrMsg);

    if(status != SQLITE_OK ){
       std::cerr << "FCSTdatabase: SQL error: " << zErrMsg << ",error code:" << status << "\n" ;

       if(status == SQLITE_BUSY){
           if(lockCounter >0){
               //If a lock occurs call this function recursively after waiting a little
               std::cerr << "FCSTdatabase: Reattempting request(" << lockCounter << ")\n";
               sleep(1);
               answer = request(sqlCmd, useCallBack, lockCounter -1); //decrement lockCounter so that we only make a limited amount of attempts
           }
           else{
               std::cerr << "FCSTdatabase: attempts failed! \n";
               return false;
           }

       }

    }
    else{
        answer = true;
    }


    sqlite3_free(zErrMsg);
    return answer;
}

//------------------------------------------------------------------------------------//
std::vector<std::vector<std::string>>
NAME::FCSTdatabase::request(const std::string&  sqlCmd){
    request(sqlCmd, true);
    return temp;
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::request_no_callback(const std::string&  sqlCmd){
    return request(sqlCmd, false);
}

//------------------------------------------------------------------------------------//

bool
NAME::FCSTdatabase::commit_data(const std::string&  model_name, const DatabaseOC& OC, const std::vector<std::string>& column_names, const std::vector<std::vector<double>>& data){
    bool answer = false;

    if((column_names.size() ==0) or (data.size() ==0)){
        std::cerr << "FCSTdatabase:No data provided for " + model_name + " whilst trying to commit to db." << std::endl;
        return answer;
    }


    //Check model does not already exist for given operating conditions.
    if(has_data(model_name,OC)){
        std::cerr << "FCSTdatabase:Data for " + model_name + " already exists in db." << std::endl;
    }
    else{
        //insert new entry to head

        std::string table_name = make_new_head_entry(model_name,OC);

        if (table_name != "NONE"){
            if (create_table(table_name, column_names)){
                if (fill_empty_table(table_name, data, column_names))
                    answer = true;
            }
        }
    }
    return answer;

}

//------------------------------------------------------------------------------------//
std::string
NAME::FCSTdatabase::make_new_head_entry(const std::string&  model_name, const DatabaseOC& OC){
    std::string answer = "NONE"; //the answer is either the new table name or NONE

    //Find how many results there are for the model name
    std::vector<std::vector<std::string>> string_answer = request("SELECT COUNT(MODEL_NAME) FROM HEAD WHERE MODEL_NAME = '" + model_name + "'");
    int num = std::atoi(string_answer[0][0].c_str());
    num++;

    if(OC.param_data.size() == 0){
        std::cerr << "FCSTdatabase:Cannot commit entry with zero operating condition information." << std::endl;;
        return answer;
    }

    answer = model_name + std::to_string(num);

    //Build  the command to create the new head entry
    std::string  sqlCmd = "INSERT INTO HEAD VALUES('" + model_name + "', '";
    int blank_OC_counter = 10; //TODO: refactor this hard coded variable to constructor

    for(int i =0; i < OC.param_data.size(); i++){
        sqlCmd += OC.param_data[i][0] + "', '" + OC.param_data[i][1]  + "','";
        blank_OC_counter--;
    }

    //we need to fill in all the blanks for OC
    for(int j =0; j < blank_OC_counter; j++){
        sqlCmd +=  "', '','";
    }
    //Finish the command string
    sqlCmd += "" + answer + "');";

    //Execute the command
    if(!request_no_callback(sqlCmd))
        answer = "NONE";
    return answer;
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::create_table(const std::string&  table_name, const std::vector<std::string>& column_names){
    bool answer = false; //True if we successfully commit the table
    std::string  sqlCmd = "CREATE TABLE " + table_name + " (";

    if(column_names.size() >0){

        //Build the command to create the table
        sqlCmd += column_names[0] + " varchar(255) ";

        for (int i = 1; i < column_names.size(); i++)
            sqlCmd += ", " + column_names[i] + " varchar(255) ";

        sqlCmd += ");";

        if(request_no_callback(sqlCmd))
           answer = true;
        else
            std::cerr << "FCSTdatabase:Cannot commit entry with zero operating condition information." << std::endl;
    }

    return answer;
}


//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::fill_empty_table(const std::string&  table_name, const std::vector<std::vector<double>>& data, const std::vector<std::string>& column_names){
    bool answer = false;

    int columns = request("pragma table_info("+ table_name +");").size();
    int rows = std::atoi(request("SELECT COUNT(*) FROM "+ table_name +";")[0][0].c_str());
    std::string  sqlCmd;


    //Check Dimensions of table and data match & make sure table is empty
    if ((rows == 0) && (columns == data[0].size())){

        bool finished = false;
        int chunk_size = 480/data[0].size(); //ensure that number of selects is under 500 (SQLITE restriction)
        for(int c = 0; c < data.size() ; c+=chunk_size){
            //Create the first line of the sql command to add a row of data
            sqlCmd = "INSERT INTO "+  table_name + " SELECT ";

            for(int a = 0; a < data[c].size(); a++){
                           sqlCmd += "'" + FcstUtilities::number_to_string(data[c][a]) + "' as '" + column_names[a] + "',";
            }
            sqlCmd.resize(sqlCmd.length() -1); //remove the last ","

            //Add the data progressively
            for(int i = c+1; i < (c + chunk_size); i++){

                //Ensure chunk does not go out of bounds
                if (i == data.size()){
                    finished = true;
                    break;
                }


                sqlCmd += " UNION SELECT ";

                for(int j = 0; j < data[i].size(); j++){
                    sqlCmd += "'" + FcstUtilities::number_to_string(data[i][j]) + "',";
                }
                sqlCmd.resize(sqlCmd.length() -1); //remove the last ","


            }

            if (request_no_callback(sqlCmd))
                        answer = true;

            if (finished){
                break;
            }
        }


    }



    return answer;
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::clear_data(const std::string&  model_name, const DatabaseOC& OC){
    bool answer = false;

    std::string table_name = find_table(model_name, OC);
    //if(table_name != "NONE"){
        //Clear head entry
        std::string sqlCmd = "DELETE FROM HEAD WHERE TABLE_REF ='"+table_name +"';";
        if (request_no_callback(sqlCmd))
            answer = true;


        //Drop the table
        sqlCmd = "DROP TABLE "+table_name +";";
        if (!request_no_callback(sqlCmd))
            answer = false; //if the above request is false then we must change answer from true to false
    //}

    return answer;
}

//------------------------------------------------------------------------------------//
std::string
NAME::FCSTdatabase::find_table(const std::string&  model_name, const DatabaseOC& OC){
    return find_table(model_name, OC, 0.00000001);
}

//------------------------------------------------------------------------------------//
std::string
NAME::FCSTdatabase::find_table(const std::string&  model_name, const DatabaseOC& OC, const double& tolerance){

    //Get the relavent head data
    std::string sqlCmd = "SELECT * FROM HEAD WHERE MODEL_NAME = '" + model_name + "'";

    std::vector<std::vector<std::string>> head_info = request(sqlCmd);
    std::vector<std::string> head_line; // A line of the head info

    bool term_found;
    std::string table_ref;

    //for each line in head_info
    for(int i = 0; i != head_info.size(); i++) {

        head_line = head_info[i];
        //note the line's table_ref
        table_ref = head_line.back();

        //for each oc in OC
        for(int j = 0; j != OC.param_data.size(); j++) { //for all the OCs
           term_found = true;

            if (!find_matching_head_term(OC.param_data[j], head_line, tolerance))
            {
                term_found = false;
                break;
            }
        }

        if(term_found){ //We  found matching operating conditions
            return table_ref;
        }
    }

    return "NONE";
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::find_matching_head_term(const std::vector<std::string>& OCparam, const std::vector<std::string>& head_line, const double& tolerance){

    double diff;
    //For all the elements in head_line
    for(int i =0; i < head_line.size() -1; i++){
        //If the OCparam name is equal to the element
        if (OCparam[0] == head_line[i]){
            //check if the value is equal within tolerance

            //Zero value case
            if(std::atof(OCparam[1].c_str()) == 0.0){
                if (std::atof(head_line[i+1].c_str()) == 0.0)
                    diff = 0;
            }
            else
                diff =((std::abs(std::atof(OCparam[1].c_str()) - std::atof(head_line[i+1].c_str())))/std::atof(OCparam[1].c_str()));

            if(diff < tolerance) //if the values match within tolerance
                return true;
        }
    }
    return false;
}

//------------------------------------------------------------------------------------//
bool
NAME::FCSTdatabase::has_data(const std::string&  model_name, const DatabaseOC& OC, const double& tolerance){
    bool answer = false;
    if (find_table(model_name, OC, tolerance) != "NONE")
        answer = true; //we have found data for given args

    return answer;
}
