//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2006-13 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: fcst_utilities.cc
//    - Description: Namespace carrying a lot of utility functions
//    - Developers: Madhur Bhaiya, Valentin N. Zingan
//    - Id: $Id: fcst_utilities.cc 1380 2013-08-22 02:05:45Z zingan $ 
//
//---------------------------------------------------------------------------

#include "utils/fcst_utilities.h"

template <typename NumType>
NumType FcstUtilities::string_to_number(const std::string& str)
{
    try
    {
        return boost::lexical_cast<NumType>(str);
    }
    catch ( boost::bad_lexical_cast const& )
    {
        AssertThrow ( false, ExcWrongString(str) );
    }
}

//---------------------------------------------------------------------------

template <typename NumType>
std::string FcstUtilities::number_to_string(const NumType &num)
{
    try
    {
        return boost::lexical_cast<std::string>(num);
    }
    catch ( boost::bad_lexical_cast const& )
    {
        AssertThrow ( false, ExcInternalError());
    }
}

//---------------------------------------------------------------------------

template <typename NumType>
std::vector<NumType> FcstUtilities::string_to_number(const std::vector<std::string>& vec)
{
    std::vector<NumType> temp( vec.size() );
    for (unsigned int i=0; i<vec.size(); ++i)
        temp[i] = string_to_number<NumType>(vec[i]);
    
    return temp;
}

//---------------------------------------------------------------------------

template <typename KeyType, typename ValueType>
std::map< KeyType, ValueType > FcstUtilities::string_to_map(const std::vector<std::string>& vec)
{
    std::map< KeyType, ValueType > return_map;
    for (std::vector<std::string>::const_iterator p = vec.begin(); p != vec.end(); ++p)
    {
        //Checking if ":" exists in a string or not.
        std::string key_str = *p;
        AssertThrow ( key_str.find(":") != std::string::npos, dealii::StandardExceptions::ExcMessage("Incorrect map string (" + key_str + "); No '':'' found for FcstUtilities::string_to_map.") );
        key_str.erase( key_str.find(":"), std::string::npos );
        while ( (key_str.length() > 0) && (std::isspace(key_str[key_str.length()-1])) )
            key_str.erase( key_str.length()-1, 1 );
        KeyType key_num = string_to_number<KeyType>(key_str);
        
        //Checking if a particular key is already defined earlier or not.
        AssertThrow ( return_map.find(key_num) == return_map.end(), dealii::StandardExceptions::ExcMessage("Key value in <" + key_str + "> appears more than once in FcstUtilities::string_to_map.") );
        
        std::string value_str = *p;
        value_str.erase (0, value_str.find(":")+1);
        while ( (value_str.length() > 0) && (std::isspace(value_str[0])) )
            value_str.erase(0,1);
        ValueType value_num = string_to_number<ValueType>(value_str);
        
        return_map[key_num] = value_num;
    }
    
    return return_map;
}

//---------------------------------------------------------------------------

template< typename KeyType, typename ValueType >
std::map< KeyType, ValueType >
FcstUtilities::string_to_map(const std::string& name)
{
    const std::vector<std::string> vec = dealii::Utilities::split_string_list(name);
    
    std::map< KeyType, ValueType > return_map;
    for (std::vector<std::string>::const_iterator p = vec.begin(); p != vec.end(); ++p)
    {
        //Checking if ":" exists in a string or not.
        std::string key_str = *p;
        AssertThrow ( key_str.find(":") != std::string::npos, dealii::StandardExceptions::ExcMessage("Incorrect map string (" + key_str + "); No '':'' found for FcstUtilities::string_to_map.") );
        key_str.erase( key_str.find(":"), std::string::npos );
        while ( (key_str.length() > 0) && (std::isspace(key_str[key_str.length()-1])) )
            key_str.erase( key_str.length()-1, 1 );
        KeyType key_num = string_to_number<KeyType>(key_str);
        
        //Checking if a particular key is already defined earlier or not.
        AssertThrow ( return_map.find(key_num) == return_map.end(), dealii::StandardExceptions::ExcMessage("Key value in <" + key_str + "> appears more than once in FcstUtilities::string_to_map.") );
        
        std::string value_str = *p;
        value_str.erase (0, value_str.find(":")+1);
        while ( (value_str.length() > 0) && (std::isspace(value_str[0])) )
            value_str.erase(0,1);
        ValueType value_num = string_to_number<ValueType>(value_str);
        
        return_map[key_num] = value_num;
    }
    
    return return_map;
}

//---------------------------------------------------------------------------

template<>
std::map< unsigned int, std::string >
FcstUtilities::string_to_map<>(const std::string& name)
{
    const std::vector<std::string> vec = dealii::Utilities::split_string_list(name);
    
    std::map< unsigned int, std::string > return_map;
    for (std::vector<std::string>::const_iterator p = vec.begin(); p != vec.end(); ++p)
    {
        //Checking if ":" exists in a string or not.
        std::string key_str = *p;
        AssertThrow ( key_str.find(":") != std::string::npos, dealii::StandardExceptions::ExcMessage("Incorrect map string (" + key_str + "); No '':'' found for FcstUtilities::string_to_map.") );
        key_str.erase( key_str.find(":"), std::string::npos );
        while ( (key_str.length() > 0) && (std::isspace(key_str[key_str.length()-1])) )
            key_str.erase( key_str.length()-1, 1 );
        unsigned int key_num = string_to_number<unsigned int>(key_str);
        
        //Checking if a particular key is already defined earlier or not.
        AssertThrow ( return_map.find(key_num) == return_map.end(), dealii::StandardExceptions::ExcMessage("Key value in <" + key_str + "> appears more than once in FcstUtilities::string_to_map.") );
        
        std::string value_str = *p;
        value_str.erase (0, value_str.find(":")+1);
        while ( (value_str.length() > 0) && (std::isspace(value_str[0])) )
            value_str.erase(0,1);
        
        return_map[key_num] = value_str;
    }
    
    return return_map;
}

//---------------------------------------------------------------------------

template< typename KeyType, typename ValueType >
std::map< KeyType, std::vector<ValueType> >
FcstUtilities::split_mapvalue_list(const std::map< KeyType, std::string >& str_map, const char delimiter)
{
    std::map< KeyType, std::vector<ValueType> > return_map;
    
    for ( typename std::map< KeyType, std::string >::const_iterator iter = str_map.begin(); iter != str_map.end(); ++iter)
    {
        AssertThrow ((iter->second).find(delimiter) != std::string::npos, dealii::StandardExceptions::ExcMessage("Delimiter \"" + std::string(1, delimiter) + "\" should exist alteast once in the map value string."));
        return_map[iter->first] = string_to_number<ValueType>( dealii::Utilities::split_string_list(iter->second, delimiter) );
    }
    
    return return_map;    
}






//---------------------------------------------------------------------------

std::string
FcstUtilities::find_fcst_root()
{
    char char_path[256];
    getcwd(char_path, 255);
    std::string path = char_path;
    int string_position = path.find("data");
    
    if (string_position == std::string::npos){
        std::cerr << "Cannot find FCST directory. Are you executing from within the data folder?" << std::endl;
        AssertThrow( false , ExcInvalidState() );
    }
    
    return path.substr(0, string_position);
}

void
FcstUtilities::print_parameter_file_XML(ParameterHandler& param)
{
    std::filebuf fb;
    fb.open("default.xml", std::ios::out);
    std::ostream out(&fb);
    param.print_parameters(out,
                            ParameterHandler::OutputStyle::XML);
    fb.close();

    //Call python script to remove certain sections  from xml parameter file
    run_python("ModifyXML.py", "default.xml");
}


void
FcstUtilities::run_python(std::string script_name,std::string arg)
{
    std::string path= FcstUtilities::find_fcst_root();
    path += "python/" + script_name;
    std::string cmd = "python " + path +" " + arg;
    system(cmd.c_str());
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit instantations
//---------------------------------------------------------------------------
template unsigned int FcstUtilities::string_to_number(const std::string&);
template double FcstUtilities::string_to_number(const std::string&);
//---------------------------------------------------------------------------
template std::vector<unsigned int> FcstUtilities::string_to_number(const std::vector<std::string>&);
template std::vector<double> FcstUtilities::string_to_number(const std::vector<std::string>&);
//---------------------------------------------------------------------------
template std::map<unsigned int, double> FcstUtilities::string_to_map(const std::vector<std::string>&);
template std::map<unsigned int, double> FcstUtilities::string_to_map(const std::string&);
template std::map<unsigned char, double> FcstUtilities::string_to_map(const std::string&);
//---------------------------------------------------------------------------
template std::map<unsigned int, std::vector<double> > FcstUtilities::split_mapvalue_list(const std::map< unsigned int, std::string >&, const char);
//---------------------------------------------------------------------------
template std::string FcstUtilities::number_to_string(const double &num);