//---------------------------------------------------------------------------
//    $Id: operating_conditions.cc 1354 2013-08-17 00:01:22Z secanell $
//
//    Copyright (C) 2009 by Marc Secanell
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include "operating_conditions.h"

//---------------------------------------------------------------------------

FuelCell::OperatingConditions::OperatingConditions()
{}

//---------------------------------------------------------------------------
FuelCell::OperatingConditions::~OperatingConditions()
{}

//---------------------------------------------------------------------------
void
FuelCell::OperatingConditions::declare_parameters (ParameterHandler &param) const
{
    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection("Operating conditions");
        {
            param.declare_entry ("Temperature cell",
                                "353", // K, or 80 Celsius
                                Patterns::Double());
            param.declare_entry ("Cathode pressure",
                                 "101325", // Pa, or 1 atm
                                 Patterns::Double(),
                                 "Pressure at the cathode channel in Pa");
            param.declare_entry ("Cathode initial oxygen mole fraction (prior to humidification)",
                                 "0.21", // 
                                 Patterns::Double());
            param.declare_entry ("Cathode relative humidity",
                                 "0.7",
                                 Patterns::Double(),
                                 "Relative humidity (fraction) in the cathode channel");
            param.declare_entry ("Anode pressure",
                                 "101325", // Pa, or 1 atm
                                 Patterns::Double(),
                                 "Pressure at the anode channel in Pa");
            param.declare_entry ("Anode relative humidity",
                                 "0.7",
                                 Patterns::Double(),
                                 "Relative humidity (fraction) in the anode channel");
            param.declare_entry ("Voltage cell",
                                 "0.6", 
                                 Patterns::Double());
            param.declare_entry("Voltage drop in the anode",
                                "0.015",
                                Patterns::Double());
            param.declare_entry ("Open circuit voltage",
                                 "1.23",
                                 Patterns::Double());
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}

//---------------------------------------------------------------------------
void
FuelCell::OperatingConditions::set_parameters(const std::vector<std::string>& name_dvar,
					      const std::vector<double>& value_dvar,
					      ParameterHandler& param) const
{
  param.enter_subsection("Fuel cell data");
  {
    param.enter_subsection("Operating conditions");
	{
	for (unsigned int i=0; i<name_dvar.size(); ++i)
	{
		if (name_dvar[i] == "V_cell")
		{
			double cell_voltage = value_dvar[i];
			if (cell_voltage < 0.0) cell_voltage *= -1.0;
			param.set("Voltage cell", cell_voltage);
		}
		else if (name_dvar[i] == "T_cell")
	  		param.set("Temperature cell", value_dvar[i]);

		else if (name_dvar[i] == "dV_a")
	  		param.set("Voltage drop in the anode", value_dvar[i]);
		
		else if (name_dvar[i] == "P_c")
	  		param.set("Cathode pressure", value_dvar[i]);

		else if (name_dvar[i] == "P_a")
	  		param.set("Anode pressure", value_dvar[i]);

		else if (name_dvar[i] == "RH_a")
	  		param.set("Anode relative humidity", value_dvar[i]);

		else if (name_dvar[i] == "RH_c")
	  		param.set("Cathode relative humidity", value_dvar[i]);

		else if (name_dvar[i] == "OCV")
	  		param.set("Open circuit voltage", value_dvar[i]);
	}
	}
	param.leave_subsection();
  }
  param.leave_subsection();
}

//---------------------------------------------------------------------------
void
FuelCell::OperatingConditions::initialize (ParameterHandler& param)
{
    R = Constants::R();
    
    param.enter_subsection("Fuel cell data");
    {
        param.enter_subsection("Operating conditions");
        {
            T_cell = param.get_double("Temperature cell");
            // Cathode
            p_c = param.get_double("Cathode pressure");
            c_c = p_c/(R*T_cell)*1E-6; //mol/cm^3
            RH_c = param.get_double("Cathode relative humidity");
            channel_oxygen_mole_fraction = param.get_double("Cathode initial oxygen mole fraction (prior to humidification)");
            // Anode
            p_a = param.get_double("Anode pressure");
            c_a = p_a/(R*T_cell)*1E-6; //mol/cm^3
            RH_a = param.get_double("Anode relative humidity");
            // Cell voltage
            V_cell = param.get_double("Voltage cell");
            dV_a = param.get_double("Voltage drop in the anode");
            OCV = param.get_double("Open circuit voltage");
            
            if (OCV > voltage_cell_th()) // NOTE: This also initializes E_th
                OCV = E_th;
        }
        param.leave_subsection();
    }
    param.leave_subsection();
}


//---------------------------------------------------------------------------
double
FuelCell::OperatingConditions::saturation_pressure()
{
  double T_celsius = T_cell - 273;
  return pow(10,(-2.1794+0.02953*T_celsius-0.000091837*pow(T_celsius,2)+0.00000014454*pow(T_celsius,3)));
}

//---------------------------------------------------------------------------
double 
FuelCell::OperatingConditions::get_x_wv()
{
  return RH_c*saturation_pressure()*101325/p_c; 
}

//---------------------------------------------------------------------------
double 
FuelCell::OperatingConditions::get_x_o2()
{
  return channel_oxygen_mole_fraction*(p_c-saturation_pressure()*RH_c*101325)/p_c; 
}

//---------------------------------------------------------------------------
double 
FuelCell::OperatingConditions::get_x_h2()
{
  return (p_a-saturation_pressure()*RH_a*101325)/p_a; 
}

//---------------------------------------------------------------------------
double 
FuelCell::OperatingConditions::voltage_cell_th()
{
  double p_h2 = p_a*get_x_h2()/101325;
  double p_o2 = p_c*get_x_o2()/101325;
  E_th  = 1.229 - (T_cell-298.15)*8.456E-4 + T_cell*4.31E-5*(log(p_h2)+0.5*log(p_o2));
  return E_th;
}


//---------------------------------------------------------------------------
void 
FuelCell::OperatingConditions::print_operating_conditions()
{
	deallog<<"========= OPERATING CONDITIONS ========"<<std::endl;
	deallog<<"Temperature: "<< T_cell <<std::endl;
	deallog<<"Cathode Pressure: "<< p_c <<std::endl;
	deallog<<"Cathode RH: "<< RH_c <<std::endl;
	deallog<<"Anode Pressure: "<< p_a <<std::endl;
	deallog<<"Anode RH: "<< RH_a <<std::endl;
	deallog<<"Open Circuit Voltage: "<< OCV <<std::endl;
	deallog<<"Cell Voltage: "<< V_cell <<std::endl;
	deallog<<"======================================="<<std::endl;
}