//---------------------------------------------------------------------------
//
//    FCST: Fuel Cell Simulation Toolbox
//
//    Copyright (C) 2013 by Energy Systems Design Laboratory, University of Alberta
//
//    This software is distributed under the MIT License.
//    For more information, see the README file in /doc/LICENSE
//
//    - Class: simulation_selector.cc
//    - Description: Used to instantiate optimization routines
//    - Developers: P. Dobson and M. Secanell
//    - $Id: simulation_selector.cc 1447 2013-08-30 21:12:32Z madhur $
//
//---------------------------------------------------------------------------

#include "utils/simulation_selector.h"

using namespace boost;

//---------------------------------------------------------------------------

template <int dim>
SimulationSelector<dim>::SimulationSelector()
{

}

//---------------------------------------------------------------------------
template <int dim>
SimulationSelector<dim>::~SimulationSelector()
{

}

//---------------------------------------------------------------------------
template <int dim>
void SimulationSelector<dim>::declare_parameters(ParameterHandler& param) const
{
     param.enter_subsection("Simulator");
     {
          param.declare_entry("simulator name",
                              "MEA",
                              Patterns::Selection("cathode | MEA | test_mesh | laplace"),
                              "");

          param.declare_entry("solver name",
                              "Newton3ppC",
                              Patterns::Selection("Linear | NewtonBasic | Newton3ppC | Newton3pp | NewtonLineSearch"),
                              "");

          param.declare_entry("solver method",
                              "adaptive refinement",
                              Patterns::Selection("adaptive refinement"),
                              "");
     }
     param.leave_subsection();
}


//---------------------------------------------------------------------------
template <int dim>
void SimulationSelector<dim>::initialize(ParameterHandler& param)
{
     param.enter_subsection("Simulator");
     {
          name_application  = param.get("simulator name");
          name_solver       = param.get("solver name");
          name_solve_method = param.get("solver method");
     }
     param.leave_subsection();
}

//---------------------------------------------------------------------------
template <int dim>
shared_ptr<AppFrame::OptimizationBlockMatrixApplication<dim> >
SimulationSelector<dim>::select_application()
{
    //-- Cathode without convection, pseudo-homogeneous
    if (name_application.compare("cathode") == 0)
    {
        deallog << "YOU ARE CURRENTLY SOLVING A CATHODE MODEL" << std::endl;
        return shared_ptr<FuelCell::Application::AppCathode<dim> > (new FuelCell::Application::AppCathode<dim>);
    }
    //-- Complete MEA without convection
    else if (name_application.compare("MEA") == 0)
    {
        deallog << "YOU ARE CURRENTLY SOLVING AN MEA MODEL" << std::endl;
        return shared_ptr<FuelCell::Application::AppPemfc<dim> > (new FuelCell::Application::AppPemfc<dim>);

    }
    else if (name_application.compare("laplace") == 0)
    {
        deallog << "YOU ARE CURRENTLY SOLVING A Laplace MODEL" << std::endl;
        return shared_ptr<FuelCell::Application::AppLaplace<dim> > (new FuelCell::Application::AppLaplace<dim>);
    }
    //run mesh test application
    else if (name_application.compare("test_mesh") == 0)
    {
        deallog << "YOU ARE CURRENTLY RUNNING THE TEST MESH APPLICATION" << std::endl;
        return shared_ptr<FuelCell::Application::AppReadMesh<dim> > (new FuelCell::Application::AppReadMesh<dim>);
    }
    else
    {
        Assert(false, ExcNotImplemented());
        deallog << "Application not implemented. See " << __FILE__ << std::endl;
    }
}

//---------------------------------------------------------------------------
template <int dim>
shared_ptr<AppFrame::ApplicationCopy> //shared_ptr<AppFrame::newtonBase>
SimulationSelector<dim>::select_solver(AppFrame::OptimizationBlockMatrixApplication<dim>* app_lin)
{
     if  (name_solver.compare("Linear") == 0)
     {
        deallog << "YOUR PROBLEM IS LINEAR" << std::endl;
        return shared_ptr<AppFrame::ApplicationCopy> (new AppFrame::ApplicationCopy(*app_lin));
     }
     else if( name_solver.compare("NewtonBasic") == 0 )
     {
        deallog << "YOU ARE USING NewtonBasic NEWTON SOLVER" << std::endl;
        return shared_ptr<AppFrame::NewtonBasic> ( new AppFrame::NewtonBasic(*app_lin) );
     }
     else if (name_solver.compare("Newton3ppC") == 0)
     {
        deallog << "YOU ARE USING Newton3ppC NEWTON SOLVER" << std::endl;
        return shared_ptr<AppFrame::Newton3ppC> (new AppFrame::Newton3ppC(*app_lin));
     }
     else if (name_solver.compare("Newton3pp") == 0)
     {
        deallog << "YOU ARE USING Newton3pp NEWTON SOLVER" << std::endl;
        return shared_ptr<AppFrame::Newton3pp> (new AppFrame::Newton3pp(*app_lin));
     }
     else if (name_solver.compare("NewtonLineSearch") == 0)
     {
        deallog << "YOU ARE USING NewtonLineSearch NEWTON SOLVER" << std::endl;
        return shared_ptr<AppFrame::NewtonLineSearch> (new AppFrame::NewtonLineSearch(*app_lin));
     }
     else
     {
        Assert(false, ExcNotImplemented());
        deallog << "This newton solver is not implemented yet! See " << __FILE__ << std::endl;
     }
}

//---------------------------------------------------------------------------
template <int dim>
shared_ptr<AppFrame::AdaptiveRefinement<dim> >
SimulationSelector<dim>::select_solver_method(AppFrame::OptimizationBlockMatrixApplication<dim>* app_lin,
                                              AppFrame::ApplicationCopy* newton_solver,
                                              const AppFrame::FEVector& solution)
{
    if (name_solve_method.compare("adaptive refinement") == 0)
    {
        return shared_ptr<AppFrame::AdaptiveRefinement<dim> > (new AppFrame::AdaptiveRefinement<dim> (*app_lin, *newton_solver, solution));
    }
    else
    {
        Assert(false, ExcNotImplemented());
        deallog << "This solver is not implemented yet! See " << __FILE__ << std::endl;
    }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit Instantiations
template class SimulationSelector<deal_II_dimension>;
