//---------------------------------------------------------------------------
//
//    Copyright (C) 2011 by M. Secanell, P Dobson, Valentin N. Zingan
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//---------------------------------------------------------------------------

#include "simulator_builder.h"

using namespace boost;

//---------------------------------------------------------------------------

template<int dim>
SimulatorBuilder<dim>::SimulatorBuilder()
:
dakota_use(false),
dakota_direct(false),
program_name("Fuel Cell"),
program_version("Version 0.1, July 2013")
{
}

template<int dim>
SimulatorBuilder<dim>::~SimulatorBuilder()
{
}

template<int dim>
void SimulatorBuilder<dim>::usage(int exit_value)
{
     //prints out a formatted usage message,
     //and exits using the exit_value
     std::cerr << "usage:\n"
             << program_name << " "
             << "[-h][-v]  "
             << "file_name \n\t"
             << "[-D parameter_file results_file] \n\t"
             << "where [] indicates optional option: \n\n\t"
             << "-h: help. \n\t\t"
             << "generates this message and exits \n\n\t"
             << "-v: version. \n\t\t"
             << "prints version information and exists \n\n\t"
             << "-D parameter_file results_file: \n\t\t"
             << "interface with optimization code DAKOTA.\n\t\t"
             << "Only for use in standard DAKOTA input file\n\t\t"
             << "'parameter_file' is read by DAKOTA and sets optimization parameters\n\t\t"
             << "'results_file' returns all necessary responses to DAKOTA\n\n\t"
             << "Authors: M. Secanell, Peter Dobson,\n\t"
             << "Ali Malekpourkoupaei and Michael Moore\n\t"
             << "Energy Systems Design Lab, University of Alberta\n\t"
             << "\t (c) Copyright 2007-2013." << std::endl;

     exit(exit_value);
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder <dim>::print_logo() const
{
     // Tell user program has started succesfully:
     deallog << "=======================================================" << std::endl;
     deallog << " WELCOME TO THE FUEL CELL SIMULATION TOOLBOX           " << std::endl;
     deallog << "                                                       " << std::endl;
     deallog << " Note: This code is developed using the deal.II finite " << std::endl;
     deallog << " element libraries (www.dealii.org).                   " << std::endl;
     deallog << "                                                       " << std::endl;
     deallog << " Acknowledgements:                                     " << std::endl;
     deallog << "        Dr. Guido Kanschat                             " << std::endl;
     deallog << "                                                       " << std::endl;
     deallog << " (c) Copyright 2007-2013.                              " << std::endl;
     deallog << " Authors: M. Secanell, Peter Dobson,                   " << std::endl;
     deallog << " Ali Malekpourkoupaei, Michael Moore,                  " << std::endl;
     deallog << " Madhur Bhaiya, Valentin N. Zingan                     " << std::endl;
     deallog << " Philip Wardlaw, and Kailyn Domican                    " << std::endl;
     deallog << " Energy Systems Design Lab, University of Alberta      " << std::endl;
     deallog << "                                                       " << std::endl;
     deallog << "=============== START COMPUTATIONS ====================" << std::endl;
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::print_timer_info() const
{
      deallog.push("MAIN");
      deallog << "The program was executed in: " << timer.wall_time() << " seconds" << std::endl;
      deallog << "=============== END ====================" << std::endl;
      deallog.pop();
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::parse_inputs(int argc, char *argv[])
{
//Auxiliary variables and flags:
     int num_input_file = 0;
     int dakota_file_on = 0;

     for (int ix = 1; ix < argc; ++ix)
     {
       //std::cout<<"argv["<< ix<<"]"<<argv[ix] <<std::endl;

          char *pchar = argv[ix];
          switch (pchar[0])
          {
                    //Look if it is a flag
               case'-':
               {
                    //std::cout<<" case \'-\' found\n";
                    switch (pchar[1])
                    {
                         case 'v':
                         {
                              std::cout << program_name << " :: "
                                      << program_version << std::endl;
                              exit(0);
                         }
                         case 'h':
                         {
                              usage(0);
                              //no necessary break because usage exits
                         }
                         case 'D':
                         {
                              dakota_file_on = 2;
                              dakota_use = true;
                              break;
                         }

                         default:
                         {
                              std::cerr << program_name
                                      << " : error : unrecognized option: - "
                                      << pchar[1] << "\n\n";
                              //print info and return error
                              usage(-1);
                         }
                    }
                    break;
               }
               default: //a file name with SimulatorBuilder options

                    if (dakota_file_on == 2)
                    {
                         dakota_file_on -= 1;
                         dakota_parameters = pchar;
                    }
                    else if (dakota_file_on == 1)
                    {
                         dakota_file_on -= 1;
                         dakota_results = pchar;
                    }
                    else
                    {
                         input_file = std::string(pchar);
                         num_input_file += 1;
                         break;
                    }
          }
     }

     if (num_input_file == 0 || num_input_file > 1)
     {
          std::cerr << program_name
                  << " : error : "
                  << num_input_file
                  << " input files specified. You need to provide only one input file \n\n";
          usage(-3);
     }

     deallog.push("MAIN");
     deallog << "Program:" << argv[0] << std::endl;
     print_logo();
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::scan()
{
  sim_selector = shared_ptr<SimulationSelector<dim> > (new SimulationSelector<dim>);
  this->declare_parameters(param);

  if(!input_file.empty())
  {
         std::size_t pos = input_file.find(".");
         std::string ext = input_file.substr(pos);

         if( ext.compare(".prm") == 0 )
         {
                deallog << "Parameters: " << input_file << std::endl;
                param.read_input(input_file,
                                 true);
         }
         else if( ext.compare(".xml") == 0 )
         {
                std::filebuf fb;
                fb.open(input_file,
                        std::ios::in);
                std::istream is(&fb);
                deallog << "Parameters: " << input_file << std::endl;
                param.read_input_from_xml(is);
         }
         else
         {
                deallog << "Extension " << ext << " is not handled" << std::endl;
                AssertThrow( false , ExcNotImplemented() );
         }
  }

  initialize(param);
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::run()
{
      if( run_tests )
      {
             timer.restart();
             run_test();
             timer.stop();

             print_timer_info();
      }
      else
      {
             if( dakota_use || dakota_direct )
             {
                    timer.restart();
                    run_optimization();
                    timer.stop();

                    print_timer_info();
             }
             else
             {
                    app_lin = sim_selector->select_application();                 // select the application that reflects the specific of the linearized problem at hand
                    newton  = sim_selector->select_solver( app_lin.get() );       // select one of the available Newton solvers to march over iterations of the original nonlinear problem
                    solver  = sim_selector->select_solver_method( app_lin.get(),
                                                                  newton.get() ); // select the refinement strategy you want to use

                    if( simulator_parameter_file_names.size() == 0 ) // we proceed as we did before
                    {
                           deallog << "Run program using input file: " << simulator_parameter_file_name << std::endl;
                           deallog.pop();

                           timer.restart();
                           solver->solve( simulator_parameter_file_name,
                                          param );
                           timer.stop();

                           print_timer_info();
                    }
                    else // something new happens ...
                    {
                           for(unsigned int i = 0;
                                            i < simulator_parameter_file_names.size();
                                          ++i) // we loop over all xxx_data.prm files related to our application
                           {
                                  deallog << "Run program using input file: " << simulator_parameter_file_names[i] << std::endl;
                                  deallog.pop();

                                  if( i != 0 ) // if this is NOT the first instantiation of this loop, then we:
                                               // - save the final solution
                                               // - reset all data pointers
                                  {
                                         const AppFrame::FEVector solution( solver->get_solution() );

                                         param.clear();

                                         solver.reset();
                                         newton.reset();
                                         app_lin.reset();

                                         app_lin = sim_selector->select_application();
                                         newton  = sim_selector->select_solver( app_lin.get() );
                                         solver  = sim_selector->select_solver_method( app_lin.get(),
                                                                                       newton.get(),
                                                                                       solution );
                                  }

                                  timer.restart();
                                  solver->solve( simulator_parameter_file_names[i],
                                                 param );
                                  timer.stop();

                                  print_timer_info();
                           }
                    }
             }
      }
}

//---------------------------------------------------------------------------
// PRIVATE:
//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::declare_parameters(ParameterHandler& param) const
{
	param.enter_subsection("Simulator");
	{
		param.declare_entry("Dakota direct",
				    "false",
							 Patterns::Bool(),
								  "");
		param.declare_entry("simulator parameter file name",
								  "data_app_pemfc.prm",
							 Patterns::Anything(),
								  "");
                param.declare_entry("simulator parameter file names",
                                    """",
                                     Patterns::List( Patterns::Anything() ),
                                    " ");
		param.declare_entry("optimization parameter file name",
								  "opt_data_app_pemfc.prm",
							 Patterns::Anything(),
								  "");
		param.declare_entry("Run tests",
								  "false",
							 Patterns::Bool(),
								  "");
		param.declare_entry("save transfer files",
									"false",
									Patterns::Bool(),
									"If set to false the mesh and solution from a parametric study will be deleted.");
	}
	param.leave_subsection();

	param.enter_subsection("Log");
	{
		param.declare_entry("Logfile name", "logfile.log", Patterns::Anything());
		param.declare_entry("FileDepth", "10000", Patterns::Integer());
		param.declare_entry("ConsoleDepth", "10000", Patterns::Integer());
		param.declare_entry("Time", "false", Patterns::Bool());
	}
	param.leave_subsection();

	sim_selector->declare_parameters(param);
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::initialize(ParameterHandler& param)
{
	param.enter_subsection("Simulator");
	{
		dakota_direct = param.get_bool("Dakota direct");
		simulator_parameter_file_name = param.get("simulator parameter file name");

                if(   !param.get("simulator parameter file names").empty()   )
                {
                       simulator_parameter_file_names = dealii::Utilities::split_string_list( param.get("simulator parameter file names") );
                }

		optimization_parameter_file_name = param.get("optimization parameter file name");
		run_tests = param.get_bool("Run tests");
		save_transfer_files = param.get_bool("save transfer files");
	}
	param.leave_subsection();

	// Set parameters for execution log
	param.enter_subsection("Log");
	{
		this->open_logfile(param.get("Logfile name"));
		deallog.depth_file(param.get_integer("FileDepth"));
		deallog.depth_console(param.get_integer("ConsoleDepth"));
		deallog.log_execution_time(param.get_bool("Time"));
	}
	param.leave_subsection();

	sim_selector->initialize(param);
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::run_optimization()
{
    deallog.pop();


    if (dakota_direct)
    {
        #ifdef _WITH_DAKOTA
        // NOTE: Must declare these in order for parameter handler to not complain when reading the parameter file specified.
        //        Not exclusively required for dakota application to run.
        Dakota::ParallelLibrary parallel_lib;
        shared_ptr<Dakota::ProblemDescDB> problem_db(new Dakota::ProblemDescDB (parallel_lib));
        SIM::DakotaApplication optimization(problem_db, optimization_parameter_file_name);
        optimization.declare_parameters(param);
        optimization.manage_inputs(param);

        Dakota::DirectApplicInterface* optimization_interface;

        if (optimization.use_NLS())
        {
            deallog<<"Entering DakotaLeastSquaresInterface"<<std::endl;
            optimization_interface = new SIM::DakotaLeastSquaresInterface<dim> (optimization, problem_db, param, sim_selector, simulator_parameter_file_name);
        }
        else
        {
            deallog<<"Entering DakotaDirectInterface"<<std::endl;
            optimization_interface = new SIM::DakotaDirectInterface<dim > (optimization, problem_db, param, sim_selector, simulator_parameter_file_name);
        }

        optimization.assign_interface(optimization_interface);
        optimization.run();

        deallog << "Optimization completed" << std::endl;
        if (!save_transfer_files)
        {
            remove( "._mesh_to_be_transfered.msh" );
            remove( ".transfer_solution.FEVector" );
        }
        //Delete
        // Allows results to be retrieved  --> TODO: needs to be redefined for multi-start methods
        // 		Dakota::RealVector vars = optimization.vars_results();
        // 		Dakota::RealVector resp = optimization.resp_results();
        // 		deallog << "Results retrieved" << std::endl;
        #else
        deallog<<"Dakota not available"<<std::endl;
        #endif
    }
    else if (dakota_use)
    {

        app_lin = sim_selector->select_application();
        newton = sim_selector->select_solver(app_lin.get());
        solver = sim_selector->select_solver_method(app_lin.get(), newton.get());

        SIM::DakotaInterface<dim> optimization (simulator_parameter_file_name,
                                                param,
                                                dakota_parameters,
                                                dakota_results,
                                                *app_lin,
                                                *newton);
        optimization.run();
        deallog << "Optimization completed" << std::endl;
        if (!save_transfer_files)
        {
            remove( "._mesh_to_be_transfered.msh" );
            remove( ".transfer_solution.FEVector" );
        }

    }
    else
    {
        abort();
    }
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::run_test()
{
	deallog <<	   "==================Running Unit Tests===================" << std::endl;

	if (FcstTestSuite::run_tests())
		deallog << "===========Unit Tests Successfully Completed===========" << std::endl;
	else
		deallog << "==========Unit Tests Completed with Errors!============" << std::endl;
}

//---------------------------------------------------------------------------
template<int dim>
void SimulatorBuilder<dim>::open_logfile (const std::string& extname)
{
				   // If a parameter file was
				   // specified, use its base name for
				   // the logfile.
  std::string name = extname;

				   // Still no name, use hostname,
				   // allowing parallel programs to
				   // write into different logfiles.
  if (name == "")
    {
      char log_name[40];
      gethostname(log_name, 19);
      name = log_name;
    }

  std::string::size_type pos = name.rfind(".log");
  if (pos < name.size())
    {
      name.erase(pos);
    }
  name += ".log";

  log_file = boost::shared_ptr<std::ofstream> (new std::ofstream(name.c_str()));
				   // Log into file
  deallog.attach(*log_file);
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Explicit Instantiations
template class SimulatorBuilder<deal_II_dimension>;