// ----------------------------------------------------------------------------
//
// FCST: Fuel Cell Simulation Toolbox
//
// Copyright (C) 2006-2013 by Energy Systems Design Laboratory, University of Alberta
//
// This software is distributed under the MIT License
// For more information, see the README file in /doc/LICENSE
//
// - Class: system_management.cc
// - Description: This class manages systems of equations
// - Developers: Valentin N. Zingan,    University of Alberta
//                        Marc Secanell Gallart, University of Alberta
//                        Madhur Bhaiya, University of Alberta
// - Id: $Id: system_management.cc 1449 2013-08-30 21:18:58Z madhur $
//
// ----------------------------------------------------------------------------

#include "system_management.h"

       //////////////////////////////////////////////////
       //////////////////////////////////////////////////
       // CONSTRUCTORS, DESTRUCTOR, AND INITIALIZATION //
       //////////////////////////////////////////////////
       //////////////////////////////////////////////////

// ---             ---
// --- Constructor ---
// ---             ---

FuelCell::SystemManagement::SystemManagement()
{
  this->set_all_solution_names();
  this->set_all_equation_names();
}

// ---             ---
// --- Constructor ---
// ---             ---

FuelCell::SystemManagement::SystemManagement(MeshWorker::BlockInfo&        block_info,
                                             Table<2, DoFTools::Coupling>& cell_couplings,
                                             Table<2, DoFTools::Coupling>& flux_couplings)
:
block_info(&block_info),
cell_couplings(&cell_couplings),
flux_couplings(&flux_couplings)
{
  this->set_all_solution_names();
  this->set_all_equation_names();
}

// ---            ---
// --- Destructor ---
// ---            ---

FuelCell::SystemManagement::~SystemManagement()
{ }

// ---                    ---
// --- declare_parameters ---
// ---                    ---

void
FuelCell::SystemManagement::declare_parameters(ParameterHandler& param) const
{
  param.enter_subsection("System management");
  {
    param.declare_entry("Number of solution variables",
                        "1",
                         Patterns::Integer(),
                        " ");

    param.enter_subsection("Solution variables");
    {
      const unsigned int n_sol_vars_max = 100;
      for(unsigned int index = 1; index <= n_sol_vars_max; ++index)
      {
             std::ostringstream streamOut;
             streamOut << index;
             const std::string name = "Solution variable " + streamOut.str();

             param.declare_entry(name.c_str(),
                                "no_name",
                                 Patterns::Anything(),
                                " ");
      }
    }
    param.leave_subsection();

    param.enter_subsection("Equations");
    {
      const unsigned int n_eqs_max = 100;
      for(unsigned int index = 1; index <= n_eqs_max; ++index)
      {
             std::ostringstream streamOut;
             streamOut << index;
             const std::string name = "Equation " + streamOut.str();

             param.declare_entry(name.c_str(),
                                "no_name",
                                 Patterns::Anything(),
                                " ");
      }
    }
    param.leave_subsection();
  }
  param.leave_subsection();
}

// ---            ---
// --- initialize ---
// ---            ---

void
FuelCell::SystemManagement::initialize(ParameterHandler& param)
{
  param.enter_subsection("System management");
  {
    n_solution_names = param.get_integer("Number of solution variables");

    param.enter_subsection("Solution variables");
    {
      for(unsigned int index = 1; index <= n_solution_names; ++index)
      {
             std::ostringstream streamOut;
             streamOut << index;
             const std::string name = "Solution variable " + streamOut.str();

             solution_names.push_back( param.get(name.c_str()) );
      }
    }
    param.leave_subsection();

    param.enter_subsection("Equations");
    {
      for(unsigned int index = 1; index <= n_solution_names; ++index)
      {
             std::ostringstream streamOut;
             streamOut << index;
             const std::string name = "Equation " + streamOut.str();

             equation_names.push_back( param.get(name.c_str()) );
      }
    }
    param.leave_subsection();
  }
  param.leave_subsection();

  this->check_solution_names();
  this->check_equation_names();
}

       ///////////////////////
       ///////////////////////
       // SERVICE FUNCTIONS //
       ///////////////////////
       ///////////////////////

// ---                      ---
// --- solution_in_userlist ---
// ---                      ---

const bool
FuelCell::SystemManagement::solution_in_userlist(const std::string& name) const
{
  const std::vector<std::string>::const_iterator iter = std::find(solution_names.begin(),
                                                                  solution_names.end(),
                                                                  name);
  if( iter != solution_names.end() )
    return true;
  else
    return false;
}

// ---                        ---
// --- solution_name_to_index ---
// ---                        ---

const unsigned int
FuelCell::SystemManagement::solution_name_to_index(const std::string& name) const
{
  const std::vector<std::string>::const_iterator iter = std::find(solution_names.begin(),
                                                                  solution_names.end(),
                                                                  name);

  AssertThrow( iter != solution_names.end(), VariableNotFoundInUserVariables("variable", name) );

  return iter - solution_names.begin();
}

// ---                        ---
// --- equation_name_to_index ---
// ---                        ---

const unsigned int
FuelCell::SystemManagement::equation_name_to_index(const std::string& name) const
{
  const std::vector<std::string>::const_iterator iter = std::find(equation_names.begin(),
                                                                  equation_names.end(),
                                                                  name);

  AssertThrow( iter != equation_names.end(), EquationNotFoundInUserEquations("equation", name) );

  return iter - equation_names.begin();
}

// ---                    ---
// --- matrix_block_index ---
// ---                    ---

const unsigned int
FuelCell::SystemManagement::matrix_block_index(const std::string& equation_name,
                                               const std::string& solution_name) const
{
  const unsigned int line   = this->equation_name_to_index(equation_name);
  const unsigned int column = this->solution_name_to_index(solution_name);

  AssertThrow( (*cell_couplings)(line,column) != 0, SystemMatrixBlockDoesNotExist("requested system matrix block",
                                                                                   line,
                                                                                   column,
                                                                                   solution_name,
                                                                                   equation_name) );

  unsigned int number = 0;

  for(unsigned int eq = 0; eq < equation_names.size(); ++eq)
    for(unsigned int var = 0; var < solution_names.size(); ++var)
    {
      if( eq == line && var == column )
        return number;

      if( (*cell_couplings)(eq,var) != 0 )
        number++;
    }
}

// ---                     ---
// --- make_cell_couplings ---
// ---                     ---

void
FuelCell::SystemManagement::make_cell_couplings(const std::vector<couplings_map>& src)
{
  if( src.size() == 0 )
  {
         std::cerr << "FuelCell::SystemManagement::make_cell_couplings function: The argument is empty" << std::endl;
         return;
  }

  for(unsigned int i = 0; i < src.size(); ++i)
         if( src[i].empty() )
         {
                std::cerr << "FuelCell::SystemManagement::make_cell_couplings function: Outer map/maps of the argument is/are empty" << std::endl;
                return;
         }

  for(unsigned int i = 0; i < src.size(); ++i)
         for(typename couplings_map::const_iterator iter  = src[i].begin();
                                                    iter != src[i].end();
                                                  ++iter)
         {
                std::map<std::string, DoFTools::Coupling> tmp = iter->second;
                if( tmp.empty() )
                {
                       std::cerr << "FuelCell::SystemManagement::make_cell_couplings function: Inner map/maps of the argument is/are empty" << std::endl;
                       return;
                }
         }

  (*cell_couplings).reinit(equation_names.size(), solution_names.size());

  for(unsigned int index = 0; index < src.size(); ++index)
  {
    couplings_map cell_couplings_i = src[index];
    couplings_map::const_iterator iter;

    for( iter = cell_couplings_i.begin(); iter != cell_couplings_i.end(); ++iter )
    {
      std::map<std::string, DoFTools::Coupling> int_map = iter->second;
      std::map<std::string, DoFTools::Coupling>::const_iterator int_iter;

      for( int_iter = int_map.begin(); int_iter != int_map.end(); ++int_iter )
      {
               const unsigned int line   = this->equation_name_to_index(iter->first);
               const unsigned int column = this->solution_name_to_index(int_iter->first);

               (*cell_couplings)(line,column) = int_iter->second;
      }
    }
  }
}

// ---                     ---
// --- make_flux_couplings ---
// ---                     ---

void
FuelCell::SystemManagement::make_flux_couplings(const std::vector<couplings_map>& src)
{
  if( src.size() == 0 )
  {
         std::cerr << "FuelCell::SystemManagement::make_flux_couplings function: The argument is empty" << std::endl;
         return;
  }

  for(unsigned int i = 0; i < src.size(); ++i)
         if( src[i].empty() )
         {
                std::cerr << "FuelCell::SystemManagement::make_flux_couplings function: Outer map/maps of the argument is/are empty" << std::endl;
                return;
         }

  for(unsigned int i = 0; i < src.size(); ++i)
         for(typename couplings_map::const_iterator iter  = src[i].begin();
                                                    iter != src[i].end();
                                                  ++iter)
         {
                std::map<std::string, DoFTools::Coupling> tmp = iter->second;
                if( tmp.empty() )
                {
                       std::cerr << "FuelCell::SystemManagement::make_flux_couplings function: Inner map/maps of the argument is/are empty" << std::endl;
                       return;
                }
         }

  (*flux_couplings).reinit(equation_names.size(), solution_names.size());

  for(unsigned int index = 0; index < src.size(); ++index)
  {
    couplings_map cell_couplings_i = src[index];
    couplings_map::const_iterator iter;

    for( iter = cell_couplings_i.begin(); iter != cell_couplings_i.end(); ++iter )
    {
      std::map<std::string, DoFTools::Coupling> int_map = iter->second;
      std::map<std::string, DoFTools::Coupling>::const_iterator int_iter;

      for( int_iter = int_map.begin(); int_iter != int_map.end(); ++int_iter )
      {
               const unsigned int line   = this->equation_name_to_index(iter->first);
               const unsigned int column = this->solution_name_to_index(int_iter->first);

               (*flux_couplings)(line,column) = int_iter->second;
      }
    }
  }
}

       /////////////////////
       /////////////////////
       // OTHER FUNCTIONS //
       /////////////////////
       /////////////////////

// ---                        ---
// --- set_all_solution_names ---
// ---                        ---

void
FuelCell::SystemManagement::set_all_solution_names()
{
  // GROUP 1

  all_solution_names.push_back("oxygen_molar_fraction");
  all_solution_names.push_back("water_molar_fraction");
  all_solution_names.push_back("hydrogen_molar_fraction");
  all_solution_names.push_back("nitrogen_molar_fraction");
  all_solution_names.push_back("protonic_electrical_potential");
  all_solution_names.push_back("electronic_electrical_potential");
  all_solution_names.push_back("membrane_water_content");
}

// ---                        ---
// --- set_all_equation_names ---
// ---                        ---

void
FuelCell::SystemManagement::set_all_equation_names()
{
  // GROUP 1

  all_equation_names.push_back("Ficks Transport Equation - oxygen");
  all_equation_names.push_back("Ficks Transport Equation - water");
  all_equation_names.push_back("Ficks Transport Equation - hydrogen");
  all_equation_names.push_back("Ficks Transport Equation - nitrogen");
  all_equation_names.push_back("Proton Transport Equation");
  all_equation_names.push_back("Electron Transport Equation");
  all_equation_names.push_back("Membrane Water Content Transport Equation");
}

// ---                      ---
// --- check_solution_names ---
// ---                      ---

void
FuelCell::SystemManagement::check_solution_names() const
{
  for(unsigned int index = 0; index < solution_names.size(); ++index)
  {
    const std::string name = solution_names[index];

    const std::vector<std::string>::const_iterator iter = std::find(all_solution_names.begin(),
                                                                    all_solution_names.end(),
                                                                    name);

    AssertThrow( iter != all_solution_names.end(), VariableNotFoundInFCSTVariables("variable", name) );
  }
}

// ---                      ---
// --- check_equation_names ---
// ---                      ---

void
FuelCell::SystemManagement::check_equation_names() const
{
  for(unsigned int index = 0; index < equation_names.size(); ++index)
  {
    const std::string name = equation_names[index];

    const std::vector<std::string>::const_iterator iter = std::find(all_equation_names.begin(),
                                                                    all_equation_names.end(),
                                                                    name);

    AssertThrow( iter != all_equation_names.end(), EquationNotFoundInFCSTEquations("equation", name) );
  }
}

       ////////////////////////
       ////////////////////////
       // ACCESSORS AND INFO //
       ////////////////////////
       ////////////////////////

// ---                   ---
// --- print_system_info ---
// ---                   ---

void
FuelCell::SystemManagement::print_system_info() const
{
  deallog << std::endl;
  deallog << std::endl;
  deallog << "------------------------------";
  deallog << std::endl;
  deallog << std::endl;
  deallog << "Available FCST solution variables:" << std::endl;
  deallog << std::endl;

  for(unsigned int index = 0; index < all_solution_names.size(); ++index)
    deallog << all_solution_names[index] << std::endl;

  deallog << std::endl;
  deallog << "Available FCST equations:" << std::endl;
  deallog << std::endl;

  for(unsigned int index = 0; index < all_equation_names.size(); ++index)
    deallog << all_equation_names[index] << std::endl;

  deallog << std::endl;
  deallog << "User defined solution variables:" << std::endl;
  deallog << std::endl;

  for(unsigned int index = 0; index < solution_names.size(); ++index)
    deallog << solution_names[index] << std::endl;

  deallog << std::endl;
  deallog << "User defined equations:" << std::endl;
  deallog << std::endl;

  for(unsigned int index = 0; index < equation_names.size(); ++index)
    deallog << equation_names[index] << std::endl;

  deallog << std::endl;
  deallog << "Number of blocks: " << block_info->global.size();
  deallog << std::endl;

  deallog << std::endl;
  deallog << "Base element: " << std::endl;
  deallog << std::endl;

  for(unsigned int index = 0; index < block_info->base_element.size(); ++index)
    deallog << "No block: " << index << ", Name of block: " << solution_names[index] << ", No fe: " << block_info->base_element[index] << std::endl;

  deallog << std::endl;
  deallog << "Cell couplings table is represented by (line - equation, column - variable):" << std::endl;
  deallog << std::endl;

  for(unsigned int eq = 0; eq < equation_names.size(); ++eq)
  {
    for(unsigned int var = 0; var < solution_names.size(); ++var)
      deallog << (*cell_couplings)(eq,var) << "   ";

    deallog << std::endl;
  }

  deallog << std::endl;
  deallog << "Flux couplings table is represented by (line - equation, column - variable):" << std::endl;
  deallog << std::endl;

  for(unsigned int eq = 0; eq < equation_names.size(); ++eq)
  {
    for(unsigned int var = 0; var < solution_names.size(); ++var)
      deallog << (*flux_couplings)(eq,var) << "   ";

    deallog << std::endl;
  }

  deallog << std::endl;
  deallog << "FCST is using the mode: ";
  if(block_info->local_renumbering.size() == 0)
    deallog << "standard deal.II mode" << std::endl;
  else
    deallog << "block-wise renumbering mode" << std::endl;
  deallog << std::endl;
  deallog << std::endl;
  deallog << "------------------------------";
  deallog << std::endl;
}