#include <analytical_agglomerate_test.h>
using namespace FuelCellShop;

std::string parseDouble(double toParse){
	return boost::lexical_cast<std::string>(toParse);
}




void AnalyticalAgglomerateTest::testAV(){
	agg.setAV(1);
	agg.loadingWeigths.push_back(1); agg.loadingWeigths.push_back(2); agg.loadingWeigths.push_back(3);
	//agg.setUpLoadings();

	std::string msg = parseDouble(agg.getAV(0.5));

	TEST_ASSERT_MSG(agg.getAV(0.5) == 0.81818181818181812, msg.c_str());
}

void AnalyticalAgglomerateTest::testAVSetup(){
	agg.setAV(1);
	agg.loadingWeigths.push_back(1); agg.loadingWeigths.push_back(2); agg.loadingWeigths.push_back(3);
	agg.setUpLoadings();

	std::string msg = parseDouble(agg.actualLoadings[0]) + " " + parseDouble(agg.actualLoadings[1]) + " " + parseDouble(agg.actualLoadings[2]);

	TEST_ASSERT_MSG(((agg.actualLoadings[0] == 0.40909090909090906 ) && (agg.actualLoadings[1] ==0.81818181818181812) &&
			(agg.actualLoadings[2] == 1.2272727272727271)), msg.c_str() );
}


void AnalyticalAgglomerateTest::testSetSolution(){
    std::map<VariableNames,SolutionVariable> solutions;
    solutions[oxygen_molar_fraction] = SolutionVariable(0.5,2,oxygen_molar_fraction);

    //Check for setting solution without providing key variables (x_O2, phi_M, and phi_S)
    TEST_THROWS_MSG(agg.set_solution(solutions, oxygen_molar_fraction, 1), ExcMessage, "An exception for missing solution was expected.");


    solutions[protonic_electrical_potential] = SolutionVariable(0.6, 2, protonic_electrical_potential);
    solutions[electronic_electrical_potential] = SolutionVariable(0.6, 2, protonic_electrical_potential);

    std::vector<VariableNames> expected_sol_names;
    expected_sol_names.push_back(oxygen_molar_fraction);
    expected_sol_names.push_back(protonic_electrical_potential);
    expected_sol_names.push_back(electronic_electrical_potential);

    agg.set_solution(solutions,oxygen_molar_fraction, 1);

    //Check that the solution names are set up correctly
    TEST_ASSERT_MSG(agg.sol_names == expected_sol_names, "Solution names inside agglomerate are incorrect");

    //Check that the agglomerate wont accept a reactant it can't solve for.
    TEST_THROWS_MSG(agg.set_solution(solutions, oxygen_concentration, 1), ExcMessage, "Agglomerate should not except this reactant");

    //Check the solution data
    TEST_ASSERT_MSG(agg.solutions[protonic_electrical_potential][0] == 0.6, "Solution inside agglomerate is incorrect");
}


void AnalyticalAgglomerateTest::testO2CurrentDensity()
{
    ParameterHandler param;
    FuelCellShop::Layer::CatalystLayer<dim>::declare_CatalystLayer_parameters("CathodeCL", param);
    FuelCellShop::Material::PolymerElectrolyteBase::declare_PolymerElectrolyte_parameters(param);
    FuelCellShop::Material::CatalystSupportBase::declare_CatalystSupport_parameters(param);
    FuelCellShop::Material::CatalystBase::declare_Catalyst_parameters(param);
    OC.declare_parameters(param);

    param.enter_subsection("Fuel cell data");
    param.enter_subsection("CathodeCL");
    param.set("Catalyst layer type", "AgglomerateCL");
    param.enter_subsection("AgglomerateCL");
    param.set("Agglomerate composition", "ionomer");
    param.set("Agglomerate solver","analytical");
    param.leave_subsection();
    param.leave_subsection();
    param.leave_subsection();

    CCL = FuelCellShop::Layer::CatalystLayer<dim>::create_CatalystLayer("CathodeCL", param);

    OC.initialize(param);

    CCL->set_reaction_kinetics("ORR");
    CCL->set_constant_solution(OC.get_pc_Pa(), total_pressure);
    CCL->set_constant_solution(OC.get_T(), temperature_of_REV);

    std::vector<FuelCellShop::SolutionVariable> sols;
    sols.push_back(FuelCellShop::SolutionVariable(0.1,1,oxygen_molar_fraction));
    sols.push_back(FuelCellShop::SolutionVariable(0.625,1,electronic_electrical_potential));
    sols.push_back(FuelCellShop::SolutionVariable(-0.1,1,protonic_electrical_potential));
    sols.push_back(FuelCellShop::SolutionVariable(8,1,membrane_water_content));
    CCL->set_solution(sols);

    std::vector<double> current(1, 0.0);
    CCL->current_density(current);

    double expected = 154.6875;
    double match = expected*0.05;
    TEST_ASSERT_DELTA_MSG(expected, current[0], match, "Current from model ionomer analytical does not match expected results!");


}


void AnalyticalAgglomerateTest::test02CurrentDerivative(){
    //std::vector<std::vector<double> >Dcurrent(3, std::vector<double>(1,0.0));

    std::vector<VariableNames> sol_names;
    sol_names.push_back(oxygen_molar_fraction);
    sol_names.push_back(protonic_electrical_potential);
    sol_names.push_back(electronic_electrical_potential);
    std::map< VariableNames, std::vector<double> > derivatives;

    CCL->set_derivative_flags(sol_names);
    CCL->derivative_current_density(derivatives);
    double dCurrentdO2 = derivatives[oxygen_molar_fraction][0];
    double dCurrentdPhi_m = derivatives[protonic_electrical_potential][0];
    double dCurrentdPhi_s = derivatives[electronic_electrical_potential][0];

    double expDCurrentDO2 = 1546.875;     double diffDCurrentDO2    = expDCurrentDO2*0.05;
    double expDCurrentPhi_m = 4555.875;  double diffDCurrentPhi_m  = expDCurrentPhi_m*0.05;
    double expDCurrentPhi_s = -4555.875;  double diffDCurrentPhi_s  = expDCurrentPhi_s*-0.05; //Difference has to be positive value

    TEST_ASSERT_DELTA_MSG(dCurrentdO2, expDCurrentDO2, diffDCurrentDO2, "Current derivative w.r.t. O2 is incorrect");
    TEST_ASSERT_DELTA_MSG(dCurrentdPhi_m, expDCurrentPhi_m, diffDCurrentPhi_m, "Current derivative w.r.t. Phi_m is incorrect");
    TEST_ASSERT_DELTA_MSG(dCurrentdPhi_s, expDCurrentPhi_s, diffDCurrentPhi_s, "Current derivative w.r.t. Phi_s is incorrect");



}

